Package: cvasi
Type: Package
Title: Calibration, Validation, and Simulation of TKTD Models
Version: 1.5.1
Authors@R: c(
    person("Nils", "Kehrein", email="nils.kehrein@gmail.com", role=c("aut", "cre")),
    person("Dirk", "Nickisch", role="aut"),
    person("Peter", "Vermeiren", role="aut"),
    person("Torben", "Wittwer", role="ctb"),
    person("Johannes", "Witt", role="ctb"),
    person("Andre", "Gergs", role="ctb")
    )
Description: Eases the use of ecotoxicological effect models. Can simulate
    common toxicokinetic-toxicodynamic (TK/TD) models such as 
    General Unified Threshold models of Survival (GUTS) and Lemna. It can
    derive effects and effect profiles (EPx) from scenarios. It supports the
    use of 'tidyr' workflows employing the pipe symbol. Time-consuming
    tasks can be parallelized.
URL: https://github.com/cvasi-tktd/cvasi
BugReports: https://github.com/cvasi-tktd/cvasi/issues
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
Imports: cli, rlang, stringr, dplyr, tibble, purrr, furrr, tidyr,
        magrittr, utils, stats, methods, grid, gridExtra, ggplot2,
        GGally, deSolve, lubridate, units, lifecycle
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Collate: 'batch.R' 'cache.R' 'class-CalibrationSet.R'
        'class-ExposureSeries.R' 'class-EffectScenario.R' 'sequence.R'
        'calibrate.R' 'class-ParameterSet.R' 'class-Transferable.R'
        'data.R' 'dose_response.R' 'effect.R' 'epx.R' 'explore_space.R'
        'fx.R' 'solver.R' 'man-lemna.R' 'model-lemna_setac.R'
        'model-magma.R' 'model-lemna_schmitt.R' 'fit_growth.R'
        'fit_tktd.R' 'get.R' 'get_param.R' 'get_tag.R' 'get_times.R'
        'globals.R' 'has.R' 'import_morse.R' 'import_toxswa.R' 'is.R'
        'lik_profile.R' 'log.R' 'man-deb.R' 'man-macrophytes.R'
        'model-algae.R' 'model-deb_abj.R' 'model-debtox.R'
        'model-deb_daphnia.R' 'model-guts.R' 'model-guts_red.R'
        'numerics.R' 'package.R' 'pll.R' 'plot.R' 'plotting.r' 'pull.R'
        'set.R' 'set_bounds.R' 'set_exposure.R' 'set_forcings.R'
        'set_init.R' 'set_param.R' 'set_times.R' 'set_transfer.R'
        'set_window.R' 'show.R' 'simulate.R' 'survival.R' 'tox_data.R'
        'utils-pipe.R'
Suggests: future, knitr, lemna, rmarkdown, roxyglobals, testthat, withr
Depends: R (>= 3.5.0)
VignetteBuilder: knitr
Config/roxyglobals/filename: globals.R
Config/roxyglobals/unique: FALSE
NeedsCompilation: yes
Packaged: 2026-04-10 14:15:38 UTC; ENCAZ
Author: Nils Kehrein [aut, cre],
  Dirk Nickisch [aut],
  Peter Vermeiren [aut],
  Torben Wittwer [ctb],
  Johannes Witt [ctb],
  Andre Gergs [ctb]
Maintainer: Nils Kehrein <nils.kehrein@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-11 09:20:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-15 05:17:27 UTC; unix
Archs: cvasi.so.dSYM

Package: cvasi
Type: Package
Title: Calibration, Validation, and Simulation of TKTD Models
Version: 1.5.1
Authors@R: c(
    person("Nils", "Kehrein", email="nils.kehrein@gmail.com", role=c("aut", "cre")),
    person("Dirk", "Nickisch", role="aut"),
    person("Peter", "Vermeiren", role="aut"),
    person("Torben", "Wittwer", role="ctb"),
    person("Johannes", "Witt", role="ctb"),
    person("Andre", "Gergs", role="ctb")
    )
Description: Eases the use of ecotoxicological effect models. Can simulate
    common toxicokinetic-toxicodynamic (TK/TD) models such as 
    General Unified Threshold models of Survival (GUTS) and Lemna. It can
    derive effects and effect profiles (EPx) from scenarios. It supports the
    use of 'tidyr' workflows employing the pipe symbol. Time-consuming
    tasks can be parallelized.
URL: https://github.com/cvasi-tktd/cvasi
BugReports: https://github.com/cvasi-tktd/cvasi/issues
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
Imports: cli, rlang, stringr, dplyr, tibble, purrr, furrr, tidyr,
        magrittr, utils, stats, methods, grid, gridExtra, ggplot2,
        GGally, deSolve, lubridate, units, lifecycle
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Collate: 'batch.R' 'cache.R' 'class-CalibrationSet.R'
        'class-ExposureSeries.R' 'class-EffectScenario.R' 'sequence.R'
        'calibrate.R' 'class-ParameterSet.R' 'class-Transferable.R'
        'data.R' 'dose_response.R' 'effect.R' 'epx.R' 'explore_space.R'
        'fx.R' 'solver.R' 'man-lemna.R' 'model-lemna_setac.R'
        'model-magma.R' 'model-lemna_schmitt.R' 'fit_growth.R'
        'fit_tktd.R' 'get.R' 'get_param.R' 'get_tag.R' 'get_times.R'
        'globals.R' 'has.R' 'import_morse.R' 'import_toxswa.R' 'is.R'
        'lik_profile.R' 'log.R' 'man-deb.R' 'man-macrophytes.R'
        'model-algae.R' 'model-deb_abj.R' 'model-debtox.R'
        'model-deb_daphnia.R' 'model-guts.R' 'model-guts_red.R'
        'numerics.R' 'package.R' 'pll.R' 'plot.R' 'plotting.r' 'pull.R'
        'set.R' 'set_bounds.R' 'set_exposure.R' 'set_forcings.R'
        'set_init.R' 'set_param.R' 'set_times.R' 'set_transfer.R'
        'set_window.R' 'show.R' 'simulate.R' 'survival.R' 'tox_data.R'
        'utils-pipe.R'
Suggests: future, knitr, lemna, rmarkdown, roxyglobals, testthat, withr
Depends: R (>= 3.5.0)
VignetteBuilder: knitr
Config/roxyglobals/filename: globals.R
Config/roxyglobals/unique: FALSE
NeedsCompilation: yes
Packaged: 2026-04-10 14:15:38 UTC; ENCAZ
Author: Nils Kehrein [aut, cre],
  Dirk Nickisch [aut],
  Peter Vermeiren [aut],
  Torben Wittwer [ctb],
  Johannes Witt [ctb],
  Andre Gergs [ctb]
Maintainer: Nils Kehrein <nils.kehrein@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-11 09:20:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 13:39:00 UTC; unix
Archs: cvasi.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(diagnostics,cvasi_simulate)
S3method(diagnostics,default)
S3method(num_info,cvasi_fit)
S3method(num_info,cvasi_simulate)
S3method(num_info,default)
S3method(plot,cvasi_drc)
S3method(plot,cvasi_simulate)
S3method(plot,lik_profile)
S3method(plot,param_space)
export("%>%")
export(Algae_Simple)
export(Algae_TKTD)
export(Algae_Weber)
export(CalibrationSet)
export(DEB_Daphnia)
export(DEB_abj)
export(DEBtox)
export(ExposureSeries)
export(GUTS_RED_IT)
export(GUTS_RED_SD)
export(GUTS_SD)
export(Lemna_SETAC)
export(Lemna_Schmitt)
export(Lemna_SchmittThold)
export(Magma)
export(Myrio)
export(Myrio_log)
export(SamX)
export(batch)
export(cache_controls)
export(calibrate)
export(caliset)
export(ci_from_hessian)
export(diagnostics)
export(dose_response)
export(effect)
export(epx)
export(epx_mtw)
export(explore_space)
export(fit_growth)
export(fit_tktd)
export(fx)
export(get_model)
export(get_model_name)
export(get_param)
export(get_tag)
export(get_times)
export(import_morse)
export(import_swash)
export(import_toxswa)
export(is_DEB)
export(is_GUTS)
export(is_GUTS_IT)
export(is_GUTS_SD)
export(is_Lemna)
export(is_LemnaThreshold)
export(is_deb)
export(is_guts)
export(is_guts_it)
export(is_guts_sd)
export(is_lemna)
export(is_scenario)
export(lik_profile)
export(log_disable)
export(log_enable)
export(log_envir)
export(log_lik)
export(log_msg)
export(log_scenarios)
export(morse)
export(no_exposure)
export(num_info)
export(parameter_set)
export(pll_debug)
export(plot_epx)
export(plot_lik_profile)
export(plot_param_space)
export(plot_ppc)
export(plot_scenario)
export(plot_sd)
export(pull_metadata)
export(sequence)
export(set_bounds)
export(set_endpoints)
export(set_exposure)
export(set_forcings)
export(set_init)
export(set_moa)
export(set_mode_of_action)
export(set_noexposure)
export(set_notransfer)
export(set_nowindow)
export(set_param)
export(set_tag)
export(set_times)
export(set_transfer)
export(set_window)
export(simulate)
export(simulate_batch)
export(solver)
export(survival)
export(td2cs)
export(tox_data)
exportClasses(Algae)
exportClasses(AlgaeSimple)
exportClasses(AlgaeSimpleScenario)
exportClasses(AlgaeTKTD)
exportClasses(AlgaeTKTDScenario)
exportClasses(AlgaeWeber)
exportClasses(AlgaeWeberScenario)
exportClasses(CalibrationSet)
exportClasses(Deb)
exportClasses(DebAbj)
exportClasses(DebDaphnia)
exportClasses(DebTox)
exportClasses(EffectScenario)
exportClasses(ExposureSeries)
exportClasses(Guts)
exportClasses(GutsIt)
exportClasses(GutsRedIt)
exportClasses(GutsRedSd)
exportClasses(GutsSd)
exportClasses(Lemna)
exportClasses(LemnaSchmitt)
exportClasses(LemnaSchmittScenario)
exportClasses(LemnaSetac)
exportClasses(LemnaSetacScenario)
exportClasses(Magma)
exportClasses(NoExposureSeries)
exportClasses(ParameterSet)
exportClasses(ScenarioSequence)
exportClasses(Transferable)
exportClasses(parameter_set)
exportMethods("[")
exportMethods("[[")
exportMethods("[[<-")
exportMethods(effect)
exportMethods(get_model)
exportMethods(get_param)
exportMethods(get_tag)
exportMethods(get_times)
exportMethods(length)
exportMethods(set_bounds)
exportMethods(set_param)
exportMethods(set_times)
importFrom(deSolve,diagnostics)
importFrom(deSolve,diagnostics.deSolve)
importFrom(deSolve,diagnostics.default)
importFrom(deSolve,ode)
importFrom(lifecycle,deprecated)
importFrom(lifecycle,is_present)
importFrom(magrittr,"%>%")
importFrom(methods,"slot<-")
importFrom(methods,.hasSlot)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,selectMethod)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(rlang,abort)
importFrom(rlang,enquos)
importFrom(rlang,eval_tidy)
importFrom(rlang,inform)
importFrom(rlang,warn)
importFrom(stats,optim)
importFrom(stats,setNames)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,installed.packages)
importFrom(utils,sessionInfo)
importFrom(utils,tail)
useDynLib(cvasi, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# cvasi: Calibration, Validation, and Simulation of TKTD models in R

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/cvasi)](https://cran.r-project.org/package=cvasi)
[![R-CMD-check](https://github.com/cvasi-tktd/cvasi/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/cvasi-tktd/cvasi/actions/workflows/R-CMD-check.yaml)
<!--[![Codecov test coverage](https://codecov.io/gh/xy/cvasi/branch/main/graph/badge.svg)](https://app.codecov.io/gh/xy/cvasi?branch=main)-->
<!-- badges: end -->

The `cvasi` package aims to ease the use of ecotox effect models by
providing an intuitive workflow. Model inputs and parameters are
encapsulated in scenario objects which can be piped to other functions.
Operations can be chained using the `tidyr` syntax. The most
time-consuming processes can be run in parallel if requested.

The package provides facilities to

- Simulate effect models such as *GUTS*, *GUTS-RED*, *DEB*, *Lemna*
  based on Schmitt et al. (2013), and algae models based on Weber et
  al. (2012),
- Calculate effect endpoints,
- Derive effect profiles (*EPx* values),
- Import exposure time-series from *FOCUS TOXSWA*,
- Import fitted parameters from *morse*,
- And much more.

A graphical user interface implemented in
[Shiny](https://posit.co/products/open-source/rstudio/) is also
available, see the *[cvasi.ui](https://github.com/cvasi-tktd/cvasi.ui/)*
package. Please have a look at the [Changelog](NEWS.md) for an overview
of user-facing updates and changes.

## Installation

Install the package from CRAN:

``` r
install.packages("cvasi", dependencies=TRUE)
```

Or install the newest development version from *GitHub*:

``` r
install.packages("remotes", dependencies=TRUE)
remotes::install_github("cvasi-tktd/cvasi", dependencies=TRUE)
```

For installing `cvasi` from *GitHub* on Windows computers, please make
sure that you also have
[*Rtools*](https://cran.r-project.org/bin/windows/Rtools/) installed on
your machine. *Rtools* are required to compile the package’s C code.

## Documentation

The package contains the following vignettes

- [User
  Manual](https://github.com/cvasi-tktd/cvasi/blob/main/doc/cvasi-1-manual.md)
- [Modeling
  Howto](https://github.com/cvasi-tktd/cvasi/blob/main/doc/cvasi-2-howto.md)

They can also be accessed locally by executing an *R* statement such as:

``` r
vignette("cvasi-1-manual", package="cvasi")
```

## Usage

Basic usage:

``` r
library(cvasi)

# create and parameterize a GUTS-RED-IT scenario
scenario <- GUTS_RED_IT() %>%
  set_param(c(kd=0.0005, hb=0, alpha=0.4, beta=1.5)) %>%
  set_exposure(data.frame(time=c(0, 100, 101, 200, 201, 400),
                          conc=c(0, 0, 0.1, 0.1, 0, 0))) %>%
  set_times(1:400)

# simulate scenario
results <- scenario %>% simulate()
tail(results)
#>     time           D H        S
#> 395  395 0.004429420 0 0.998654
#> 396  396 0.004427206 0 0.998654
#> 397  397 0.004424993 0 0.998654
#> 398  398 0.004422781 0 0.998654
#> 399  399 0.004420570 0 0.998654
#> 400  400 0.004418360 0 0.998654

# ... and plot simulation results
plot(results)
```

<img src="doc/figures/readme-unnamed-chunk-5-1.png" width="70%" style="display: block; margin: auto;" />

Calculation of effects:

``` r
# calculate effect level
scenario %>% effect()
#> # A tibble: 1 × 4
#>   scenario         L L.dat.start L.dat.end
#>   <list>       <dbl>       <dbl>     <dbl>
#> 1 <GutsRdIt> 0.00135           1       400

# create a dose-response curve
drc <- scenario %>% dose_response()
head(drc)
#>   endpoint        mf      effect
#> 1        L  3.812500 0.009914431
#> 2        L  4.799653 0.013956165
#> 3        L  6.042405 0.019598667
#> 4        L  7.606938 0.027460989
#> 5        L  9.576567 0.038353465
#> 6        L 12.056184 0.053333780

# plot the dose-response curve
plot(drc)
```

<img src="doc/figures/readme-unnamed-chunk-6-1.png" width="60%" style="display: block; margin: auto;" />

``` r

# derive EPx values
scenario %>% epx()
#> # A tibble: 1 × 3
#>   scenario   L.EP10 L.EP50
#>   <list>      <dbl>  <dbl>
#> 1 <GutsRdIt>   19.0   82.1
```

Multiple scenarios can be processed in parallel without modifications to
the workflow:

``` r
# enable parallel processing
future::plan(future::multisession)

# derive EPx for a list of 100 scenarios in parallel
rep(c(scenario), 100) %>% epx()

# disable parallel processing
future::plan(future::sequential)
```

## License

The package and its source code is free and open-source software
available under the [GPL-3.0
license](https://github.com/cvasi-tktd/cvasi/blob/main/LICENSE.md).

## Issues

If you find any issues or bugs within the package, please create a [new
issue](https://github.com/cvasi-tktd/cvasi/issues) on GitHub.

## Contributing

Contributions to the project are welcome! Please have a look at the
[Contribution
Guidelines](https://github.com/cvasi-tktd/cvasi/blob/main/CONTRIBUTING.md)
before submitting a Pull Request.

## Acknowledgements

Financial support for creation and release of this software project was
provided by Bayer Crop Science. This R package started as an internal
project at Bayer Crop Science and the project owners would like to thank
the people who have contributed (in no particular order):

Nils Kehrein, Johannes Witt, André Gergs, Thomas Preuss, Julian
Heinrich, Zhenglei Gao, Tjalling Jager, Dirk Nickisch, Torben Wittwer,
and Peter Vermeiren.

This folder contains exposure time series of the fungicide propiconazole which
were predicted using FOCUS surface water models. The data has been published
as part of the supporting information (Appendix E) by EFSA (2018).

Original filenames:
Propiconazole apple simulations.xlsx
Propiconazole spring cereals simulations.xlsx

Literature:
EFSA Scientific Opinion on TKTD models, 2018. doi:10.2903/j.efsa.2018.5377

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  fig.path = "../doc/figures/howto-",
  comment = "#>"
)

## ----setup, include=FALSE-----------------------------------------------------
library(cvasi)
library(dplyr)
#options(conflicts.policy=FALSE,warn.conflicts=FALSE)

## -----------------------------------------------------------------------------
# Create a new scenario object
myscenario <- Lemna_Schmitt()

# Get model name
get_model(myscenario)

# Set a custom tag to identify the scenario
myscenario %>%
  set_tag("Lab experiment #1") -> myscenario

# Get custom tag
get_tag(myscenario)

# The tag is also displayed when printing scenario properties
myscenario

# Accessing scenario slots and their default values
myscenario@forcings.req # forcings required for effect calculations
myscenario@endpoints    # available effect endpoints
myscenario@control.req  # are control runs required for effect calculation?

## ----eval=FALSE---------------------------------------------------------------
# # Call the help page of effect scenarios class
# ?scenarios

## ----eval=FALSE---------------------------------------------------------------
# # Call the help page of the biomass transfer class
# ?Transferable

## -----------------------------------------------------------------------------
# The example scenario `metsulfuron` based on the Lemna model by Schmitt et al. (2013)
# is modified by setting a new exposure series and initial state. Then, it is
# simulated.
  metsulfuron %>%
    set_noexposure() %>%  # set no exposure (i.e., a control run)
    set_init(c(BM = 50)) %>%  # set initial biomass
    simulate()

## -----------------------------------------------------------------------------
# Initialize the random numbers generator
set.seed(123)
# Generating a random exposure series spanning 14 days
random_conc <- runif(15, min=0, max=0.1)
exposure_profile <- data.frame(time=0:14, conc=random_conc)
# Run EPx calculations
minnow_it %>%
  set_exposure(exposure_profile) %>%  # set a specific exposure scenario
  epx()  # run EPx calculations

## -----------------------------------------------------------------------------
# Derive effect levels of all exposure windows
metsulfuron %>% 
  set_window(length=7, interval=1) %>%
  effect(max_only=FALSE)

## -----------------------------------------------------------------------------
# Only report effect levels larger than 1e-5 = 0.001%
metsulfuron %>% 
  set_window(length=7, interval=1) %>%
  effect(max_only=FALSE, marginal_effect=1e-5)

## -----------------------------------------------------------------------------
set.seed(123)
# Generate a random exposure series spanning 14 days
ts <- data.frame(time = 0:14,
                 conc = runif(15, min=0, max=0.1))

# Run EPx calculations for a window length of 7 days and a step size of 1 day
metsulfuron %>%
  set_exposure(ts) %>%
  set_window(length=7, interval=1) %>%
  effect(max_only=FALSE) -> results
results

# Create a plot of effects over time
library(ggplot2)
ggplot(results) +
  geom_point(aes(dat.start, BM*100)) +
  labs(x="Start of window (day)", y="Effect on biomass (%)")

## -----------------------------------------------------------------------------
metsulfuron %>%
  set_init(c(BM=1)) %>%
  set_noexposure() %>%
  set_transfer(interval=3, biomass=1) %>%
  simulate() -> result
result

library(ggplot2)
ggplot(result) +
  geom_line(aes(time, BM)) +
  labs(x="Time (days)", y="Biomass (g_dw/m2)", title="Biomass transfer every three days")


## -----------------------------------------------------------------------------
metsulfuron %>%
  set_init(c(BM=1)) %>%
  set_noexposure() %>%
  set_transfer(times=c(3, 6), biomass=c(1, 0.5)) %>%
  simulate() -> result2

library(ggplot2)
ggplot(result2) +
  geom_line(aes(time, BM)) +
  labs(x="Time (days)", y="Biomass (g_dw/m2)", title="Biomass transfer at custom time points")

## ----eval=FALSE---------------------------------------------------------------
# # Call the help page of set_transfer
# ?set_transfer

## ----warning=FALSE------------------------------------------------------------
library(dplyr)
# set k_phot_fix, k_resp and Emax to the defaults recommended in Klein et al. (2022)
# for Tier 2C studies. Set initial biomass to 12.0 (original data is in fronds
# rather than biomass, but for sake of simplicity, no conversion was performed).
control <- metsulfuron %>% 
  set_param(c(k_phot_fix=TRUE, k_resp=0, Emax=1)) %>% 
  set_init(c(BM=12)) %>%
  set_noexposure() %>%
  set_bounds(list(k_phot_max=c(0, 0.5)))
# `metsulfuron` is an example scenario based on Schmitt et al. (2013) and therefore
# uses the historical, superseded nomenclature by Schmitt et al. We recommend using
# the newer SETAC Lemna model by Klein et al. (2022) for current applications,
# see `Lemna_SETAC()` for details.

# Get control data from Schmitt et al. (2013)
obs_control <- schmitt2013 %>%
  filter(trial == "T0")

# Fit parameter `k_phot_max` to observed data: `k_phot_max` is a physiological
# parameter which is typically fitted to the control data
fit1 <- calibrate(
  x = control,
  par = c(k_phot_max = 1),
  data = obs_control,
  output = "BM",
  method="Brent" # Brent is recommended for one-dimensional optimization
)
fit1$par

# Update the scenario with fitted parameter and simulate it
fitted_growth <- control %>% 
  set_param(fit1$par) %>%
  set_bounds(list(
    EC50=c(0, 1000),
    b=c(0.1, 10), 
    P_up=c(0, 0.1)
  ))
sim_mean <- fitted_growth %>%
  simulate() %>%
  mutate(trial="control")

# Exposure and observations in long format for plotting
treatments <- obs_control %>%
  select(time, conc) %>%
  mutate(trial="control")
obs_mean <- obs_control %>%
  select(time, BM=obs) %>%
  mutate(trial="control")

# Plot results
plot_sd(
  model_base = fitted_growth,
  treatments = treatments,
  rs_mean = sim_mean,
  obs_mean = obs_mean
)

## ----warning=FALSE------------------------------------------------------------
# Use all trials and exposure series from Schmitt et al. (2013) as input for fitting
fit2 <- calibrate(
  fitted_growth,
  data=schmitt2013,
  par=c(EC50=0.3, b=4.16, P_up=0.005),
  output="BM",
  method="L-BFGS-B",
  verbose=FALSE
)
fit2$par

# Update the scenario with fitted parameter and simulate all trials
fitted_tktd <- fitted_growth %>%
  set_param(fit2$par)

treatments <- schmitt2013 %>% select(time, conc, trial)
rs_mean <- fitted_tktd %>%
  batch(treatments) %>%
  simulate()

# Observations in long format for plotting
obs_mean <- schmitt2013 %>% select(time, obs, trial)

# Plot results
plot_sd(
  model_base = fitted_tktd,
  treatments = treatments,
  rs_mean = rs_mean,
  obs_mean = obs_mean
)

## -----------------------------------------------------------------------------
# using the calibration set and calibrated parameters from the previous Lemna 
# Schmitt example, a likelihood profiling is done
# We set parameter boundaries to constrain the likelihood profiling within these 
# boundaries (this is optional)

# Calibrate a model to a list of calibration sets
fit <- calibrate(
  fitted_growth,
  data=schmitt2013,
  par=c(EC50=0.3, b=4.16, P_up=0.005),
  output="BM",
  method="L-BFGS-B",
  verbose=FALSE
)

# Conduct profiling
res <- lik_profile(x = fitted_growth,
                   data = schmitt2013,
                   par = fit$par, # the parameter values after calibration
                   output = "BM", # the observational output against which to 
                                  # compare the model fit
                   bounds = list(EC50_int = list(0.1, 4), 
                                 b = list(1, 5),
                                 P = list(0.0001, 0.2)))
# Visualise profiling results
plot(res)

# Access 95% confidence intervals of profiled parameters
res$EC50$confidence_interval
res$b$confidence_interval


## -----------------------------------------------------------------------------
# Create calisets
cs <- td2cs(fitted_growth, tox_data(schmitt2013))
# Define tags and weights
lab_tags <- c("A", "B", "B", "A", "B", "B", "B")
lab_weights <- c(1, 1.5, 1.5, 1, 1.5, 1.5, 1.5)

for(i in 1:length(cs)){
  cs[[i]]@tag <- lab_tags[i]
  cs[[i]]@weight <- lab_weights[i]
}

res_2 <- lik_profile(x = cs, # the calibration set
                   par = fit$par, # the parameter values after calibration
                   output = "BM", # the observational output against which to 
                   # compare the model fit
                   bounds = list(EC50_int = list(0.1, 4), 
                                 b = list(1, 5),
                                 P = list(0.0001, 0.2)),
                   individual = TRUE)
                   
plot(res_2)



## ----eval=FALSE---------------------------------------------------------------
# # Call the help page for more information about the parameter space explorer
# ?explore_space
# 

## -----------------------------------------------------------------------------

# Calibrate a model to a list of calibration sets
cs <- td2cs(fitted_growth, tox_data(schmitt2013))
fit <- calibrate(
  cs,
  par=c(EC50=0.3, b=4.16, P_up=0.005),
  output="BM",
  method="L-BFGS-B",
  verbose=FALSE
)
# conduct profiling
res <- lik_profile(x = cs, # the calibration set
                   par = fit$par, # the parameter values after calibration
                   output = "BM", # the observational output against which to 
                                  # compare the model fit
                   bounds = list(EC50_int = list(0.1, 4), 
                                 b = list(1, 5),
                                 P = list(0.0001, 0.2)))
# Conduct space exploration
res_space <- explore_space(
  x = cs,
  par = fit2$par,
  res = res, # output of the likelihood profiling function
  output = "BM")

# Visualize the parameter space
plot(res_space)


## -----------------------------------------------------------------------------
# A base scenario is created for the whole period, but parameters are only
# valid until day 7
sc1 <- metsulfuron %>%
  set_times(0:14)

# A parameter change occurs at day 7: k_loss increases from 0 to 0.1 d-1
sc2 <- sc1 %>%
  set_param(c(k_loss=0.1))

# Combine both scenarios to a sequence with a break at *t = 7*:
# scenario `sc1` will be simulated for *t = [0, 7]*, and
# scenario `sc2` will be simulated for *t = [7, 14]*.
sq <- sequence(list(sc1, sc2), breaks=7)
simulate(sq)

## ----eval=FALSE---------------------------------------------------------------
# # Call the help page of `sequence`
# ?sequence

## -----------------------------------------------------------------------------
# Simulations with a maximum solver step length of hmax=0.01
metsulfuron %>%
  set_times(0:7) %>%
  simulate(hmax=0.1)

## ----eval=FALSE---------------------------------------------------------------
# future::plan(future::multisession)

## ----eval=FALSE---------------------------------------------------------------
# vignette("future-1-overview", package="future")

## ----warning=FALSE------------------------------------------------------------
## An exemplary implementation of the GUTS-RED-SD TKTD model ##
# Model ODEs following the deSolve specification
sd_ode <- function(t, state, params) {
  with(as.list(c(state, params)), {
    dDw <- kd * (Cw(t) - Dw)            # dDw/dt, scaled damage
    dH <- kk * max(0, Dw - z)           # dH/dt, cumulative hazard
    
    list(c(dDw, dH))                    # return derivatives
  })
}

## ----eval=FALSE---------------------------------------------------------------
# vignette("deSolve", package="deSolve")

## ----warning=FALSE------------------------------------------------------------
## Properties of a sample scenario ##
init <- c(Dw=0, H=0)                         # initial state
times <- 0:5                                 # output time points [0,5]
param <- c(kd=22, hb=0.01, z=0.5, kk=0.08)   # model parameters
exp <- data.frame(time=0:5,                  # exposure time-series
                  conc=c(0,1,1,0.5,0.6,0.2))

# Create a linear interpolation of the exposure time-series
expf <- approxfun(x=exp$time, y=exp$conc, method="linear", rule=2)
# Extend parameter set by interpolated exposure series
paramx <- as.list(c(param, Cw=expf))

# Numerically solve the ODE
deSolve::ode(y=init, times=times, parms=paramx, func=sd_ode)

## -----------------------------------------------------------------------------
## Integrate a new model class into the package workflow ##
# Create a unique class that derives from 'EffectScenario'
setClass("MyGuts", contains="EffectScenario")

# Create an object of the new class and assign scenario properties
new("MyGuts", name="My custom model") %>%
  set_init(init) %>%
  set_times(times) %>%
  set_param(param) %>%
  set_exposure(exp, reset_times=FALSE) %>%
  set_endpoints("L") -> myscenario

myscenario

## -----------------------------------------------------------------------------
# the actual function calling deSolve can have a different signature
solver_myguts <- function(scenario, ...) {
  # get relevant data from scenario
  init <- scenario@init
  param <- scenario@param
  times <- scenario@times
  exp <- scenario@exposure@series
  if(nrow(exp) == 1) { # extend exposure series to have at least two rows
    row2 <- exp[1,]
    row2[[1]] <- row2[[1]]+1
    exp <- rbind(exp, row2)
  }
  
  # Create a linear interpolation of the exposure time-series
  expf <- approxfun(x=exp[,1], y=exp[,2], method="linear", rule=2)
  # Extend parameter set by interpolated exposure series
  paramx <- as.list(c(param, Cw=expf))
  
  # pass everything to ODE solver for numerical integration
  as.data.frame(deSolve::ode(y=init, times=times, parms=paramx, func=sd_ode, ...))
}

## Overload the solver() function ##
# The functions signature, i.e. the number and names of its arguments, must stay as is
setMethod("solver", "MyGuts", function(scenario, ...) solver_myguts(scenario, ...))

## -----------------------------------------------------------------------------
myscenario %>% simulate()

## -----------------------------------------------------------------------------
## Overload effect endpoint calculation ##
# fx() is called by effect()
setMethod("fx", "MyGuts", function(scenario, ...) fx_myguts(scenario, ...))

# @param scenario Scenario object to assess
# @param ... any additional parameters
fx_myguts <- function(scenario, ...) {
  # simulate the scenario (it is already clipped to the moving exposure window)
  out <- simulate(scenario, ...)
  # only use model state at the end of the simulation
  end <- tail(out, 1)
  # length of the simulated period, first column of result contains time
  t <- tail(out[, 1], 1)- out[[1, 1]]
  # calculate survival according to EFSA Scientific Opinion on TKTD models
  # p. 33, doi:10.2903/j.efsa.2018.5377
  survival <- exp(-end$H) * exp(-scenario@param$hb * t)
  return(c("L"=survival))
}

# Derive effect levels for our sample scenario
myscenario %>% effect()

## -----------------------------------------------------------------------------
## Get all model parameters
# Lemna model parameters taken from file 'mm2.r' included
# in supporting material of Schmitt et al. (2013)
param_study <- list(
  #     - Effect -
  Emax     = 0.784,   # [same as conc. data]  maximum effect concentration
  EC50     = 0.3,     # [same as conc. data]  Midpoint of effect curve
  b        = 4.16,    # [-]  Slope of effect curve
  #     - Toxicokinetics -
  P_up     = 0.0054,  # [cm/d]  Permeability for uptake
  AperBM   = 1000,    # [cm^2/g_dw]  Frond area/dry weight
  Kbm      = 0.75,    # []  Biomass(fw)-water partition coefficient
  P_Temp   = F,       # Switch for temperature dependence of cuticle permeability
  MolWeight = 381,    # [g/mol] Molmass of molecule (determines Q10_permeability)
  #     - Fate of biomass -
  k_phot_fix  = F,    # If True k_G_max is not changed by environmental factors
  k_phot_max  = 0.47, # [1/d]  Maximum photosynthesis rate
  k_resp   = 0.05,    # [1/d]  Respiration rate at ref. temperature
  k_loss   = 0.0,     # [1/d]  Some rate of loss (e.g. Flow rate)
  #     - Temperature dependence -
  Tmin     = 8.0  ,   # [°C]  Minimum growth temperature 
  Tmax     = 40.5 ,   # [°C]  Maximum growth temperature
  Topt     = 26.7 ,   # [°C]  Optimum growth temperature 
  t_ref    = 25,      # [°C]  Reference temperature for respiration rate
  Q10      = 2,       # [-]   Temperature dependent factor for respiration rate
  #     - Light dependence (Linear) -
  k_0      = 3 ,      # [1/d]  Intercept of linear part
  a_k      = 5E-5 ,   # [(1/d)/(kJ/m^2/d)]   Slope of linear part
  #     - Phosphorus dependence (Hill like dependence) -
  C_P      = 0.3,     # [mg/L]  Phosporus concentration in water
  CP50     = 0.0043,  # [mg/L]  P-conc. where growth rate is half
  a_P      = 1,       # []  Hill coefficient
  KiP      = 101,     # [mg/L]  P-inhibition constant for very high P-conc.
  #     - Nitrogen dependence (Hill like dependence) -
  C_N      = 0.6,     # [mg/L]  Nitrogen concentration in water
  CN50     = 0.034,   # [mg/L]  N-conc. where growth rate is half
  a_N      = 1,       # []  Hill coefficient
  KiN      = 604,     # [mg/L]  n-inhibition constant for very high P-conc.
  #     - Density dependence -
  BM50     = 176,     # [g_dw/m^2] Cut off BM
  #     - Others -
  mass_per_frond = 0.0001,  # [g_dw/frond]  Dry weight per frond
  BMw2BMd  = 16.7     # [g_fresh/g_dry]  Fresh- / dryweight 
)


## Define the forcing variables
forc_temp <- data.frame(t = 0, tmp = 12) # [°C]  Current temperature (may also be a table)
forc_rad  <- data.frame(t = 0, rad = 15000) # [kJ/m²/d]  Radiation  (may also be given as table)


## Define a simple exposure pattern
# t 0..6  concentration 1 ug/L
# t 7..14 concentration 0 ug/L
exposure <- data.frame(time = 0:14, 
                       conc = c(rep(1, 7), rep(0, 8))
                       )


## Set initial values 
# given in file 'mmc2.r' of Schmitt et al. (2013)
init <- c(
  BM       = 50,     # [g_dw/m^2]  Dry Biomass dry weight per m2
  E        = 1,      # (0-1)  (Toxic) Effect = Factor on growth rate  (Range: 0 - 1, 1=no effect)
  M_int    = 0       # [ug]   Amount of toxicant in biomass
)

## create a scenario object, containing the model (with parameters) and the exposure time-series
Lemna_Schmitt() %>%               # the Lemna model by Schmitt et al. (2013)
  set_tag("metsulfuron") %>%      # set a tage for the specific implementation of the model
  set_init(init) %>%              # set the starting values (as prepared above)
  set_param(param_study) %>%      # set the parameters (as prepared above)
  set_exposure(exposure) %>%      # set the exposure scenario (exposure time series)
  set_forcings(temp=forc_temp, rad=forc_rad) -> metsulfuron # set the external forcing 
# variables, and save everything as an object



## ----out.width = "100%"-------------------------------------------------------
## simulate with model, under a range of different exposure scenarios
# create several exposure scenarios
exp_scen <- schmitt2013 %>% select(time, conc, trial)

# simulate for all these scenarios
results <- metsulfuron %>%
  batch(exp_scen) %>%
  simulate()

# plot results
plot_sd(
  model_base = metsulfuron,
  treatments = exp_scen,
  rs_mean = results,
)


## simulate with model, under a range of different exposure scenarios, and including 
## a biomass transfer
# simulate for all scenarios
results <- metsulfuron %>%
  set_transfer(interval = c(5), biomass = 10) %>% # implement a biomass transfer every 5 days
  batch(exp_scen) %>%
  simulate()

# plot results
plot_sd(
  model_base = metsulfuron,
  treatments = exp_scen,
  rs_mean = results,
)

#' Set exposure time-series
#'
#' *Exposure* refers to the toxicant concentration an organism is exposed to.
#' In case of aquatic organisms, this would commonly be the concentration of a
#' toxicant in water. Other interpretations are possible depending on model
#' assumptions.
#'
#' Exposure time-series are generally represented by a `data.frame` containing two
#' columns. The first column for time, the second representing the
#' exposure level. The ordering of columns is mandatory. The column names
#' are non-relevant but sensible names may help documenting the
#' scenario and its data. The `data.frame`'s rows must be ordered chronologically.
#' A time-series can consist of only a single row; in this case it will represent constant
#' exposure.
#'
#' For convenience, a time-series with zero exposure can be set using
#' [set_noexposure()].
#'
#' ### Computational efficiency
#' Handling time-series is a costly task for the ODE solver due to consistency
#' checks and interpolation between time steps. How the solver interpolates
#' the time-series can be controlled by optional arguments to functions
#' such as [simulate()] and [effect()]. Please refer to [simulate()] for a brief
#' overview and [deSolve::forcings] for a detailed description.
#'
#' Exposure time-series should be kept as short as possible and as complex as
#' needed for optimal computational efficiency.
#'
#' ### Output times
#' By default, the exposure time-series' time points will also be used as output
#' times of the scenario. Any output times previously set by [set_times()] will
#' be lost. If this behavior is undesired, set the function argument
#' `reset_times=FALSE`.
#'
#' ### Multiple exposure series and scenarios
#' The functions supports modifying multiple scenarios at once: by
#' calling it with lists of [scenario] and [ExposureSeries]
#' objects. The cartesian product of all scenarios and exposure series will
#' be returned, iff the parameter `expand = TRUE` is set.
#'
#' As an example for the *expand* mode, two scenarios `A` and `B` and one
#' exposure series `g` will result in two scenarios `Ag` and `Bg`, both
#' using exposure series `g`. Two scenarios `A` and `B` as well as two
#' exposure seres `g` and `h` will result in four scenarios `Ag`,`Ah`,`Bg`,
#' and `Bh`.
#'
#' @param scenarios `vector` of [scenarios]
#' @param series `vector` of  [ExposureSeries] objects or a single `data.frame`
#' @param reset_times `logical`, if `TRUE`, the exposure time-series' time points
#'  will be set as output times. Defaults to `TRUE`
#' @param ... additional arguments
#'
#' @return `list` of `EffectScenario` objects
#' @include class-EffectScenario.R class-ExposureSeries.R
#' @export
#' @examples
#' # Set a data.frame as exposure series
#' df <- data.frame(time=c(0, 1, 2, 3), conc=c(1, 1, 0, 0))
#' Lemna_Schmitt() %>% set_exposure(df)
#'
#' # Create and set an ExposureSeries object
#' es1 <- ExposureSeries(df)
#' Lemna_Schmitt() %>% set_exposure(es1)
#'
#' # By default, the time points of the exposure series will also be used as
#' # as output times. To avoid overriding existing output times, set reset_times=FALSE
#' Lemna_Schmitt() %>%
#'   set_times(0:10) %>%
#'   set_exposure(es1, reset_times=FALSE)
#'
#' # Setting two series with one function call, creates two scenarios
#' es2 <- ExposureSeries(data.frame(time=5:10, conc=1))
#' Lemna_Schmitt() %>% set_exposure(c(es1, es2))
setGeneric("set_exposure",
           function(scenarios, series, ...) standardGeneric("set_exposure"),
           signature = c("scenarios", "series")
)

#' @rdname set_exposure
setMethod("set_exposure", c("ANY","ANY"), function(scenarios, series, ...) {
  stop("argument types not supported")
})

# single data.frame object, i.e. a time-series
#' @noRd
set_exposure_dfr <- function(scenarios, series, reset_times=TRUE) {
  if(is(series[, 1], "units") | is(series[, 2], "units")) {
    series <- units::drop_units(series)
  }
  scenarios@exposure <- ExposureSeries(series)
  if(reset_times) {
    scenarios <- set_times(scenarios, dplyr::pull(series, 1))
  }
  scenarios
}

# single ExposureSeries object
#' @noRd
set_exposure_exs <- function(scenarios, series, reset_times=TRUE) {
  # coerce to data.frame to avoid issues with data.frame-like types
  df <- as.data.frame(series@series)
  # remove any `units` information to avoid issues for now. this is not ideal but we
  # would need to check and/or remove units in many places where the series
  # is used.
  # TODO check if we can keep units
  if(is(df[, 1], "units") | is(df[, 2], "units")) {
    df <- units::drop_units(df)
  }

  # assign checked data.frame to scenario but keep also any additional data from
  # exposure series object
  series@series <- df
  scenarios@exposure <- series
  if(reset_times) {
    scenarios <- set_times(scenarios, dplyr::pull(series@series, 1))
  }

  scenarios
}

# "No exposure series" object, make sure that we do not try to reset the
# output times while updating the scenario's series.
#' @noRd
set_exposure_noexs <- function(scenarios, series, ...) {
  set_exposure_exs(scenarios, series, reset_times=FALSE)
}

# list of exposure series
#' @noRd
set_exposure_lst <- function(scenarios, series, ...) {
  lapply(series, function(sc, exs) set_exposure(sc, exs, ...), sc=scenarios)
}

#' @rdname set_exposure
setMethod("set_exposure", c("EffectScenario","data.frame"), function(scenarios, series, ...) set_exposure_dfr(scenarios, series, ...))

#' @rdname set_exposure
setMethod("set_exposure", c("EffectScenario","ExposureSeries"), set_exposure_exs)

#' @rdname set_exposure
setMethod("set_exposure", c("EffectScenario","NoExposureSeries"), set_exposure_noexs)

#' @rdname set_exposure
setMethod("set_exposure", c("EffectScenario","list"), set_exposure_lst)

#' @rdname set_exposure
setMethod("set_exposure", c("list","list"), function(scenarios, series, ...) {
  unlist(lapply(scenarios, set_exposure, series=series, ...))
})

#' @rdname set_exposure
setMethod("set_exposure", c("list","ANY"), function(scenarios, series, ...) {
  lapply(scenarios, set_exposure, series=series, ...)
})

#' @rdname set_exposure
setMethod("set_exposure", c("ScenarioSequence","ANY"), function(scenarios, series, ...) {
  scenarios@scenarios <- set_exposure(scenarios@scenarios, series, ...)
  scenarios
})

#' Set zero exposure
#'
#' The scenarios current exposure is replaced by a constant exposure time-series
#' of value zero(`0.0`). Output times are unaffected.
#'
#' @param x vector of [scenarios]
#'
#' @return vector of [scenarios]
#' @export
#'
#' @examples
#' # Derive effect size in sample scenario without toxicant exposure
#' minnow_it %>%
#'   set_noexposure() %>%
#'   effect()
set_noexposure <- function(x) {
  if(is.vector(x)) {
    return(sapply(x, function(sc) set_noexposure(sc)))
  }

  set_exposure(x, no_exposure(), reset_times=FALSE)
}


# Check if object represents a valid exposure series
# @param series a data.frame or ExposureSeries object
# @return none
check_exposure <- function(series) {
  # unbox
  if(is(series, 'ExposureSeries')) {
    series <- series@series
  }
  if(!is.data.frame(series)) {
    stop("Argument `series` must be a `data.frame` or `ExposureSeries` object.")
  }

  warn <- c()
  err <- c()
  if(nrow(series) == 0) {
    err <- c(err, "Exposure series must have at least one row.")
  }
  if(length(series) < 2) {
    err <- c(err, "Exposure series must have at least two columns.")
  }
  else {
    if(length(series) > 2)
      warn <- c(warn, "Exposure series has more than two columns, additional columns are ignored.")

    time <- dplyr::pull(series, 1)
    if(!all(is.numeric(time)))
      err <- c(err, "First column (time) must contain numeric values only.")
    else
    {
      if(is(time, "units"))
        time <- units::drop_units(time)
      # no missing, infinite, or NaN values
      if(any(is.na(time) | is.nan(time) | is.infinite(time)))
        err <- c(err, "First column (time) contains invalid values (e.g. NA, NaN, Inf).")
      # time must be in ascending order
      if(any(tidyr::replace_na(diff(time), 0) < 0))
        err <- c(err, "First column (time) must be in ascending order.")
    }

    exp <- dplyr::pull(series, 2)
    if(!all(is.numeric(exp)))
      err <- c(err, "Second column (exposure) must contain numeric values only.")
    else {
      # no missing, infinite, or NaN values
      if(any(is.na(exp) | is.nan(exp) | is.infinite(exp)))
        err <- c(err, "Second column (exposure) contains invalid values (e.g. NA, NaN, Inf).")
    }
  }

  if(length(warn) > 0) {
    if(length(warn) > 1)
      warn <- paste("*", warn)
    warning(warn, call.=FALSE)
  }
  if(length(err) > 0) {
    err <- c("Exposure series is malformed.", paste("*", err))
    stop(err, call.=FALSE)
  }
}

test_that("basic arguments", {
  # single scenario, basic data.frame
  sc <- new("EffectScenario", name="a") %>% set_times(1:3)
  rs <- set_exposure(sc, data.frame(t=1, c=2), FALSE)
  expect_true(is_scenario(rs))
  expect_equal(rs@exposure@series$c, c(2))

  # single scenario, exposure series object
  es <- ExposureSeries(data.frame(t=1:2, c=2))
  rs <- set_exposure(sc, es, reset_times=TRUE)
  expect_true(is_scenario(rs))
  expect_equal(rs@exposure@series, es@series)
  expect_equal(rs@times, c(1:2))

  # single scenario, exposure series object, do not reset times
  rs <- set_exposure(sc, es, reset_times=FALSE)
  expect_true(is_scenario(rs))
  expect_equal(rs@exposure@series, es@series)
  expect_equal(rs@times, 1:3)

  # drop additional columns
  expect_warning(rs <- set_exposure(sc, data.frame(a=0, b=1, c=2), reset_times=FALSE))
  expect_equal(length(rs@exposure@series), 2)
  expect_equal(names(rs@exposure@series), c("a", "b"))
})

test_that("vectorized arguments", {
  sc <- new("EffectScenario", name="a") %>% set_times(1:3)
  sc2 <- new("EffectScenario", name="b") %>% set_times(1:3)
  es <- ExposureSeries(data.frame(t=1, c=2))
  es2 <- ExposureSeries(data.frame(t=1,c=3))

  # list of scenarios, single series
  rs <- set_exposure(list(sc,sc2), es, reset_times=FALSE)
  expect_type(rs, "list")
  expect_equal(length(rs), 2)
  expect_true(all(is_scenario(rs)))
  expect_equal(rs[[1]]@exposure@series, es@series)
  expect_equal(rs[[1]]@times, 1:3)
  expect_equal(rs[[1]]@name,  "a")
  expect_equal(rs[[2]]@exposure@series, es@series)
  expect_equal(rs[[2]]@times, 1:3)
  expect_equal(rs[[2]]@name,  "b")

  # single scenario, list of exposure series
  rs <- set_exposure(sc, list(es, es2), reset_times=FALSE)
  expect_type(rs, "list")
  expect_equal(length(rs), 2)
  expect_true(all(is_scenario(rs)))
  expect_equal(rs[[1]]@exposure@series, es@series)
  expect_equal(rs[[1]]@times, 1:3)
  expect_equal(rs[[2]]@exposure@series, es2@series)
  expect_equal(rs[[2]]@times, 1:3)

  # list of scenarios, list of exposure series
  rs <- set_exposure(list(sc,sc2), list(es,es2), reset_times=FALSE)
  expect_type(rs, "list")
  expect_equal(length(rs), 4)
  expect_true(all(is_scenario(rs)))
  expect_equal(rs[[1]]@exposure@series, es@series)
  expect_equal(rs[[1]]@times, 1:3)
  expect_equal(rs[[1]]@name,  "a")
  expect_equal(rs[[2]]@exposure@series, es2@series)
  expect_equal(rs[[2]]@times, 1:3)
  expect_equal(rs[[2]]@name,  "a")
  expect_equal(rs[[3]]@exposure@series, es@series)
  expect_equal(rs[[3]]@times, 1:3)
  expect_equal(rs[[3]]@name,  "b")
  expect_equal(rs[[4]]@exposure@series, es2@series)
  expect_equal(rs[[4]]@times, 1:3)
  expect_equal(rs[[4]]@name,  "b")
})

test_that("scenario sequence", {
  sc <- minnow_it %>% set_noexposure()
  suppressMessages(seq <- sequence(list(sc, sc), breaks=2))

  df <- data.frame(foo=1, bar=2)
  seq <- seq %>% set_exposure(df, reset_times=FALSE)
  expect_equal(seq@scenarios[[1]]@exposure@series, df)
  expect_equal(seq@scenarios[[2]]@exposure@series, df)
})

test_that("set_noexposure", {
  sc <- new("EffectScenario", name="a") %>%
    set_exposure(data.frame(time=1, conc=2), FALSE) %>%
    set_times(1:3)

  rs <- sc %>% set_exposure(no_exposure())
  # times unmodified
  expect_equal(rs@times, sc@times)
  # exposure set to zero
  expect_true(all(rs@exposure@series[,2] == 0))

  rs <- set_noexposure(sc)
  # times unmodified
  expect_equal(rs@times, sc@times)
  # exposure set to zero
  expect_true(all(rs@exposure@series[,2] == 0))

  # multiple scenarios at once
  rs <- list(sc, sc) %>% set_noexposure()
  expect_true(all(rs[[1]]@exposure@series[,2] == 0))
  expect_true(all(rs[[2]]@exposure@series[,2] == 0))
})

test_that("non-standard argument types", {
  # check that data.frame-like types, such as tibbles, work as expected
  tib <- tibble::tibble(t=0:3, c=0)
  rs <- set_exposure(new("EffectScenario", name="a"), tib)
  expect_equal(rs@exposure@series, as.data.frame(tib))

  # data.frames can contain numerical columns which carry units
  df <- data.frame(time=0:3, conc=0)
  units(df$time) <- units::as_units("days")
  units(df$conc) <- units::as_units("ug/L")
  sc <- set_exposure(new("EffectScenario", name="a"), df)
  # units must not be present afterwards
  expect_false(any(has_units(sc@exposure@series[, 1])))
  expect_false(any(has_units(sc@exposure@series[, 2])))
})

test_that("invalid arguments", {
  sc <- minnow_it
  es <- sc@exposure
  # exposure series too short
  expect_error(set_exposure(sc, data.frame(t=0, c=0)))

  # nonsense
  expect_error(set_exposure(sc, 1))
  expect_error(set_exposure(1, es))
  # invalid time-series structure
  expect_error(set_exposure(1, data.frame()))
  expect_error(set_exposure(1, data.frame(time=1)))
  expect_error(set_exposure(1, data.frame(time="1")))
  expect_error(set_exposure(1, data.frame(time=1, conc="a")))
})

test_that("check_exposure", {
  # no series
  expect_error(check_exposure())
  expect_error(check_exposure(foo))
  expect_error(check_exposure(1))
  expect_error(check_exposure(NULL))
  # too few columns
  expect_error(check_exposure(data.frame()), "two columns")
  expect_error(check_exposure(data.frame(a=0)), "two columns")
  # empty series
  expect_error(check_exposure(data.frame(t=numeric(0), e=numeric(0))), "least one row")
  # non-numeric columns
  expect_error(check_exposure(data.frame(t="foo", e=0)), "numeric values")
  expect_error(check_exposure(data.frame(t=0, e="bar")), "numeric values")
  expect_error(check_exposure(data.frame(t="foo", e="bar")), "numeric values")
  # missing/invalid values
  expect_error(check_exposure(data.frame(t=NA_real_, e=0)), "invalid values")
  expect_error(check_exposure(data.frame(t=Inf, e=0)), "invalid values")
  expect_error(check_exposure(data.frame(t=NaN, e=0)), "invalid values")
  expect_error(check_exposure(data.frame(t=0, e=NA_real_)), "invalid values")
  expect_error(check_exposure(data.frame(t=0, e=Inf)), "invalid values")
  expect_error(check_exposure(data.frame(t=0, e=NaN)), "invalid values")
  # time not sorted
  expect_error(check_exposure(data.frame(t=c(0,1,0), e=0)), "ascending order")

  # additional columns
  expect_warning(check_exposure(data.frame(t=0, e=0, f=0)), "additional columns")

  # valid argument
  check_exposure(data.frame(t=0:3, e=0))
  # valid arguments with units
  foo <- units::set_units(1, sec)
  check_exposure(data.frame(t=c(foo, foo), e=0))
  check_exposure(data.frame(t=0, e=c(foo, foo)))
})

test_that("ExposureSeries constructor", {
  # shortened test set
  expect_error(ExposureSeries())
  expect_error(ExposureSeries(1))
  expect_error(ExposureSeries(data.frame()), "two columns")
  expect_error(ExposureSeries(data.frame(t=numeric(0), e=numeric(0))), "least one row")
  expect_error(ExposureSeries(data.frame(t="foo", e=0)), "numeric values")
  expect_error(ExposureSeries(data.frame(t=NA_real_, e=0)), "invalid values")
  expect_warning(es <- ExposureSeries(data.frame(t=0, e=0, f=0)), "additional columns")
  # where additional columns dropped?
  expect_equal(length(es@series), 2)

  # valid arguments
  df <- data.frame(t=0:3, e=0)
  es <- ExposureSeries(df)
  expect_equal(es@series, df)
  # valid arguments with units
  foo <- units::set_units(1, sec)
  df <- data.frame(t=c(foo, foo), e=0)
  es <- ExposureSeries(df)
  expect_equal(es@series, df)
})

#
# Lemna simulation without transfers to new medium.
# Reference values were calculated using the original model of Schmitt et al. (1993)
# doi.org/10.1016/j.ecolmodel.2013.01.017
#
test_that("Lemna_Schmitt simulation", {
  tol <- 1e-5
  # no exposure
  metsulfuron %>%
    set_exposure(no_exposure(), FALSE) %>%
    set_times(0:14) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:14)

  expect_equal(out[1,"BM"], 50)
  expect_equal(out[8,"BM"], 70.52071, tolerance=tol)
  expect_equal(out[15,"BM"], 91.13051, tolerance=tol)

  expect_equal(out[,"E"], rep(1,15))
  expect_equal(out[,"M_int"], rep(0,15))
  expect_equal(out[,"C_int"], rep(0,15))

  expect_equal(out[1,"FrondNo"], 500000)
  expect_equal(out[8,"FrondNo"], 705207.1, tolerance=tol)
  expect_equal(out[15,"FrondNo"], 911305.1, tolerance=tol)

  # constant exposure
  metsulfuron %>%
    set_exposure(data.frame(t=0, c=1), FALSE) %>%
    set_times(0:14) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:14)

  expect_equal(out[1,"BM"], 50)
  expect_equal(out[8,"BM"], 51.62533, tolerance=tol)
  expect_equal(out[15,"BM"], 50.32514, tolerance=tol)

  expect_equal(out[,"E"], rep(1,15))

  # M_int not reported by original model, cannot test
  # expect_equal(out[,"M_int"], NA)

  expect_equal(out[1,"C_int"], 0)
  expect_equal(out[8,"C_int"], 0.6898937, tolerance=tol)
  expect_equal(out[15,"C_int"], 0.7206542, tolerance=tol)

  expect_equal(out[1,"FrondNo"], 500000)
  expect_equal(out[8,"FrondNo"], 516253.3, tolerance=tol)
  expect_equal(out[15,"FrondNo"], 503251.4, tolerance=tol)

  # step function exposure
  metsulfuron %>%
    set_exposure(data.frame(t=0:14,c=c(rep(1,7),rep(0,8)))) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:14)

  expect_equal(out[1,"BM"], 50)
  expect_equal(out[8,"BM"], 51.63740, tolerance=tol)
  expect_equal(out[15,"BM"], 65.95882, tolerance=tol)

  expect_equal(out[,"E"], rep(1,15))

  # M_int not reported by original model, cannot test
  # expect_equal(out[,"M_int"], NA)

  expect_equal(out[1,"C_int"], 0)
  expect_equal(out[8,"C_int"], 0.54974771, tolerance=tol)
  expect_equal(out[15,"C_int"], 0.01825732, tolerance=tol)

  expect_equal(out[1,"FrondNo"], 500000)
  expect_equal(out[8,"FrondNo"], 516374.0, tolerance=tol)
  expect_equal(out[15,"FrondNo"], 659588.2, tolerance=tol)
})

test_that("Lemna_SchmittThold simulation", {
  sc <- Lemna_SchmittThold() %>% set_all(metsulfuron)
  sc@init <- c(sc@init, AUC=0)

  out_orig <- simulate(metsulfuron, method="lsoda")

  # no threshold set, no influence on results
  expect_equal(simulate(sc, method="lsoda")[-5], out_orig, ignore_attr=TRUE)

  # results should be identical for very large, i.e. irrelevant, thresholds
  sc %>% set_param(c(threshold=1e10)) %>% simulate(method="lsoda") -> out
  expect_equal(out[-5], out_orig, ignore_attr=TRUE)

  # plausibility of threshold dynamics
  sc %>% set_param(c(threshold=2)) %>% simulate(method="lsoda") -> out
  expect_equal(out$AUC, c(0:6, rep(6.5, 8)), tolerance=1e-5, ignore_attr=TRUE) # AUC values
  expect_equal(out[1:3,][-5], out_orig[1:3,], tolerance=1e-5, ignore_attr=TRUE) # unaffected growth until threshold
  expect_true(all(diff(out$BM[-c(1,2)]) < 0)) # BM declines after threshold is exceeded

  # compare to values from Schmitt et al. implementation
  # the results are not extremely accurate because trapz(), which is used in
  # the Schmitt et al. model for AUC calculation, deviates visibly from exact results
  expect_equal(out[3,"BM"], 52.43495, tolerance=1e-3) # before threshold
  expect_equal(out[4,"BM"], 51.42340, tolerance=1e-3) # after threshold
  expect_equal(out[14,"BM"], 41.97315, tolerance=1e-3)

  # error on missing AUC state variable
  expect_error(Lemna_SchmittThold() %>% set_all(metsulfuron) %>% set_param(c(threshold=1)) %>% simulate(),
               regexp="AUC state variable")
})

#
# Lemna simulation without transfers to new medium.
# Reference values were calculated using the original model of Schmitt et al. (1993)
# doi.org/10.1016/j.ecolmodel.2013.01.017
#
test_that("Lemna simulation, regular transfers", {
  tol <- 1e-4
  ## no exposure, no transfer
  metsulfuron %>%
    set_init(c(BM=0.0012,E=1,M_int=0)) %>%
    set_exposure(no_exposure(), FALSE) %>%
    set_transfer(interval=7) %>%
    set_times(0:7) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:7)

  expect_equal(out[1,"BM"], 0.0012, tolerance=tol)
  expect_equal(out[8,"BM"], 0.002177663, tolerance=tol)

  expect_equal(out[,"E"], rep(1,8))
  expect_equal(out[,"M_int"], rep(0,8))
  expect_equal(out[,"C_int"], rep(0,8))

  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[8,"FrondNo"], 21.77663, tolerance=tol)

  ## no exposure, one transfer
  metsulfuron %>%
    set_init(c(BM=0.0012,E=1,M_int=0)) %>%
    set_exposure(no_exposure(), FALSE) %>%
    set_transfer(interval=7) %>%
    set_times(0:14) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:14)

  expect_equal(out[1,"BM"], 0.0012, tolerance=tol)
  expect_equal(out[8,"BM"], 0.002177663, tolerance=tol)
  expect_equal(out[9:15,"BM"], out[2:8,"BM"], tolerance=tol)

  expect_equal(out[,"E"], rep(1,15))
  expect_equal(out[,"M_int"], rep(0,15))
  expect_equal(out[,"C_int"], rep(0,15))

  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[8,"FrondNo"], 21.77663, tolerance=tol)
  expect_equal(out[9:15,"FrondNo"], out[2:8,"FrondNo"], tolerance=tol)

  ## no exposure, multiple transfers
  metsulfuron %>%
    set_init(c(BM=0.0012,E=1,M_int=0)) %>%
    set_noexposure() %>%
    set_transfer(interval=7) %>%
    set_times(0:28) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:28)

  expect_equal(out[1,"BM"], 0.0012, tolerance=tol)
  expect_equal(out[8,"BM"], 0.002177663, tolerance=tol)
  expect_equal(out[9:15,"BM"], out[2:8,"BM"], tolerance=tol)
  expect_equal(out[9:15,"BM"], out[2:8,"BM"], tolerance=tol)
  expect_equal(out[16:22,"BM"], out[2:8,"BM"], tolerance=tol)
  expect_equal(out[23:29,"BM"], out[2:8,"BM"], tolerance=tol)

  expect_equal(out[,"E"], rep(1,29))
  expect_equal(out[,"M_int"], rep(0,29))
  expect_equal(out[,"C_int"], rep(0,29))

  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[8,"FrondNo"], 21.77663, tolerance=tol)
  expect_equal(out[9:15,"FrondNo"], out[2:8,"FrondNo"], tolerance=tol)
  expect_equal(out[16:22,"FrondNo"], out[2:8,"FrondNo"], tolerance=tol)
  expect_equal(out[23:29,"FrondNo"], out[2:8,"FrondNo"], tolerance=tol)

  ## no exposure, irregular simulation periods with transfer
  # sim stops before transfer
  metsulfuron %>%
    set_init(c(BM=0.0012,E=1,M_int=0)) %>%
    set_noexposure() %>%
    set_transfer(interval=7) %>%
    set_times(0:6) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:6)
  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[2,"FrondNo"], 13.06635, tolerance=tol)
  expect_equal(out[7,"FrondNo"], 19.99945, tolerance=tol)

  # sim entails 1st transfer
  metsulfuron %>%
    set_init(c(BM=0.001836730,E=1,M_int=0)) %>%
    set_noexposure() %>%
    set_transfer(interval=7) %>%
    set_times(5:10) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 5:10)
  expect_equal(out[1,"FrondNo"], 18.36730, tolerance=tol) # t=5
  expect_equal(out[2,"FrondNo"], 19.99945, tolerance=tol) # t=6
  expect_equal(out[3,"FrondNo"], 21.77663, tolerance=tol) # t=7
  expect_equal(out[4,"FrondNo"], 13.06635, tolerance=tol) # t=8, frond number reset
  expect_equal(out[5,"FrondNo"], 14.22745, tolerance=tol) # t=9
  expect_equal(out[6,"FrondNo"], 15.49173, tolerance=tol) # t=10

  # sim starts and ends in 2nd interval, no transfer occurring
  metsulfuron %>%
    set_init(c(BM=0.0012,E=1,M_int=0)) %>%
    set_noexposure() %>%
    set_transfer(interval=7) %>%
    set_times(7:10) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 7:10)
  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[2,"FrondNo"], 13.06635, tolerance=tol)
  expect_equal(out[4,"FrondNo"], 15.49173, tolerance=tol)

  ## constant exposure, no transfer
  metsulfuron %>%
    set_init(c(BM=0.0012, E=1, M_int=0)) %>%
    set_exposure(data.frame(t=0, c=1), FALSE) %>%
    set_transfer(interval=7) %>%
    set_times(0:7) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:7)
  expect_equal(out[1,"BM"], 0.0012)
  expect_equal(out[8,"BM"], 0.001334016, tolerance=tol)
  expect_equal(out[,"E"], rep(1,8))
  expect_equal(out[1,"M_int"], 0)
  expect_equal(out[1,"C_int"], 0)
  expect_equal(out[8,"C_int"], 0.6804143, tolerance=tol)
  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[8,"FrondNo"], 13.34016, tolerance=tol)

  ## constant exposure, one transfer
  metsulfuron %>%
    set_init(c(BM=0.0012, E=1, M_int=0)) %>%
    set_exposure(data.frame(t=0, c=1), FALSE) %>%
    set_transfer(interval=7,biomass=0.0012) %>%
    set_times(0:14) %>%
    simulate(method="lsoda", hmax=0.1) -> out

  expect_equal(out[,"time"], 0:14)
  expect_equal(out[1,"BM"], 0.0012)
  expect_equal(out[8,"BM"], 0.001334016, tolerance=tol)
  expect_equal(out[9,"BM"], 0.001203635, tolerance=3e-4)
  expect_equal(out[15,"BM"], 0.001226337, tolerance=8e-4)
  expect_equal(out[,"E"], rep(1,15))
  expect_equal(out[1,"M_int"], 0)
  expect_equal(out[1,"C_int"], 0)
  expect_equal(out[8,"C_int"], 0.6804143, tolerance=tol)
  expect_equal(out[9,"C_int"], 0.6916387, tolerance=tol)
  expect_equal(out[15,"C_int"], 0.7099914, tolerance=tol)
  expect_equal(out[1,"FrondNo"], 12)
  expect_equal(out[8,"FrondNo"], 13.34016, tolerance=1e-4)
  expect_equal(out[9,"FrondNo"], 12.03918, tolerance=tol)
  expect_equal(out[15,"FrondNo"], 12.27103, tolerance=tol)
})

test_that("Lemna simulation irregular transfers", {
  tol <- 1e-5
  ## irregular transfer times

  # vanilla use case, one transfer in the middle of the simulated period
  metsulfuron %>%
    set_init(c(BM=0.001, E=1, M_int=0)) %>%
    set_exposure(data.frame(t=0, c=0), FALSE) %>%
    set_transfer(interval=7, biomass=0.001) %>%
    set_times(0:14) -> sc.reg
  sc.reg %>% set_transfer(times=c(7)) -> sc.irg

  expect_equal(simulate(sc.irg, method="lsoda", hmax=0.1),
               simulate(sc.reg, method="lsoda", hmax=0.1))


  sc.reg %>%
    set_transfer(interval=5) %>%
    set_times(0:12) %>%
    simulate(method="lsoda", hmax=0.01) -> out.reg
  sc.reg %>%
    set_transfer(times=c(5,10,12,14,16)) %>%
    set_times(0:20) %>%
    simulate(method="lsoda", hmax=0.01) -> out.irg
  # same results until first transfer?
  expect_equal(out.irg[1:6,-1], out.reg[1:6,-1], tolerance=tol)
  # same results in 2nd interval?
  expect_equal(out.irg[6:11,-1], out.reg[6:11,-1], tolerance=tol)
  # same results at beginning of 3rd interval?
  expect_equal(out.irg[6:8,"BM"], out.reg[11:13,"BM"], ignore_attr=T, tolerance=tol)
  # same results at consecutive irregular intervals?
  expect_equal(out.irg[11:13,"BM"], out.reg[6:8,"BM"], ignore_attr=T, tolerance=tol)
  expect_equal(out.irg[14:15,"BM"], out.reg[7:8,"BM"], ignore_attr=T, tolerance=tol)
  expect_equal(out.irg[16:17,"BM"], out.reg[7:8,"BM"], ignore_attr=T, tolerance=tol)
  expect_equal(out.irg[18:21,"BM"], out.reg[7:10,"BM"], ignore_attr=T, tolerance=tol)

  # one transfer at end of period
  metsulfuron %>%
    set_init(c(BM=0.001, E=1, M_int=0)) %>%
    set_exposure(data.frame(t=0, c=0), FALSE) %>%
    set_times(0:14) %>%
    set_transfer(interval=-1) -> sc.reg
  sc.reg %>% set_transfer(times=c(14)) -> sc.irg
  expect_equal(simulate(sc.reg), simulate(sc.irg), tolerance=tol)

  # one transfer at beginning of period
  sc.reg %>% set_transfer(times=c(0)) -> sc.irg
  expect_equal(simulate(sc.reg), simulate(sc.irg), tolerance=tol)

  # no relevant transfers in period
  sc.reg %>% set_transfer(times=c(21)) -> sc.irg
  expect_equal(simulate(sc.reg), simulate(sc.irg), tolerance=tol)
})

test_that("Lemna solver", {
  # Schmitt model
  expect_equal(as.data.frame(solver(metsulfuron))$time, metsulfuron@times)

  # Schmitt Threshold model
  Lemna_SchmittThold() %>%
    set_init(metsulfuron@init) %>%
    set_param(metsulfuron@param) %>%
    set_param(c("threshold"=2)) %>%
    set_forcings(metsulfuron@forcings) %>%
    set_exposure(metsulfuron@exposure@series) -> sc
  expect_equal(as.data.frame(solver(sc))$time, sc@times)

  # SETAC model
  focusd1 %>% set_times(0:10) -> sc
  expect_equal(as.data.frame(solver(sc))$time, sc@times)
})

# test routine ---------------------------------------------------------------------
# Simulation with the GUTS model, compare against reference values.
# Dataset A is part of EFSA's *GUTS* ringtest to compare results from software
# implementations of [GUTS-RED models][GUTS-RED-models] (EFSA 2018). The
# ringtest focused on [GUTS-RED-IT][GUTS_RED_IT] and [GUTS-RED-SD][GUTS_RED_SD]
# models.
test_that("GUTS-RED-IT simulation", {
  tol <- 1e-4

  # define GUTS model
  guts_it <- GUTS_RED_IT() %>%
    set_endpoints("S") %>%
    set_param(c(hb = 0, kd = 0.5, alpha = 5, beta = 5)) %>%
    set_noexposure() %>%
    set_bounds(list(
      kd = list(0.001, 1),
      alpha = list(0.1, 10),
      beta = list(0.1, 10),
      hb = list(0, 1)
    ))

  #####  optimise kD and TD parameters  ##########################

  # fit background mortality hb
  # GUTS IT
  dt_control <- ringtest_it %>%
    dplyr::filter(replicate == "Control") %>%
    dplyr::mutate(S = Nsurv / max(Nsurv)) %>%
    dplyr::select(time, S)

  fit <- calibrate(
    x = guts_it,
    par = c(hb = 0),
    data = dt_control,
    output = "S",
    method = "Brent",
    verbose=FALSE
  )

  guts_it <- guts_it %>% set_param(c(hb = as.numeric(fit$par)))

  #####

  data <- ringtest_it %>%
    dplyr::group_by(replicate) %>%
    dplyr::mutate(S = Nsurv / max(Nsurv)) %>%
    dplyr::select(time, S, replicate, conc) %>%
    dplyr::ungroup()

  calib_par <- calibrate(guts_it,
                         data = data,
                         par = c(kd = 0.5, alpha = 5, beta = 5),
                         output = "S",
                         maxsteps = 10^6,
                         approx = "linear",
                         hmax = 0.01,
                         verbose = FALSE,
                         hessian = TRUE
  )

  guts_it <- guts_it %>%
    set_param(calib_par$par)

  suppressMessages({
  res <- lik_profile(
    x = guts_it,
    data = data,
    par = calib_par$par,
    output = "S",
    pars_bound <- list(
      hb = list(0, 1),
      kd = list(0.001, 1),
      alpha = list(0.1, 5),
      beta = list(0.1, 5)
    ),
    hmax = 0.01,
    maxsteps = 10^6,
    data_type = "continuous"
  )
  })

  # # tests for best fit parameters
  expect_equal(round(guts_it@param$hb,6), 0.044089, tolerance = tol)
  expect_equal(round(guts_it@param$kd,6), 0.774101, tolerance = tol)
  expect_equal(round(guts_it@param$alpha,6), 5.563681, tolerance = tol)
  expect_equal(round(guts_it@param$beta,6), 5.085245, tolerance = tol)
  expect_equal(res$kd$confidence_interval[1], 0.7076738, tolerance = tol)
  expect_equal(res$kd$confidence_interval[2], 0.8490582, tolerance = tol)
  expect_equal(res$alpha$confidence_interval[1], 5.451594, tolerance = tol)
  expect_equal(res$alpha$confidence_interval[2], 5.563681, tolerance = tol)
  expect_equal(res$beta$confidence_interval[1], 4.514981, tolerance = tol)
  expect_equal(res$beta$confidence_interval[2], 5.085245, tolerance = tol)
})

# Extensive tests of model behavior and outputs are part of the automated model
# verification, see `test-verify-guts_red.R`

test_that("background mortality", {
  # RED-IT
  ctrl <- minnow_it %>% set_noexposure() %>% set_param(c(hb=0))
  mort <- ctrl %>% set_param(c(hb=1))

  expect_equal(simulate(ctrl)$S, rep(1, 5))
  expect_true(all(diff(simulate(mort)$S) < 0)) # survival prob is strictly decreasing

  # RED-SD
  ctrl <- minnow_sd %>% set_noexposure() %>% set_param(c(hb=0))
  mort <- ctrl %>% set_param(c(hb=1))

  expect_equal(simulate(ctrl)$S, rep(1, 5))
  expect_true(all(diff(simulate(mort)$S) < 0)) # survival prob is strictly decreasing
})

test_that("C code errors", {
  # invalid parameter values
  expect_error(minnow_it %>% set_param(c(kd=-1)) %>% simulate(), "smaller than zero")
  expect_error(minnow_it %>% set_param(c(hb=-1)) %>% simulate(), "smaller than zero")

  # invalid exposure values
  expect_error(minnow_it %>% set_exposure(data.frame(t=0, -1), FALSE) %>% simulate())

  # invalid parameter values
  expect_error(minnow_sd %>% set_param(c(kd=-1)) %>% simulate(), "smaller than zero")
  expect_error(minnow_sd %>% set_param(c(hb=-1)) %>% simulate(), "smaller than zero")
  expect_error(minnow_sd %>% set_param(c(kk=-1)) %>% simulate(), "smaller than zero")
  expect_error(minnow_sd %>% set_param(c(z=-1)) %>% simulate(), "smaller than zero")

  # invalid exposure values
  expect_error(minnow_sd %>% set_exposure(data.frame(t=0, -1), FALSE) %>% simulate())
})

test_that("output variables", {
  rs <- minnow_it %>% simulate(nout=0)
  rs2 <- minnow_it %>% simulate(nout=1)

  expect_equal(length(names(rs2)), length(names(rs)) + 1)
  # S is calculated on the return value of deSolve and therefore appears at the
  # end of the columns, after the output variable Cw, unfortunately
  expect_equal(names(rs2), c("time","D","H","Cw","S"))
  expect_equal(rs2$Cw, minnow_it@exposure@series[,2]) # external concentration

  rs <- minnow_sd %>% simulate(nout=0)
  rs2 <- minnow_sd %>% simulate(nout=1)

  expect_equal(length(names(rs2)), length(names(rs)) + 1)
  # S is calculated on the return value of deSolve and therefore appears at the
  # end of the columns, after the output variable Cw, unfortunately
  expect_equal(names(rs2), c("time","D","H","Cw","S"))
  expect_equal(rs2$Cw, minnow_sd@exposure@series[,2]) # external concentration
})

test_that("solvers", {
  # RED-IT
  expect_equal(solver(minnow_it)$time, minnow_it@times)
  # RED-SD
  expect_equal(solver(minnow_it)$time, minnow_it@times)
})

test_that("effects", {
  # RED-IT
  ctrl <- minnow_it %>% set_noexposure() %>% simulate()
  trl <- minnow_it %>% simulate()
  myeffect <- 1 - tail(trl$S, n=1) / tail(ctrl$S, n=1)
  expect_equal(myeffect, effect(minnow_it)$L)

  # RED-SD
  ctrl <- minnow_sd %>% set_noexposure() %>% simulate()
  trl <- minnow_sd %>% simulate()
  myeffect <- 1 - tail(trl$S, n=1) / tail(ctrl$S, n=1)
  expect_equal(myeffect, effect(minnow_sd)$L)
})

# test routine ---------------------------------------------------------------------
# Simulation with the GUTS model, compare against reference values.
# Dataset A is part of EFSA's *GUTS* ringtest to compare results from software
# implementations of [GUTS-RED models][GUTS-RED-models] (EFSA 2018). The
# ringtest focused on [GUTS-RED-IT][GUTS_RED_IT] and [GUTS-RED-SD][GUTS_RED_SD]
# models.
test_that("GUTS-RED-SD simulation", {
  tol <- 1e-4

  # GUTS SD
  guts_sd <- GUTS_RED_SD() %>%
    set_endpoints("S") %>%
    set_param(c(hb = 0, kd = 0.6, z = 5, kk = 0.5)) %>%
    set_noexposure() %>%
    set_bounds(list(
      kd = c(0.001, 1),
      z = c(0.01, 5),
      kk = c(0.01, 5),
      hb = c(0, 0.5)
    ))

  #####  optimise kD and TD parameters  ##########################
  # GUTS SD
  dt_control <- ringtest_sd %>%
    dplyr::filter(replicate == "Control") %>%
    dplyr::mutate(S = Nsurv / max(Nsurv)) %>%
    dplyr::select(time, S)

  fit <- calibrate(
    x = guts_sd,
    par = c(hb = 0),
    data = dt_control,
    output = "S",
    method = "Brent",
    verbose=FALSE
  )

  guts_sd <- guts_sd %>% set_param(c(hb = as.numeric(fit$par)))

  #####

  data <- ringtest_sd %>%
    dplyr::group_by(replicate) %>%
    dplyr::mutate(S = Nsurv / max(Nsurv)) %>%
    dplyr::select(time, S, replicate, conc) %>%
    dplyr::ungroup()

  calib_par <- calibrate(guts_sd,
                         data = data,
                         par = c(kd = 1, z = 0.1, kk = 0.1),
                         output = "S",
                         maxsteps = 10^6,
                         approx = "linear",
                         hmax = 0.01,
                         verbose = FALSE,
                         hessian = TRUE
  )

  guts_sd <- guts_sd %>%
    set_param(calib_par$par) %>%
    set_noexposure()

  suppressMessages({
  res <- lik_profile(
    x = guts_sd,
    data = data,
    output = "S",
    par = calib_par$par,
    bounds = list(
      kd = c(0.001, 1),
      z = c(0.01, 10),
      kk = c(0.01, 5),
      hb = c(0, 1)
    ),
    refit = FALSE,
    type = "fine",
    method = "Brent",
    data_type = "continuous"
  )
  })

  # # tests for best fit parameters
  expect_equal(round(guts_sd@param$hb,6), 0.011794, tolerance = tol)
  expect_equal(round(guts_sd@param$kd,6), 0.685552, tolerance = tol)
  expect_equal(round(guts_sd@param$z,6), 3.003922, tolerance = tol)
  expect_equal(round(guts_sd@param$kk,6), 0.728990, tolerance = tol)
  expect_equal(res$kd$confidence_interval[1], 0.6671638, tolerance = tol)
  expect_equal(res$kd$confidence_interval[2], 0.7041692, tolerance = tol)
  expect_equal(res$kk$confidence_interval[1], 0.6889944, tolerance = tol)
  expect_equal(res$kk$confidence_interval[2], 0.7706515, tolerance = tol)
  expect_equal(res$z$confidence_interval[1], 2.964857, tolerance = tol)
  expect_equal(res$z$confidence_interval[2], 3.041375, tolerance = tol)
})

test_that("arg x=scenario (Lemna_Schmitt)", {
  sc <- metsulfuron %>% set_noexposure()
  df <- data.frame(t=0:2, o=12)
  td <- tox_data(df)
  # w/o arg data
  expect_error(fit_growth(sc), "data. is missing")

  # w/ arg data=data.frame
  fit_growth(sc, data=df)
  succeed()

  # w/ arg data=tox_data
  fit_growth(sc, data=td)
  succeed()

  # w/ arg data, but no exposure info anywhere. this can happen if no exposure
  # series was set for a newly created scenario.
  sc2 <- metsulfuron
  sc2@exposure@series <- data.frame(t=numeric(0), c=numeric(0))
  fit_growth(sc2, data=df)
  succeed()
})

test_that("arg x=scenario w/ exposure", {
  sc <- metsulfuron %>% set_exposure(data.frame(time=0:2, 1))
  df <- data.frame(t=0:2, o=12)

  # w/ arg data=data.frame
  expect_error(fit_growth(sc, data=df), "must not include exposure")
})

test_that("arg x=sequence", {
  sc <- metsulfuron %>% set_noexposure()
  sq <- sequence(list(sc, sc), breaks=1)
  df <- data.frame(t=0:2, o=12)

  # w/ arg data=data.frame
  fit_growth(sq, data=df)
  succeed()
})

test_that("arg x=caliset", {
  sc <- metsulfuron %>% set_noexposure()
  df <- data.frame(t=0:2, o=12)
  cs <- caliset(sc, df)

  expect_error(fit_growth(cs), "not supported")
})

test_that("arg x=list of calisets", {
  sc <- metsulfuron %>% set_noexposure()
  df <- data.frame(t=0:2, o=12)
  cs <- list(caliset(sc, df))

  fit_growth(cs)
  succeed()

  # w/ arg data=fails
  expect_error(fit_growth(cs, data=df), "not supported")
})

test_that("arg x=unsupported scenario", {
  sc <- new("EffectScenario") %>% set_noexposure()
  df <- data.frame(t=0:2, o=12)

  # scenario w/o data
  expect_error(fit_growth(sc), "data. is missing")
  # scenario w/ data
  expect_error(fit_growth(sc, data=df), "not supported yet")
  # caliset
  cs <- caliset(sc, data=df)
  expect_error(fit_growth(cs), "not supported")
})

# arg x=fit: does this need support?

effect_IT <- function(df) {
  minnow_it %>% set_exposure(df) %>% effect(ep_only=TRUE, rtol=0) %>% unname()
}

effect_SD <- function(df) {
  minnow_sd %>% set_exposure(df) %>% effect(ep_only=TRUE, rtol=0, hmax=0.01) %>% unname()
}

window_IT <- function(df,window=-1,interval=-1) {
  mod <- minnow_it
  if(window > 0)
    mod <- set_window(mod, window, interval)
  mod %>%
    set_exposure(df) %>%
    effect(max_only=FALSE, hmax=0.001, method="ode45") %>%
    dplyr::mutate(L=round(L, 5)) %>% # avoid marginal differences due to numerical errors
    dplyr::arrange(dplyr::desc(L)) %>%
    dplyr::slice(1) -> tib
  c(tib[[1,"dat.start"]],tib[[1,"dat.end"]])
}

test_that("return types", {
  sc <- minnow_it

  expect_true(is.numeric(effect(sc, max_only=TRUE, ep_only=TRUE)))
  expect_true(is.data.frame(effect(sc, max_only=TRUE, ep_only=FALSE)))
  expect_true(is.data.frame(effect(sc, max_only=FALSE, ep_only=TRUE)))
  expect_true(is.data.frame(effect(sc, max_only=FALSE, ep_only=FALSE)))
  expect_true(is.numeric(effect(sc)$L))

  sc <- metsulfuron
  expect_true(is.numeric(effect(sc, max_only=TRUE, ep_only=TRUE)))
  expect_true(is.data.frame(effect(sc, max_only=TRUE, ep_only=FALSE)))
  expect_true(is.data.frame(effect(sc, max_only=FALSE, ep_only=TRUE)))
  expect_true(is.data.frame(effect(sc, max_only=FALSE, ep_only=FALSE)))
  expect_true(is.numeric(effect(sc)$BM))
  expect_true(is.numeric(effect(sc)$r))
})

test_that("return names", {
  # A GUTS-RED-IT model has only one effect endpoint
  sc <- minnow_it

  fxn <- function(...) {
    names(effect(sc, ...))
  }

  expect_equal(fxn(ep_only=FALSE, max_only=FALSE), c("scenario","L","dat.start","dat.end"))
  expect_equal(fxn(ep_only=FALSE, max_only=TRUE),  c("scenario","L","L.dat.start","L.dat.end"))
  expect_equal(fxn(ep_only=TRUE,  max_only=FALSE), c("L"))
  expect_equal(fxn(ep_only=TRUE,  max_only=TRUE),  c("L"))

  expect_equal(fxn(ep_only=FALSE, max_only=FALSE, marginal_effect=1e-3), c("scenario","L","dat.start","dat.end"))
  expect_equal(fxn(ep_only=TRUE,  max_only=TRUE,  marginal_effect=1e-3), c("L"))

  expect_equal(fxn(ep_only=FALSE, max_only=FALSE, factor=2), c("scenario","L","dat.start","dat.end"))
  expect_equal(fxn(ep_only=TRUE,  max_only=TRUE,  factor=2), c("L"))

  # The Lemna model has two effect endpoints
  sc <- metsulfuron

  expect_equal(fxn(ep_only=FALSE, max_only=FALSE), c("scenario","BM","r","dat.start","dat.end"))
  expect_equal(fxn(ep_only=FALSE, max_only=TRUE),  c("scenario","BM","BM.dat.start","BM.dat.end","r","r.dat.start","r.dat.end"))
  expect_equal(fxn(ep_only=TRUE,  max_only=FALSE), c("BM","r"))
  expect_equal(fxn(ep_only=TRUE,  max_only=TRUE),  c("BM","r"))

  expect_equal(fxn(ep_only=FALSE, max_only=FALSE, marginal_effect=1e-3), c("scenario","BM","r","dat.start","dat.end"))
  expect_equal(fxn(ep_only=TRUE,  max_only=TRUE,  marginal_effect=1e-3), c("BM","r"))

  expect_equal(fxn(ep_only=FALSE, max_only=FALSE, factor=2), c("scenario","BM","r","dat.start","dat.end"))
  expect_equal(fxn(ep_only=TRUE,  max_only=TRUE,  factor=2), c("BM","r"))
})

test_that("factor argument", {
  sc <- minnow_it
  sc0 <- sc %>% set_noexposure()
  sc10 <- sc %>% set_exposure(data.frame(t=0, c=10), reset_times=FALSE)
  sc20 <- sc %>% set_exposure(data.frame(t=0, c=20), reset_times=FALSE)

  expect_equal(effect(sc, factor=0, ep_only=TRUE), effect(sc0, ep_only=TRUE))
  expect_equal(effect(sc, factor=10, ep_only=TRUE), effect(sc10, ep_only=TRUE))
  expect_equal(effect(sc, factor=20, ep_only=TRUE), effect(sc20, ep_only=TRUE))
})

test_that("marginal effects", {
  source(test_path("dummy.R"), local = TRUE)

  # nothing changes if effect larger than threshold
  sc <- new("DummyScenario", endpoints=c("foo","bar"), fx=c("foo"=1))
  expect_equal(effect(sc, ep_only=TRUE, marginal_effect=1e-4), c(foo=1, bar=NA_real_))

  # round effect value to zero
  sc <- new("DummyScenario", endpoints=c("foo","bar"), fx=c("foo"=1e-5))
  expect_equal(effect(sc, ep_only=TRUE, marginal_effect=1e-4), c(foo=0, bar=NA_real_))
  sc <- new("DummyScenario", endpoints=c("foo","bar"), fx=c("foo"=-1e-5))
  expect_equal(effect(sc, ep_only=TRUE, marginal_effect=1e-4), c(foo=0, bar=NA_real_))

  sc <- new("DummyScenario", endpoints="foo", fx=c("foo"=-1e-5))
  expect_equal(effect(sc, ep_only=TRUE, marginal_effect=1e-4), c(foo=0))

  # test one 'real' scenario
  sc <- minnow_it
  expect_equal(effect(sc, ep_only=TRUE, marginal_effect=1e-4), c(L=0))

  # questionable inputs
  expect_warning(effect(sc, marginal_effect=0.1), "larger than 1%")
})

test_that("generic effect calculation", {
  # no effect
  expect_equal(calc_effect(1, 1), 0)
  expect_equal(calc_effect(0, 0), 0)
  expect_equal(calc_effect(-1, -1), 0)
  # positive finite effects, i.e. treatment is lower than control
  expect_equal(calc_effect(0.9, 1), 0.1)
  expect_equal(calc_effect(0,  1), 1)
  expect_equal(calc_effect(-1, 1), 2)

  expect_equal(calc_effect(-1.1, -1), 0.1)
  expect_equal(calc_effect(-2,   -1), 1)
  expect_equal(calc_effect(-3,   -1), 2)
  # negative finite effects, i.e. treatment is higher than control
  expect_equal(calc_effect(1.1, 1), -0.1)
  expect_equal(calc_effect(2,   1), -1)
  expect_equal(calc_effect(3,   1), -2)

  expect_equal(calc_effect(-0.9, -1), -0.1)
  expect_equal(calc_effect( 0,   -1), -1)
  expect_equal(calc_effect( 1,   -1), -2)
  # infinite effects, i.e. control is zero
  expect_equal(calc_effect(0.1,   0), -Inf)
  expect_equal(calc_effect(-0.1,  0), Inf)
})

test_that("GUTS-RED pulsed exposure", {
  # allowed numerical tolerance between model runs
  tol <- 0.001
  # define some exposure patterns
  conc <- 10
  constant <- data.frame(t=c(1,10),c=c(conc))
  pulse_start <- data.frame(t=c(1,10,10.001,20),c=c(conc,conc,0,0))
  pulse_mid <- data.frame(t=c(1,4.999,5,14,14.001,20),c=c(0,0,conc,conc,0,0))
  pulse_end <- data.frame(t=c(1,10.999,11,20),c=c(0,0,conc,conc))

  # IT model
  expect_equal(effect_IT(constant), 0.578722, tolerance=tol)
  expect_equal(effect_IT(constant), effect_IT(pulse_start), tolerance=tol)
  expect_equal(effect_IT(constant), effect_IT(pulse_mid), tolerance=tol)
  expect_equal(effect_IT(constant), effect_IT(pulse_end), tolerance=tol)
  # SD model
  expect_equal(effect_SD(constant), 0.9376296, tolerance=tol)
  expect_equal(effect_SD(constant), effect_SD(pulse_start), tolerance=tol)
  expect_equal(effect_SD(constant), effect_SD(pulse_mid), tolerance=tol)
  expect_equal(effect_SD(constant), effect_SD(pulse_end), tolerance=tol)
})

test_that("GUTS-RED background mortality", {
  # hb>0 should not affect effect levels
  expect_error(effect(minnow_it %>% set_param(c(hb=1)), minnow_it))
  expect_error(effect(minnow_sd %>% set_param(c(hb=1)), minnow_sd))
})

#test if function correctly finds maximum effect within a time-series
test_that("window finding GUTS", {
  # define some exposure patterns
  conc <- 10
  constant <- data.frame(t=1:10,c=c(conc))
  pulse_start <- data.frame(t=1:20,c=c(rep(conc,5),rep(0,15)))
  pulse_mid <- data.frame(t=1:20,c=c(rep(0,5),rep(conc,5),rep(0,10)))
  pulse_end <- data.frame(t=1:20,c=c(rep(0,15),rep(conc,5)))

  # full range of profile
  expect_equal(window_IT(constant), c(1,10))
  expect_equal(window_IT(pulse_start), c(1,20))
  # should return first window of multiple candidates of equal effect
  expect_equal(window_IT(constant,window=5,interval=1), c(1,6))
  # identification of max effect window, full exposure
  expect_equal(window_IT(pulse_start,window=5,interval=1), c(1,6))
  expect_equal(window_IT(pulse_mid,window=5,interval=1), c(5,10))
  expect_equal(window_IT(pulse_end,window=5,interval=1), c(15,20))
  # partial exposure
  expect_equal(window_IT(pulse_start,window=10,interval=1), c(1,11))
  expect_equal(window_IT(pulse_mid,window=10,interval=1), c(1,11))
  expect_equal(window_IT(pulse_end,window=10,interval=1), c(10,20))
})

test_that("window finding DEB", {
  tol = 0.1 # relative error of 10%, large tolerance because DEBtool results are uncertain
  ## results for Americamysis bahia
  # no exposure
  expect_equal(effect(set_exposure(americamysis,data.frame(t=1:28,c=0)),ep_only=TRUE),
               c(L=0,R=0),
               tolerance=tol)
  # medium exposure
  expect_equal(effect(americamysis,ep_only=TRUE),
               c(L=1-.89151/1.0041,R=1-5.0677/9.4513),
               tolerance=tol)
  # high exposure
  expect_equal(effect(set_exposure(americamysis,data.frame(t=1:28,c=3.46)),ep_only=TRUE),
               c(L=1-0.60046/1.0041,R=1),
               tolerance=tol)
  # effects for all windows
  # provide controls as argument

})

test_that("window finding Lemna", {
  metsulfuron %>% set_window(5,1) -> lemna
  conc <- 1
  pulse_start <- data.frame(t=1:20,c=c(rep(conc,5),rep(0,15)))
  pulse_mid <- data.frame(t=1:20,c=c(rep(0,5),rep(conc,5),rep(0,10)))
  pulse_end <- data.frame(t=1:20,c=c(rep(0,14),rep(conc,5),0))

  effect(set_exposure(lemna,pulse_start)) -> estr
  effect(set_exposure(lemna,pulse_mid)) -> emid
  effect(set_exposure(lemna,pulse_end)) -> eend

  # time of peak effect
  expect_equal(estr$BM.dat.start, 1)
  expect_equal(estr$BM.dat.end, 6)
  expect_equal(emid$BM.dat.start, 6)
  expect_equal(emid$BM.dat.end, 11)
  expect_equal(eend$BM.dat.start, 15)
  expect_equal(eend$BM.dat.end, 20)

  # absolute value of effect
  expect_gt(estr$BM, 0.19)
  expect_equal(estr$BM, emid$BM, tolerance=1e-5)
  expect_equal(estr$BM, eend$BM, tolerance=1e-5)
})

test_that("Lemna effects", {
  # no effects
  metsulfuron %>%
    set_exposure(data.frame(t=0:14,c=0)) -> sc
  expect_equal(sc %>% effect(ep_only=TRUE), c("BM"=0,"r"=0))

  # check that r does not exceed 1.0 in case of negative growth rates
  metsulfuron %>%
    set_exposure(data.frame(t=0:14, c=1000)) -> sc
  sc %>% simulate() -> out
  # check that we indeed have negative growth
  expect_true(out$BM[1] > tail(out$BM,1))
  # r effect endpoint should be greater than 1.0
  sc %>% effect() -> efx
  expect_gt(efx$r[1], 1)
  # all endpoints contained in output?
  expect_true(all(c("BM","r") %in% names(efx)))
})

test_that("Magma", {
  Magma() %>%
    set_init(c(BM=1)) %>%
    set_exposure(metsulfuron@exposure@series) %>%
    set_param(c(EC50_int=0.5,b=0.2,P=0.01,r_DW_TSL=1)) %>%
    set_endpoints(c("BM","r")) %>%
    effect() -> efx
  # all endpoints contained in output?
  expect_true(all(c("BM","r") %in% names(efx)))
})

test_that("general arguments", {
  metsulfuron %>% set_window(7,1) -> lemna

  # results for all relevant windows
  effect(lemna, max_only=FALSE, hmax=1, method="ode45") -> ep
  expect_equal(ep$dat.start,    0:7)
  expect_equal(ep$dat.end,      7:14)
  expect_true(all(ep$BM[1:7] > 0))
  expect_true(ep$BM[8] == 0)

  effect(lemna, hmax=1) -> epmax
  expect_equal(epmax$BM[1], max(ep$BM), tolerance=1e-5)

  # factor applied to exposure series
  factor <- 3.21
  lemna2 <- lemna
  lemna2@exposure@series[,2] <- lemna2@exposure@series[,2]*factor
  expect_equal(effect(lemna, factor=factor), effect(lemna2), tolerance=1e-5)
})

test_that("invalid arguments", {
  ## invalid inputs
  # multiple scenarios
  expect_error(effect(c(americamysis, americamysis)))
  expect_error(effect(list(minnow_it, minnow_sd)))
})

test_that("GUTS-RED sequence", {
  # continuous sim
  simulate(minnow_it, 0:4, hmax=.1, method="ode45") -> out.orig
  class(out.orig) <- c("cvasi_simulate", "data.frame")

  # sequential sim
  list(minnow_it %>% set_times(c(0,1,1.5)),
       minnow_it %>% set_times(c(1.5,2:4))) -> seq
  simulate(sequence(seq),hmax=0.1,method="ode45") -> out.ss
  # remove intermediate step at t=1.5
  out.ss <- out.ss[-3,]

  # results should be exactly the same when using a Runge-Kutta scheme
  expect_equal(out.ss[,1], out.orig[,1])
  expect_equal(out.ss[,2], out.orig[,2])
  expect_equal(out.ss[,3], out.orig[,3])
})

test_that("Lemna sequence", {
  tol=1e-5

  ##
  ## transfer occurs during simulation of a scenario
  metsulfuron %>%
    set_exposure(data.frame(t=0:14,c=0)) %>%
    set_transfer(interval=7, biomass=50) -> lemna
  simulate(lemna, method="lsoda") -> out.orig
  class(out.orig) <- c("cvasi_simulate", "data.frame")

  suppressWarnings(simulate(sequence(list(lemna)), method="lsoda") -> out.ss)
  expect_equal(out.ss, out.orig)

  ##
  ## transfer occurs between sequences
  list(lemna %>% set_times(0:7),
       lemna %>% set_times(7:14)) -> seq
  simulate(sequence(seq), method="lsoda") -> out.ss
  expect_equal(out.ss, out.orig)

  ##
  ## deactivate the effect of any residues in the 2nd scenario
  lemna %>%
    set_noexposure() %>%
    set_times(0:7) %>%
    simulate(method="lsoda") -> out.noex

  list(lemna %>% set_times(0:7),
       lemna %>% set_param(c(Emax=0)) %>% set_times(7:14)) -> seq
  simulate(sequence(seq), method="lsoda") -> out.ss

  # 1st scenario should give the same results as before
  expect_equal(out.ss[0:7,], out.orig[0:7,])
  # 2nd scenario should show uninhibited growth
  expect_equal(out.ss[9:15,"BM"], out.noex[2:8,"BM"], tolerance=tol)
  expect_equal(out.ss[9:15,"E"], out.noex[2:8,"E"], tolerance=tol)
  expect_equal(out.ss[9:15,"FrondNo"], out.noex[2:8,"FrondNo"], tolerance=tol)
  # but internal mass should still behave the same as before
  expect_equal(out.ss[9:15,"M_int"], out.orig[9:15,"M_int"], tolerance=tol)

  ## transfer in irregular intervals
  metsulfuron %>% set_transfer(times=c(5,10,12), biomass=50) -> lemna
  simulate(lemna, method="lsoda") -> out.orig

  # transfers during each sequence
  list(lemna %>% set_transfer(times=5) %>% set_times(0:7),
       lemna %>% set_transfer(times=c(10,12)) %>% set_times(7:14)) -> seq
  simulate(sequence(seq), method="lsoda") -> out.ss
  expect_equal(out.ss, out.orig, tolerance=tol, ignore_attr=TRUE)

  # transfer occurring at end of a sequence
  list(lemna %>% set_transfer(times=5) %>% set_times(0:5),
       lemna %>% set_transfer(times=c(10,12)) %>% set_times(5:14)) -> seq
  simulate(sequence(seq), method="lsoda") -> out.ss
  expect_equal(out.ss, out.orig, tolerance=tol, ignore_attr=TRUE)

})

test_that("output times", {
  # break included in output times
  sc <- minnow_it %>% set_times(0:6)
  sq <- sequence(list(sc, sc), breaks=3)
  rs1 <- simulate(sq)
  expect_equal(rs1$time, 0:6)

  # break included multiple times
  tms <- c(0:3, 3, 4:6)
  sc <- minnow_it %>% set_times(tms)
  sq <- sequence(list(sc, sc), breaks=3)
  rs1 <- simulate(sq)
  expect_equal(rs1$time, tms)

  # break not included
  sq <- sequence(list(sc, sc), breaks=3.1)
  rs2 <- simulate(sq)
  expect_equal(rs2$time, tms)
  expect_equal(rs2, rs1, tolerance=1e-5, ignore_attr=TRUE)
})

test_that("included outputs", {
  # single break
  sc1 <- minnow_it %>% set_times(0:3)
  sc2 <- minnow_it %>% set_times(3:6)
  sq <- sequence(list(sc1, sc2))
  rs <- simulate(sq)
  expect_equal(rs$time, 0:6)

  sq@inc_start <- c(TRUE, FALSE)
  sq@inc_end <- c(FALSE, FALSE)
  rs <- simulate(sq)
  expect_equal(rs$time, c(0, 1, 2, 4, 5))


  sq@inc_start <- c(TRUE, TRUE)
  sq@inc_end <- c(FALSE, FALSE)
  rs <- simulate(sq)
  expect_equal(rs$time, c(0, 1, 2, 3, 4, 5))

  sq@inc_start <- c(FALSE, FALSE)
  sq@inc_end <-  c(TRUE, TRUE)
  rs <- simulate(sq)
  expect_equal(rs$time, c(1, 2, 3, 4, 5, 6))

  # multiple breaks
  sc <- minnow_it %>% set_times(0:6)
  sq <- sequence(list(sc, sc, sc), breaks=c(2.1, 4))
  expect_equal(sq@inc_start, c(TRUE, FALSE, FALSE))
  expect_equal(sq@inc_end, c(FALSE, TRUE, TRUE))
  rs <- simulate(sq)
  expect_equal(rs$time, 0:6)

  sq@inc_start <- c(TRUE, TRUE, TRUE)
  rs <- simulate(sq)
  expect_equal(rs$time, c(0, 1, 2, 2.1, 3, 4, 4, 5, 6))

  # 1st scenario without output
  sc <- minnow_it %>% set_times(0:6)
  sq <- sequence(list(sc, sc), breaks=3) %>% set_times(3:6)
  rs <- simulate(sq)
  expect_equal(rs$time, 3:6)

  # 2ns scenario without output
  sc <- minnow_it %>% set_times(0:6)
  sq <- sequence(list(sc, sc), breaks=3) %>% set_times(0:3)
  rs <- simulate(sq)
  expect_equal(rs$time, 0:3)

  sq <- sq %>% set_times(0:2)
  rs <- simulate(sq)
  expect_equal(rs$time, 0:2)
})

test_that("simulate", {
  sc <- americamysis
  sc@param["p_M"] <- NULL
  expect_error(simulate(sc), "missing parameter")

  # optional output variables
  rs1 <- simulate(americamysis %>% set_nowindow)
})

test_that("effect", {
  sc <- americamysis %>% set_window(length=-1)
  ctrl <- sc %>% set_noexposure() %>% simulate()
  t1 <- sc %>% simulate()

  myeffect <- 1 - tail(t1$L, n=1)/tail(ctrl$L, n=1)
  expect_equal(effect(sc)$L[1], myeffect, tolerance=1e-5)
})

test_that("regular intervals, starting and ending on transfer", {
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_times(0:10) %>%
    set_noexposure() %>%
    set_transfer(interval=5, biomass=1)

  ref <- sc %>%
    set_notransfer() %>%
    set_times(0:5) %>%
    simulate()

  rs <- simulate(sc)
  # first batch must equal to ref sim
  expect_equal(rs[1:6, ], ref, tolerance=1e-6, ignore_attr=TRUE)
  # second batch must also equal to ref sim, apart from the timestamp and sim start (t==0)
  expect_equal(rs[7:11, -1], ref[2:6, -1], tolerance=1e-6, ignore_attr=TRUE)

  # negative starting time
  rs <- sc %>%
    set_times(-5:5) %>%
    simulate()
  expect_equal(rs[1:6, -1], ref[, -1], tolerance=1e-6, ignore_attr=TRUE)
  expect_equal(rs[7:11, -1], ref[2:6, -1], tolerance=1e-6, ignore_attr=TRUE)
})

test_that("regular intervals, not starting on transfer", {
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_times(1:10) %>%
    set_noexposure() %>%
    set_transfer(interval=5, biomass=1)

  ref <- sc %>%
    set_notransfer() %>%
    set_times(0:5) %>%
    simulate()

  rs <- simulate(sc)
  expect_equal(rs[1:5, -1], ref[1:5, -1], tolerance=1e-6, ignore_attr=TRUE)
  expect_equal(rs[6:10, -1], ref[2:6, -1], tolerance=1e-6, ignore_attr=TRUE)

  # negative starting time
  rs <- sc %>%
    set_times(-4:5) %>%
    simulate()
  expect_equal(rs[1:5, -1], ref[1:5, -1], tolerance=1e-6, ignore_attr=TRUE)
  expect_equal(rs[6:10, -1], ref[2:6, -1], tolerance=1e-6, ignore_attr=TRUE)
})

test_that("regular intervals, not ending on transfer", {
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_times(0:9) %>%
    set_noexposure() %>%
    set_transfer(interval=5, biomass=1)

  ref <- sc %>%
    set_notransfer() %>%
    set_times(0:5) %>%
    simulate()

  rs <- simulate(sc)
  expect_equal(rs[1:6, ], ref[1:6, ], tolerance=1e-6, ignore_attr=TRUE)
  expect_equal(rs[7:10, -1], ref[2:5, -1], tolerance=1e-6, ignore_attr=TRUE)

  # negative starting time
  rs <- sc %>%
    set_times(-5:4) %>%
    simulate()
  expect_equal(rs[1:6, -1], ref[1:6, -1], tolerance=1e-6, ignore_attr=TRUE)
  expect_equal(rs[7:10, ], ref[2:5, ], tolerance=1e-6, ignore_attr=TRUE)
})

test_that("regular intervals, no transfers relevant", {
  # starts and ends transfer
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_times(0:5) %>%
    set_noexposure() %>%
    set_transfer(interval=5, biomass=1)
  rs <- simulate(sc)
  ref <- sc %>%
    set_notransfer() %>%
    simulate()

  expect_equal(rs, ref, tolerance=1e-6)

  # negative starting time
  rs <- sc %>%
    set_times(-5:0) %>%
    simulate()
  expect_equal(rs[, -1], ref[, -1], tolerance=1e-6, ignore_attr=TRUE)

  # starts and ends somewhere in between transfers
  sc <- sc %>% set_times(1:4)
  rs <- sc %>% simulate()
  ref <- sc %>%
    set_notransfer() %>%
    simulate()

  expect_equal(rs, ref, tolerance=1e-6)
})


test_that("regular intervals with redundant time-points", {
  times <- c(0, 0, 1:5, 5, 6, 6, 7:10)

  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_noexposure() %>%
    set_times(times) %>%
    set_transfer(interval=5, biomass=1)
  rs <- simulate(sc)

  ref <- sc %>%
    set_notransfer() %>%
    set_times(times[times <= 5]) %>%
    simulate()

  expect_equal(rs$time, times)
  expect_equal(rs[1:8, ], ref, tolerance=1e-6)
  expect_equal(rs[9:14, -1], ref[c(3, 3:7), -1], tolerance=1e-6, ignore_attr=TRUE)
})

test_that("transfer not in output times", {
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_noexposure() %>%
    set_times(c(1, 10)) %>%
    set_transfer(interval=5, biomass=1)
  rs <- simulate(sc)

  ref <- sc %>%
    set_notransfer() %>%
    set_times(0:5) %>%
    simulate()

  expect_equal(rs[, 1], c(1, 10))
  expect_equal(rs[1, -1], ref[1, -1], tolerance=1e-6, ignore_attr=TRUE)
  expect_equal(rs[2, -1], ref[6, -1], tolerance=1e-6, ignore_attr=TRUE)
})

test_that("custom time points", {
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_noexposure() %>%
    set_transfer(times=c(3, 8), biomass=1)
  rs <- simulate(sc)
  ref <- sc %>%
    set_notransfer() %>%
    simulate()

  expect_equal(rs[1:4, -1], ref[1:4, -1], tolerance=1e-6) # first batch
  expect_equal(rs[5:9, -1], ref[2:6, -1], tolerance=1e-6, ignore_attr=TRUE) # second batch
  expect_equal(rs[10:15, -1], ref[2:7, -1], tolerance=1e-6, ignore_attr=TRUE) # third batch
})

test_that("custom biomass", {
  metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_noexposure() %>%
    set_transfer(times=c(2,4,6,8,10,12), biomass=c(1,2,1,2,1,2)) %>%
    simulate() -> rs

  expect_equal(rs[1:3,-1], rs[1:3,-1], tolerance=1e-6, ignore_attr=T)
  expect_equal(rs[2:3,-1], rs[4:5,-1], tolerance=1e-6, ignore_attr=T)
  expect_equal(rs[2:3,-1], rs[8:9,-1], tolerance=1e-6, ignore_attr=T)
  expect_equal(rs[2:3,-1], rs[12:13,-1], tolerance=1e-6, ignore_attr=T)
  expect_equal(rs[6:7,-1], rs[10:11,-1], tolerance=1e-6, ignore_attr=T)
  expect_gt(rs$BM[6], rs$BM[4]+1)
  expect_gt(rs$BM[7], rs$BM[5]+1)
})

test_that("non-standard biomass compartment name", {
  sc <- metsulfuron
  sc@transfer.comp.biomass <- "Foo"
  sc@init <- c(Foo=1, E=1, M_int=0)
  sc %>%
    set_transfer(times=c(2,3,6), biomass=c(1,1,1)) %>%
    simulate() -> rs

  expect_contains(names(rs), "Foo")
})

# if a scenario cannot be simulated entirely, the solver may return NaN values. the
# function should raise an informative warning/message if this occurs.
test_that("handle numerical instability", {
  sc <- metsulfuron %>%
    set_init(c(BM=1)) %>%
    set_times(0:5) %>%
    set_transfer(interval=3, biomass=1)
  # modify exposure series so that we will have invalid state variables from a
  # certain point onwards
  sc@forcings$temp <- data.frame(time=c(0, 2, 2.1, 5), conc=c(12, 12, -1e10, 12))
  expect_warning(simulate(sc), "contains NaN")
})

#test_that("multiple compartments", {
#
#})

#' Import `morse` model parameters
#'
#' Loads GUTS model parameters which were fitted by the morse package.
#'
#' @param fit Either a string with a file path to an *.Rdata* or *.RDS* file
#'   containing a *morse* fit, or a *morse* fit object itself
#' @param find_sd a logical value. If `TRUE`, it will try to find fitted parameters of a *GUTS-RED-SD* model
#' @param find_it  a logical value. If `TRUE`, it will try to find fitted parameters of a *GUTS-RED-IT* model
#' @param reset_hb a logical value. If `TRUE`, the background hazard rate `hb` is set to zero
#' @param params `character`, if set to `"estim"` then only the best-fit parameters are
#'  imported, else all parameter sets in the MCM chains are returned
#' @param mcmc_size optional `integer`, sets the maximum number of imported parameter sets per MCMC.
#'   By default, all MSMS parameter samples are imported.
#' @param file *deprecated*`,` alias for parameter `fit`
#' @param find.SD *deprecated*`,` alias for parameter `find_sd`
#' @param find.IT *deprecated*`,` alias for parameter `find_it`
#' @param reset.hb *deprecated*`,` alias for parameter `rest_hb`
#' @param mcmc.size *deprecated*`,` alias for parameter `mcmc_size`
#'
#' @return list of [parameter_set] objects
#' @importFrom lifecycle is_present deprecated
#' @export
#'
#' @examples
#' # import all parameter fits
#' try(import_morse("path/to/morse_fit.RData"))
#'
#' # import parameters for a specific model
#' try(import_morse("path/to/morse_fit.RData", find_it=TRUE, find_sd=FALSE))
#'
#' # modify model objects
#' try(models %>% set_param(import_morse("path/to/morse_fit.RData")))
import_morse <- function(fit, find_sd=TRUE, find_it=TRUE, reset_hb=FALSE,
                         params=c("estim", "all"), mcmc_size, find.SD=deprecated(),
                         find.IT=deprecated(), reset.hb=deprecated(), mcmc.size=deprecated(),
                         file=deprecated()) {
  if(is_present(file)) {
    lifecycle::deprecate_soft("1.3.0", "import_morse(file)", "import_morse(fit)")
    fit <- file
  }
  if(is_present(find.SD)) {
    lifecycle::deprecate_soft("1.3.0", "import_morse(find.SD)", "import_morse(find_sd)")
    find_sd <- find.SD
  }
  if(is_present(find.IT)) {
    lifecycle::deprecate_soft("1.3.0", "import_morse(find.IT)", "import_morse(find_it)")
    find_it <- find.IT
  }
  if(is_present(reset.hb)) {
    lifecycle::deprecate_soft("1.3.0", "import_morse(reset.hb)", "import_morse(reset_hb)")
    reset_hb <- reset.hb
  }
  if(is_present(mcmc.size)) {
    lifecycle::deprecate_soft("1.3.0", "import_morse(mcmc.size)", "import_morse(mcmc_size)")
    mcmc_size <- mcmc.size
  }

  if(missing(mcmc_size))
    mcmc_size <- -1 # load all by default
  params <- match.arg(params)

  fits <- list()
  # is fit a path to a data file?
  if(is.character(fit))
  {
    log_msg("\n++ Loading parameters from morse object file")
    log_msg("file: ", fit)
    if(!file.exists(fit))
      stop("morse object file does not exist")

    if(endsWith(tolower(fit), ".rdata")) {
      ex <- new.env()
      suppressWarnings(load(fit, envir=ex))
      fit <- ex
    } else if(endsWith(tolower(fit), ".rds")) {
      fit <- list(survfit=suppressWarnings(readRDS(fit)))
    }
  }
  else if(is(fit, "survFit")) {
    fit <- list(survfit=fit)
  } else {
    stop("unsupported type: ", is(fit))
  }

  # try to find fitted parameters automatically
  if(find_sd | find_it) {
    for(nm in names(fit)) {
      var <- fit[[nm]]
      if(is(var, "survFit")) {
        if(find_it & "alpha" %in% var$estim.par$parameters)
          fits[["GUTS-RED-IT"]] <- nm
        if(find_sd & "kk" %in% var$estim.par$parameters)
          fits[["GUTS-RED-SD"]] <- nm
      }
    }

    if(find_it & !("GUTS-RED-IT" %in% names(fits))) {
      warning("Could not find GUTS-RED-IT fitted parameters")
    }
    if(find_sd & !("GUTS-RED-SD" %in% names(fits))) {
      warning("Could not find GUTS-RED-SD fitted parameters")
    }
    if(length(fits) == 0)
      stop("Aborting import")
  }

  # extract parameter sets
  ps <- c()
  for(nm in names(fits)) {
    morse_fit <- fit[[fits[[nm]]]]

    # only load best fit parameters
    if(params == "estim") {
      p <- morse_fit$estim.par$median
      names(p) <- morse_fit$estim.par$parameters
      if(reset_hb) {
        p["hb"] <- 0 # reset background hazard rate to zero
      }

      log_msgf(nm, ": best-fit imported")
      log_msg("  ",paste(names(p), unlist(p), sep="=", collapse=", "))
      ps <- c(ps, parameter_set(model=nm, param=as.list(p)))
    }
    # load all MCMC shots
    else {
      # import all
      if(mcmc_size <= 0)
        do.call("rbind", morse_fit$mcmc) -> morse_all
      else { # select a certain number of rows from each MCM chain
        morse_all <- data.frame()
        for(i in seq(length(morse_fit$mcmc)))
          morse_all <- rbind(morse_all, morse_fit$mcmc[[i]][seq(mcmc_size), ])
      }

      # rename and convert parameter values
      morse_all[, c(paste0(morse_fit$estim.par$parameters, "_log10"))] %>%
        tibble::as_tibble() %>%
        dplyr::mutate_all(function(x) 10^x) %>%
        dplyr::rename_with(.fn=function(x) sub("_log10", "", x)) -> morse_all
      if(reset_hb) {
        morse_all$hb <- c(0)
      }

      log_msgf(paste0(nm, ": MCMC samples imported (n=", nrow(morse_all), ")"))
      for(i in seq(nrow(morse_all))) {
        ps <- c(ps, parameter_set(model=nm, param=as.list(morse_all[i, ])))
        #log_msg("  ",paste(names(morse.all),unlist(morse.all[i,]),sep="=",collapse=", "))
      }
    }
  }
  return(ps)
}

# for backwards compatibility
#' @param ... Arguments passed on to `import_morse()`
#' @describeIn import_morse deprecated alias
#' @export
morse <- function(...) {
  lifecycle::deprecate_soft("1.3.0", "morse()", "import_morse()")
  import_morse(...)
}

test_that("import from RDS", {
  fp <- test_path("../data/morse/fit_sd.rds")
  fit <- suppressMessages(import_morse(fp, find_sd=TRUE, find_it=FALSE,
                                       param="estim", reset_hb=FALSE))

  expect_equal(length(fit), 1)
  fit <- fit[[1]]
  expect_true(is_parameter_set(fit))
  expect_equal(fit@model, "GUTS-RED-SD")
  expect_true(fit@param$kd > 0)
  expect_true(fit@param$hb > 0)
  expect_true(fit@param$z > 0)
  expect_true(fit@param$kk > 0)


  fp <- test_path("../data/morse/fit_it.rds")
  fit <- suppressMessages(import_morse(fp, find_sd=FALSE, find_it=TRUE,
                                       param="estim", reset_hb=FALSE))

  expect_equal(length(fit), 1)
  fit <- fit[[1]]
  expect_true(is_parameter_set(fit))
  expect_equal(fit@model, "GUTS-RED-IT")
  expect_true(fit@param$kd > 0)
  expect_true(fit@param$hb > 0)
  expect_true(fit@param$alpha > 0)
  expect_true(fit@param$beta > 0)
})

test_that("import from RData", {
  fp <- test_path("../data/morse/fit_sd.rdata")
  fit <- suppressMessages(import_morse(fp, find_sd=TRUE, find_it=FALSE,
                                       param="estim", reset_hb=FALSE))

  expect_equal(length(fit), 1)
  fit <- fit[[1]]
  expect_true(is_parameter_set(fit))
  expect_equal(fit@model, "GUTS-RED-SD")
  expect_true(fit@param$kd > 0)
  expect_true(fit@param$hb > 0)
  expect_true(fit@param$z > 0)
  expect_true(fit@param$kk > 0)
})

test_that("import from object", {
  fp <- test_path("../data/morse/fit_sd.rds")
  mfit <- readRDS(fp)
  fit <- suppressMessages(import_morse(mfit, find_sd=TRUE, find_it=FALSE,
                                       param="estim", reset_hb=FALSE))

  expect_equal(length(fit), 1)
  fit <- fit[[1]]
  expect_true(is_parameter_set(fit))
  expect_equal(fit@model, "GUTS-RED-SD")
  expect_true(fit@param$kd > 0)
  expect_true(fit@param$hb > 0)
  expect_true(fit@param$z > 0)
  expect_true(fit@param$kk > 0)
})

test_that("reset hb", {
  fp <- test_path("../data/morse/fit_sd.rds")
  fit <- suppressMessages(import_morse(fp, find_sd=TRUE, find_it=FALSE,
                                       param="estim", reset_hb=FALSE))
  fit <- fit[[1]]
  expect_true(fit@param$hb > 0)

  fit <- suppressMessages(import_morse(fp, find_sd=TRUE, find_it=FALSE,
                                       param="estim", reset_hb=TRUE))
  fit <- fit[[1]]
  expect_true(fit@param$hb == 0)
})

test_that("import MCMC samples", {
  fp <- test_path("../data/morse/fit_sd.rds")
  fit <- suppressMessages(import_morse(fp, find_sd=TRUE, find_it=FALSE,
                                       param="all", reset_hb=FALSE))
  expect_equal(length(fit), 150)
  expect_true(all(sapply(fit, slot, name="model") == "GUTS-RED-SD"))

  fit <- fit[[1]]
  expect_true(fit@param$kd > 0)
  expect_true(fit@param$hb > 0)
  expect_true(fit@param$z > 0)
  expect_true(fit@param$kk > 0)

  fit <- suppressMessages(import_morse(fp, find_sd=TRUE, find_it=FALSE,
                                       param="all", mcmc_size=10))
  expect_equal(length(fit), 30)
  expect_true(all(sapply(fit, slot, name="model") == "GUTS-RED-SD"))
})

test_that("deprecated parameters", {
  fp <- test_path("../data/morse/fit_sd.rds")
  expect_warning(suppressMessages(import_morse(fp, find.IT=FALSE)))
  expect_warning(suppressMessages(import_morse(file=fp, find_it=FALSE)))
  expect_warning(suppressMessages(import_morse(fp, find_it=FALSE, reset.hb=TRUE)))
  expect_warning(suppressMessages(import_morse(fp, find_it=FALSE, mcmc.size=1)))

  fp <- test_path("../data/morse/fit_it.rds")
  expect_warning(suppressMessages(import_morse(fp, find.SD=FALSE)))
})

test_that("invalid arguments", {
  expect_error(import_morse("foo"))
  expect_error(import_morse(list()))
})

#
# The code below contains an alternative model implementation of the GUTS-RED-IT
# and GUTS-RED-SD models which was published with the `morse` package, version
# 3.2.7 and before. The `morse` implementation is used as a means to compare model
# results numerically.
#


# wrapper functions extract LPx values from morse's return values in a format
# that is compatible with this package
morse_sd <- function(scenario) {
  param <- scenario@param
  exp <- scenario@exposure@series

  rs <- SurvSD_ode(
    Cw=exp[, 2],
    time=exp[, 1],
    kk=param[["kk"]],
    kd=param[["kd"]],
    z=param[["z"]],
    hb=param[["hb"]]
  ) %>%
    dplyr::select(time, D, H, S=dtheo)
  class(rs) <- c("cvasi_simulate", "data.frame")
  rs
}

morse_it <- function(scenario) {
  param <- scenario@param
  exp <- scenario@exposure@series

  rs <- SurvIT_ode(
    Cw=exp[, 2],
    time=exp[, 1],
    kd=param[["kd"]],
    hb=param[["hb"]],
    alpha=param[["alpha"]],
    beta=param[["beta"]]
  ) %>%
    dplyr::mutate(H=0) %>%
    dplyr::select(time, D, H, S=dtheo)
  class(rs) <- c("cvasi_simulate", "data.frame")
  rs
}

# Code below taken from morse package, version 3.2.7 (2020-11-11)
# https://github.com/pveber/morse/blob/c1637ee7811d4369942d9ac479fa681285f0ae5f/R/predict_ode.survFit.R
#
# Some parts have been commented out because they are not needed for our
# use case.
SurvSD_ode <- function(Cw, time, replicate=NULL, kk, kd, z, hb, mcmc_size = 1, interpolate_length = NULL, interpolate_method = "linear"){

  ## external signal with several rectangle impulses
  signal <- data.frame(times = time,
                       import = Cw)

  sigimp <- stats::approxfun(signal$times, signal$import, method = interpolate_method, rule = 2)

  if(!is.null(interpolate_length)){
    times <- seq(min(time), max(time), length = interpolate_length)
  } else{
    times <- signal$times
  }

  ## The parameters
  parms  <- list( kd = kd,
                  kk = kk,
                  z = z,
                  hb = hb,
                  mcmc_size = mcmc_size)

  ## Start values for steady state
  xstart <- c(D = rep(0, mcmc_size),
              H = rep(0, mcmc_size))

  ## Solve model
  out <- ode(y = xstart,
             times = times,
             func = model_SD,
             parms,
             input = sigimp)

  dtheo <- exp(- out[, grep("H", colnames(out))] )

  # Manage vector case
  #if(mcmc_size == 1){
  #  q50 = dtheo
  #  qinf95 = dtheo
  #  qsup95 = dtheo
  #} else{
  #  q50 = apply(dtheo, 1, quantile, probs = 0.5, na.rm = TRUE)
  #  qinf95 = apply(dtheo, 1, quantile, probs = 0.025, na.rm = TRUE)
  #  qsup95 = apply(dtheo, 1, quantile, probs = 0.975, na.rm = TRUE)
  #}

  #dtheo <- as.data.frame(dtheo) %>%
  #  mutate(time = out[, "time"],
  #         conc = out[, "signal"]
  #         replicate = rep(replicate, nrow(out)),
  #         q50 = q50,
  #         qinf95 = qinf95,
  #         qsup95 = qsup95
  #         )

  out <- as.data.frame(out)
  out$dtheo <- dtheo

  return(out)
}

model_SD <- function(t, State, parms, input)  {
  with(as.list(c(parms, State)), {

    conc_ext = input(t)

    D = State[1:mcmc_size]
    H = State[(mcmc_size+1):(2*mcmc_size)]

    dD <- kd * (conc_ext - D)     # internal concentration
    dH <- kk * pmax(D - z, 0) + hb # risk function

    res <- c(dD, dH)
    list(res, signal = conc_ext)
  })
}


# Survival function for "IT" model with external concentration changing with time
#
# @param Cw A scalar of external concentration
# @param time A vector of time
# @param kk a vector of parameter
# @param kd a vector of parameter
# @param z a vector of parameter
# @param hb a vector of parameter
#
#
# @return A matrix generate with coda.samples() function
#

SurvIT_ode <- function(Cw, time, replicate=NULL, kd, hb, alpha, beta, mcmc_size = 1, interpolate_length = NULL, interpolate_method = "linear"){

  ## external signal with several rectangle impulses
  signal <- data.frame(times = time,
                       import = Cw)

  sigimp <- stats::approxfun(signal$times, signal$import, method = interpolate_method, rule = 2)

  if(!is.null(interpolate_length)){
    times <- seq(min(time), max(time), length = interpolate_length)
  } else{
    times <- signal$times
  }

  ## The parameters
  parms  <- list( kd = kd,
                  alpha = alpha,
                  beta = beta,
                  mcmc_size = mcmc_size)

  ## Start values for steady state
  xstart <- c(D = rep(0, mcmc_size))

  ## Solve model
  out <- ode(y = xstart,
             times = times,
             func = model_IT,
             parms,
             input = sigimp)

  D <- out[, grep("D", colnames(out))]
  cumMax_D <- if(is.null(dim(D))) cummax(D) else apply(D, 2, cummax)
  thresholdIT <- t(1 / (1 + (t(cumMax_D) / parms$alpha)^(-parms$beta)))

  dtheo <- (1 - thresholdIT) * exp(times %*% t(-hb))

  # Manage vector case
  #if(mcmc_size == 1){
  #  q50 = dtheo
  #  qinf95 = dtheo
  #  qsup95 = dtheo
  #} else{
  #  q50 = apply(dtheo, 1, quantile, probs = 0.5, na.rm = TRUE)
  #  qinf95 = apply(dtheo, 1, quantile, probs = 0.025, na.rm = TRUE)
  #  qsup95 = apply(dtheo, 1, quantile, probs = 0.975, na.rm = TRUE)
  #}

  #dtheo <- as.data.frame(dtheo) %>%
  #  mutate(time = out[, "time"],
  #         conc = out[, "signal"]
  #         replicate = rep(replicate, nrow(out)),
  #         q50 = q50,
  #         qinf95 = qinf95,
  #         qsup95 = qsup95
  #         )

  out <- as.data.frame(out)
  out$dtheo <- as.numeric(dtheo)

  return(out)
}

model_IT <- function(t, State, parms, input) {
  with(as.list(c(parms, State)), {
    conc_ext <- input(t)

    D = State[1:mcmc_size]

    dD <- kd*(conc_ext - D)    # internal damage

    list(dD = dD, signal = conc_ext)
  })
}

################
##
## Test of model implementation
##
################

# GUTS model verification as conducted in
# EFSA Scientific Opinion on TKTD models, pp. 36, doi:10.2903/j.efsa.2018.5377
test_that("EFSA classical exposure scenario", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=seq(0,7,0.01),
                    conc=c(rep(5,401), rep(0,300)))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) -> sd
  res_cvasi <- simulate(sd)
  res_morse <- morse_sd(sd)
  # test that results are equal using a sensible numerical precision
  expect_equal(res_cvasi, res_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) -> it
  res_cvasi <- simulate(it)
  res_morse <- morse_it(it)
  # test that results are equal using a sensible numerical precision
  expect_equal(res_cvasi, res_morse, tolerance = 1e-3, ignore_attr=TRUE)
})

# GUTS model verification as conducted in
# EFSA Scientific Opinion on TKTD models, pp. 36, doi:10.2903/j.efsa.2018.5377
test_that("EFSA pulsed scenarios", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # repeat tests for (selected) exposure multiplication factors 1..50
  for(MF in c(1,2,5,10,25,50))
  {
    # Code and parameters from EFSA (2018), Appendix E
    # > 01.GUTS-implementation-check.R
    exp <- data.frame(time=seq(0,14,0.01),
                      conc=c(rep(0,301), rep(5 * MF,200), rep(0,900)))

    kD <- 0.3 # unit: [time^-1]
    hb <- 0 # background mortality rate [time^-1]
    bw <- 0.5 # unit: 1/[D]
    zw <- 2.5 # unit: [D]
    mw <- 2.5 # unit: [D]
    beta <- 2 # dimensionless

    # test GUST-RED-SD scenario
    GUTS_RED_SD() %>%
      set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
      set_exposure(exp) -> sd
    res_cvasi <- simulate(sd)
    res_morse <- morse_sd(sd)
    # test that results are equal using a sensible numerical precision
    expect_equal(res_cvasi, res_morse, tolerance = 1e-3, ignore_attr=TRUE)

    # test GUST-RED-IT scenario
    GUTS_RED_IT() %>%
      set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
      set_exposure(exp) -> it
    res_cvasi <- simulate(it)
    res_morse <- morse_it(it)
    # test that results are equal using a sensible numerical precision
    expect_equal(res_cvasi, res_morse, tolerance = 1e-3, ignore_attr=TRUE)
  }
})

# GUTS model verification as conducted in
# EFSA Scientific Opinion on TKTD models, pp. 36
# doi:10.2903/j.efsa.2018.5377
test_that("EFSA extreme scenarios: zero exposure, no background mortality", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=seq(0,7,0.01), conc=rep(0,701))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) -> sd
  sd_cvasi <- simulate(sd)
  sd_morse <- morse_sd(sd)
  # test that results are equal using a sensible numerical precision
  expect_equal(sd_cvasi, sd_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) -> it
  it_cvasi <- simulate(it)
  it_morse <- morse_it(it)
  # test that results are equal using a sensible numerical precision
  expect_equal(it_cvasi, it_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # internal concentrations must be identical in both models
  expect_equal(sd_cvasi$D, it_cvasi$D, tolerance = 0.001, ignore_attr=TRUE)
})

# GUTS model verification as conducted in
# EFSA Scientific Opinion on TKTD models, pp. 36
# doi:10.2903/j.efsa.2018.5377
test_that("EFSA extreme scenarios: high exposure over days 1-4", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=seq(0,7,0.01),
                    conc=c(rep(100,401), rep(0,300)))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) -> sd
  sd_cvasi <- simulate(sd)
  sd_morse <- morse_sd(sd)
  # test that results are equal using a sensible numerical precision
  expect_equal(sd_cvasi, sd_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) -> it
  it_cvasi <- simulate(it)
  it_morse <- morse_it(it)
  # test that results are equal using a sensible numerical precision
  expect_equal(it_cvasi, it_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # internal concentrations must be identical in both models
  expect_equal(sd_cvasi$D, it_cvasi$D, tolerance = 0.001, ignore_attr=TRUE)
})

# GUTS model verification as conducted in
# EFSA Scientific Opinion on TKTD models, pp. 36
# doi:10.2903/j.efsa.2018.5377
test_that("EFSA extreme scenarios: zero exposure with slight background mortality", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=seq(0,7,0.01), conc=rep(0,701))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0.05 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) -> sd
  sd_cvasi <- simulate(sd)
  sd_morse <- morse_sd(sd)
  # test that results are equal using a sensible numerical precision
  expect_equal(sd_cvasi, sd_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) -> it
  it_cvasi <- simulate(it)
  it_morse <- morse_it(it)

  # remove column H (hazard) from results because it is not explicitly calculated
  # by the morse model implementation, therefore we cannot check it
  it_cvasi <- dplyr::select(it_cvasi, !H)
  it_morse <- dplyr::select(it_morse, !H)

  # test that results are equal using a sensible numerical precision
  expect_equal(it_cvasi, it_morse, tolerance = 1e-3, ignore_attr=TRUE)

  # internal concentrations must be identical in both models
  expect_equal(sd_cvasi$D, it_cvasi$D, tolerance = 0.001, ignore_attr=TRUE)
})

################
##
## Snapshots which provide a constant baseline for model outputs.
## The results have been compared manually with the graphs in EFSA (2018).
##
################

test_that("snapshot: zero exposure, no bg mortality", {
  exp <- data.frame(time=seq(0,7,0.01), conc=rep(0,701))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) %>%
    simulate() -> res_sd
  #ggplot2::ggplot(res_sd) + ggplot2::geom_line(aes(time, D))
  #ggplot2::ggplot(res_sd) + ggplot2::geom_line(aes(time, S*100))
  expect_snapshot_value(res_sd, style="json2", tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) %>%
    simulate() -> res_it
  #ggplot2::ggplot(res_it) + ggplot2::geom_line(aes(time, D))
  #ggplot2::ggplot(res_it) + ggplot2::geom_line(aes(time, S*100))
  expect_snapshot_value(res_it, style="json2", tolerance = 1e-3, ignore_attr=TRUE)
})

test_that("snapshot: high exposure, no bg mortality", {
  exp <- data.frame(time=seq(0,7,0.01),
                    conc=c(rep(100,401), rep(0,300)))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) %>%
    simulate() -> res_sd
  #ggplot2::ggplot(res_sd) + ggplot2::geom_line(aes(time, D))
  #ggplot2::ggplot(res_sd) + ggplot2::geom_line(aes(time, S*100))
  expect_snapshot_value(res_sd, style="json2", tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) %>%
    simulate() -> res_it
  #ggplot2::ggplot(res_it) + ggplot2::geom_line(aes(time, D))
  #ggplot2::ggplot(res_it) + ggplot2::geom_line(aes(time, S*100))
  expect_snapshot_value(res_it, style="json2", tolerance = 1e-3, ignore_attr=TRUE)
})

test_that("snapshot: no exposure, slight bg mortality", {
  exp <- data.frame(time=seq(0,7,0.01), conc=rep(0,701))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0.05 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # test GUST-RED-SD scenario
  GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) %>%
    simulate() -> res_sd
  #ggplot2::ggplot(res_sd) + ggplot2::geom_line(aes(time, D))
  #ggplot2::ggplot(res_sd) + ggplot2::geom_line(aes(time, S*100))
  expect_snapshot_value(res_sd, style="json2", tolerance = 1e-3, ignore_attr=TRUE)

  # test GUST-RED-IT scenario
  GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp) %>%
    simulate() -> res_it
  #ggplot2::ggplot(res_it) + ggplot2::geom_line(aes(time, D))
  #ggplot2::ggplot(res_it) + ggplot2::geom_line(aes(time, S*100))
  expect_snapshot_value(res_it, style="json2", tolerance = 1e-3, ignore_attr=TRUE)
})

################
##
## Test of model sensitivity
##
################
# Instead of calculating local parameter sensitivities, each parameter is
# varied by a factor and simulation outputs are numerically compared. If
# outputs match, then it follows that the sensitivities are also identical.
#
# Exposure series is identical to the one used in file `01.GUTS-implementation-check.R`
# by EFSA (2018).
test_that("EFSA sensitivity: GUTS-RED-IT", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # parameters to modify
  for(p in c("kd", "alpha", "beta")) {
    # factors to apply to parameter values
    for(f in c(0.75, 0.8, 0.9, 1, 1.1, 1.3, 1.5, 1.75, 2)) {
      # set up basic scenario
      scenario <- GUTS_RED_IT() %>%
        set_param(c(kd=0.3, hb=0, alpha=2.5, beta=2)) %>%
        set_exposure(data.frame(time= seq(0,5,0.01), conc=c(rep(0,301), rep(15,200))))
      # modify parameter value
      scenario <- scenario %>% set_param(setNames(scenario@param[[p]] * f, p))
      # simulate and compare
      expect_equal(simulate(scenario), morse_it(scenario), tolerance=0.001, ignore_attr=TRUE)
    }
  }
})

test_that("EFSA sensitivity: GUTS-RED-SD", {
  source(test_path("morse_guts-red.R"), local=TRUE)

  # parameters to modify
  for(p in c("kd", "z", "kk")) {
    # factors to apply to parameter values
    for(f in c(0.75, 0.8, 0.9, 1, 1.1, 1.3, 1.5, 1.75, 2)) {
      # set up basic scenario
      scenario <- GUTS_RED_SD() %>%
        set_param(c(kd=0.3, hb=0, kk=0.5, z=2.5)) %>%
        set_exposure(data.frame(time= seq(0,5,0.01), conc=c(rep(0,301), rep(15,200))))
      # modify parameter value
      scenario <- scenario %>% set_param(setNames(scenario@param[[p]] * f, p))
      # simulate and compare
      expect_equal(simulate(scenario), morse_sd(scenario), tolerance=0.001, ignore_attr=TRUE)
    }
  }
})


################
##
## Test of EPx calculation
##
################
test_that("EFSA LPx propiconazole", {
  filename <- test_path("../data/EFSA_propiconazole/focus.rds")
  skip_if_not(file.exists(filename))

  # load exposure series
  exp <- readRDS(filename)

  # Helper functions to return LP50 for propiconazole, model parameters
  # listed in Table 3 of EFSA (2018), p 56. Technically, the parameters' units
  # do not fit to the unit of the exposure series (umol/L vs ug/L)
  mf_sd <- function(exp) {
    GUTS_RED_SD() %>%
      set_param(c(kd=2.179, hb=0.02683, kk=0.1237, z=16.89)) %>%
      set_exposure(exp) %>%
      epx(level=50, hmax=0.01, method="ode45") %>%
      dplyr::pull(L.EP50) %>%
      round(digits=0)
  }

  mf_it <- function(exp) {
    GUTS_RED_IT() %>%
      set_param(c(kd=0.7194, hb=0.01579, alpha=17.69, beta=6.7)) %>%
      set_exposure(exp) %>%
      epx(level=50, hmax=0.01, method="ode45") %>%
      dplyr::pull(L.EP50) %>%
      round(digits=0)
  }

  ## Some LPx differ slightly from the reported values in EFSA (2018).
  ## However, it is reasonable to assume that our results are more precise than
  ## the ad-hoc model implementations provided to EFSA.

  # LPx values reported in Table 4, page 66, of EFSA (2018)
  # Apples, R1 pond
  expect_equal(mf_sd(exp$R1pond), 17)
  expect_equal(mf_it(exp$R1pond), 17)
  # Apples, R2 stream
  expect_equal(mf_sd(exp$R2stream), 44)
  expect_equal(mf_it(exp$R2stream), 49)
  # Cereals, D1 ditch
  expect_equal(mf_sd(exp$D1ditch), 3)
  expect_equal(mf_it(exp$D1ditch), 3)
  # Cereals, D1 stream
  expect_equal(mf_sd(exp$D1stream), 20, tolerance=0.06)
  expect_equal(mf_it(exp$D1stream), 24)
  # Cereals, D3 ditch
  expect_equal(mf_sd(exp$D3ditch), 12)
  expect_equal(mf_it(exp$D3ditch), 16)
  # Cereals, D4 pond
  expect_equal(mf_sd(exp$D4pond), 12)
  expect_equal(mf_it(exp$D4pond), 12)
  # Cereals, D4 stream
  expect_equal(mf_sd(exp$D4stream), 205, tolerance=0.02)
  expect_equal(mf_it(exp$D4stream), 250, tolerance=0.02)
  # Cereals, D5 pond
  expect_equal(mf_sd(exp$D5pond), 12)
  expect_equal(mf_it(exp$D5pond), 12)
  # Cereals, D5 stream
  expect_equal(mf_sd(exp$D5stream), 195, tolerance=0.02)
  expect_equal(mf_it(exp$D5stream), 237, tolerance=0.01)
  # Cereals, R4 stream
  expect_equal(mf_sd(exp$R4stream), 39)
  expect_equal(mf_it(exp$R4stream), 39)
})

#' Effect level
#'
#' Derives the effect level due to toxicant exposure in the supplied scenarios.
#' Either relative to a control scenario or derived directly from model
#' endpoints, depending on model type. For scenarios with moving exposure windows,
#' the maximum effect is returned.
#'
#' By default, only the maximum effect in all moving exposure windows will
#' be returned. If argument `max_only=FALSE` is set, the returned table will
#' be converted to long-format and will contain effect levels for each
#' assessed exposure window.
#'
#' ## Calculation
#'
#' Effects are calculated similarly to *relative errors*, i.e. the difference
#' between control and treatment scenarios is divided by the absolute value
#' of the control. Effects are usually in the interval `[0,1]`, but values
#' larger than one or smaller than zero can occur. As a special case, if the
#' endpoint from the control scenario is zero, then the effect is either
#' - zero, if also the treatment is zero
#' - positive infinity, if the treatment is smaller than zero
#' - negative infinity, if the treatment is greater than zero
#'
#' As an example, a control scenario achieves a biomass of
#' 1.0 and the treatment scenario achieves a biomass of 0.9, the effect will
#' be equal to 0.1 or 10%. However, effects can take on any real value. If,
#' for example, the biomass of the previously mentioned treatment scenario drops
#' below zero, then an effect larger than 1.0 will be calculated If, instead,
#' the biomass in the treatment scenario is greater than in the control, then the
#' effect will be negative.
#'
#' ### Output formatting
#' Start and end time of exposure windows can be disabled by setting `ep_only=TRUE`.
#' Effect levels smaller than a certain threshold can be automatically set to
#' zero (`0.0`) to avoid spurious effect levels introduced by numerical errors.
#' Set `marginal_effect` to an adequate value less than 1%.
#'
#' ### Computational efficiency
#' Calculations can be sped up by providing a `data.frame` of pre-calculated
#' control scenarios for each assessed time window. As control scenarios are
#' by definition independent of any exposure multiplication factor, they can
#' be reused for repeated calculations, e.g. to derive effect profiles or
#' dose-response relationships.
#'
#' @param x a [scenario] objects
#' @param factor optional numeric value which scales the exposure time-series
#' @param ep_only logical, if TRUE only effect endpoints are returned as a vector
#' @param max_only `logical`, if `TRUE` only the maximum effect is returned, else
#' results for all effect windows are reported
#' @param marginal_effect `numeric`, if set, any effect smaller than this threshold will
#' be reported as zero to exclude pseudo-effects originating from small numerical
#' errors
#' @param ... additional parameters passed on to [simulate()]
#'
#' @return a `tibble`, by default containing scenarios, effect levels, and the
#' exposure window where the maximum effect level occurred. The number of columns
#' depends on the enabled effect endpoints and function arguments.
#'
#' By default, the first column, named `scenarios`, contains the original scenario
#' objects that were the basis of the calculation. For each effect endpoint, it
#' will be followed by one column with the maximum effect level and two columns
#' containing start and end time of the associated exposure window. If exposure
#' windows are disabled, the columns will just contain the start and end time of
#' the simulation. The effect level column will have the name of the effect
#' endpoint, start and end time will additionally have the suffixes `.dat.start`
#' and `.dat.end`, respectively.
#' @export
setGeneric("effect", function(x, ...) standardGeneric("effect"), signature="x")

#' @autoglobal
effect_scenario <- function(x, factor=1, max_only=TRUE, ep_only=FALSE, marginal_effect, .cache, ...) {
  if(is.vector(x))
    stop("vectors of scenarios not supported")
  eps <- get_endpoints(x)
  if(length(eps) == 0)
    warning("no endpoints selected", call.=FALSE)

  # apply multiplication factor
  if(is_sequence(x)) {
    for(i in seq_along(x))
      x[[i]]@exposure@series[, 2] <- x[[i]]@exposure@series[, 2] * factor
  } else {
    x@exposure@series[, 2] <- x@exposure@series[, 2] * factor
  }

  # is a cache set? it would contain window candidates as well as results of
  # control simulations (i.e. with zero exposure)
  if(missing(.cache)) {
    .cache <- cache_windows(x, ...)
  }
  ctrl_req <- is_control_required(x)

  # calculate effects for all windows
  efx <- do.call(rbind, lapply(seq_along(.cache$windows), function(i) {
    # current exposure window
    win <- .cache$windows[[i]]
    # derive endpoints rel to ctrl if required, otherwise use values as they are
    rs <- fx(clip_scenario(x, window=win), ...)
    if(ctrl_req) {
      rs <- calc_effect(rs, .cache$controls[[i]])
    }
    # make sure all endpoints exist
    missing <- eps[is.na(rs[eps])]
    if(length(missing) > 0) {
      rs[missing] <- NA_real_
    }
    # return effects including exposure window
    c(win, rs)
  }))

  # set effects with an absolute value smaller than the marginal effect
  # threshold to zero
  if(!missing(marginal_effect)) {
    if(marginal_effect > 0.01) {
      warning("marginal effect threshold is larger than 1%")
    }
    efx[, eps] <- ifelse(abs(efx[, eps]) < marginal_effect, 0, efx[, eps])
  }

  # find maximum of all endpoints and corresponding time points
  if(max_only) {
    # check if all endpoints are present in result set
    #if(length(setdiff(eps, names(efx))) > 0)
    #  stop(paste("endpoint(s)",paste(setdiff(eps, names(efx)),collapse=","),"missing in effect result"))

    rs <- list()
    for(ep in eps) {
      idx <- which.max(efx[, ep])
      # idx can be empty if all endpoint values are NA, make sure we select some row
      idx <- ifelse(length(idx) == 0, 1, idx)
      max.efx <- efx[idx, c(ep, "window.start", "window.end")]
      names(max.efx) <- c(ep, paste0(ep, ".dat.start"), paste0(ep, ".dat.end"))
      rs <- append(rs, max.efx)
    }

    # discard timestamps and scenario?
    if(ep_only)
      rs <- unlist(rs[eps])
    else {
      rs <- append(list(scenario=list(x)), rs)
      rs <- tibble::as_tibble(rs)
    }
    return(rs)
  }
  # return endpoints only, for all windows
  else if(ep_only) {
    return(as.data.frame(subset(efx, select=eps)))
  }

  # rename date columns, move to back
  efx <- dplyr::relocate(as.data.frame(efx), window.start, window.end, .after=dplyr::last_col()) %>%
    dplyr::rename(dat.start=window.start, dat.end=window.end)

  # add scenario object
  efx$scenario <- list(x)
  dplyr::relocate(efx, scenario)
}

#' @describeIn effect Default for all generic [scenarios]
#' @include class-EffectScenario.R
#' @export
setMethod("effect", "EffectScenario", function(x, factor=1, max_only=TRUE, ep_only=FALSE, marginal_effect, ...)
  effect_scenario(x, factor=factor, max_only=max_only, ep_only=ep_only, marginal_effect=marginal_effect, ...))
#setMethod("effect", "EffectScenario", effect_scenario)

#' @describeIn effect For scenario [sequences][sequence]
#' @include sequence.R
#' @export
setMethod("effect", "ScenarioSequence", function(x, ...) effect_scenario(x, ...))

# Calculates the effect based on simulation and control. Takes into
# account some edge cases where sim/control are zero which would otherwise
# introduce NaNs.
#
# @param sim Usually a scenario with non-zero exposure, i.e. a treatment
# @param control A scenario with zero exposure
# @return Any real value
calc_effect <- function(sim, control) {
  rs <- (control - sim) / abs(control) # effect relative to control
  # handle zeroes in control endpoints
  idx_c0 <- control == 0
  if(any(idx_c0, na.rm=TRUE)) {
    rs[idx_c0]  <- -sign(sim[idx_c0]) * Inf # infinite effect if sim <> 0 but control is zero
    rs[idx_c0 & sim == 0] <- 0      # no effect if control and sim are zero
  }
  rs
}


#### Helper functions ####
clip_scenario <- function(scenario, window) {
  if(missing(window))
    stop("nothing to clip")
  if(length(window) != 2)
    stop("window must consist of a start and end time point")
  if(any(is.na(window)))
    stop("window contains invalid values")

  # select relevant range
  times <- get_times(scenario)
  times <- times[times >= min(window) & times <= max(window)]
  # make sure that start and end of window are included
  if(length(times) == 0)
    times <- window[[1]]
  else if(times[1] > window[1])
    times <- c(window[[1]], times)
  if(tail(times,1) < window[2])
    times <- c(times, window[[2]])

  # update output times
  scenario <- set_times(scenario, times)
  # do not touch exposure & forcing time-series of sequences, it's too fragile
  if(is_sequence(scenario)) {
    return(scenario)
  }

  # clip exposure series
  exposure <- clip_forcings(scenario@exposure@series, window)
  # clip any forcings data
  forcings <- scenario@forcings
  for(fnm in names(forcings))
    forcings[[fnm]] <- clip_forcings(forcings[[fnm]], window)

  # update other scenario settings
  scenario@exposure@series <- exposure
  scenario@forcings <- forcings

  scenario
}

#
# Minimze forcing time-series to reduce numerical overhead in deSolve
#
# It's okay if the resulting data.frame does not (fully) cover the requested
# window as forcings will be interpolated and extended by the solver as needed.
# Because of subtle interpolation issues, we also have to include the row just
# before and after our window of interest.
#
clip_forcings <- function(df, window) {
  if(nrow(df)<=2)
    return(df)
  df[seq(max(1,sum(df[,1]<window[1])), min(nrow(df),sum(df[,1]<=window[2])+1)),]
}

test_that("clip_forcings", {
  # short forcings series
  df <- data.frame(t=0,v=0)
  expect_equal(clip_forcings(df,c(0,1)), df)
  expect_equal(clip_forcings(df,c(10,11)), df)

  df <- data.frame(t=0:1,v=0)
  expect_equal(clip_forcings(df,c(0,1)), df)
  expect_equal(clip_forcings(df,c(10,11)), df)

  # vanilla use case
  df <- data.frame(t=c(0,1,1.5,2,3,4,4.5,5),v=23)
  expect_equal(clip_forcings(df,c(0,1)), data.frame(t=c(0,1,1.5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(0,1.1)), data.frame(t=c(0,1,1.5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(0,1.5)), data.frame(t=c(0,1,1.5,2),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(0,1.6)), data.frame(t=c(0,1,1.5,2),v=23), ignore_attr=TRUE)

  expect_equal(clip_forcings(df,c(1.5,4)), data.frame(t=c(1,1.5,2,3,4,4.5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(1.6,4)), data.frame(t=c(1.5,2,3,4,4.5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(2,4)), data.frame(t=c(1.5,2,3,4,4.5),v=23), ignore_attr=TRUE)

  expect_equal(clip_forcings(df,c(4,5)), data.frame(t=c(3,4,4.5,5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(4.1,5)), data.frame(t=c(4,4.5,5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(4.5,5)), data.frame(t=c(4,4.5,5),v=23), ignore_attr=TRUE)

  # window not (fully) included in time-series
  expect_equal(clip_forcings(df,c(-1,1)), data.frame(t=c(0,1,1.5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(4,5.1)), data.frame(t=c(3,4,4.5,5),v=23), ignore_attr=TRUE)
  expect_equal(clip_forcings(df,c(10,11)), data.frame(t=5,v=23), ignore_attr=TRUE)
})

test_that("clip_scenario", {
  # constant exposure series
  Lemna_Schmitt() %>%
    set_exposure(data.frame(t=0:1, c=2)) %>%
    set_times(0:10) %>%
    set_forcings(temp=data.frame(t=0, temp=20), rad=data.frame(t=-1:20, rad=42)) -> sc
  clipped <- clip_scenario(sc, c(0,2.1))

  expect_equal(clipped@times, c(0, 1, 2, 2.1))
  expect_equal(clipped@exposure@series, data.frame(t=0:1, c=2))
  expect_equal(clipped@forcings[["temp"]], data.frame(t=0, temp=20))
  expect_equal(clipped@forcings[["rad"]], data.frame(t=c(-1, 0, 1, 2, 3), rad=42), ignore_attr=TRUE)

  # complex exposure series
  sc <- sc %>% set_exposure(data.frame(t=2:4,c=23))
  clipped <- clip_scenario(sc, c(1,3))

  expect_equal(clipped@times, c(1,2,3))
  expect_equal(clipped@exposure@series, data.frame(t=c(2,3,4),c=23))
  expect_equal(clipped@forcings[["temp"]], data.frame(t=0,temp=20))
  expect_equal(clipped@forcings[["rad"]], data.frame(t=c(0,1,2,3,4),rad=42), ignore_attr=TRUE)

  # window is between two exposure time points
  sc <- sc %>% set_exposure(data.frame(t=c(0,10), c=0))
  clipped <- clip_scenario(sc, c(3,4))
  expect_equal(clipped@times, c(3,4))
  expect_equal(clipped@exposure@series, data.frame(t=c(0,10),c=0))
})

#' Macrophyte models
#'
#' Population models of standard test macrophytes, such as *Lemna spp.*
#'
#' Available macrophyte models:
#' - [Lemna][Lemna-models]
#' - [Magma][Magma()]
#'
#' @inheritSection Transferable Biomass transfer
#'
#' @name Macrophyte-models
#' @family macrophyte models
#' @family models
#' @seealso [Scenarios]
NULL

# Dummy scenario class used for unit testing without the need to do actual
# time-consuming simulations
setClass("DummyScenario", contains=c("EffectScenario","Transferable"),
  slots=list(
    fx="ANY",
    simulate="ANY",
    solver="ANY"
  ),
  prototype=list(
    fx=NULL,
    simulate=NULL,
    solver=data.frame(time=0:10, A=1, B=2),

    name="DummyScenario",
    param=list(baz=42),
    param.req=c("baz"),
    param.bounds=list(baz=c(0, 1)),
    init=c(A=1, B=2),
    times=0:10,
    endpoints=c("foo"),
    exposure=ExposureSeries(data.frame(t=0, c=1)),
    control.req=FALSE
  )
)

# Mock results from [solver()]
setMethod("solver", "DummyScenario", function(scenario, ...) solver_dummy(scenario, ...))
solver_dummy <- function(scenario, ...) {
  x <- scenario@solver
  if(is.function(x))
    return(x(scenario, ...))
  if(is.data.frame(x))
    return(x)
  stop("invalid return type for dummy solver")
}

# Mock results from [simulate()]
setMethod("simulate", "DummyScenario", function(x, ...) simulate_dummy(scenario=x, ...))
simulate_dummy <- function(scenario, ...) {
  x <- scenario@simulate
  if(is.null(x))
    return(simulate_scenario(scenario, ...))
  if(is.function(x))
    return(x(scenario, ...))
  if(is.data.frame(x))
    return(x)
  stop("invalid return type for dummy simulation")
}

# Mock results from [fx()]
setMethod("fx", "DummyScenario", function(scenario, ...) fx_dummy(scenario, ...))
fx_dummy <- function(scenario, ...) {
  x <- scenario@fx
  if(is.function(x))
    return(x(scenario, ...))
  if(is.numeric(x))
    return(x)
  stop("invalid return type for dummy fx")
}

#' Disable parallelization for debugging
#'
#' In certain cases it might be beneficial to disable parallel execution
#' of e.g. effect profile calculations. By disabling, all processes run
#' sequentially and instantly pass messages to the console which would be
#' delayed during parallel processing. This makes it easier to pinpoint
#' problems within the data or algorithm.
#'
#' @param state `logical`, if `TRUE` then parallelization is disabled
#' @return no return value
#' @export
pll_debug <- function(state=TRUE) {
  if(state)
    message("cvasi: parallelization is disabled")
  else
    message("cvasi: parallelization can be enabled")
  options(cvasi.pll.off=state)
}

get_endpoints <- function(x) {
  if(is.vector(x))
    return(lapply(x, get_endpoints))

  if(is_sequence(x))
    return(get_endpoints(x[[1]]))
  x@endpoints
}

# Get forcings
get_forcings <- function(scenario) {
  if(is.vector(scenario)) stop("vector arguments not supported")
  return(scenario@forcings)
}

# Get names of required forcings
get_req_forcings <- function(scenario) {
  if(is.vector(scenario)) stop("vector arguments not supported")
  return(scenario@forcings.req)
}

# Get names of state variables
get_vars <- function(x) {
  if(is.vector(x)) stop("vector arguments not supported")
  return(names(x@init))
}

# Get settings for moving windows
get_window <- function(x) {
  if(is.vector(x) & length(x) > 1) {
    return(lapply(x, get_window))
  }
  if(is_scenario(x)) {
    return(list(length=x@window.length, interval=x@window.interval))
  } else if(is_sequence(x)) {
    return(get_window(x[[1]]))
  }
  stop("Type not supported")
}

#' Get model name
#'
#' Returns the unique model name that is associated with a scenario, e.g.
#' `GUTS-RED-IT`. The function supports vectorized arguments.
#'
#' @param x (vector of) [scenarios] or [`parameter_set`] objects
#'
#' @return vector of `character`
#' @aliases get_model,ANY-method get_model,list-method get_model,EffectScenario-method
#'   get_model,ParameterSet-method get_model_name
#' @export
#'
#' @examples
#' # returns `GUTS-RED-IT`
#' get_model(minnow_it)
setGeneric("get_model", function(x) standardGeneric("get_model"))
#' @export
setMethod("get_model", "ANY", function(x) stop("model name not supported for given object type"))
#' @export
setMethod("get_model", "list", function(x) sapply(x, get_model))
#' @export
setMethod("get_model", "EffectScenario", function(x) return(x@name))
#' @export
setMethod("get_model", "ParameterSet", function(x) return(x@model))
# get_model-method Deprecated alias for `get_name()`
#' @export
get_model_name <- function(x) {
  # todo add lifecycle notice
  return(get_model(x))
}

test_that("get_model", {
  s1 <- new("EffectScenario", name="foo")
  s2 <- new("EffectScenario", name="bar")
  p3 <- new("ParameterSet", model="baz")

  # empty argument
  expect_equal(get_model(list()), list())
  # single argument
  expect_equal(get_model(s1), s1@name)
  expect_equal(get_model(p3), p3@model)
  # vectorized arguments
  expect_equal(get_model(c(s1,s2)), c(s1@name, s2@name))
  expect_equal(get_model(c(s1,p3)), c(s1@name, p3@model))
  # invalid arguments
  expect_error(get_model(assessment()))
})

test_that("get_window", {
  sc <- new("EffectScenario") %>% set_window(length=-1, interval=-1)
  expect_equal(get_window(sc), list(length=-1, interval=-1))

  sc <- new("EffectScenario") %>% set_window(length=2, interval=3)
  l <- list(length=2, interval=3)
  expect_equal(get_window(sc), l)

  expect_equal(get_window(list(sc, sc)), list(l, l))
})

#' Calculate a dose response curve
#'
#' Returns a `data.frame` with points on the dose response curve for the given
#' effect scenario.
#'
#' Derives a dose response curve from a [scenario]. The result will
#' cover the requested range of effect levels. The tested multiplication factors
#' can be chosen by different strategies, i.e. a `vanilla` approach using a
#' fixed set of factors, or `decadic` and `exponential` approaches
#' employing logarithmic and exponential factor scaling, respectively.
#'
#' @param scenario used for calculation
#' @param range numeric vector specifying the required range of effect levels in
#'   percent (%), defaults to `c(1,99)`
#' @param n minimum number of points on the dose response curve
#' @param strategy controls how multiplication factors are chosen, `vanilla` uses a fixed
#' set of multiplication factors, `decadic` and `exponential` have varying step lengths.
#' @param verbose  logical, set to `TRUE` for additional status messages
#' @param ... additional arguments passed on to [effect()]
#'
#' @return `data.frame` with two columns, i.e. `mf` and `effect`
#' @export
#' @global endpoint
#'
#' @examples
#' # basic dose response curve
#' minnow_sd %>% dose_response()
#'
#' # modify the minimum number of points on the curve
#' minnow_sd %>% dose_response(n=10)
#'
#' # select a subset of the effect range
#' minnow_sd %>% dose_response(range=c(10,20))
#'
#' # use an alternative strategy for the selection of multiplication factors
#' minnow_sd %>% dose_response(strategy="decadic")
#'
#' # provide additional output how multiplication factors were selected
#' minnow_sd %>% dose_response(verbose=TRUE)
dose_response <- function(scenario, range=c(1, 99), n=20, strategy=c("exponential", "decadic", "vanilla"),
                          verbose=FALSE, ...)
{
  if(length(scenario) > 1 | is.data.frame(scenario))
    stop("multiple scenarios supplied")
  if(is.vector(scenario))
    scenario <- scenario[[1]]

  strategy <- match.arg(strategy)
  if(length(range) == 0)
    stop("effect range required")
  else if(length(unique(range)) == 1) {
    min.effect <- max(1, range[1] - 1)
    max.effect <- min(99, range[1] + 1)
  } else {
    min.effect <- max(min(range) - 1, 1)
    max.effect <- min(max(range) + 1, 99)
  }

  # sequential processing requested?
  if(getOption("cvasi.pll.off", FALSE))
    map2 <- purrr::map2_dfr
  else
    map2 <- furrr::future_map2_dfr
  # check if controls are present
  cache <- cache_windows(scenario, ...)

  epx.min <- epx(scenario, level=min.effect, .cache=cache, ...) %>%
    dplyr::rename_with(~gsub("\\.EP\\d+", "", .x))
  epx.max <- epx(scenario, level=max.effect, .cache=cache, ...) %>%
    dplyr::rename_with(~gsub("\\.EP\\d+", "", .x))
  drc <- data.frame(endpoint=character(), mf=numeric(), effect=numeric())

  for(ep in scenario@endpoints)
  {
    f.min <- epx.min[[1, ep]]
    f.max <- epx.max[[1, ep]]
    if(is.na(f.min) | is.na(f.max)) {
      warning("Cannot determine EPx range, skipping endpoint '", ep, "'", call.=FALSE)
      next
    }

    # Decadic strategy has a constant step length in decimal multiples of powers of ten
    if(strategy=="decadic")
    {
      dec.step <- 2
      do.repeat <- T
      while(do.repeat)
      {
        dec.step <- dec.step / 2
        if(verbose)
          message("  decadic step length ", dec.step)
        if(dec.step < 1e-30)
          stop("Refinement failed, step length too small")
        do.repeat <- F
        # list of factors to calculate effects for
        mf <- c()
        for(i in seq(log10(f.min), 3))
          mf <- c(mf, seq(1, 9, dec.step) * 10^i)
        # remove factors larger than upper boundary
        if(f.max > 0)
          mf <- mf[mf <= f.max]
        # more factors requested than present?
        do.repeat <- length(mf) < n
      }
    }
    # Exponential strategy has a constant step width in the powers of ten
    else if(strategy=="exponential")
    {
      pow.step <- 0.2
      do.repeat <- T
      while(do.repeat)
      {
        pow.step <- pow.step / 2
        if(verbose)
          message(paste("  exponential step length", pow.step))
        if(pow.step < 1e-30)
          stop("Refinement failed, step length too small")
        do.repeat <- F
        # list of factors to calculate effects for
        mf <- exp(log(10) * seq(log(f.min) / log(10), 10, pow.step))
        # remove factors out of target range
        mf <- mf[mf >= f.min]
        if(f.max > 0)
          mf <- mf[mf<=f.max]
        # more factors requested than present?
        do.repeat <- length(mf)<n
      }
    }
    # Vanilla approach as implemented in an older risk assessment
    else if(strategy=="vanilla") {
      mf <- c(1:99,	1:9 * 100, 2:10 * 500, 6:10 * 1000)
      # remove factors out of target range
      mf <- mf[mf >= f.min]
      if(f.max > 0)
        mf <- mf[mf <= f.max]
      if(length(mf) < n)
        stop("Refinement not supported for vanilla strategy")
    }

    if(verbose)
      message("  calculating ", length(mf), " new effect levels")
    # calculate effects
    efx <- map2(rep(c(scenario), length(mf)), mf, effect, ep_only=TRUE, .cache=cache, ...)
    dplyr::bind_rows(drc, tibble::tibble(mf=mf, effect=efx[[ep]], endpoint=ep)) -> drc
  }
  drc <- drc %>%
    dplyr::arrange(endpoint, mf) %>%
    as.data.frame()
  class(drc) <- c("cvasi_drc", class(drc))
  attr(drc, "scenario") <- scenario

  drc
}

#' Effect scenario classes
#'
#' The `EffectScenario` class is the base for all of the basic scenario
#' types and models. It contains slots for data and settings that are
#' required by most models such as a vector of model parameters and a vector
#' of initial states. For each particular model, the class's slots are
#' filled with certain default or fixed values. Some models derive from this
#' class and add slots to store additional data.
#'
#' Certain behaviors that are required to model complex processes cannot be
#' represented by a single EffectScenario. As an example, the parameters of a scenario
#' are generally fixed during the simulated time period. In order to represent a
#' change in parameter values, the original scenario would
#' need to split into two scenarios *A* and *B* which differ by parameter values
#' and simulated time period. By combining these scenarios to a *scenario sequence*,
#' the sequence would be treated as a single, complex scenario. See
#' [sequence()] for more information.
#'
#' ### Parameters
#' Most parameters are represented by numerical types but other types
#' are possible depending on model. Please refer to the model description
#' which parameters are required and in which unit. Some or all parameters may
#' be required to start a simulation. If required parameters are missing,
#' simulation will fail with an error message.
#'
#' ### Initial state
#' The *initial state* represents the starting values of state variables when
#' starting a simulation. A scenario's default initial state may be insufficient
#' to get sensible results. It is advisable to set an initial state explicitly
#' when creating a new scenario, see [set_init()].
#'
#' In theory, a scenario's state variables can be renamed by modifying the names
#' of the initial state vector. However, this is strongly
#' discouraged as this will affect other routines such as [effect()] and [epx()]
#' and may render results useless.
#'
#' ### Exposure
#' *Exposure* refers to the concentration of toxicant an organism is exposed to.
#' In case of aquatic organisms, this would commonly be the concentration of a
#' toxicant in water. Other interpretations are possible depending on model
#' assumptions.
#'
#' Exposure time-series are generally represented by a `data.frame` containing two
#' columns. The first column representing time, the second representing the
#' exposure level. The ordering of columns is mandatory. The column names
#' are essentially irrelevant but sensible names may help documenting the
#' scenario and its data. The rows must be ordered chronologically. A time-series
#' can consist of only a single row; in this case it will represent constant
#' exposure. Exposure time-series are set to a scenario using [set_exposure()].
#'
#' Handling time-series is a costly task for the ODE solver due to consistency
#' checks and interpolation between time steps. How the solver interpolates
#' the time-series can be controlled by certain arguments to functions
#' such as [simulate()] and [effect()]. Please refer to [simulate()] for a brief
#' overview and [deSolve::forcings] for a detailed description.
#'
#' Exposure time-series should be kept as short as possible and as complex as
#' needed for optimal computational efficiency.
#'
#' ### Environmental forcings
#' *Forcings* generally refer to model parameters that change over time as part
#' of an external function such as environmental temperature and exposure levels.
#' Due to the importance of exposure in regulatory assessments, this R package
#' explicitly distinguishes between environmental forcings and exposure. However,
#' the same restrictions and features apply to both of them.
#'
#' Forcing time-series are handled the same way as exposure time-series, i.e.
#' they are represented by a `data.frame` containing two columns. The first column
#' representing time, the second representing the parameter that is a function of
#' time. The ordering of columns is mandatory. The rows must be ordered
#' chronologically. Forcings time-series are set using [set_forcings()].
#' Please refer to the *Exposure* section for more information on how time-series
#' are handled.
#'
#' ### Output times
#' A scenario's simulated time period is defined by its minimum and maximum output
#' time. Simulation results will only be returned for the defined output times
#' even though the ODE solver may use smaller time steps between output times.
#' Output times can be explicitly set using [set_times()]. The number and
#' distance of output times may have influence on the precision of simulation
#' results and numerical stability, cf. [simulate()].
#'
#' Be aware that [set_exposure()] will overwrite previously defined output times
#' if not requested otherwise.
#'
#' ### Effects
#' Generally, all state variables can be used as effect endpoints but models may
#' provide additional endpoints. Use [set_endpoints()] to enable or disable
#' endpoints for a scenario.
#'
#' Some scenarios or models require control runs to calculate effects under
#' exposure. Generally, control simulations will run automatically
#' where needed.
#'
#' ### Moving exposure windows
#' The time frame relevant for effects may be much shorter than the assessed
#' exposure time-series for certain organisms. This fact can be represented
#' by moving exposure windows which divide a long time period in a number of
#' consecutive windows of the same length. Each window is simulated individually and effects are
#' calculated. By default, methods such as [effect()] will only return the
#' maximum effect of all considered windows but detailed results can be presented
#' on demand.
#'
#' To use moving exposure windows, the exposure time-series must be regular,
#' i.e. must have an equidistant step length in time. The length of the window
#' is defined as the number of time steps of the exposure time-series. As an
#' example, assume the time-series has daily granularity and a moving window
#' of seven days length is required. In this case, the moving window must have
#' a length of seven (7) time steps. If the exposure time-series had hourly granularity,
#' the same window would need to have a length of 168 (=7*24) time steps. Please
#' refer to [set_window()] for details.
#'
#' @slot name `character`, unique model name
#' @slot tag `character`, an optional identifier
#' @slot param `list` of parameter key-value pairs
#' @slot param.bounds named `list` of parameter boundaries
#' @slot param.req `character` vector of required parameters
#' @slot forcings `list` of `data.frame`s representing forcing time-series
#' @slot forcings.req `character` vector or required model forcings data, e.g. temperature
#' @slot init `list` of initial model states
#' @slot times `numeric` vector of output times, beginning and end also define
#'  the simulated period
#' @slot endpoints `character` vector of endpoints to calculate results for
#' @slot exposure `data.frame` with two columns representing an exposure time-series
#' @slot control `list` of named numerical vectors, contains the control values
#' for all relevant moving windows
#' @slot control.req `logical`, if `TRUE` then control values are required to
#'   calculate effects
#' @slot window.length `numeric`, maximum length of the simulated period, if
#'   `window.length` is shorter than the exposure pattern, then all possible
#'   exposure sub-patterns are evaluated for effect calculation. This is also
#'   referred to as a moving window approach.
#' @slot window.interval `numeric`, interval determining distance between moving
#'   windows during effect calculation. First window starts at first time point
#'   in exposure pattern.
#'
#' @name Scenarios
#' @aliases scenario scenarios EffectScenario-class
#' @family scenario-related
NULL



#' @include class-ExposureSeries.R
#' @export
setClass("EffectScenario",
   slots = list(
     name="character",
     param="list",
     param.req="character",
     param.bounds="list",
     forcings="list",
     forcings.req="character",
     init="numeric",
     times="numeric",
     endpoints="character",
     exposure="ExposureSeries",
     control="ANY",
     control.req="logical",
     window.length="numeric",
     window.interval="numeric",
     tag="character"
   ),
   prototype = list(
     name="undefined",
     param.bounds=list(),
     control.req=TRUE,
     window.length=-1,
     window.interval=-1,
     tag=NA_character_
   )
)

#' Survival rate
#'
#' `r lifecycle::badge("deprecated")`
#' Derives the survival rate of individuals for
#' *Reduced GUTS* models. Function was replaced by output of [simulate()] and
#' will be removed in a later version.
#'
#' The survival rate describes the survival probability at each
#' time point. The function simulates the *GUTS* scenario and appends a column
#' `survival` to the simulation result. A value of one (`1.0`) denotes that
#' all individuals survive. A value of zero (`0.0`) denotes that no individuals
#' survived.
#'
#' Only available for *Reduced GUTS* models, see [GUTS-RED-models].
#' The equations were described by EFSA (2018).
#'
#' @param scenario an `EffectScenario` to simulate
#' @param ... additional parameters passed on to [simulate()]
#'
#' @return a `data.frame` containing simulation results
#'
#' @export
#' @inherit GUTS-RED-models references
#' @seealso [GUTS-RED-models]
#'
#' @autoglobal
#' @examples
#' # calculate survival rate
#' minnow_it %>% survival()
#'
#' # plot survival over time based on a random exposure profile
#' minnow_sd %>%
#'   set_exposure(data.frame(t=1:100, c=runif(100)*10)) %>%
#'   survival() -> df
#' plot(df$time, df$survival, "l")
survival <- function(scenario, ...) {
  if(length(scenario) > 1 | is.data.frame(scenario))
    stop("multiple scenarios supplied")

  lifecycle::deprecate_soft("1.2.0", "survival()", "simulate()")
  if(is_guts(scenario)) {
    df <- simulate(scenario, ...) %>%
    dplyr::rename(survival=S)
    return(df)
  }
  stop("model not supported")
}

test_that("basic functionality", {
  # GUTS-RED-IT
  sim <- minnow_it %>% simulate()
  expect_warning(surv <- minnow_it %>% survival())
  expect_equal(surv$survival, sim$S)
  # GUTS-RED-SD
  sim <- minnow_sd %>% simulate()
  expect_warning(surv <- minnow_sd %>% survival())
  expect_equal(surv$survival, sim$S)
})

test_that("unsupported model", {
  expect_error(expect_warning(survival(metsulfuron)))
})

# for internal use only
#' @noRd
setClass("SimulationBatch",
  slots=list(
   scenarios="list",
   id_col="character",
   format="character",
   select="ANY"
  )
)

#' Batch simulation of multiple exposure levels
#'
#' `r lifecycle::badge("experimental")`
#'
#' A convenience function to simulate a single base scenario with one or more
#' exposure levels. The functions aims at reproducing the setup and result format
#' of common effect studies.
#'
#' Simulating a [scenario] is generally limited to assessing a single exposure
#' series. However, laboratory experiments commonly examine the effects of
#' multiple exposure levels on a biological system.
#' A *batch simulation* approach involves running multiple simulations with
#' varying exposure or *treatment* conditions. To illustrate: if the objective is
#' to examine the impact of a chemical on cell growth, multiple scenarios need
#' to be simulated to reproduce the cell growth dynamics under varying
#' concentrations of the assessed chemical. Each simulation run will represent a
#' specific exposure level, ranging from low to high concentrations of the
#' chemical.
#'
#' To simulate the conditions of such a laboratory experiment, the scenarios and
#' exposure levels can either be created and simulated individually, or the
#' `batch()` function can be used for ease of use.
#'
#' ### Exposure series
#' The set of exposure levels can be represented by one of the following types:
#'
#' - A (named) list: Each element represents an exposure level or exposure
#'   series. An exposure level can be represented by a constant numeric, a
#'   `data.frame` with two columns, or an ExposureSeries object. The names of
#'   the list elements specify the study ID.
#' - Or alternatively, a `data.frame` with three columns: One column for time,
#'   one for the exposure level, and one character column to specify the
#'   study IDs.
#'
#' Each exposure level will be simulated using the base scenario. If the exposure
#' levels are provided as a named list, the names will also appear in the return
#' value of `simulate()`. This behavior can be used, for example, to define unique
#' study IDs for particular exposure levels.
#'
#' ### Exposure IDs
#' The list of exposure levels can be supplied as a named list. The names
#' will be used as unique (study) IDs, so that the simulation results belonging
#' to any exposure level can be identified in the output.
#' If no IDs are defined by the user, generic IDs of the form `'trial{n}'` will
#' be assigned, with `{n}` being replaced by consecutive integers starting at one.
#'
#' If the batch is passed on to [simulate()]`, the IDs will be contained in its
#' return value, e.g. as a dedicated column (long format) or as part of the
#' column names (wide format).
#'
#' ### Output format
#' The return value of [simulate()] is by default in long format, i.e. it will
#' contain one row for each output time and exposure level. It is possible to
#' pivot the tabular data to wide format, by setting the argument
#' `format = 'wide'`.
#'
#' In wide format, the output columns of each exposure level are pasted
#' next to each other. If more than one column is pivoted per exposure level,
#' then the exposure or study ID is added as a suffix to column names.
#' If the output per exposure level contains only a single column (besides time
#' and the exposure ID itself), then original column name is dropped and only
#' exposure IDs are used. See the examples section for reference.
#'
#' ### Select output columns
#' Often, only a single output column is of interest in batch simulations,
#' such as the number of surviving individuals. To ease the interpretation and
#' handling of the output of batch simulations, the columns contained in the
#' output of each simulated exposure level can be filtered. One or more columns
#' can be selected. By default, no filtering of output columns is conducted.
#'
#' As an example, to create an overview of survival probabilities (*S*) in the
#' GUTS-RED-IT example scenario `minnow_it`:
#'
#' ```
#' minnow_it %>%
#'   batch(exposure=list(0, 5, 10), select="S", format="wide") %>%
#'   simulate()
#' ```
#'
#' @param scenario a [scenario] object
#' @param exposure a named `list()` or a `data.frame` with three columns
#' @param id_col `character`, name of column in resulting ´data.frame` which
#'    contains a trial's name or ID
#' @param format `character`, set to `'long'` for long tabular format, or
#'    `'wide'` for wide format
#' @param times_from `character`, set to `'scenario'` to use output times from
#'    scenario, or `'exposure'` to take output times from each exposure series
#' @param select optional `character` vector to select columns from the
#'    simulation output
#' @return a simulation batch object
#' @export
#' @examples
#' # Simulate a batch experiment with three constant exposure levels of
#' # 0.0, 2.0, and 5.0 µmol/L
#' simulate(batch(minnow_it, list(0, 2, 5)))
#'
#' # Alternatively, in tidyr style syntax
#' trials_list1 <- list(0, 2, 5)
#' minnow_it %>%
#'   batch(trials_list1) %>%
#'   simulate()
#'
#' # Assign unique IDs to each exposure level
#' trials_list2 <- list(Control=0, TrialA=2, TrialB=5)
#' minnow_it %>%
#'   batch(trials_list2) %>%
#'   simulate()
#'
#' # Alternatively, define multiple exposure levels in a single data.frame
#' trials_table <- data.frame(time=c(0, 0, 0),
#'                            conc=c(0, 2, 5),
#'                            trial=c("Control", "TrialA", "TrialB"))
#' minnow_it %>%
#'   batch(trials_table) %>%
#'   simulate()
#'
#' # Limit simulation output to column 'S' (survival probability)
#' minnow_it %>%
#'   batch(trials_list2, select="S") %>%
#'   simulate()
#'
#' # Return data in wide-format, unique IDs will be used as column names
#' minnow_it %>%
#'   batch(trials_list2, select="S", format="wide") %>%
#'   simulate()
batch <- function(scenario, exposure, id_col="trial", format=c("long", "wide"),
                  times_from=c("scenario", "exposure"), select=NULL) {
  # make sure that it's a scenario
  if(!all(is_scenario(scenario)))
    stop("argument 'scenario' must contain a scenario object")
  if(length(scenario) != 1)
    stop("argument 'scenario' must contain only a single element")
  if(missing(exposure))
    stop("argument 'exposure' is missing")

  # convert data.frame with three columns to a named list of data.frames
  # which have two columns
  if(is.data.frame(exposure)) {
    if(length(exposure) != 3)
      stop("argument 'exposure' is a data.frame; it must have exactly three columns")
    if(!(id_col %in% names(exposure)))
      stop("argument 'id_col' is not a column in exposure data.frame")
    if(any(is.null(exposure[, id_col]) | is.na(exposure[, id_col])))
      stop("the study ID column contains missing values (NA or NULL)")

    ids <- exposure[, id_col]
    exposure <- dplyr::select(exposure, !dplyr::any_of(id_col))
    exposure <- split(exposure, ids)
    # make sure we have the original ordering, and not something lexicographically
    # sorted by `split()`
    exposure <- exposure[unique(ids)]
  }

  # verify format of exposure argument
  if(!is.list(exposure))
    stop("argument 'exposure' must be a list")
  if(length(exposure) == 0)
    stop("argument 'exposure' must not be empty")

  # verify types of exposure series, convert where necessary
  for(i in seq_along(exposure)) {
    val <- exposure[[i]]
    if(is(val, "ExposureSeries")) {
      # nothing else to do
    } else if(is.data.frame(val)) {
      exposure[[i]] <- ExposureSeries(val)
    } else if(is.numeric(val) & length(val) == 1) { # atomic value
      exposure[[i]] <- ExposureSeries(data.frame(time=0, conc=val))
    }
  }

  ids <- names(exposure)
  # some or all ids missing
  if(is.null(ids)) {
    message("Elements in 'exposure' are not named, assigning generic ids 'trial{n}'")
    names(exposure) <- paste0("trial", seq(length(exposure)))
  }

  ids_empty <- which(names(exposure) == "")
  if(length(ids_empty) > 0) {
    message("Some elements in 'exposure' are not named, assigning generic ids 'trial{n}'")
    nms <- names(exposure)
    nms[ids_empty] <- paste0("trial", ids_empty)
    names(exposure) <- nms
  }

  # other arguments
  format <- match.arg(format)
  times_from <- match.arg(times_from)

  # create list of scenarios
  lst <- lapply(exposure, function(exp) {
    set_exposure(scenario, exp, reset_times=(times_from == "exposure"))
    })

  new("SimulationBatch",
    scenarios=lst,
    id_col=id_col,
    format=format,
    select=select
  )
}

test_that("ids", {
  base <- minnow_it %>% set_times(0:4)

  suppressMessages(b <- batch(base, list("bar"=0, "baz"=1), id_col="foo"))
  df <- simulate(b)
  expect_contains(names(df), "foo")
  expect_equal(df[, "foo"], rep(c("bar", "baz"), each=5))
})

test_that("format argument", {
  base <- minnow_it

  # long format
  suppressMessages(b <- batch(base, list(foo=0), format="long"))
  df <- simulate(b)
  expect_equal(names(df), c("time","D","H","S","trial"))

  # wide format
  suppressMessages(b <- batch(base, list(foo=0), format="wide"))
  df <- simulate(b)
  expect_equal(names(df), c("time","D_foo","H_foo","S_foo"))

  # wide format, single column
  suppressMessages(b <- batch(base, list(foo=0), format="wide", select="S"))
  df <- simulate(b)
  expect_equal(names(df), c("time","foo"))
})

test_that("select argument", {
  base <- minnow_it

  suppressMessages(b <- batch(base, list(0), select="S"))
  df <- simulate(b)
  expect_equal(names(df), c("time","S", "trial"))
})

test_that("exposure levels", {
  base <- new("EffectScenario")

  # single level
  suppressMessages(b <- batch(base, list(23)))
  expect_s4_class(b, "SimulationBatch")
  expect_equal(length(b@scenarios), 1)
  expect_equal(b@scenarios[[1]]@exposure@series[,2], 23)

  # multiple levels
  suppressMessages(b <- batch(base, list(0, 1, 2, 3)))
  expect_s4_class(b, "SimulationBatch")
  expect_equal(length(b@scenarios), 4)
  expect_equal(b@scenarios[[1]]@exposure@series[,2], 0)
  expect_equal(b@scenarios[[4]]@exposure@series[,2], 3)
})

test_that("list of exposure types", {
  base <- new("EffectScenario")

  # constant values
  suppressMessages(b <- batch(base, list(1, 2)))
  expect_equal(length(b@scenarios), 2)
  expect_equal(b@scenarios[[1]]@exposure@series[,2], 1)
  expect_equal(b@scenarios[[2]]@exposure@series[,2], 2)

  # data.frames
  suppressMessages(b <- batch(base, list(
    data.frame(t=0, c=3),
    data.frame(t=0, c=4)
  )))
  expect_s4_class(b, "SimulationBatch")
  expect_equal(length(b@scenarios), 2)
  expect_equal(b@scenarios[[1]]@exposure@series[,2], 3)
  expect_equal(b@scenarios[[2]]@exposure@series[,2], 4)

  # ExposureSeries objects
  suppressMessages(b <- batch(base, list(
    ExposureSeries(data.frame(t=0, c=5)),
    ExposureSeries(data.frame(t=0, c=6))
  )))
  expect_s4_class(b, "SimulationBatch")
  expect_equal(length(b@scenarios), 2)
  expect_equal(b@scenarios[[1]]@exposure@series[,2], 5)
  expect_equal(b@scenarios[[2]]@exposure@series[,2], 6)
})

test_that("data.frame of exposure series", {
  base <- new("EffectScenario")

  # single trial
  df1 <- data.frame(t=0:4, c=1, trial="foo")
  suppressMessages(b <- batch(base, df1))
  expect_equal(length(b@scenarios), 1)
  expect_equal(names(b@scenarios), "foo")
  expect_equal(b@scenarios[[1]]@exposure@series[,1], 0:4)
  expect_equal(b@scenarios[[1]]@exposure@series[,2], rep(1, 5))

  # multiple trials, custom id column name
  df2 <- rbind(data.frame(t=0, c=1, baz="foo"),
               data.frame(t=0, c=2, baz="bar"))
  suppressMessages(b <- batch(base, df2, id_col="baz"))
  expect_equal(length(b@scenarios), 2)
  expect_equal(names(b@scenarios), c("foo","bar"))
  expect_equal(b@scenarios[[1]]@exposure@series[,2], 1)
  expect_equal(b@scenarios[[2]]@exposure@series[,2], 2)

  ## error modes
  # too few columns
  expect_error(batch(base, data.frame(t=0:1, trial="foo")))
  # too many columns
  expect_error(batch(base, data.frame(t=0:1, conc=0, trial="foo", bar="baz")))
  # id column not present
  expect_error(batch(base, data.frame(t=0:1, conc=0, bar="baz")))
  # missing trial ids
  expect_error(batch(base, data.frame(t=0:1, conc=0, trial=c("foo", NA_character_))))
})

test_that("id column", {
  base <- minnow_it

  # default value
  suppressMessages(b <- batch(base, list(0)))
  expect_true(is.character(b@id_col))

  # custom value
  suppressMessages(b <- batch(base, list(0), id_col="foo"))
  expect_equal(b@id_col, "foo")
})

test_that("unique exposure ids", {
  base <- minnow_it

  # automatic ids
  suppressMessages(b <- batch(base, list(0, 1)))
  expect_equal(names(b@scenarios), c("trial1", "trial2"))

  suppressMessages(b <- batch(base, list(foo=0, 1)))
  expect_equal(names(b@scenarios), c("foo", "trial2"))

  # manual ids
  b <- batch(base, list(a=0, b=1))
  expect_equal(names(b@scenarios), c("a", "b"))
})

test_that("unique exposure ids", {
  base <- new("EffectScenario") %>% set_times(0:10)

  # times from scenario
  suppressMessages(b <- batch(base, list(0), times_from="scenario"))
  expect_equal(b@scenarios[[1]]@times, 0:10)
  suppressMessages(b <- batch(base, list(data.frame(t=1, c=0)), times_from="scenario"))
  expect_equal(b@scenarios[[1]]@times, 0:10)

  # times from exposure
  expect_error(suppressMessage(batch(base, list(0), times_from="exposure")))
  suppressMessages(b <- batch(base, list(data.frame(t=1:4, c=0)), times_from="exposure"))
  expect_equal(b@scenarios[[1]]@times, 1:4)
})

test_that("format argument", {
  base <- new("EffectScenario")

  suppressMessages(b <- batch(base, list(0), format="long"))
  expect_equal(b@format, "long")
  suppressMessages(b <- batch(base, list(0), format="wide"))
  expect_equal(b@format, "wide")
})

test_that("select argument", {
  base <- new("EffectScenario")

  suppressMessages(b <- batch(base, list(0), select=NULL))
  expect_null(b@select)
  suppressMessages(b <- batch(base, list(0), select="foo"))
  expect_equal(b@select, "foo")
})

test_that("invalid arguments", {
  expect_error(batch())
  expect_error(batch(NULL, list(0)))
  expect_error(batch(list(), list(0)))
  expect_error(batch(list(1), list(0)))
  expect_error(batch(list(minnow_it, minnow_it), list(0)))

  expect_error(batch(minnow_it))
  expect_error(batch(minnow_it, 1))
  expect_error(batch(minnow_it, list()))
})

#' SWASH project exposure profile import
#'
#' Read all TOXSWA files within a SWASH project directory.
#'
#' @param swash_dir path to the SWASH project directory
#' @param ... arguments passed on to [import_toxswa()]
#'
#' @return a list of imported exposure series, see [import_toxswa()] for details
#' @export
import_swash <- function(swash_dir, ...){
  toxwa_file_regex <- "^\\d*\\.out$"
  f <- list.files(swash_dir,
                  pattern = toxwa_file_regex,
                  full.names = TRUE)
  import_toxswa(f, ...)
}

#' Import *TOXSWA* exposure series
#'
#' Read one or more *TOXSWA* exposure series from *TOXSWA*'s `.out` files. By default, the
#' concentration dissolved in water (*ConLiqWatLay*) at the end of the simulated waterbody
#' (i.e. at the maximum of the *x* dimension) is returned. The unit of the time
#' scale as well as of the imported model output variable can be scaled as needed.
#'
#' The numerical time scale is shifted to always start at time zero (0.0).
#' Numerical columns of the returned *data.frame* objects will be of type
#' [units::units]. Please be aware that the use of `units` objects may not be
#' supported by all functions in this package. However, [set_times()] and
#' [set_exposure()] can handle `units` objects safely.
#'
#' Incomplete list of alternative *TOXSWA* v5.5.3 output variables:
#' - *ConLiqWatLay*: Concentration dissolved in water (g/m3)
#' - *ConLiqSed*: Concentration in pore water sediment (g/m3)
#' - *ConSysWatLay*: Total concentration in water (g/m3)
#' - *CntSorSusSol*: Content sorbed to suspended solids (g/kg)
#' - *CntSorSed*: Content sorbed to sediment (g/kg)
#'
#' @param files vector of strings with absolute or relative paths to files
#' @param alias optional vector with strings, will be used as an alias to identify
#'   a TOXSWA series instead of its filename
#' @param output_var character, single output variable from *TOXSWA* that is
#'   imported, defaults to *ConLiqWatLay*
#' @param output_unit character, target unit of the imported output variable,
#'   defaults to *ug/L*, syntax must be compatible with [units::units()]
#' @param time_unit character, target unit of the imported time scale,
#'   defaults to *days*, syntax must be compatible with [units::units()]
#' @param substance optional vector of characters, if set, only the substance
#'   codes defined in this vector are imported
#' @param split logical, if `TRUE` then one series will be returned for each
#'   substance found in the *TOXSWA* files, else all substances per file will
#'   be in one *data.frame*. Defaults to `TRUE`
#' @export
#' @autoglobal
#' @return list of *data.frame* objects with exposure series. Each *data.frame* has at
#'   least three columns:
#'   - `time`: numerical time scale, always starts at zero
#'   - `timestamp`: time as datetime objects such as `POSIXct`
#'   - one or more additional columns for each imported substance
#'
import_toxswa <- function(files, alias=NA, output_var="ConLiqWatLay", output_unit="ug/L",
                           time_unit="days", substance=NULL, split=TRUE) {
  if(length(files) == 0) {
    stop("files vector is empty")
  }
  if(length(output_var) != 1)
    stop("Exactly one output variable must be selected")

  # vectorized file list
  if(length(files) > 1) {
    return(purrr::map2(files, alias, import_toxswa, output_var=output_var, output_unit=output_unit,
                  time_unit=time_unit, substance=substance, split=split) %>%
           purrr::list_flatten())
  }

  # for now we assume that `files` contains a path to a file
  if(any(is.na(files)) | !is.character(files))
    stop("files vector contains invalid values")
  if(!file.exists(files))
    stop("file does not exists: ", files)

  headers <- paste0("V", seq(1, max(utils::count.fields(files, sep=""))))
  df <- utils::read.table(files, header=FALSE, col.names=headers, fill=TRUE)

  comments <- df %>%
    dplyr::filter(V1=="*") %>%
    tidyr::unite(col="V1", dplyr::everything(), sep=" ", na.rm=TRUE) %>%
    dplyr::mutate(V1=trimws(V1)) %>%
    dplyr::pull(V1)
  data <- df %>%
    dplyr::filter(V1 != "*")
  rm(df, headers)

  # check toxswa version
  version_string <- comments[stringr::str_detect(comments, "TOXSWA version")]
  if(length(version_string) == 0)
    stop("TOXSWA version missing")
  if(!stringr::str_detect(version_string, " (4|5\\.5\\.3)$"))
    stop("TOXSWA version not supported")

  # get unit of chosen output var
  unit_str <- comments[stringr::str_detect(comments, paste0("^\\* Unit for ", output_var, " is "))]
  if(length(unit_str) == 0)
    stop("Unit of output var missing")
  if(length(unit_str) > 1)
    stop("Output variable is not unique")
  unit_str <- stringr::str_extract(unit_str, "\\((.*?)\\)", group=1)
  if(is.na(unit_str) | unit_str == "")
    stop("Unit of output var missing")
  # replace any '.' by a space to avoid issues with `units` package
  unit_str <- stringr::str_replace(unit_str, "\\.", " ")

  # is output var included in timerseries?
  data <- data %>%
    dplyr::filter(stringr::str_detect(tolower(V3), pattern=paste0("^", tolower(output_var), "(_\\w+)?$"))) %>%
    dplyr::select(dplyr::where(~ sum(!is.na(.x)) > 0)) %>% # remove any columns containing only NA
    dplyr::select(dplyr::where(~ sum(.x != "") > 0)) # remove any columns containing only emtpy strings
  if(nrow(data) == 0)
    stop("No data for output variable in TOXSWA file")

  data <- data %>%
    dplyr::select(c(dplyr::any_of(c(1, 2, 3)), dplyr::last_col())) %>%
    dplyr::rename(time=1, timestamp=2, name=3, value=4) %>%
    dplyr::mutate(value=units::set_units(as.numeric(value), unit_str, mode="standard"))

  # convert units of output if requested
  if(!is.null(output_unit)) {
    units(data$value) <- units::as_units(output_unit)
  }

  # check if variable is substance-dependent (which it should be), and
  # pivot by substance
  if(any(stringr::str_detect(data$name, "_")))
  {
    data <- data %>%
      dplyr::mutate(name=stringr::str_extract(name, "_(.+)$", group=1)) %>%
      tidyr::pivot_wider()
  }
  else if(!is.null(substance)) {
    stop("Output var is not substance dependent, check inputs")
  }

  # is chosen substance part of result?
  if(!is.null(substance)) {
    if(!all(substance %in% names(data)))
      stop("Selected substance code not found in file")

    data <- data %>% dplyr::select(c(1, 2, !!substance))
  }

  data <- data %>%
    dplyr::mutate(time=as.numeric(time),
                  time=time - time[[1]],
                  time=units::set_units(time, "days"),
                  time=units::set_units(time, time_unit, mode="standard"),
                  timestamp=lubridate::dmy_hm(timestamp))
  # set unit information as attribute for backwards compat
  myunits <- c("time"=units::deparse_unit(data$time),
               "conc"=units::deparse_unit(as.data.frame(data)[,3]))
  attr(data, "units") <- myunits

  # use file w/o extension as fallback option, otherwise use supplied alias
  if(is.na(alias))
    fn <- tools::file_path_sans_ext(basename(files))
  else
    fn <- alias

  result <- list()
  # split series by substance?
  if(split) {
    nms <- names(data)
    for(i in seq(3, length(data))) {
      key <- paste0(fn, "_", nms[[i]])
      result[[key]] <- dplyr::select(data, dplyr::all_of(c(1, 2, i)))
    }
  } else {
    result[[fn]] <- data
  }
  result
}

test_that("basic import", {
  fn <- testthat::test_path(file.path("../data/TOXSWA", "testonil.out"))

  rs <- import_toxswa(fn, time_unit="days", output_unit="ug/L", split=FALSE)
  expect_equal(length(rs), 1)
  expect_equal(names(rs), "testonil")
  expect_equal(length(rs[[1]]), 4)
  expect_equal(names(rs[[1]]), c("time","timestamp","substance","metabolite"))
  expect_equal(as.numeric(rs[[1]]$time), c(0, 0.041, 0.083, 0.125, 0.166))
  expect_equal(rs[[1]]$timestamp[1], lubridate::ymd_hms("1992-01-01 00:30:00"))
  expect_equal(rs[[1]]$timestamp[2], lubridate::ymd_hms("1992-01-01 01:30:00"))
  expect_equal(as.numeric(rs[[1]]$substance), c(2000, 0, 0, 0, 0))
  expect_equal(as.numeric(rs[[1]]$metabolite), c(4000, 0, 0, 0, 0))
  expect_equal(units::deparse_unit(rs[[1]]$time), "d")
  expect_equal(units::deparse_unit(rs[[1]]$substance), "ug L-1")
  expect_equal(units::deparse_unit(rs[[1]]$metabolite), "ug L-1")

  # multiple files
  rs <- import_toxswa(c(fn, fn), split=FALSE)
  expect_equal(length(rs), 2)
  expect_equal(names(rs), c("testonil", "testonil"))
  expect_equal(rs[[1]], rs[[2]])

  # no file
  expect_error(import_toxswa(NULL))
  expect_error(import_toxswa("foo/bar.baz"))
})

test_that("unit conversion", {
  fn <- testthat::test_path(file.path("../data/TOXSWA", "testonil.out"))
  # time units
  rsd <- import_toxswa(fn, time_unit="days", split=FALSE)
  rsh <- import_toxswa(fn, time_unit="hours", split=FALSE)
  expect_equal(as.numeric(rsd[[1]]$time) * 24,
               as.numeric(rsh[[1]]$time))
  # output var units
  rsm <- import_toxswa(fn, output_unit="mg/L", split=FALSE)
  rsu <- import_toxswa(fn, output_unit="ug/L", split=FALSE)
  expect_equal(as.numeric(rsm[[1]]$substance) * 1000,
               as.numeric(rsu[[1]]$substance))
  expect_equal(as.numeric(rsm[[1]]$metabolite) * 1000,
               as.numeric(rsu[[1]]$metabolite))

  # invalid unit
  expect_error(import_toxswa(fn, output_unit="foo"))
})

test_that("filter substance", {
  fn <- testthat::test_path(file.path("../data/TOXSWA", "testonil.out"))

  rs <- import_toxswa(fn, substance="substance", split=FALSE)
  expect_equal(names(rs[[1]]), c("time", "timestamp", "substance"))
  rs <- import_toxswa(fn, substance="metabolite", split=FALSE)
  expect_equal(names(rs[[1]]), c("time", "timestamp", "metabolite"))

  expect_error(import_toxswa(fn, substance="foo"))
})

test_that("split by substance", {
  fn <- testthat::test_path(file.path("../data/TOXSWA", "testonil.out"))

  rs <- import_toxswa(fn, split=TRUE)
  expect_equal(length(rs), 2)
  expect_equal(names(rs[[1]]), c("time", "timestamp", "substance"))
  expect_equal(names(rs[[2]]), c("time", "timestamp", "metabolite"))
})

test_that("aliases", {
  fn <- testthat::test_path(file.path("../data/TOXSWA", "testonil.out"))

  rs <- import_toxswa(fn, alias="foo", split=FALSE)
  expect_equal(length(rs), 1)
  expect_equal(names(rs), "foo")

  rs <- import_toxswa(fn, alias="foo", split=TRUE)
  expect_equal(length(rs), 2)
  expect_equal(names(rs), c("foo_substance", "foo_metabolite"))
})

test_that("invalid file contents", {
  # no expected contents
  withr::with_file("file1", {
    writeLines("foo", "file1")
    expect_error(import_toxswa("file1"), "version missing")
  })

  # unsupported TOXSWA version
  withr::with_file("file1", {
    writeLines("* FOCUS  TOXSWA version   : 99", "file1")
    expect_error(import_toxswa("file1"), "version not supported")
  })

  # output unit missing
  withr::with_file("file1", {
    writeLines("* FOCUS  TOXSWA version   : 4", "file1")
    expect_error(import_toxswa("file1"), "Unit of output var missing")
  })

  # output unit not found
  withr::with_file("file1", {
    writeLines(c("* FOCUS  TOXSWA version   : 4",
                 "* Unit for ConLiqWatLay is ()"), "file1")
    expect_error(import_toxswa("file1"), "Unit of output var missing")
  })

  # output unit not unique
  withr::with_file("file1", {
    writeLines(c("* FOCUS  TOXSWA version   : 4",
                 "* Unit for ConLiqWatLay is (g.m-3)",
                 "* Unit for ConLiqWatLay is (kg.m-3)"), "file1")
    expect_error(import_toxswa("file1"), "Output variable is not unique")
  })

  # no data for output
  withr::with_file("file1", {
    writeLines(c("* FOCUS  TOXSWA version   : 4",
                 "* Unit for ConLiqWatLay is (g.m-3)"), "file1")
    expect_error(import_toxswa("file1"), "No data")
  })
})

test_that("large number of columns", {
  # It's a frequent issue with most functions such as read.delim(), readr::read_*
  # that not all columns are actually imported into the same data.frame row
  withr::with_file("file1", {
    writeLines(c("* FOCUS  TOXSWA version   : 4",
                 "* Unit for ConLiqWatLay is (g.m-3)",
                 " 0  01-Jan-1992-00h30 ConLiqWatLay_par 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 42"), "file1")
    rs <- import_toxswa("file1", output_unit="g/m3")
    expect_equal(as.numeric(rs[[1]]$par), 42)
  })
})

########################
## Class definitions
########################
#' @export
setClass("Guts", contains="EffectScenario",
         slots=list("dose_metric" = "character", "scaled_ci" = "logical"),
         prototype=list("dose_metric" = "D", "scaled_ci" = FALSE)
)

#' @export
setClass("GutsSd", contains="Guts")

#' @export
setClass("GutsIt", contains="Guts")


########################
## Constructors
########################

#' GUTS-SD scenario
#'
#' `r lifecycle::badge("experimental")`<br>
#' Full *General Unified Threshold models of Survival* (GUTS) with stochastic
#' death (*SD*). The model was defined by Jager et al. (2011). It is
#' compatible with the *Full GUTS* model as described by EFSA (2018), but
#' some parameter names may differ.
#'
#' @section State variables:
#' The following list describes the default names and standard units of *GUTS*
#' state variables:
#' * `Ci`, (scaled) internal concentration (conc)
#' * `D`, (scaled) damage (*)
#' * `H`, cumulative hazard (-)
#'
#' The state variables are initialized with zero by default.
#'
#' @section SD model parameters:
#' The set of parameters and their names follows the definition by Jager et al
#' (2011). The actual number of required parameters depends on the selected model
#' variant, i.e. if the internal concentration is scaled or not, as well as the
#' selected dose metric. The full set of parameters is as follows
#'
#' * `ki`, accumulation rate into body (time^-1)
#' * `ke`, elimination rate (time^-1)
#' * `Kiw`, scaling constant for external concentration (*)
#' * `kr`, damage recovery rate (time^-1)
#' * `kk`, killing rate constant (time^-1)
#' * `hb`, background hazard rate (time^-1)
#' * `z`, threshold for effects (*)
#'
#'
#' @section Effects:
#' The effect endpoint `L` (lethality) is available for *GUTS* models.
#' A value of zero (`0.0`) denotes *no effect* on organism survival. A value of
#' one (`1.0`) denotes a lethality rate of 100%, i.e. no survivors.
#'
#' The survival probability `S` is available in the return value of [simulate()].
#'
#' @references
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#' Jager T., Albert C., Preuss T.G., and Ashauer R., 2011: General Unified Threshold
#' Model of Survival - a Toxicokinetic-Toxicodynamic Framework for Ecotoxicology.
#' Environ. Sci. Technol. 45(7), pp. 2529-2540. \doi{10.1021/es103092a}
#'
#'
#' @param scaled_ci `logical`, switch to enable scaling. If `TRUE`, the model
#'    will use the scaled internal concentration `Ci*`. Else no scaling. Default
#'    value is `FALSE`.
#' @param dose_metric `character`, selects the dose metric. `'D'` to use damage, `'Ci'` to
#'   for internal concentration, or `'Cw'` for the external concentration. Default
#'   value is `'D'`.
#' @return an S4 object of type `GutsSd-class`
#' @export
#' @family GUTS models
#' @aliases Guts-class GutsSd-class GUTS-SD
#' @seealso [GUTS-RED-SD]
#' @importFrom methods new
GUTS_SD <- function(scaled_ci=FALSE, dose_metric=c("D", "Ci", "Cw")) {
  # Check arguments
  if(!is.logical(scaled_ci)) {
    stop("Argument `scaled_ci` must be a logical value")
  }
  dose_metric <- match.arg(dose_metric)
  # Plausibility check
  if(scaled_ci & dose_metric == "Cw") {
    stop("Internal concentration `Ci` cannot be scaled, if dose metric M=Cw")
  }

  init <- c(Ci=NA_real_, D=NA_real_, H=0)
  preq <- c("kk", "hb", "z")

  # Parameters required for (scaled) damage ODE
  if(dose_metric == "D") {
    preq <- c("kr", preq)
    init["D"] <- 0
  }

  # Parameters required for internal concentration ODE
  if(scaled_ci) { # scaled Ci
    preq <- c("ke", "Kiw", preq)
    init["Ci"] <- 0
  } else if(dose_metric != "Cw") { # Any other case, unless Ci disabled
    preq <- c("ki", "ke", preq)
    init["Ci"] <- 0
  }

  param <- as.list(setNames(rep(NA_real_, length(preq)), preq))
  if(scaled_ci) {
    param["Kiw"] <- 1
  }

  new("GutsSd",
      name = paste0("GUTS-SD (", ifelse(scaled_ci, "scaled Ci, ", ""), "M=", dose_metric, ")"),
      scaled_ci = scaled_ci,
      dose_metric = dose_metric,
      param.req = preq,
      param = param,
      endpoints = c("L"),
      init = init,
      control.req=FALSE
  ) %>% set_noexposure()
}

# TODO currently only implements exponentially distributed individual thresholds z
#   but other types such as log-normal or log-logistic may be necessary,
#   cf Jager et al (2011), Table 3
#' GUTS-IT scenario
#'
#' `r lifecycle::badge("experimental")`<br>
#' Full *General Unified Threshold models of Survival* (GUTS) with Individual Tolerance
#  Distribution (*IT*). The model was defined by Jager et al. (2011). It is
#' compatible with the *Full GUTS* model as described by EFSA (2018), but
#' some parameter names may differ.
#'
#' @section State variables:
#' The following list describes the default names and standard units of *GUTS*
#' state variables:
#' * `Ci`, (scaled) internal concentration (conc)
#' * `D`, (scaled) damage (*)
#' * `H`, cumulative hazard (-)
#'
#' The state variables are initialized with zero by default.
#'
#' @section IT model parameters:
#' The set of parameters and their names follows the definition by Jager et al
#' (2011). The actual number of required parameters depends on the selected model
#' variant, i.e. if the internal concentration is scaled or not, as well as the
#' selected dose metric. The full set of parameters is as follows
#'
#' * `ki`, accumulation rate into body (time^-1)
#' * `ke`, elimination rate (time^-1)
#' * `Kiw`, scaling constant for external concentration (*)
#' * `kr`, damage recovery rate (time^-1)
#' * `hb`, background hazard rate (time^-1)
#' * `alpha`, median of thresholds (conc)
#' * `beta`, shape parameter (-)
#'
#' @section Effects:
#' The effect endpoint `L` (lethality) is available for *GUTS* models.
#' A value of zero (`0.0`) denotes *no effect* on organism survival. A value of
#' one (`1.0`) denotes a lethality rate of 100%, i.e. no survivors.
#'
#' The survival probability `S` is available in the return value of [simulate()].
#'
#' @references
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#' Jager T., Albert C., Preuss T.G., and Ashauer R., 2011: General Unified Threshold
#' Model of Survival - a Toxicokinetic-Toxicodynamic Framework for Ecotoxicology.
#' Environ. Sci. Technol. 45(7), pp. 2529-2540. \doi{10.1021/es103092a}
#'
#'
#' @param scaled_ci `logical`, switch to enable scaling. If `TRUE`, the model
#'    will use the scaled internal concentration `Ci*`. Else no scaling. Default
#'    value is `FALSE`.
#' @param dose_metric `character`, selects the dose metric. `'D'` to use damage, `'Ci'` to
#'   for internal concentration, or `'Cw'` for the external concentration. Default
#'   value is `'D'`.
#' @return an S4 object of type `GutsIt-class`
# @export
#' @family GUTS models
#' @aliases GutsIt-class GUTS-IT
#' @seealso [GUTS-RED-IT]
#' @importFrom methods new
GUTS_IT <- function(scaled_ci=FALSE, dose_metric=c("D", "Ci", "Cw")) {
  # Check arguments
  if(!is.logical(scaled_ci)) {
    stop("Argument `scaled_ci` must be a logical value")
  }
  dose_metric <- match.arg(dose_metric)
  # Plausibility check
  if(scaled_ci & dose_metric == "Cw") {
    stop("Internal concentration `Ci` cannot be scaled, if dose metric M=Cw")
  }

  init <- c(Ci=NA_real_, D=NA_real_, H=0)
  preq <- c("hb", "alpha", "beta")

  # Parameters required for (scaled) damage ODE
  if(dose_metric == "D") {
    preq <- c("kr", preq)
    init["Ci"] <- 0
    init["D"] <- 0
  }

  # Parameters required for internal concentration ODE
  if(scaled_ci) { # scaled Ci
    preq <- c("ke", "Kiw", preq)
    init["Ci"] <- 0
  } else if(dose_metric != "Cw") { # Any other case, unless Ci disabled
    preq <- c("ki", "ke", preq)
    init["Ci"] <- 0
  }

  param <- as.list(setNames(rep(NA_real_, length(preq)), preq))
  if(scaled_ci) {
    param["Kiw"] <- 1
  }

  new("GutsIt",
      name = paste0("GUTS-IT (", ifelse(scaled_ci, "scaled Ci, ", ""), "M=", dose_metric, ")"),
      scaled_ci = scaled_ci,
      dose_metric = dose_metric,
      param.req = preq,
      param = param,
      endpoints = c("L"),
      init = init,
      control.req=FALSE
  ) %>% set_noexposure()
}

########################
## Simulation
########################

# @param scenario Scenario object
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param method string, numerical solver used by [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_gutssd <- function(scenario, method="ode45", hmax=0.1, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)
  init <- scenario@init

  # Convert parameters for scaled Ci, if necessary
  if(scenario@scaled_ci) {
    params["ki"] = params["ke"] * params["Kiw"]
  }

  # Set dose_metric switch value
  if(scenario@dose_metric == "D") {
    dm_val <- 0
  } else if(scenario@dose_metric == "Ci") {
    dm_val <- 1
    init[["D"]] <- NA_real_
  } else if(scenario@dose_metric == "Cw") {
    dm_val <- 2
    init[["Ci"]] <- NA_real_
    init[["D"]] <- NA_real_
  } else {
    stop("Slot `dose_metric` has unsupported value")
  }
  params["dose_metric"] <- dm_val

  # check if all required parameters are present
  params_missing <- is.na(params[scenario@param.req])
  if(any(params_missing)) {
    stop("Missing parameters: ", paste(scenario@param.req[params_missing], collapse=", "))
  }

  # make sure that parameters are present and in required order
  param_order <- c("ki", "ke", "kr", "kk", "hb", "z", "dose_metric")
  # not all parameters expected by the model's C code may be present, make sure
  # that all parameters are present and initialized with `NA`
  params <- setNames(params[param_order], param_order)

  df <- ode2df(ode(y=init, times=scenario@times, parms=params, dllname="cvasi",
                   initfunc="gutssd_init", func="gutssd_func", initforc="gutssd_forc",
                   forcings=scenario@exposure@series, outnames=c("Cw", "M"),
                   method=method, hmax=hmax, ...))

  # Derive survival probability, refs:
  # Jager et al (2011) DOI: 10.1021/es103092a
  # EFSA Scientific Opinion, p. 33, DOI: 10.2903/j.efsa.2018.5377
  df$S <- exp(-df$H) # background hazard rate `hb` included in state variable `H` (if enabled)
  df
}

solver_gutsit <- function(scenario, method="ode45", hmax=0.1, nout=0, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)
  init <- scenario@init

  # Convert parameters for scaled Ci, if necessary
  if(scenario@scaled_ci) {
    params["ki"] = params["ke"] * params["Kiw"]
  }

  # Set dose_metric switch value
  if(scenario@dose_metric == "D") {
    dm_val <- 0
    init[["CMax"]] <- init[["D"]]
  } else if(scenario@dose_metric == "Ci") {
    dm_val <- 1
    init[["CMax"]] <- init[["Ci"]]
    init[["D"]] <- NA_real_
  } else if(scenario@dose_metric == "Cw") {
    dm_val <- 2
    init[["CMax"]] <- 0
    init[["Ci"]] <- NA_real_
    init[["D"]] <- NA_real_
    # activate Cw as additional output of the solver
    nout <- max(nout, 1)
  } else {
    stop("Slot `dose_metric` has unsupported value")
  }
  params["dose_metric"] <- dm_val

  # check if all required parameters are present
  params_missing <- is.na(params[scenario@param.req])
  if(any(params_missing)) {
    stop("Missing parameters: ", paste(scenario@param.req[params_missing], collapse=", "))
  }

  # make sure that parameters are present and in required order
  param_order <- c("ki", "ke", "kr", "hb", "dose_metric")
  # not all parameters expected by the model's C code may be present, make sure
  # that all parameters are present and initialized with `NA`
  params_ode <- setNames(params[param_order], param_order)

  df <- ode2df(ode(y=init, times=scenario@times, parms=params_ode, dllname="cvasi",
                   initfunc="gutsit_init", func="gutsit_func", initforc="gutsit_forc",
                   forcings=scenario@exposure@series, outnames=c("Cw", "M"), nout=nout,
                   method=method, hmax=hmax, ...))

  # `CMax` approximates `cummax(M)` but may underestimate its value in case of
  # M = Cw, because Cw has no known derivative. Therefor we apply a *parallel*
  # maximum on `M` and `CMax`. However, if there too few output time points this
  # might not improve the result.
  if(scenario@dose_metric == "Cw") {
    df$CMax <- pmax(df$CMax, df$Cw)
  }
  # None of the dose metrics can, by definition, become negative. However, the
  # numerical schemes may sometimes return negative values due to interpolation.
  df$CMax <- pmax(0, df$CMax)

  # Derive survival probability, refs:
  # Jager et al (2011) DOI: 10.1021/es103092a
  # EFSA Scientific Opinion, p. 33, DOI: 10.2903/j.efsa.2018.5377
  FS <- (1 / (1 + (df$CMax / params[["alpha"]])^(-params[["beta"]])))
  df$S <- (1 - FS) * exp(-df$H)
  df$CMax <- NULL
  df
}

#' @include solver.R
#' @describeIn solver Numerically integrates GUTS-SD scenarios
setMethod("solver", "GutsSd", function(scenario, ...) solver_gutssd(scenario, ...))

#' @include solver.R
#' @describeIn solver Numerically integrates GUTS-IT scenarios
setMethod("solver", "GutsIt", function(scenario, ...) solver_gutsit(scenario, ...))

#' @include fx.R
#' @describeIn fx Calculates lethality of [GUTS-SD] scenarios
setMethod("fx", "GutsSd", function(scenario, ...) fx_gutsredsd(scenario, ...))

#' @include fx.R
#' @describeIn fx Calculates lethality of [GUTS-IT] scenarios
setMethod("fx", "GutsIt", function(scenario, ...) fx_gutsredit(scenario, ...))

test_that("Constructor", {
  expect_equal(GUTS_SD(scaled_ci=TRUE)@scaled_ci, TRUE)
  expect_equal(GUTS_SD(scaled_ci=FALSE)@scaled_ci, FALSE)
  expect_equal(GUTS_SD(dose_metric="D")@dose_metric, "D")
  expect_equal(GUTS_SD(dose_metric="Ci")@dose_metric, "Ci")
  expect_equal(GUTS_SD(dose_metric="Cw")@dose_metric, "Cw")
  expect_equal(is(GUTS_SD())[1], "GutsSd")

  expect_equal(GUTS_IT(scaled_ci=TRUE)@scaled_ci, TRUE)
  expect_equal(GUTS_IT(scaled_ci=FALSE)@scaled_ci, FALSE)
  expect_equal(GUTS_IT(dose_metric="D")@dose_metric, "D")
  expect_equal(GUTS_IT(dose_metric="Ci")@dose_metric, "Ci")
  expect_equal(GUTS_IT(dose_metric="Cw")@dose_metric, "Cw")
  expect_equal(is(GUTS_IT())[1], "GutsIt")
})

test_that("SD: parameters", {
  #
  # Non-scaled Ci
  #
  expect_equal(GUTS_SD(scaled_ci=FALSE, dose_metric="D")@param.req, c("ki", "ke", "kr", "kk", "hb", "z"))
  expect_equal(GUTS_SD(scaled_ci=FALSE, dose_metric="Ci")@param.req, c("ki", "ke", "kk", "hb", "z"))
  expect_equal(GUTS_SD(scaled_ci=FALSE, dose_metric="Cw")@param.req, c("kk", "hb", "z"))

  expect_equal(GUTS_IT(scaled_ci=FALSE, dose_metric="D")@param.req, c("ki", "ke", "kr", "hb", "alpha", "beta"))
  expect_equal(GUTS_IT(scaled_ci=FALSE, dose_metric="Ci")@param.req, c("ki", "ke", "hb", "alpha", "beta"))
  expect_equal(GUTS_IT(scaled_ci=FALSE, dose_metric="Cw")@param.req, c("hb", "alpha", "beta"))

  #
  # Scaled Ci
  #
  expect_equal(GUTS_SD(scaled_ci=TRUE, dose_metric="D")@param.req, c("ke", "Kiw", "kr", "kk", "hb", "z"))
  expect_equal(GUTS_SD(scaled_ci=TRUE, dose_metric="Ci")@param.req, c("ke", "Kiw", "kk", "hb", "z"))
  expect_equal(GUTS_IT(scaled_ci=TRUE, dose_metric="D")@param.req, c("ke", "Kiw", "kr", "hb", "alpha", "beta"))
  expect_equal(GUTS_IT(scaled_ci=TRUE, dose_metric="Ci")@param.req, c("ke", "Kiw", "hb", "alpha", "beta"))
  # default value for Kiw
  expect_equal(GUTS_SD(scaled_ci=TRUE, dose_metric="D")@param[["Kiw"]], 1)
  expect_equal(GUTS_SD(scaled_ci=TRUE, dose_metric="Ci")@param[["Kiw"]], 1)
  expect_equal(GUTS_IT(scaled_ci=TRUE, dose_metric="D")@param[["Kiw"]], 1)
  expect_equal(GUTS_IT(scaled_ci=TRUE, dose_metric="Ci")@param[["Kiw"]], 1)
  # invalid combination of arguments
  expect_error(GUTS_SD(scaled_ci=TRUE, dose_metric="Cw"))
  expect_error(GUTS_IT(scaled_ci=TRUE, dose_metric="Cw"))
})

# Tests the basic behavior of the model, checks for plausible changes in state
# variables and concrete values where possible
# Basis for tests: GUTS-SD with non-scaled Ci, and dose metric M=D
test_that("SD: solver", {
  tol <- 1e-5
  base <- GUTS_SD(scaled_ci=FALSE, dose_metric="D") %>%
    set_param(c(ki=1, ke=1, kr=1, kk=1, z=1, hb=0)) %>%
    set_noexposure() %>%
    set_times(0:5)

  # no exposure -> no damage, hazard, etc
  rs <- simulate(base)
  expect_equal(names(rs), c("time","Ci","D","H","S"))
  expect_equal(rs$time, 0:5)
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # no exposure, but background mortality
  rs <- base %>% set_param(c(hb=1)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 0))
  expect_equal(rs$H, 0:5, tolerance=tol)
  expect_equal(rs$S, exp(-(0:5)), tolerance=tol)

  # update base scenario, disable all processes by default
  base <- base %>%
    set_param(c(ki=0, ke=0, kr=0, kk=0, z=0, hb=0)) %>%
    set_exposure(data.frame(time=0:5, conc=1))

  # with exposure, uptake only (ki > 0)
  rs <- base %>% set_param(c(ki=1)) %>% simulate()
  expect_equal(rs$Ci, 0:5, tolerance=tol)
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # no exposure, elimination only, (ke > 0)
  rs <- base %>% set_init(c(Ci=1)) %>% set_noexposure() %>% set_param(c(ke=1)) %>% simulate()
  expect_equal(rs$Ci, exp(-(0:5)), tolerance=tol)
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # with exposure, uptake and elimination (ki, ke > 0)
  rs <- base %>% set_param(c(ki=1)) %>% simulate()
  expect_true(all(diff(rs$Ci) > 0)) # strictly increasing Ci
  expect_true(all(rs$Ci <= 0:5)) # Ci smaller compared to scenario w/o elimination
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # no exposure, repair only (kr > 0)
  rs <- base %>% set_init(c(D=1)) %>% set_noexposure() %>% set_param(c(kr=1)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_equal(rs$D, exp(-(0:5)), tolerance=tol)
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # with exposure, uptake and repair (ki, kr > 0)
  rs <- base %>% set_param(c(ki=1, kr=1)) %>% simulate()
  expect_true(all(diff(rs$Ci) > 0)) # continuously increasing Ci
  expect_true(all(diff(rs$D) > 0))  # continuously increasing D
  expect_true(all(rs$D <= rs$Ci))   # D strictly smaller than Ci (due to repair)
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # constant damage, killing rate only (kk > 0)
  rs <- base %>% set_init(c(D=1)) %>% set_param(c(kk=1)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 1))
  expect_equal(rs$H, 0:5, tolerance=tol)
  expect_equal(rs$S, exp(-(0:5)), tolerance=tol)

  # constant damage, killing rate, threshold set (kk == z, kk, z > 0)
  rs <- base %>% set_init(c(D=1)) %>% set_param(c(kk=1, z=1)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 1))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # dose metric: Ci
  rs <- GUTS_SD(scaled_ci=FALSE, dose_metric="Ci") %>%
    set_init(c(Ci=1)) %>%
    set_param(c(ki=0, ke=0, kk=1, z=0, hb=0)) %>%
    set_noexposure() %>%
    set_times(0:5) %>%
    simulate()
  expect_true(all(rs$Ci == 1))
  expect_true(all(is.na(rs$D)))
  expect_equal(rs$H, 0:5, tolerance=tol)
  expect_equal(rs$S, exp(-(0:5)), tolerance=tol)

  # dose metric: Cw
  rs <- GUTS_SD(scaled_ci=FALSE, dose_metric="Cw") %>%
    set_init(c(D=1)) %>%
    set_param(c(kk=1, z=0, hb=0)) %>%
    set_exposure(data.frame(t=0, c=1), reset_times=FALSE) %>%
    set_times(0:5) %>%
    simulate()

  expect_true(all(is.na(rs$Ci)))
  expect_true(all(is.na(rs$D)))
  expect_equal(rs$H, 0:5, tolerance=tol)
  expect_equal(rs$S, exp(-(0:5)), tolerance=tol)
})

test_that("IT: solver", {
  tol <- 1e-5
  base <- GUTS_IT(scaled_ci=FALSE, dose_metric="D") %>%
    set_param(c(ki=1, ke=1, kr=1, alpha=1, beta=1, hb=0)) %>%
    set_noexposure() %>%
    set_times(0:5)

  # no exposure -> no damage, hazard, etc
  rs <- simulate(base)
  expect_equal(names(rs), c("time","Ci","D","H","S"))
  expect_equal(rs$time, 0:5)
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # no exposure, but background mortality
  rs <- base %>% set_param(c(hb=1)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 0))
  expect_equal(rs$H, 0:5, tolerance=tol)
  expect_equal(rs$S, exp(-(0:5)), tolerance=tol)

  # update base scenario, disable all processes by default
  base <- base %>%
    set_param(c(ki=0, ke=0, kr=0, alpha=1, beta=1, hb=0)) %>%
    set_exposure(data.frame(time=0:5, conc=1))

  # with exposure, uptake only (ki > 0)
  rs <- base %>% set_param(c(ki=1)) %>% simulate()
  expect_equal(rs$Ci, 0:5, tolerance=tol)
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # no exposure, elimination only, (ke > 0)
  rs <- base %>% set_init(c(Ci=1)) %>% set_noexposure() %>% set_param(c(ke=1)) %>% simulate()
  expect_equal(rs$Ci, exp(-(0:5)), tolerance=tol)
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # with exposure, uptake and elimination (ki, ke > 0)
  rs <- base %>% set_param(c(ki=1)) %>% simulate()
  expect_true(all(diff(rs$Ci) > 0)) # strictly increasing Ci
  expect_true(all(rs$Ci <= 0:5)) # Ci smaller compared to scenario w/o elimination
  expect_true(all(rs$D == 0))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 1))

  # no exposure, repair only (kr > 0)
  rs <- base %>% set_init(c(D=1)) %>% set_noexposure() %>% set_param(c(kr=1)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_equal(rs$D, exp(-(0:5)), tolerance=tol)
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 0.5))

  # with exposure, uptake and repair (ki, kr > 0)
  rs <- base %>% set_param(c(ki=1, kr=1)) %>% simulate()
  expect_true(all(diff(rs$Ci) > 0)) # continuously increasing Ci
  expect_true(all(diff(rs$D) > 0))  # continuously increasing D
  expect_true(all(rs$D <= rs$Ci))   # D strictly smaller than Ci (due to repair)
  expect_true(all(rs$H == 0))
  expect_true(all(diff(rs$S) < 0))  # strictly decreasing

  # scale the median of thresholds, alpha
  rs <- base %>% set_init(c(D=1)) %>% set_param(c(alpha=2)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 1))
  expect_true(all(rs$H == 0))
  expect_equal(rs$S, rep(2/3, nrow(rs)), tolerance=tol)

  # scale the shape factor, beta
  rs <- base %>% set_init(c(D=1)) %>% set_param(c(beta=2)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 1))
  expect_true(all(rs$H == 0))
  expect_true(all(rs$S == 0.5))

  # scale alpha and beta
  rs <- base %>% set_init(c(D=1)) %>% set_param(c(alpha=2, beta=2)) %>% simulate()
  expect_true(all(rs$Ci == 0))
  expect_true(all(rs$D == 1))
  expect_true(all(rs$H == 0))
  expect_equal(rs$S, rep(1 - 1 / (1 + (1/2)^-2), nrow(rs)))

  # dose metric: Ci
  rs <- GUTS_IT(scaled_ci=FALSE, dose_metric="Ci") %>%
    set_init(c(Ci=1)) %>%
    set_param(c(ki=0, ke=0, alpha=1, beta=1, hb=0)) %>%
    set_noexposure() %>%
    set_times(0:5) %>%
    simulate()
  expect_true(all(rs$Ci == 1))
  expect_true(all(is.na(rs$D)))
  expect_true(all(rs$H == 0))
  expect_equal(rs$S, rep(0.5, nrow(rs)), tolerance=tol)

  # dose metric: Cw
  rs <- GUTS_IT(scaled_ci=FALSE, dose_metric="Cw") %>%
    set_init(c(D=1)) %>%
    set_param(c(alpha=1, beta=1, hb=0)) %>%
    set_exposure(data.frame(t=0, c=1), reset_times=FALSE) %>%
    set_times(0:5) %>%
    simulate()

  expect_true(all(is.na(rs$Ci)))
  expect_true(all(is.na(rs$D)))
  expect_true(all(rs$H == 0))
  expect_equal(rs$S, rep(0.5, nrow(rs)), tolerance=tol)
})

test_that("SD: parameters passed to solver", {
  # set up mocked deSolve function
  capture <- list()
  myode <- function(y, times, parms, ...) {
    capture$y <<- y
    capture$times <<- times
    capture$parms <<- parms
    data.frame(H=0)
  }

  # scenario: non-scaled Ci, M=D
  sc <- GUTS_SD(scaled_ci=FALSE, dose_metric="D") %>%
    set_param(c(ki=1, ke=1, kr=1, kk=1, hb=0, z=1)) %>%
    set_noexposure() %>%
    set_times(0:1)

  # execute
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$y, sc@init)
  expect_equal(capture$times, sc@times)
  expect_equal(capture$parms, c(unlist(sc@param), c("dose_metric"=0)))

  # scenario: non-scaled Ci, M=Ci
  sc <- GUTS_SD(scaled_ci=FALSE, dose_metric="Ci") %>%
    set_param(c(ki=1, ke=1, kk=1, hb=0, z=1))
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$parms[["dose_metric"]], 1)

  # scenario: non-scaled Ci, M=Cw
  sc <- GUTS_SD(scaled_ci=FALSE, dose_metric="Cw") %>%
    set_param(c(kk=1, hb=0, z=1))
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$parms[["dose_metric"]], 2)

  # scenario: scaled Ci, M=D
  sc <- GUTS_SD(scaled_ci=TRUE, dose_metric="D") %>%
    set_param(c(ke=3, Kiw=4, kr=1, kk=1, hb=0, z=1))
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$parms[["ke"]], 3)
  expect_equal(capture$parms[["ki"]], 3 * 4) # Kiw := ki/ke
  expect_equal(capture$parms[["dose_metric"]], 0)
})

test_that("IT: parameters passed to solver", {
  # set up mocked deSolve function
  capture <- list()
  myode <- function(y, times, parms, ...) {
    capture$y <<- y
    capture$times <<- times
    capture$parms <<- parms
    data.frame(H=0, CMax=0, Cw=0)
  }

  # scenario: non-scaled Ci, M=D
  sc <- GUTS_IT(scaled_ci=FALSE, dose_metric="D") %>%
    set_param(c(ki=1, ke=1, kr=1, hb=0, alpha=1, beta=1)) %>%
    set_noexposure() %>%
    set_times(0:1)
  myodeparam <- unlist(sc@param[c("ki","ke","kr","hb")])
  myinit <- c(sc@init, "CMax"=0)

  # execute
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$y, myinit)
  expect_equal(capture$times, sc@times)
  expect_equal(capture$parms, c(myodeparam, c("dose_metric"=0)))

  # scenario: non-scaled Ci, M=Ci
  sc <- GUTS_IT(scaled_ci=FALSE, dose_metric="Ci") %>%
    set_param(c(ki=1, ke=1, hb=0, alpha=1, beta=1)) %>%
    set_noexposure() %>%
    set_times(0:1)
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$parms[["dose_metric"]], 1)

  # scenario: non-scaled Ci, M=Cw
  sc <- GUTS_IT(scaled_ci=FALSE, dose_metric="Cw") %>%
    set_param(c(hb=0, alpha=1, beta=1)) %>%
    set_noexposure() %>%
    set_times(0:1)
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$parms[["dose_metric"]], 2)

  # scenario: scaled Ci, M=D
  sc <- GUTS_IT(scaled_ci=TRUE, dose_metric="D") %>%
    set_param(c(ke=3, Kiw=4, kr=5, hb=0, alpha=1, beta=1))
  with_mocked_bindings(
    solver(sc),
    ode = myode
  )
  expect_equal(capture$parms[["ke"]], 3)
  expect_equal(capture$parms[["ki"]], 3 * 4) # Kiw := ki/ke
  expect_equal(capture$parms[["dose_metric"]], 0)
})

test_that("SD: effects", {
  sc  <- GUTS_SD(scaled_ci=FALSE, dose_metric="D") %>%
    set_init(c(Ci=1, D=0.5, H=0.2)) %>%
    set_param(c(ki=1, ke=1, kr=1, kk=1, z=0.3, hb=0)) %>%
    set_noexposure() %>%
    set_times(0:5)

  expect_equal(
    effect(sc, ep_only=TRUE, max_only=FALSE)$L, # lethality calculated by effect/fx
    1 - tail(simulate(sc), n=1)$S,              # lethality calculated manually
    ignore_attr=TRUE,
    tolerance=1e-5
  )
})

test_that("IT: effects", {
  sc  <- GUTS_IT(scaled_ci=FALSE, dose_metric="D") %>%
    set_init(c(Ci=1, D=0.5, H=0.2)) %>%
    set_param(c(ki=0, ke=0, kr=0, hb=0, alpha=1, beta=1)) %>%
    set_noexposure() %>%
    set_times(0:1)

  expect_equal(
    effect(sc, ep_only=TRUE, max_only=FALSE)$L, # lethality calculated by effect/fx
    1 - tail(simulate(sc), n=1)$S,              # lethality calculated manually
    ignore_attr=TRUE,
    tolerance=1e-5
  )
})

# additional model verification is implemented in test `verify-guts`

#
# Verification of the GUTS-SD model implementation by comparing results
# of this package's model with published results.
#


# GUTS-SD with scaled internal concentration Ci* and dose metric M=Ci*.
# This model is functionally identical to GUTS-RED-SD.
#
# Values are compared to the GUTS-RED-SD implementation which was verified
# using results from EFSA (2018), DOI: 10.2903/j.efsa.2018.5377
test_that("GUTS-SD scaled Ci, M=Ci", {

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=c(0, 1, 2, 3, 4, 4.0001, 5, 6, 7),
                    conc=c(rep(5, 5), rep(0, 4)))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # reproduces results shown in Figure 8, p. 37, from EFSA (2018)
  # constant exposure for 4 days

  efsa <- GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp)
  rf <- efsa %>% simulate(method="ode45", hmax=0.01)

  local <- GUTS_SD(scaled_ci=TRUE, dose_metric="Ci") %>%
    set_param(c(ke=kD, kk=bw, hb=hb, z=zw)) %>%
    set_exposure(exp)
  rs <- local %>%
    simulate(method="ode45", hmax=0.01) %>%
    dplyr::select(time, D=Ci, H, S)

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)

  # reproduces selected results shown in Figure 9, p. 38, from EFSA (2018)
  # pulsed exposure from day 3 to 5
  exp <- data.frame(time=c(0:3, 3:5, 5:14),
                    conc=c(rep(0, 4), rep(50, 3), rep(0, 10))) # low concentration
  rf <- efsa %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01)

  rs <- local %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01) %>%
    dplyr::select(time, D=Ci, H, S)

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)

  exp <- data.frame(time=c(0:3, 3:5, 5:14),
                    conc=c(rep(0, 4), rep(200, 3), rep(0, 10))) # high concentration
  rf <- efsa %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01)

  rs <- local %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01) %>%
    dplyr::select(time, D=Ci, H, S)

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)
})

# GUTS-IT with scaled internal concentration Ci* and dose metric M=Ci*.
# This model is functionally identical to GUTS-RED-IT
#
# Values are compared to the GUTS-RED-IT implementation which was verified
# using results from EFSA (2018), DOI: 10.2903/j.efsa.2018.5377
test_that("GUTS-IT scaled Ci, M=Ci", {

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=c(0, 1, 2, 3, 4, 4.0001, 5, 6, 7),
                    conc=c(rep(5, 5), rep(0, 4)))
  kD <- 0.3 # unit: [time^-1]
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # reproduces results shown in Figure 8, p. 37, from EFSA (2018)
  # constant exposure for 4 days

  efsa <- GUTS_RED_IT() %>%
    set_param(c(kd=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp)
  rf <- efsa %>% simulate(method="ode45", hmax=0.01)

  local <- GUTS_IT(scaled_ci=TRUE, dose_metric="Ci") %>%
    set_param(c(ke=kD, hb=hb, alpha=mw, beta=beta)) %>%
    set_exposure(exp)
  rs <- local %>%
    simulate(method="ode45", hmax=0.01) %>%
    dplyr::select(time, D=Ci, H, S)

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)

  # reproduces selected results shown in Figure 9, p. 38, from EFSA (2018)
  # pulsed exposure from day 3 to 5

  exp <- data.frame(time=c(0:3, 3:5, 5:14),
                    conc=c(rep(0, 4), rep(50, 3), rep(0, 10))) # low concentration
  rf <- efsa %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01)

  rs <- local %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01) %>%
    dplyr::select(time, D=Ci, H, S)

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)

  exp <- data.frame(time=c(0:3, 3:5, 5:14),
                    conc=c(rep(0, 4), rep(200, 3), rep(0, 10))) # high concentration
  rf <- efsa %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01)

  rs <- local %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.01) %>%
    dplyr::select(time, D=Ci, H, S)

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)
})


# GUTS-SD with scaled internal concentration Ci* and dose metric M=Ci*.
# This model is functionally identical to GUTS-RED-SD.
#
# Values are compared to results reported by Jager at al. (2011), DOI: 10.1021/es103092a
# Figure 3, for the effect of naphtalene on fathead minnows. Numerical values
# were not reported by Jager et al., but results matched by visual comparison.
test_that("scaled Ci, M=Ci", {
  ## SD model

  # medium concentration of 47.2 uM
  rs1 <- GUTS_SD(scaled_ci=TRUE, dose_metric="Ci") %>%
    set_param(c( ke=5.53, kk=0.102, hb=0, z=44)) %>%
    set_exposure(data.frame(t=0, c=47.2), reset_times=FALSE) %>%
    set_times(seq(0, 4, 0.1)) %>%
    simulate(hmax=0.01)
  # manual inspection of results
  plot(rs1)
  expect_snapshot_value(rs1, style="serialize", ignore_attr=TRUE, tolerance=1e-3)

  # high concentration of 80.4 uM
  rs2 <- GUTS_SD(scaled_ci=TRUE, dose_metric="Ci") %>%
    set_param(c( ke=5.53, kk=0.102, hb=0, z=44)) %>%
    set_exposure(data.frame(t=0, c=80.4), reset_times=FALSE) %>%
    set_times(seq(0, 4, 0.1)) %>%
    simulate(hmax=0.01)
  # manual inspection of results
  # plot(rs2)
  expect_snapshot_value(rs2, style="serialize", ignore_attr=TRUE, tolerance=1e-3)

  ## IT model

  # TODO requires implementation of log-normal distributed thresholds z (see SI of Jager et al (2011))
  # medium concentration of 47.2 uM
  #rs3 <- GUTS_IT(scaled_ci=TRUE, dose_metric="Ci") %>%
  #  set_param(c( ke=1.48, hb=0, median=46.4, spred=10.4)) %>%
  #  set_exposure(data.frame(t=0, c=47.2), reset_times=FALSE) %>%
  #  set_times(seq(0, 4, 0.1)) %>%
  #  simulate(hmax=0.01)
  ## manual inspection of results
  ## plot(rs3)
  #expect_snapshot_value(rs1, style="serialize", ignore_attr=TRUE, tolerance=1e-3)

  # TODO test high concentration
})


# GUTS-SD with dose metric M=Cw.
# This model is functionally identical to GUTS-RED-SD for sufficiently large `kd`
#
# Verification by plausibility check against GUTS-RED-SD
test_that("GUTS-SD M=Cw", {

  # Code and parameters from EFSA (2018), Appendix E
  # > 01.GUTS-implementation-check.R
  exp <- data.frame(time=c(0, 1, 2, 3, 4, 4.0001, 5, 6, 7),
                    conc=c(rep(5, 5), rep(0, 4)))
  kD <- 100000 # unit: [time^-1], value was modified for this test by expert decision
  hb <- 0 # background mortality rate [time^-1]
  bw <- 0.5 # unit: 1/[D]
  zw <- 2.5 # unit: [D]
  mw <- 2.5 # unit: [D]
  beta <- 2 # dimensionless

  # reproduces results shown in Figure 8, p. 37, from EFSA (2018)
  # constant exposure for 4 days
  rf <- GUTS_RED_SD() %>%
    set_param(c(kd=kD, hb=hb, kk=bw, z=zw)) %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.001, rtol=1e-8, maxsteps=1e7) %>%
    dplyr::select(!D)

  rs <- GUTS_SD(dose_metric="Cw") %>%
    set_param(c(kk=bw, hb=hb, z=zw)) %>%
    set_exposure(exp) %>%
    simulate(method="ode45", hmax=0.001) %>%
    dplyr::select(time, H, S) # D is not calculated for this model variant

  expect_equal(rs, rf, ignore_attr=TRUE, tolerance=1e-3)
})


# TODO Some model variants not explicitly covered yet, needs additional literature sources
# non-scaled Ci, dose metrics M=D, Ci
# scaled Ci, dose metric M=D

# checks if all arguments are of type parameter_set
#' @importFrom methods is
is_parameter_set <- function(x) {
  if(is.list(x))
    return(sapply(x, is_parameter_set))
  is(x, "ParameterSet")
}

# checks if a parameter_set objects matches a certain scenario
is_param_match <- function(scenario, param, ignore_tags=FALSE) {
  if(is.vector(scenario))
    stop("scenario vectors not supported")
  if(!is_scenario(scenario))
    stop("scenario must be of type EffectScenario")
  if(is.vector(param))
    return(sapply(param, function(ps) is_param_match(scenario, ps)))
  if(!is_parameter_set(param))
    stop("param must be of type parameter_set")

  # if both tags are NA, then match by name only
  if(ignore_tags | all(is.na(get_tag(c(scenario, param)))))
    return(get_model(scenario) == get_model(param))
  # if one tag is NA, then no match possible
  if(any(is.na(get_tag(c(scenario, param)))))
    return(FALSE)
  # otherwise match by name and tag
  get_model(scenario) == get_model(param) & get_tag(scenario) == get_tag(param)
}

#' Test if argument is an effect scenario
#'
#' Supports vectorized arguments.
#'
#' @param x Some value or object
#' @return vector of `logical` values
#' @export
#' @examples
#' # returns `TRUE`
#' is_scenario(minnow_it)
#'
#' # returns `FALSE`
#' is_scenario(list())
#'
is_scenario <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_scenario))
  }
  is(x, "EffectScenario")
}

is_sequence <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_sequence))
  }

  is(x, "ScenarioSequence")
}

is_caliset <- function(x) {
  if(is.list(x))
    return(sapply(x, is_caliset))

  is(x, "CalibrationSet")
}

is_exp_series <- function(x) {
  if(is.list(x))
    return(sapply(x, is_exp_series))
  is(x, "ExposureSeries")
}

#' Test if argument is a GUTS scenario
#'
#' @param x vector of `EffectScenario` objects
#'
#' @return vector of `logical` values
#' @export
#' @examples
#' # returns `TRUE`
#' is_guts(minnow_it)
#' is_guts(GUTS_RED_IT())
#'
#' # returns `c(TRUE,TRUE,TRUE)`
#' is_guts(c(minnow_it, minnow_it, minnow_it))
#'
#' # returns `FALSE`
#' is_guts_sd(minnow_it)
is_guts <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_guts))
  }
  if(!is_scenario(x))
    return(FALSE)

  startsWith(get_model(x),"GUTS-")
}

#' @export
#' @describeIn is_guts Deprecated alias.
is_GUTS <- function(x) {
  is_guts(x)
}

#' @describeIn is_guts Test if argument is a GUTS-IT scenario
#' @export
is_guts_it <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_guts_it))
  }
  if(!is_scenario(x))
    return(FALSE)

  name <- get_model(x)
  startsWith(name,"GUTS-") & endsWith(name,"-IT")
}

#' @export
#' @describeIn is_guts Deprecated alias.
is_GUTS_IT <- function(x) {
  is_guts_it(x)
}

#' @describeIn is_guts Test if argument is a GUTS-SD scenario
#' @export
is_guts_sd <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_guts_sd))
  }
  if(!is_scenario(x))
    return(FALSE)

  name <- get_model(x)
  startsWith(name,"GUTS-") & endsWith(name,"-SD")
}

#' @export
#' @describeIn is_guts Deprecated alias.
is_GUTS_SD <- function(x) {
  is_guts_sd(x)
}

#' Test if argument is a DEB scenario
#'
#' @param x vector of `EffectScenario` objects
#' @return vector of `logical` values
#' @export
is_deb <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_deb))
  }
  if(!is_scenario(x))
    return(FALSE)

  startsWith(get_model(x), "DEB")
}

#' @export
#' @describeIn is_deb Deprecated alias.
is_DEB <- function(x) {
  is_deb(x)
}

#' Test if argument is a Lemna scenario
#'
#' @param x vector of [scenario] objects
#' @return vector of logical values
#' @export
is_lemna <- function(x) {
  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_lemna))
  }
  if(!is_scenario(x))
    return(FALSE)

  is(x, "Lemna")
}

#' @export
#' @describeIn is_lemna Deprecated alias.
is_Lemna <- function(x) {
  is_lemna(x)
}

#' Test if argument is a LemnaThreshold scenario
#'
#' `r lifecycle::badge("deprecated")`
#'
#' @param x vector of [scenarios] objects
#' @return vector of `logical` values
#' @seealso [is_lemna()]
#' @export
is_LemnaThreshold <- function(x) {
  lifecycle::deprecate_warn("1.5.0", "is_LemnaThreshold()")

  if(is.list(x)) {
    if(length(x) > 0)
      return(sapply(x, is_LemnaThreshold))
  }
  if(!is_scenario(x))
    return(FALSE)

  get_model(x) == "Lemna_SchmittThold"
}

# Check if an exposure series is complete, i.e. regular
#
is_regular_series <- function(x) {
  if(is(x,"ExposureSeries")) {
    if("regular.series" %in% names(x@meta))
      return(x@meta$regular.series)
    return(FALSE)
  }
  else if(is.vector(x) & is.numeric(x) & length(x) > 0) {
    x_diff <- unique(diff(x))
    if(length(x_diff) <= 1)
      return(TRUE)

    # disregard numerical issuesm i.e. very small deviations from a single value
    x_min <- min(x_diff[which(x_diff > 0)])
    return(all(1 - abs(x_diff/x_min) < 1e-10))
  }
  stop("type not supported")
}

# Check if scenarios requires a control for effect calculation
#
is_control_required <- function(x) {
  if(is.vector(x) & length(x) > 1)
    return(sapply(x, is_control_required))
  if(is_sequence(x))
    return(is_control_required(x[[1]]))
  if(!is_scenario(x))
    stop("Argument `x` must be a scenario")

  x@control.req
}

# Check if an R package is locally available
#
# The check is case-sensitive
# @param name `character`, name of an R package, e.g. 'dplyr'
# @return `logical`, `TRUE` if package is installed, else `FALSE`
#' @importFrom utils installed.packages
is_pckg_avail <- function(name) {
  require(package=name, quietly=TRUE, character.only=TRUE)
}

# Check if an exposure series represents zero/no exposure
#
# @param x `EffectScenario` or `ExposureSeries` object
# @return `logical`, `TRUE` if no/zero exposure
is_no_exposure <- function(x) {
  if(is.list(x))
    return(sapply(x, is_no_exposure))
  if(all(is_sequence(x))) # ordering of these checks is important! due to length(sequence) > 1 in most cases
    return(all(sapply(x@scenarios, is_no_exposure)))
  if(length(x) > 1)
    return(sapply(x, is_no_exposure))

  if(is.vector(x))
    x <- x[[1]] # unbox
  if(is(x, "EffectScenario"))
    x <- x@exposure
  if(!is(x, "ExposureSeries"))
    return(FALSE)
  if(nrow(x@series) == 0) {
    return(TRUE)
  } else if(nrow(x@series) == 1) {
    if(x@series[[1,2]] == 0)
      return(TRUE)
  } else {
    if(all(x@series[,2] == 0))
      return(TRUE)
  }
  return(FALSE)
}

test_that("dynamic call routing: list of calisets", {
  # create dummy caliset
  cs <- caliset(scenario=Lemna_SETAC(), data=data.frame(time=0:1, obs=0))
  my_tktd <- function(x, ...) { return(TRUE) }

  # check if dynamic call is routed correctly
  rs <- FALSE
  with_mocked_bindings(
    rs <- fit_tktd(list(cs)),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_true(rs)
})

test_that("dynamic call routing: scenario sequence", {
  # create dummy sequence
  sc <- Lemna_SETAC() %>% set_times(0:8)
  sq <- sequence(list(sc, sc), breaks=3)
  my_tktd <- function(x, ...) { return(TRUE) }

  # check if dynamic call is routed correctly
  rs <- FALSE
  with_mocked_bindings(
    rs <- fit_tktd(sq, data=data.frame(time=0:1, obs=0)),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_true(rs)
})

test_that("arg x=scenario", {
  ## Valid arguments
  # scenario with data
  sc <- Lemna_SETAC() %>% set_times(0:8)
  my_tktd <- function(x, ...) { return(x) }

  rs <- NULL
  with_mocked_bindings(
    rs <- fit_tktd(sc, data=data.frame(time=0:1, obs=0)),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_true(is.list(rs))
  expect_true(is_caliset(rs[[1]]))
  expect_equal(rs[[1]]@scenario, sc)

  # sequence with data
  sq <- sequence(list(sc, sc), breaks=3)

  rs <- NULL
  with_mocked_bindings(
    rs <- fit_tktd(sq, data=data.frame(time=0:1, obs=0)),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_true(is.list(rs))
  expect_true(is_caliset(rs[[1]]))
  expect_equal(rs[[1]]@scenario, sq)

  ## Invalid arguments
  # list of scenarios with data
  expect_error(fit_tktd(list(sc), data=data.frame(time=0:1, obs=0)), "not supported")
  # scenario without data
  expect_error(fit_tktd(sc), "is a scenario")

})

test_that("arg x=cvasi_fit", {
  fit <- list(par=c("k_photo_max"=0.123456))
  class(fit) <- c("cvasi_fit", "list")
  cs <- caliset(Lemna_SETAC(), data.frame(time=0:1, obs=0))
  csref <- caliset(Lemna_SETAC() %>% set_param(fit$par), data.frame(time=0:1, obs=0))
  my_tktd <- function(x, ...) { return(x) }

  ## Valid arguments
  # data=single caliset
  rs <- NULL
  with_mocked_bindings(
    rs <- fit_tktd(fit, data=cs),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_equal(rs, list(csref))

  # data=list of calisets
  rs <- NULL
  with_mocked_bindings(
    rs <- fit_tktd(fit, data=list(cs, cs)),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_equal(rs, list(csref, csref))

  ## Invalid arguments
  # data=missing
  expect_error(fit_tktd(fit), "data. is missing")
  # data=data.frame
  expect_error(fit_tktd(fit, data=data.frame(time=0:2, obs=0)), "only contain caliset")
})

test_that("arg x=caliset(s)", {
  ## Valid arguments
  sc <- Lemna_SETAC() %>% set_times(0:8)
  cs <- caliset(sc, data.frame(time=0:1, obs=0))
  my_tktd <- function(x, ...) { return(x) }

  # single caliset
  rs <- NULL
  with_mocked_bindings(
    rs <- fit_tktd(cs),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_true(is.list(rs))
  expect_equal(length(rs), 1)
  expect_true(is_caliset(rs[[1]]))
  expect_equal(rs[[1]]@scenario, sc)

  # multiple calisets
  rs <- NULL
  with_mocked_bindings(
    rs <- fit_tktd(list(cs, cs)),
    fit_tktd_lemna_setac = my_tktd
  )
  expect_true(is.list(rs))
  expect_equal(length(rs), 2)
  expect_equal(rs, list(cs, cs))

  ## Invalid arguments
  # caliset with data
  expect_error(fit_tktd(cs, data=data.frame(t=0, o=0)), "alisets.*cannot be used together")
})

test_that("arg x=unsupported model", {
  source(test_path("dummy.R"), local = TRUE)
  sc <- new("DummyScenario")
  df <- data.frame(t=0, obs=0)

  # args: scenario + data
  expect_error(fit_tktd(sc, data=df), "Scenario type not supported")

  # args: caliset
  cs <- caliset(sc, data=df)
  expect_error(fit_tktd(cs), "Scenario type not supported")
})

test_that("is_scenario", {
  expect_true(is_scenario(GUTS_RED_IT()))
  expect_true(is_scenario(new("EffectScenario")))
  expect_equal(is_scenario(c(GUTS_RED_IT(),Lemna_Schmitt())), c(TRUE,TRUE))
  expect_equal(is_scenario(c(DEB_abj(),NA)), c(TRUE,FALSE))
  # non-scenario arguments
  expect_false(is_scenario(c()))
  expect_false(is_scenario(NA))
  expect_false(is_scenario(NULL))
  expect_false(is_scenario(1))
  expect_equal(is_scenario(1:5), FALSE)
})

test_that("is_sequence", {
  sc <- new("EffectScenario") %>% set_times(0:6)
  suppressWarnings(sq <- sequence(list(sc)))
  expect_true(is_sequence(sq))
  expect_equal(is_sequence(list(sq, sq)), c(TRUE, TRUE))

  expect_false(is_sequence(minnow_it))
  expect_false(is_sequence(1))
  expect_error(is_sequence())
})

test_that("is_caliset", {
  cs <- caliset(minnow_it, data.frame(t=0, n=0))
  expect_true(is_caliset(cs))
  expect_equal(is_caliset(list(cs, cs)), c(TRUE, TRUE))

  expect_false(is_caliset(minnow_it))
  expect_false(is_caliset(1))
  expect_error(is_caliset())
})

test_that("is_parameter_set", {
  expect_true(is_parameter_set(parameter_set("test")))
  expect_equal(is_parameter_set(c(parameter_set("test"),parameter_set("test"))), c(TRUE,TRUE))
  # invalid arguments
  expect_false(is_parameter_set(NA))
  expect_false(is_parameter_set(NULL))
  expect_false(is_parameter_set(1))
  expect_equal(is_parameter_set(1:5), FALSE)
})

test_that("is_param_match", {
  sc_unt <- new("EffectScenario", name="foo", tag="")
  sc_tgd <- new("EffectScenario", name="foo", tag="bar")
  sc_other <- new("EffectScenario", name="baz")
  ps_unt <- parameter_set(model="foo", tag="")
  ps_tgd <- parameter_set(model="foo", tag="bar")
  ps_other <- parameter_set(model="baz", tag="")

  # match
  expect_true(is_param_match(sc_unt, ps_unt))
  expect_true(is_param_match(sc_tgd, ps_tgd))
  # no match
  expect_false(is_param_match(sc_unt, ps_tgd))
  expect_false(is_param_match(sc_unt, ps_other))
  expect_false(is_param_match(sc_tgd, ps_unt))
  expect_false(is_param_match(sc_tgd, ps_other))
  expect_false(is_param_match(sc_other, ps_unt))
  expect_false(is_param_match(sc_other, ps_tgd))
  # vectorised parameter sets
  expect_equal(c(FALSE,FALSE), is_param_match(sc_unt, c(ps_tgd, ps_tgd)))
  expect_equal(c(TRUE,FALSE),  is_param_match(sc_unt, c(ps_unt, ps_tgd)))
  expect_equal(c(TRUE,FALSE),  is_param_match(sc_unt, c(ps_unt, ps_other)))
  expect_equal(c(TRUE,TRUE),   is_param_match(sc_tgd, c(ps_tgd, ps_tgd)))

  # invalid inputs
  expect_error(is_param_match(NULL, NULL))
  expect_error(is_param_match(c(), c()))
  expect_error(is_param_match(sc_unt, c()))
  expect_error(is_param_match(c(), ps_unt))
})

test_that("is_guts", {
  expect_true(is_guts(GUTS_RED_IT()))
  expect_true(is_guts(GUTS_RED_SD()))
  expect_equal(is_guts(c(GUTS_RED_IT(),GUTS_RED_IT())), c(TRUE,TRUE))

  # deprecated alias
  expect_true(is_GUTS(GUTS_RED_IT()))

  # invalid arguments
  expect_false(is_guts(list()))
  expect_false(is_guts(NA))
  expect_false(is_guts(NULL))
  expect_false(is_guts(1))
  expect_equal(is_guts(1:5), FALSE)
})

test_that("is_guts_it", {
  expect_true(is_guts_it(GUTS_RED_IT()))
  expect_equal(is_guts_it(c(GUTS_RED_IT(),GUTS_RED_IT())), c(TRUE,TRUE))

  # deprecated alias
  expect_true(is_GUTS_IT(GUTS_RED_IT()))

  # invalid arguments
  expect_false(is_guts_it(list()))
  expect_false(is_guts_it(GUTS_RED_SD()))
  expect_false(is_guts_it(NA))
  expect_false(is_guts_it(NULL))
  expect_false(is_guts_it(1))
  expect_equal(is_guts_it(1:5), FALSE)
})

test_that("is_guts_sd", {
  expect_true(is_guts_sd(GUTS_RED_SD()))
  expect_equal(is_guts_sd(c(GUTS_RED_SD(),GUTS_RED_SD())), c(TRUE,TRUE))

  # deprecated alias
  expect_true(is_GUTS_SD(GUTS_RED_SD()))

  # invalid arguments
  expect_false(is_guts_sd(list()))
  expect_false(is_guts_sd(GUTS_RED_IT()))
  expect_false(is_guts_sd(NA))
  expect_false(is_guts_sd(NULL))
  expect_false(is_guts_sd(1))
  expect_equal(is_guts_sd(1:5), FALSE)
})

test_that("is_deb", {
  expect_true(is_deb(DEB_abj()))
  expect_equal(is_deb(c(DEB_abj(),DEB_abj())), c(TRUE,TRUE))

  # deprecated alias
  expect_true(is_DEB(DEB_abj()))

  # invalid arguments
  expect_false(is_deb(list()))
  expect_false(is_deb(GUTS_RED_IT()))
  expect_false(is_deb(NA))
  expect_false(is_deb(NULL))
  expect_false(is_deb(1))
  expect_equal(is_deb(1:5), FALSE)
})

test_that("is_lemna", {
  expect_true(is_lemna(Lemna_Schmitt()))
  expect_true(is_lemna(Lemna_SchmittThold()))
  expect_true(is_lemna(Lemna_SETAC()))
  expect_equal(is_lemna(c(Lemna_Schmitt(),Lemna_Schmitt())), c(TRUE,TRUE))

  # deprecated alias
  expect_true(is_Lemna(Lemna_Schmitt()))

  # invalid arguments
  expect_false(is_lemna(list()))
  expect_false(is_lemna(GUTS_RED_IT()))
  expect_false(is_lemna(NA))
  expect_false(is_lemna(NULL))
  expect_false(is_lemna(1))
  expect_equal(is_lemna(1:5), FALSE)
})

test_that("is_LemnaThreshold", {
  lifecycle::expect_deprecated(expect_true(is_LemnaThreshold(Lemna_SchmittThold())))
  lifecycle::expect_deprecated(expect_equal(is_LemnaThreshold(c(Lemna_SchmittThold(),Lemna_SchmittThold())), c(TRUE,TRUE)))
  # invalid arguments
  lifecycle::expect_deprecated(expect_false(is_LemnaThreshold(list())))
  lifecycle::expect_deprecated(expect_false(is_LemnaThreshold(Lemna_Schmitt())))
  lifecycle::expect_deprecated(expect_false(is_LemnaThreshold(NA)))
  lifecycle::expect_deprecated(expect_false(is_LemnaThreshold(NULL)))
  lifecycle::expect_deprecated(expect_false(is_LemnaThreshold(1)))
  lifecycle::expect_deprecated(expect_equal(is_LemnaThreshold(1:5), FALSE))
})

test_that("is_regular_series", {
  # exposure series
  expect_false(is_regular_series(ExposureSeries(series=data.frame(t=0:5, c=0:5))))
  expect_true(is_regular_series(ExposureSeries(series=data.frame(t=0:5, c=0:5), meta=list(regular.series=TRUE))))
  # some numeric vector of output times
  expect_true(is_regular_series(1))
  expect_true(is_regular_series(0:10))
  expect_true(is_regular_series(seq(0,10,0.1)))
  expect_true(is_regular_series(seq(0,10,0.3)))
  expect_true(is_regular_series(seq(0,10.12,0.127)))
  # invalid arguments
  expect_error(is_regular_series(NA))
  expect_error(is_regular_series(NULL))
  expect_error(is_regular_series(c()))
  expect_error(is_regular_series(c(NA,NA)))
})

test_that("is_control_required", {
  expect_true(is_control_required(Lemna_Schmitt()))
  expect_true(is_control_required(DEB_abj()))
  expect_equal(is_control_required(c(Lemna_Schmitt(),Lemna_Schmitt())), c(TRUE,TRUE))
  # invalid arguments
  expect_false(is_control_required(GUTS_RED_IT()))
  expect_error(is_control_required(NA))
  expect_error(is_control_required(NULL))
  expect_error(is_control_required(1))
  expect_error(is_control_required(1:5))
})

# TOOD support algae and other primary producer models as well
# TODO output messages for fitted parameter(s), starting values, output variable
#      reuse outputs from calibrate()


#' Fit growth parameters
#'
#' `r lifecycle::badge("experimental")` <br>
#' High-level function to fit growth and/or loss parameters to observed data. It eases the
#' use of [calibrate()] by providing sensible defaults for various (model dependent)
#' settings.
#'
#' @param x a [scenario] or a list of [caliset] objects
#' @param data toxicological trial data to fit growth parameters to: required if
#'      `x` is either a [scenario] or the result of a fit. See Section *Data*
#'      for details.
#' @param par named vector, of parameters to fit and their starting values
#' @param log_scale logical, if `TRUE` then fitting will be performed on log-transformed
#'   observations and predictions, else the data will be used as-is
#' @param verbose logical, if `TRUE` then info messages are printed to the console
#' @param ... additional arguments passed through to [calibrate()]
#' @aliases fit_growth,list,missing-method fit_growth,EffectScenario,ANY-method
#'   fit_growth,ScenarioSequence,ANY-method
#' @returns a list
#' @export
#' @examples
#' # Use experimental data from control trial ('T0') to fit growth rate
#' ctrl <- schmitt2013[schmitt2013$trial == "T0" ,]
#'
#' # Set up a scenario, provide dummy parameter values where necessary
#' sc <- Lemna_Schmitt() %>%
#'   set_init(c(BM=0.0012)) %>%
#'   set_param(c(EC50=1, b=1, P_up=1))
#'
#' # Run fitting routine
#' fit_growth(sc, data=ctrl, verbose=TRUE)
setGeneric("fit_growth", function(x, data, ...) standardGeneric("fit_growth"))

#' @describeIn fit_growth Default handler
setMethod("fit_growth", c("ANY", "ANY"), function(x, data, ...) stop("Object type or argument combination not supported"))

# Wrapper for list of caliset objects
#' @noRd
setMethod("fit_growth", c("list", "missing"), function(x, data, ...) fit_growth_list(x, data=data, ...))

# Wrapper for [scenario]s
#' @noRd
setMethod("fit_growth", c("EffectScenario", "ANY"), function(x, data, ...) fit_growth_scenario(x=x, data=data, ...))

# Wrapper for [sequence]s
#' @noRd
setMethod("fit_growth", c("ScenarioSequence", "ANY"), function(x, data, ...) fit_growth_scenario(x=x, data=data, ...))

# Wrapper for single scenario + dataset
fit_growth_scenario <- function(x, data, ...) {
  # if scenario type not supported yet, we may inadvertently land here
  if(is.list(x) & all(is_caliset(x))) {
    stop("Scenario type not supported yet")
  }

  # normal call handling starts here...
  if(missing(data)) {
    stop("Argument 'data' is missing")
  }
  # check if dataset is represented by tabular data
  if(is.data.frame(data)) {
    data <- tox_data(data)
  }
  else if(is(data, "ToxData")) {
    # nothing to do
  } else {
    stop("Argument 'data' has unsupported type")
  }
  # create calisets
  x <- td2cs(x, data)
  fit_growth(x, ...)
}

# Wrapper for lists of calisets
#' @importFrom methods selectMethod
fit_growth_list <- function(x, ...) {
  # check that list only contains caliset object
  if(!all(is_caliset(x))) {
    stop("Argument 'x' must only contain caliset objects")
  }
  # then use scenario in first caliset to decide which method to call
  sc <- x[[1]]@scenario
  if(is_sequence(x[[1]]@scenario))
    sc <- sc[[1]]

  cls <- class(sc)[1]
  fun <- selectMethod("fit_growth", signature=c(x=cls, data="missing"))
  fun(x, ...)
}

#' @describeIn fit_growth Fit growth parameters of [Lemna_SETAC] scenarios
#' @include model-lemna_setac.R
setMethod("fit_growth", c("LemnaSetac", "missing"), function(x, data, par, log_scale=TRUE, verbose=FALSE, ...)
  fit_growth_lemna(x, data=data, par=par, log_scale=log_scale, verbose=verbose, ...))

#' @describeIn fit_growth Fit growth parameters of [Lemna_Schmitt] scenarios
#' @include model-lemna_schmitt.R
setMethod("fit_growth", c("LemnaSchmitt", "missing"), function(x, data, par, log_scale=TRUE, verbose=FALSE, ...)
  fit_growth_lemna_schmitt(x, data=data, par=par, log_scale=log_scale, verbose=verbose, ...))


#' @autoglobal
fit_growth_lemna <- function(x, par, data=NULL, log_scale=TRUE, verbose=FALSE,
                             maxsteps=10^6, lower=NULL, upper=NULL, ...) {
  # TODO this is not a good solution, as every model-dep function must check this.
  #      better to not pass arguments through to calibrate() in the future
  if(!is.null(lower)) {
    stop("Argument 'lower' is not allowed, use `set_bounds()` instead")
  }
  if(!is.null(upper)) {
    stop("Argument 'upper' is not allowed, use `set_bounds()` instead")
  }

  # is it a list of calisets?
  if(is.vector(x)) {
    if(!all(is_caliset(x))) {
      stop("Argument 'x' must either be a scenario or a list of calisets")
    }
  } else if(is_scenario(x) | is_sequence(x)) {
    # check if a dataset was supplied as well, then create calisets
    if(is.null(data)) {
      stop("Argument 'data' is missing")
    }
    if(is.matrix(data)) {
      data <- as.data.frame(data)
    }
    if(is.data.frame(data)) {
      data <- tox_data(data)
    }
    else if(is(data, "ToxData")) {
      # nothing to do
    } else {
      stop("Argument 'data' has unsupported type")
    }
    x <- td2cs(x, data)
  }

  # make sure that no exposure is set in any scenario/caliset
  for(i in seq_along(x)) {
    if(!is_no_exposure(x[[i]]@scenario)) {
      stop("Scenarios/observed data must not include exposure to fit growth parameters")
    }
    # check if all datasets start at time=0, which is an implicit assumption in
    # the following calculation of statistics etc
    if(min(x[[i]]@data[, 1]) != 0) {
      stop("Scenarios/datasets must start at time zero (t=0)")
    }
    # make sure that each scenario has an exposure series set
    x[[i]]@scenario <- set_noexposure(x[[i]]@scenario)
  }

  # determine parameter to fit: for lemna, use fit 'k_photo_max'
  par_nms <- "k_photo_max"

  # if no starting value set, use a default
  if(missing(par)) {
    par <- c("k_photo_max" = 0.35)
  }

  # check for invalid values
  if(any(is.na(par) | is.nan(par) | is.infinite(par))) {
    stop("Starting value of `", par_nms,"` is invalid")
  }

  # determine variable to fit on
  output_nm <- "FrondNo"

  # run fitting routine
  fit <- calibrate(x, par=par, output=output_nm, hessian=TRUE, log_scale=log_scale,
                   err_fun="sse", verbose=verbose, maxsteps=maxsteps, ...)

  # remove data points at time==0, as these are the initial conditions and not predictions
  df <- fit$data %>% dplyr::filter(time > 0)

  ## Calculate Statistics
  # Degrees of Freedom: number of data points minus number of fitted parameters, excluding t=0
  fit$dof <- nrow(df) - length(par)

  # limits of the 95% confidence interval
  # TODO likelihood profiling as alternative
  fit$ci <- ci_from_hessian(fit, dof=fit$dof, level=0.95)

  # NRMSE
  fit$nrmse <-  1 / mean(df[, 2]) * sqrt(1 / nrow(df) * sum((df[, 2] - df[, 3])^2))

  fit
}

fit_growth_lemna_schmitt <- function(x, par, ...) {
  if(missing(par)) {
    par <-  c("k_phot_max" = 0.35)
  }
  fit_growth_lemna(x=x, par=par, ...)
}


#' @describeIn fit_growth Fit growth parameters of [Magma] scenarios
#' @include model-magma.R
setMethod("fit_growth", c("Magma", "missing"), function(x, data, par, log_scale=TRUE, verbose=FALSE, ...)
  fit_growth_magma(x, data=data, par=par, log_scale=log_scale, verbose=verbose, ...))

#' @autoglobal
fit_growth_magma <- function(x, par, data=NULL, log_scale=TRUE, verbose=FALSE,
                             maxsteps=10^6, lower=NULL, upper=NULL, ...) {
  # TODO this is not a good solution, as every model-dep function must check this.
  #      better to not pass arguments through to calibrate() in the future
  if(!is.null(lower)) {
    stop("Argument 'lower' is not allowed, use `set_bounds()` instead")
  }
  if(!is.null(upper)) {
    stop("Argument 'upper' is not allowed, use `set_bounds()` instead")
  }

  # is it a list of calisets?
  if(is.vector(x)) {
    if(!all(is_caliset(x))) {
      stop("Argument 'x' must either be a scenario or a list of calisets")
    }
  } else if(is_scenario(x) | is_sequence(x)) {
    # check if a dataset was supplied as well, then create calisets
    if(is.null(data)) {
      stop("Argument 'data' is missing")
    }
    if(is.matrix(data)) {
      data <- as.data.frame(data)
    }
    if(is.data.frame(data)) {
      data <- tox_data(data)
    }
    else if(is(data, "ToxData")) {
      # nothing to do
    } else {
      stop("Argument 'data' has unsupported type")
    }
    x <- td2cs(x, data)
  }

  # make sure that no exposure is set in any scenario/caliset
  for(i in seq_along(x)) {
    if(!is_no_exposure(x[[1]]@scenario)) {
      stop("Scenarios/observed data must not include exposure to fit growth parameters")
    }
    # check if all datasets start at time=0, which is an implicit assumption in
    # the following calculation of statistics etc
    if(min(x[[1]]@data[, 1]) != 0) {
      stop("Scenarios/datasets must start at time zero (t=0)")
    }
  }

  # determine parameter to fit
  par_nms <- "mu_control"
  # TODO check if all calisets use the same growth model?
  is_log_growth <- x[[1]]@scenario@growth_model == "log"
  if(is_log_growth) {
    par_nms <- c(par_nms, "D_L")
  }

  # if no starting value set, use a default
  if(missing(par)) {
    par <- c("mu_control" = 0.47) # magic value: 0.47 is the default value proposed by model authors
    if(is_log_growth) {
      par["D_L"] <- 1 # magic value: no specific reason
    }
  }

  # check for invalid values
  if(any(is.na(par) | is.nan(par) | is.infinite(par))) {
    stop("Starting value of '", paste(par_nms, collapse=","), "' is invalid")
  }

  # determine variable to fit on
  output_nm <- "TSL"

  # run fitting routine
  fit <- calibrate(x, par=par, output=output_nm, hessian=TRUE, log_scale=log_scale,
                   err_fun="sse", verbose=verbose, maxsteps=maxsteps, ...)

  # remove data points at time==0, as these are the initial conditions and not predictions
  df <- fit$data %>% dplyr::filter(time > 0)

  ## Calculate Statistics
  # Degrees of Freedom: number of data points minus number of fitted parameters, excluding t=0
  fit$dof <- nrow(df) - length(par)

  # limits of the 95% confidence interval
  # TODO likelihood profiling as alternative
  fit$ci <- ci_from_hessian(fit, dof=fit$dof, level=0.95)

  # NRMSE
  fit$nrmse <-  1 / mean(df[, 2]) * sqrt(1 / nrow(df) * sum((df[, 2] - df[, 3])^2))

  fit
}


#' Parameter confidence intervals from fit
#'
#' Calculates parameter confidence intervals for a fit returned by [calibrate()].
#' The fit must provide a valid *Hessian* matrix.
#'
#' @param fit return value from [calibrate()]
#' @param dof integer, Degrees of Freedom, commonly the number of independent observations
#'   minus the number of fitted parameters
#' @param level numeric, desired confidence level, i.e a value between zero and one.
#'   Defaults to 95% (`0.95`).
#' @returns name list of numeric vectors of length two, the elements
#'   representing the lower and upper confidence limit, respectively.
#' @export
ci_from_hessian <- function(fit, dof, level=0.95) {
  if(missing(fit))
    stop("Argument 'fit' is missing")
  if(!inherits(fit, "cvasi_fit"))
    stop("Argument 'fit' has incorrect type, not a result from `calibrate()`")
  if(!("hessian" %in% names(fit)))
     stop("Argument 'fit' does not contain a Hessian matrix")
  if(missing(dof))
    stop("Argument 'dof' is missing")
  if(any(dof <= 0 | is.na(dof) | is.infinite(dof)))
    stop("argument 'dof' is out of range")
  if(!is.numeric(level))
    stop("Argument 'level' must be numeric")
  if(any(level <= 0 | level >= 1 | is.na(level) | is.infinite(level)))
    stop("Argument 'level' is out of range")

  par_names <- names(fit$par)
  std_err <- try(sqrt(diag(solve(fit$hessian)) * fit$value / dof), silent=TRUE)
  if(inherits(std_err, "try-error")) {
    warning("Hessian cannot be inverted: ", as.character(std_err), call.=FALSE)
    std_err <- rep(NA_real_, length(par_names))
  }
  std_err <- setNames(std_err, par_names)

  # limits of the confidence interval
  ci <- lapply(par_names, function(nm) {
    ci_lower <- (1 - level)/2
    ci_upper <- 1 - ci_lower
    unname(fit$par[[nm]] + stats::qnorm(p=c(ci_lower, ci_upper)) * std_err[[nm]])
  })
  ci <- setNames(ci, par_names)
  ci
}

# TODO rename and overload profile() function
# TODO remove parameter 'bounds'
# TODO accept a cvasi_fit object as arg 'x'

#' Likelihood profiling
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' The aim of the function is two-fold: 1) estimate a 95% confidence
#' around each parameter of a calibrated model, and 2) see if perhaps a local minimum was found rather
#' than a global minimum. To achieve this, the likelihood profiling goes through
#' every parameter one by one. For each parameter,
#' the model is sequentially refit with the parameter value set to
#' increasingly lower and higher values, and the likelihood of the model given the
#' data calculated (using [log_lik()]). The likelihood is then compared
#' to the likelihood of the original model (using a likelihood ratio). This leads
#' to the development of a likelihood profile, from which a plot a 95%
#' confidence interval for the parameter is derived.
#'
#' The idea of the function is a variable stepwise algorithm:
#' When the likelihood ratio changes very little (less than `l_crit_min`), the stepsize is
#' increased (up to a maximum, specified by `f_step_max`). When the lik.
#' ratio changes too much (more than `l_crit_max`), the algorithm tries again
#' with a smaller stepsize (also bound to a minimum: `f_step_min`). Note that
#' the stepsize is used as a fraction of the parameter value that is tried.
#' To prevent very small stepsizes when the value goes towards zero (as can
#' be the case for effect thresholds), an absolute minimum
#' stepsize (`f_step_abs`), which is specified as a fraction of the best
#' parameter value (`Xhat`) (unless it is zero, then the algorithm takes
#' something small).
#'
#' Note that the likelihood of the model given the data can be calculated across
#' all datasets provided in the calibration set `x`, or calculated separately for
#' each individual dataset before being combined into one likelihood (by adjusting
#' the optional parameter `individual`). The latter
#' has the advantage that different datasets can be given different weights in
#' the likelihood calculation (using the "weight" slot of the [caliset] objects,`x`).
#' Further, for continuous data (e.g. biomass), the likelihood considers the variance (standard
#' deviation) in the log likelihood calculation, which can vary between datasets
#' when the likelihood is calculated for each dataset separately before combining
#' into an overall likelihood. The latter could be relevant when factors might lead
#' to variability between datasets (e.g. different labs, different animal culture,...)
#'
#' To conduct the likelihood calculations on separate datasets, the parameter `individual`
#' which by default is 'FALSE' can be set to 'TRUE'. Then, then log likelihoods
#' are calculated for each dataset individually (or in subgroups, using the "tag" names of the
#' [caliset] object, if provided, to group datasets with the same "tag"
#' before calculating the log likelihood). Subsequently, the log
#' likelihoods for the subsets are combined into an overall likelihood (considering
#' the *set* weights provided in the "weight" slot of the [caliset] object).
#' Note that for each *set* only 1 weight can be provided (i.e. not individual
#' weights for each datapoint within the *set*), and that *set* with the same tag
#' should have identical weight.
#'
#' The function was inspired by a MatLab BYOM v.6.8 procedure, created by
#' Tjalling Jager. For details, please refer to BYOM (http://debtox.info/byom.html)
#' as well as Jager (2021).
#'
#' @references
#' Jager T, 2021: Robust Likelihood-Based Optimization and Uncertainty Analysis
#' of Toxicokinetic-Toxicodynamic Models. Integrated Environmental Assessment and
#' Management 17:388-397. \doi{10.1002/ieam.4333}
#'
#' @param x either a single [scenario] or a list of [caliset] objects
#' @param par named vector - parameters (names and values) to be profiled
#' @param output character vector, name of output column of [simulate()] that
#'  is used in calibration
#' @param data only needed if `x` is a [scenario]
#' @param bounds optional list of lists (including lower and upper bound): uses defaults in `x` object, but
#'  can be overwritten here (e.g. bounds <- list(k_resp = list(0,10), k_phot_max = list(0,30)) )
#' @param refit if `TRUE` (default), refit if a better minimum is found
#' @param type `"fine"` or `"coarse"` (default) likelihood profiling
#' @param individual if `FALSE` (default), the log likelihood is calculated across
#' the whole dataset. Alternatively, if `TRUE`, log likelihoods are calculated for
#' each (group of) *set*(s) individually.
#' @param break_prof If `TRUE`, then stop the profiling if a better optimum is located.
#'    Default is `FALSE`.
#' @param log_scale `FALSE` (default), option to calculate the log likelihood on a
#' log scale (i.e., observations and predictions are log transformed during calculation)
#' @param data_type Character argument, `"continuous"` (default) or `"count"`, to specify the data type
#' for the log likelihood calculations.
#' @param ... additional parameters passed on to [calibrate()] and [simulate()]. To avoid
#'   parameter confusion, use argument `method` to select optimization algorithms
#'   of `calibrate()` and argument `ode_method` to select numerical integration
#'   schemes of package `deSolve`.
#' @autoglobal
#' @examples
#' # Example with Lemna model - physiological params
#' library(dplyr)
#'
#' # observations - control run
#' obs <- schmitt2013 %>%
#'   filter(trial == "T0")
#'
#' # update metsulfuron
#' myscenario <- metsulfuron %>%
#'   set_param(c(k_phot_fix = TRUE, Emax = 1)) %>%
#'   set_init(c(BM = 0.0012)) %>%
#'   set_noexposure() %>%
#'   set_bounds(list(k_phot_max=c(0, 1)))
#'
#' fit <- calibrate(
#'   x = myscenario,
#'   par = c(k_phot_max = 1),
#'   data = obs,
#'   output = "FrondNo",
#'   method = "Brent"
#' )
#'
#' # Likelihood profiling
#' \donttest{
#' res <- lik_profile(
#'   x = myscenario,
#'   data = obs,
#'   output = "FrondNo",
#'   par = fit$par,
#'   bounds = list(
#'     k_phot_max = list(0, 30)
#'   ),
#'   refit = FALSE,
#'   type = "fine",
#'   method = "Brent"
#' )
#' # plot
#' plot(res)
#' }
#'
#' @returns A list containing, for each parameter profiled, the likelihood
#' profiling results as a dataframe;
#' the 95% confidence interval; the original parameter value; the likelihood plot object; and
#' the recalibrated parameter values (in case a lower optimum was found)
#' @export

# # License information related to MatLab BYOM v.6.8:
# Copyright (c) 2012-2023 Tjalling Jager
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so

lik_profile <- function(x,
                        par,
                        output,
                        data = NULL,
                        bounds = NULL,
                        refit = TRUE,
                        type = c("coarse", "fine"),
                        individual = FALSE,
                        break_prof = FALSE, # typically we do not want to stop
                        # profiling until we have all parameters, and then make a decision based on parameter space
                        log_scale = FALSE,
                        data_type = c("continuous", "count"),
                        ...) {
  # check inputs
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # convert to list of calibration sets if needed
  if (length(x) == 1) {
    if (is_scenario(x)) {
      if (is.null(data)) {
        stop("Scenario provided, but argument 'data' is missing")
      } else {
        #message("Scenario converted to calibration set")
        if(is.data.frame(data))
          data <- tox_data(data)
        x <- td2cs(x, data, output_var=output)
      }
    }
  }

  # check if type and data_type is one of the 2 options
  type <- match.arg(type)
  data_type <- match.arg(data_type)

  # check correct input parameters
  check_inputs_lik_prof(
    par = par,
    x = x,
    output = output,
    type = type,
    individual = individual,
    data_type = data_type
  )


  # update scenario of calibrationset with the provided par
  for (i in seq_along(x)) {
    x[[i]]@scenario <- x[[i]]@scenario %>%
      set_param(par)
  }

  # initialize parameter list
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # names and boundaries of free parameters
  pnames <- names(par)
  pfree <- list(
    values = par, # free param values
    bounds = get_bounds(x[[1]]@scenario)
  )

  # replace defaults if custom boundaries are given
  if (!is.null(bounds)) {
    check_bounds(bounds)
    pfree$bounds[names(bounds)] <- bounds
  }

  # check that each parameter to profile has boundaries set
  missing <- setdiff(names(par), names(pfree$bounds))
  if(length(missing) > 0) {
    stop("Parameter boundaries missing for parameter", ifelse(length(missing)>1, "s", ""), ": ",
         paste(missing, collapse=", "))
  }

  # Settings for profiling
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # cutoffs for the Chi-square, at a alpha = 0.05 significance level,
  # these values are used in the chi-square test to determine if a fit is
  # significantly different from a previous one (i.e. larger than the cutoff given here)
  chi_crit_j <- stats::qchisq(p=0.95, df=length(pnames))
  chi_crit_s <- stats::qchisq(p=0.95, df=1) # when profiling, we have nested models, differing
  # by 1 parameter (the profiled parameter which is fixed). The likelihood ratio
  # of such nested models follows a Chi-square distribution with 1 degree of freedom


  # setting for profile type
  # values taken from BYOM
  if (type == "fine") {
    # number of iterations
    max_iter <- 50 * (length(pnames) - 1)
    # criteria for change in likelihood ratio
    l_crit_max <- 1 # maximum change in likelihood, above this value, stepsize is decreased
    l_crit_min <- 0.2 # minimum change in likelihood, below this value, stepsize is increased
    l_crit_stop <- chi_crit_s + 5 # stop when likelihood ratio reaches this value
    # settings for step size
    f_step_min <- 0.001 # min stepsize (as fraction of value that is tried)
    f_step_max <- 30 # max stepsize (as fraction of value that is tried)
  } else if (type == "coarse") {
    # number of iterations
    max_iter <- 30 * (length(pnames) - 1)
    # criteria for change in likelihood ratio
    l_crit_max <- 2
    l_crit_min <- 0.7
    l_crit_stop <- chi_crit_s + 2
    # settings for step size
    f_step_min <- 0.005
    f_step_max <- 30
  }


  # The actual profiling
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # a list to store results
  res_list <- list()
  # (important) arguments to this function
  arg_list = list(
    "refit"=refit,
    "type"=type,
    "individual"=individual,
    "log_scale"=log_scale,
    "data_type"=data_type,
    "chi_crit_s"=chi_crit_s
  )
  attr(res_list, "args") <- arg_list

  # profile parameter by parameter
  for (par_select in pnames) {
    # profile the parameter
    res_list[[par_select]] <- suppressWarnings(profile_par(
      par_select = par_select, # par = the parameter to profile
      x = x,
      pfree = pfree,
      type = type,
      output = output,
      max_iter = max_iter,
      f_step_min = f_step_min,
      f_step_max = f_step_max,
      chi_crit_j = chi_crit_j,
      chi_crit_s = chi_crit_s,
      l_crit_max = l_crit_max,
      l_crit_min = l_crit_min,
      l_crit_stop = l_crit_stop,
      refit = refit,
      individual = individual,
      log_scale = log_scale,
      data_type = data_type,
      ...
    ))

    # check if better optimum was found for this parameter (if so, stop profiling)
    if (break_prof) {
      if (any(res_list[[par_select]]$likelihood_profile[, "log_lik_rat"] < 0.00)) {
        # give warning to user
        warning(paste0("Better parameter value found for ", par, ", profiling halted"))
        break
      }
    }
  } # end of parameter for loop

  # Return result list
  class(res_list) <- c(class(res_list), "lik_profile")
  return(res_list)
} # end of lik_profile function



# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Check inputs for likelihood profiling
#
# @description internal function to check if the parameters of the [lik_profile()]
# function are set correctly
#
# @param x either a single [scenario] or a list of [caliset] objects
# @param par named vector - parameters (names and values) to be profiled
# @param output character vector - the output from the [scenario] or [caliset] that is used in calibration
# @param type "fine" or "coarse" (default) likelihood profiling
# @param individual if 'FALSE' (default), the log likelihood is calculated across
# the whole dataset. Alternatively, if 'TRUE', log likelihoods are calculated for
# each (group of) *set*(s) individually.
# @param data_type Character argument, "continuous" (default) or "count", to specify the data type
# for the log likelihood calculations.
#
# @return x as a list of [caliset], and objects error message when needed

check_inputs_lik_prof <- function(par,
                                  x,
                                  output,
                                  type,
                                  individual,
                                  data_type) {

  # check if attempt to profile more params than possible ~~~~~~~~~~~~~~~~~~~
  if (length(par) > 10) {
    stop("attempt to profile more parameters that function allows, reduce no. of parameters")
  }

  # check model (x)~~~~~~~~~~~~~~~~~~~
  # check if Calibrationset or EffectScenario is provided, and convert EffectScenario to CalibrationSet
  if (is.list(x)) {
    if (is(x[[1]]) != "CalibrationSet") {
      stop("incorrect model specified, please provide a scenario or a list of calibration sets")
    }
  } else { ## input is not a list
    if (is_scenario(x)) {
      # all fine
    } else { # input is not a list, and also not an EffectScenario
      stop("incorrect model specified, please provide a scenario or a list of calibration sets")
    }
  }

  # check if the obs match the data_type
  all <- lapply(x, slot, name = "data")
  observed_data_type <- unlist(lapply(all, function(x) is(x[[2]])))
  stopifnot(any(observed_data_type =="numeric" ))
  if(data_type == "count"){
    stopifnot(any(observed_data_type == "integer"))
  }

  # check par ~~~~~~~~~~~~~~~~~~~
  # check if parameters are provided as named vector
  stopifnot(mode(par) == "numeric")
  if (is.null(names(par))) {
    stop("parameters are not provided as a named vector")
  }

  # check output ~~~~~~~~~~~~~~~~~~~
  stopifnot(mode(output) == "character")


  # check data weights and tags if "individual = TRUE" is chosen ~~~~~~~~~~~~~~~
  if (individual == TRUE) {
    data_weights <- lapply(x, slot, name = "weight")
    data_tag <- unlist(lapply(x, slot, name = "tag"))

    # checks to ensure data weights are given properly
    if (length(data_weights) != length(x)) {
      stop("Please ensure all datasets have a weight assigned to them")
    }

    if (any(unlist(lapply(data_weights, length)) > 1)) {
      stop("Please ensure each individual datasets has only 1 weight assigned to them")
    }

    if (any(is.na(unlist(data_weights)))) {
      stop("Please ensure all datasets have a weight assigned to them (NA is not allowed)")
    }

    # checks to ensure appropriate tags are given
    if (!is.null(data_tag)) {
      for (i in seq_along(unique(data_tag))) {
        index <- which(data_tag == unique(data_tag)[i])
        if (length(unique(unlist(data_weights[index]))) != 1) {
          stop("Please ensure datasets with the same tag, have the same weight assigned to them")
        }
      }
    } else { # if no data_tags are given, make sure that all weights are equal
      if (length(unique(unlist(data_weights))) != 1) {
        stop("Please ensure tag names are provided, when weights are assigned to individual data subsets")
      }
    }
  }
  else {
    data_weights <- as.vector(sapply(x, slot, name="weight"))
    if(length(unique(data_weights)) != 1) {
      warning("Assigned dataset weights will be ignored for profiling if argument `individual=FALSE` ...")
    }
  }
}


# TODO make print statements optional
#  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Parameter profiling
#
# @description For a specific parameter, this function will sequentially refit
# a model with the parameter value set to
# increasingly lower and higher values, and the likelihood of the model given the
# data calculated (using function [log_lik()]).
#
# The function was inspired by the MatLab BYOM procedure by
# Tjalling Jager, http://debtox.info/byom.html, as described in Jager 2021:
# Robust Likelihood-Based Optimization and Uncertainty Analysis of Toxicokinetic-
# Toxicodynamic Models" Integrated Environmental Assessment and Management
# 17:388-397 DOI: 10.1002/ieam.4333
#
# @param par_select `character`, the parameter to be profiled
# @param x list of [caliset] objects
# @param pfree list of parameter values and their bounds for the profiled parameter
# @param type "fine" or "coarse" (default) likelihood profiling
# @param output `character` vector - the output from the [caliset] that is used during calibration
# @param max_iter `numeric`, maximum number of profiling iterations to attempts
# @param f_step_min, `numeric`,min stepsize (as fraction of value that is tried)
# @param f_step_max, `numeric`,max stepsize (as fraction of value that is tried)
# @param chi_crit_s, `numeric`,cutoffs for the Chi-square under a 95% probability
# @param l_crit_max, `numeric`,maximum change in likelihood, above this value, stepsize is decreased
# @param l_crit_min, `numeric`,minimum change in likelihood, below this value, stepsize is increased
# @param l_crit_stop, `numeric`,stop when likelihood ratio reaches this value
# @param refit if 'TRUE' (default), refit if a better minimum is found
# @param individual if 'FALSE' (default), the log likelihood is calculated across
# the whole dataset. Alternatively, if 'TRUE', log likelihoods are calculated for
# each (group of) *set*(s) individually.
# @param log_scale 'FALSE' (default), option to calculate the log likelihood on a
# log scale (i.e., observations and predictions are log transformed during calculation)
# @param data_type Character argument, "continuous" (default) or "count", to specify the data type
# for the log likelihood calculations.
# @param ... additional parameters passed on to [stats::optim()] and [calibrate()]
#
# @return A list of containing the likelihood profiling results as a dataframe;
# the 95% confidence interval; the original parameter value; the likelihood plot object; and
# the recalibrated parameter values (in case a lower optimum was found)
#' @autoglobal
profile_par <- function(par_select,
                        x, pfree, type, output,
                        max_iter, f_step_min, f_step_max,
                        chi_crit_j, chi_crit_s, l_crit_max, l_crit_min, l_crit_stop,
                        refit, individual, log_scale, data_type, ...) {
  # initialize
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # initialize empty list to store results for this param
  param_res <- list()


  # print name of parameter being profiled
  message("Profiling: ", par_select)

  # get best fit value (i.e., calibrated value)
  x_hat_orig <- pfree$values[[par_select]]

  # get all data, and weights
  all_data <- lapply(x, slot, name = "data")
  data_tag <- unlist(lapply(x, slot, name = "tag"))
  names(all_data) <- data_tag
  data_weights <- unlist(lapply(x, slot, name = "weight"))

  # Make predictions with the original model
  rs <- eval_cs(x, output=output, verbose=FALSE, .ignore_method=TRUE, ...)

  # Calulate log likelihood with original model
  #   Option 1: calculation of loglik across all datasets
  if (individual == FALSE) {
    # calc log lik
    ll_orig <- log_lik(
      npars = length(pfree$values), # Note: only the free params (i.e. ones that you did the calibration on)
      obs = rs$obs,
      pred = rs$pred,
      log_scale = log_scale,
      data_type = data_type
    )
  } else {
    #  Option 2: calculation for individual sub-datasets, which are then combined
    ll_list <- list()
    subsets <- data.frame(obs=rs$obs,
                     pred=rs$pred,
                     wgts=rs$wgts,
                     tags=unlist(rs$tags)) %>%
      dplyr::group_by(tags) %>%
      dplyr::group_split()

    for (i in seq_along(subsets)) {
      ll_list[[i]] <- log_lik(
        npars = length(pfree$values), # Note: only the free params (i.e. ones that you did the calibration on)
        obs = subsets[[i]]$obs,
        pred = subsets[[i]]$pred,
        log_scale = log_scale,
        data_type = data_type
      ) * unique(subsets[[i]]$wgts)
    }
    ll_orig <- sum(unlist(ll_list))
  }


  # additional setting for profile type
  # values taken from BYOM
  if (type == "fine") {
    if (x_hat_orig == 0) { # if parameter value is zero
      f_step_abs <- 1e-4 # This is just a low value that should be ok in most situations
    } else {
      f_step_abs <- 1e-3 * abs(x_hat_orig) # smallest stepsize (in absolute sense)
    }
  } else if (type == "coarse") {
    if (x_hat_orig == 0) {
      f_step_abs <- 1e-3
    } else {
      f_step_abs <- 1e-2 * abs(x_hat_orig)
    }
  }

  message("start param value: ", round(x_hat_orig, 3), "\nLL:", round(ll_orig, 3))

  # refit the model with the par fixed to lower values, and then higher values
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # initialize list to store results during while loop
  log_lik_rat <- c(0)
  par_value <- c(x_hat_orig)
  results <- list(
    loglik = ll_orig,
    log_lik_rat = log_lik_rat,
    par_value = par_value
  )

  # initialize while loop
  reach_limit <- FALSE
  x_try <- NA
  prof_dir <- -1 # first we go down
  change_dir <- FALSE # we are not yet ready to change profiling direction
  iter <- 2 # 2 because first one will be the original value

  while (reach_limit == FALSE) {
    # jump to a lower/upper par value ~~~~~~~~~~~~~~~~~

    if (is.na(x_try)) { # for 1st iteration (and 1st time changing direction), start from the minimum step size (as in BYOM)
      f_step <- f_step_min * abs(x_hat_orig)
      x_try <- x_hat_orig + prof_dir * f_step
    } else { # L. 619 in BYOM
      if (f_step * abs(x_try) >= f_step_abs) { # if stepsize is larger or equal to abs minimum stepsize
        x_try <- x_try + prof_dir * f_step * abs(x_try) # make new x_try value
      } else {
        if (f_step * abs(x_try) < f_step_abs) { # if stepsize is smaller that abs minimum stepsize
          x_try <- x_try + prof_dir * f_step_abs
        } # make new x_try value
        f_step <- f_step_abs / abs(x_try) # define f_step
      }
    }

    # check that values are within the specified parameter bounds
    if (prof_dir == -1) { # while going down
      if (x_try < pfree$bounds[[par_select]][1]) { # if value is below parameter minimum bound
        x_try <- pfree$bounds[[par_select]][[1]] # profile one last time, using the min bound value
        change_dir <- TRUE # then change direction
        warning("Minimum parameter bound reached")
      }
    } else {
      if (prof_dir == 1) { # while going up
        if (x_try > pfree$bounds[[par_select]][2]) { # if value is above parameter maximum bound
          x_try <- pfree$bounds[[par_select]][[2]] # profile one last time, using the max bound value
          reach_limit <- TRUE # stop profiling
          warning("Maximum parameter bound reached")
        }
      }
    }

    # save new param value, x_try
    results[["par_value"]][iter] <- x_try



    # refit model and get likelihood ~~~~~~~~~~~~~~~

    # assign name to x_try
    names(x_try) <- par_select

    # fix the parameter value to the new (lower/higher) value
    for (i in seq_along(x)) {
      x[[i]]@scenario <- x[[i]]@scenario %>% set_param(param = x_try)
    }

    # give initials for remaining free params
    prof_param_index <- which(names(pfree[["values"]]) == par_select)
    pfree_remaining <- unlist(pfree[["values"]])[-prof_param_index]

    # Make predictions with the newly fitted model
    rs <- eval_cs(x, output=output, verbose=FALSE, .ignore_method=TRUE, ...)

    # Calulate log likelihood with the newly fitted model
    #   Option 1: calculation of loglik across all datasets
    if (individual == FALSE) {
      # calc log lik
      ll_new <- log_lik( # list with only 1 entry
        npars = length(pfree$values), # Note: only the free params (i.e. ones that you did the calibration on)
        obs = rs$obs,
        pred = rs$pred,
        log_scale = log_scale,
        data_type = data_type
      )
    } else {
      #  Option 2: calculation for individual sub-datasets, which are then combined
      ll_list <- list()
      subsets <- data.frame(obs=rs$obs,
                            pred=rs$pred,
                            wgts=rs$wgts,
                            tags=unlist(rs$tags)) %>%
        dplyr::group_by(tags) %>%
        dplyr::group_split()

      for (i in seq_along(subsets)) {
        ll_list[[i]] <- log_lik(
          npars = length(pfree$values), # Note: only the free params (i.e. ones that you did the calibration on)
          obs = subsets[[i]]$obs,
          pred = subsets[[i]]$pred,
          log_scale = log_scale,
          data_type = data_type
        ) * unique(subsets[[i]]$wgts)
      }
      ll_new <- sum(unlist(ll_list))
    }

    # save new LL
    results[["loglik"]][iter] <- ll_new

    # save -2*ln(likelihood ratio) comparing the full (original) model with the nested (1 param fixed) model
    results[["log_lik_rat"]][iter] <- -2 * (ll_new - ll_orig) # matlab BYOM L. 704

    # difference with ll from previous iteration (needed to decide on step size etc.) # matlab BYOM L. 705
    if (iter == 1) {
      delta_l <- abs(results[["loglik"]][iter] - ll_orig)
    } else {
      delta_l <- abs(results[["loglik"]][iter] - results[["loglik"]][iter - 1])
    }

    # evaluate likelihood ~~~~~~~~~~~~~~~~~~~

    # X2 difference test to compare nested models
    if (results[["log_lik_rat"]][iter] > chi_crit_s) { # if the point is outside the 95% conf. region
      if (prof_dir == -1) { # while going down
        change_dir <- TRUE # then change direction
        message("hit 95% Conf Region, changing direction")
      } else {
        if (prof_dir == 1) { # while going up
          reach_limit <- TRUE # stop profiling
          message("hit 95% Conf Region, end of profiling")
        }
      }
    }

    if (delta_l > l_crit_stop) { # stop while-loop if probability is high enough
      reach_limit <- TRUE
      message("critical limit reached, end of profiling")
    } else {
      if (delta_l < l_crit_min) { # if change in likelihood is very small
        f_step <- min(2 * f_step, f_step_max) # double the stepsize, unless greater than max step
      }
    }


    # if needed, change direction
    if (change_dir == TRUE) {
      prof_dir <- 1 # change direction
      x_try <- x_hat_orig # restart from original param value
    }

    iter <- iter + 1
  } # end of while loop


  # reform results list into dataframe
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  res_df <- data.frame(
    par_value = results[["par_value"]],
    log_lik_rat = results[["log_lik_rat"]],
    loglik = results[["loglik"]]
  )


  # get prof region
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  min <- min(res_df$par_value)
  max <- max(res_df$par_value)

  # get 95% CI
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # indicate the first value
  res_df$start <- "No"
  res_df[1, "start"] <- "Yes"
  res_df$start <- as.factor(res_df$start)

  # order dataset, from low to high param value
  res_df <- res_df[order(res_df$par_value), ]

  orig_ind <- which(res_df$start == "Yes")
  # lower 95
  low_df <- res_df[1:orig_ind, ]
  if (min(low_df$log_lik_rat) < 0) { # in case a lower minimum is found within the data subset
    min_ind <- which(low_df$log_lik_rat == min(low_df$log_lik_rat))
    low_df <- low_df[1:min_ind, ]
  }
  if (nrow(low_df) > 1 && !all(low_df[, "log_lik_rat"] == 0)) {
    # needs to be at least 2 values, at least one with a ll not 0
    f_inter <- stats::approxfun(
      y = low_df$par_value,
      x = low_df$log_lik_rat, rule = 2
    )
    low95 <- f_inter(chi_crit_s)
  } else {
    low95 <- low_df$par_value
    if (length(low95) > 1) {
      low95 <- min(low95)
    } # in case multiple values with ll of 0, take the min param value
  }
  # upper 95
  up_df <- res_df[orig_ind:nrow(res_df), ]
  if (min(up_df$log_lik_rat) < 0) { # in case a lower minimum is found within the data subset
    min_ind <- which(up_df$log_lik_rat == min(up_df$log_lik_rat))
    up_df <- up_df[min_ind:nrow(up_df), ]
  }
  if (nrow(up_df) > 1 && !all(up_df[, "log_lik_rat"] == 0)) {
    # needs to be at least 2 values, at least one with a ll not 0
    f_inter <- stats::approxfun(
      y = up_df$par_value,
      x = up_df$log_lik_rat, rule = 2
    )
    up95 <- f_inter(chi_crit_s)
  } else {
    up95 <- up_df$par_value
    if (length(up95) > 1) {
      up95 <- max(up95)
    } # in case multiple values with ll of 0, take the max param value
  }

  # save all outputs for the parameter
  param_res <- list(
    likelihood_profile = res_df,
    confidence_interval = c(low95, up95),
    prof_region = c(min, max),
    orig_par_value = x_hat_orig,
    fit_new = NULL
  )

  # recalibrate if needed
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # Check in results if any log likelihood ratio is lower than the original
  if (refit == TRUE) {
    if (any(res_df[["loglik"]] > ll_orig + 0.01)) {
      # only if the difference is at least 0.01 as in BYOM (value taken from BYOM)
      # this ignores tiny, irrelevant, improvement
      best_par_value <- res_df[which(res_df$log_lik_rat == min(res_df$log_lik_rat)), "par_value"]
      names(best_par_value) <- par_select
      # update par value in model
      for (i in seq_along(x)) {
        x[[i]]@scenario <- x[[i]]@scenario %>% set_param(param = best_par_value)
      }
      # recalibrate
      fit_new <- calibrate(
        x = x,
        par = pfree_remaining,
        output = output,
        log_scale = log_scale,
        verbose = FALSE,
        ...
      )
      # save recalibrated result
      param_res[["fit_new"]] <- list(
        best_fit_param = c(best_par_value, fit_new$par),
        recalibration_outputs = fit_new
      )
    } else { # if no better optimum was found for this parameter
      param_res[["fit_new"]] <- "No recalibration was necessary."
    }
  }
  return(param_res)
} # end of profile_par function


#' Calculate log likelihood
#'
#' @description Calculates the sum of log likelihoods of each observation given
#' the model parameterization.
#'
#' Current implementations enable log likelihood calculations for:
#' 1) continuous data, considering a normal distribution around the prediction
#' for each datapoint,
#' 2) count data, considering a multinomial distribution for data reporting the
#' number of survivors over time.
#'
#' The log likelihood calculation for count data was inspired by a MatLab BYOM v.6.8
#' procedure, created by Tjalling Jager. For details, please refer to BYOM
#' (http://debtox.info/byom.html) as well as Jager (2021).
#'
#' @references
#' Jager T, 2021. Robust Likelihood-Based Optimization and Uncertainty Analysis
#' of Toxicokinetic-Toxicodynamic Models. Integrated Environmental Assessment and
#' Management 17:388-397. \doi{10.1002/ieam.4333}
#'
#' @param obs numeric vector of observed values
#' @param pred numeric vector of predicted values
#' @param data_type determines the if likelihood profiling is conducted for
#' `"continuous"` (default) or `"count"` data
#' @param log_scale `FALSE` (default), option to calculate the log likelihood on a
#' log scale (i.e., observations and predictions are log transformed during calculation)
#' @param npars named numeric vector of parameters that the model was calibrated on,
#' required for `"continuous"` data type, optional for `"count"`.
#'
#' @return the log likelihood value
#' @export
#'
#' @examples
#' # simple example for continuous data #####
#' # observations
#' obs <- c(12, 38, 92, 176, 176, 627, 1283, 2640)
#' # intercept, a, and slope, b, of a Poisson regression fitted through obs
#' pars <- c(a = 2, b = 0.73)
#' # predictions with the Poisson regression
#' pred <- c(15.43, 32.15, 66.99, 139.57, 290.82, 605.94, 1262.52, 2630.58)
#' # example plot
#' plot(seq(1:length(obs)), obs)
#' lines(seq(1:length(obs)), pred)
#' log_lik(
#'   obs = obs,
#'   pred = pred,
#'   npars = length(pars),
#' )
#'
#' # example with count data and GUTS model #####
#' library(dplyr)
#' # observational data
#' dt <- ringtest_c %>% filter(replicate == "E")
#' myexposure <- dt %>% select(time, conc)
#' obs <- dt %>%
#'   mutate(S=Nsurv / max(Nsurv)) %>%
#'   select(time, S)
#' # GUTS model
#' GUTS_RED_IT() %>%
#'   set_param(c(hb = 0)) %>%
#'   set_exposure(myexposure) -> myscenario
#' # fit
#' fit <- calibrate(
#'   x = myscenario,
#'   par = c(kd=1.2, alpha=9.2, beta=4.3),
#'   data = obs,
#'   output = "S")
#' # update
#' myscenario <- myscenario %>% set_param(fit$par)
#' # simulate
#' pred <- myscenario %>% simulate()
#' pred <- pred$S #* max(obs$S)
#' obs <- obs$S
#' # calc likelihood
#' log_lik(obs,
#'     pred,
#'     data_type = "count")
#'

log_lik <- function(obs,
                    pred,
                    data_type = c("continuous", "count"),
                    log_scale = FALSE,
                    npars = NULL){

  # general checks
  stopifnot(length(obs) == length(pred))
  if(any(is.na(c(obs, pred)))){
    warning("missing values in observations or predictions")
  }
  data_type <- match.arg(data_type)

  # transformations
  if(log_scale == TRUE){
    obs = log(obs)
    pred = log(pred)
  }

  # Likelihoods
  if(data_type == "continuous"){

    # specific checks
    stopifnot(is.numeric(npars))
    stopifnot(length(npars) == 1)

    # calculations
    k <- npars
    res <- obs - pred
    n <- length(res)
    SSE <- sum(res^2)
    sigma <- sqrt(SSE / (n - k))
    sigma_unbiased <- sigma * sqrt((n - k) / n)

    # log likelihood for normal probability density
    LL <- sum(log(stats::dnorm(x = obs, mean = pred, sd = sigma_unbiased)))
    return(LL)
  }
  if(data_type == "count"){

    # calculations
    Ndeaths <- -diff(obs)       # numbers of deaths
    Mdeaths <- -diff(pred)      # conditional probabilities of deaths
    Mdeaths <- max(Mdeaths,1e-50)   # otherwise we get problems taking the logarithm
    pred    <- max(pred,1e-50)      # otherwise we get problems taking the logarithm

    # log likelihood for count data (multinomial)
    LL  <- sum( Ndeaths * log(Mdeaths) )
    return(LL)
  }

}

test_that("log-likelihood gives expected error", {

  # error for too many params
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                           par = Lemna_Schmitt()@param,
                           data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                             obs = c(12, 38, 92, 176, 176, 627, 1283, 2640)),
                           output = "BM")))

  # error for data_type
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                                            par = c(k_phot_max = 5.6,
                                                    k_resp = 1.9),
                                            data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                                              obs = c(12, 38, 92, 176, 176, 627, 1283, 2640)),
                                            output = "BM",
                                            data_type = "foo")))

  # error for data_type
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                                            par = c(k_phot_max = 5.6,
                                                    k_resp = 1.9),
                                            data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                                              obs = c(12, 38, 92, 176.04, 176, 627, 1283, 2640)),
                                            output = "BM",
                                            data_type = "count")))

  # error for data_type
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                                            par = c(k_phot_max = 5.6,
                                                    k_resp = 1.9),
                                            data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                                              obs = c(12, 38, 92, "foo", 176, 627, 1283, 2640)),
                                            output = "BM",
                                            data_type = "count")))

  # error for model
  # # error because of misspecified model (not calibration set nor scenario)
  # expect_error(lik_profile(x = "model",
  #                          par = c(k_phot_max = 5.6,
  #                                  k_resp = 1.9),
  #                          output = "BM"))

  # error because data missing for EffectScenario
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                           output = "BM",
                           par = c(k_phot_max = 5.6,
                                         k_resp = 1.9))))

  # errors for pars_profile
  # error because parameter vector is not a vector
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                           output = "BM",
                           data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                             obs = c(12, 38, 92, 176, 176, 627, 1283, 2640)),
                           par = Lemna_Schmitt()@param[c(1,2)])))

  # # error because parameter vector is not named
  # expect_error(lik_profile(x = Lemna_Schmitt(),
  #                          output = "BM",
  #                          data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
  #                                            obs = c(12, 38, 92, 176, 176, 627, 1283, 2640)),
  #                          par = as.numeric(Lemna_Schmitt()@param[c(1,2)])))

  # errors for output
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                           output = list("BM"),
                           data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                             obs = c(12, 38, 92, 176, 176, 627, 1283, 2640)),
                           par = c(k_phot_max = 5.6,
                                            k_resp = 1.9))))

  # errors for profile type
  expect_error(suppressMessages(lik_profile(x = Lemna_Schmitt(),
                           output = "BM",
                           data = data.frame(t = c(0, 3, 5, 7, 7.01, 10, 12, 14),
                                             obs = c(12, 38, 92, 176, 176, 627, 1283, 2640)),
                           par = c(k_phot_max = 5.6,
                                            k_resp = 1.9),
                           type = "slow")))

})



test_that("likelihood profiling works", {

  # Inputs for likelihood profiling
  obs <- schmitt2013 %>%
    dplyr::filter(trial == "T0") %>%
    dplyr::select(time, BM=obs)
  params <- c(k_phot_max = 3.979200,
              k_resp =  7.700779e-08)
  sc <- metsulfuron %>%
    set_param(params)

  # Likelihood profiling
  suppressMessages(
    res <- lik_profile(
      x = sc,
      data = obs,
      output = "BM",
      par = params[1],
      type = "fine"
    )
  )

  # tests
  expect_equal(length(res$k_phot_max$confidence_interval), 2)
  expect_equal(dim(res$k_phot_max$likelihood_profile)[2], 4)
  expect_equal(res$k_phot_max$orig_par_value, 3.9792)
})

test_that("parameter space explorer works", {

  # exposure - control run
  exp <- data.frame(time = c(0,3,5,7,7.01,10,12,14),
                    conc = rep(0, 8))

  # observations - control run
  obs <- data.frame(time = c(0,3,5,7,7.01,10,12,14),
                    BM = c(12,38,92,176,176,627,1283,2640))

  # parameters after calibration
  params <- c(k_phot_max = 5.663571,
              k_resp = 1.938689)

  bounds = list(k_phot_max = list(0,30),
                    k_resp = list(0,10))


  # update metsulfuron
  sc <- metsulfuron %>%
    set_init(c(BM  = 5, E = 1,  M_int = 0)) %>%
    set_param(list(k_0 = 5E-5,
                   a_k =  0.25,
                   BM50 = 17600,
                   mass_per_frond = 0.1)) %>%
    set_exposure(exp) %>%
    set_param(params)

  sc <- sc %>%
    set_bounds(bounds)

  # Likelihood profiling
  suppressMessages(
    res <- lik_profile(x = sc,
                       data = obs,
                       output = "BM",
                       par = params,
                       refit = FALSE,
                       type = "fine",
                       method="Brent")
  )

  # parameter space explorer
  suppressMessages(
    Par_exp <- explore_space(x = list(caliset(sc, obs)),
                  par = params,
                  res = res,
                  output = "BM",
                  sample_size = 1000,
                  max_iter = 1,   # for speed, here put to 1, please increase for improved results
                  nr_accept = 100)
  )

  # tests
  expect_equal(ncol(Par_exp), 5)
  expect_equal(Par_exp[1,4], 0)
  expect_equal(Par_exp[1,5], "Original fit")
})

test_that("log-likelihood works: continous data, incl. log scale", {

  npars <- 2
  obs <- c(12, 38, 92, 176, 176, 627, 1283, 2640)
  pred <- c(5.0, 21.256, 55.586, 144.162, 144.845, 574.043, 1323.999, 2632.258)
  res <- log_lik(npars = npars,
                 obs = obs,
                 pred = pred)
  res_log <- log_lik(npars = npars,
                   obs = obs,
                   pred = pred,
                   log_scale = TRUE)

  expect_equal(res, -39.074, tolerance = 10e-3)  # expected value comes from previous cvasi runs
  expect_equal(res_log, -4.50445, tolerance = 10e-3) # expected value comes from previous cvasi runs

})

test_that("log-likelihood works: count data", {

  # observational data
  dt <- ringtest_c %>%
    dplyr::filter(replicate == "E")
  myexposure <- dt %>%
    dplyr::select(time, conc)
  obs <- dt %>%
    dplyr::mutate(S=Nsurv / max(Nsurv)) %>%
    dplyr::select(time, S)
  # GUTS model
  GUTS_RED_IT() %>%
    set_param(c(hb=0)) %>%
    set_exposure(myexposure) -> myscenario
  # fit
  suppressWarnings({
    fit <- calibrate(
      x = myscenario,
      par = c(kd=1.2, alpha=9.2, beta=4.3),
      data = obs,
      output = "S",
      verbose=FALSE)
  })
  # update
  myscenario <- myscenario %>%
    set_param(fit$par)
  # simulate
  pred <- myscenario %>%
    simulate()
  # calc likelihood
  res <- log_lik(obs = obs$S,
          pred =  pred$S,
          data_type = "count")

  expect_equal(res, -0.21, tolerance = 10e-3) # expected value comes from previous cvasi runs
})



test_that("log-likelihood gives expected error", {

  npars <- 2
  obs <- c(12, 38, 92, 176, 176, 627, 1283, 2640)
  pred <- c(5.0, 21.256, 55.586, 144.162, 144.845, 574.043, 1323.999, 2632.258)

  expect_error(log_lik(npars = npars,
                       obs = obs[-1],
                       pred = pred))

  expect_error(log_lik(npars = "test",
                       obs = obs,
                       pred = pred))

  expect_error(log_lik(npars = npars,
                       obs = obs,
                       pred = pred,
                       data_type = "foo"))

})

# TODO make print statements optional
# TODO reuse/Abstract profiling code from lik_profile()

#' Explore parameter space
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' The function is aimed at getting an idea of how the parameter space
#' of a model behaves, so that parameter identifiability problems and correlations
#' between parameters can be explored. Therefore, the function samples a large
#' number of parameter sets by randomly drawing from each parameter's 95%
#' confidence interval (generated by [lik_profile()]). It then
#' checks how many of the parameter sets are within acceptable limits by comparing
#' the likelihood ratio of a parameter set vs. the original parameter set against
#' a chi-square distribution as degrees of freedom (df) the total number of profile
#' parameters (outer rim) or one df (inner rim). If needed, the function resamples
#' until at least `nr_accept` parameters sets are within the inner rim
#'
#' @param x a list of [caliset] objects
#' @param par best fit parameters from joined calibration
#' @param res output of [lik_profile()] function
#' @param output character vector, name of output column of [simulate()] that
#'  is used in calibration
#' @param data only needed if `x` is a [scenario]
#' @param sample_size number of samples to draw from each parameter interval
#' @param max_iter max number of iterations to redraw samples (within a smaller space), and repeat the process
#' @param nr_accept threshold for number of points sampled within the inner circle
#' @param sample_factor multiplication factor for sampling (95% interval * sample factor)
#' @param individual if `FALSE` (default), the log likelihood is calculated across
#' the whole dataset. Alternatively, if `TRUE`, log likelihoods are calculated for
#' each (group of) *set*(s) individually.
#' @param log_scale `FALSE` (default), option to calculate the log likelihood on a
#' log scale (i.e., observations and predictions are log transformed during calculation)
#' @param data_type Character argument, `"continuous"` (default) or `"count"`, to specify the data type
#' for the log likelihood calculations.
#' @param max_runs *deprecated* alias for `max_iter` parameter
#' @param ... additional parameters passed through to [simulate()]
#'
#' @return a list containing a plot to explore the parameter space, and the `data.frame`
#' supporting it
#'
#' @seealso [plot.param_space]
#' @export
#' @autoglobal
#' @importFrom lifecycle is_present deprecated
#' @examples
#' \donttest{
#' library(dplyr)
#' # Example with Lemna model - physiological params
#' # Before applying the function, a model needs to be calibrated and its parameters profiled
#' # Inputs for likelihood profiling
#'
#' # observations - control run
#' obs <- schmitt2013 %>%
#'   filter(trial == "T0")
#'
#' # parameters after calibration
#' params <- c(
#'   k_phot_max = 5.663571,
#'   k_resp = 1.938689,
#'   Topt = 26.7
#' )
#'
#' # set parameter boundaries (if different from defaults)
#' bounds <- list(
#'   k_resp = list(0, 10),
#'   k_phot_max = list(0, 30),
#'   Topt = list(20, 30)
#' )
#'
#' # update metsulfuron
#' myscenario <- metsulfuron %>%
#'   set_init(c(BM = 1.2, E = 1, M_int = 0)) %>%
#'   set_param(list(
#'     k_0 = 5E-5,
#'     a_k = 0.25,
#'     BM50 = 17600,
#'     mass_per_frond = 0.1
#'   )) %>%
#'   set_noexposure() %>%
#'   set_param(params) %>%
#'   set_bounds(bounds)
#'
#' # Likelihood profiling
#' res <- lik_profile(
#'   x = myscenario,
#'   data = obs,
#'   output = "FrondNo",
#'   par = params,
#'   refit = FALSE,
#'   type = "fine",
#'   method = "Brent"
#' )
#' # plot
#' plot(res)
#'
#' # parameter space explorer
#' set.seed(1) # for reproducibility
#' res_space <- explore_space(
#'   x = myscenario,
#'   data = obs,
#'   par = params,
#'   res = res,
#'   output = "FrondNo",
#'   sample_size = 1000,
#'   max_iter = 20,
#'   nr_accept = 100)
#'
#' plot(res_space)
#'
#' }
explore_space <- function(x,
                          par,
                          res,
                          output,
                          data,
                          sample_size = 1000,
                          max_iter = 30,
                          nr_accept = 100,
                          sample_factor = 1.2,
                          individual = FALSE,
                          log_scale = FALSE,
                          data_type = c("continuous", "count"),
                          max_runs=deprecated(),
                          ...) {
  if(is_present(max_runs)) {
    lifecycle::deprecate_warn("1.5.0", "cvasi::explore_space(max_runs)", "cvasi::explore_space(max_iter)")
    max_iter <- max_runs
  }

  # some checks
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # convert to list of calibration sets if needed
  if (length(x) == 1) {
    if (is_scenario(x)) {
      if(is.null(data)) {
        stop("Scenario provided, but argument 'data' is missing")
      } else {
        #message("Scenario converted to calibration set")
        if(is.data.frame(data))
          data <- tox_data(data)
        x <- td2cs(x, data, output_var=output)
      }
    }
  }
  # check sensible use of function
  if(length(output) > 1){
    stop("Parameter `output` must have length of one")
  }
  if (length(names(res)) <= 1) {
    stop("parameter space exploration only makes sense for 2 or more parameters")
  }
  # check if the same output is used throughout
  #if(!(output %in% x[[1]]@scenario@endpoints)){
  #  stop("Specified output not present in scenario object (x)")
  #}
  #if(!(output %in% colnames(x[[1]]@data))){
  #  stop("Specified output not present in data of scenario object (x)")
  #}
  # check if explored parameters are part of the model
  if (any(names(res) %in% names(x[[1]]@scenario@param) == "FALSE")) {
    stop("Profiled parameters (in res) are not part of the scenario object (x)")
  }
  # check if boundaries for explored parameters are available
  if(any(names(res) %in% names(x[[1]]@scenario@param.bounds) == "FALSE")){
    stop("No parameter boundaries available for explored parameters, please set bounds in scenario object (x)")
  }
  if(!is.numeric(sample_factor))
    stop("Argument 'sample_factor' must be numeric")
  if(length(sample_factor) != 1)
    stop("Argument 'sample_factor' must be of length one")
  if(is.na(sample_factor) | is.nan(sample_factor))
    stop("Argument 'sample_factor' has invalid value")
  if(sample_factor <= 1)
    stop("Argument 'sample_factor' out of range, must be greater than 1.0")


  # check that a correct option for data_type is entered
  data_type <- match.arg(data_type)


  # calculate log likelihood with original model, for the profiled pars
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  # get all data, and weights
  all_data <- lapply(x, slot, name = "data")
  data_tag <- unlist(lapply(x, slot, name = "tag"))
  names(all_data) <- data_tag
  data_weights <- unlist(lapply(x, slot, name = "weight"))

  # Make predictions with the original model
  rs <- eval_cs(set_param(x, par), output=output, verbose=FALSE, .ignore_method=TRUE, ...)

  # Calulate log likelihood with the original model
  #   Option 1: calculation of loglik across the whole dataset
  if (individual == FALSE) {
    # calc log lik
    ll_orig <- log_lik(
      npars = length(res),
      obs = rs$obs,
      pred = rs$pred,
      log_scale = log_scale,
      data_type = data_type
    )
  } else {
    #  Option 2: calculation for individual sub-datasets, which are then combined
    ll_list <- list()
    subsets <- data.frame(obs=rs$obs,
                          pred=rs$pred,
                          wgts=rs$wgts,
                          tags=unlist(rs$tags)) %>%
      dplyr::group_by(tags) %>%
      dplyr::group_split()

    for (i in seq_along(subsets)) {
      ll_list[[i]] <- log_lik(
        npars = length(res),
        obs = subsets[[i]]$obs,
        pred = subsets[[i]]$pred,
        log_scale = log_scale,
        data_type = data_type
      ) * unique(subsets[[i]]$wgts)
    }
    ll_orig <- sum(unlist(ll_list))
  }

  # obtain a (random uniform) sample from each parameter's profile region,
  # and include the original param values
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  par_sample_list <- list()
  for (i in 1:length(res)) {
    par_sample_list[[i]] <- c(
      res[[i]]$orig_par_value, # original par value is put first
      stats::runif(sample_size,
        min = res[[i]]$prof_region[1] / sample_factor,
        max = res[[i]]$prof_region[2] * sample_factor
      )
    )
    names(par_sample_list)[[i]] <- names(res[i])
  }
  par_sample <- data.frame(par_sample_list)

  # calculate LL, LLR for each parameter set
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  LL <- NULL
  ## TODO parallelize sample evaluation
  for (i in 1:nrow(par_sample)) {

    # make predictions
    rs <- eval_cs(set_param(x, par_sample[i, ]), output=output, verbose=FALSE, .ignore_method=TRUE, ...)

    # calculate log likelihood
    #   Option 1: calculation of loglik across the whole dataset
    if (individual == FALSE) {
      # calc log lik
      ll_new <- log_lik( # list with only 1 entry
        npars = length(res),
        obs = rs$obs,
        pred = rs$pred,
        log_scale = log_scale,
        data_type = data_type
      )
    } else {
      #  Option 2: calculation for individual sub-datasets, which are then combined
      ll_list <- list()
      subsets <- data.frame(obs=rs$obs,
                            pred=rs$pred,
                            wgts=rs$wgts,
                            tags=unlist(rs$tags)) %>%
        dplyr::group_by(tags) %>%
        dplyr::group_split()

      for (i in seq_along(subsets)) {
        ll_list[[i]] <- log_lik(
          npars = length(res),
          obs = subsets[[i]]$obs,
          pred = subsets[[i]]$pred,
          log_scale = log_scale,
          data_type = data_type
        ) * unique(subsets[[i]]$wgts)
      }
      ll_new <- sum(unlist(ll_list))
    }

    LL <- c(LL, sum(unlist(ll_new)))
  } # end of loop per sample i

  # add likelihood to the dataframe
  par_sample$LL <- LL

  # calc likelihood ratio: -2*ln(likelihood ratio)
  par_sample$LLR <- -2 * (par_sample$LL - ll_orig) # matlab BYOM L. 704

  # X2 difference test to compare nested models
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # cutoffs for the Chi-square test, for alpha = 0.05 significance level,
  # these values are used in the chi-square test to determine if a fit is significantly
  # different from a previous one (i.e. larger than the cutoff given here)
  chi_crit_j <- stats::qchisq(p=0.95, df=length(res)) # cut-off for the 95% profile region
  chi_crit_s <- stats::qchisq(p=0.95, df=1) # cut-off for single par. CI

  # clean dataset based on criteria
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  par_sample$LLR_accept <- ifelse(par_sample$LLR < chi_crit_j, "accept", "reject") # accept combinations within the profile region
  par_sample$LLR_quality <- ifelse(par_sample$LLR < chi_crit_s, "Inner", "Outer") # give preference to those with 95% CI inner rim
  # calculate nr of points within inner rim
  inner_ind <- which(par_sample$LLR_quality == "Inner") # remember which ones are in the inner rim (used/updated later in the while loop)
  inner_size <- length(inner_ind)

  # Additional sampling (repeating steps above with slight modifications)
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  sample_iter <- 1
  while (inner_size < nr_accept) { # number of accepted points should be >= nr_accept

    message("resampling parameter space, iteration: ", sample_iter)

    # obtain a (random normal) sample based on the previously "inner" rim samples
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ## FIXME TODO if 1st sampling resulted in a very small sample size of 'accepted'
    ##    then re-sampling might fail or will be unstable, 1st sampling must be repeated
    ##    until sample size n >= 10?
    par_sample_list <- list()
    for (i in 1:length(res)) {
      par_sample_list[[i]] <- stats::rnorm(sample_size,
        mean = mean(par_sample[inner_ind, i]),
        sd = stats::sd(par_sample[inner_ind, i]) * 1.5
      )
      names(par_sample_list)[[i]] <- names(res[i])
    }
    par_sample_new <- data.frame(par_sample_list)

    # ensure that values are all within param bounds
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    for(i in 1:length(res)){ # for each parameter
      nm <- names(res)[i]
      too_small <- which(par_sample_new[, i] < x[[1]]@scenario@param.bounds[[nm]][1])
      too_large <- which(par_sample_new[, i] > x[[1]]@scenario@param.bounds[[nm]][2])
      if(length(c(too_small, too_large)) > 0){ # remove samples outside boundaries (if present)
        par_sample_new <- par_sample_new[-c(too_small, too_large), ]
        if (nrow(par_sample_new) == 0) {
          stop("all sampled parameter values out of bounds")
        }
      }
    }

    # calculate LL, LLR for each parameter set
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    LL <- NULL
    ## TODO parallize enumeration, combine code with initial LL/LLR loop
    for (i in 1:nrow(par_sample_new)) {

      rs <- eval_cs(set_param(x, par_sample_new[i, ]), output=output, verbose=FALSE, .ignore_method=TRUE, ...)

      # calculate log likelihood
      #   Option 1: calculation of loglik across the whole dataset
      if (individual == FALSE) {
        # calc log lik
        ll_new <- log_lik( # list with only 1 entry
          npars = length(res),
          obs = rs$obs,
          pred = rs$pred,
          log_scale = log_scale,
          data_type = data_type
        )
      } else {
        #  Option 2: calculation for individual sub-datasets, which are then combined
        ll_list <- list()
        subsets <- data.frame(obs=rs$obs,
                              pred=rs$pred,
                              wgts=rs$wgts,
                              tags=unlist(rs$tags)) %>%
          dplyr::group_by(tags) %>%
          dplyr::group_split()

        for (i in seq_along(subsets)) {
          ll_list[[i]] <- log_lik(
            npars = length(res),
            obs = subsets[[i]]$obs,
            pred = subsets[[i]]$pred,
            log_scale = log_scale,
            data_type = data_type
          ) * unique(subsets[[i]]$wgts)
        }
        ll_new <- sum(unlist(ll_list))
      }

      LL <- c(LL, sum(unlist(ll_new)))
    } # end of loop per sample i

    # add likelihood to the dataframe
    par_sample_new$LL <- LL

    # calc likelihood ratio: -2*ln(likelihood ratio)
    par_sample_new$LLR <- -2 * (par_sample_new$LL - ll_orig) # matlab BYOM L. 704


    # clean dataset based on criteria
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    par_sample_new$LLR_accept <- ifelse(par_sample_new$LLR < chi_crit_j, "accept", "reject") # accept combinations within the profile region
    par_sample_new$LLR_quality <- ifelse(par_sample_new$LLR < chi_crit_s, "Inner", "Outer") # give preference to those with 95% CI inner rim

    # join with previous df
    par_sample <- rbind(par_sample, par_sample_new)

    # calculate nr of points within inner rim (and update inner_ind)
    inner_ind <- which(par_sample$LLR_quality == "Inner") # remember which ones are in the inner rim (used/updated later in the while loop)
    inner_size <- length(inner_ind)

    sample_iter <- sample_iter + 1
    if (sample_iter > max_iter) {
      message("Maximum number of iterations exceeded, increase 'max_iter' for improved results")
      break()
    }
  } # end while loop

  # find best LLR value, and compare with original LLR (which is per definition = 0)
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if (min(par_sample$LLR, na.rm=TRUE) < 0) {
    message("better optimum found in space explorer")
    best_fit_ind <- which.min(par_sample$LLR)
    par_sample$LLR_quality[best_fit_ind] <- "Best fit"
  }
  par_sample[1, "LLR_quality"] <- "Original fit"


  # clean and return
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  output <- par_sample %>%
    dplyr::filter(LLR_accept == "accept") %>%
    dplyr::select(!c(LLR_accept))

  class(output) <- c("param_space", class(output))
  return(output)
}

#' Creates plot of model results (uncertainties optional)
#'
#' `r lifecycle::badge("experimental")`
#'
#' All parameter combinations and exposure patterns are simulated and the mean
#' of predictions is derived for a single study. The uncertainty is
#' passed to the function due to computation time. Results are displayed by
#' plotting the time series including the uncertainty interval. Observation data
#' can be optionally displayed. Data should be provided in long format. Function
#' plots the time (column 1) and the predictions (column 2, can be changed by
#' the user plot_col)
#'
#' @param model_base effect scenario object with mean parameters
#' @param treatments treatments exposure levels as data frame
#' @param rs_mean `data.frame`, model results best fit params
#' @param rs_range `data.frame`, uncertainties as data frame
#' @param obs_mean `data.frame`, observation data with means per treatment level
#' @param obs_full `data.frame`, full set including results for replicates
#' @param x_breaks optional vector of breaks of x-axis
#' @param y_lim optional vector containing limits of y-axis
#' @param grid_labels optional labels of grid headers
#' @param grid_ncol optional number of grid columns
#' @param plot_col output column which should be plotted
#' @param y_title optional title of y-axis
#' @param ... any additional parameters
#' @return a ggplot2 plot object
#' @global trial
#' @export
#'
#' @examples
#' set.seed(124)
#' exposure <- data.frame(
#'   time = 0:21,
#'   conc = rnorm(n = 22, mean = 0.1, sd = 0.06),
#'   trial = "T1"
#' )
#' forcings <- list(temp = 12, rad = 15000)
#' param <- list(EC50 = 0.3, b = 4.16, P_up = 0.0054)
#' inits <- list(BM = 0.0012, E = 1, M_int = 0)
#'
#' scenario <- Lemna_Schmitt() %>%
#'   set_forcings(forcings) %>%
#'   set_param(param) %>%
#'   set_init(inits)
#'
#' sim_result <- simulate_batch(
#'   model_base = scenario,
#'   treatments = exposure,
#'   param_sample = NULL
#' )
#'
#' plot_sd(
#'   model_base = scenario,
#'   treatments = exposure,
#'   rs_mean = sim_result
#' )
plot_sd <- function(model_base,
                    treatments,
                    rs_mean,
                    rs_range = NULL,
                    obs_mean = NULL,
                    obs_full = NULL,
                    x_breaks = NULL,
                    y_lim = NULL,
                    grid_labels = NULL,
                    grid_ncol = 2,
                    plot_col = 2,
                    y_title = NULL, ...) {
  # check format
  if (!is.data.frame(treatments) && !is.null(treatments)) {
    stop("treatments not a data.frame")
  }
  if (!is.data.frame(rs_range) && !is.null(rs_range)) {
    stop("rs_range not a data.frame")
  }
  if (!is.data.frame(obs_mean) && !is.null(obs_mean)) {
    stop("obs_mean not a data.frame")
  }
  if (!is.data.frame(obs_full) && !is.null(obs_full)) {
    stop("obs_full not a data.frame")
  }

  no_trials <- length(unique(rs_mean$trial))

  # ToDo check class calibration set / set more than one exposure in an effect
  # scenario object as an alternative
  # ToDo check style guide
  # ToDo combine treatments and data (obs_mean)
  # ToDo how to handle replicates? Plot replicates? Plot uncertainties?

  exp_plot <- treatments # long format
  exp_plot$trial <- as.factor(exp_plot$trial)

  # try to guess the simulated time period
  t_max <- max(treatments[, 1]) # 1st column should contain time
  # check if replicates are present if obs_full is not null, otherwise there is
  # something funny going on with the column names. was data imported with
  # read.csv(..., check.names=FALSE) ?? read.csv should not be part of the plot
  # function. Transformation should be done before plotting? Since dplyr/tidy
  # is used, we should use always long format?
  if (!is.null(obs_full)) {
    if (length(unique(names(obs_full))) == length(names(obs_full))) {
      warning("no replicates found: check col names
              in full experimental dataset")
    }
  }

  # try to find the max number of data for setting y-axis
  obs_max <- rs_mean %>%
    dplyr::select(dplyr::all_of(plot_col)) %>%
    max()

  if (!is.null(obs_full)) {
    obs_plot <- obs_full
    obs_max <- max(obs_max, obs_plot[,2], na.rm=TRUE) #2 = data, observations
    obs_plot$trial <- as.factor(obs_plot$trial) #trial
  } else if (!is.null(obs_mean)) {
    obs_plot <- obs_mean
    obs_max <- max(obs_max, obs_plot[,2], na.rm=TRUE)
    obs_plot$trial <- as.factor(obs_plot$trial) #trial
  } else {
    obs_plot <- NULL
  }

  obs_max <- obs_max * 1.10 # 10 per cent for the symbols

  # determine max of x-axis and scale the x value
  exp_max <- max(exp_plot[,2]) #?? col instead of name?
  if (obs_max > exp_max * 10 | exp_max > obs_max) {
    exp_plot[,2] <- exp_plot[,2] / exp_max * obs_max
  }

  # color scheme
  col_palette <- c(
    "#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
    "#0072B2", "#D55E00", "#CC79A7"
  )
  if (length(obs_mean) - 1 > length(col_palette)) {
    stop("color palette has too few elements, please amend")
  }
  # random colors instead?
  # set default breaks for x axis
  if (is.null(x_breaks)) {
    x_breaks <- grDevices::axisTicks(usr = range(treatments[, 1]),
                                     log = FALSE,
                                     axp = NULL,
                                     nint = 5)
  }
  # set default y axis limits
  if (is.null(y_lim)) {
    y_lim <- c(0, obs_max)
  }
  # set default labels
  if (is.null(grid_labels)) {
    # grid_labels is an optional function parameter
    grid_labels <- unique(exp_plot$trial)
    # create labeller for grid headers
    names(grid_labels) <- unique(exp_plot$trial)
  }
  f_lb <- ggplot2::labeller(trial = grid_labels)

  # set y title
  if (is.null(y_title)) {
    y_title <- names(rs_mean)[plot_col]
  }

# create plot
  # one plot
  if(no_trials == 1){ #ToDo? Exposure with second y-axis
    plot <- ggplot2::ggplot() +
      # best fit
      ggplot2::geom_line(data = rs_mean, ggplot2::aes(x = rs_mean[,1], #time
                                                      y = rs_mean[, plot_col], #output selected by user
                                                      color = "black"),
                         #show.legend = TRUE
      ) +
      # exposure profiles
      ggplot2::geom_area(data = exp_plot, ggplot2::aes(x = exp_plot[,1], #time
                                                       y = exp_plot[,2], #conc
                                                       fill = "grey"),
                         alpha = 0.12,
                         position = "identity",
                         #show.legend = TRUE
      )

    # colors
    plot <- plot +
      ggplot2::scale_color_manual(values = "black", name = "Prediction") +
      ggplot2::scale_fill_manual(values = "grey", name = "Exposure") +
      ggplot2::scale_x_continuous(name = "Time (days)", breaks = x_breaks) +
      ggplot2::scale_y_continuous(name = y_title) +
      #ggplot2::scale_y_continuous(sec.axis = sec_axis(~./1,name = "Exposure")) +
      # (name of column)
      ggplot2::coord_cartesian(xlim = c(0, t_max), ylim = y_lim) + # boundaries
      ggplot2::theme_bw() +
      ggplot2::theme(legend.position = "top")

  } else {
    plot <- ggplot2::ggplot() +

      # best fit
      ggplot2::geom_line(data = rs_mean, ggplot2::aes(x = rs_mean[,1], #time
                                                      y = rs_mean[, plot_col], #output selected by user
                                                      color = trial)
      ) +
      # exposure profiles
      ggplot2::geom_area(data = exp_plot, ggplot2::aes(x = exp_plot[,1], #time
                                                       y = exp_plot[,2], #conc
                                                       fill = "black"),
                         alpha = 0.12,
                         position = "identity"
      ) +
      # Facet by trial
      ggplot2::facet_wrap(~ trial, ncol = grid_ncol, labeller = f_lb)

    # colors
    plot <- plot +
      ggplot2::scale_color_manual(values = col_palette) +
      ggplot2::scale_fill_manual(values = col_palette) +
      ggplot2::scale_x_continuous(name = "Time (days)", breaks = x_breaks) +
      ggplot2::scale_y_continuous(name = y_title) +
      # (name of column)
      ggplot2::coord_cartesian(xlim = c(0, t_max), ylim = y_lim) + # boundaries
      ggplot2::theme_bw() +
      ggplot2::theme(legend.position = "none")

    if (!is.null(obs_plot)) {
      plot <- plot + # observations, data
        ggplot2::geom_point(data = obs_plot, ggplot2::aes(x = obs_plot[,1], #time
                                                          y = obs_plot[,2], #observations
                                                          bg = trial),
                            size = 3, shape = 22
        ) +
        # Facet by trial
        ggplot2::facet_wrap(~ trial, ncol = grid_ncol, labeller = f_lb)
    }

    if (!is.null(rs_range)) {
      plot <- plot + # uncertainties, data
        ggplot2::geom_ribbon(data = rs_range, ggplot2::aes(x = rs_range[,1], #time
                                                           ymin = rs_range[,2], #min
                                                           ymax = rs_range[,3], #max
                                                           fill = trial),
                             alpha = 0.35
        ) +
        # Facet by trial
        ggplot2::facet_wrap(~ trial, ncol = grid_ncol, labeller = f_lb)
    }
  }
  print(plot)
}

#' Creates a PPC plot for a single dataset
#'
#' `r lifecycle::badge("experimental")`
#'
#' A sample of parameters representing the uncertainty within the dataset
#' is passed to the function. All parameter combinations and exposure patterns
#' are simulated and the range of predicted frond numbers is derived for a
#' single study. The uncertainty is displayed by a Posterior Predictive Plot
#' (PPC). The data (rs_mean, obs_mean and obs_full) must have the
#' following format (col1 = time, col2 = data of interest, col3 = trial name).
#' Data for uncertainties (rs_range) must have the format: col1 = time,
#' col2 = lower boundaries, col3 = upper boundaries, col4 = trial. The user
#' should take care of the input data and consider whether control data and
#' data at time zero should be included in the model check.
#'
#' @param rs_mean `data.frame`, model results best fit params
#' @param col_number, column to plot, default = 2
#' @param rs_range `data.frame`, predictions (min, max from param.sample run)
#' @param obs_mean `data.frame`, observations with means per treatment level
#' @param obs_full `data.frame`, full data set including results for replicates
#' @param xy_lim optional `numeric`, limits of x and y axis for plotting
#' @param study optional `string`, name of study which can be used as key
#' @global time trial
#' @export
#' @global time trial
#'
#' @return a ggplot2 plot object
plot_ppc <- function(rs_mean,
                     rs_range,
                     col_number = 2,
                     obs_mean = NULL,
                     obs_full = NULL,
                     xy_lim = NULL,
                     study = NULL) {
  # Check if 'out.file' is in the arguments
  if ("outfile" %in% names(args)) {
    lifecycle::deprecate_stop(details = "The 'outfile' argument is no longer
                              supported. Please adjust your code accordingly.")
  }

  if (!is.data.frame(obs_mean) && !is.null(obs_mean)) {
    stop("obs_mean not a data.frame")
  }
  if (!is.data.frame(obs_full) && !is.null(obs_full)) {
    stop("obs_full not a data.frame")
  }

  # try to find the maximumn value of observations
  if (is.null(xy_lim)) {
    if (!is.null(obs_full)) {
      xy_lim <- max(obs_full[, col_number]) # observation data
    } else {
      xy_lim <- max(obs_mean[, col_number]) # observation data
    }
  }

  # check if replicates are present, otherwise there is something funny going
  # on with the column names. was data imported with
  # read.csv(..., check.names=FALSE) ??
  if (!is.null(obs_full)) {
    if (length(unique(names(obs_full))) == length(names(obs_full))) {
      warning("no replicates found: check col names
              in full experimental dataset")
    }
  }

  # add experimental data to table
  if (!is.null(obs_full)) {
    obs_plot <- obs_full
  } else if (!is.null(obs_mean)) {
    obs_plot <- obs_mean
  } else {
    obs_plot <- stop("No observation data are available!")
  }

  #prepare data frame for plot_ppc (pred, obs, min, max)
  rs_ppc <- data.frame(time = rs_mean$time, trial = rs_mean$trial)
  rs_ppc$pred <- rs_mean[, (ncol(rs_mean) - 1)]
  rs_ppc$obs <- obs_plot$data
  rs_ppc$min <- rs_range$min
  rs_ppc$max <- rs_range$max

  # prepare data frame for plot_ppc (pred, obs, min, max)
  rs_ppc <- data.frame(time = rs_mean[, 1], trial = rs_mean$trial)
  rs_ppc$pred <- rs_mean[, col_number]
  rs_ppc$obs <- obs_plot[, col_number] # observation data
  rs_ppc$min <- rs_range[, col_number] # lower boundaries
  rs_ppc$max <- rs_range[, col_number + 1] # upper boundaries

  # Add PPC column
  rs_ppc$PPC <- rs_ppc$obs >= rs_ppc$min & rs_ppc$obs <= rs_ppc$max
  rs_ppc$PPC <- as.factor(ifelse(rs_ppc$PPC, "blue", "orange"))

  # set study key if requested
  if (!is.null(study)) {
    rs_ppc$study <- as.factor(study)
  } # necessary when more than 1 study is available

  # create plot
  plot <- plot_ppc_combi(rs_ppc, xy_lim = xy_lim)

  print(plot)

  # return ggplot2 object
  return(plot)
}

#' Create PPC plot for one or more datasets
#'
#' `r lifecycle::badge("experimental")`
#'
#' The function expects a data.frame with five mandatory and one optional
#' column. The mandatory columns are as follows:
#' - `pred`: mean of predictions e.g. frond number for lemna
#' - `max`: maximum of predictions
#' - `min`: minimum of predictions
#' - `obs`: observations
#' - `PPC`: color code
#' The optional column is to be named `study` and contains a study identifier.
#' If more than one study identifier is present in the table, individual
#' studies will be plotted in different colors and a legend will be displayed.
#' The function is called by plot_ppc where the column names are defined
#' (see rs_ppc object).
#'
#' @param table `data.frame` containing return values of calls to `plot_ppc()`
#' @param xy_lim optional `numeric`, limits of x and y axis for plotting
#' @return a ggplot2 plot object
#' @autoglobal
plot_ppc_combi <- function(table, xy_lim = NULL) {
  if (!is.data.frame(table)) stop("table not a data.frame")

  if (!("obs" %in% names(table) &&
          "pred" %in% names(table) &&
          "min" %in% names(table) &&
          "max" %in% names(table) &&
          "PPC" %in% names(table)
      )) {
    stop("Required variables (obs, pred, min, max) are missing in the dataset.")
  }

  # try to find the max number in experiments
  if (is.null(xy_lim)) {
    xy_lim <- max(table[, c("pred", "obs")])
  }

  if (("study" %in% names(table))) {
    n_studies <- length(unique(table$study))
  } else {
    n_studies <- 1
    table$study <- "black"
  }

  # calculate PPC
  n <- nrow(table)
  n_fit <- nrow(dplyr::filter(table, obs <= max & obs >= min))
  ppc <- n_fit / n * 100

  # calculate NRMSE
  nrmse <- 1 / mean(table$obs) * sqrt(1 / n * sum((table$obs - table$pred)^2))

  # color scheme
  col_palette <- c(
    "#E69F00", "#56B4E9", "#009E73", "#F0E442",
    "#0072B2", "#D55E00", "#CC79A7"
  )
  if (n_studies == 1) {
    col_palette <- c("black")
  }
  if (n_studies > length(col_palette)) {
    stop("color palette has too few elements, please amend")
  }

  # create plot
  title <- paste0(
    "NRMSE: ",
    round(nrmse * 100, 1),
    "% & PPC: ",
    round(ppc, 1),
    "%"
  )

  plot <- ggplot2::ggplot(table) +
    ggplot2::geom_abline(linetype = "dashed") +
    ggplot2::geom_linerange(
      ggplot2::aes(obs,
        ymin = min,
        ymax = max,
        color = PPC
      ),
      alpha = 0.7,
      linewidth = 1.1
    ) +
    ggplot2::scale_color_identity()

  plot <- plot +
    ggplot2::geom_point(
      ggplot2::aes(obs,
        pred,
        color = study
      ),
      size = 1.5
    ) +
    ggplot2::coord_cartesian(xlim = c(0, xy_lim), ylim = c(0, xy_lim)) +
    ggplot2::labs(x = "Observed", y = "Predicted", title = title) +
    ggplot2::theme_bw()

  if (n_studies == 1) { # hide legend if only one study supplied
    plot <- plot + ggplot2::theme(legend.position = "none")
  }

  return(plot)
}

#' Plot EPx values
#'
#' `r lifecycle::badge("experimental")`
#'
#' @param EPx_ts the result of `epx_mtw`, ie. a tibble with window.start,
#' window.end, endpoint, level and EPx
#' @param exposure_ts an exposure time series with columns for time 'time' and
#' concentration 'conc'
#' @param draw Should the whole plot be drawn? If FALSE the exposure plot and
#' the EPx plot are returned as a list for later modification
#' @param time_col the name of the time column in the exposure dataset
#' @param conc_col the name of the concentration column in the exposure dataset
#' @param epx_x_title title of the x-axis of the epx panel
#' @param conc_y_title title of the y-axis of the concentration panel
#'
#' @return a grid of ggplots
#' @export
#' @autoglobal
#'
#' @examples
#' ti <- 0:21
#' expo <- abs(0.01*ti + rnorm(length(ti), 0, 0.05))
#' exposure <- data.frame(time = ti, conc = expo)
#' metsulfuron_epx_mtw <- metsulfuron %>%
#' set_exposure(exposure) %>%
#' epx_mtw(level = 10, factor_cutoff = 1000)
#' metsulfuron_epx_mtw
#' plot_epx(EPx_ts = metsulfuron_epx_mtw,
#' exposure_ts = exposure, conc_y_title = "env. concentration [µg/L]")
plot_epx <- function(EPx_ts, exposure_ts, draw = TRUE, time_col = "time", conc_col = "conc",
                     epx_x_title = "Start time", conc_y_title = "Exposure conc.") {

  plot_dat <- EPx_ts %>%
    dplyr::select(!c("window.end", "level"))

  # EPx plot ----
  ep_level <- unique(EPx_ts[["level"]])
  if (length(ep_level) > 1) {
    stop("only single EPx level accepted for plotting")
  }
  epx_y_title <- parse(text = paste0('EP[', ep_level, ']'))
  epx_plot_x_lim <- EPx_ts %>%
    dplyr::select(window.start, window.end) %>%
    range()

  len_ts <- plot_dat %>%
    dplyr::group_by(endpoint) %>%
    dplyr::summarise(len = length(window.start)) %>%
    dplyr::pull(len) %>%
    min()
  if (len_ts == 1){
    epx_plot_base <- ggplot2::ggplot(plot_dat) +
      ggplot2::geom_point(ggplot2::aes(window.start, EPx),
                          color = "orange")
  } else {
    epx_plot_base <- ggplot2::ggplot(plot_dat) +
      ggplot2::geom_line(ggplot2::aes(window.start, EPx),
                         color = "orange",
                         linewidth = 1.25)
  }
  epx_plot <- epx_plot_base +
    ggplot2::ylab(epx_y_title) +
    ggplot2::xlab(epx_x_title) +
    ggplot2::ylim(0, NA) +
    ggplot2::xlim(epx_plot_x_lim) +
    ggplot2::facet_wrap(endpoint ~ ., ncol = 1, scales="free_y")

  # Concentration plot ----
  conc_plot <- ggplot2::ggplot(exposure_ts) +
    ggplot2::geom_area(ggplot2::aes(.data[[time_col]], .data[[conc_col]]),
                       alpha = 0.5,
                       fill = "blue",
                       position = "identity") +
    ggplot2::xlim(epx_plot_x_lim) +
    ggplot2::ylab(conc_y_title) +
    ggplot2::theme(axis.title.x = ggplot2::element_blank(),
                   axis.text.x = ggplot2::element_blank(),
                   axis.ticks.x = ggplot2::element_blank())

  # Bind and align the two plots ----
  conc_plot_ <- ggplot2::ggplotGrob(conc_plot)
  epx_plot_ <- ggplot2::ggplotGrob(epx_plot)
  plots <- rbind(conc_plot_, epx_plot_, size = "last")
  plots$widths <- grid::unit.pmax(conc_plot_$widths, epx_plot_$widths)

  if (draw) {
    grid::grid.newpage()
    grid::grid.draw(plots)
  } else {
    return(list(conc_plot, epx_plot))
  }
}

#' Creates a prediction plot for one effect scenario
#'
#' `r lifecycle::badge("deprecated")`
#'
#' This function has been deprecated and replaced by the generic [plot()].
#'
#' Sometimes it is helpful if the user can plot results of one effect
#' scenario. This is for instance the case for test simulations or predictions
#' for one profile. This function runs the simulation for one effect scenario
#' and plots the results. Function plots the time (column 1) and the predictions
#' (column 2, can be changed by the user plot_col)
#'
#' @param model_base effect scenario object with mean parameters
#' @param plot_col output column which should be plotted, default = 2
#' @param trial_number name for model run (if available tag is used)
#' @return plot of the results for one `effect scenario`
#' @export
#'
#' @examples
#' # Please use `plot()` instead
#' metsulfuron %>%
#'   simulate() %>%
#'   plot()
plot_scenario <- function(model_base, plot_col = 2, trial_number = NULL) {
  lifecycle::deprecate_warn("1.3.1", "plot_scenario()", "plot()")

  # set name of run
  if (is.null(trial_number)) {
    if (is.na(model_base@tag)) {
      trial_number <- "T1"
    } else {
      trial_number <- model_base@tag
    }
  }

  # run model
  sim_result <- model_base %>% simulate()
  sim_result$trial <- trial_number

  # prepare exposure data
  exposure <- model_base@exposure@series
  exposure$trial <- trial_number

  # plot results
  plot_sd(model_base = model_base,
    treatments = exposure,
    rs_mean = sim_result,
    obs_mean = NULL,
    plot_col = plot_col
  )
}

# These tests are very basic for now. Either they need to be extended or
# the plotting functions to be removed/amended.

test_that("plot_sd", {
  rs <- minnow_it %>%
    simulate() %>%
    dplyr::mutate(trial="foo")

  plot_sd(
    model_base = minnow_it,
    treatments = minnow_it@exposure@series %>% dplyr::mutate(trial="foo"),
    rs_mean = rs
  )
  succeed()
})

test_that("plot_ppc", {

  rs <- minnow_it %>%
    simulate() %>%
    dplyr::mutate(trial="foo")
  rr <- rs %>%
    dplyr::select(1, Y=2, X=2, 3)

  plot_ppc(rs_mean=rs,
           rs_range=rr,
           obs_mean=rs
  )
  succeed()
})

test_that("plot_epx", {
  rs <- metsulfuron %>%
          epx_mtw(level = 10, factor_cutoff = 1000)
  plot_epx(rs, metsulfuron@exposure@series)
  succeed()
})

test_that("plot_scenario", {
  lifecycle::expect_deprecated({
    plot_scenario(minnow_it)
  })
  succeed()
})

# Creates a cache of data required for effect calculation
#
# Will store all relevant moving windows as well as effects from control
# runs, if needed. All data is stored in an environment object, because these
# are the only objects which are passed by reference by R. This should allow
# some efficiency gains w.r.t. to memory usage and processing time.
#
# @param x scenario object
# @param ... additional parameters passed through to [simulate()]
# @return `environment` object with two lists: `windows` and `controls`
cache_windows <- function(x, ...) {
  if(is.vector(x))
    stop("Argument `x` must be a scenario")

  env <- new.env(hash=FALSE, size=10)
  # determine moving windows
  env$windows <- window_candidates(x)
  env$controls <- list()

  if(!is_control_required(x)) {
    return(env)
  }

  # set exposure to zero
  y <- set_noexposure(x)

  # If the control scenario has no or constant forcing functions, then we only
  # need to simulate a single control, because the results for all windows will
  # be identical anyhow
  if(has_constant_forcings(y))
  {
    y <- y %>%
      clip_scenario(env$windows[[1]]) %>%
      set_window(length=-1)
    y@control.req <- FALSE
    efx <- effect(y, ep_only=TRUE, ...)
    env$controls <- lapply(env$windows, FUN=function(win) return(efx))
  }
  # simulate all windows individually
  else
  {
    # disable control calculation
    if(is_sequence(y))
      y[[1]]@control.req <- FALSE
    else
      y@control.req <- FALSE
    # derive effect
    df <- effect(y, max_only=FALSE, ep_only=TRUE, .cache=env, ...)
    nms <- names(df)
    env$controls <- lapply(seq(1, nrow(df)), function(i) setNames(df[i, ], nms))
  }

  env
}

#' Cache control simulations
#'
#' `r lifecycle::badge("deprecated")`
#'
#' Handling of cached control simulations has been modified and is solely
#' managed by package functions. `cache_controls()` is no longer needed and
#' will raise an error if called.
#'
#' @param x parameter not used
#' @param ... parameters not used
#' @return nothing
#' @export
cache_controls <- function(x, ...) {
  lifecycle::deprecate_stop("1.5.0", "cache_controls()", details="Caching of control scenarios is handled by package functions")
}

# Returns all possible windows, i.e. simulation periods to evaluate for effects
# @param scenario a [scenario] or [sequence]
window_candidates <- function(scenario) {
  times <- get_times(scenario)
  if(length(times) == 0)
    stop("Scenario has no output times")
  t_start <- times[[1]]
  t_end <- times[[length(times)]]

  # if no moving windows defined, then use whole period as single window
  win <- get_window(scenario)
  if(win$length <= 0)
    return(list(c("window.start"=t_start, "window.end"=t_end)))
  if(win$interval <= 0)
    stop("Scenario's window.interval has invalid value")

  # create list of candidate windows
  win.starts <- seq(t_start, max(t_end - win$length, t_start), win$interval)
  lapply(win.starts, function(ws) c("window.start"=ws, "window.end"=ws + win$length))
}

#' Start and stop logging
#'
#' @param file `character`, file name or path to a log file
#' @param append `logical`, if `TRUE` output will be appended to an existing log file,
#' otherwise the log file will be replaced
#' @param envir log will be automatically disabled if `environment` is exited,
#' set to `NULL` to disable
#' @return no return value
#' @export
log_enable <- function(file=NULL, append=TRUE, envir=parent.frame()) {
  if(is.character(file)) {
    #if(!is.null(envir))
    #  withr::defer(log_disable(), envir=envir)
    file = file(file, open=ifelse(append,"at","wt"))
    sink(file=file, type="output", append=append)
    options(cvasi.log.file=file)
  }

  options(cvasi.log=TRUE)
  log_msg("++ Log enabled: ", Sys.time())
  if(is.character(file)) {
    log_msg("++ Log file:    ", file)
  }
}

#' @export
#' @rdname log_enable
log_disable <- function() {
  log_msg("++ Log disabled: ",Sys.time())
  # close file connection and sink
  f <- getOption("cvasi.log.file",default=NULL)
  if(!is.null(f)) {
    sink(type="output")
    suppressWarnings(sink())
    try(close(f), silent=TRUE)
    options(cvasi.log.file=NULL)
  }

  options(cvasi.log=FALSE)
}

#' Add a log message
#'
#' Message will only appear in the console or in log file if logging was
#' enabled using `log_enable()`.
#'
#' @param ... elements will be concatenated using `paste0()`
#' @return no return value
#' @export
#' @examples
#' log_msg("this message will not appear")
#'
#' log_enable()
#' log_msg("this message will appear")
#' log_msg("a number of ","elements to ",42," concatenate")
log_msg <- function(...) {
  if(getOption("cvasi.log", default=FALSE)) {
    f <- getOption("cvasi.log.file",default=NULL)
    if(!is.null(f)) {
      write(paste0(...), stdout()) # write to log
    }
    message(paste0(...)) # write to console
  }
}

# Add a log message to console
#
# Output to log file is optional, output to console mandatory
log_msgf <- function(...) {
  if(getOption("cvasi.log", default=FALSE)) {
    f <- getOption("cvasi.log.file",default=NULL)
    if(!is.null(f)) {
      write(paste0(...), stdout()) # write to log
    }
  }
  message(paste0(...)) # write to console
}

#' Log R environment properties
#'
#' @return no return value
#' @export
#' @importFrom utils sessionInfo
log_envir <- function() {
  si <- sessionInfo()

  # imitates the behavior of print.sessionInfo() but looks nicer
  log_msg("\n++ Session info")
  log_msg(si$R.version$version.string)
  log_msg("Platform: ", si$platform)
  log_msg("Running under: ", si$running)

  log_msg("\nMatrix products: ", si$matprod)
  log_msg("Locale: ", gsub(";", ", ", si$locale))

  log_msg("\nAttached base packages:")
  log_msg(paste(si$basePkgs, sep="", collapse=", "))
  log_msg("\nOther attached packages:")
  log_msg(format_table(data.frame(name=sapply(si$otherPkgs, `[[`, "Package"), version=sapply(si$otherPkgs, `[[`, "Version"))))
  log_msg("\nLoaded via namespace (not attached):")
  log_msg(format_table(data.frame(name=sapply(si$loadedOnly, `[[`, "Package"), version=sapply(si$loadedOnly, `[[`, "Version"))))
}

# Logs the contents of a table
#
# Values are padded with white-spaces to the longest string in each column,
# the first column is left aligned, the remaining right-aligned, the table
# is sorted by the first column
#
#' @global .
format_table <- function(table) {
  table %>%
    dplyr::arrange(.[1]) %>%
    dplyr::mutate(dplyr::across(1, .fns=~ stringr::str_pad(.x, max(nchar(.x)), side="right", pad=" "))) %>%
    dplyr::mutate(dplyr::across(-1, .fns=~ stringr::str_pad(.x, max(nchar(.x)), side="left", pad=" "))) %>%
    tidyr::unite(col="V1", tidyr::everything(), sep=" ") %>%
    dplyr::pull(1) %>%
    paste(collapse="\n")
}

#' Log scenario properties
#'
#' @param x vector of `EffectScenario` objects
#' @param header `logical`, if `TRUE` a header line will be printed
#' @return unmodified argument `x`
#' @export
log_scenarios <- function(x, header=TRUE) {
  if(header)
    log_msg("\n++ Scenario overview")
  if(is.vector(x))
    sapply(x, log_scenarios, header=FALSE)
  else {
    show_scenario(x, inline=TRUE, show_exposure_series=FALSE)
    log_msg("")
  }
  x
}

#' Set scenario parameters
#'
#' Modifies the parameters of one or more [scenario] objects.
#'
#' Most parameters are represented by numerical types but other types
#' are possible depending on model. Please refer to the model description
#' which parameters are required and in which unit. Some or all parameters may
#' be required to start a simulation. If required parameters are missing,
#' simulation will fail with an error message.
#'
#' @param x object(s) to modify
#' @param param named numeric `vector` with parameter names and value OR a
#'   list of \linkS4class{parameter_set} objects
#' @return Vector of modified objects
#' @export
#' @include class-EffectScenario.R class-ParameterSet.R
#' @examples
#' Lemna_Schmitt() %>% set_param(c(Emax=1,EC50=0.12))
setGeneric("set_param", function(x, param) standardGeneric("set_param"))

#' @rdname set_param
#' @export
setMethod("set_param", c("EffectScenario","vector"), function(x, param) {
  # Supported cases: 'param' is either
  # a) a named numerical vector of key/value pairs
  # b) a vector/list of parameter_set objects
  if(is.atomic(param) | all(sapply(param, is.atomic))) {
    # if required parameters are set, check if arguments conform
    if(length(x@param.req) > 0) {
      unused <- setdiff(names(param),x@param.req)
      if(length(unused)>0)
        warning(paste("ignoring parameters",paste(unused,collapse=",")))

      ints <- intersect(x@param.req,names(param))
      x@param[ints] <- param[ints]
    } else { # otherwise, just use values as they are
      x@param[names(param)] <- param
    }
  }
  else if(all(is_parameter_set(param))) {
    x <- set_param_multi(list(x), param)[[1]]
  }
  else {
    stop("list of parameters contains inconsistent or invalid types")
  }
  x
})

# single scenarios, single parameter_set
#' @rdname set_param
#' @export
setMethod("set_param", c("EffectScenario","ParameterSet"), function(x, param) {
  if(is_param_match(x, param)) {
    x <- set_param(x, param@param)
    return(x)
  }
  stop("model name & tag do not match parameter set")
})

# multiple scenarios, single parameter_set
#' @rdname set_param
#' @export
setMethod("set_param", c("list","ParameterSet"), function(x, param) {
  sapply(x, set_param, param)
})

# multiple scenarios, EITHER single parameter set OR list of parameter_set objects
#' @rdname set_param
#' @export
setMethod("set_param", c("list","vector"), function(x, param) {
  # match scenarios to parameter sets?
  if(all(is_scenario(x)) & all(is_parameter_set(param)))
    return(set_param_multi(x, param))
  # default
  sapply(x, set_param, param)
})

# for a scenario sequence, apply parameters to each scenario in sequence
#' @rdname set_param
#' @export
setMethod("set_param", c("ScenarioSequence","ANY"), function(x, param) {
  x@scenarios <- lapply(x@scenarios, set_param, param)
  x
})

# for calisets
#' @rdname set_param
#' @export
setMethod("set_param", c("CalibrationSet", "ANY"), function(x, param) {
  x@scenario <- set_param(x@scenario, param)
  x
})

#' Match scenarios to parameter sets 1:1
#' @param scenarios `list` of `EffectScenario` objects
#' @param param `list` of `parameter_set` objects
#' @noRd
#' @global tag model name desc
set_param_multi <- function(scenarios, param) {
  df <- tibble::tibble(ps=param, model=get_model(param), tag=get_tag(param)) %>%
    dplyr::mutate(tag=ifelse(is.na(tag), "NA", tag))
  rs <- data.frame(model=character(0), tag=character(0), n=numeric(0))

  for(i in seq_along(scenarios)) {
    sc <- scenarios[[i]]
    # check if there's a unique parameter set available
    md <- get_model(sc)
    tg <- ifelse(is.na(get_tag(sc)), "NA", get_tag(sc))
    matches <- dplyr::filter(df, model==md & tag==tg)
    rs <- dplyr::bind_rows(rs, list(model=md, tag=tg, n=nrow(matches)))

    # only update scenario's parameters if it's a unique match
    if(nrow(matches) == 1)
      scenarios[[i]] <- set_param(sc, matches$ps[[1]])
  }
  # generate some overview of what happened
  log_msgf("+ Matching scenarios to multiple parameter sets")
  rs %>%
    dplyr::mutate(name=paste0(paste(model, tag, sep="#"), ": "),
                  desc=dplyr::case_when(n==0 ~ "no matches",
                                        n==1 ~ "unique match",
                                        n>1 ~ "ambiguous")) %>%
    dplyr::select(name, desc) %>%
    format_table() %>%
    log_msgf()
  if(any(rs$n > 1))
    stop("match of scenarios to parameter sets must be unambiguous")
  if(any(rs$n == 0))
    warning("could not match all scenarios to parameter sets")

  scenarios
}

#' @export
setClass("ExposureSeries",
         slots=list(
           series="data.frame",
           dates="ANY",
           meta="list",
           context="list",
           file="character"
         )
)

#' @export
setClass("NoExposureSeries", contains="ExposureSeries")



#' Exposure time-series
#'
#' Creates an object that encapsulates an exposure time-series with its
#' metadata, such as formatted datetime strings and file name where the
#' series was loaded from. [no_exposure()] is shorthand to create a time-series
#' of constant zero exposure.
#'
#' @slot dates original time points of time-series, e.g. time stamps of the form `2000-01-01 12:00`
#' @slot file `character`, file name where data originates from, may be empty
#' @slot meta `list`, contains metadata
#' @slot context `list`, contains contextual metadata, such as project ids
#' @slot series `data.frame` containing the actual time-series
#' @param series `data.frame` with two columns containing a time-series
#' @param dates `vector`, optional original list of time stamps
#' @param file `character`, optional file name where data originates from
#' @param meta `list`, optional metadata
#' @param context `list` optional contextual metadata such as project ids
#' @return an S4 object of type [ExposureSeries-class]
#'
#' @seealso [no_exposure()]
#' @aliases ExposureSeries-class NoExposureSeries-class
#' @export
ExposureSeries <- function(series, dates, file, meta, context) {
  check_exposure(series)
  # coerce to data.frame to avoid issues with data.frame-like types
  series <- dplyr::select(as.data.frame(series), c(1, 2))

  if(missing(file))
    file <- "unknown"
  if(missing(meta))
    meta <- list()
  if(missing(context))
    context <- list()
  if(missing(dates))
    dates <- series[,1]

  new("ExposureSeries",
      series=series,
      dates=dates,
      file=file,
      meta=meta,
      context=context
  )
}


#' Zero exposure
#'
#' Creates an [ExposureSeries] with zero concentration. When setting
#' the zero exposure, pay attention not to accidentally reset the output times
#' of your scenario as the zero exposure series contains only a single time point.
#' See the examples.
#'
#' @return  an S4 object of type [ExposureSeries-class]
#' @seealso [set_noexposure()]
#' @export
#' @examples
#' # Set exposure to zero, but keep the original output times
#' minnow_it %>%
#'   set_noexposure() %>%
#'   simulate()
no_exposure <- function() {
  new("NoExposureSeries", series=data.frame(time=0, conc=0), file="no exposure")
}

library(readxl)
library(dplyr)
library(ggplot2)

# Imports exposure series used for model validation in EFSA (2018), pp. 65,
# and saves them as an RDS object to be available for unit tests.
#
# EFSA Scientific Opinion on TKTD models, 2018. doi:10.2903/j.efsa.2018.5377

dir <- "tests/data/EFSA_propiconazole"

focus <- list()
# Apple scenarios
focus$R1pond <- read_xlsx(file.path(dir, "Propiconazole apple simulations.xlsx"), sheet="R1pond") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$R2stream <- read_xlsx(file.path(dir, "Propiconazole apple simulations.xlsx"), sheet="R2stream") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)

# Cereal scenarios
focus$D1ditch <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D1 ditch") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$D1stream <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D1 stream") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$D3ditch <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D3 dicth") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$D4pond <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D4 pond") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$D4stream <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D4 stream") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$D5pond <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D5 pond") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$D5stream <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="D5 stream") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)
focus$R4stream <- read_xlsx(file.path(dir, "Propiconazole spring cereals simulations.xlsx"), sheet="R4 stream") %>%
  mutate(days=as.numeric(difftime(Time, Time[[1]], units="days"))) %>%
  select(days, conc_ugL=3)

saveRDS(focus, file.path(dir, "focus.rds"))

rm(focus)

#' Dynamic Energy Budget (DEB) models
#'
#' Supported models:
#' * [DEB_abj]
#' * [DEBtox]
#'
#' @name DEB-models
#' @family DEB models
#' @family models
#' @aliases Deb-class
NULL

# Parent class for all DEB related models/scenarios
#' @export
setClass("Deb", contains="EffectScenario")

#' Simulate an effect scenario
#'
#' The supplied `EffectScenario` is passed on to the ODE solver for numerical
#' integration. Internally, `simulate()` is split up into several functions
#' dedicated to particular models, e.g. one for GUTS and one for Lemna type
#' models. The package will take care of using the correct function
#' for each model when `simulate()` is called.
#'
#' Simulation results are returned as a time-series for each state
#' variable. Some models provide additional properties describing the model state,
#' e.g. the internal concentration of a toxicant within the organism. Refer
#' to the respective [scenario] for more information.
#'
#' Additional arguments to `simulate()` will be passed on to [deSolve::ode()]
#' which enables control of the numerical integration parameters.
#'
#' ### Output times and windows
#' The minimum and maximum of given time points define the simulated
#' period. However, the simulation can also be limited to
#' a subset of time points by enabling a moving exposure window, see [set_window()].
#'
#' Results will be returned for each output time point. Precision of the numeric solver
#' may be affected by chosen output times in certain cases. Hence, small deviations
#' in results should be expected if different output times are set. This effect
#' can be mitigated by either defining are sufficiently small time step for the solver
#' using argument `hmax` or by decreasing the error tolerances `atol` and `rtol`.
#' These arguments are passed to the solver, see e.g. [deSolve::lsoda()] for details.
#'
#' ### Optional output variables
#' Some models support adding intermediary model variables to the return value
#' of `simulate()`. Analyzing the additional outputs may be helpful to understand
#' model behavior and support finding potential issues in model parameterization.
#'
#' Optional outputs are enabled by setting the parameter `nout` to a value greater
#' than zero. If `nout` is set to `n`, then the first `n` optional output columns
#' will be returned along the normal simulation result.
#'
#' Which optional outputs are available depends on the model/scenario at hand.
#' Please refer to the model documentation for details.
#' As an example, the [GUTS-RED-IT][GUTS_RED_IT()] model supports adding the
#' external toxicant concentration to the output by setting `nout=1`:
#'
#' `minnow_it %>% simulate(nout=1)`
#'
#' ### Numerical precision and stability
#' Each model was assigned a default ODE solver which handles most of the
#' occurring inputs well. In most cases, this will be an explicit numerical
#' scheme of the Runge-Kutta family with variable step width. For certain extreme
#' parameters settings, such as very high uptake/permeability of the contaminant
#' or exposure series which represent step functions, the numerical approximation
#' might deteriorate and yield invalid results. In this case try to decrease the
#' allowed max step width by setting the argument `hmax` with various values.
#' Start with `hmax=1` and decrease the value by orders of 10. It is not
#' possible or desirable to reduce `hmax` to extremely small values, as the
#' ODE solver will require more CPU time and simulation will become inefficient.
#'
#' Oftentimes, it will be computational more efficient to adapt the solver's
#' error tolerances `atol` and `rtol` than reducing the step width `hmax` to achieve
#' stable numerics. Start by decreasing deSolve's default values by
#' orders of ten until the simulation yields acceptable results, see e.g.
#' [deSolve::lsoda()] for more information on error tolerances.
#'
#' As an alternative to adapting solver parameters, it might be worthwhile to
#' try other numerical schemes which can handle stiff ODEs, such as Radau, LSODA,
#' or LSODES. To change solvers, set the `method` argument.
#' To select e.g. the Radau scheme, set `method="radau"`. For LSODA, set `method="lsoda"`.
#' Radau performs better than LSODA in some cases, as the latter method can return
#' biologically nonsensical results without raising an error. See [deSolve::ode()]
#' for details on available ODE solvers.
#'
#'
#' @param x [scenario] to simulate
# @param times optional `numeric` vector of time points for which results are returned,
#   overrides settings of the scenario
#' @param ... additional parameters passed on to ODE solver
#'
#' @return A `data.frame` with the time-series of simulation results
#' @export
#' @examples
#' # base R syntax
#' simulate(minnow_sd)
#' # tidy syntax with the same result
#' minnow_sd %>% simulate()
#'
#' # Extend the simulated time frame to the interval [0, 10]
#' minnow_sd %>%
#'   set_times(seq(0, 10)) %>%
#'   simulate()
#'
#' # Use an alternative exposure profile, but keep the original output times
#' minnow_sd %>%
#'   set_exposure(data.frame(t=0, c=10), reset_times=FALSE) %>%
#'   simulate()
#'
#' ##
#' ## Precision of results
#'
#' # A large number of output times forces smaller solver time steps
#' minnow_it %>%
#'   set_times(seq(0, 10, 0.001)) %>%
#'   simulate() %>%
#'   tail()
#'
#' # Defining only two output times allows the ODE solver to make larger steps
#' # in time during numerical integration. However, results can become
#' # imprecise.
#' minnow_long <- minnow_it %>% set_times(c(0, 10))
#' minnow_long %>% simulate()
#'
#' # Numerical precision of results can be increased by limiting the solver's
#' # maximum step length in time using argument `hmax`.
#' minnow_long %>% simulate(hmax=0.005)
#'
#' # A similar numerical precision can be achieved by switching to an alternative
#' # numerical integration scheme, such as the Radau scheme, without limiting
#' # the step length.
#' minnow_long %>% simulate(method="radau")
#'
#' # Reducing the step length even further may increase numerical precision, but
#' # may exceed the solver's allowed number of integration steps per output interval.
#' # The following simulation will be aborted with a solver error:
#' try(
#'   minnow_long %>% simulate(hmax=0.001)
#' )
#'
#' # However, the solver's maximum number of allowed steps can be increased,
#' # if needed, using the argument `maxsteps`:
#' minnow_long %>% simulate(hmax=0.001, maxsteps=10^5)
setGeneric("simulate", function(x, ...) standardGeneric("simulate"), signature="x")

#' Default for all models using moving exposure windows
#' @rdname simulate
#' @include class-EffectScenario.R
setMethod("simulate", "EffectScenario", function(x, ...) simulate_scenario(x, ...))

#' @rdname simulate
#' @include class-Transferable.R
setMethod("simulate", "Transferable", function(x, ...) simulate_transfer(scenario=x, ...))

#' @rdname simulate
#' @include sequence.R
setMethod("simulate", "ScenarioSequence", function(x, ...) simulate_seq(seq=x, ...))

#' @rdname simulate
#' @include batch.R
setMethod("simulate", "SimulationBatch", function(x, ...) simulate_batch2(batch=x, ...))


# Wrapper for solver function to enforce setting of a S3 class for all simulation
# results.
#' @importFrom utils capture.output
simulate_scenario <- function(x, times, .suppress=FALSE, ...) {
  # set times vector, mostly for backwards compatibility
  if(!missing(times)) {
    x <- set_times(x, times)
  }
  if(length(x@times) < 2) {
    stop("Scenario must have at least two output time points for simulation.")
  }

  # listof conditions raised by deSolve
  has_error <- FALSE
  has_warning <- FALSE
  conds <- list()
  out <- capture.output(
           withCallingHandlers(
             rs <- withRestarts(solver(scenario=x, ...), muffleStop=function(x) data.frame() ),
             error=function(x) {
               conds <<- append(conds, list(c("error", x$call, x$message)))
               has_error <<- TRUE
               invokeRestart("muffleStop")
             },
             warning=function(x) {
               conds <<- append(conds, list(c("warning", x$call, x$message)))
               has_warning <<- TRUE
               tryInvokeRestart("muffleWarning")
             })
           )
  # convert deSolve return type to an actual data.frame
  if(is(rs, "deSolve")) {
    rs <- ode2df(rs)
  }
  class(rs) <- c("cvasi_simulate", "data.frame")
  was_aborted <- num_aborted(rs)
  attr(rs, "cvasi_status") <- "success"

  # suppress any conditions, this simplifies the calling code and also
  # drastically reduces the time spent in this function
  if(.suppress)
  {
    if(was_aborted)
      attr(rs, "cvasi_status") <- "aborted"
    else if(has_error)
      attr(rs, "cvasi_status") <- "error"
    else if(has_warning)
      attr(rs, "cvasi_status") <- "warning"
  }
  else
  {
    # save any output messages from deSolve, i.e. error output
    attr(rs, "desolve_output") <- out
    attr(rs, "desolve_conds") <- conds
    # set custom class for usability features such as plotting

    attr(rs, "cvasi_status") <- "success"
    if(was_aborted) {
      attr(rs, "cvasi_status") <- "aborted"
      warn1 <- which.min(sapply(conds, function(lst) lst[[1]] == "warning"))
      if(length(warn1) == 0) {
        conds <- list(list("", "", "An unknown error occurred."))
        warn1 <- 1
      }
      warn(c("Simulation aborted.", conds[[warn1]][[3]]),
           footer = "Please run `num_info()` on result to get help on numerical issues.")
    }
    else if(has_error) {
      # find all error messages
      err <- sapply(conds, function(lst) ifelse(lst[[1]] == "error", unlist(tail(lst, n=1)), NA_character_))
      abort(c("Simulation failed", err[!is.na(err)]), class="cvasi_error")
    }
    else if(has_warning) {
      attr(rs, "cvasi_status") <- "warning"
      warn1 <- which.min(sapply(conds, function(lst) lst[[1]] == "warning"))
      warn(c("Issues during simulation.", conds[[warn1]][[3]]),
           footer = "Please run `num_info()` on result to get help on numerical issues.")
    }
  }
  rs
}

# Simulate a scenario where (biomass) quantities are transferred and reset
# at certain time points
#
# The base scenario is split at each transfer time point and each period
# is simulated individually. After each transfer, biomass and other
# state variables are modified to a new initial state where simulation
# continues in the next period.
#
# @param scenario EffectScenario to simulate
# @param times optional numerical vector overriding the scenario's output times
# @param .in_sequence optional logical, set to TRUE if simulated scenario is part
#   of a scenario sequence, then it will return a proposed initial state for
#   simulate_seq to use
# @param ... additional parameters passend on to [solver()]
# @return data.frame
simulate_transfer <- function(scenario, times, .in_sequence=FALSE, ...) {
  if(!has_transfer(scenario)) # shortcut if no transfers were defined
    return(simulate_scenario(scenario, times=times, ...))
  if(!missing(times))
    scenario <- set_times(scenario, times)

  times <- scenario@times
  t_min <- min(times)
  t_max <- max(times)

  # find transfer time points that occur during the simulated period
  tr_points <- vector("numeric")
  ends_on_transfer <- FALSE
  if(has_regular_transfer(scenario)) {
    seq_start <- ceiling(t_min / scenario@transfer.interval) * scenario@transfer.interval
    seq_end <- t_max
    if(seq_start <= seq_end) # avoid errors in case no transfer occurs
      tr_points <- seq(seq_start, seq_end, scenario@transfer.interval)
  } else {
    tr_points <- scenario@transfer.times
  }
  # does simulation period end on a transfer?
  if(length(tr_points) > 0)
    ends_on_transfer <- tr_points[length(tr_points)] == t_max
  # limit vector to transfer time points which occur during the simulated period.
  # but exclude the starting time as a potential transfer
  tr_points <- tr_points[tr_points > t_min & tr_points <= t_max]

  # if no transfer occurs during simulated period -> early exit
  if(length(tr_points) == 0 & !ends_on_transfer)
    return(simulate_scenario(set_times(scenario, times), ...))

  # add all transfer time points to the output times vector
  tr_inserted <- setdiff(tr_points, times)
  if(length(tr_inserted) > 0)
    times <- sort(c(times, tr_inserted))

  # biomass level after each transfer
  if(length(scenario@transfer.biomass) == 1)
    tr_biomass <- rep(scenario@transfer.biomass, times=length(tr_points))
  else if(length(scenario@transfer.biomass) == length(tr_points))
    tr_biomass <- scenario@transfer.biomass
  else
    stop("length of biomass and transfer times vectors do not match", call.=FALSE)

  df <- data.frame()
  t_start <- t_min
  # add last time point to also simulate the remainder of the scenario, if it exists
  if(!ends_on_transfer) {
    tr_points <- c(tr_points, t_max)
    # repeat the last value, although it will never be used
    tr_biomass <- c(tr_biomass, tail(tr_biomass, n=1))
  }
  # simulate population until next transfer to new medium
  for(i in seq_along(tr_points))
  {
    tr <- tr_points[i]
    period <- times[times >= t_start & times <= tr]
    out <- simulate_scenario(set_times(scenario, period), ...)

    # select rows/time points:
    # 1) that do NOT occur directly after a transfer, i.e. the start of a simulation
    selector <- i == 1 | period != t_start
    # 2) that were NOT inserted for technical reasons
    if(length(tr_inserted) > 0) {
      selector <- selector & !(period %in% tr_inserted)
    }
    df <- rbind(df, out[selector, ])

    # Transfer a fixed number of biomass to a new medium:
    # use last state as starting point
    init <- unlist(tail(out, 1)[names(scenario@init)])
    # scale internal toxicant mass in proportion to new biomass
    BM.fac <- tr_biomass[i] / init[[scenario@transfer.comp.biomass]]
    # reset biomass
    init[[scenario@transfer.comp.biomass]] <- tr_biomass[i]
    # scale other compartments relative to new biomass
    init[scenario@transfer.comp.scaled] <- init[scenario@transfer.comp.scaled] * BM.fac
    # in case of e.g. numerical issues, NaN can occur in sim output and new initial states
    # TODO check if error due to e.g. numerics, to provide a targeted messages
    if(any(is.nan(init))) {
      warning("Simulation result contains NaN values, please check parameters and numerical stability.", call.=FALSE)
      break
    }
    scenario <- set_init(scenario, init)
    t_start <- tr
  }

  # add metadata at which system state the subsequent simulation should start
  # if they were simulated in a sequence, cf. simulate_seq()
  if(.in_sequence) {
    if(ends_on_transfer)
      attr(df, "next_init") <- init
    else
      attr(df, "next_init") <- tail(df, 1)[names(scenario@init)]
  }
  rownames(df) <- NULL
  class(df) <- c("cvasi_simulate", "data.frame")
  df
}

# Simulate a sequence of scenarios
#
# The sequence of scenarios is simulated end to end, i.e. the next scenario
# starts where the last one ended. Start and end time points must match
# exactly, state variables of all scenarios must be identical
#
# @return data.frame
simulate_seq <- function(seq, times, ...) {
  if(length(seq) == 0)
    stop("No scenarios in sequence")
  if(!missing(times)) {
    seq <- set_times(seq, times)
  }
  # for backwards compatibility, include all breaks in output times
  # since v1.5.0
  if(methods::.hasSlot(seq, "inc_start")) {
    inc_start <- seq@inc_start
    inc_end <- seq@inc_end
  } else {
    inc_start <- c(TRUE, rep(FALSE, length(seq) - 1))
    inc_end <- rep(TRUE, length(seq))
  }

  df <- data.frame() # result table
  init <- seq[[1]]@init
  t_start <- NA_real_
  nms <- names(init)

  for(i in seq_along(seq))
  {
    # skip scenario if no output times defined
    if(length(seq[[i]]@times) == 0) {
      next
    }
    # get the first output time from the sequence
    if(is.na(t_start)) {
      t_start <- seq[[i]]@times[1]
    }

    # get current scenario and re-set init state
    sc <- set_init(seq[[i]], init)
    # check if current scenario starts where the previous ended
    if(sc@times[1] != t_start)
      stop("Scenarios are not on a continuous time scale")
    # we simulate each scenario as is
    out <- simulate(sc, .in_sequence=TRUE, ...)

    # which system state to use as initial state for next scenario
    # in sequence?
    # a) if simulate() returned suitable metadata then use that
    # b) otherwise, use last system state of last scenario
    init <- attr(out, "next_init", exact=TRUE)
    if(is.null(init)) {
      init <- tail(out, 1)[nms]
    }

    # filter certain output times from result if they are not supposed to be reported
    if(!inc_start[[i]]) {
      out <- out[sc@times != t_start, ]
    }
    if(!inc_end[[i]]) {
      t_end <- sc@times[length(sc@times)]
      out <- out[out[, 1] != t_end, ]
    }

    t_start <- tail(sc@times, 1)
    # reported result
    df <- rbind(df, out)
  }
  attr(df, "next_init") <- NULL
  rownames(df) <- NULL
  class(df) <- c("cvasi_simulate", "data.frame")
  df
}

# Simulate a batch of scenarios
#
# A batch consist of a base scenario and number of exposure series that are
# simulated with said base scenario. This is intended to replicate the setup
# of ecotox effect studies.
#
# This function is not intended to be invoked by the user, but the user is
# expected to make a call such as `simulate(batch(myscenario, list(1, 2, 3)))`.
# Parameters that influence the format of the return value are documented
# by [batch()].
simulate_batch2 <- function(batch, times, ...) {
  # some basic checks
  if(!is.character(batch@id_col))
    stop("id column is missing")
  if(batch@format != "long" & batch@format != "wide")
    stop("invalid format selected")
  if(!missing(times))
    batch@scenarios <- lapply(batch@scenarios, set_times, times)

  # TODO parallelization strategy should be controllable by user
  df <- furrr::future_map2_dfr(
    batch@scenarios,
    names(batch@scenarios),
    function(sc, nm, id_col) {
      rs <- simulate(sc, ...)
      rs[[id_col]] <- nm
      rs
    },
    id_col=batch@id_col
  )

  # if the user specified one or more columns to select, then we will select
  # a subset from the simulation results, but always time (1st colum), and
  # the (trial) ids as well
  if(!is.null(batch@select)) {
    # magic value 1: we assume that the first column contains time
    df <- dplyr::select(df, 1, !c(1, batch@id_col) & batch@select, batch@id_col)
  }
  # pivot to wide format?
  if(batch@format == "wide") {
    df <- tidyr::pivot_wider(df, id_cols=1, names_from=batch@id_col, values_from=!c(1, batch@id_col))
  }

  df
}

#' Batch simulation using multiple exposure series
#'
#' `r lifecycle::badge("deprecated")`
#'
#' A convenience function to simulate a single base scenario with one or more
#' exposure series. This aims at reproducing the setup and results of common
#' effect studies.
#'
#' A scenario contains only one exposure series. However, laboratory experiments
#' commonly examine the effects of multiple exposure levels on a biological system.
#' A batch simulation approach would involve running multiple simulations with
#' varying exposure or treatment conditions. To illustrate, if the objective is
#' to examine the impact of a substance on cell growth, the simulation model
#' could be designed to replicate the cell growth dynamics under varying
#' concentrations of the substance. Each simulation run would represent a
#' specific exposure level, ranging from low to high concentrations of the
#' chemical. To simulate such a laboratory experiment, the simulate_batch
#' function can be used. All exposure series are saved in the treatment argument.
#' The first column contains the time, the second column the concentration, and
#' the third column the trial name (exposure level, e.g. 'T1', 'T2', 'T3').
#'
#' @param model_base effect scenario object with mean parameters
#' @param treatments treatments exposure levels as data frame (time, conc, trial)
#' @param param_sample *deprecated* parameter, no longer in use
#' @param ... additional parameters passed through to [simulate()]
#' @return a `data.frame` with simulation results
#' @export
#' @examples
#' t1 <- data.frame(time=0:10, conc=0, trial="control")  # 1st treatment level
#' t2 <- data.frame(time=0:10, conc=1, trial="T1")       # 2nd treatment level
#' treatments <- rbind(t1, t2)
#'
#' metsulfuron %>%
#'   simulate_batch(treatments)
simulate_batch <- function(model_base, treatments, param_sample=deprecated(), ...) {
  lifecycle::deprecate_soft("1.5.0", "simulate_batch()", details="Please use `simulate(batch())` instead")
  if(is_present(param_sample)) {
    if(!is.null(param_sample)) {
      lifecycle::deprecate_stop("1.3.0", "simulate_batch(param_sample)")
    }
  }
  # Check if columns 'time' and 'conc' exist
  if(!("time" %in% colnames(treatments)) || !("conc" %in% colnames(treatments))) {
    stop("Columns 'time' and/or 'conc' not found in treatments dataframe.")
  }
  exp_levels <- unique(treatments[,3])
  treatments[,3] <- factor(treatments[,3], levels = exp_levels)
  # simulate scenarios with a defined step width in time
  t_max <- max(treatments[, 1]) # 1st column should contain time
  sim_times <- seq(0, t_max, 0.1) # magic number: 0.1
  # simulate base scenario (mean line)
  # Initialize an empty list to store data frames
  list_of_effect_sets <- list()
  # create list of effect scenario objects for each trial
  list_of_effect_sets <- lapply(split(treatments, treatments$trial), function(x){
    model_base %>%
      set_exposure(data.frame(time = x[,1], conc = x[,2]))
  })
                                                # time, conc

  # simulate
  df <- data.frame()
  df_one_run <- data.frame()
  simulated_results <- purrr::map(list_of_effect_sets, ~ .x %>% simulate(), ...)
  # add trial name
  simulated_results <- purrr::map2(
    simulated_results,
    exp_levels,
    ~ dplyr::mutate(.x, trial = .y)
  )
  df <- dplyr::bind_rows(simulated_results)
  class(df) <- c("cvasi_simulate", "data.frame")
  return(df)
}

test_that("arg times", {
  sc <- minnow_it
  orig_times <- sc@times
  new_times <- orig_times + 1

  rs <- simulate(sc)
  expect_equal(rs$time, orig_times)

  rs <- simulate(sc, times=new_times)
  expect_equal(rs$time, new_times)

  # invalid
  sc@times <- 1
  expect_error(simulate(sc), "at least two output time")
})

#TODO in simulate_scenario, it appears if the conditions array may sometimes have an unknown format?
# ─cvasi:::simulate_scenario(scenario, times = times, ...) at cvasi/R/simulate.R:244:3
# 7.       └─rlang::abort(c("Simulation failed", err[!is.na(err)]), class = "cvasi_error") at cvasi/R/simulate.R:216:7

test_that("simulate_batch", {
  t1 <- data.frame(time=0:10, conc=0, trial="control")
  t2 <- data.frame(time=0:10, conc=1, trial="T1")
  treatments <- rbind(t1, t2)

  rs1 <- metsulfuron %>% set_exposure(t1[,c(1,2)]) %>% simulate()
  rs2 <- metsulfuron %>% set_exposure(t2[,c(1,2)]) %>% simulate()
  lifecycle::expect_deprecated({
    rsb <- metsulfuron %>% simulate_batch(treatments)
  })

  expect_equal(dplyr::select(rsb, !trial), rbind(rs1, rs2), ignore_attr=T, tolerance=1e-5)
  expect_equal(rsb$trial, rep(c("control","T1"), each=nrow(rs1)))
})

test_that("deprecated parameter", {
  expect_error({
    lifecycle::expect_deprecated({
      simulate_batch(metsulfuron, data.frame(), param_sample=23)
    })
  })
})

########################
## Organizing man pages
########################

# nada

########################
## Class definitions
########################

#' Magma scenario class
#'
#' This entry documents the class definition used for [Magma]-type scenarios.
#' For details regarding the model, please refer to the [Magma model][Magma()] manual.
#'
#' @include class-Transferable.R
#' @slot growth_model *character*, selects the growth model, such as *exponential*
#'    or *logistic* growth.
#' @export
#' @seealso [Macrophyte-models]
setClass("Magma",
         contains=c("Transferable", "EffectScenario"),
         slots=c("growth_model"="character")
         )

# for backwards compatibility only
setClass("Myriophyllum", contains="Magma")
setClass("MyrioExp", contains="Myriophyllum")
setClass("MyrioExpScenario", contains="MyrioExp")
setClass("MyrioLog", contains="Myriophyllum")
setClass("MyrioLogScenario", contains="MyrioLog")


########################
## Constructor
########################

#' Magma model (Witt et al., submitted)
#'
#' The *Magma* model interprets the Tier 2C version of the *Lemna* model by
#' [Klein et al. (2021)][Lemna_SETAC()], as a generic macrophyte model.
#' It is mathematically equivalent to the Tier 2C version of the model
#' by Klein et al. (2021) with the recommended Tier 2C settings `k_photo_fixed=TRUE`
#' and `k_resp=0`.
#'
#' In particular, the growth model is a simple exponential growth model,
#' which is considered to be the typical situation for a laboratory macrophyte
#' study. Instead of frond numbers as for Lemna, the biomass is also returned as
#' total shoot length (*TSL*) in simulation results. Consequently, the model has
#' the additional parameter `r_DW_TSL` (dry weight per total shoot length ratio)
#' instead of `r_DW_FN` (dry weight per frond number ratio). A model variant
#' with an option for logistic growth is provided as well.
#'
#' @section State variables:
#' The model has two state variables:
#' - `BM`, Biomass (g dw)
#' - `M_int`, Mass of toxicant in plant population (ng)
#'
#' @section Model parameters:
#' The growth model can either simulate exponential growth (the default) or
#' logistic growth. For logistic growth, an additional parameter `D_L` describing
#' the limit density or carrying capacity needs to be provided.
#'
#' - Growth model
#'   - `mu_control`, Maximum photosynthesis rate (d-1), default: `0.47`
#'   - (optional) `D_L`, Limit density (g dw)
#'
#' - Concentration response (Toxicodynamics)
#'   - `EC50_int`, Internal concentration resulting in 50% effect (ug L-1)
#'   - `E_max`, Maximum inhibition (-), default: `1`
#'   - `b`, Slope parameter (-)
#'
#' - Internal concentration (Toxicokinetics)
#'   - `P`, Permeability (cm d-1)
#'   - `r_A_DW`, Area per dry-weight ratio (cm2 g-1), default: `1000`
#'   - `r_FW_DW`, Fresh weight per dry weight ratio (-), default: `16.7`
#'   - `r_FW_V`, Fresh weight density (g cm-3), default: `1`
#'   - `r_DW_TSL`, Dry weight per total shoot length ratio  (g dw cm-1)
#'   - `K_pw`, Partitioning coefficient plant:water (-), default: `1`
#'   - `k_met`, Metabolisation rate (d-1), default: `0`
#'
#' @section Environmental factors:
#'
#' None.
#'
#' @section Parameter boundaries:
#' Default values for parameter boundaries are set for all parameters by expert
#' judgement, for calibration purposes. Values can be modified using [set_bounds()].
#'
#' @section Simulation output:
#' Simulation results will contain the state variables biomass (`BM`) and
#' mass of internal toxicant (`M_int`).
#'
#' It is possible to amend the output of [simulate()] with additional model
#' quantities that are not state variables, for e.g. debugging purposes or to
#' analyze model behavior. To enable or disable additional outputs, use the
#' optional argument `nout` of [simulate()]. As an example, set `nout=2` to
#' enable reporting of total shoot length (`TSL`) and internal concentration
#' (`C_int`). Set `nout=0` to disable additional outputs. The default is `nout=1`.
#'
#' The available output levels are as follows:
#' - `nout` >= 1: `TSL`, total shoot length (cm)
#' - `nout` >= 2: `C_int`, internal concentration (ug L-1)
#' - `nout` >= 3: `f_photo`, photosynthesis dependency function (-)
#' - `nout` >= 4: `C_int_unb`, unbound internal concentration (ug L-1)
#' - `nout` >= 5: `C_ext`, external concentration (ug L-1)
#' - `nout` >= 6: `dBM`, biomass derivative (g dw d-1)
#' - `nout` >= 7: `dM_int`, mass of toxicant in plants derivative (ng d-1)
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `hmax = 0.1`<br>
#'    Maximum step length in time suitable for most simulations.
#'
#' @inheritSection Lemna_SETAC Effects
#' @inheritSection Transferable Biomass transfer
#' @param growth `character`, growth model to simulate: `"exp"` for exponential
#'    growth or `"log"` for logistic growth. Default is `"exp"`.
#' @references
#' Witt et al., submitted
#'
#' Klein J., Cedergreen N., Heine S., Reichenberger S., Rendal C.,
#' Schmitt W., Hommen U., 2021: *Refined description of the Lemna TKTD growth model
#' based on Schmitt et al. (2013) - equation system and default parameters*.
#' Report of the working group *Lemna* of the SETAC Europe Interest Group Effect
#' Modeling. Version 1.1, uploaded on 09 May 2022.
#' https://www.setac.org/group/effect-modeling.html
#'
#' @return an S4 object of type `Magma`
#' @seealso [Macrophyte-models], [Lemna-models], [Transferable], [Scenarios]
#' @family macrophyte models
#' @export
Magma <- function(growth=c("exp", "log")) {
  growth <- match.arg(growth)

  sc <- new("Magma",
      name=paste0("Magma-", growth),
      # `k_photo_max` only included for backwards compatibility of Myriophyllum-type scenarios
      param.req=c("mu_control", "E_max", "EC50_int", "b", "P", "r_A_DW",
                  "r_FW_DW", "r_FW_V", "r_DW_TSL", "K_pw", "k_met", "k_photo_max"),
      # default values as defined by Witt et al.
      param=list(mu_control=0.1, E_max=1, r_A_DW=1000, r_FW_DW=16.7, r_FW_V=1,
                 K_pw=1, k_met=0),
      # boundary presets defined by expert judgement
      param.bounds=list(mu_control=c(0, 1), E_max=c(0, 1), EC50_int=c(0, 1e6),
                        b=c(0.1, 20), P=c(0, 100)),
      endpoints=c("BM", "r"),
      control.req=TRUE,
      init=c(BM=0, M_int=0),
      transfer.comp.biomass="BM",
      transfer.comp.scaled="M_int",
      growth_model=growth
  ) %>% set_noexposure()

  # Adapt list of parameters for logistic growth
  if(growth == "log") {
    sc@param.req <- c(sc@param.req, "D_L")
    sc@param.bounds <- append(sc@param.bounds, list("D_L"=c(0, 1e6)))
  }

  sc
}

# constructors for backwards-compatibility
#' @export
#' @rdname Magma
Myrio <- function() {
  lifecycle::deprecate_soft(when="1.5.0", "Myrio()", "Magma()")
  Magma(growth="exp")
}

#' @export
#' @rdname Magma
Myrio_log <- function() {
  lifecycle::deprecate_soft(when="1.5.0", "Myrio_log()", "Magma(growth=\"log\")")
  Magma(growth="log")
}



########################
## Simulation
########################

# Numerically solve Magma scenarios
#
# @param scenario
# @param ... additional parameters passed on to [deSolve::ode()]
# @return data.frame
#' @include solver.R
#' @importFrom methods .hasSlot
solver_magma <- function(scenario, nout=1, method="lsoda", hmax=0.1, ...) {
  # for backwards compatibility with Myriophyllum-type scenarios
  if(!.hasSlot(scenario, "growth_model")) {
    # identify growth model to use
    gm <- "exp"
    if(is(scenario, "MyrioLog")) {
      gm <- "log"
    }
    # mock non-existing slot
    attr(scenario, "growth_model") <- gm

    # use `k_photo_max` as an alias of `mu_control`
    scenario@param["mu_control"] <- scenario@param["k_photo_max"]
  }
  # end of backwards compatibility

  params <- scenario@param

  # supplement parameter list with a placeholder value to ensure correct
  # parameter order
  if(scenario@growth_model == "exp") {
    params$growthno <- 1   # magic value: 1:=exp. growth, constant defined in model's C code
    params$D_L <- NA_real_ # magic value: dummy value w/o meaning, reqd to ensure parameter order
  } else {
    params$growthno <- 2 # magic value: 2:=log. growth, constant defined in model's C code
  }

  # make sure that parameters are present and in required order
  params_order <- c("mu_control", "growthno", "D_L", "E_max", "EC50_int", "b",
                    "P", "r_A_DW", "r_FW_DW", "r_FW_V", "r_DW_TSL", "K_pw",
                    "k_met")
  if(is.list(params)) {
    params <- unlist(params)
  }

  # check for missing parameters
  params_missing <- setdiff(params_order, names(params))
  if(length(params_missing)>0) {
    stop(paste("missing parameters:", paste(params_missing, collapse=",")))
  }

  # reorder parameters for deSolve
  params <- params[params_order]
  forcings <- list(scenario@exposure@series)

  # set names of additional output variables
  outnames <- c("TSL", "C_int", "f_photo", "C_int_unb", "C_ext", "dBM", "dM_int")

  ode(y=scenario@init, times=scenario@times, parms=params, dllname="cvasi",
        initfunc="magma_init", func="magma_func", initforc="magma_forc",
        forcings=forcings, nout=nout, method=method, hmax=hmax,
        outnames=outnames, ...)
}

#' @describeIn solver Numerically integrates `Magma` scenarios
setMethod("solver", "Magma", function(scenario, ...) solver_magma(scenario, ...))

########################
## Effects
########################

#' @include fx.R model-lemna_setac.R
#' @describeIn fx Effect at end of simulation of [Magma] scenarios
setMethod("fx", "Magma", function(scenario, ...) fx_lemna(scenario, ...))

test_that("Magma simulation", {
  nms <- intersect(Magma()@param.req, names(focusd1@param))
  params <- focusd1@param[nms]
  params$r_DW_TSL <- focusd1@param$r_DW_FN
  params$mu_control <- focusd1@param$k_photo_max

  # set up a Magma scenario with Lemna parameters for testing purposes
  Magma(growth="exp") %>%
    set_init(focusd1@init) %>%
    set_param(params) %>%
    set_exposure(focusd1@exposure@series) %>%
    set_times(0:20) -> sc

  focusd1 %>%
    set_param(c("k_photo_fixed"=1,"k_loss"=0)) %>%
    set_times(0:20) -> orig

  # equal results if all else is identical
  expect_equal(as.data.frame(solver(sc))$TSL, as.data.frame(solver(orig))$FrondNo)
  # check that TSL scales with r_DW_TSL
  r_DW_TSL <- focusd1@param$r_DW_FN * 10
  sc <- set_param(sc, c("r_DW_TSL"=r_DW_TSL))
  expect_equal(as.data.frame(solver(sc))$TSL, as.data.frame(solver(orig))$FrondNo/10)

  # Magma with logistic growth
  params$D_L <- focusd1@param$BM_L

  Magma(growth="log") %>%
    set_init(focusd1@init) %>%
    set_param(params) %>%
    set_exposure(focusd1@exposure@series) %>%
    set_times(0:20) %>%
    simulate(nout=7) %>%
    as.data.frame() -> out_log
  # we should have almost reached the density limit
  expect_equal(tail(out_log$BM, 1), params$D_L, tolerance=0.1)
  # dBM should decrease continuously because initial state is above 50% of BM_L
  expect_true(all(out_log$dBM > 0))
  expect_true(all(out_log$BM == sort(out_log$BM)))

  # error if r_DW_TSL is missing
  sc@param$r_DW_TSL <- NULL
  expect_error(solver(sc), regexp="missing.*r_DW_TSL")
})

# Scenario sequence class
#' @export
#' @aliases sequence
#' @include sequence.R
setClass("ScenarioSequence",
         slots=list(scenarios="list",
                    breaks="numeric",
                    inc_start="logical",
                    inc_end="logical"))

#' Sequence of scenarios
#'
#' Scenario sequences can be used to implement changes in model parameters
#' over time, which otherwise would remain constant for the duration of a simulation.
#' A sequence of scenarios is treated as a single scenario and each scenario
#' is simulated one after the other. If scenario `n` in a sequence was simulated,
#' scenario `n+1` will start off in the model state where `n` had ended.
#'
#' Sequences are generally treated the same as scenarios: Sequences can be
#' simulated, as well as effects and EPx can be derived.
#'
#' ### Requirements
#' All scenarios in a sequence must fulfill the following requirements:
#'
#' * All scenarios must have identical state variables
#' * The *output times* of all scenarios must represent a continuous time series
#'   without gaps or overlaps
#'
#' Using the `breaks` argument, the function can split up the scenarios'
#' output times at the given break points. The break points must be within
#' the interval defined by the superset of all output times in the sequence.
#'
#' @param seq list of [scenario] objects
#' @param breaks optional vector of *numerics*, scenarios' output times will
#'   be modified so that one scenario ends at the break and the next one begins
#' @return an S4 object of type [ScenarioSequence-class]
#' @examples
#' # Create a scenario with background mortality only
#' scen1 <- minnow_it %>%
#'   set_noexposure() %>%
#'   set_times(0:10)
#' # Modify a scenario parameter, e.g. set background mortality to zero
#' scen2 <- scen1 %>% set_param(c(hb=0))
#'
#' # Create a sequence of scenarios, scenario #1 will be simulated for the
#' # time period [0, 4], and #2 for [4, 10]
#' sq <- sequence(list(scen1, scen2), breaks=c(4))
#'
#' # Simulate the sequence: the mortality stops after t=4.0, due to scenario #2
#' # being simulated after t=4.0, which disabled the background mortality
#' simulate(sq)
#' # Effect endpoints can also be calculated
#' effect(sq)
#'
#' @name sequence
#' @aliases ScenarioSequence-class
#' @seealso [sequence.extract]
#' @export
sequence <- function(seq, breaks=NULL) {
  if(missing(seq))
    stop("Argument 'seq' is missing")
  if(!is.list(seq))
    stop("Argument 'seq' must be of type list")
  if(length(seq) == 0)
    stop("Argument `seq` is empty")
  if(length(seq) == 1)
    warning("Argument `seq` has ony a single element")
  if(any(!is_scenario(seq)))
     stop("Argument `seq` must contain scenario objects only")
  if(!is.null(breaks)) {
    if(any(!is.numeric(breaks)))
      stop("Argument 'breaks' must be vector of numerics")
    if(length(breaks) + 1 != length(seq))
      stop("Length of argument 'breaks' does not fit to length of sequence")
  }

  obj <- new("ScenarioSequence", scenarios=seq)
  obj@inc_start <- c(TRUE, rep(FALSE, length(obj) - 1))
  obj@inc_end <- rep(TRUE, length(obj))
  if(!is.null(breaks)) {
    obj@breaks <- breaks
    obj <- split_sequence(obj, .messages=FALSE)
  } else {
    obj@breaks <- breaks_from_sequence(obj)
  }
  check_sequence(obj)
  obj <- mock_as_scenario(obj)
  obj
}

#' Extract and replace elements of a sequence
#'
#' The array accessor generics allow extracting and replacing scenarios within
#' am existing sequence. `[` and `[[` work identical to
#'
#' @param x [sequence]
#' @param i index of elements to extract or replace
#' @param j *not used*
#' @param value new scenario
#' @name sequence.extract
#' @rdname sequence.extract
#' @aliases [[<-,ScenarioSequence,numeric,ANY,ANY-method
#' @return various
#' @examples
#' # create a sequence
#' seq <- sequence(list(minnow_it, minnow_it), breaks=3)
#'
#' seq[1]       # first element, as a list of scenarios
#' seq[c(1)]    # the same
#' seq[c(1, 2)] # both elements as a list of scenarios
#' seq[[1]]     # first element as a scenario
#'
#' # replacing single elements
#' seq[[1]] <- minnow_sd %>% set_times(1:3)
NULL


#' @export
#' @describeIn sequence.extract Returns a list of scenarios from the sequence.
setMethod("[", c("ScenarioSequence","numeric","missing","missing"), function(x, i) {
  if(any(is.na(i)))
    stop("Argument `i` must not contain NAs")
  if(any(i < 1 | i > length(x@scenarios)))
    stop("Index is out of bounds.")

  x@scenarios[i]
})

#' @export
#' @describeIn sequence.extract Returns a single scenario from the sequence.
setMethod("[[", c("ScenarioSequence", "numeric"), function(x, i) {
  if(length(i) != 1)
    stop("Index must be of length one.")
  if(any(i < 1 | i > length(x@scenarios)))
    stop("Index is out of bounds.")

  x@scenarios[[i]]
})

#' @export
#' @describeIn sequence.extract Replaces a single scenario in the sequence.
setMethod("[[<-", c("ScenarioSequence","numeric","missing","EffectScenario"), function(x, i, j, value) {
  if(any(is.na(i))) {
    stop("Argument `i` must not contain NAs")
  }

  if(length(i) != 1) {
    stop("Index must be of length one")
  }
  if(any(i < 1 | i > length(x@scenarios))) {
    stop("Index is out of bounds")
  }

  x@scenarios[[i]] <- value
  check_sequence(x)
  if(i == 1) {
    x <- mock_as_scenario(x)
  }

  x
})

# do not export, re-assigning scenarios has complicated consequences
# @describeIn sequence.extract Replaces a single scenario in the sequence.
setMethod("[[<-", c("ScenarioSequence","numeric","ANY","ANY"), function(x, i, j, value) {
  stop("Assigned type not supported")
})

#' @export
#' @describeIn sequence.extract Returns the number of scenarios in the sequence.
setMethod("length", "ScenarioSequence", function(x) {
  length(x@scenarios)
})

breaks_from_sequence <- function(seq) {
  if(any(!is_sequence(seq))) {
    stop("Argument `seq` must be a sequence")
  }

  brks <- numeric(0)
  for(i in seq_along(seq@scenarios)) {
    brks <- c(brks, tail(get_times(seq@scenarios[[i]]), n=1))
  }
  # drop last item
  head(brks, n=-1)
}

split_sequence <- function(x, .messages=FALSE) {
  # check for backwards compatibility
  if(!methods::.hasSlot(x, "breaks")) {
    stop("Sequence has outdated format, please create a new sequence object")
  }

  breaks <- x@breaks
  n <- length(x)
  if(n == 1) {
    return(x)
  }

  if(.messages) {
    cli::cli_text("Splitting sequence at {n} break{?s} ...")
  }

  x@inc_start <- rep(TRUE, n)
  x@inc_end <- rep(TRUE, n)
  is_relevant <- rep(FALSE, n)
  for(i in seq_along(x))
  {
    is_last <- i == length(x)
    t_start <- if(i == 1) -Inf else breaks[[i - 1]]
    t_end <- if(is_last) Inf else breaks[[i]]
    times <- get_times(x[[i]])

    # is there overlap of the scenario and the time period we are selecting for?
    if(min(times) < t_end & max(times) > t_start) {
      times <- times[times >= t_start & times <= t_end] # reframe output times
      is_relevant[[i]] <- TRUE
    } else { # no overlap
      times <- numeric(0)
      x@inc_start[[i]] <- FALSE
      x@inc_end[[i]] <- FALSE
    }
    x@scenarios[[i]]@times <- times
  }

  # check if start and end time need to be included in the output times
  for(i in seq_along(x))
  {
    if(is_relevant[[i]])
    {
      t_start <- if(i == 1) -Inf else breaks[[i - 1]]
      t_end <- if(i == n) Inf else breaks[[i]]
      times <- get_times(x[[i]])
      any_before <- any(head(is_relevant, i - 1))
      any_after <- any(tail(is_relevant, n - i))

      # continue where previous scenario ended
      if(length(times) == 0) {
        times <- t_start
        x@inc_start[[i]] <- FALSE
        x@inc_end[[i]] <- FALSE
      }
      if(any_before)
      {
        x@inc_start[[i]] <- FALSE
        if(times[[1]] != t_start) {
          times <- c(t_start, times)
        }
      }
      if(any_after && times[[length(times)]] != t_end) {
        times <- c(times, t_end)
        x@inc_end[[i]] <- FALSE
      }
      x@scenarios[[i]]@times <- times
    }

    if(.messages)
    {
      if(is_relevant[[i]])
        cli::cli_text(" Scenario #", i, ": simulated period [", min(times), ", ", max(times), "]")
      else
        cli::cli_text(" Scenario #", i, ": not simulated")
    }
  }

  x
}

# check validity of sequence elements
check_sequence <- function(seq) {
  if(!all(is_scenario(seq@scenarios))) {
    stop("sequence does not contain a list of effect scenarios")
  }
  # check if start and end of output times match between subsequent scenarios
  for(i in seq_along(seq)) {
    cur <- seq[[i]]
    if(is.null(cur)) {
      next
    }
    if(length(cur@times) == 0) {
      next
    }
    # skip further check for first element in sequence
    if(i == 1) {
      next
    }

    prev <- seq[[i-1]]
    if(is.null(prev)) { # nothing to compare to
      next
    }
    if(length(prev@times) == 0) {
      next
    }

    end <- tail(prev@times, 1)
    start <- head(cur@times, 1)
    if(end < start) {
      stop(sprintf("output time gap between scenario #%d  and #%d: [_, %g], [%g, _]", i-1, i, end, start))
    } else if(start < end) {
      stop(sprintf("output time overlap between scenario #%d and #%d: [_, %g], [%g, _]", i-1, i, end, start))
    }
  }
}

# Sets a scenario's slot values as attributes of a sequence.
#
# This emulates or mocks the existence of slots in the sequence
# object, at least for read-only access. Writing to these emulated slots will
# fail with an error.
#
# This is a workaround to make code work with sequence objects, which is not
# aware of sequences and cannot handle them, e.g. some model-specific `fx()`
# functions. Will not mock the contents of exposure and forcings slots, as these
# are too difficult to handle as the actual time-series may also depend on
# solver settings.
#
# @param sq [sequence] where slots are emulated
# @param scenario [scenario] to mock slots from
# @return modified [sequence]
mock_as_scenario <- function(sq, scenario=NULL) {
  if(!is_sequence(sq))
    stop("Argument `sq` must be a sequence")
  if(is.null(scenario))
    scenario <- sq[[1]]
  if(!is_scenario(scenario))
    stop("Argument `scenario` must be a scenario")

  # these slots/attributes will never be set/mocked
  protected <- slotNames(new("ScenarioSequence"))
  exclude <- c(protected, "times", "exposure", "forcings")

  nms <- slotNames(scenario)
  if(any(protected %in% nms)) {
    warning("Mocked scenario contains slots with protected names: ", paste(intersect(nms, protected), collapse=", "))
  }
  nms <- setdiff(nms, exclude)
  # mock slots by setting an attribute of the same name, this hack works because S4
  # slots are implemented by R using attributes.
  for(nm in nms) {
    attr(sq, nm) <- slot(scenario, nm)
  }
  # special handling for output times
  attr(sq, "times") <- get_times(sq)
  # A sequence's exposure and forcings should never ever be accessed by using slots.
  # Setting them to NA should make sure that any code accessing them fails.
  attr(sq, "exposure") <- NA
  attr(sq, "forcings") <- NA

  sq
}

test_that("supported object types", {
  # empty argument
  expect_equal(get_tag(list()), list())
  # single scenario
  s1 <- new("EffectScenario") %>% set_tag("foo") %>% set_times(0:5)
  expect_equal(get_tag(s1), "foo")
  # multiple scenarios
  s2 <- new("EffectScenario") %>% set_tag("bar") %>% set_times(0:5)
  expect_equal(get_tag(list(s1, s2)), list("foo", "bar"))
  # scenario sequence
  suppressMessages(seq <- sequence(list(s1, s2), breaks=3))
  expect_equal(get_tag(seq), "foo")
  # parameter set
  ps <- parameter_set("xy", tag="baz")
  expect_equal(get_tag(ps), "baz")
})

test_that("unsupported object types", {
  # invalid arguments
  expect_error(get_tag(1))
})

test_that("basic use", {
  sc <- new("EffectScenario")
  tms <- 0:10

  sc <- set_times(sc, tms)
  expect_equal(sc@times, tms)
})

test_that("set units", {
  sc <- new("EffectScenario")
  tms <- units::set_units(0:10, "h")
  sc <- set_times(sc, tms)
  expect_equal(sc@times, 0:10)
})

test_that("sequence", {
  sc <- new("EffectScenario") %>% set_times(0:5)
  sq <- sequence(list(sc, sc), breaks=3)

  expect_equal(get_times(sq), 0:5)

  sq <- sq %>% set_times(1:5)
  expect_equal(get_times(sq), 1:5)

  sq <- sq %>% set_times(10:15)
  expect_equal(get_times(sq), 10:15)
})

test_that("supported types", {
  param <- list("foo"=1, "bar"=2)
  # empty argument
  expect_equal(get_param(list()), list())
  # scenario
  sc <- new("EffectScenario") %>%
    set_param(param) %>%
    set_times(0:5)
  expect_equal(get_param(sc), param)
  # list of scenarios
  expect_equal(get_param(c(sc, sc)), list(param, param))

  # sequence
  sq <- sequence(list(sc, sc), breaks=3)
  expect_equal(get_param(sq), param)

  # parameter set
  ps <- parameter_set("foobar", param)
  expect_equal(get_param(ps), param)
})

test_that("unsupported types", {
  expect_error(get_param(1))
})

test_that("get_rparam", {
  rp <- c("foo", "bar")
  sc1 <- new("EffectScenario", param.req=rp) %>% set_times(1:5)

  # empty argument
  expect_equal(get_rparam(list()), list())
  # single scenario
  expect_equal(get_rparam(sc1), rp)
  # multiple scenarios
  expect_equal(get_rparam(list(sc1, sc1)), list(rp, rp))
  # sequence
  suppressMessages(sq <- sequence(list(sc1, sc1), breaks=3))
  expect_equal(get_rparam(sq), rp)

  # invalid argument
  expect_error(get_rparam(1))
})

test_that("supported types", {
  tms <- 0:5
  sc <- new("EffectScenario") %>% set_times(tms)

  # scenarios
  expect_equal(get_times(sc), tms)
  expect_equal(get_times(list(sc, sc)), list(tms, tms))

  # sequences
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(get_times(sq), tms)
  sq <- sequence(list(sc, sc), breaks=3)
  expect_equal(get_times(sq), tms)
})

test_that("unsupported types", {
  expect_error(get_times(1), "not supported")
})

test_that("arg scenario=invalid", {
  sc <- minnow_it

  expect_error(caliset(), "scenario. is missing")
  expect_error(caliset(scenario=NULL), "must be a scenario")
  expect_error(caliset(scenario=list()), "must be a scenario")
  expect_error(caliset(scenario=list(sc, sc)), "must be a scenario")
  expect_error(caliset(scenario=1), "must be a scenario")
})

test_that("arg data=invalid", {
  sc <- minnow_it

  expect_error(caliset(sc), "is missing")
  expect_error(caliset(sc, 1), "data.frame")
  expect_error(caliset(sc, data.frame()), "two columns")
  expect_error(caliset(sc, data.frame(t=1)), "two columns")
  expect_error(caliset(sc, data.frame(t=numeric(0), o=numeric(0))), "not be empty")
  expect_error(caliset(sc, data.frame(t=1, o="a")), "be numerical")
  expect_error(caliset(sc, data.frame(t="a", o=1)), "be numerical")
  expect_error(caliset(sc, data.frame(t="a", o="a")), "be numerical")
  # 1st column
  expect_error(caliset(sc, data.frame(t=NA, o=1)), "invalid value")
  expect_error(caliset(sc, data.frame(t=NaN, o=1)), "invalid value")
  expect_error(caliset(sc, data.frame(t=Inf, o=1)), "invalid value")
  expect_error(caliset(sc, data.frame(t=c(1, 0), o=1)), "ascending order")
  # 2nd column
  expect_error(caliset(sc, data.frame(t=1, o=NaN)), "invalid value")
  expect_error(caliset(sc, data.frame(t=1, o=Inf)), "invalid value")
})

test_that("arg data=observed include NA", {
  sc <- minnow_it
  data <- data.frame(times=0:3, obs=c(0, NA_real_, 1, 2))

  cs <- suppressMessages(caliset(sc, data=data))
  expect_equal(cs@scenario, sc)
  expect_equal(cs@data, data[!is.na(data$obs), ])
})

test_that("arg weight=invalid", {
  sc <- minnow_it
  df1 <- data.frame(t=1, o=2)
  dfn <- data.frame(t=1:8, o=2)

  expect_error(caliset(sc, df1, NA))
  expect_error(caliset(sc, df1, NaN))
  expect_error(caliset(sc, df1, Inf))
  expect_error(caliset(sc, df1, "foo"))

  expect_error(caliset(sc, df1, 1:3))
  expect_error(caliset(sc, dfn, 1:3))
})

test_that("caliset creation", {
  sc <- minnow_it
  df <- data.frame(t=1:10, o=0)
  wgt1 <- 1
  wgtn <- 2:11
  tag <- list("foo", 23)

  cs <- caliset(sc, df, wgt1, tag)
  expect_equal(cs@scenario, sc)
  expect_equal(cs@data, df)
  expect_equal(cs@weight, rep(wgt1, nrow(df)))
  expect_equal(cs@tag, tag)

  cs <- caliset(sc, df, wgtn)
  expect_equal(cs@weight, wgtn)
  expect_null(cs@tag)
})

test_that("caliset with sequences", {
  sc <- new("EffectScenario") %>% set_times(0:6)
  sq <- sequence(list(sc, sc), breaks=3)
  cs <- caliset(sq, data.frame(t=0:10, n=0))

  expect_equal(cs@scenario, sq)
})

test_that("sequence creation", {
  sc <- new("EffectScenario") %>% set_times(0:10)

  # single element
  expect_warning(sequence(list(sc)))
  # multiple elements
  expect_error(sequence(list(sc, sc)), "time overlap")
  sc2 <- sc %>% set_times(10:20)
  expect_equal(sequence(list(sc, sc2))@scenarios, list(sc, sc2))
})

test_that("split_sequence", {
  # scenarios overlap
  sc <- new("EffectScenario") %>% set_times(0:10)

  # single scenario, no changes
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(sq[[1]]@times, 0:10)

  # two scenarios, both scenarios include the break
  sq <- sequence(list(sc, sc), breaks=5)
  expect_equal(sq[[1]]@times, 0:5)
  expect_equal(sq[[2]]@times, 5:10)
  expect_equal(sq@inc_start, c(TRUE, FALSE))
  expect_equal(sq@inc_end, c(TRUE, TRUE))

  # break before the first scenario
  sq <- sequence(list(sc, sc), breaks=-1)
  expect_equal(sq[[1]]@times, integer(0))
  expect_equal(sq[[2]]@times, 0:10)
  expect_equal(sq@inc_start, c(FALSE, TRUE))
  expect_equal(sq@inc_end, c(FALSE, TRUE))

  # break after the second scenario
  sq <- sequence(list(sc, sc), breaks=11)
  expect_equal(sq[[1]]@times, 0:10)
  expect_equal(sq[[2]]@times, integer(0))
  expect_equal(sq@inc_start, c(TRUE, FALSE))
  expect_equal(sq@inc_end, c(TRUE, FALSE))


  suppressMessages(l <- sequence(list(sc, sc), breaks=1)@scenarios)
  expect_equal(l[[1]]@times, c(0, 1))
  expect_equal(l[[2]]@times, 1:10)

  suppressMessages(l <- sequence(list(sc, sc), breaks=9)@scenarios)
  expect_equal(l[[1]]@times, 0:9)
  expect_equal(l[[2]]@times, 9:10)

  suppressMessages(l <- sequence(list(sc, sc), breaks=0.5)@scenarios)
  expect_equal(l[[1]]@times, c(0, 0.5))
  expect_equal(l[[2]]@times, c(0.5, 1:10))

  suppressMessages(l <- sequence(list(sc, sc), breaks=9.5)@scenarios)
  expect_equal(l[[1]]@times, c(0:9, 9.5))
  expect_equal(l[[2]]@times, c(9.5, 10))

  # scenarios have a gap
  sc1 <- new("EffectScenario") %>% set_times(0:4)
  sc2 <- new("EffectScenario") %>% set_times(6:10)

  suppressMessages(l <- sequence(list(sc1, sc2), breaks=3)@scenarios)
  expect_equal(l[[1]]@times, 0:3)
  expect_equal(l[[2]]@times, c(3, 6:10))

  suppressMessages(l <- sequence(list(sc1, sc2), breaks=4)@scenarios)
  expect_equal(l[[1]]@times, 0:4)
  expect_equal(l[[2]]@times, c(4, 6:10))

  suppressMessages(l <- sequence(list(sc1, sc2), breaks=5)@scenarios)
  expect_equal(l[[1]]@times, 0:5)
  expect_equal(l[[2]]@times, 5:10)

  suppressMessages(l <- sequence(list(sc1, sc2), breaks=6)@scenarios)
  expect_equal(l[[1]]@times, c(0:4, 6))
  expect_equal(l[[2]]@times, 6:10)

  suppressMessages(l <- sequence(list(sc1, sc2), breaks=6.5)@scenarios)
  expect_equal(l[[1]]@times, c(0:4, 6.5))
  expect_equal(l[[2]]@times, c(6.5, 7:10))
})

test_that("set new times", {
  sc1 <- new("EffectScenario") %>% set_times(0:3)
  sc2 <- new("EffectScenario") %>% set_times(3:6)

  # no breaks
  suppressWarnings(sq <- sequence(list(sc1)) %>% set_times(1:5))
  expect_equal(sq[[1]]@times, 1:5)

  # some breaks
  sq <- sequence(list(sc1, sc2)) %>% set_times(1:5)
  expect_equal(sq[[1]]@times, 1:3)
  expect_equal(sq[[2]]@times, 3:5)

  # no output for 1st scenario in sequence
  sq <- sequence(list(sc1, sc2)) %>% set_times(5:7)
  expect_equal(sq[[1]]@times, numeric(0))
  expect_equal(sq[[2]]@times, 5:7)

  # no output for 2nd scenario in sequence
  sq <- sequence(list(sc1, sc2)) %>% set_times(1:3)
  expect_equal(sq[[1]]@times, 1:3)
  expect_equal(sq[[2]]@times, numeric(0))
})

test_that("invalid arguments", {
  sc <- new("EffectScenario") %>% set_times(0:10)

  expect_error(sequence())
  expect_error(sequence("foo"))
  expect_error(sequence(list()))
  expect_warning(sequence(list(sc)))
  expect_error(sequence(list(sc, "foo")))
  expect_error(sequence(list(sc, sc), breaks="foo"))
  expect_error(sequence(list(sc, sc), breaks=c(1, 2)))
})

test_that("breaks_from_sequence", {
  sc1 <- new("EffectScenario") %>% set_times(0:5)

  # no breaks
  suppressWarnings(sq <- sequence(list(sc1)))
  expect_equal(breaks_from_sequence(sq), numeric(0))
  # one break
  suppressMessages(sq <- sequence(list(sc1, sc1), breaks=3))
  expect_equal(breaks_from_sequence(sq), 3)
  # multiple breaks
  suppressMessages(sq <- sequence(list(sc1, sc1, sc1), breaks=c(2,3.1)))
  expect_equal(breaks_from_sequence(sq), c(2, 3.1))
})

test_that("included outputs", {
  # include breaks if scenarios were split by user
  sc1 <- minnow_it %>% set_times(0:3)
  sc2 <- minnow_it %>% set_times(3:6)
  sq <- sequence(list(sc1, sc2))
  expect_equal(sq@inc_start, c(TRUE, FALSE))
  expect_equal(sq@inc_end, c(TRUE, TRUE))

  # remove breaks from output if it was automatically split
  sc1 <- minnow_it %>% set_times(0:2)
  sc2 <- minnow_it %>% set_times(3:6)
  sq <- sequence(list(sc1, sc2), breaks=3)
  expect_equal(sq@inc_start, c(TRUE, FALSE))
  expect_equal(sq@inc_end, c(FALSE, TRUE))

  # multiple breaks
  sc <- minnow_it %>% set_times(0:6)
  sq <- sequence(list(sc, sc, sc), breaks=c(2.1, 4))
  expect_equal(sq@inc_start, c(TRUE, FALSE, FALSE))
  expect_equal(sq@inc_end, c(FALSE, TRUE, TRUE))
})

test_that("sequence_check", {
  sc <- new("EffectScenario")

  # gap in output times
  sc1 <- sc %>% set_times(1:2)
  sc2 <- sc %>% set_times(4:5)
  expect_error(sequence(list(sc1, sc2)), "time gap")
  # output time overlap
  sc1 <- sc %>% set_times(1:2)
  sc2 <- sc %>% set_times(1:3)
  expect_error(sequence(list(sc1, sc2)), "time overlap")
})

test_that("array access", {
  sc1 <- new("EffectScenario") %>% set_times(0:3)
  sc2 <- new("EffectScenario") %>% set_times(3:6)
  sc3 <- new("EffectScenario") %>% set_times(3:9)
  sq <- sequence(list(sc1, sc2))

  # multiple elements
  expect_equal(sq[1], list(sc1))
  expect_equal(sq[2], list(sc2))
  expect_equal(sq[c(1, 2)], list(sc1, sc2))

  # single elements
  expect_equal(sq[[1]], sc1)
  expect_equal(sq[[2]], sc2)

  # assignments
  sq[[2]] <- sc3
  expect_equal(sq[[2]], sc3)

  # invalid arguments
  expect_error(sq[0], "out of bounds")
  expect_error(sq[10], "out of bounds")
  expect_error(sq[c(0, 42)], "out of bounds")

  expect_error(sq[[0]], "out of bounds")
  expect_error(sq[[10]], "out of bounds")
  expect_error(sq[[c(1, 2)]], "length one")

  expect_error(sq[[c(1, 2)]] <- sc3, "length one")
  expect_error(sq[[0]] <- sc3, "out of bounds")
  expect_error(sq[[10]] <- sc3, "out of bounds")
})

test_that("mocking scenarios", {
  sq <- new("ScenarioSequence")
  sc <- new("EffectScenario", tag="foo", exposure=no_exposure()) %>% set_times(0:5)

  # slot not emulated yet -> fail
  expect_error(sq@tag)

  # mock from supplied scenario
  mock <- mock_as_scenario(sq, sc)
  expect_equal(mock@tag, sc@tag)

  # writing should fail
  expect_error(mock@tag <- "bar")
  # some slot should never be mocked
  expect_equal(mock@exposure, NA)

  # mock from first scenario in sequence
  sc1 <- sc %>% set_tag("baz")
  sq <- sequence(list(sc1, sc), breaks=3)
  mock <- mock_as_scenario(sq)
  expect_equal(mock@tag, sc1@tag)
  expect_equal(mock@times, sc@times)
})

test_that("mocked slots", {
  # newly created
  sc1 <- new("EffectScenario", tag="foo", times=0:5)
  sc2 <- sc1 %>% set_tag("bar")
  sq <- sequence(list(sc1, sc2), breaks=3)
  expect_equal(sq@tag, sc1@tag)

  # first scenario updated
  sq[[1]] <- sq[[1]] %>% set_tag("baz")
  expect_equal(sq[[1]]@tag, "baz")
  expect_equal(sq@tag, "baz")
})

test_that("with simulate()", {
  tol <- 1e-5

  # guts red it
  sc <- minnow_it
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(simulate(sq), simulate(sc), tolerance=tol)

  # lemna
  sc <- metsulfuron
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(simulate(sq), simulate(sc), tolerance=tol)

  # multiple elements in sequence
  suppressWarnings(sq <- sequence(list(sc, sc), breaks=3.1))
  expect_equal(simulate(sq), simulate(sc), ignore_attr=TRUE, tolerance=tol)
})

test_that("with fx()", {
  tol <- 1e-5

  # guts red it
  sc <- minnow_it
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(fx(sq), fx(sc), tolerance=tol)

  # lemna
  sc <- metsulfuron
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(fx(sq), fx(sc), tolerance=tol)
})

test_that("with effect()", {
  tol <- 1e-5

  # no moving windows
  sc <- minnow_it
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(effect(sq, ep_only=TRUE), effect(sc, ep_only=TRUE), ignore_attr=TRUE, tolerance=tol)

  # with moving windows
  sc <- minnow_it %>% set_window(length=2, interval=2)
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(effect(sq, ep_only=TRUE, max_only=FALSE),
               effect(sc, ep_only=TRUE, max_only=FALSE),
               ignore_attr=TRUE, tolerance=tol)

  sc <- minnow_it %>%
    set_exposure(data.frame(t=0:4, c=c(0, 0, 0, 1, 1))) %>%
    set_window(length=2, interval=2)
  sq <- sequence(list(sc, sc), breaks=2)
  expect_equal(effect(sq, ep_only=TRUE, max_only=FALSE),
               effect(sc, ep_only=TRUE, max_only=FALSE),
               ignore_attr=TRUE, tolerance=tol)
})

test_that("with epx()", {
  tol <- 1e-5

  # no moving windows
  sc <- minnow_it
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(epx(sc, ep_only=TRUE),
               epx(sq, ep_only=TRUE),
               tolerance=tol, ignore_attr=TRUE)

  # with moving windows
  sc <- sc %>%
    set_exposure(data.frame(t=0:4, c=c(0, 0, 0, 1, 1))) %>%
    set_window(length=2, interval=1)
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(epx(sc, ep_only=TRUE),
               epx(sq, ep_only=TRUE),
               tolerance=tol, ignore_attr=TRUE)
})

test_that("single scenario", {
  sc <- GUTS_RED_IT()

  # basic use
  expect_equal(set_init(sc, c(D=23))@init, c(D=23, H=0))
  expect_equal(set_init(sc, c(D=23,H=42))@init, c(D=23, H=42))
  expect_equal(set_init(sc, c())@init, c(D=0, H=0))

  # lists
  expect_equal(set_init(sc, list(D=23))@init, c(D=23, H=0))
})

test_that("vector of scenarios", {
  sc1 <- GUTS_RED_IT() %>% set_init(c(D=1))
  sc2 <- GUTS_RED_IT() %>% set_init(c(D=2))

  rs <- set_init(c(sc1, sc2), c(D=23))
  expect_equal(rs[[1]]@init, c(D=23, H=0))
  expect_equal(rs[[2]]@init, c(D=23, H=0))
})

test_that("scenario w/o state vars", {
  sc <- new("EffectScenario")
  # gracefully accept any parameter if init is empty
  expect_equal(set_init(sc, list(a=1, b=2))@init, c(a=1, b=2))
  # ... but reject additional values on the second call
  expect_warning(sc %>% set_init(c(a=1)) %>% set_init(c(b=2)), "not state variables")
})

test_that("scenario sequence", {
  sc1 <- minnow_it %>% set_init(c(D=1))
  sc2 <- minnow_it %>% set_init(c(D=2))
  suppressMessages(seq <- sequence(list(sc1, sc2), breaks=2))

  seq <- seq %>% set_init(c(D=5))
  expect_equal(seq@scenarios[[1]]@init, c(D=5, H=0))
  expect_equal(seq@scenarios[[2]]@init, c(D=5, H=0))
})

test_that("invalid arguments", {
  sc <- GUTS_RED_IT()

  expect_error(set_init(sc))
  expect_error(set_init(sc, c(D=23, 42)), "unnamed elements")
  expect_error(set_init(sc, c(D="a")), "must be numeric")
  expect_warning(set_init(sc, c(D=NaN)), "invalid values")

  expect_warning(set_init(sc, c(foo=0)), "not state variable")
})

test_that("vanilla usage", {
  sc1 <- new("EffectScenario", name="foo")
  sc2 <- new("EffectScenario", name="foo", tag="bar")

  p <- list(kd=1, hb=2, alpha=3, beta=4)
  ps1 <- parameter_set("foo", p)

  ## vanilla scenarios
  # single scenario, set vector of atomic values
  expect_equal(set_param(sc1, unlist(p))@param, p)
  # single scenario, single parameter set
  expect_equal(set_param(sc1, ps1)@param, p)
  # multiple scenarios, atomic vector
  lst <- set_param(c(sc1,sc1), unlist(p))
  expect_equal(lst[[1]]@param, p)
  expect_equal(lst[[2]]@param, p)
  # multiple scenarios, parameter set
  lst <- set_param(c(sc1,sc1), ps1)
  expect_equal(lst[[1]]@param, p)
  expect_equal(lst[[2]]@param, p)
})

test_that("special cases", {
  sc1 <- minnow_it
  sc2 <- sc1 %>% set_times(sc1@times + max(sc1@times))
  p <- list(kd=1)
  ps1 <- parameter_set(sc1@name, p, tag=sc1@tag)

  # all scenarios within a sequence need to be modified
  sequence(seq=c(sc1, sc2)) %>%
    set_param(ps1) -> seq
  expect_equal(length(seq@scenarios), 2)
  expect_equal(seq@scenarios[[1]]@param$kd, p$kd)
  expect_equal(seq@scenarios[[2]]@param$kd, p$kd)
})

test_that("arg x=ScenarioSequence", {
  sc <- GUTS_RED_IT() %>% set_times(0:5)
  sq <- sequence(c(sc, sc), breaks=3)
  pv <- c(hb=23)
  ps <- parameter_set(sc@name, pv)

  ## single sequence
  # param=vector
  foo <- set_param(sq, pv)
  expect_equal(foo@scenarios[[1]]@param, as.list(pv))
  expect_equal(foo@scenarios[[2]]@param, as.list(pv))
  # param=ParameterSet
  foo <- set_param(sq, ps)
  expect_equal(foo@scenarios[[1]]@param, as.list(pv))
  expect_equal(foo@scenarios[[2]]@param, as.list(pv))

  ## multiple sequences
  lst <- list(sq, sq)
  # param=vector
  foo <- set_param(lst, pv)
  expect_equal(length(foo), 2)
  expect_equal(foo[[1]]@scenarios[[1]]@param, as.list(pv))
  expect_equal(foo[[1]]@scenarios[[2]]@param, as.list(pv))
  expect_equal(foo[[2]]@scenarios[[1]]@param, as.list(pv))
  expect_equal(foo[[2]]@scenarios[[2]]@param, as.list(pv))
  # param=ParameterSet
  foo <- set_param(lst, ps)
  expect_equal(length(foo), 2)
  expect_equal(foo[[1]]@scenarios[[1]]@param, as.list(pv))
  expect_equal(foo[[1]]@scenarios[[2]]@param, as.list(pv))
  expect_equal(foo[[2]]@scenarios[[1]]@param, as.list(pv))
  expect_equal(foo[[2]]@scenarios[[2]]@param, as.list(pv))
})

test_that("arg x=CalibrationSet", {
  sc <- GUTS_RED_IT() %>% set_times(0:5)
  cs <- caliset(sc, data=data.frame(t=0, o=0))
  pv <- c(hb=23)
  ps <- parameter_set(sc@name, pv)

  ## single sequence
  # param=vector
  foo <- set_param(cs, pv)
  expect_equal(foo@scenario@param, as.list(pv))
  # param=ParameterSet
  foo <- set_param(cs, ps)
  expect_equal(foo@scenario@param, as.list(pv))

  ## multiple sequences
  lst <- list(cs, cs)
  # param=vector
  foo <- set_param(lst, pv)
  expect_equal(length(foo), 2)
  expect_equal(foo[[1]]@scenario@param, as.list(pv))
  expect_equal(foo[[2]]@scenario@param, as.list(pv))
  # param=ParameterSet
  foo <- set_param(lst, ps)
  expect_equal(length(foo), 2)
  expect_equal(foo[[1]]@scenario@param, as.list(pv))
  expect_equal(foo[[2]]@scenario@param, as.list(pv))
})

test_that("invalid arguments", {
  sc1 <- new("EffectScenario", name="foo")
  sc2 <- new("EffectScenario", name="foo", tag="bar")

  p <- list(kd=1)
  ps1 <- parameter_set("foo", p)
  ps2 <- parameter_set("foo", tag="bar", param=p)

  # multiple scenarios, one mismatch
  expect_error(set_param(c(sc1, sc2), ps1))
  # multiple parameter sets match, i.e. ambiguous assignments
  suppressMessages(expect_error(set_param(sc1, list(ps1, ps1))))
  # inconsistent types
  expect_error(set_param(sc1, list(1, ps1)))
  # model & tag dont match
  expect_error(set_param(sc1, ps2))
  # warn if invalid parameters were passed as argument
  sc3 <- new("EffectScenario", param.req=c("a"))
  expect_warning(set_param(sc3, c("b"=1)))
})

test_that("scenario", {
  sc <- new("EffectScenario", param.bounds=list())
  expect_equal(get_bounds(sc), list())

  sc <- new("EffectScenario", param.bounds=list(foo=c(0, 1)))
  expect_equal(get_bounds(sc), list(foo=c(0, 1)))
})

test_that("sequence", {
  sc <- new("EffectScenario", param.bounds=list()) %>% set_times(0:2)
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(get_bounds(sq), list())

  sc <- new("EffectScenario", param.bounds=list(foo=c(0, 1)))
  suppressWarnings(sq <- sequence(list(sc)))
  expect_equal(get_bounds(sq), list(foo=c(0, 1)))
})

test_that("caliset", {
  sc <- new("EffectScenario", param.bounds=list())
  cs <- caliset(sc, data.frame(t=0, n=0))
  expect_equal(get_bounds(cs), list())

  sc <- new("EffectScenario", param.bounds=list(foo=c(0, 1)))
  cs <- caliset(sc, data.frame(t=0, n=0))
  expect_equal(get_bounds(cs), list(foo=c(0, 1)))
})

test_that("invalid arguments", {
  expect_error(get_bounds(), "missing")
  expect_error(get_bounds(NULL), "must be a scenario")
  expect_error(get_bounds(NA), "must be a scenario")
  sc <- new("EffectScenario")
  expect_error(get_bounds(list(sc, sc)), "length one")
})

########################
## Class definition
########################

#' @export
#' @include man-deb.R
setClass("DebAbj", contains="Deb")


########################
## Constructor
########################

#' DEB_abj
#'
#' Creates a *DEB abj* scenario. The *abj* model with type M acceleration is
#' like model *std*, but acceleration occurs between birth and metamorphosis (V1-morph).
#' Isomorphy is assumed before and after acceleration. Metamorphosis is before
#' puberty and occurs at maturity `E_Hj`, which might or might not correspond with
#' changes in morphology. The *abj* model is a one-parameter extension of model *std*
#' [(DEB Wiki)](https://debportal.debtheory.org/docs/Typified_models.html).
#'
#' ## State variables
#'
#' The following list describes the default names and standard units of the model's
#' state variables:
#' * `L`, structural length (cm)
#' * `E`, energy reserve (J)
#' * `H`, energy invested in maturity (J)
#' * `R`, reproduction buffer (J)
#' * `cV`, internal concentration ([C])
#' * `Lmax`, maximum structural length (cm)
#'
#' All state variables are initialized with zero. See [set_init()] on how to set
#' the initial state.
#'
#' ## Parameters
#'
#' The following model parameters are required:
#' * `p_M`, vol-spec somatic maintenance (J/d.cm^3)
#' * `v`, energy conductance (cm/d)
#' * `k_J`, maturity maint rate coefficient (1/d)
#' * `p_Am`, surface-area specific maximum assimilation rate (J/d.cm^2)
#' * `kap`, allocation fraction to soma (-)
#' * `E_G`, spec cost for structure (J/cm^3)
#' * `f`, scaled functional response (-)
#' * `E_Hj`, maturity at metamorphosis (J)
#' * `E_Hp`, maturity at puberty (J)
#' * `kap_R`, reproduction efficiency (-)
#' * `L_b`, structural length at birth (cm)
#' * `L_j`, structural length at metamorphosis (cm)
#' * `ke`, elimination rate constant (d-1)
#' * `c0`, no-effect concentration sub-lethal ([C])
#' * `cT`, tolerance concentration ([C])
#' * `MoA`, mode of action switch (-)
#'
#' ## Mode of Actions
#'
#' Any combination of the following mode of actions (MoA) can be considered by
#' the model:
#' * `MoA = 1`: effect on feeding
#' * `MoA = 2`: effect on maintenance costs
#' * `MoA = 4`: effect on overhead costs for making an egg
#' * `MoA = 8`: hazard during oogenesis
#' * `MoA = 16`: energy conductance
#'
#' To activate more than one MoA, simply add up the corresponding
#' codes. To disable all MoAs, set the parameter to zero.
#' See also [set_mode_of_action()].
#'
#' ## Effects
#'
#' The state variables *L* (structural length) and *R* (reproduction buffer) are
#' set as effect endpoints by default. All state variables are available as
#' potential endpoints. The list of considered endpoints can be modified
#' by using [set_endpoints()].
#'
#' To calculate effects, each *DEB* scenario is simulated twice: One simulation
#' which considers exposure to a toxicant and one simulation without exposure, i.e.
#' a control. See also [effect()].
#'
#' @section Simulation output:
#' Simulation results will contain the state variables.
#' It is possible to amend the output of [simulate()] with additional model
#' quantities that are not state variables, for e.g. debugging purposes or to
#' analyze model behavior. To enable or disable additional outputs, use the
#' optional argument `nout` of [simulate()]. As an example, set `nout=2` to
#' enable reporting of the acceleration factor (`MV`) and the mobilization flux
#' (`pC`). Set `nout=0` to disable additional outputs (default).
#'
#' The available output levels are as follows:
#' - `nout` >= 1: `MV` acceleration factor (-)
#' - `nout` >= 2: `pC` mobilization flux (J/d)
#' - `nout` >= 3: `pA` assimilation flux (J/d)
#' - `nout` >= 4: `pJ` energy invested in maturity flux (J/d)
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' @return an S4 object of type [DebAbj-class]
#' @export
#' @family DEB models
#' @aliases DebAbj-class
#' @examples
#' # Create an abj scenario from scratch and simulate it
#' DEB_abj() %>%
#'   set_init(c(L=0.02,E=0.1,H=0.01)) %>%
#'   set_param(c(p_M=3000,v=0.02,k_J=0.6,p_Am=300,kap=0.9,E_G=4000,f=1,
#'               E_Hj=0.05,E_Hp=0.3,kap_R=0.9,ke=1,c0=0,cT=1,L_b=0.02,
#'               L_j=0.04,MoA=0)) %>%
#'   set_noexposure() %>%
#'   set_times(0:10) %>%
#'   simulate()
#'
#' # Print information about sample scenario 'americamysis'
#' americamysis
#'
#' # Simulate 'americamysis' scenario
#' americamysis %>% simulate()
DEB_abj <- function() {
  new("DebAbj",
      name="DEB_abj",
      param.req=c("p_M","v","k_J","p_Am","kap","E_G","f","E_Hj","E_Hp","kap_R",
                  "ke","c0","cT","L_b","L_j","MoA"),
      endpoints=c("L","R"),
      control.req=TRUE,
      init=c("L"=0,"E"=0,"H"=0,"R"=0,"cV"=0,"Lmax"=0),
      exposure=no_exposure()
  ) %>% set_noexposure()
}


########################
## Simulation
########################

#' @importFrom deSolve ode
solver_deb_abj <- function(scenario, method="lsoda", ...) {
  # make sure that parameters are present and in required order
  params.req <- c("p_M","v","k_J","p_Am","kap","E_G","f","E_Hj","E_Hp","kap_R","ke","c0",
                  "cT","L_b","L_j","MoA")
  # additional output variables
  outnames <- c("MV","pC","pA","pJ")

  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # check for missing parameters
  params.missing <- setdiff(params.req, names(params))
  if(length(params.missing)>0)
    stop(paste("missing parameters:",paste(params.missing,sep=",",collapse=",")))

  # reorder parameters for deSolve
  params <- params[params.req]

  # create forcings list
  forcings <- list(scenario@exposure@series)

  # run solver
  ode(y=scenario@init, times=scenario@times, parms=params, method=method,
        dllname="cvasi", initfunc="deb_abj_init", func="deb_abj_func",
        initforc="deb_abj_forc", forcings=forcings, outnames=outnames, ...)
}
#' @include solver.R
#' @describeIn solver Numerically integrates DEB_abj scenarios
setMethod("solver", "DebAbj", function(scenario, ...) solver_deb_abj(scenario, ...))


########################
## Effects
########################

# uses the default fx implementation, nothing more required

# TODO output messages for fitted parameter(s), starting values, output variable
#      reuse outputs from calibrate()


#' Fit TK/TD parameters
#'
#' `r lifecycle::badge("experimental")` <br>
#' High-level function to fit TK/TD parameters to observed data. It eases the
#' use of [calibrate()] by providing sensible defaults for various (model dependent)
#' settings.
#'
#' ### Data
#'
#' The function can be used in three basic ways: Fit parameters of
#'
#' 1) a single [scenario] to a data set
#' 2) a list of [caliset]s
#' 3) a list of [caliset]s in a chain of `fit` functions
#'
#' For option 1), scenario and data are supplied separately. In this case,
#' supported types of argument `data` are `data.frame`s and [tox_data] objects.
#' Any `data.frame` must have a format that is compatible with the [tox_data()]
#' function.
#'
#' In option 2), all conditions are fully described by a list of calibration sets.
#' The user is responsible to set up all [caliset] objects to their needs.
#'
#'
#' Option 3) is for convenience purposes and allows the chaining of fit functions.
#' The latter alternative accepts the return value of e.g. [fit_growth()]
#' as argument `x` and the fitted parameter values are applied to all
#' calibration sets in argument `data`.
#'
#' @param x a [scenario] or a list of [caliset] objects
#' @param data toxicological trial data to fit TK/TD parameters to: required if
#'      `x` is either a [scenario] or the result of a fit. See Section *Data*
#'      for details.
#' @param par named vector, of parameters to fit and their starting values
#' @param log_scale logical, if `TRUE` then fitting will be performed on log-transformed
#'   observations and predictions, else the data will be used as-is
#' @param verbose logical, if `TRUE` then info messages are printed to the console
#' @param ... additional arguments passed through to [calibrate()]
#' @aliases fit_tktd,ANY,ANY-method fit_tktd,list,missing-method fit_tktd,cvasi_fit,ANY-method
#'    fit_tktd,CalibrationSet,ANY-method fit_tktd,CalibrationSet,missing-method
#'    fit_tktd,EffectScenario,ANY-method fit_tktd,ScenarioSequence,ANY-method
#' @returns a list
#' @export
#' @examples
#' # Use experimental data from control trial ('T0') to fit growth rate
#' ctrl <- schmitt2013[schmitt2013$trial == "T0" ,]
#'
#' # Set up a scenario, provide dummy parameter values where necessary
#' sc <- Lemna_Schmitt() %>%
#'   set_init(c(BM=0.0012)) %>%
#'   set_param(c(EC50=1, b=1, P_up=1))
#'
#' # Run fitting routine
#' fit_growth(sc, data=ctrl, verbose=TRUE)
#'
#' # Use fitted growth parameter to adapt scenario
#' sc2 <- sc %>% set_param(c(k_phot_max=0.43925))
#'
#' # Use experimental ecotox data for various concentrations of 'metsulfuron-methyl'
#' trials <- schmitt2013[schmitt2013$trial != "T0" ,]
#'
#' # Fit remaining TK/TD parameters
#' fit_tktd(sc2, data=trials, verbose=TRUE)

setGeneric("fit_tktd", function(x, data, ...) standardGeneric("fit_tktd"))

# Default handler
#' @noRd
setMethod("fit_tktd", c("ANY", "ANY"), function(x, data, ...) stop("Object type or argument combination not supported"))

# Wrapper for [sequence]s
#' @noRd
setMethod("fit_tktd", c("ScenarioSequence", "ANY"), function(x, data, ...) fit_tktd_scenario(x=x, data=data, ...))

# Wrapper for [scenario]s
#' @noRd
setMethod("fit_tktd", c("EffectScenario", "ANY"), function(x, data, ...) fit_tktd_scenario(x=x, data=data, ...))

# Wrapper for list of caliset objects
#' @noRd
setMethod("fit_tktd", c("list", "missing"), function(x, data, ...) fit_tktd_list(x, ...))

#' @noRd
setMethod("fit_tktd", c("CalibrationSet", "missing"), function(x, data, ...) fit_tktd_caliset(x=x, ...))

# Make user aware of invalid argument combination
#' @noRd
setMethod("fit_tktd", c("CalibrationSet", "ANY"), function(x, data, ...)
  stop("Calisets (argument 'x') cannot be used together with argument 'data'")
)

#' @noRd
setOldClass("cvasi_fit")

# @describeIn fit_tktd Wrapper for chaining fit_* functions
#' @noRd
setMethod("fit_tktd", c("cvasi_fit", "ANY"), function(x, data, ...) fit_tktd_fit(x=x, data=data, ...))

# Wrapper for single scenario + dataset
fit_tktd_scenario <- function(x, data, ...) {
  # if scenario type not supported yet, we may inadvertently land here
  if(is.list(x) & all(is_caliset(x))) {
    stop("Scenario type not supported yet")
  }

  # normal call handling starts here...
  if(missing(data)) {
    stop("Argument 'data' is missing")
  }
  # check if dataset is represented by tabular data
  if(is.data.frame(data)) {
    data <- tox_data(data)
  }
  else if(is(data, "ToxData")) {
    # nothing to do
  } else {
    stop("Argument 'data' has unsupported type")
  }
  # create calisets
  x <- td2cs(x, data)
  fit_tktd(x, ...)
}

# Wrapper for single caliset objects
fit_tktd_caliset <- function(x, ...) {
  fit_tktd(list(x), ...)
}

# Wrapper for cvasi_fit objects, allows chaining of fit* functions, such as
# scenario %>%
#  fit_growth(data=calisets1) %>%
#  fit_tktd(data=calisets2)
fit_tktd_fit <- function(x, data, ...) {
  if(missing(data)) {
    stop("Argument 'data' is missing")
  }

  # check data type, only works for (list of) calisets
  if(!all(is_caliset(data))) {
    stop("Argument 'data' must only contain caliset objects")
  }
  if(!is.vector(data)) {
    data <- list(data)
  }
  # apply fitted parameter(s) to all scenarios
  for(i in seq_along(data)) {
    data[[i]]@scenario <- data[[i]]@scenario %>% set_param(x$par)
  }
  # continue with normal fitting
  fit_tktd_list(data)
}

# Wrapper for lists of calisets
#' @importFrom methods selectMethod
fit_tktd_list <- function(x, ...) {
  # check that list only contains caliset object
  if(!all(is_caliset(x))) {
    stop("Argument 'x' must only contain caliset objects")
  }
  # then use scenario in first caliset to decide which method to call
  sc <- x[[1]]@scenario
  if(is_sequence(x[[1]]@scenario))
    sc <- sc[[1]]

  cls <- class(sc)[1]
  fun <- selectMethod("fit_tktd", signature=c(x=cls, data="missing"))
  fun(x, ...)
}

#' @describeIn fit_tktd Fit TK/TD parameters of [Lemna_SETAC] scenarios
#' @include model-lemna_schmitt.R
#' @include model-lemna_setac.R
setMethod("fit_tktd", c("LemnaSetac", "missing"), function(x, data, par, log_scale=TRUE, verbose=FALSE, ...)
  fit_tktd_lemna_setac(x, data=data, par=par, log_scale=log_scale, verbose=verbose, ...))
#' @describeIn fit_tktd Fit TK/TD parameters of [Lemna_Schmitt] scenarios
setMethod("fit_tktd", c("LemnaSchmitt", "missing"), function(x, data, par, log_scale=TRUE, verbose=FALSE, ...)
  fit_tktd_lemna_schmitt(x, data=data, par=par, log_scale=log_scale, verbose=verbose, ...))

#' @autoglobal
fit_tktd_lemna_setac <- function(x, par, data, log_scale, verbose=TRUE,
                                 lower=NULL, upper=NULL, ...) {
  # TODO this is not a good solution, as every model-dep function must check this.
  #      better to not pass arguments through to calibrate() in the future
  if(!is.null(lower)) {
    stop("Argument 'lower' is not allowed, use `set_bounds()` instead")
  }
  if(!is.null(upper)) {
    stop("Argument 'upper' is not allowed, use `set_bounds()` instead")
  }

  if(any(is_scenario(x))) {
    stop("Argument 'x' is a scenario and must be combined with argument 'data'")
  }
  if(!missing(data)) {
    stop("Argument 'data' not supported here")
  }
  # is it a list of calisets?
  if(is.vector(x)) {
    if(!all(is_caliset(x))) {
      stop("Argument 'x' must only contain calisets")
    }
  }

  for(i in seq_along(x)) {
    # check if all datasets start at time=0, which is an implicit assumption in
    # the following calculation of statistics etc
    if(min(x[[1]]@data[, 1]) != 0) {
      stop("Scenarios/datasets must start at time zero (t=0)")
    }
  }

  # if no starting value set, use default values
  if(missing(par)) {
    par <- c(EC50_int=1, b=2, P=0.1)
  }
  if(is.list(par)) {
    par <- unlist(par)
  }
  # check for invalid values
  if(any(is.na(par) | is.nan(par) | is.infinite(par))) {
    stop("Starting values contain invalid values")
  }

  # determine variable to fit on
  output_nm <- "FrondNo"

  # run fitting routine
  fit <- calibrate(x, par=par, output=output_nm, hessian=TRUE, log_scale=log_scale,
                   err_fun="sse", verbose=verbose, ...)

  if(length(fit$par) == 0) {
    warning("Optimized parameters missing, cannot compute residuals")
    return(fit)
  }

  # remove data points at time==0, as these are the initial conditions and not predictions
  df <- fit$data %>% dplyr::filter(time > 0)

  ## Calculate Statistics
  fit$shapiro <- stats::shapiro.test(fit$data$residuals)

  # Degrees of Freedom: number of data points minus number of fitted parameters, excluding t=0
  fit$dof <- nrow(df) - length(par)

  # limits of the 95% confidence interval
  # TODO likelihood profiling as alternative
  fit$ci <- ci_from_hessian(fit, dof=fit$dof, level=0.95)

  # NRMSE
  fit$nrmse <-  1 / mean(df[, 2]) * sqrt(1 / nrow(df) * sum((df[, 2] - df[, 3])^2))

  fit
}

fit_tktd_lemna_schmitt <- function(x, par, ...) {
  if(missing(par)) {
    par <-  c(EC50=1, b=2, P_up=0.1)
  }
  fit_tktd_lemna_setac(x=x, par=par, ...)
}

#' @describeIn fit_tktd Fit TK/TD parameters of [Magma] scenarios
#' @include model-magma.R
setMethod("fit_tktd", c("Magma", "missing"), function(x, data, par, log_scale=TRUE, verbose=FALSE, ...)
  fit_tktd_magma(x, data=data, par=par, log_scale=log_scale, verbose=verbose, ...))

#' @autoglobal
fit_tktd_magma <- function(x, par, data, log_scale, verbose=TRUE, lower=NULL, upper=NULL, ...) {
  # TODO this is not a good solution, as every model-dep function must check this.
  #      better to not pass arguments through to calibrate() in the future
  if(!is.null(lower)) {
    stop("Argument 'lower' is not allowed, use `set_bounds()` instead")
  }
  if(!is.null(upper)) {
    stop("Argument 'upper' is not allowed, use `set_bounds()` instead")
  }

  if(any(is_scenario(x))) {
    stop("Argument 'x' is a scenario and must be combined with argument 'data'")
  }
  if(!missing(data)) {
    stop("Argument 'data' not supported here")
  }
  # is it a list of calisets?
  if(is.vector(x)) {
    if(!all(is_caliset(x))) {
      stop("Argument 'x' must only contain calisets")
    }
  }

  for(i in seq_along(x)) {
    # check if all datasets start at time=0, which is an implicit assumption in
    # the following calculation of statistics etc
    if(min(x[[1]]@data[, 1]) != 0) {
      stop("Scenarios/datasets must start at time zero (t=0)")
    }
  }

  # if no starting value set, use default values
  if(missing(par)) {
    par <- c(EC50_int=1, E_max=1, b=1)
  }
  if(is.list(par)) {
    par <- unlist(par)
  }
  # check for invalid values
  if(any(is.na(par) | is.nan(par) | is.infinite(par))) {
    stop("Starting values contain invalid values")
  }

  # determine variable to fit on
  output_nm <- "FrondNo"

  # run fitting routine
  fit <- calibrate(x, par=par, output=output_nm, hessian=TRUE, log_scale=log_scale,
                   err_fun="sse", verbose=verbose, ...)

  if(length(fit$par) == 0) {
    warning("Optimized parameters missing, cannot compute residuals")
    return(fit)
  }

  # remove data points at time==0, as these are the initial conditions and not predictions
  df <- fit$data %>% dplyr::filter(time > 0)

  ## Calculate Statistics
  fit$shapiro <- stats::shapiro.test(fit$data$residuals)

  # Degrees of Freedom: number of data points minus number of fitted parameters, excluding t=0
  fit$dof <- nrow(df) - length(par)

  # limits of the 95% confidence interval
  # TODO likelihood profiling as alternative
  fit$ci <- ci_from_hessian(fit, dof=fit$dof, level=0.95)

  # NRMSE
  fit$nrmse <-  1 / mean(df[, 2]) * sqrt(1 / nrow(df) * sum((df[, 2] - df[, 3])^2))

  fit
}

########################
## Organizing man pages
########################

#' Lemna models
#'
#' Overview of supported *Lemna* models
#'
#' - [Lemna_Schmitt()] by Schmitt *et al.* (2013)
#' - [Lemna_SETAC()] by Klein *et al.* (2021)
#'
#' @inheritSection Transferable Biomass transfer
#' @name Lemna-models
#' @family Lemna models
#' @family models
#' @seealso [Macrophyte-models]
#' @aliases Lemna-class
NULL

# Generic Lemna class
#' @include class-EffectScenario.R class-Transferable.R
#' @export
setClass("Lemna", contains=c("Transferable", "EffectScenario"))

########################
## Class definition
########################

# Lemna model class (Klein et al. 2021)
#' @export
#' @include man-lemna.R
setClass("LemnaSetac", contains="Lemna")
# for backwards compatibility
#' @export
setClass("LemnaSetacScenario", contains="LemnaSetac")


########################
## Constructor
########################

#' Lemna model v1.3 (Klein et al. 2021)
#'
#' The model was described and published by the SETAC Europe Interest Group
#' Effect Modeling (Klein et al. 2021). It is based on the *Lemna* model
#' by Schmitt (2013). The model is a mechanistic combined
#' toxicokinetic-toxicodynamic (TK/TD) and growth model for the aquatic
#' macrophytes *Lemna spp.*. The model simulates the development of Lemna biomass
#' under laboratory and environmental conditions. Growth of the Lemna population
#' is simulated on basis of photosynthesis and respiration rates which are
#' functions of environmental conditions. The toxicodynamic sub-model describes
#' the effects of growth-inhibiting substances by a respective reduction in the
#' photosynthesis rate based on internal concentrations.
#'
#' @section State variables:
#' The model has two state variables:
#' - `BM`, Biomass (g dw)
#' - `M_int`, Mass of toxicant in plant population (ng)
#'
#' The  units of the state variables `BM` and `M_int` are to be interpreted
#' as *per vessel* in laboratory tests, and as *per square meter* for field populations.
#' The same interpretation applies to all parameters relating biomass and toxicant mass to
#' other entities.
#'
#' @section Model parameters:
#' - Growth model
#'   - `k_photo_fixed`, Model switch for unlimited growth conditions (TRUE/FALSE)
#'   - `k_photo_max`, Maximum photosynthesis rate (d-1)
#'   - `k_loss`, Reference loss rate (d-1)
#'   - `BM_threshold`, Lower biomass abundance threshold,  (g dw)
#'   - `BM_min`, Reservoir for biomass recovery,  (g dw)
#'
#' - Temperature response of photosynthesis
#'   - `T_opt`, Optimum growth temperature (°C)
#'   - `T_min`, Minimum growth temperature (°C)
#'   - `T_max`, Maximum growth temperature (°C)
#'
#' - Temperature response of biomass loss rate
#'   - `Q10`, Temperature coefficient (-)
#'   - `T_ref`, Reference temperature for response=1 (°C)
#'
#' - Irradiance reponse of photosynthesis
#'   - `alpha`, Slope of irradiance response (m2 d kJ-1)
#'   - `beta`, Intercept of irradiance response (-)
#'
#' - Nutrient response of photosynthesis
#'   - `N_50`, Half-saturation constant of Nitrogen (mg N L-1)
#'   - `P_50`, Half-saturation constant of Phosphorus (mg P L-1)
#'
#' - Density dependence of photosynthesis
#'   - `BM_L`, Carrying capacity (g dw)
#'
#' - Concentration response (Toxicodynamics)
#'   - `EC50_int`, Internal concentration resulting in 50% effect (ug L-1)
#'   - `E_max`, Maximum inhibition (-)
#'   - `b`, Slope parameter (-)
#'
#' - Internal concentration (Toxicokinetics)
#'   - `P`, Permeability (cm d-1)
#'   - `r_A_DW`, Area per dry-weight ratio (cm2 g-1)
#'   - `r_FW_DW`, Fresh weight per dry weight ratio (-)
#'   - `r_FW_V`, Fresh weight density (g cm-3)
#'   - `r_DW_FN`, Dry weight per frond ratio  (g dw)
#'   - `K_pw`, Partitioning coefficient plant:water (-)
#'   - `k_met`, Metabolisation rate (d-1)
#'
#' @section Forcings:
#' External concentrations, alias *exposure*,  have to be provided as *ug L-1*.
#'
#' Besides exposure, the model requires four environmental properties as
#' input:
#' - `tmp`, ambient temperature (°C)
#' - `irr`, irradiance (kJ m-2 d-1)
#' - `P`, Phosphorus concentration (mg P L-1)
#' - `N`, Nitrogen concentration (mg N L-1)
#'
#' The following constant default values are used for these properties:
#' - `tmp` = 12 °C
#' - `irr` = 15,000 kJ m-2 d-1
#' - `P` = 0.3 mg L-1
#' - `N` = 0.6 mg L-1
#'
#' Forcings time-series are represented by `data.frame` objects consisting of two
#' columns. The first for time and the second for the environmental factor in question.
#'
#' Entries of the `data.frame` need to be ordered chronologically. A time-series
#' can consist of only a single row; in this case it will represent constant
#' environmental conditions. See [scenarios] for more details.
#'
#' @section Effects:
#' Supported effect endpoints include *BM* (biomass) and *r* (average
#' growth rate during simulation). The effect on biomass is calculated from
#' the last state of a simulation. Be aware that endpoint *r* is incompatible
#' with biomass transfers.
#'
#' @section Simulation output:
#' For reasons of convenience, the return value of [simulate()] contains by
#' default two additional variables derived from simulation results:
#' the internal concentration `C_int` as well as the number of fronds `FrondNo`.
#' These can be disabled by setting the argument `nout = 0`.
#'
#' The available output levels are as follows:
#' - `nout >= 1`: `C_int`, internal concentration (mass per volume)
#' - `nout >= 2`: `FrondNo`, frond number (-)
#' - Response functions
#'   - `nout >= 3`: `f_loss`, respiration dependency function (-)
#'   - `nout >= 4`: `f_photo`, photosynthesis dependency function (-)
#'   - `nout >= 5`: `fT_photo`, temperature response of photosynthesis (-)
#'   - `nout >= 6`: `fI_photo`, irradiance response of photosynthesis (-)
#'   - `nout >= 7`: `fP_photo`, phosphorus response of photosynthesis (-)
#'   - `nout >= 8`: `fN_photo`, nitrogen response of photosynthesis (-)
#'   - `nout >= 9`: `fBM_photo`, density response of photosynthesis (-)
#'   - `nout >= 10`: `fCint_photo`, concentration response of photosynthesis (-)
#' - Environmental variables
#'   - `nout >= 11`: `C_int_unb`, unbound internal concentration (mass per volume)
#'   - `nout >= 12`: `C_ext`, external concentration (mass per volume)
#'   - `nout >= 13`: `Tmp`, temperature (deg C)
#'   - `nout >= 14`: `Irr`, irradiance (kJ m-2 d-1)
#'   - `nout >= 15`: `Phs`, Phosphorus concentration (mg P L-1)
#'   - `nout >= 16`: `Ntr`, Nitrogen concentration (mg N L-1)
#' - Derivatives
#'   - `nout >= 17`: `dBM`, biomass derivative (g dw m-2 d-1)
#'   - `nout >= 18`: `dM_int`, mass of toxicant in plants derivative (mass per m2 d-1)
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `hmax = 0.1`<br>
#'    Maximum step length in time suitable for most simulations.
#'
#' @inheritSection Transferable Biomass transfer
#'
#' @section Model history and changes:
#' - cvasi v1.0.0
#'   - Model and parameters as described by Klein et al. (2022) report version 1.1
#' - cvasi v1.5.0
#'   - Default value of parameter `beta` modified due to typo in previous report
#'     versions, now conforms with Klein et al. (2025), report version 1.2
#'   - New value: `beta=0.25`, old value: `beta=0.025`
#'
#' @references
#' Klein J., Cedergreen N., Heine S., Reichenberger S., Rendal C.,
#' Schmitt W., Hommen U., 2021: *Refined description of the Lemna TKTD growth model
#' based on Schmitt et al. (2013) - equation system and default parameters*.
#' Report of the working group *Lemna* of the SETAC Europe Interest Group Effect
#' Modeling. Version 1.3, uploaded in July 2025.
#' https://www.setac.org/group/effect-modeling.html
#'
#'
#' Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013:
#' *Mechanistic TK/TD-model simulating the effect of growth inhibitors on
#' Lemna populations*. Ecol Model 255, pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
#'
#' @return an S4 object of type [LemnaSetac-class]
#' @seealso [Lemna-models], [Macrophyte-models], [Transferable], [Scenarios]
#' @family Lemna models
#' @family macrophyte models
#' @aliases LemnaSetacScenario-class LemnaSetac-class
#' @export
Lemna_SETAC <- function() {
  new("LemnaSetac",
      name="Lemna_SETAC",
      # Recommended parameter values according to Klein et al. (2021)
      param=list(
        k_photo_fixed=TRUE, # model switch, if TRUE then f_loss = 1 and f_photo = fCint_photo()
        k_photo_max = 0.47,    # max photosynthesis rate (d-1)
        k_loss = 0.05,         # reference loss rate (d-1)
        BM_min = 5e-4,         # threshold density for setting dBM/dt to zero (g dw m-2)

        # response parameters
        T_opt = 26.7,   # optimum growth temperature (deg C)
        T_min = 8,      # minimum growth temperature (deg C)
        T_max = 40.5,   # maximum growth temperature (deg C)
        Q10 = 2,        # temperature coefficient (-)
        T_ref = 25,     # ref temperature for response=1 (°C)
        alpha = 5e-5,   # slope of irradiance response of photosynthesis (m2 d kJ-1)
        beta = 0.25,    # intercept of irradiance response of photosynthesis (-)
        N_50 = 0.034,   # half-saturation constant of Nitrogen response (mg N L-1)
        P_50 = 0.0043,  # half-saturation constant of Phosphorus response (mg P L-1)
        BM_L = 177,     # carrying capacity (g dw m-2)

        # toxicodynamic parameters
        E_max = 1,      # maximum inhibition (-), substance specific
        EC50_int = NA,  # int. conc. resulting in 50% effect (ug L-1), substance specific
        b = NA,         # slope parameter (-), substance specific

        # toxicokinetic parameters
        P = NA,         # permeability (cm d-1), substance specific
        r_A_DW = 1000,  # area per dry-weight ratio (cm2 g-1)
        r_FW_DW = 16.7, # fresh weight per dry weight ratio (-)
        r_FW_V = 1,     # fresh weight density (g cm-3)
        r_DW_FN = 1e-4, # dry weight per frond ratio (g dw)
        K_pw = 1,       # partitioning coefficient plant:water (-), generally substance specific
        k_met = 0),
      # parameters required for the model to run
      param.req=c("k_photo_fixed", "k_photo_max", "k_loss", "BM_min", "T_opt",
                  "T_min", "T_max", "Q10", "T_ref", "alpha", "beta", "N_50",
                  "P_50", "BM_L", "E_max", "EC50_int", "b", "P", "r_A_DW",
                  "r_FW_DW", "r_FW_V", "r_DW_FN", "K_pw", "k_met"),
      # boundary presets defined by expert judgement
      param.bounds=list(k_photo_max=c(0, 1), EC50_int=c(0, 1e6), b=c(0.1, 20),
                        P=c(0, 100)),
      endpoints=c("BM", "r"),
      forcings.req=c("tmp", "irr", "P", "N"),
      forcings=list(tmp=data.frame(time=0, tmp=12),
                    irr=data.frame(time=0, irr=15000),
                    P=data.frame(time=0, P=0.3),
                    N=data.frame(time=0, N=0.6)
      ),
      control.req=TRUE,
      init=c(BM=0.0012, M_int=0),
      transfer.comp.biomass="BM",
      transfer.comp.scaled="M_int"
  ) %>% set_noexposure()
}


########################
## Simulation
########################

# Numerically integrate Lemna_SETAC scenarios
#
# @param scenario an EffectScenario object
# @param nout number of additional output variables, see [deSolve::ode()]
# @param method numerical integration method, see [deSolve::ode()]
# @param hmax numeric, set max step length in time, defaults to `0.1`
# @param ... additional parameters passed on to [deSolve::ode()]
#
# @return data.frame
#' @importFrom deSolve ode
solver_lemna_setac <- function(scenario, nout=2, method="lsoda", hmax=0.1, ...) {
  params <- scenario@param
  if(is.list(params)) params <- unlist(params)

  # check for missing parameters
  params.missing <- setdiff(scenario@param.req, names(params))
  if(length(params.missing)>0)
    stop(paste("missing parameters:", paste(params.missing, sep=",", collapse=", ")))

  # reorder parameters for deSolve
  params <- params[scenario@param.req]
  # check if any parameter is negative
  if(any(params < 0, na.rm=TRUE))
    stop(paste("parameter out of bounds:",paste(names(params)[which(params < 0)], collapse=", ")))
  if(any(is.na(params)))
    stop(paste("parameter missing:",paste(names(params)[which(is.na(params))], collapse=", ")))

  envir <- scenario@forcings
  # Environmental factors are irrelevant in case of unlimited growth, removing any
  # superfluous time-series data will reduce simulation overhead/runtime
  if(params["k_photo_fixed"]) {
    forcings <- list(scenario@exposure@series,
                     data.frame(time=0, tmp=0),
                     data.frame(time=0, irr=0),
                     data.frame(time=0, P=0),
                     data.frame(time=0, N=0))
  } else {
    forcings <- list(scenario@exposure@series, envir$tmp, envir$irr, envir$P, envir$N)
  }

  # set names of additional output variables
  outnames <- c("C_int", "FrondNo", "f_loss", "f_photo", "fT_photo", "fI_photo",
                "fP_photo", "fN_photo", "fBM_photo", "fCint_photo", "C_int_unb",
                "C_ext", "Tmp", "Irr", "Phs", "Ntr", "dBM", "dM_int")
  # run deSolve
  ode(y=scenario@init, times=scenario@times, parms=params, dllname="cvasi",
        initfunc="lemna_setac_init", func="lemna_setac_func", initforc="lemna_setac_forc",
        forcings=forcings, nout=nout, outnames=outnames, method=method, hmax=hmax, ...)
}
#' @include solver.R
#' @describeIn solver Numerically integrates Lemna_SETAC scenarios
setMethod("solver", "LemnaSetac", function(scenario, ...) solver_lemna_setac(scenario, ...) )


########################
## Effects
########################

# Calculate effect of Lemna scenario
fx_lemna <- function(scenario, ...) {
  efx_r <- "r" %in% scenario@endpoints
  # TODO move to a validate_scenario function, this takes precious time on every effect() call
  if(efx_r & has_transfer(scenario))
    stop("endpoint r is incompatible with biomass transfers")

  out <- simulate(scenario, ...)
  if(!num_success(out)) {
    return(NA)
  }

  efx <- c("BM"=tail(out$BM, 1))
  if(efx_r) # we skip the log() operation if we can
    efx["r"] <- log(tail(out$BM,1) / out$BM[1]) / (tail(out[,1],1) - out[1,1])

  efx
}

#' @include fx.R
#' @describeIn fx Effect at end of simulation of [Lemna-models]
setMethod("fx", "Lemna", function(scenario, ...) fx_lemna(scenario, ...))

# converts a scenario object defined by the `lemna` package to a scenario
# of the `cvasi` package
lemna2scenario <- function(sc) {
  if(!is(sc, "lemna_scenario"))
    stop("argument is not a lemna scenario")

  forc <- sc$envir
  forc$conc <- NULL

  Lemna_SETAC() %>%
    set_init(sc$init) %>%
    set_param(sc$param) %>%
    set_exposure(sc$envir$conc) %>%
    set_times(sc$times) %>%
    set_forcings(forc)
}

# Make sure that simulation results of this package and 'lemna' are identical,
# the 'lemna' package vouches for correct model implementation
test_that("lemna::lemna, k_phot0_fixed=FALSE", {
  skip_on_os("mac") # macos numerics always deviate in some unforeseeable way
  skip_if_not_installed("lemna")

  origR <- lemna::lemna(lemna::focusd1, times=0:20, hmax=0.01, ode_mode="r")
  origC <- lemna::lemna(lemna::focusd1, times=0:20, hmax=0.01, ode_mode="c")

  lemna::focusd1 %>%
    lemna2scenario() %>%
    set_times(0:20) %>%
    simulate(hmax=0.01) -> out

  expect_equal(out, origR, ignore_attr=TRUE, tolerance=1e-3)
  expect_equal(out, origC, ignore_attr=TRUE, tolerance=1e-8)
})

test_that("lemna::lemna, k_phot0_fixed=TRUE", {
  skip_on_os("mac") # macos numerics always deviate in some unforeseeable way
  skip_if_not_installed("lemna")

  lemna_sc <- lemna::focusd1
  lemna_sc$param$k_photo_fixed <- 1
  origR <- lemna::lemna(lemna_sc, times=0:20, hmax=0.01, ode_mode="r")
  origC <- lemna::lemna(lemna_sc, times=0:20, hmax=0.01, ode_mode="c")

  lemna2scenario(lemna_sc) %>%
    set_times(0:20) %>%
    simulate(hmax=0.01) -> out

  expect_equal(out, origR, ignore_attr=TRUE, tolerance=1e-3)
  expect_equal(out, origC, ignore_attr=TRUE, tolerance=1e-8)
})

########################
## Class definition
########################

# Lemna model class (Schmitt et al. 2013)
#' @export
#' @include man-lemna.R
setClass("LemnaSchmitt", contains="Lemna")
# for backwards compatibility
#' @export
setClass("LemnaSchmittScenario", contains="LemnaSchmitt")


########################
## Constructor
########################

#' Lemna model (Schmitt et al. 2013)
#'
#' The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
#' growth model for the aquatic macrophytes *Lemna spp.*
#' The model simulates the development of *Lemna* biomass under laboratory and
#' environmental conditions and was developed by Schmitt *et al.* (2013). Growth
#' of the *Lemna* population is simulated on basis of photosynthesis and respiration
#' rates which are functions of environmental conditions.
#' The toxicodynamic sub-model describes the effects of growth-inhibiting
#' substances by a respective reduction in the photosynthesis rate based on
#' internal concentrations. This is the historical version of the Lemna model.
#' For current uses, we recommend the Lemna (SETAC) model, which is a more recent
#' version of the Schmitt model.
#'
#' Constructors to ease creation of scenarios based on the *Lemna* model by
#' Schmitt *et al.* (2013).
#' A variant of this *Lemna* model, `Lemna_SchmittThold()`, provides an additional
#' cumulative exposure threshold parameter. The Lemna biomass stops growing
#' if the integral of exposure over time exceeds the threshold. The integral
#' of exposure is internally accounted for by an additional state variable
#' `AUC` (Area Under Curve).
#'
#' @section State variables:
#' The following list describes the default names and standard units of the model's
#' state variables:
#' * BM, g_dw/m2, dry weight biomass per square meter
#' * E, -, effect \[0,1\]
#' * M_int, ug, internal toxicant mass
#' * AUC, ug/L, cumulative exposure (**only** for `LemnaThreshold` model)
#'
#' Biomass (BM) and internal toxicant mass (M_int) are initialized to zero by
#' default. See [set_init()] on how to set the initial states.
#'
#' @section Model parameters:
#' The following model parameters are required:
#' * Fate and biomass
#'   * `k_phot_fix`, logical, TRUE then k_phot_max is not changed by environmental factors, else FALSE
#'   * `k_phot_max`, 1/d, maximum photosynthesis rate
#'   * `k_resp`, 1/d, respiration rate
#'   * `k_loss`, 1/d, rate of loss (e.g. flow rate)
#'   * `mass_per_frond`, g_dw/frond, dry weight per frond
#'   * `BMw2BMd`, g_fw/g_dw, Fresh weight/dry weight
#' * Effect
#'   * `Emax`, -, maximum effect \[0,1\]
#'   * `EC50`, ug/L, midpoint of effect curve
#'   * `b`, -, slope of effect curve
#' * Toxicokinetics
#'   * `P_up`, cm/d, Permeability for uptake
#'   * `AperBM`, cm2/g_dw, A_leaf / d_leaf = 1/d_leaf (for circular disc, d=0.05 cm)
#'   * `Kbm`, -, Biomass(fw) : water partition coefficient
#'   * `P_Temp`, logical, TRUE to enable temperature dependence of cuticle permeability, else FALSE
#'   * `MolWeight`, g/mol, Molmass of molecule (determines Q10_permeability)
#' * Temperature dependence
#'   * `Tmin`, deg C, minimum temperature for growth
#'   * `Tmax`, deg C, maximum temperature for growth
#'   * `Topt`, deg C, optimal temperature for growth
#'   * `t_ref`, deg C, reference temperature for respiration rate
#'   * `Q10`, -, temperature dependence factor for respiration rate
#' * Light dependence
#'   * `k_0`, 1/d, light dependence: intercept of linear part
#'   * `a_k`, (1/d)/(kJ/m2.d), light dependence: slope of linear part
#' * Phosphorus dependence (Hill like dep.)
#'   * `C_P`, mg/L, phosphorus concentration in water
#'   * `CP50`, mg/L, phosphorus conc. where growth rate is halfed
#'   * `a_p`, -, Hill coefficient
#'   * `KiP`, mg/L, p-inhibition constant for very high p-conc.
#' * Nitrogen dependence (Hill like dep.)
#'   * `C_N`, mg/L, nitrogen concentration in water
#'   * `CN50`, mg/L, n-conc. where growth rate is halfed
#'   * `a_N`, -, Hill coefficient
#'   * `KiN`, mg/L, n-inhibition constant for very high p-conc.
#' * Density dependence
#'   * `BM50`, g_dw/m2, cut off BM
#'
#' The `Lemna_SchmittThold` model requires the following additional parameter:
#' * `threshold`, ug/L, cumulative exposure threshold
#'
#' @section Forcings:
#' Besides exposure, the model requires four environmental properties as
#' time-series input:
#' - `temp`, temperature (°C)
#' - `rad`, global irradiation (kJ m-2 d-1)
#'
#' The following constant default values are used for these properties:
#' - `temp` = 12 °C
#' - `rad` = 15,000 kJ m-2 d-1
#'
#' Forcings time-series are represented by `data.frame` objects consisting of two
#' columns. The first for time and the second for the environmental factor in question.
#'
#' Entries of the `data.frame` need to be ordered chronologically. A time-series
#' can consist of only a single row; in this case it will represent constant
#' environmental conditions. See [scenarios] for more details.
#'
#' @section Effects:
#' Supported effect endpoints include *BM* (biomass) and *r* (average
#' growth rate during simulation). The effect on biomass is calculated from
#' the last state of a simulation. Be aware that endpoint *r* is incompatible
#' with frond transfers.
#'
#' @section Parameter boundaries:
#' Default values for parameter boundaries are set for all parameters by expert
#' judgement, for calibration purposes. Values can be access from the object, and
#' defaults overwritten.
#'
#' @section Simulation output:
#' Simulation results will contain two additional columns besides state variables:
#' * `C_int`, ug/L, internal concentration of toxicant
#' * `FrondNo`, -, number of fronds
#'
#' It is possible to amend the output of [simulate()] with additional model
#' quantities that are not state variables, for e.g. debugging purposes or to
#' analyze model behavior. To enable or disable additional outputs, use the
#' optional argument `nout` of [simulate()], see examples below. `nout=1`
#' enables reporting of internal concentration (C_int), `nout=14` enables all
#' additional outputs, and `nout=0` will disable additional outputs.
#'
#' The available output levels are as follows:
#' * `nout >= 1`: `C_int`, internal concentration (ug/L)
#' * `nout >= 2`: `FrondNo`, number of fronds (-)
#' * `nout >= 3`: `C_int_u`, unbound internal concentration (ug/l)
#' * Growth and TK/TD
#'   * `nout >= 4`: `BM_fresh`, fresh weight biomass (g_fw/m2)
#'   * `nout >= 5`: `k_photo_eff`, current photosynthesis rate (1/d)
#'   * `nout >= 6`: `k_resp_eff`, current respiration rate (1/d)
#'   * `nout >= 7`: `f_Eff`, toxic effect factor (-)
#'   * `nout >= 8`: `P_up_eff`, current permeability for uptake (cm/d)
#' * Environmental variables
#'   * `nout >= 9`: `actConc`, current toxicant concentration in surrounding medium (ug/L)
#'   * `nout >= 10`: `actTemp`, current environmental temperature (deg C)
#'   * `nout >= 11`: `actRad`, current environmental radiation (kJ/m2.d)
#' * Derivatives
#'   * `nout >= 12`: `d BM/dt`, current change in state variable BM
#'   * `nout >= 13`: `d E/dt`, current change in effect
#'   * `nout >= 14`: `d M_int/dt`, current change in internal toxicant mass
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `hmax = 0.1`<br>
#'    Maximum step length in time suitable for most simulations.
#'
#' @inheritSection Transferable Biomass transfer
#'
#' @param param optional named `list` or `vector` of model parameters
#' @param init optional named numeric `vector` of initial state values
#'
#' @return an S4 object of type [LemnaSchmitt-class]
#' @seealso [Lemna-models], [Macrophyte-models], [Transferable], [Scenarios]
#' @references
#' Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013:
#' *Mechanistic TK/TD-model simulating the effect of growth inhibitors on
#' Lemna populations*. Ecol Model 255, pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
#'
#' @family Lemna models
#' @family macrophyte models
#' @aliases LemnaSchmitt-class LemnaSchmittScenario-class
#' @export
Lemna_Schmitt <- function(param, init) {
  new("LemnaSchmitt",
      name="Lemna_Schmitt",
      param.req=c("Emax", "EC50", "b", "P_up", "AperBM", "Kbm", "P_Temp",
                  "MolWeight", "k_phot_fix", "k_phot_max", "k_resp", "k_loss",
                  "Tmin", "Tmax", "Topt", "t_ref", "Q10", "k_0", "a_k", "C_P",
                  "CP50", "a_P", "KiP", "C_N", "CN50", "a_N", "KiN", "BM50",
                  "mass_per_frond", "BMw2BMd"),
      # default values as defined by Schmitt et al. (2013)
      param=list(Emax=1, AperBM=1000, Kbm=1, P_Temp=FALSE,
                 MolWeight=390.4, k_phot_fix=TRUE, k_phot_max=0.47, k_resp=0.05, k_loss=0.0,
                 Tmin=8.0, Tmax=40.5, Topt=26.7, t_ref=25, Q10=2, k_0=3, a_k=5E-5, C_P=0.3,
                 CP50=0.0043, a_P=1, KiP=101, C_N=0.6, CN50=0.034, a_N=1, KiN=604, BM50=176,
                 mass_per_frond=0.0001, BMw2BMd=16.7),
      # boundary presets defined by expert judgement
      param.bounds=list(Emax=c(0, 1),  EC50=c(0, 1e6), b=c(0.1, 20), P_up=c(0, 100),
                        k_phot_max=c(0,1)),
      endpoints=c("BM","r"),
      forcings.req=c("temp", "rad"),
      forcings=list(temp=data.frame(time=0, tmp=12),
                    rad=data.frame(time=0, irr=15000)),
      control.req=TRUE,
      init=c(BM=0.0012, E=1, M_int=0),
      transfer.interval=-1,
      transfer.biomass=0.0012,
      transfer.comp.biomass="BM",
      transfer.comp.scaled="M_int"
  )  %>% set_noexposure() -> o
  if(!missing(param))
    o <- set_param(o, param)
  if(!missing(init))
    o <- set_init(o, init)

  o
}

#' @describeIn Lemna_Schmitt model variant with cumulative exposure threshold
#' @export
Lemna_SchmittThold <- function(param, init) {
  o <- Lemna_Schmitt()
  o@name <- "Lemna_SchmittThold"
  o@param.req <- c(o@param.req, "threshold")
  o@init <- c(o@init, AUC=0)

  if(!missing(param))
    o <- set_param(o, param)
  if(!missing(init))
    o <- set_init(o, init)

  o
}


########################
## Simulation
########################

# @param scenario Scenario object
# @param nout `numeric`, number of additional output variables, `nout=1` appends
#   the internal concentration `C_int`, the maximum number is 13
# @param method string, numerical solver used by [deSolve::ode()]
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_lemna_schmitt <- function(scenario, nout=2, method="lsoda", hmax=0.1, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # make sure that parameters are present and in required order
  params.req <- c("Emax","EC50","b","P_up","AperBM","Kbm","P_Temp","MolWeight",
                  "k_phot_fix","k_phot_max","k_resp","k_loss","Tmin","Tmax",
                  "Topt","t_ref","Q10","k_0","a_k","C_P","CP50","a_P","KiP",
                  "C_N","CN50","a_N","KiN","BM50","mass_per_frond","BMw2BMd")
  # additional output variables
  outnames <- c("C_int","FrondNo","C_int_u","BM_fresh","k_phot_eff","k_resp_eff","f_Eff",
                "P_up_eff","actConc","actTemp","actRad","dBMdt","dEdt","dM_intdt")

  # check for missing parameters
  params.missing <- setdiff(params.req, names(params))
  if(length(params.missing)>0)
    stop(paste("missing parameters:",paste(params.missing,sep=",",collapse=",")))

  # make sure exposure AUC parameter exists
  # magic value: -1 denotes that threshold is not considered
  if(!("threshold" %in% names(params))) {
    params["threshold"] <- -1
  }
  # disable threshold for scenarios were effects were disabled
  if(params["Emax"] <= 0 | params["threshold"] <= 0) {
    params["threshold"] <- -1
  }

  # reorder parameters for deSolve
  params <- params[c(params.req, "threshold")]
  # check if any parameter is negative apart from threshold
  if(any(head(params,-1)<0)) {
    stop(paste("parameter out of bounds: ",paste(names(params)[which(head(params,-1)<0)], collapse=",")))
  }

  # create forcings list
  # TODO check if it can be activated without issues
  #if(params["k_phot_fix"]==FALSE)
  forcings <- list(scenario@exposure@series, scenario@forcings$temp, scenario@forcings$rad)
  #else  # temp and radiation are not required by model if k_phot_fix==TRUE
  #  forcings <- list(exposure, data.frame(t=0,temp=-1), data.frame(t=0,rad=-1))

  # run solver
  ode(y=scenario@init, times=scenario@times, parms=params, dllname="cvasi",
        initfunc="lemna_schmitt_init", func="lemna_schmitt_func", initforc="lemna_schmitt_forc",
        forcings=forcings, nout=nout, outnames=outnames, method=method,
        hmax=hmax, ...)
}
#' @include solver.R
#' @describeIn solver Numerically integrates Lemna_Schmitt scenarios
setMethod("solver", "LemnaSchmitt", function(scenario, ...) solver_lemna_schmitt(scenario, ...) )


########################
## Effects
########################

# effects are calculated them same way as for `LemnaSetac`, see
# model-lemna_setac.R for details

test_that("set_transfer default handler", {
  expect_error(new("EffectScenario") %>% set_transfer(interval=1), "not support")
  expect_error("foo" %>% set_transfer(interval=1), "not support")
})

test_that("set_transfer interval", {
  # no transfer
  Lemna_Schmitt() %>% set_transfer(interval=-1) -> sc
  expect_equal(sc@transfer.interval, -1)
  expect_equal(sc@transfer.times, numeric(0))

  ## transfer interval
  Lemna_Schmitt() %>% set_transfer(interval=5) -> sc
  expect_equal(sc@transfer.interval, 5)
  expect_equal(sc@transfer.times, numeric(0))

  ## transfer times
  Lemna_Schmitt() %>% set_transfer(times=seq(5)) -> sc
  expect_equal(sc@transfer.interval, -1)
  expect_equal(sc@transfer.times, seq(5))
  # auto-ordering of times
  Lemna_Schmitt() %>% set_transfer(times=c(1,5,2)) -> sc
  expect_equal(sc@transfer.times, c(1,2,5))

  # using interval and times in conjunction
  Lemna_Schmitt() %>% set_transfer(interval=-2, times=c(1)) -> sc
  expect_equal(sc@transfer.interval, -1)
  expect_equal(sc@transfer.times, c(1))
  Lemna_Schmitt() %>% set_transfer(interval=1, times=c()) -> sc
  expect_equal(sc@transfer.interval, 1)
  expect_equal(sc@transfer.times, numeric(0))

  # re-setting options
  Lemna_Schmitt() %>%
    set_transfer(interval=5) %>%
    set_transfer(times=seq(5)) -> sc
  expect_equal(sc@transfer.interval, -1)
  expect_equal(sc@transfer.times, seq(5))

  Lemna_Schmitt() %>%
    set_transfer(times=seq(5)) %>%
    set_transfer(interval=5) -> sc
  expect_equal(sc@transfer.interval, 5)
  expect_equal(sc@transfer.times, numeric(0))
})

test_that("set_transfer biomass", {
  ## biomass
  Lemna_Schmitt() %>% set_transfer(biomass=23) -> sc
  expect_equal(sc@transfer.biomass, 23)
  # one biomass amount for each transfer
  Lemna_Schmitt() %>% set_transfer(times=0:2, biomass=1:3) -> sc
  expect_equal(sc@transfer.biomass, 1:3)

  ## compartments
  Lemna_Schmitt() %>%
    set_transfer(scaled_comp=c("E","M_int")) -> sc
  expect_equal(sc@transfer.comp.scaled, c("E","M_int"))
  Lemna_Schmitt() %>%
    set_transfer(scaled_comp=c("E")) %>%
    set_transfer(scaled_comp=c("M_int")) -> sc
  expect_equal(sc@transfer.comp.scaled, c("M_int"))

  # setting all at once
  Lemna_Schmitt() %>%
    set_transfer(interval=-2, times=c(1,2), biomass=c(3,4), scaled_comp="E") -> sc
  expect_equal(sc@transfer.interval, -1)
  expect_equal(sc@transfer.times, c(1,2))
  expect_equal(sc@transfer.biomass, c(3,4))
  expect_equal(sc@transfer.comp.scaled, "E")
})

test_that("set_transfer invalid arguments", {
  sc <- Lemna_Schmitt()
  # invalid inputs
  expect_error(set_transfer(sc, interval=NA), "numeric")
  expect_error(set_transfer(sc, interval=2, times=2), "cannot be used")
  expect_error(set_transfer(sc, times=c(2,NA,3)), "invalid value")
  expect_error(set_transfer(sc, times=1, biomass="foo"), "invalid value")
  expect_error(set_transfer(sc, times=1, biomass=c(1, NA_real_)), "invalid value")
  expect_error(set_transfer(sc, times=1, biomass=1:3), "length of")
  expect_error(set_transfer(sc, interval=1, biomass=1:3), "length of")
  # invalid object state
  expect_error(set_transfer(sc, interval=1, scaled_comp=character(0)), "not be empty")
  expect_error(set_transfer(sc, interval=1, scaled_comp="foo"), "state variables")
  sc2 <- sc
  sc2@transfer.comp.biomass <- NA_character_
  expect_error(set_transfer(sc2, interval=1), "biomass state variable.*not set")
  sc2@transfer.comp.biomass <- character(0)
  expect_error(set_transfer(sc2, interval=1), "biomass state variable.*length 1")
})

test_that("set_notransfer", {
  tr <- Lemna_Schmitt() %>% set_transfer(interval=5)

  # single scenario
  expect_false(has_transfer(tr %>% set_notransfer()))
  # multiple scenarios
  expect_equal(has_transfer(list(tr, tr) %>% set_notransfer()), c(FALSE, FALSE))
})

#' Set window length
#'
#' Exposure windows are defined as a period of time at the scale of the exposure series.
#' As an example: if an exposure series has an hourly time step, a window length
#' of `24` will consider the exposure within 24 hours intervals for effect
#' calculation. The same applies for the window interval, i.e. the period between
#' considered exposure windows. Set `length=-1` to disable moving windows.
#'
#' @param x vector of `EffectScenario` objects
#' @param length `numeric`, length of exposure window to consider for effect
#'  calculation, set `length=-1` to disable moving windows
#' @param interval `numeric`, interval between considered exposure windows
#'
#' @return modified `EffectScenario` objects
#' @include class-EffectScenario.R
#' @export
#' @examples
#' # Calculate the maximum effect for all windows of 10 days length
#' metsulfuron %>%
#'   set_window(length=10, interval=1) %>%
#'   effect()
#'
#' # Disable moving exposure windows
#' metsulfuron %>%
#'   set_nowindow() %>%
#'   effect()
set_window <- function(x, length, interval) {
  # vectorized input
  if(is.vector(x)) {
    return(sapply(x, function(sc) set_window(sc, length=length, interval=interval)))
  }

  if(!missing(length)) {
    if(!is.numeric(length) | base::length(length) > 1)
      stop("length must be a scalar")
    if(length == 0)
      stop("interval must be larger than zero")
    if(length < 0) {
      length <- -1
      x@window.interval <- -1
    }

    x@window.length <- length
  }
  if(!missing(interval)) {
    if(!is.numeric(interval) | base::length(interval) > 1)
      stop("interval must be a scalar")
    if(interval <= 0) {
      interval <- -1
      x@window.length <- -1
    }

    x@window.interval <- interval
  }

  if(x@window.interval > 0 & x@window.length < 0)
    stop("window interval defined, but window length is missing")
  if(x@window.interval < 0 & x@window.length > 0)
    stop("window length defined, but window interval is missing")
  x
}

#' @describeIn set_window Disable moving windows
#' @export
set_nowindow <- function(x) {
  # vectorized input
  if(is.vector(x)) {
    return(sapply(x, set_nowindow))
  }

  set_window(x, length=-1, interval=-1)
}

test_that("no windows", {
  sc <- new("EffectScenario") %>% set_times(1:7) %>% set_window(length=-1)
  sc@control.req <- FALSE

  env <- cache_windows(sc)
  expect_equal(env$windows, list(c("window.start"=1, "window.end"=7)))
  expect_equal(env$controls, list())
})

test_that("windows", {
  sc <- new("EffectScenario") %>% set_times(1:7) %>% set_window(length=3, interval=2)
  sc@control.req <- FALSE
  windws <- list(
    c("window.start"=1, "window.end"=4),
    c("window.start"=3, "window.end"=6)
  )

  # from scenario
  env <- cache_windows(sc)
  expect_equal(env$windows, windws)
  expect_equal(env$controls, list())

  # from sequence
  sq <- sequence(list(sc, sc), breaks=2)
  env <- cache_windows(sq)
  expect_equal(env$windows, windws)
  expect_equal(env$controls, list())
})

test_that("controls", {
  sc <- minnow_it %>%
    set_param(c(kd=1000, hb=0)) %>%
    set_times(1:7) %>%
    set_window(length=3, interval=2)
  sc@control.req <- TRUE
  scc <- sc %>% set_noexposure()
  scv <- sc %>% set_exposure(data.frame(t=0:10, c=runif(11)), reset_times=FALSE)
  windws <- list(
    c("window.start"=1, "window.end"=4),
    c("window.start"=3, "window.end"=6)
  )
  effcts <- list(c(L=0), c(L=0))

  # from scenario with constant forcings
  env <- cache_windows(scc)
  expect_equal(env$windows, windws)
  expect_equal(env$controls, effcts)

  # from scenario with variable forcings
  env <- cache_windows(scv)
  expect_equal(env$windows, windws)
  expect_equal(env$controls, effcts)

  # from sequence
  sq <- sequence(list(scv, scv), breaks=2)
  env <- cache_windows(sq)
  expect_equal(env$windows, windws)
  expect_equal(env$controls, effcts)
})

test_that("control value", {
  sc <- metsulfuron %>%
    set_times(1:7) %>%
    set_window(length=3, interval=2)
  scc <- sc %>% set_noexposure()
  scc@control.req <- FALSE

  env <- cache_windows(sc)
  expect_equal(length(env$controls), 2)

  # first window
  expect_equal(env$controls[[1]], scc %>% set_times(1:4) %>% effect(ep_only=TRUE), tolerance=1e-5)
  # second window
  expect_equal(env$controls[[2]], scc %>% set_times(3:6) %>% effect(ep_only=TRUE), tolerance=1e-5)
})

test_that("invalid arguments", {
  expect_error(cache_windows(1))
  expect_error(cache_windows(c(1, 2)))
})

test_that("window_candidates", {
  sc <- new("EffectScenario")
  sc@control.req <- FALSE

  # no moving windows
  sc <- sc %>% set_times(0:5) %>% set_nowindow()
  win <- cache_windows(sc)$windows
  expect_equal(win, list(c(window.start=0, window.end=5)))

  # scenario starting at zero
  sc <- sc %>% set_times(0:5) %>% set_window(length=1, interval=1)
  win <- dplyr::bind_rows(cache_windows(sc)$windows)
  expect_equal(win, data.frame(window.start=0:4, window.end=1:5), ignore_attr=TRUE)

  # scenario starting at non-zero
  sc <- sc %>% set_times(1:5) %>% set_window(length=1, interval=1)
  win <- dplyr::bind_rows(cache_windows(sc)$windows)
  expect_equal(win, data.frame(window.start=1:4, window.end=2:5), ignore_attr=TRUE)

  # ending not on last time point
  sc <- sc %>% set_times(1:6) %>% set_window(length=2, interval=3)
  win <- dplyr::bind_rows(cache_windows(sc)$windows)
  expect_equal(win, data.frame(window.start=c(1,4), window.end=c(3,6)), ignore_attr=TRUE)

  # window length does not conform to output time step
  sc <- sc %>% set_times(0:3) %>% set_window(length=0.5, interval=1)
  win <- dplyr::bind_rows(cache_windows(sc)$windows)
  expect_equal(win, data.frame(window.start=c(0:2), window.end=0:2 + 0.5), ignore_attr=TRUE)

  # window interval does not conform to output time step
  sc <- sc %>% set_times(0:3) %>% set_window(length=1, interval=0.5)
  win <- dplyr::bind_rows(cache_windows(sc)$windows)
  expect_equal(win, data.frame(window.start=seq(0, 2, 0.5), window.end=seq(1, 3, 0.5)), ignore_attr=TRUE)

  # window longer than scenario times
  sc <- sc %>% set_times(0:3) %>% set_window(length=5, interval=1)
  win <- cache_windows(sc)$windows
  expect_equal(win, list(c(window.start=0, window.end=5)))

  # no output times
  sc@times <- numeric(0)
  expect_error(cache_windows(sc))
})

# Check if simulation result contain an error raised by deSolve
#
# @param x return value of [simulate()]
# @return logical, `TRUE` means an error occurred during integration, else `FALSE`
num_error <- function(x) {
  #if(!is(x, "cvasi.simulate"))
  #  stop("Argument `x` must be a simulation result")
  status <- attr(x, "cvasi_status")
  if(is.null(status))
    return(FALSE)

  return(status == "error")
}

num_success <- function(x) {
  status <- attr(x, "cvasi_status")
  if(is.null(status))
    return(FALSE)

  return(status == "success")
}

# Check if simulation was aborted - only for internal use
num_aborted <- function(x) {
  diagn <- attr(x, "desolve_diagn")
  if(!is.null(diagn))
    return(diagn$istate[[1]] == -1 | diagn$istate[[1]] == -2)

  return(FALSE)
}

# Helper method to convert a deSolve result type to a data.frame while also keeping
# deSolve's diagnostic info in place using attribute `desolve_diagn`
ode2df <- function(x) {
  df <- as.data.frame(x)
  attr(df, "desolve_diagn") <- attributes(x)
  df
}

#' Print information about numerical solver result
#'
#' `r lifecycle::badge("experimental")`
#'
#' Prints information on the status of a return value from [simulate()], e.g.
#' if it was successful and what, if any, issues occurred. Also provides tips
#' on solving frequently occurring issues.
#'
#' The function requires certain metadata which is created by [deSolve::ode()]
#' and is passed through to the result of `simulate()`. The metadata may be lost
#' if the `data.frame` returned by `simulate()` is converted or cast to other types.
#'
#' @param obj Return value of [simulate()]
#' @seealso [diagnostics()]
#' @export
#' @examples
#' # A simulation without any issues
#' minnow_it %>% simulate() %>% num_info()
#'
#' # A simulation which terminated early due to the solver
#' # taking too many numerical steps
#' rs <- suppressWarnings(minnow_it %>% simulate(hmax=1e-80))
#' num_info(rs)
#'
#' # Print deSolve diagnostics for additional information
#' diagnostics(rs)
num_info <- function(obj) {
  UseMethod("num_info")
}

#' @rdname num_info
#' @export
num_info.default <- function(obj) {
  stop("Type is not supported to provide numerical information.")
}

#' @rdname num_info
#' @export
num_info.cvasi_simulate <- function(obj) {
  diag <- attr(obj, "desolve_diagn")
  status <- attr(obj, "cvasi_status")

  if(status == "success") {
    cli::cli_text("Simulation status: ", cli::col_green(toupper(status)))
  } else {
    cli::cli_text("Simulation status: ", cli::col_red(toupper(status)))
  }

  # TODO do state variables contain invalid values?

  if(is.null(diag)) {
    cli::cli_text(cli::col_grey("Object does not contain solver diagnostics info."))
  }
  else
  {
    # return code
    rc <- diag$istate[[1]]

    # get issue description
    if(rc < 0) {
      tx <- tryCatch(capture.output(diagnostics.cvasi_simulate(obj)), error=function(e) "")
      idesc <- trimws(tx[which(startsWith(tx, "  return code")) + 1])
    }

    # maxsteps exceeded
    if(rc == -1)
    {
      steps <- diag$istate[[2]]

      cli::cli_h1("Numerical Steps")
      l1 <- cli::cli_ul()
      cli::cli_li(idesc)
      if(!is.null(diag$rstate)) {
        cli::cli_li("Simulation terminated early at t={.val {signif(diag$rstate[[3]], digits=3)}}")
      } else {
        cli::cli_li("Simulation terminated early")
      }
      def <- ifelse(steps == 5000, " (default value)", "")
      cli::cli_li("Argument {.field maxsteps}={.val {steps}}{def}")

      cli::cli_text(" ")
      cli::cli_text(cli::col_blue("Potential Solutions"))
      cli::cli_li("Increase the number of allowed steps with argument {.field maxsteps}")
      cli::cli_li("Increase the value of arguments {.field {c('hmax','atol','rtol')}}")
      cli::cli_li("Try a different ODE solver with argument {.field method}, such as {.val {c('lsoda','lsode','ode45','rk4')}}")
      cli::cli_end(l1)
    }

    # excess accuracy requested
    if(rc == -2)
    {
      steps <- diag$istate[[2]]

      cli::cli_h1("Numerical Precision")
      l1 <- cli::cli_ul()
      cli::cli_li(idesc)
      cli::cli_li("Simulation terminated early at t={.val {signif(diag$rstate[[1]], digits=3)}}")

      cli::cli_text(" ")
      cli::cli_text(cli::col_blue("Potential Solutions"))
      cli::cli_li("Increase the value of arguments {.field {c('hmax','atol','rtol')}}")
      cli::cli_end(l1)
    }
  }

  output <- attr(obj, "desolve_output")
  if(!is.null(output))
  {
    if(length(output) > 0) {
      output <- stringr::str_trim(output)
      output <- output[output != ""]
      cli::cli_h1("deSolve Output")
      l3 <- cli::cli_ol()
      cli::cli_li(c(output))
      cli::cli_end(l3)
    }
  }

  conds <- attr(obj, "desolve_conds")
  if(!is.null(conds))
  {
    if(length(conds) > 0)
    {
      cli::cli_h1("deSolve Messages")
      l4 <- cli::cli_ol()
      for(m in conds)
      {
        if(m[[1]] == "warning")
          cli::cli_li(cli::cli_text(cli::col_red("WARNING:"), " ", stringr::str_to_sentence(m[[3]])))
        else
          cli::cli_li(cli::cli_text(cli::bg_red(cli::col_white("ERROR:")), " ", stringr::str_to_sentence(m[[3]])))
      }
      cli::cli_end(l4)
    }
  }
}

#' @rdname num_info
#' @export
num_info.cvasi_fit <- function(obj) {
  num_info.cvasi_simulate(obj)
}

#' @importFrom deSolve diagnostics diagnostics.default diagnostics.deSolve
#' @export
NULL

#' Diagnostics of solvers
#'
#' Prints several diagnostics of the simulation to the console, e.g. number of
#' steps taken, the last step size, etc. The information is provided by
#' [deSolve::diagnostics()].
#'
#' @seealso [deSolve::diagnostics()]
#' @param obj return value of a simulation
#' @param ... unused parameters
#' @rdname diagnostics
#' @export
diagnostics <- function(obj, ...) {
  UseMethod("diagnostics")
}

#' @rdname diagnostics
#' @export
diagnostics.default <- function(obj, ...) {
  deSolve::diagnostics.deSolve(obj)
}

#' @rdname diagnostics
#' @export
diagnostics.cvasi_simulate <- function(obj, ...) {
  # translate our metadata to something that is compatible with deSolve's format
  diag <- attr(obj, "desolve_diagn")
  if(is.null(diag))
    stop("Object does not contain solver diagnostics info.")
  attr(obj, "istate") <- diag$istate
  attr(obj, "rstate") <- diag$rstate
  attr(obj, "type") <- diag$type
  class(obj) <- "deSolve"
  deSolve::diagnostics.deSolve(obj)
}

test_that("raised warnings", {
  # sim aborted
  expect_warning(minnow_it %>% simulate(maxsteps=10), "Simulation aborted")

  # suppress warnings
  foo <- minnow_it %>% simulate(maxsteps=10, .suppress=TRUE)
  expect_true(num_aborted(foo))

  # warning raised by solver
  source(test_path("dummy.R"), local = TRUE)
  instable <- new("DummyScenario", solver=function(...) {
    warning("planned warning")
    df <- data.frame(t=0, A=1)
    attr(df, "desolve_diagn") <- list(istate=-42) # magic value from deSolve, cf. [num_info()]
    df
  })
  expect_warning(rs <- simulate(instable), "planned warning")
})

test_that("raised errors", {
  # unknown symbol
  expect_error(minnow_it %>% simulate(hmax=foobar), "foobar' not found")

  # invalid argument to solver
  expect_error(minnow_it %>% simulate(hmax=-1), "non-negative value")

  # suppress errors
  foo <- minnow_it %>% simulate(hmax=-1, .suppress=TRUE)
  expect_true(num_error(foo))
})

# TODO test calibrate with method L-BFGS-B for a problem which is not computable
#   this way the optim_set will return the error value which needs to be finite
#   for L-BFGS-B to complete

test_that("fit to simple dataset", {
  tol <- 1e-5

  ## set up a scenario to create perfect fit data
  rs <- simulate(minnow_it)
  # modify scenario by setting parameter `kd` to quasi-random value
  tofit <- minnow_it %>%
    set_param(c(kd=0.01)) %>%
    set_bounds(list(kd=c(0.001, 10)))

  # calibrate modified scenario on synthetic data
  calib <- calibrate(tofit,
                     par=c(kd=0.1),
                     data=rs[, c(1, 2)],
                     output="D",
                     method="Brent",
                     verbose=FALSE)

  expect_equal(calib$par[["kd"]],
               minnow_it@param$kd,
               tolerance=tol)

  # expect a warning with Nelder-Mead but result should be good nonetheless
  expect_warning(calibrate(tofit,
                           par=c(kd=0.1),
                           data=rs[, c(1, 2)],
                           output="D",
                           method="Nelder-Mead",
                           control=list(reltol=1e-12),
                           verbose=FALSE) -> calib)
  expect_equal(calib$par[["kd"]],
               minnow_it@param$kd,
               tolerance=tol)
})

test_that("fit to dataset with replicates", {
  tol <- 1e-5

  ##
  ## set up a scenario to create perfect fit data
  ##
  rs <- minnow_it %>%
    set_times(rep(minnow_it@times, each=2)) %>%
    simulate()

  # modify scenario by setting parameter `kd` to quasi-random value
  tofit <- minnow_it %>%
    set_param(c(kd=0.1)) %>%
    set_bounds(list(kd=c(0.001, 10)))

  # calibrate modified scenario on synthetic data
  calib <- calibrate(tofit,
                     par=c(kd=0.1),
                     data=rs[, c(1, 2)],
                     output="D",
                     method="Brent",
                     verbose=FALSE)

  expect_equal(calib$par[["kd"]],
               minnow_it@param$kd,
               tolerance=tol)
})

test_that("fit to complex dataset", {
  ##
  ## fit a parameter to several synthetic datasets created using
  ## various values for parameter `kd`
  ##
  minnow_it %>%
    simulate() -> rs.ideal # original kd=1.2296
  minnow_it %>%
    set_param(c(kd=1.1)) %>%
    simulate() %>%
    dplyr::mutate(trial="low") -> rs.lo
  minnow_it %>%
    set_param(c(kd=1.33)) %>%
    simulate() %>%
    dplyr::mutate(trial="high")  -> rs.hi

  set.seed(123)
  df <- dplyr::bind_rows(rs.lo, rs.hi)

  # add noisy replicates
  for(i in seq(10)) {
    rs.ideal %>%
      dplyr::mutate(D=D + stats::rnorm(dplyr::n(), sd=0.05), trial=as.character(i)) %>%
      dplyr::mutate(D=pmax(D, 0)) %>%
      dplyr::bind_rows(df) -> df
  }

  # modify scenario by setting parameter `kd` to quasi-random value
  tofit <- minnow_it %>%
    set_param(c(kd=0.1)) %>%
    set_bounds(list(kd=c(0.001, 10)))
  # fit to data
  calibrate(tofit,
            par=c(kd=0.1),
            data=df[, c("time", "D", "trial")],
            output="D",
            method="Brent",
            verbose=FALSE) -> calib
  # we have to use lower precision for comparison purposes, but result is
  # derived from noisy data, so this is OK
  expect_equal(calib$par[["kd"]], minnow_it@param$kd, tolerance=0.01)


  ##
  ## fit two parameters to the previous data
  ##
  #suppressWarnings(
  #  calibrate(tofit,
  #    par=c(hb=0.1, kd=0.1),
  #    data=df,
  #    by="trial",
  #    output="D",
  #    verbose=FALSE)) -> calib
  ## result of Nelder-Mead fit is sensitive to start conditions
  ## hb is irrelevant because it has no influence on 'D' (internal damage/conc)
  #expect_equal(calib$par[["kd"]],
  #             minnow_it@param$kd,
  #             tolerance=0.01)


  # fit with box constraints
  #calibrate(tofit,
  #          par=c(hb=0.1, kd=1),
  #          data=df,
  #          by="trial",
  #          output="D",
  #          method="L-BFGS-B",
  #          verbose=FALSE) -> calib
  #expect_equal(calib$par[["kd"]],
  #             minnow_it@param$kd,
  #             tolerance=0.01)
})


test_that("fit to calibration set", {
  ## fit a parameter to several synthetic datasets created using
  ## various values for parameter `kd`
  minnow_it %>%
    simulate() %>%
    dplyr::select(time, D) -> rs.ideal

  # modify scenario by setting parameter `kd` to quasi-random value
  tofit <- minnow_it %>%
    set_param(c(kd=0.1)) %>%
    set_bounds(list(kd=c(0.001, 10)))
  # create a single calibration set
  cs <- caliset(tofit, rs.ideal)

  calibrate(cs,
            par=c(kd=0.1),
            output="D",
            method="Brent",
            verbose=FALSE) -> calib

  expect_equal(calib$par[["kd"]],
               minnow_it@param$kd,
               tolerance=1e-5)


  # create additional synthetic data using different values for `kd`
  minnow_it %>%
    set_param(c(kd=1.1296)) %>%
    simulate() %>%
    dplyr::select(time, D) -> rs.lo
  minnow_it %>%
    set_param(c(kd=1.3296)) %>%
    simulate() %>%
    dplyr::select(time, D) -> rs.hi

  # create list of several calibration sets
  cs <- list(
    caliset(tofit, rs.ideal),
    caliset(tofit, rs.lo),
    caliset(tofit, rs.hi)
  )

  calibrate(cs,
            par=c(kd=0.1),
            output="D",
            method="Brent",
            verbose=FALSE) -> calib

  expect_equal(calib$par[["kd"]],
               minnow_it@param$kd,
               tolerance=0.01)
})

# TODO optim raises one warning about the selected method and then additional
#   ones originating from the dummy scenario, the test does not account for the former
test_that("failed simulations during fitting", {
  source(test_path("dummy.R"), local = TRUE)
  fail <- new("DummyScenario", simulate=function(...) stop("dummy scenario failed"))

  # simulation fails completely
  suppressWarnings( # suppress any additional warnings
    expect_warning(
      calibrate(fail, par=c("baz"=0), data=data.frame("t"=0:2, "a"=0), output="a", verbose=FALSE),
      "simulations failed"
    )
  )

  # some runs are unstable
  instable <- new("DummyScenario", solver=function(...) {
    warning("planned warning")
    df <- data.frame(t=0:2, A=1)
    attr(df, "desolve_diagn") <- list(istate=c(-1, 100), rstate=1) # magic value from deSolve, cf. [num_info()]
    df
  }) %>%
    set_bounds(list(baz=c(0, 1)))
  expect_warning(rs <- calibrate(instable, par=c("baz"=0), data=data.frame("t"=0:2, "A"=0),
                                 output="A", verbose=FALSE, method="Brent"))
  expect_true(num_aborted(rs))
  expect_equal(attr(rs, "desolve_diagn"), list(istate=c(-1, 100), rstate=1))
})

test_that("fit with weights", {
  tofit <- minnow_it %>%
    set_bounds(list(kd=c(0.001, 10)))
  # synthetic dataset #1, should be irrelevant due to small weights
  tofit %>%
    simulate() %>%
    dplyr::select(time, D) -> rs1
  # synthetic dataset #2, the only relevant one due to large weights
  tofit %>%
    set_param(c(kd=0.5)) %>%
    simulate() %>%
    dplyr::select(time, D) -> rs2

  # create list containing calibration sets
  cs <- list(
    caliset(tofit, rs1, weight=0.0001),
    caliset(tofit, rs2, weight=1000)
  )

  calibrate(cs,
            par=c(kd=minnow_it@param$kd),
            output="D",
            method="Brent",
            verbose=FALSE) -> calib

  expect_equal(calib$par[["kd"]],
               0.5,
               tolerance=1e-5)

})

test_that("arg x=scenario invalid", {
  sc <- new("EffectScenario")
  # data not a data.frame
  expect_error(calibrate(sc, data=1), "must be a .data.frame.")
  # data empty
  expect_error(calibrate(sc, data=data.frame()), "is empty")
  # output not in data
  df <- data.frame(t=0:3, foo=1, bar=2)
  expect_error(calibrate(sc, data=df, output="baz"), ".par. is missing")
  # group by not a single character
  lifecycle::expect_deprecated(expect_error(calibrate(sc, data=df, output="foo", by=c("t", "foo")), "length one"))
})

test_that("arg x=calisets invalid", {
  sc <- new("EffectScenario") %>% set_times(0:5) %>% set_param(c(foo=1, bar=2))
  cs <- caliset(sc, data.frame(t=0:5, bar=1))

  # data supplied
  expect_error(calibrate(cs, data=data.frame()), "cannot be used in combination")
  # not all elements are calisets
  expect_error(calibrate(list(cs, 1)), "only contain caliset")
  # par is non-numeric
  expect_error(calibrate(cs, par=sc), "non-numerical elements")
  expect_error(calibrate(cs, par=c("foo"="b")), "not determine starting value")
  # not all elements in par are named
  expect_error(calibrate(cs, par=c(1)), "must be named")
  expect_error(calibrate(cs, par=c(foo=1, 2)), "must be named")
  # ... are actual parameters of the scenario
  expect_error(calibrate(cs, par=c(baz=0)), "not scenario parameters")
  # output missing
  expect_error(calibrate(cs, par=c(foo=0)), "output. is missing")
  # output invalid length
  expect_error(calibrate(cs, par=c(foo=0), output=character(0)), "output. must be of length one")
  expect_error(calibrate(cs, par=c(foo=0), output=c("a","b")), "output. must be of length one")
  # output not a string
  expect_error(calibrate(cs, par=c(foo=0), output=1), "output. must be a string")
})

test_that("fit with error functions", {
  tol <- 1e-5
  rs <- simulate(minnow_it)
  tofit <- minnow_it %>%
    set_param(c(kd=0.01)) %>%
    set_bounds(list(kd=c(0.001, 10)))

  # use an alternative error function
  tofit <- tofit %>% set_init(rs[2,c("D", "H")])
  calib <- calibrate(tofit,
                     par=c(kd=0.1),
                     data=rs[-1, c(1, 2)],
                     output="D",
                     err_fun="sse",
                     log_scale=TRUE,
                     method="Brent",
                     verbose=FALSE)
  expect_equal(calib$par[["kd"]],
               minnow_it@param$kd,
               tolerance=tol)

  # use a custom error function
  myerr <- function(obs, pred, ...) sum((log(obs) - log(pred))^2)
  calib2 <- calibrate(tofit, par=c(kd=0.1), data=rs[-1, c(1, 2)],
                      output="D", err_fun=myerr, method="Brent",
                      verbose=FALSE)
  expect_equal(calib2$par[["kd"]],
               calib$par[["kd"]],
               tolerance=tol)
})

test_that("SSE errfun", {
  # basic use
  expect_equal(sse(c(1), c(1)), 0)
  expect_equal(sse(c(1), c(0)), 1)
  expect_equal(sse(c(2), c(0)), 4)
  expect_equal(sse(c(0), c(1)), 1)
  expect_equal(sse(c(0), c(2)), 4)
  expect_equal(sse(c(1, 1), c(1, 1)), 0)
  expect_equal(sse(c(2, 2), c(1, 1)), 2)
  expect_equal(sse(c(3, 3), c(1, 1)), 8)
  # weights
  expect_equal(sse(c(1), c(1), c(1)), 0)
  expect_equal(sse(c(3), c(1), c(1)), 4)
  expect_equal(sse(c(3), c(1), c(0.1)), 0.4)
  expect_equal(sse(c(3, 1), c(1, 1), c(0.1)), 0.4)
  # sizes of arguments dont match
  expect_error(sse(1, numeric(0)), "observed and predicted")
  expect_error(sse(1:3, 1), "observed and predicted")
  expect_error(sse(1, 1:3), "observed and predicted")
  expect_error(sse(1, 1, numeric(0)), "weights and observed")
  expect_error(sse(1, 1, 1:3), "weights and observed")
  # invalid/missing values
  expect_equal(sse(NA_real_, 1), NA_real_)
})

test_that("Log SSE errfun", {
  # basic use
  expect_equal(log_sse(1, 1), 0)
  expect_equal(log_sse(exp(1), 1), 1)
  # TODO how to deal with zeros in data?
  expect_equal(log_sse(c(1, 1), c(1, 1)), 0)
  expect_equal(log_sse(c(exp(1), exp(1)), c(1, 1)), 2)
  # weights
  expect_equal(log_sse(1, 1, 1), 0)
  expect_equal(log_sse(exp(1), 1, 1), 1)
  expect_equal(log_sse(exp(1), 1, 0.1), 0.1)
  # sizes of arguments dont match
  expect_error(log_sse(1, numeric(0)), "observed and predicted")
  expect_error(log_sse(1:3, 1), "observed and predicted")
  expect_error(log_sse(1, 1:3), "observed and predicted")
  expect_error(log_sse(1, 1, numeric(0)), "weights and observed")
  expect_error(log_sse(1, 1, 1:3), "weights and observed")
  # invalid/missing values
  expect_equal(log_sse(NA_real_, 1), NA_real_)
})

test_that("input check: fitted parameter", {
  # parameter to fit not set yet, process should succeed
  val <- minnow_it@param[["kd"]]
  sc <- minnow_it
  sc@param[["kd"]] <- NULL
  moptim <- function(...) list(convergence=0, par=list(kd=0))

  with_mocked_bindings(
    calibrate(sc, par=c(kd=val), data=data.frame(t=0:5, S=0), output="S", verbose=FALSE, method="Nelder-Mead"),
    optim=moptim
  )
  succeed()

  # generic scenario which has an empty param.req slot
  sc <- minnow_it
  sc@param.req <- character(0)
  with_mocked_bindings(
    calibrate(sc, par=c(kd=val), data=data.frame(t=0:5, S=0), output="S", verbose=FALSE, method="Nelder-Mead"),
    optim=moptim
  )
  succeed()

  # parameter does not belong to model
  sc <- minnow_it
  expect_error(with_mocked_bindings(
    calibrate(sc, par=c(foo=1), data=data.frame(t=0:5, S=0), output="S", verbose=FALSE, method="Nelder-Mead"),
    optim=moptim
  ),
  "not scenario parameters")
})

# starting values for optimization are given by user or, if nothing set by
# user, are derived from the scenario's parameter values
test_that("starting value from scenario", {
  sc <- minnow_it %>% set_param(c(kd=42, hb=123))
  data <- data.frame(time=0:5, S=1, id="foo")
  par <- c("kd")
  moptim <- function(par, ...) list(convergence=0, par=as.list(par))

  # starting value taken from scenario's parameter list
  rs <- with_mocked_bindings(calibrate(sc, data=data, par="kd", output="S", verbose=FALSE, method="Nelder-Mead"),
                       optim = moptim)
  expect_equal(names(rs$par), "kd")
  expect_equal(rs$par$kd, 42)

  # one value from parameters, one supplied as argument
  rs <- with_mocked_bindings(calibrate(sc, data=data, par=list("kd", "hb"=1), output="S", verbose=FALSE, method="Nelder-Mead"),
                             optim = moptim)
  expect_equal(names(rs$par), c("kd", "hb"))
  expect_equal(rs$par$kd, 42)
  expect_equal(rs$par$hb, 1)

  # parameter not set
  expect_error(
    with_mocked_bindings(calibrate(sc, data=data, par="foo", output="S", verbose=FALSE, method="Nelder-Mead"),
                         optim = moptim),
    "not determine starting value"
  )
})

# calibrate() proposes a sensible method if nothing specified by user,
# otherwise user knows best, probably
test_that("optim method", {
  sc <- minnow_it %>% set_param(c(kd=42, hb=123)) %>% set_bounds(list(kd=c(0, 1)))
  data <- data.frame(time=0:5, S=1, id="foo")
  par <- c("kd")
  moptim <- function(par, method=NA_character_, ...) list(convergence=0, par=par, method=method)

  # single parameter -> use Brent
  rs <- with_mocked_bindings(calibrate(sc, data=data, par=c("kd"), output="S", verbose=FALSE),
                             optim = moptim)
  expect_equal(rs$method, "Brent")

  # multiple parameters -> use Nelder-Mead
  rs <- with_mocked_bindings(calibrate(sc, data=data, par=c("kd", "hb"), output="S", verbose=FALSE),
                             optim = moptim)
  expect_equal(rs$method, "Nelder-Mead")

  # custom method selected
  rs <- with_mocked_bindings(calibrate(sc, data=data, par=c("kd", "hb"), output="S", verbose=FALSE, method="foo"),
                             optim = moptim)
  expect_equal(rs$method, "foo")
})

# bounds are derived from scenario's parameter.bounds
test_that("parameter bounds", {
  sc <- minnow_it %>% set_bounds(list(kd=c(-2, -1), hb=c(-4, -3)))
  data <- data.frame(time=0:5, S=1, id="foo")
  moptim <- function(par, output, upper, lower, ...) list(convergence=0, par=par, upper=upper, lower=lower)

  # Methods that support boundaries
  with_mocked_bindings(rs <- calibrate(sc, data=data, par="kd", output="S", verbose=FALSE, method="Brent"),
                       optim = moptim)
  expect_equal(rs$lower, -2)
  expect_equal(rs$upper, -1)

  with_mocked_bindings(rs <- calibrate(sc, data=data, par=c("kd", "hb"), output="S", verbose=FALSE, method="L-BFGS-B"),
                       optim = moptim)
  expect_equal(rs$lower, c(-2, -4))
  expect_equal(rs$upper, c(-1, -3))

  # No bounds set
  expect_error(
    with_mocked_bindings(rs <- calibrate(sc, data=data, par="alpha", output="S", verbose=FALSE, method="Brent"),
      optim = moptim)
  )

  # Bounds not supported by method
  with_mocked_bindings(rs <- calibrate(sc, data=data, par="kd", output="S", verbose=FALSE, method="Nelder-Mead"),
                       optim = moptim)
  expect_equal(rs$lower, -Inf)
  expect_equal(rs$upper, Inf)
})

# outdate arguments should still work but raise a warning
test_that("deprecated arguments", {
  sc <- minnow_it %>% set_bounds(list(kd=c(0, 1)))
  data <- data.frame(time=0:5, S=1, id="foo")
  par <- c(kd = 0)
  moptim <- function(output, ...) list(convergence=0, par=list(kd=0), output=output)

  # no lifecycle messages
  with_mocked_bindings(calibrate(sc, data=data, par=par, output="S", verbose=FALSE),
                       optim = moptim)
  succeed()

  # outdated arguments
  with_mocked_bindings(
    lifecycle::expect_deprecated(rs <- calibrate(sc, data=data, par=par, endpoint="S", verbose=FALSE)),
    optim = moptim
  )
  expect_equal(rs$output, "S")

  with_mocked_bindings(
    lifecycle::expect_deprecated(calibrate(sc, data=data, par=par, output="S", by="id", verbose=FALSE)),
    optim = moptim
  )
})

#' Calls ODE solver for a particular model
#'
#' Please refer to the *Modeling Howto* vignette on how to implement custom
#' models by overloading the `solver` function.
#'
#' Some solvers may set reasonable default values for e.g. maximum step
#' length in time (`hmax`), but not all do. Please check the model documentation for
#' details.
#'
#' @param scenario [scenario] object
#' @param ... additional parameters passed on to [deSolve::ode()]
#' @param method string, selects the numerical solver used by [deSolve::ode()]
#' @param hmax numeric, sets the maximum step length in time, see [deSolve::ode()]
#' @param nout numeric, the  number of additional output variables returned by
#'          the model, see `vignette("compiledCode", "deSolve")` for details.
#'          If and which output variables are available depends on the scenario
#'          type, please refer to the documentation of the model in question.
#' @return `data.frame` with simulation results
#' @export
setGeneric("solver",
           function(scenario, ...) standardGeneric("solver"),
           signature = "scenario"
)

# Default solver which uses the model's name to switch between solver calls
solver_default <- function(scenario, ...) {
  stop("cannot simulate unknown model type, solver missing")
}

#' @describeIn solver Default solver, raises an error
setMethod("solver", "ANY", solver_default)

########################
## Organizing man pages
########################

#' GUTS-RED models
#'
#' Reduced *General Unified Threshold models of Survival* (GUTS) with stochastic
#' death (*SD*) and individual tolerance (*IT*)
#'
#' The TKTD models *[GUTS-RED-SD][GUTS_RED_SD()]* and *[GUTS-RED-IT][GUTS_RED_IT()]*
#' were described by EFSA (2018).
#' GUTS-RED models assume a one-compartment model which directly links
#' external concentration to the scaled damage. The scaled damage is given in
#' units of concentration, equal to the units of measurement in the external
#' medium, e.g. ug/L. The damage dynamics is connected to an individual hazard
#' state variable, resulting in simulated mortality when an internal damage
#' threshold is exceeded. The death mechanisms stochastic death (*SD*) and
#' individual threshold (*IT*) are extreme cases of the *GUTS* theory.
#'
#' For *SD* models, the threshold parameter for lethal effects is fixed and
#' identical for all individuals of a group, meaning that the variance of the
#' threshold values is zero. Hence, the killing rate relates the probability of
#' a mortality event in proportion to the scaled damage. For *IT* models,
#' the thresholds for effects are distributed among individuals of a group.
#' Mortality of an individual follows immediately once the individual's tolerance
#' is exceeded. Meaning in model terms that the killing rate is set to
#' infinity (EFSA 2018).
#'
#' @section State variables:
#' The following list describes the default names and standard units of *GUTS-RED*
#' state variables:
#' * `D`, scaled damage (conc)
#' * `H`, cumulative hazard (-)
#'
#' The state variables are initialized with zero by default.
#'
#' @section SD model parameters:
#' * `kd`, dominant rate constant (time^-1)
#' * `hb`, background hazard rate (time^-1)
#' * `z`, threshold for effects (conc)
#' * `kk`, killing rate constant (time^-1)
#'
#' @section IT model parameters:
#' * `kd`, dominant rate constant (time^-1)
#' * `hb`, background hazard rate (time^-1)
#' * `alpha`, median of thresholds (conc)
#' * `beta`, shape parameter (-)
#'
#' @section Effects:
#' The effect endpoint `L` (lethality) is available for *GUTS-RED* models.
#' A value of zero (`0.0`) denotes *no effect* on organism survival. A value of
#' one (`1.0`) denotes a lethality rate of 100%, i.e. no survivors.
#'
#' The survival probability `S` is available in the return value of [simulate()].
#'
#' @references
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#' @name GUTS-RED-models
#' @family GUTS-RED models
#' @family models
NULL


########################
## Class definitions
########################

#' @export
setClass("GutsRedSd", contains="EffectScenario")

#' @export
setClass("GutsRedIt", contains="EffectScenario")


########################
## Constructors
########################

#' GUTS-RED-IT scenario
#'
#' Reduced *General Unified Threshold models of Survival* (GUTS) with
#' individual tolerance (*IT*).
#'
#' @inheritSection GUTS-RED-models State variables
#' @inheritSection GUTS-RED-models IT model parameters
#' @inheritSection GUTS-RED-models Effects
#' @section Simulation output:
#' The return value of [simulate()] will contain values for the state variables,
#' as well as an additional column `S` which represents the survival probability
#' for each time point. `S` is calculated as described in EFSA (2018) as
#' *S = (1- F(t))*. The background hazard rate `hb` is already considered in state
#' variable `H` and therefore does not occur as an additional term to derive `S`.
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' @inherit GUTS-RED-models references
#'
#' @section Model history and changes:
#' - cvasi v1.0.0
#'   - Model and parameters as described by EFSA PPR (2018).
#' - cvasi v1.5.0
#'   - Cumulative damage over time will be calculated on ODE level, instead on
#'     results of [simulate()], to avoid influence of selected output times
#'     on predicted survival.
#'
#' @param param optional named `list` or `vector` with model parameters
#' @param init optional named numeric `vector` to use as initial state
#' @return an S4 object of type [GutsRedIt-class]
#'
#' @export
#' @family GUTS-RED models
#' @aliases GutsRedIt-class GUTS-RED-IT
#' @importFrom methods new
GUTS_RED_IT <- function(param, init) {
  new("GutsRedIt",
      name = "GUTS-RED-IT",
      param.req = c("kd", "hb", "alpha", "beta"),
      endpoints = c("L"),
      init = c(D=0, H=0),
      control.req=FALSE
  ) %>% set_noexposure() -> o
  if(!missing(param))
    o <- set_param(o, param)
  if(!missing(init))
    o <- set_init(o, init)

  o
}


#' GUTS-RED-SD scenario
#'
#' Reduced *General Unified Threshold models of Survival* (GUTS) with stochastic
#' death (*SD*).
#'
#' @inheritSection GUTS-RED-models State variables
#' @inheritSection GUTS-RED-models SD model parameters
#' @inheritSection GUTS-RED-models Effects
#' @inheritParams GUTS_RED_IT
#' @section Simulation output:
#' The return value of [simulate()] will contain values for the state variables,
#' as well as an additional column `S` which represents the survival probability
#' for each time point. `S` is calculated as described in EFSA (2018) as
#' *S = exp(-H)*. The background hazard rate `hb` is already considered in state
#' variable `H` and therefore does not occur as an additional term to derive `S`.
#'
#' @inheritSection GUTS_RED_IT Solver settings
#' @inherit GUTS-RED-models references
#' @return an S4 object of type [GutsRedSd-class]
#'
#' @export
#' @family GUTS-RED models
#' @aliases GutsRedSd-class GUTS-RED-SD
#' @importFrom methods new
GUTS_RED_SD <- function(param, init) {
  new("GutsRedSd",
      name = "GUTS-RED-SD",
      param.req = c("kd", "hb", "z", "kk"),
      endpoints = c("L"),
      init = c(D=0, H=0),
      control.req=FALSE
  ) %>% set_noexposure() -> o
  if(!missing(param))
    o <- set_param(o, param)
  if(!missing(init))
    o <- set_init(o, init)

  o
}


########################
## Simulation
########################

# @param scenario Scenario object
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param method string, numerical solver used by [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_gutsredsd <- function(scenario, method="lsoda", hmax=1, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # check if all required parameters are present
  params_missing <- is.na(params[scenario@param.req])
  if(any(params_missing)) {
    stop("Missing parameters: ", paste(scenario@param.req[params_missing], collapse=", "))
  }

  # make sure that parameters are present and in required order
  params <- params[c("kd", "hb", "z", "kk")]

  df <- ode2df(ode(y=scenario@init, times=scenario@times, parms=params, dllname="cvasi",
                    initfunc="gutsredsd_init", func="gutsredsd_func", initforc="gutsredsd_forc",
                    forcings=scenario@exposure@series, outnames=c("Cw"),
                    method=method, hmax=hmax, ...))
  # Derive survival probability, see EFSA Scientific Opinion on TKTD models, p. 33
  # doi:10.2903/j.efsa.2018.5377
  df$S <- exp(-df$H) # background hazard rate included in H (if enabled)
  df
}
#' @include solver.R
#' @describeIn solver Numerically integrates GUTS-RED-SD scenarios
setMethod("solver", "GutsRedSd", function(scenario, ...) solver_gutsredsd(scenario, ...))

# @param scenario Scenario object
# @param method string, numerical solver used by [deSolve::ode()]
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_gutsredit <- function(scenario, method="lsoda", hmax=1, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # check if all required parameters are present
  params_missing <- is.na(params[scenario@param.req])
  if(any(params_missing)) {
    stop("Missing parameters: ", paste(scenario@param.req[params_missing], collapse=", "))
  }

  # make sure that parameters are present and in required order
  odeparams <- params[c("kd","hb")]

  # CMax := cumulative maximum of D(t)
  CMax_init <- scenario@init[["D"]]
  df <- ode2df(ode(y=c(scenario@init, "CMax"=CMax_init), times=scenario@times, parms=odeparams, dllname="cvasi",
                initfunc="gutsredit_init", func="gutsredit_func", initforc="gutsredit_forc",
                forcings=scenario@exposure@series, outnames=c("Cw"),
                method=method, hmax=hmax, ...))

  # The internal damage can, by definition, not become negative. However, the
  # numerical schemes may sometimes return negative values due to interpolation.
  df$CMax <- pmax(0, df$CMax)

  # Derive survival probability, EFSA Scientific Opinion on TKTD models, p. 33
  # doi:10.2903/j.efsa.2018.5377
  FS <- (1 / (1 + (df$CMax / params["alpha"])^(-params["beta"])))
  df$S <- (1 - FS) * exp(-df$H)
  df$CMax <- NULL
  df
}
#' @describeIn solver Numerically integrates GUTS-RED-IT scenarios
setMethod("solver", "GutsRedIt", function(scenario, ...) solver_gutsredit(scenario, ...) )


########################
## Effects
########################

# Calculate effect of GUTS-RED-SD scenario
fx_gutsredsd <- function(scenario, ...) {
  # we save the control run if we just set the background mortality to zero
  # as it would cancel out, anyways
  if(scenario@param$hb > 0)
    scenario@param$hb <- 0

  res <- simulate(scenario, ...)
  if(!num_success(res)) {
    return(c("L"=NA_real_))
  }

  c("L"=1 - tail(res$S, n=1))
}

# Calculate effect of GUTS-RED-IT scenario
fx_gutsredit <- function(scenario, ...) {
  # we avoid the control run if we just set the background mortality to zero
  # as it would cancel out anyways
  if(scenario@param$hb > 0) {
    scenario <- set_param(scenario, c(hb=0))
  }

  res <- simulate(scenario, ...)
  if(!num_success(res)) {
    return(c("L"=NA_real_))
  }

  c("L"=1 - tail(res$S, n=1))
}

#' @include fx.R
#' @describeIn fx Calculates lethality of [GUTS-RED-SD] scenarios
setMethod("fx", "GutsRedSd", function(scenario, ...) fx_gutsredsd(scenario, ...))
#' @describeIn fx Calculates lethality of [GUTS-RED-IT] scenarios
setMethod("fx", "GutsRedIt", function(scenario, ...) fx_gutsredit(scenario, ...))

test_that("arg sets", {
  # empty set
  rs <- eval_cs(list(), output="foo")
  expect_contains(names(rs), c("pred","obs","wgts","times","tags")) # partial check only
  expect_length(rs$obs, 0)
  expect_length(rs$pred, 0)
  expect_length(rs$times, 0)

  # single set
  sc <- GUTS_RED_IT()
  cs <- caliset(sc, data.frame(t=0:2, o=3:5), weight=6:8, tag="foo")
  pred <- 1:3
  rs <- with_mocked_bindings(eval_cs(list(cs), output="bar"),
    simulate=function(x, ...) {
      data.frame(time=get_times(x), "bar"=pred)
    })
  expect_equal(rs$obs, cs@data[, 2])
  expect_equal(rs$pred, pred)
  expect_equal(rs$times, cs@data[, 1])
  expect_equal(rs$wgts, cs@weight)
  expect_equal(rs$tags, as.list(rep(cs@tag, nrow(cs@data))))

  # multiple sets
  cs2 <- caliset(sc, data.frame(t=1:3, o=4:6), weight=7:9, tag="bar")
  rs <- with_mocked_bindings(eval_cs(list(cs, cs2), output="bar"),
    simulate=function(x, ...) {
     data.frame(time=get_times(x), "bar"=pred)
    })
  expect_equal(rs$obs, c(cs@data[, 2], cs2@data[, 2]))
  expect_equal(rs$pred, c(pred, pred))
  expect_equal(rs$times, c(cs@data[, 1], cs2@data[, 1]))
  expect_equal(rs$wgts, c(cs@weight, cs2@weight))
  expect_equal(rs$tags, as.list(c(rep(cs@tag, nrow(cs@data)),
                                rep(cs2@tag, nrow(cs2@data)))))
})

test_that("arg output", {
  sc <- GUTS_RED_IT()
  cs <- caliset(sc, data.frame(t=0:2, o=3))
  rs <- with_mocked_bindings(eval_cs(list(cs), output="baz"),
    simulate=function(x, ...) {
     data.frame(time=get_times(x), bar=8, baz=9)
    })
  expect_equal(rs$pred, c(9, 9, 9))
})

test_that("arg method/ode_method", {
  sc <- GUTS_RED_IT()
  cs <- caliset(sc, data.frame(t=0:1, o=3))
  fsim <- function(x, method=0, ...) data.frame(time=get_times(x), foo=method)

  # no method requested
  rs <- with_mocked_bindings(eval_cs(list(cs), output="foo"), simulate=fsim)
  expect_equal(rs$pred, c(0, 0))
  # arg 'method' set
  rs <- with_mocked_bindings(eval_cs(list(cs), output="foo", method="bar"), simulate=fsim)
  expect_equal(rs$pred, c("bar", "bar"))
  # arg 'ode_method' set
  rs <- with_mocked_bindings(eval_cs(list(cs), output="foo", ode_method="bar"), simulate=fsim)
  expect_equal(rs$pred, c("bar", "bar"))
  # arg 'ode_method' overwrites arg 'method'
  rs <- with_mocked_bindings(eval_cs(list(cs), output="foo", ode_method="bar", method="baz"), simulate=fsim)
  expect_equal(rs$pred, c("bar", "bar"))
})

test_that("arg verbose", {
  sc <- GUTS_RED_IT()
  cs <- caliset(sc, data.frame(t=0:1, o=3))
  fsim <- function(x, method=0, ...) data.frame(time=get_times(x), foo=NA)

  # message
  expect_message(with_mocked_bindings(eval_cs(list(cs), output="foo", verbose=TRUE), simulate=fsim), "contains NA")
  # no message
  expect_no_message(with_mocked_bindings(eval_cs(list(cs), output="foo", verbose=FALSE), simulate=fsim))
})

test_that("arg .suppress", {
  sc <- GUTS_RED_IT()
  cs <- caliset(sc, data.frame(t=0:1, o=3))
  ferror <- function(x, ...) stop("foobar")

  # error: try-error object
  expect_warning(
      with_mocked_bindings(rs <- eval_cs(list(cs), output="foo", verbose=FALSE, .suppress=FALSE), simulate=ferror),
      "with caliset #1"
  )
  expect_true(rs$is_error)
  expect_false(rs$is_issue)
  expect_equal(rs$err_msg, "foobar")

  # issue: desolve aborted
  faborted <- function(x, ...) {
    df <- data.frame(time=0:1, foo=0)
    attr(df, "desolve_diagn") <- list(istate=-1)
    df
  }
  with_mocked_bindings(rs <- eval_cs(list(cs), output="foo", verbose=FALSE, .suppress=TRUE), simulate=faborted)
  expect_false(rs$is_error)
  expect_true(rs$is_issue)
  expect_equal(rs$err_msg, "simulation terminated early")
})

#
# DEBtox model as implemented in BYOM, DEBtox2019 module, version 4.7, last
# modified 13 Dec 2022. Source: https://www.debtox.info/byom.html
#
# See Jager (2020, DOI: 10.1016/j.ecolmodel.2019.108904 ) for details.
#


########################
## Class definition
########################

#' @include man-deb.R
#' @export
setClass("DebTox", contains="Deb")


########################
## Constructor
########################

#' DEBtox model
#'
#' Creates a  *DEBtox* scenario as described by Jager (2020). It
#' represents a simplified *DEBtox* model based on *DEBkiss*. In the *BYOM*
#' application \[[link](https://www.debtox.info/byom.html)\], this model is
#' referred to as *DEBtox 2019*, version 4.7.
#' It supports an optional feature of the *ERA special* model variant, which
#' can consider a reference *Lm* parameter to compare results of multiple
#' datasets.
#'
#' ## State variables
#'
#' The following list describes the names and units of the
#' model's state variables:
#'  - `D`, scaled damage (\[C\])
#'  - `L`, body length (mm)
#'  - `R`, cumulative reproduction (-)
#'  - `S`, survival probability (-)
#'
#' State variables `D`, `L`, and `R` are initialized with zero. Variable `S`
#' is initialized with one (`1.0`). See [set_init()] on how to set
#' the initial state manually.
#'
#' ## Parameters
#'
#' The following parameters are required:
#'  - General
#'    - `L0`, body length at start (mm)
#'    - `Lp`, body length at puberty (mm)
#'    - `Lm`, maximum body length (mm)
#'    - `rB`, von Bertalanffy growth rate constant (1/d)
#'    - `Rm`, maximum reproduction rate (#/d)
#'    - `f`, scaled functional response (-)
#'    - `hb`, background hazard rate (d-1)
#'    - `a`, Weibull background hazard coefficient (-). Set to `1` to disable.
#'  - Extra parameters
#'    - `Lf`, body length at half-saturation feeding (mm)
#'    - `Lj`, body length at which acceleration stops (mm)
#'    - `Tlag`, lag time for start development (d)
#'  - TK/TD parameters
#'    - `kd`, dominant rate constant (d-1)
#'    - `zb`, effect threshold energy budget (\[C\])
#'    - `bb`, effect strength energy-budget effects (1/\[C\])
#'    - `zs`, effect threshold survival (\[C\])
#'    - `bs`, effect strength survival (1/(\[C\] d))
#'  - Other parameters (formerly globals in *BYOM*)
#'    - `FBV`, dry weight egg as fraction of structural body weight (-)
#'    - `KRV`, part. coeff. repro buffer and structure (kg/kg) (for losses with reproduction)
#'    - `kap`, approximation for kappa (for starvation response)
#'    - `yP`, product of yVA and yAV (for starvation response)
#'    - `Lm_ref`, optional reference max length for scaling rate constants (mm).
#'       Set to zero to disable the reference length. Disabled by default.
#'    - `len`, a switch to control body length dynamics: `1` organism can shrink,
#'         `2` organism cannot shrink. Default value is `1`.
#'    - `Tbp`, optional brood-pouch delay (d). Set to `NA` or zero to disable.
#'          Default value is `0`.
#'    - `MoA`, mode of action switches (-). Default value is `0`.
#'    - `FB`, feedback on damage dynamics switches (-). Default value is `0`.
#'
#' A reference `Lm_ref` is needed to properly compare different data sets,
#' or when calibrating on more than one data set. If `Lm` differs, one would not
#' want to have different rate constants at the same length.
#'
#' ### Mode of Action
#'
#' Any combination of the following mode of actions (*MoA*) can be considered by
#' the model:
#'  - `MoA = 1`: assimilation/feeding
#'  - `MoA = 2`: costs for maintenance
#'  - `MoA = 4`: costs for growth and reproduction
#'  - `MoA = 8`: costs for reproduction
#'  - `MoA = 16`: hazard for reproduction
#'
#' To activate more than one mode of action, simply add up the corresponding
#' codes and set parameter `MoA` to the desired value. To disable all mode of
#' actions, set parameter `MoA` to zero. See also [set_moa()].
#'
#' As an example, to consider effects on feeding and maintenance, set the
#' mode of action to three (`3`):
#'
#' `DEBtox() %>% set_param(c(MoA=3))`
#'
#' ### Feedbacks
#'
#' Any combination of the following damage feedbacks can be considered by
#' the model:
#'  - `1`: surf:vol scaling uptake rate
#'  - `2`: surf:vol scaling elimination rate
#'  - `4`: growth dilution
#'  - `8`: losses with reproduction
#'
#' To activate more than one feedback, simply add up the corresponding
#' codes. To disable all feedbacks, set the parameter to zero.
#'
#' ## Effects
#'
#' The state variables *L* (body length), *R* (cumulative reproduction), and
#' *S* (survival probability) are set as effect endpoints by default. All state
#' variables are available as potential endpoints. The list of considered
#' endpoints can be modified by using [set_endpoints()].
#'
#' To calculate effects, each *DEBtox* scenario is simulated twice: One simulation
#' which considers exposure to a toxicant and one simulation without exposure, i.e.
#' a control. See also [effect()].
#'
#' ## Simulation output
#'
#' The following intermediary model variables can be added to the model
#' output on demand. Simply set the optional parameter `nout` to the
#' required output level and pass it to [simulate()].
#'
#'  - `nout` >= 1: `f`, actual scaled response
#'  - `nout` >= 2: `fR`, actual f considering starvation
#'  - `nout` >= 3: `kd`, actual kd
#'  - `nout` >= 4: `s`, stress level
#'  - `nout` >= 5: `h`, hazard rate
#'  - `nout` >= 6: `sA`, stress factor on assimilation/feeding
#'  - `nout` >= 7: `sM`, stress factor on maintenance
#'  - `nout` >= 8: `sG`, stress factor on growth costs
#'  - `nout` >= 9: `sR`, stress factor on reproduction costs
#'  - `nout` >= 10: `sH`, stress factor on hazard to reproduction
#'  - `nout` >= 11: `xu`, damage feedback factor for surf:vol scaling uptake rate
#'  - `nout` >= 12: `xe`, damage feedback factor for surf:vol scaling elimination rate
#'  - `nout` >= 13: `xG`, damage feedback factor for growth dilution
#'  - `nout` >= 14: `xR`, damage feedback factor for losses with repro
#'
#' ## Solver settings
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `method = 'ode45'`<br>
#'    Selects the Dormand-Prince 4(5) method of the Runge-Kutta family, see
#'    [deSolve::rkMethod()] for details.
#'
#' ## Model history and changes
#'
#' - cvasi v1.0.0
#'    - The `DEB_Daphnia()` model implemented BYOM's *DEBtox 2019*
#'      model version 4.5
#' - cvasi v1.2.0
#'    - The model equations were updated to conform with BYOM's *DEBtox 2019*
#'      version 4.7. This introduced a new model parameter `a`, the Weibull
#'      background hazard coefficient, and limited the maximum hazard rate to
#'      99% per hour.
#'    - The scenario constructor was renamed to `DEBtox()`.
#'    - Additional intermediary model variables available as optional simulation
#'      output
#'
#' @return an S4 object of type [DebTox-class]
#' @export
#' @family DEB models
#' @aliases DEBtox DEBtox2019 DebTox-class DEB_Daphnia DebDaphnia-class
#' @references
#' Jager T, 2020: Revisiting simplified DEBtox models for analysing
#' ecotoxicity data. Ecol Model 416. \doi{10.1016/j.ecolmodel.2019.108904}
#'
#' Romoli et al., 2024: Environmental risk assessment with energy budget
#' models: a comparison between two models of different complexity.
#' Environ Toxicol Chem 43(2):440-449. \doi{10.1002/etc.5795}
DEBtox <- function() {
  new("DebTox",
      name="DEBtox",
      param=list("Lm_ref"=0, "len"=1, "Tbp"=0, "MoA"=0, "FB"=0),
      param.req=c("L0", "Lp", "Lm", "rB", "Rm", "f", "hb", "a", "Lf", "Lj", "Tlag",
                  "kd", "zb", "bb", "zs", "bs", "FBV", "KRV", "kap", "yP", "Lm_ref",
                  "len", "Tbp", "MoA", "FB"),
      endpoints=c("L","R","S"),
      control.req=TRUE,
      init=c("D"=0, "L"=0, "R"=0, "S"=1)
  ) %>% set_noexposure()
}


########################
## Simulation
########################

#' @importFrom deSolve ode
solver_debtox <- function(scenario, method="ode45", ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  if(scenario@init[["L"]] < params[["L0"]])
    warning("initial L is smaller than parameter L0")

  # make sure that parameters are present and in required order
  params_req <- c("L0", "Lp", "Lm", "rB", "Rm", "f", "hb", "a", "Lf", "Lj", "Tlag",
                  "kd", "zb", "bb", "zs", "bs", "FBV", "KRV", "kap", "yP",
                  "Lm_ref", "MoA", "FB")
  # names of additional output columns
  outnames <- c("f", "fR", "kd", "s", "h", "sA", "sM", "sG", "sR", "sH",
                "xu", "xe", "xG", "xR")

  # check for missing parameters
  params.missing <- is.na(params[params_req])
  if(any(params.missing))
    stop(paste("missing parameters:", paste(params_req[params.missing], sep=",", collapse=",")))

  # This is a quick check to make sure that acceleration cannot continue
  # after puberty. So, Lj should be smaller than Lp. This is done to ensure
  # consistency with standard DEB, where acceleration stops at a maturity
  # level, and maturity does not increase anymore after puberty. However,
  # there is no strong theoretical reason why animals should stop
  # accelerating at puberty, so this part could be commented out.
  if(params[["Lj"]] > params[["Lp"]])
    stop("Lj > Lp, acceleration cannot continue after puberty")

  # only one forcings time-series: exposure
  forcings <- list(scenario@exposure@series)

  # This is a means to include a delay caused by the brood pouch in species
  # like Daphnia. The repro data are for the appearance of neonates, but egg
  # production occurs earlier. This globally shifts the model output in this
  # function below. This way, the time vector in the data does not need to be
  # manipulated, and the model plots show the neonate production as expected.
  has_bp <- min(params[["Tbp"]], na.rm=TRUE) > 0 # consider brood-pouch delay?
  times <- scenario@times
  if(has_bp) {
    # Extra times needed to calculate brood-pouch delay
    tbp <- times[times > params[["Tbp"]]] - params[["Tbp"]]
    ordering <- order(c(times, tbp))
    # Boolean vector to identify newly added time points
    loct <- c(rep(F, length(times)), rep(T, length(tbp)))[ordering]
    # Full vector with original and new time points, sorted
    times <- sort(c(times, tbp))
  }

  # run solver
  out <- ode2df(ode(y=scenario@init, times=times, parms=params[params_req], dllname="cvasi",
             initfunc="debtox_init", func="debtox_func", initforc="debtox_forc",
             forcings=forcings, outnames=outnames, method=method, ...))

  # When animal cannot shrink in length (but does on weight!)
  if(min(params[["len"]], na.rm=TRUE) == 2) {
    # Find cumulative maximum of body length and replace values in output
    out[, "L"] <- cummax(out[, "L"])
  }

  # Make sure survival does not get negative
  # In some cases, it may become just a bit negative, which means zero.
  out[, "S"] <- pmax(0, out[, "S"])

  # If we need a brood-pouch delay ...
  if(has_bp) {
    # Find where the extra brood-pouch time points are in the output
    Xbp <- out[loct, "R"]
    # Sub-select the output to return only data for the original time points
    out <- out[!loct, ]
    # Put in the brood-pouch reproduction values we asked for, initial lag phase
    # set to zero.
    out[, "R"] <- c(rep(0, nrow(out) - length(Xbp)), Xbp)
  }

  out
}

#' @include solver.R
#' @describeIn solver Numerically integrates *DEBtox* scenarios
setMethod("solver", "DebTox", solver_debtox)


########################
## Effects
########################

# uses the default fx implementation, nothing more required

# simulations to verify the model implementation
test_that("verify: constant exposure", {
  source(test_path("init-simulate_Daphnia.R"), local=TRUE)

  param <- tfs_param
  param$f <- 0.88

  tfs_result %>%
    dplyr::filter(set=="set3" & run=="control") %>%
    dplyr::pull(time) -> times

  for(i in seq(2, length(tfs_exposure$cst))) {
    exposure <- tfs_exposure$cst[,c(1,i)]

    DEBtox() %>%
      set_init(tfs_init) %>%
      set_param(param) %>%
      set_times(times) %>%
      set_exposure(exposure, reset_times = FALSE) %>%
      simulate(nout=0) -> out

    runstr <- ifelse(i == 2, "control", paste0("CST_",(i-2)))
    tfs_result %>%
      dplyr::filter(set == "set3" & run == runstr) %>%
      dplyr::select(!c(set, run)) -> verify
    expect_equal(out, verify, tolerance=0.1, ignore_attr=TRUE)
  }
})

test_that("verify: variable exposure 1", {
  source(test_path("init-simulate_Daphnia.R"), local=TRUE)

  tfs_result %>%
    dplyr::filter(set=="set1" & run=="control") %>%
    dplyr::pull(time) -> times

  for(i in seq(2, length(tfs_exposure$var1))) {
    exposure <- tfs_exposure$var1[,c(1,i)]

    DEBtox() %>%
      set_init(tfs_init) %>%
      set_param(tfs_param) %>%
      set_times(times) %>%
      set_exposure(exposure, reset_times = FALSE) %>%
      simulate() -> out

    runstr <- ifelse(i == 2, "control", paste0("VAR1_",(i-2)))
    tfs_result %>%
      dplyr::filter(set == "set1" & run == runstr) %>%
      dplyr::select(!c(set, run)) -> verify
    expect_equal(out, verify, tolerance=1e-3, ignore_attr=TRUE)
  }
})

test_that("verify: variable exposure 2", {
  source(test_path("init-simulate_Daphnia.R"), local=TRUE)

  tfs_result %>%
    dplyr::filter(set=="set2" & run=="control") %>%
    dplyr::pull(time) -> times

  for(i in seq(2, length(tfs_exposure$var1))) {
    exposure <- tfs_exposure$var2[,c(1,i)]

    DEBtox() %>%
      set_init(tfs_init) %>%
      set_param(tfs_param) %>%
      set_times(times) %>%
      set_exposure(exposure, reset_times = FALSE) %>%
      simulate(hmax=0.01) -> out

    runstr <- ifelse(i == 2, "control", paste0("VAR2_",(i-2)))
    tfs_result %>%
      dplyr::filter(set == "set2" & run == runstr) %>%
      dplyr::select(!c(set, run)) -> verify
    expect_equal(out, verify, tolerance=1e-3, ignore_attr=TRUE)
  }
})

# model behavior
test_that("brood-pouch delay", {
  source(test_path("init-simulate_Daphnia.R"), local=TRUE)

  # irregular time points
  dmagna %>%
    set_param(c(Tbp=3)) %>%
    simulate() -> res_bp
  dmagna %>%
    set_param(c(Tbp=0)) %>%
    simulate() -> res_no
  # last time point with zero reproduction should be roughly shifted by Tbp
  expect_equal(
    res_bp %>% dplyr::filter(R==0) %>% dplyr::pull(time) %>% max(),
    res_no %>% dplyr::filter(R==0) %>% dplyr::pull(time) %>% max() + 3,
    tolerance=0.1
  )
  expect_equal(res_bp$L, res_no$L)

  # regular time points
  dmagna %>%
    set_times(0:21) %>%
    set_param(c(Tbp=3)) %>%
    simulate() -> res_bp
  dmagna %>%
    set_times(0:21) %>%
    set_param(c(Tbp=0)) %>%
    simulate() -> res_no
  expect_equal(
    res_bp %>% dplyr::filter(R==0) %>% dplyr::pull(time) %>% max(),
    res_no %>% dplyr::filter(R==0) %>% dplyr::pull(time) %>% max() + 3
  )
  expect_equal(tail(res_bp$R, 19), head(res_no$R, 19))
  expect_equal(res_bp$L, res_no$L)
})

#' @export
setClass("ParameterSet",
         slots=list(
           model="character",
           tag="character",
           param="list"
         )
)

# for backwards compatability
#' @export
setClass("parameter_set", contains="ParameterSet")


#' Set of model parameters
#' @param model `character`, a string containing a model name, e.g. `"GUTS-RED-IT"`
#' @param tag `character`, an optional identifier
#' @param param named `list` of model parameters
#' @slot model `character`, a string containing a model name, e.g. `"GUTS-RED-IT"`
#' @slot tag `character`, an optional identifier
#' @slot param named `list` of model parameters
#' @return an S4 object of type [ParameterSet-class]
#'
#' @export
#' @aliases ParameterSet-class parameter_set-class
#' @examples
#' # create a parameter set and assign it
#' ps <- parameter_set("GUTS-RED-IT", list(kd=0.12, hb=0.3))
#' GUTS_RED_IT() %>% set_param(ps)
#'
#' # multiple scenarios can be modified at once
#' c(GUTS_RED_IT(), GUTS_RED_IT()) %>%
#'   set_param(ps)
#'
#' # model names must match, otherwise an error will be raised
#' try(GUTS_RED_SD() %>% set_param(ps))
parameter_set <- function(model, param=list(), tag=NA_character_) {
  if(!is.character(model))
    stop("model name must be of type character")
  if(!is.character(tag))
    stop("model tag must be of type character")
  if(length(dim(param)) > 0)
    stop("list of parameters must not be a matrix or data.frame")
  # convert vector to list
  if(is.vector(param) & !is.list(param)) {
    param <- as.list(param)
  }
  # check that list only contains atomic types, otherwise we might accept weird objects
  if(!all(sapply(param, is.atomic))) {
    stop("list of parameters must only  contain atomic types")
  }

  new("ParameterSet",
      model=model,
      tag=tag,
      param=as.list(param)
  )
}

test_that("set parameters", {
  nm <- "foo"
  tag <- "bar"
  p1 <- c("a"=1, "b"=2)
  p2 <- list("a"=1, "b"=2)

  # name
  expect_equal(parameter_set(nm)@model, nm)
  # tag
  expect_equal(parameter_set(nm, tag=tag)@tag, tag)
  # parameters
  expect_equal(parameter_set(nm, param=p1)@param, p2) # must be a list
  expect_equal(parameter_set(nm, param=p2)@param, p2)
})


test_that("invalid parameters", {
  # name: invalid type
  expect_error(parameter_set(list()))
  # name missing
  expect_error(parameter_set(tag="bar"))
  # tag: invalid type
  expect_error(parameter_set("foo", tag=list()))
  # param: invalid type
  expect_error(parameter_set("foo", param=data.frame()))
  expect_error(parameter_set("foo", param=list(data.frame())))
  expect_error(parameter_set("foo", param=list("a"=1, "b"=list())))
})

#' Set time-dependent parameters
#'
#' Parameters which change their value over time are referred to as *forcings*.
#' If and what parameters can vary over time depends on the model in question.
#' In many cases, *forcings* represent time-series of environmental properties.
#'
#' Forcing time-series are always represented by a
#' `data.frame` containing two columns. The first column representing time,
#' the second representing the parameter that is a function of time. The
#' ordering of columns is mandatory. The column names are essentially irrelevant
#' but may help documenting the scenario and its data. The rows must be
#' ordered chronologically. A time-series can consist of only a single row; in
#' this case it will represent constant conditions.
#'
#' Handling forcing time-series is a costly task for the ODE solver due to consistency
#' checks and interpolation between timesteps. How the solver interpolates
#' the forcing time-series can be controlled by certain arguments to functions
#' such as [simulate()] and [effect()]. Please refer to [simulate()] for a brief
#' overview and [deSolve::forcings] for a detailed description.
#'
#' Forcing time-series should be kept as short as possible and as complex as
#' needed for optimal computational efficiency.
#'
#' @param x (vector of) [scenario] objects
#' @param ... named argument list to set as forcings
#'
#' @return Modified [scenarios]
#' @export
#' @include class-EffectScenario.R
#' @rdname set_forcings
#' @examples
#' # constant values will be automatically converted to a data.frame
#' Lemna_Schmitt() %>% set_forcings(temp=20) -> lemna
#' lemna@forcings
#'
#' # setting multiple forcings at once
#' df <- data.frame(t=0:14, temp=rnorm(15, mean=20)) # random temperature series
#' Lemna_Schmitt() %>% set_forcings(temp=df, rad=15000) -> lemna
#' lemna@forcings
#'
#' # forcings can also be supplied as a named list
#' Lemna_Schmitt() %>% set_forcings(list(temp=20, rad=15000)) -> lemna
#' lemna@forcings
setGeneric("set_forcings", function(x,...) standardGeneric("set_forcings"), signature="x")

# For EffectScenario objects
#' @rdname set_forcings
setMethod("set_forcings", "EffectScenario", function(x, ...) set_forc_default(x, ...))
# For vectorized input/scenarios
#' @rdname set_forcings
setMethod("set_forcings", "list", function(x, ...) set_forc_list(x, ...))

#
# Set forcings for any kind of EffectScenario object
#
#' @importFrom rlang enquos eval_tidy
set_forc_default <- function(x, ...) {
  lst <- list()
  qqs <- rlang::enquos(...)
  nms <- names(qqs)
  # use standard or non-standard evaluation?
  # if 1st and only argument is a list but not a data.frame, then use std eval
  if(length(qqs)==1 &
     is.list(rlang::eval_tidy(qqs[[1]])) &
     !is.data.frame(rlang::eval_tidy(qqs[[1]]))) {
    lst <- rlang::eval_tidy(qqs[[1]])
    nms <- names(lst)
  } else { # otherwise look for any number of named arguments
    for(i in 1:length(qqs))
      lst <- append(lst, list(rlang::eval_tidy(qqs[[i]])))
    names(lst) <- nms
  }

  unused <- setdiff(nms, get_req_forcings(x))
  if(length(unused)>0)
    warning(paste("unused forcing series:",paste(unused,sep=",",collapse=",")))
  ints <- intersect(nms, get_req_forcings(x))
  if(length(lst)-length(unused)-length(ints)>0)
    warning("unnamed forcing series were ignored")

  for(nm in ints) {
    data <- lst[[nm]]
    if(is.numeric(data) & length(data) == 1) {
      data <- data.frame(t=c(0), V1=c(data))
      names(data) <- c("time", nm)
    }
    if(!is.data.frame(data))
      stop(paste("forcing series",nm,"has invalid type"))
    if(length(data)!=2)
      stop(paste("forcing series",nm,"must have two columns"))
    x@forcings[[nm]] <- data
  }
  x
}

set_forc_list <- function(x, ...) {
  return(lapply(x, function(sc) set_forcings(sc, ...)))
}

#' @include class-EffectScenario.R
NULL

#' Set effect endpoints
#'
#' Effect endpoints calculated by functions such as [effect()] and [epx()]
#' can be enabled and disabled. If an endpoint is not required for an assessment,
#' it should be disabled for reasons of computational efficiency. Please refer
#' to the model description for a list of available endpoints.
#'
#' @param x vector of `EffectScenario` objects
#' @param endpoints `character` vector of endpoint names
#'
#' @return Modified `EffectScenario` objects
#' @export
#' @examples
#' # Only enable reproduction (R) endpoint for americamysis scenario
#' americamysis %>%
#'   set_endpoints("R") %>%
#'   effect()
#'
#' # Enable endpoints length (L) and reproduction (R)
#' americamysis %>%
#'   set_endpoints(c("L","R")) %>%
#'   effect()
set_endpoints <- function(x, endpoints) {
  if(is.vector(x)) {
    return(sapply(x, function(sc) set_endpoints(sc,endpoints)))
  }
  if(length(endpoints) == 0)
    return(x)

  x@endpoints <- unlist(endpoints)
  x
}

#' Set mode of action
#'
#' Updates the model parameter `MoA` to a certain value
#'
#' @param x vector of [scenarios]
#' @param code a code for a mode of action, refer to model description for
#'   details
#'
#' @return modified [scenarios]
#' @export
#'
#' @examples
#' # Set MoA=8, i.e. hazard during oogenesis
#' americamysis %>%
#'   set_mode_of_action(8) %>%
#'   effect(method="ode45")
#'
#' # alternative approach using the parameter directly
#' americamysis %>%
#'   set_param(c(MoA=8)) %>%
#'   effect(method="ode45")
set_mode_of_action <- function(x, code) {
  if(is.vector(x))
    return(sapply(x, set_mode_of_action, code))

  if(!("MoA" %in% x@param.req))
    stop("model does not support modes of action")
  set_param(x, c(MoA=code))
}

#' @describeIn set_mode_of_action Shorthand version
#' @export
set_moa <- function(x, code) {
  set_mode_of_action(x, code)
}


# Set/copy all slots to a new scenario
#
# Can be used to copy scenario properties to a new scenario object of another
# model, e.g. copy properties of a Lemna_Schmitt scenario to Lemna_SchmittThold.
#
# @param x an `EffectScenario` to modify
# @param tmpl an `EffectScenario` to copy properties from
# @return modified `EffectScenario` objects
# @examples
# # Apply scenario properties from one model to another
# Lemna_SchmittThold() %>%
#   set_all(metsulfuron)
#' @importFrom methods slot slotNames slot<-
set_all <- function(x, tmpl) {
  if(is.vector(x))
    return(sapply(x, set_all, tmpl))

  if(!is_scenario(x) | !is_scenario(tmpl))
    stop("argument is not an EffectScenario object")

  trgt_slots <- slotNames(x)
  tmpl_slots <- slotNames(tmpl)
  exclude <- c("name","param.req","forcings.req")

  for(nm in intersect(trgt_slots, tmpl_slots)) {
    if(nm %in% exclude) next

    slot(x, nm) <- slot(tmpl, nm)
  }

  x
}


#' Set a tag
#'
#' Sets the user-defined, custom tag of a scenario. Tags
#' can be helpful to quickly distinguish scenarios by e.g. a user-specified
#' string.
#'
#' The function supports vectorized inputs.
#' If arguments `x` and `tag` are vectors, then tags are assigned to scenarios
#' on a 1:1 basis. If the length of both vectors do not match, an error is raised.
#'
#' @param x (vector of) [scenario] objects
#' @param tag (vector of) `character` or any other object
#'
#' @return (vector of) modified [scenarios]
#' @seealso [get_tag()]
#' @export
#'
#' @examples
#' # set a custom tag
#' myscenario <- GUTS_RED_SD() %>% set_tag("My Custom Tag")
#'
#' # returns `My Custom Tag`
#' get_tag(myscenario)
#'
#' # the tag also appears in the scenario overview
#' myscenario
set_tag <- function(x, tag) {
  if(is.vector(x)) {
    if(length(tag) == 1)
      return(sapply(x, set_tag, tag=tag))
    else if(length(tag) == length(x))
      return(purrr::map2(x, tag, set_tag))
    else
      stop("Argument `x` and `tag` must have the same length", call.=FALSE)
  }

  if(!is_scenario(x))
    stop("Argument `x` is not a scenario", call.=FALSE)
  if(length(tag) > 1)
    stop("Argument `x` and `tag` must have the same length", call.=FALSE)

  x@tag <- tag
  x
}

effect4tolerance <- function(tolerance) {
  mf <- epx(minnow_it,level=50,effect_tolerance=tolerance,ep_only=TRUE)[[1,1]]
  effect(minnow_it, factor=mf, ep_only=TRUE)
}

test_that("effect levels", {
  # avoid some common bogus messages on package load
  suppressMessages(suppressWarnings(require(purrr, quietly=TRUE)))

  ## requested effect levels
  expect_named(epx(minnow_it,level=23),c("scenario","L.EP23"))
  expect_named(epx(minnow_it,level=c(10,20)),c("scenario","L.EP10","L.EP20"))
})

test_that("effect_tolerance", {
  ## effect.tolerance
  ## result are getting more precise if we decrease the tolerance
  exact <- effect4tolerance(1e-10)
  expect_gt(abs(exact-effect4tolerance(1e-1)), abs(exact-effect4tolerance(1e-2)))
  expect_gt(abs(exact - effect4tolerance(1e-1)), 0.01)
  expect_lt(abs(exact - effect4tolerance(1e-1)), 1e-1)
  expect_lt(abs(exact - effect4tolerance(1e-2)), 1e-2)
  expect_lt(abs(exact - effect4tolerance(1e-3)), 1e-3)
  expect_lt(abs(exact - effect4tolerance(1e-4)), 1e-4)
  rm(exact)
})

test_that("min/max factor", {
  scen <- set_exposure(minnow_it, data.frame(t=1:5,c=1e-5)) # low exposure
  expect_warning(epx(scen,level=50,max_factor=1e3),regexp="multiplication factor out of range")
  scen <- set_exposure(minnow_it, data.frame(t=1:5,c=1e5)) # high exposure
  expect_warning(epx(scen,level=50,min_factor=1e-3),regexp="multiplication factor out of range")
  rm(scen)
  # one of two effect level fails, i.e. BM.EP50
  expect_warning(metsulfuron %>% set_endpoints("BM") %>% epx())
  suppressWarnings(rs <- metsulfuron %>% set_endpoints("BM") %>% epx())
  expect_equal(rs[[1,"BM.EP10"]], 0.395, tolerance=0.1)
  expect_equal(rs[[1,"BM.EP50"]], NA_real_)
})

test_that("factor cutoff", {
  cutoff <- 2000
  scen <- set_exposure(minnow_it, data.frame(t=1:5,c=1e-5))
  expect_true(all(epx(scen,ep_only=TRUE) > cutoff))
  expect_true(all(epx(scen,factor_cutoff=cutoff,ep_only=TRUE) == cutoff))
  rm(cutoff)

})

test_that("ep_only", {
  source(test_path("dummy.R"), local = TRUE)

  sc <- new("DummyScenario", endpoints="foo", fx=function(scenario, ...) {
    return(c(foo=scenario@exposure@series[1, 2]))
  })

  # single scenario, single EPx level
  rs <- epx(sc, level=10, ep_only=TRUE)
  expect_equal(nrow(rs), 1)
  expect_equal(names(rs), "foo.EP10")
  expect_equal(rs[[1, 1]], 0.1)

  # two scenarios, a two EPx levels
  rs <- epx(c(sc, sc), level=c(10, 50), ep_only=TRUE, effect_tolerance = 1e-5)
  expect_equal(nrow(rs), 2)
  expect_equal(names(rs), c("foo.EP10","foo.EP50"))
  expect_equal(rs[[1, 1]], 0.1)
  expect_equal(rs[[1, 2]], 0.5, tolerance=1e-5)
})

test_that("long_format", {
  source(test_path("dummy.R"), local = TRUE)
  sc <- new("DummyScenario", endpoints="foo", fx=function(scenario, ...) {
    return(c(foo=scenario@exposure@series[1, 2]))
  })

  # single scenario, single EPx level
  rs <- epx(sc, level=10, long_format=TRUE)
  expect_equal(nrow(rs), 1)
  expect_equal(names(rs), c("scenario","endpoint","level","EPx"))
  # single scenario, two EPx levels
  rs <- epx(sc, level=c(10, 50), long_format=TRUE)
  expect_equal(nrow(rs), 2)
  # two scenarios, a two EPx levels
  rs <- epx(c(sc,sc), level=c(10, 50), long_format=TRUE)
  expect_equal(nrow(rs), 4)
  expect_equal(names(rs), c("scenario","endpoint","level","EPx"))
  # again, but only endpoints=TRUE
  rs <- epx(c(sc,sc), level=c(10, 50), long_format=TRUE, ep_only=TRUE)
  expect_equal(nrow(rs), 4)
  expect_equal(names(rs), c("endpoint","level","EPx"))
})

test_that("error handling", {
  source(test_path("dummy.R"), local = TRUE)

  # single scenario, simulation fails
  sc <- new("DummyScenario", endpoints="foo", fx=function(...) stop("planned error"))

  expect_warning(rs <- epx(sc, level=10))
  expect_equal(names(rs), c("scenario","foo.EP10", "error"))
  expect_true(is.na(rs[['foo.EP10']]))
  expect_equal(rs[['error']], "planned error")

  # single scenario, two levels, one without valid results
  sc <- new("DummyScenario", endpoints="foo", fx=function(scenario, ...) {
    c <- scenario@exposure@series[1, 2]
    if(c > 60) return(NA_real_)
    return(c(foo=c/100))
  })

  expect_warning(rs <- epx(sc, level=c(10,90)), "Some scenarios have failed")
  expect_equal(names(rs), c("scenario","foo.EP10", "foo.EP90"))
  expect_equal(rs[['foo.EP10']], 10)
  expect_true(is.na(rs[['foo.EP90']]))

  # two scenarios, one passes, one fails
  sc1 <- new("DummyScenario", endpoints="foo", simulate=function(...) stop("planned error"))
  sc2 <- new("DummyScenario", endpoints="foo", fx=function(scenario, ...) {
    return(c(foo=scenario@exposure@series[1, 2]))
  })

  expect_warning(rs <- epx(list(sc1, sc2))) # continue with warning
  expect_true(all(!is.na(rs[2,c("foo.EP10","foo.EP50")]))) # EPx present
  expect_true(is.na(rs[2,"error"])) # no error msg
  expect_true(all(is.na(rs[1,c("foo.EP10","foo.EP50")]))) # no EPx
  expect_true(!is.na(rs[1,"error"])) # error msg
  rm(rs)

  # two scenarios, both failed
  expect_warning(rs <- epx(list(sc1, sc1), level=10))
  expect_equal(names(rs), c("scenario","foo.EP10", "error"))
  expect_true(all(is.na(rs[1:2,2])))
  expect_true(all(!is.na(rs[1:2,3])))
})

test_that("set_window", {
  # set both arguments at once
  Lemna_Schmitt() %>%
    set_window(length=1, interval=2) -> sc
  expect_equal(sc@window.length, 1)
  expect_equal(sc@window.interval, 2)

  # set only a single argument: length
  Lemna_Schmitt() %>%
    set_window(interval=1, length=2) %>%
    set_window(length=42) -> sc
  expect_equal(sc@window.length, 42)
  expect_equal(sc@window.interval, 1)

  # set only a single argument: interval
  Lemna_Schmitt() %>%
    set_window(interval=1, length=2) %>%
    set_window(interval=42) -> sc
  expect_equal(sc@window.length, 2)
  expect_equal(sc@window.interval, 42)

  # disable both by setting one to -1
  Lemna_Schmitt() %>%
    set_window(length=1, interval=2) %>%
    set_window(length=-1) -> sc
  expect_equal(sc@window.length, -1)
  expect_equal(sc@window.interval, -1)

  Lemna_Schmitt() %>%
    set_window(length=1, interval=2) %>%
    set_window(interval=-1) -> sc
  expect_equal(sc@window.length, -1)
  expect_equal(sc@window.interval, -1)

  # vectorized scenario argument
  c(Lemna_Schmitt(), Lemna_Schmitt()) %>% set_window(length=1, interval=2) -> scs
  expect_equal(length(scs), 2)
  expect_equal(scs[[1]]@window.length, 1)
  expect_equal(scs[[2]]@window.length, 1)
  expect_equal(scs[[1]]@window.interval, 2)
  expect_equal(scs[[2]]@window.interval, 2)

  # disable windows
  sc <- new("EffectScenario") %>%
    set_window(length=1, interval=2) %>%
    set_nowindow()
  expect_equal(sc@window.length, -1)
  expect_equal(sc@window.interval, -1)

  sc <- new("EffectScenario") %>%
    set_window(length=1, interval=2)
  scl <- set_nowindow(c(sc, sc))
  expect_equal(scl[[1]]@window.interval, -1)
  expect_equal(scl[[2]]@window.interval, -1)

  # invalid inputs
  sc <- new("EffectScenario")
  expect_error(set_window(sc, length=0))
  expect_error(set_window(sc, length=c(1,2)))
  expect_error(set_window(sc, interval=c(1,2)))
  expect_error(set_window(sc, interval=1))
  expect_error(set_window(sc, length=1))
})

test_that("set_endpoints", {
  # single scenario
  expect_equal(set_endpoints(new("EffectScenario"), "A")@endpoints, "A")
  expect_equal(set_endpoints(new("EffectScenario"), c("A","B"))@endpoints, c("A","B"))

  # vectorized input
  expect_equal(set_endpoints(c(new("EffectScenario"),new("EffectScenario")), "A") %>%
                 sapply(function(x) slot(x,"endpoints")), c("A","A"))

  # invalid input ?
})

test_that("set_forcings", {
  sc <- new("EffectScenario", forcings.req=c("rad","tmp"))
  # std evaluation
  expect_equal(set_forcings(sc, list(rad=data.frame(time=0:2, rad=0)))@forcings, list(rad=data.frame(time=0:2, rad=0)))
  expect_equal(set_forcings(sc, list(rad=1))@forcings, list(rad=data.frame(time=0, rad=1)))
  expect_equal(set_forcings(sc, list(rad=1, tmp=2))@forcings, list(rad=data.frame(time=0, rad=1),
                                                                   tmp=data.frame(time=0, tmp=2)))
  # non-standard evaluation
  expect_equal(set_forcings(sc, rad=1)@forcings, list(rad=data.frame(time=0, rad=1)))
  expect_equal(set_forcings(sc, rad=1, tmp=2)@forcings, list(rad=data.frame(time=0, rad=1),
                                                             tmp=data.frame(time=0, tmp=2)))
  ts <- data.frame(time=c(0,1,2), val=c(2.1,3,4.2))
  expect_equal(set_forcings(sc, rad=ts)@forcings, list(rad=ts))

  # vectorized scenarios
  expect_equal(set_forcings(c(sc,sc), rad=1) %>% sapply(function(x) slot(x, "forcings")),
               list(rad=data.frame(time=0, rad=1), rad=data.frame(time=0, rad=1)))

  # invalid series name
  expect_warning(set_forcings(sc, foo=1), regexp="unused forcing")
  expect_warning(set_forcings(sc, list(foo=1)), regexp="unused forcing")
  # unnamed series
  expect_warning(set_forcings(sc, list(1)), regexp="unnamed forcing")
  # wrong number of columns in data.frame
  expect_error(set_forcings(sc, rad=data.frame(a=0, b=1, c=2)), regexp="two columns")
  # vectorized input in non-std eval
  expect_error(set_forcings(sc, rad=c(1,2,3)), regexp="rad.*invalid type")
})

test_that("set_moa", {
  sc <- new("EffectScenario", param.req=c("foo", "MoA"))
  # std input
  expect_equal(set_mode_of_action(sc, 1)@param, list(MoA=1))
  expect_equal(set_moa(sc, 1)@param, list(MoA=1))
  # vectorized scenarios
  expect_equal(set_moa(c(sc,sc), 1) %>% sapply(function(x) slot(x, "param")), list(MoA=1, MoA=1))

  # MoA not supported
  expect_error(set_moa(new("EffectScenario"), 1), regexp="does not support modes of action")
})

#' Get output times
#'
#' @param x (vector of) [scenario] objects
#'
#' @return (list of) times vector
#' @aliases get_times,ANY-method get_times,list-method  get_times,EffectScenario-method
#'   get_times,ScenarioSequence-method
#' @seealso [set_times()]
#' @include class-EffectScenario.R sequence.R
#' @export
#'
#' @examples
#' # Create a scenario
#' myscenario <- GUTS_RED_IT() %>% set_times(0:5)
#' # Returns the defined output times
#' get_times(myscenario)
setGeneric("get_times", function(x) standardGeneric("get_times"))

#' @export
setMethod("get_times", "ANY", function(x) stop("Object type is not supported"))

#' @export
setMethod("get_times", "list", function(x) lapply(x, get_times))

#' @export
setMethod("get_times", "EffectScenario", function(x) return(x@times))

#' @export
setMethod("get_times", "ScenarioSequence", function(x) get_times_seq(x))

get_times_seq <- function(x) {
  # for backwards compatibility, since v1.5.0
  if(methods::.hasSlot(x, "inc_start")) {
    inc_start <- x@inc_start
    inc_end <- x@inc_end
  } else {
    inc_start <- c(TRUE, rep(FALSE, length(x) - 1))
    inc_end <- rep(TRUE, length(x))
  }

  times <- c()
  for(i in seq_along(x)) {
    tms <- get_times(x[[i]])
    if(!inc_start[[i]]) {
      tms <- tms[ tms != tms[1] ]
    }
    if(!inc_end[[i]]) {
      tms <- tms[ tms != tms[length(tms)] ]
    }
    times <- c(times, tms)
  }
  times
}

#' Biomass transfer class
#'
#' By inheriting from class `Transferable`, a scenario's behavior can be
#' extended to support transfer and reset of biomass at dedicated points during
#' simulation.
#'
#' @section Biomass transfer:
#' Models supporting biomass transfer can be instructed to move a fixed amount
#' of biomass to a new medium after a period of time. This feature replicates
#' a procedure occurring in e.g. *Lemna* effect studies and may be necessary to
#' recreate study results.
#'
#' The biomass transfer feature assumes that always a fixed amount of
#' biomass is transferred. Transfers can occur at any fixed point in time or
#' in regular intervals. During a transfer, the biomass is reset to the
#' transferred amount and additional compartments can be scaled 1:1 accordingly,
#' to e.g. reflect the change in internal toxicant mass when biomass is modified.
#' Transfer settings can be modified using [set_transfer()].
#'
#' If a transfer occurs, simulation results of that time point will report the model state
#' **before** the transfer. Be aware that if transfers are defined using the
#' `interval` argument, the transfers will always occur relative to time point
#' zero (`t = 0`). As an example, setting a regular transfer of seven days,
#' `interval =  7`, will result at transfers occurring at time points which are
#' integer multiplicates of seven, such as `t=0`, `t=7`, `t=14` and so forth.
#' The starting and end times of a scenario do not influece **when** a regular
#' transfer occurs, only **if** it occurs.
#'
#' @seealso [set_transfer()]
#' @slot transfer.times  `numeric`, vector of custom time points at which transfers occur,
#'  e.g. `c(2,5,14)`
#' @slot transfer.interval `numeric`, length of regular interval until biomass transfer to new
#'  medium, regular transfers always occur relative to time point zero
#' @slot transfer.biomass `numeric`, amount of biomass transferred to new medium
#' @slot transfer.comp.biomass `character` state variable which describes
#'   biomass
#' @slot transfer.comp.scaled `character` vector of state variable which will
#'   be scaled 1:1 when biomass is modified, e.g. internal toxicant mass
#' @name Transferable
#' @family scenario-related
#' @aliases Biomass-transfer Transferable-class
#' @examples
#' # Simulation without biomass transfers
#' metsulfuron %>%
#'   set_noexposure() %>%
#'   set_notransfer() %>%
#'   simulate()
#'
#' # With biomass transfer every 7 days, biomass is reset to 50 *g/m²* on transfer
#' metsulfuron %>%
#'   set_noexposure() %>%
#'   set_transfer(interval=7, biomass=50) %>%
#'   simulate()
NULL

#' @export
setClass("Transferable",
         slots=list(
           transfer.times="numeric",
           transfer.interval="numeric",
           transfer.biomass="numeric",
           transfer.comp.biomass="character",
           transfer.comp.scaled="character"
         ),
         prototype=list(
           transfer.times=vector("numeric"),
           transfer.interval=-1,
           transfer.biomass=NA_real_,
           transfer.comp.biomass=NA_character_,
           transfer.comp.scaled=character(0)
         )
)

# Class definition
#' @aliases tox_data
setClass("ToxData",
  slots = list(data="ANY",
               exposure="ANY",
               qty_cols="character"),
  prototype = list(data=list(),
                   exposure=list(),
                   qty_cols=character(0))
)

# TODO overload plot() to display tox data
# TODO overload show to have some basic stats
# TODO overload as.data.frame() to return the tox data (we ignore exposure for now)
# TODO examples
#
#' Prepare ecotox study data for fitting
#'
#' Takes ecotox study data in *long-form* tabular format and prepares it for
#' parameter fitting. It supports extracting (optional) exposure concentration from
#' tabular data (useful for e.g. studies of acute toxicity) or exposure series
#' can also provided as individual time-series.
#'
#' # Tabular format
#' The long-form tabular data must have at least two and at most four columns.
#' The position of the columns define what they represent, the column names
#' are ignored:
#'
#' - First column: time
#' - Second column: observed quantity, e.g. number of individuals
#' - (optional) Third column: Trial or treatment ID
#' - (optional) Fourth column: Concentration
#'
#' The first two columns, time and observed quantity, must always be present.
#' The third column, trial ID, is used to split the table by treatment so
#' that trials can later be handled individually. The fourth column, concentration,
#' can be used to also define the exposure level during the experiments.
#'
#' # Explicit exposure series
#' As an alternative to defining concentrations along observed data, exposure
#' can also be passed as a list of exposure levels and series with
#' argument `exposure`. It can be used to provide exposure series for each
#' trial. The following object types are supported to define exposure:
#'
#' - Numerical constants
#' - Tabular data, e.g. `data.frame`s
#' - [exposure series][ExposureSeries] objects
#'
#' If exposure is constant over time, exposure can be defined using a single
#' constant value. More complex exposure time-series can be defined using e.g.
#' `data.frame`s. Tabular data must have two columns with the first column
#' representing time and the second column representing exposure/concentrations.
#'
#'
#' @param data a `data.frame` with at least two and at most four columns; the first
#'   column must represent time, the second an observed quantity, the optional third
#'   a trial or treatment ID, and the optional fourth the concentration during
#'   the experiment
#' @param exposure an optional named list; names must correspond to trial IDs used for
#'   the `data` argument; values can be numeric constants, data.frames, or an
#'   [ExposureSeries] object
#' @return a `ToxData` object
#' @aliases ToxData-class
#' @export
#' @examples
#' library(dplyr)
#'
#' mydata <- schmitt2013 %>% tox_data()
tox_data <- function(data, exposure=NULL) {
  if(missing(data))
    stop("Argument 'data' is missing")
  if(is.matrix(data))
    data <- as.data.frame(data)
  if(!is.data.frame(data))
    stop("Argument 'data' must be a data.frame")

  # additional checks on `data`
  errs <- c()
  has_conc <- ncol(data) >= 4
  has_trial <- ncol(data) >= 3
  if(ncol(data) > 4)
    warning("Argument `data` has more than four columns, additional columns will be ignored")
  if(ncol(data) == 0) {
    errs <- c(errs, "Argument 'data' is empty")
  }
  if(ncol(data) >= 1) {
    if(!is.numeric(dplyr::pull(data, 1))) # time
      errs <- c(errs, "Argument 'data': first column must be numeric")
  }
  if(ncol(data) >= 2) {
    if(!is.numeric(dplyr::pull(data, 2))) # quantity
      errs <- c(errs, "Argument 'data': second column must be numeric")
  }
  # no check of trial id, yet
  if(has_conc) {
    if(!is.numeric(dplyr::pull(data, 4))) # concentrations
      errs <- c(errs, "Argument 'data': fourth column must be numeric")

    # if `exposure` argument is present AND `data` contains concentrations,
    # then it's unclear what to do
    if(!is.null(exposure)) {
      stop("Confusion about exposure series: argument 'data' contains concentrations, but argument 'exposure' is also set")
    }
    # select exposure data as time, conc, trial id
    exposure <- dplyr::select(data, 1, 4, 3)
    # drop concentration from observed data
    data <- dplyr::select(data, 1, 2, 3)
  }
  if(!has_trial) {
    # assign default trial name
    data <- dplyr::mutate(data, trial="trial")
    has_trial <- TRUE
  }

  if(length(errs) > 0) {
    stop("\n* ", paste(errs, collapse="\n* "))
  }

  # convert inputs to named list of exposure series
  if(!is.null(exposure)) {
    exposure <- tox_exposure(exposure)
  } else {
    exposure <- list()
  }

  nms <- names(data)
  # split into list of observations per trial
  if(has_trial) {
    data <- data %>%
      dplyr::mutate(dplyr::across(3, ~ factor(., levels=unique(.)))) %>%
      dplyr::filter(dplyr::if_any(2, ~ !is.na(.))) %>%
      dplyr::group_by(dplyr::pick(3))
    ids <- dplyr::group_keys(data) %>% dplyr::pull(1)
    # sort by time and split by trial id
    data <- data %>%
      dplyr::arrange(dplyr::pick(1)) %>%
      dplyr::group_split(.keep=FALSE)
    data <- setNames(as.list(data), ids)
    for(i in seq_along(data)) {
      data[[i]] <- as.data.frame(data[[i]]) # get rid of tibbles
    }
  }
  # table contains data of a single trial
  else {
    data <- list(dplyr::arrange(data, dplyr::pick(1))) # order by time
  }

  new("ToxData", data=data, exposure=exposure, qty_cols=nms[2])
}

# Function for user convenience, intended to work similar to [morse::survData()]
#' @autoglobal
surv_data <- function(data) {
  # coerce argument to data.frame, e.g. to convert morse's data types such as `survDataCstExp`
  data <- as.data.frame(data)

  # TODO check that columns exist
  # TODO check data types

  # reorder and select columns
  qty <- data %>% dplyr::select(time, Nsurv, replicate)
  expo <- data %>% dplyr::select(time, conc, replicate)

  tox_data(qty, expo)
}

# Converts argument to a named list of objects that can be used as
# exposure series, i.e. as arguments to [set_exposure()]. Element names
# will be used as trial ids and vice versa in order to later associate
# exposure series to observed data sets.
#
# @param exp a numeric vector, a data.frame, an exposure series, or a (named) list
#   of any of the aforementioned types
# @return named list of [ExposureSeries] objects
tox_exposure <- function(exp) {
  if(missing(exp))
    stop("Argument 'exp' is missing")
  # argument must either be a named vector/list or a data.frame with three columns
  if(is.vector(exp)) {
    # check that names exist and none are empty
    nms <- names(exp)
    if(is.null(nms) | any(nms == ""))
      stop("Some trial ids are empty or missing")
    exp <- as.list(exp)
  }
  else if(is.data.frame(exp)) {
    exp <- as.data.frame(exp) # get rid of tibbles
    if(ncol(exp) < 3)
      stop("Argument must have at least three columns")
    if(!is.numeric(exp[, 1]))
      stop("First column must be numeric")
    if(!is.numeric(exp[, 2]))
      stop("Second column must be numeric")

    # split by trial id
    exp <- exp %>%
      dplyr::mutate(dplyr::across(3, ~ factor(., levels=unique(.)))) %>%
      dplyr::filter(dplyr::if_any(2, ~ !is.na(.))) %>%
      dplyr::group_by(dplyr::pick(3))
    lst <- dplyr::group_split(exp, .keep=FALSE)
    nms <- dplyr::group_keys(exp) %>% dplyr::pull(1)
    exp <- setNames(as.list(lst), nms) # get rid of tibble prototype
    for(i in seq_along(exp)) {
      exp[[i]] <- as.data.frame(exp[[i]]) # get rid of tibble format
    }
  }
  else {
    stop("Argument is of invalid type")
  }

  # check if list elements have valid types, convert where possible
  for(i in seq_along(exp)) {
    elem <- exp[[i]]
    nm <- names(exp)[[i]]
    # constant value
    if(is.numeric(elem)) {
      if(length(elem) != 1)
        stop("Exposure value of trial '", nm, "' must have length one")
      exp[[i]] <- data.frame(time=0, conc=elem)
    }
    # data.frame
    else if(is.data.frame(elem)) {
      if(ncol(elem) != 2)
        stop("Exposure series of trial '", nm, "' must have two columns")
    }
    # exposure series
    else if(all(is_exp_series(elem)) & length(elem) == 1) {
      # no checks for now
    } else {
      stop("Exposure series of trial '", nm, "' has invalid type")
    }
  }

  if(length(exp) == 0)
    stop("Argument does not contain any exposure information")

  # check names
  nms <- names(exp)
  if(is.null(nms)) {
    nms <- c("")
  }
  if(any(nms == "")) {
    stop("Some trial ids are empty or missing")
  }
  if(any(table(nms) > 1)) {
    stop("Trial ids are not unique")
  }

  exp
}

#' Create *calisets* from *tox_data*
#'
#' This is a convenience function which eases the creation of [caliset]s
#' from a base scenario and a [tox_data] object. The scenario
#' will be used as is. If exposure series are defined by the `tox_data`
#' object, then these will be assigned to the scenario(s) accordingly.
#'
#' @param scenario a base [scenario]
#' @param data return value of [tox_data()]
#' @param output_var optional *character*, rename observed data column to this value
#' @return list of [caliset]s
#' @export
#' @examples
#' # Import trial data from Schmitt et al. (2013), including exposure
#' mydata <- tox_data(schmitt2013)
#'
#' # Example trial contained in data set: trial 'T0.32'
#' mydata@data[["T0.32"]]        # observed quantities
#' mydata@exposure[["T0.32"]]    # associated exposure series
#'
#' # Create list of calisets from full data set
#' lst <- td2cs(Lemna_SETAC(), mydata)
#'
#' # Example caliset representing conditions of trial 'T0.32'
#' lst[[2]]
td2cs <- function(scenario, data, output_var=NULL) {
  if(is.list(scenario))
    stop("Argument 'scenario' must be of length one")
  if(!is_scenario(scenario) & !is_sequence(scenario))
    stop("Argument 'scenario' must be a scenario")

  if(missing(data))
    stop("Argument 'data' is missing")
  if(!is(data, "ToxData"))
    stop("Argument 'data' must be a tox_data object")

  if(!is.null(output_var)) {
    if(!is.character(output_var))
      stop("Argument 'output_var' must be a string")
    if(length(output_var) != 1)
      stop("Argument 'output_var' must be of length one")
  }

  lst <- list()
  nms <- names(data@data)
  for(i in seq_along(data@data)) {
    # create a copy of the scenario, as it might get modified
    sc <- scenario
    nm <- nms[[i]]

    # make sure data series is a `data.frame` and rename column with observed data
    df <- as.data.frame(data@data[[i]])
    # if the name of the observed variable is defined
    if(!is.null(output_var)) {
      df <- setNames(df, c(names(df)[[1]], output_var))
    }
    # look if an exposure series exists for that trial id
    if(nm %in% names(data@exposure)) {
      sc <- set_exposure(sc, data@exposure[[nm]], reset_times=FALSE)
    }
    lst[[i]] <- caliset(sc, data=df, tag=nm)
  }
  lst
}

test_that("arg:data invalid", {
  expect_error(tox_data(list(1, 2)), "be a data.frame")
  expect_warning(tox_data(data.frame(t=0, n=1, t=2, c=3, foo=4)), "more than four columns")
  expect_error(tox_data(data.frame(t="foo", n=0)), "first column must be numeric")
  expect_error(tox_data(data.frame(t=0, n="foo")), "second column must be numeric")
  expect_error(tox_data(data.frame(t=0, n=1, t=2, conc="foo")), "fourth column must be numeric")
  expect_error(tox_data(data.frame(t=0, n=1, t=2, conc=3), exposure=list()), "Confusion about exposure")
})

test_that("arg:data", {
  df <- data.frame(t=0:4, n=1)

  # various types of data arguments
  # matrix
  dat <- tox_data(as.matrix(df))@data
  expect_equal(length(dat), 1) # list of trial data
  dat <- dat[[1]] # unnest list
  expect_true(is.data.frame(dat))
  expect_equal(length(dat), 2)
  expect_equal(names(dat), names(df))
  expect_equal(dplyr::select(dat, 1, 2), df, ignore_attr=TRUE)

  # data.frame
  dat <- tox_data(df)@data
  expect_equal(length(dat), 1) # list of trial data
  dat <- dat[[1]] # unnest list
  expect_true(is.data.frame(dat))
  expect_equal(length(dat), 2)
  expect_equal(names(dat), names(df))
  expect_equal(dplyr::select(dat, 1, 2), df, ignore_attr=TRUE)

  # tibble
  dat <- tox_data(tibble::as_tibble(df))@data
  expect_equal(length(dat), 1) # list of trial data
  dat <- dat[[1]] # unnest list
  expect_true(is.data.frame(dat))
  expect_equal(length(dat), 2)
  expect_equal(names(dat), names(df))
  expect_equal(dplyr::select(dat, 1, 2), df, ignore_attr=TRUE)
})

test_that("arg:data=data.frame format", {
  # basic data, two cols
  df2 <- data.frame(t=0:4, n=1)
  dat <- tox_data(df2)
  # observed
  expect_true(is(dat, "ToxData"))
  expect_true(is.list(dat@data))
  expect_equal(length(dat@data), 1)
  expect_equal(names(dat@data), "trial") # default trial id
  expect_true(is.data.frame(dat@data[[1]]))
  expect_equal(dat@data[[1]], df2, ignore_attr=TRUE)
  expect_equal(names(dat@data[[1]]), names(df2))
  # exposure
  expect_true(is.list(dat@exposure))
  expect_equal(length(dat@exposure), 0)
  # qty
  expect_equal(dat@qty_cols, "n")

  # with trial id, three cols
  df3 <- data.frame(t=0:4, n=1, t="foo")
  dat <- tox_data(df3)
  # single trial
  expect_equal(length(dat@data), 1)
  expect_equal(names(dat@data), "foo")
  expect_equal(dat@data[[1]], df3[, c(1, 2)], ignore_attr=TRUE)
  expect_equal(length(dat@exposure), 0)
  expect_equal(dat@qty_cols, "n")
  # multiple trials
  df3a <- data.frame(t=0:2, n=1, trial="foo")
  df3b <- data.frame(t=0:2, n=2, trial="bar")
  df3 <- rbind(df3a, df3b)
  dat <- tox_data(df3)

  expect_equal(length(dat@data), 2)
  expect_equal(names(dat@data), c("foo", "bar"))
  expect_equal(dat@data[["foo"]], df3a[, c(1, 2)], ignore_attr=TRUE)
  expect_equal(dat@data[["bar"]], df3b[, c(1, 2)], ignore_attr=TRUE)
  expect_false(tibble::is_tibble(dat@data[["foo"]]))
  expect_equal(length(dat@exposure), 0)
  expect_equal(dat@qty_cols, "n")

  # with concentration, four cols
  df4a <- data.frame(t=0:2, n=3, trial="foo", c=5)
  df4b <- data.frame(t=0:2, n=4, trial="bar", c=c(6, 6, 7))
  df4 <- rbind(df4a, df4b)
  dat <- tox_data(df4)

  expect_equal(length(dat@data), 2)
  expect_equal(names(dat@data), c("foo", "bar"))
  expect_equal(dat@data[["foo"]], df4a[, c(1, 2)], ignore_attr=TRUE)
  expect_equal(dat@data[["bar"]], df4b[, c(1, 2)], ignore_attr=TRUE)
  expect_false(tibble::is_tibble(dat@data[["foo"]]))
  expect_equal(length(dat@exposure), 2)
  expect_equal(names(dat@exposure), c("foo", "bar"))
  expect_equal(dat@exposure[["foo"]], df4a[, c(1, 4)], ignore_attr=TRUE)
  expect_equal(dat@exposure[["bar"]], df4b[, c(1, 4)], ignore_attr=TRUE)
  expect_equal(dat@qty_cols, "n")
})

test_that("arg:data=data.frame with NAs", {
  # basic data, two cols
  df2 <- data.frame(t=0:2, n=c(1, NA_real_, 2))
  dat <- tox_data(df2)
  # observed without NAs
  expect_equal(length(dat@data), 1)
  expect_equal(dat@data[[1]], df2 %>% dplyr::filter(!is.na(n)))

  # with concentration, four cols
  df4 <- data.frame(t=0:3, n=c(1, NA_real_, 2, 3), trial="foo", c=c(0, 1, NA_real_, 2))
  dat <- tox_data(df4)

  expect_equal(length(dat@data), 1)
  # observed without NAs
  expect_equal(dat@data[[1]], df4 %>% dplyr::filter(!is.na(n)) %>% dplyr::select(t, n))
  # exposure without NAs
  expect_equal(dat@exposure[[1]], df4 %>% dplyr::filter(!is.na(c)) %>% dplyr::select(t, c))
})

# TODO test tox_data with exposure argument
# TODO test surv_data

test_that("tox_exposure", {
  # named numeric vector/list
  vec <- c(foo=1)
  exp <- tox_exposure(vec)

  expect_equal(length(exp), 1)
  expect_equal(names(exp), names(vec))
  expect_equal(exp[[1]][, 2], vec[[1]])

  vec <- c(foo=1, bar=2)
  exp <- tox_exposure(vec)

  expect_equal(length(exp), 2)
  expect_equal(names(exp), names(vec))
  expect_equal(exp[[1]][, 2], vec[[1]])
  expect_equal(exp[[2]][, 2], vec[[2]])

  vec <- list(foo=1, bar=2)
  exp <- tox_exposure(vec)

  expect_equal(length(exp), 2)
  expect_equal(names(exp), names(vec))
  expect_equal(exp[[1]][, 2], vec[[1]])
  expect_equal(exp[[2]][, 2], vec[[2]])

  # data.frame w/o trial ids -> fails
  df <- data.frame(time=0, conc=1)
  expect_error(tox_exposure(df), "at least three columns")

  # data.frame with trial id
  df <- data.frame(time=0, conc=1, trial="foo")
  exp <- tox_exposure(df)

  expect_equal(length(exp), 1)
  expect_equal(names(exp), "foo")
  expect_equal(exp[[1]], df[, c(1, 2)], ignore_attr=TRUE)

  dfa <- data.frame(time=0:1, conc=1, trial="foo")
  dfb <- data.frame(time=0:1, conc=c(2, 3), trial="bar")
  df <- rbind(dfa, dfb)
  exp <- tox_exposure(df)

  expect_equal(length(exp), 2)
  expect_equal(sort(names(exp)), sort(unique(df$trial)))
  expect_equal(exp[["foo"]], dfa[, c(1, 2)], ignore_attr=TRUE)
  expect_equal(exp[["bar"]], dfb[, c(1, 2)], ignore_attr=TRUE)

  # list of various object types
  lst <- list(foo=1, bar=data.frame(t=1, c=2))
  exp <- tox_exposure(lst)

  expect_equal(length(exp), 2)
  expect_equal(sort(names(exp)), sort(unique(df$trial)))
  expect_equal(exp[["foo"]][, 2], lst$foo, ignore_attr=TRUE)
  expect_equal(exp[["bar"]], lst$bar[, c(1, 2)], ignore_attr=TRUE)
})

test_that("tox_exposure invalid arg", {
  expect_error(tox_exposure())
  # missing trial ids
  expect_error(tox_exposure(c(1, 2)), "empty or missing")
  expect_error(tox_exposure(list(1, 2)), "empty or missing")
  expect_error(tox_exposure(list(foo=1, 2)), "empty or missing")
  expect_error(tox_exposure(data.frame(time=0, conc=1, trial=c("foo","bar",""))), "empty or missing")

  # invalid data.frame format
  expect_error(tox_exposure(data.frame(t=0, c=0)), "at least three columns")
  expect_error(tox_exposure(data.frame(t="foo", c=0, tr="bar")), "must be numeric")
  expect_error(tox_exposure(data.frame(t=0, c="foo", tr="bar")), "must be numeric")

  # other invalid argument
  expect_error(tox_exposure(minnow_it), "invalid type")

  # invalid list elements
  expect_error(tox_exposure(list(foo=1, bar="a")), "invalid type")
})

test_that("td2cs", {
  ## valid arguments
  sc <- Lemna_SETAC() %>% set_times(0:4)
  # single tox_data dataset
  td <- tox_data(data.frame(time=0, obs=0, trial="foo"))
  cs <- td2cs(sc, td)
  expect_equal(length(cs), 1)
  expect_true(all(is_caliset(cs)))
  expect_equal(cs[[1]]@scenario, sc)
  expect_equal(cs[[1]]@data, td@data[[1]])
  expect_equal(cs[[1]]@tag, "foo")

  # scenario is a sequence, single tox_data dataset
  sq <- sequence(list(sc, sc), breaks=2)
  td <- tox_data(data.frame(time=0, obs=0, trial="foo"))
  cs <- td2cs(sq, td)
  expect_equal(length(cs), 1)
  expect_true(all(is_caliset(cs)))
  expect_equal(cs[[1]]@scenario, sq)
  expect_equal(cs[[1]]@data, td@data[[1]])
  expect_equal(cs[[1]]@tag, "foo")

  # multiple tox_data datasets
  td <- tox_data(data.frame(time=0, obs=0:1, trial=c("foo", "bar")))
  cs <- td2cs(sc, td)
  expect_equal(length(cs), 2)
  expect_true(all(is_caliset(cs)))
  expect_equal(cs[[1]]@scenario, sc)
  expect_equal(cs[[2]]@scenario, sc)
  expect_equal(cs[[1]]@data, td@data[[1]])
  expect_equal(cs[[2]]@data, td@data[[2]])
  expect_equal(cs[[1]]@tag, "foo")
  expect_equal(cs[[2]]@tag, "bar")

  # single tox_data dataset, multiple exposure series
  exp <- list("baz"=data.frame(t=0, c=0), "foo"=data.frame(t=0, c=1))
  td <- tox_data(data.frame(time=0, obs=0, trial="foo"), exposure=exp)
  cs <- td2cs(sc, td)
  expect_equal(length(cs), 1)
  expect_equal(cs[[1]]@scenario@param, sc@param) # use parameter list as a proxy, bc they do differ in exposure
  expect_equal(cs[[1]]@scenario@times, sc@times) # use parameter list as a proxy, bc they do differ in exposure
  expect_equal(cs[[1]]@data, td@data[[1]])
  expect_equal(cs[[1]]@tag, "foo")
  expect_equal(cs[[1]]@scenario@exposure@series, exp$foo)

  # single tox_data dataset, rename output var
  td <- tox_data(data.frame(time=0, foo=0))
  cs <- td2cs(sc, td, output_var="bar")
  expect_equal(length(cs), 1)
  expect_true(all(is_caliset(cs)))
  expect_equal(cs[[1]]@scenario, sc)
  expect_equal(cs[[1]]@data, data.frame(time=0, bar=0))

  ## invalid arguments
  sc <- Lemna_SETAC()
  expect_error(td2cs(), "scenario. is missing")
  expect_error(td2cs(list(sc)), "scenario. must be of length one")
  expect_error(td2cs("foo"), "scenario. must be a scenario")

  expect_error(td2cs(sc), "data. is missing")
  expect_error(td2cs(sc, data="foo"), "data. must be a tox_data object")

  expect_error(td2cs(sc, td, output_var=1), "must be a string")
  expect_error(td2cs(sc, td, output_var=c("foo", "bar")), "must be of length one")
})

test_that("cvasi_drc", {
  df <- dose_response(minnow_it)
  plot(df)
  succeed()
})

test_that("cvasi_simulate", {
  df <- simulate(minnow_it)
  plot(df)
  succeed()
})

test_that("lik_profile", {
  data <- schmitt2013 %>% dplyr::filter(trial == "T0")

  # update metsulfuron
  scenario <- metsulfuron %>%
    set_param(c(k_phot_fix = TRUE, Emax = 1)) %>%
    set_init(c(BM = 12)) %>%
    set_bounds(list(k_phot_max=c(0, 1)))

  # Likelihood profiling
  suppressMessages(
    res <- lik_profile(
    x = td2cs(scenario, tox_data(data)),
    data = data,
    output = "BM",
    par = c("k_phot_max"=0.43),
    bounds = list(
      k_phot_max = list(0, 1)
    ),
    refit=FALSE
  ))
  # plot
  plot(res)
  succeed()
})

#' Pull metadata from scenarios
#'
#' The method pulls available metadata from scenario objects and returns a
#' table with additional columns. If the argument already was a `data.frame`
#' object, the columns are appended. May overwrite existing columns of the
#' same name.
#'
#' @param x vector of [scenarios] or a `data.frame` containing a column
#'   `scenario` with `EffectScenario` objects
#' @param model `logical`, if `TRUE` then model metadata is pulled
#' @param exposure  `logical`, if `TRUE` then exposure series metadata is pulled
#'
#' @return a `data.frame`
#' @export
#' @global scenario context
#'
#' @examples
#' metsulfuron %>%
#'   pull_metadata()
pull_metadata <- function(x, model=TRUE, exposure=TRUE) {
  if(is_scenario(x))
    x <- tibble::tibble(scenario=c(x))
  else if(is.vector(x))
    x <- tibble::tibble(scenario=x)
  if(!is.data.frame(x))
    stop("argument x has unsupported type", call. = FALSE)

  # model & scenario related metadata
  if(model) {
    x %>%
      dplyr::rowwise() %>%
      dplyr::mutate(ModelName=scenario@name,
                    ScenarioTag=scenario@tag,
                    Window_Length=scenario@window.length,
                    Window_Interval=scenario@window.interval) -> x
  }
  # exposure series related metadata
  if(exposure) {
    # pull metadata from TOXSWA series
    x %>%
      dplyr::rowwise() %>%
      dplyr::mutate(context=list(scenario@exposure@context)) %>%
      tidyr::unnest_wider(context, names_repair="minimal") -> x
  }

  dplyr::ungroup(x)
}

# TODO provide sensible default value for output by model type

#' Fit model parameters to experimental data
#'
#' The function `calibrate()` performs the calibration (fitting) of model
#' parameters to observed data. The data can originate from one or more experiments or
#' trials. Experimental conditions, such as model parameters and exposure
#' level, can differ between trials; fitting can be performed on all datasets
#' at the same time.
#'
#' Fitting of model parameters can be performed in two ways:
#' 1) A single [scenario] is fitted to a single dataset.
#'   The dataset must represent a time-series of an output variable of the
#'   model, e.g. observed biomass over time (effect data). The dataset can represent
#'   results of one or more experimental replicates under identical conditions.
#' 2) One or more datasets of observed data are fitted each to a scenario which
#'   describes the experimental conditions during observation, such as exposure
#'   level and environmental properties. Each combination of dataset and scenario
#'   is represented by a [calibration set][caliset]. During fitting,
#'   all *calibration sets* are evaluated and a total error term is calculated
#'   over all observed and predicted values.
#'
#' ### Observed data
#' Experimental, or effect, data must be supplied as a `data.frame` in long format
#' with at least two columns: the first column contains `numeric` timestamps and
#' the remaining columns must contain the observed quantity. The observed quantitiy
#' must match in unit and meaning with the model output defined by argument `output`.
#'
#' As an example, the simulation result of [Lemna_Schmitt] model contains the
#' output column *biomass* (`BM`), amongst others. To fit model parameters of said
#' *Lemna_Schmitt* scenario based on observed biomass, the observed data must
#' represent biomass as well. A minimal observed dataset could look like this:
#'
#' ```
#' observed <- data.frame(time=c(0,  7, 14, 21),
#'                        BM=c( 12, 23, 37, 56))
#' ```
#'
#' ### Error function
#' The error function quantifies the deviations between simulated and observed data.
#' The decision for an error function can have influence on the result of the
#' fitting procedure. Two error functions are pre-defined by the package and
#' can be selected by the user, but custom error functions can be used as well.
#'
#' Available pre-defined error functions:
#'
#' * `"sse"`: Sum of squared errors
#'
#' By default, the sum of squared errors is used as the error function which
#' gets minimized during fitting. A custom error function must accept four vectorized
#' arguments and return a numeric of length one, i.e. the total error value.
#' The arguments to the error function will all have the same length and are
#' defined as follows:
#'
#' * First argument: all observed data points
#' * Second argument: all simulated data points
#' * Third argument: optional weights for each data point
#' * Fourth argument: a list of optional caliset tags
#'
#' You can choose to ignore certain arguments, such as weights and tags, in your
#' custom error function. An example of a custom error function which returns
#' the sum of absolute errors looks as follow:
#' ```
#' my_absolute_error <- function(observed, predicted, weights, tags) {
#'   sum(abs(observed - predicted))
#' }
#' ```
#'
#' As tags are optional, the fourth argument may be a list containing `NULL` values.
#' The fourth argument can be used to pass additional information to the error
#' function: For example, the tag may identify the study from where the data
#' originates from and the error function could group and evaluate the data
#' accordingly.
#'
#' @name calibrate
#' @param x either a single [scenario]. [sequence], or a list of [caliset] objects
#'    to be fitted. If only a scenario or sequence is supplied, the additional argument
#'    `data` is required.
#' @param par named numeric vector with parameters to fit and their start values
#' @param output `character`, name of a single output column of [simulate()] to
#'   optimize on
#' @param err_fun a `character` choosing one of the pre-defined error functions,
#'   or alternatively a function implementing a custom error function. Defaults
#'   to *Sum of squared errors* (`"sse"`).
#' @param log_scale logical, if `TRUE` then observed and predicted values are
#'   log-transformed before the error function is evaluated. Defaults to `FALSE`.
#' @param verbose `logical`, if `TRUE` then debug outputs are displayed during
#'   optimization
#' @param data a `data.frame` or return value of [tox_data()]; the scenarios's
#'   output is fitted to the (observed) data. See [tox_data()] for valid tabular
#'   data formats.
#' @param ode_method optional `character` to select an ODE solver for [simulate()]
#' @param ... additional parameters passed on to [stats::optim()] and [simulate()]
#'
#' @returns A list of fitted parameters (as produced by [stats::optim()])
#' is returned.
#'
#' @aliases calibrate,EffectScenario-method calibrate,CalibrationSet-method
#'    calibrate,list-method calibrate,ScenarioSequence-method
#' @examples
#' # Create an artificial data set of observed frond numbers.
#' # It assumes exponential growth with an effective growth rate of 0.38
#' trial <- data.frame(time=0:14,
#'                    fronds=12 * exp(0:14 * 0.38))
#' plot(trial)
#'
#' # Create a Lemna scenario that represents unrestricted, exponential growth.
#' scenario <- Lemna_Schmitt() %>%
#'   set_param(c(k_phot_max=1, k_resp=0, EC50=1, b=1, P_up=1)) %>%
#'   set_init(c(BM=12)) %>%
#'   set_noexposure()
#'
#' # Fit scenario parameter 'k_phot_max' to observed frond numbers:
#' fit <- calibrate(
#'   scenario,
#'   par="k_phot_max",
#'   data=trial,
#'   output="BM"
#' )
#'
#' # The fitted value of 'k_phot_max' matches the effective growth rate which
#' # was used to create the artificial data set:
#' fit$par
NULL

#' @noRd
#' @export
setGeneric("calibrate", function(x, ...) standardGeneric("calibrate"), signature="x")

#' @noRd
#' @include sequence.R
setMethod("calibrate", "ScenarioSequence", function(x, ...) {
    calibrate_scenario(x=x, ...)
  }
)

#' @describeIn calibrate Fit single [scenario] to a ([tox_data]) dataset
#' @include class-EffectScenario.R
setMethod("calibrate", "EffectScenario", function(x, output, data, ...) {
    calibrate_scenario(x=x, output=output, data=data, ...)
  }
)

#' @autoglobal
calibrate_scenario <- function(x, data, endpoint=deprecated(), output, by=deprecated(), ...)
{
  if(lifecycle::is_present(endpoint)) { # for backwards compatibility
    lifecycle::deprecate_warn("1.1.0", "calibrate(endpoint)", "calibrate(output)")
    output <- endpoint
  }
  if(lifecycle::is_present(by)) {
    lifecycle::deprecate_warn("1.5.0", "calibrate(by)", details="Argument 'data' must use a position-based input format, see `?tox_data` for details")
    if(length(by) != 1)
      stop("Argument `by` must be of length one")
    if(!is.character(by))
      stop("Argument `by` must be a character string")
    if(!(by %in% names(data)))
      stop("Argument `by` is not a column in argument `data`")
    # workaround for backwards compatibility
    data <- data %>% dplyr::relocate(!!by, .after=3)
  }

  if(length(x) != 1) {
    stop("Argument 'x' must be of length one")
  }
  if(!(is_scenario(x) | is_sequence(x))) {
    stop("Argument 'x' must be a scenario")
  }
  # coerce matrix to data.frame
  if(is.matrix(data)) {
    data <- as.data.frame(data)
  }
  if(is.data.frame(data)) {
    data <- tox_data(data)
  } else if(is(data, "ToxData")) {
    # nothing to do
  } else {
    stop("Argument 'data' must be a `data.frame` or return value of `tox_data()`")
  }

  if(length(data@data) == 0) {
    stop("Argument 'data' is empty")
  }

  # convert tox_data object to list of calisets
  lst <- td2cs(x, data)
  calibrate(x=lst, output=output, ...)
}

#' @noRd
#' @include class-CalibrationSet.R
setMethod("calibrate", "CalibrationSet", function(x, ...) calibrate_list(x=list(x), ...))

# @describeIn calibrate Fit to list of [caliset]s
#' @rdname calibrate
setMethod("calibrate", "list", function(x, par, output, err_fun="sse", log_scale=FALSE, verbose=TRUE, ode_method=NULL, ...)
  calibrate_list(x=x, par=par, output=output, err_fun=err_fun, log_scale=log_scale, verbose=verbose, ode_method=ode_method, ...))

#' @importFrom lifecycle deprecated
#' @importFrom stats optim
#' @autoglobal
calibrate_list <- function(x, par, output, endpoint=deprecated(), metric_fun=deprecated(),
                          metric_total=deprecated(), err_fun="sse", log_scale=FALSE,
                          hessian=TRUE, as_tibble=deprecated(), catch_errors=deprecated(),
                          verbose=TRUE, data, ode_method=NULL, lower=NULL, upper=NULL,
                          method=NULL, ...) {
  if(!missing(data)) {
    stop("Argument 'data' cannot be used in combination with calisets")
  }
  if(!all(sapply(x, function(y) is(y, "CalibrationSet")))) {
    stop("Argument 'x' must only contain caliset objects")
  }

  ##
  ## Parameters to fit and their starting values
  ##
  if(missing(par)) {
    stop("Argument 'par' is missing")
  }
  # vector of parameter names
  if(is.character(par)) {
    par <- as.list(par)
  }
  # (a mix of) parameter names and/or name/starting value pairs
  if(is.list(par))
  {
    # if `par`contains character elements, then it's a parameter name. get
    # associated starting value from scenario's parameters
    is_nm <- sapply(par, is.character)
    if(any(is_nm)) {
      missing_nms <- unlist(par[is_nm])
      all_nms <- names(par)
      all_nms[which(is_nm)] <- missing_nms
      par_val <- x[[1]]@scenario@param[missing_nms]
      par_val <- setNames(par_val, missing_nms) # make sure every element has a name
      par[which(is_nm)] <- par_val
      par <- setNames(par, all_nms)

      # did we find all values we are looking for?
      if(any(sapply(par, is.null))) {
        stop("Could not determine starting value for parameters '", paste(missing_nms, collapse="', '"), "'")
      }
    }

    # finally, convert list to named numerical vector
    par <- unlist(par)
  }
  if(!is.numeric(par) | !is.vector(par)) {
    stop("Argument 'par' contains non-numerical elements")
  }
  if(any(is.na(par) | is.infinite(par))) {
    stop("Argument 'par' contains invalid starting values such as NA or Infinity")
  }
  if(length(par) == 0) {
    stop("Argument 'par' is empty")
  }
  nms <- names(par)
  nms <- nms[nms != ""]
  if(length(nms) < length(par)) {
    stop("All elements of argument 'par' must be named")
  }

  unused <- setdiff(nms, c(names(get_param(x[[1]]@scenario)), get_rparam(x[[1]]@scenario)))
  if(length(unused) > 0) {
    stop(paste("Argument 'par' contains elements which are not scenario parameters: ", paste(unused, collapse=", ")))
  }
  rm(nms, unused)

  ##
  ## Auto-select a sensible optimization method
  ##
  # TODO use L-BFGS-B if bounds set/available for all parameters
  if(is.null(method)) {
    if(length(par) == 1)
      method <- "Brent"
    else
      method <- "Nelder-Mead"
  }

  ##
  ## Output variable
  ##
  if(lifecycle::is_present(endpoint)) {
    lifecycle::deprecate_warn("1.1.0", "calibrate(endpoint)", "calibrate(output)")
    output <- endpoint
  }
  if(missing(output)) {
    stop("Argument 'output' is missing")
  }
  if(length(output) != 1) {
    stop("Argument 'output' must be of length one")
  }
  if(!is.character(output)) {
    stop("Argument 'output' must be a string")
  }

  ##
  ## Parameter boundaries
  ##
  # Override boundaries using arguments to function?
  if(!is.null(lower) | !is.null(upper)) {
    lifecycle::deprecate_soft("1.5.0", "cvasi::calibrate(lower)", "cvasi::set_bounds()")
    lifecycle::deprecate_soft("1.5.0", "cvasi::calibrate(upper)", "cvasi::set_bounds()")

    if(!(method %in% c("Brent", "L-BFGS-B"))) {
      stop("Selected optimization method does not support parameter bounds")
    }
    if(length(par) != length(lower) | length(par) != length(upper)) {
      stop("Length of arguments 'lower' and 'upper' must be equal to length of 'par'")
    }
    if(any(is.na(lower) | is.na(upper))) {
      stop("Arguments 'lower' and 'upper' must not contain NA")
    }
    # reformart boundaries
    new_bounds <- list()
    for(i in seq_along(par)) {
      new_bounds[[names(par)[i]]] <- c(lower[[i]], upper[[i]])
    }
    # update scenarios in calibration sets with new parameter boundaries
    for(i in seq_along(x)) {
      x[[i]]@scenario <- x[[i]]@scenario %>% set_bounds(new_bounds)
    }
    rm(i, new_bounds)
  }

  # Boundaries can only be used for optimization methods Brent and L-BFGS-B
  if(method %in% c("Brent", "L-BFGS-B")) {
    # Collect relevant boundaries from 1st caliset
    lower <- upper <- numeric(0)
    bounds <- x[[1]]@scenario@param.bounds
    for(i in seq_along(par)) {
      nm <- names(par)[i]
      # if no boundary is set, use inifity
      if(!(nm %in% names(bounds))) {
        lower[[i]] <- -Inf
        upper[[i]] <- Inf
      } else {
        lower[[i]] <- bounds[[nm]][[1]]
        upper[[i]] <- bounds[[nm]][[2]]
      }
    }

    if(any(is.infinite(lower) | is.infinite(upper))) {
      stop("Methods Brent and L-BFGS-B require all parameter boundaries to be defined.")
    }

    rm(i, nm, bounds)
  } # Else, set values to defaults
  else {
    lower <- rep(-Inf, length(par))
    upper <- rep(Inf, length(par))
  }

  ##
  ## Error function
  ##
  err_desc <- "n.a."
  if(lifecycle::is_present(metric_fun)) {
    lifecycle::deprecate_warn("1.1.0", "calibrate(metric_fun)", "calibrate(err_fun)")
    err_fun <- metric_fun
  }
  if(lifecycle::is_present(metric_total)) {
    lifecycle::deprecate_stop("1.1.0", "calibrate(metric_total)", details="Has been removed without replacement. Please use 'err_fun' instead.")
  }

  ## Choose an error functions
  # One of the defaults selected?
  if(is.character(err_fun)) {
    err_fun <- err_fun[[1]]
    # one of the pre-defined error functions
    if(err_fun == "sse") {
      err_f <- sse
      err_desc <- "Sum of squared errors"
    } else if(err_fun == "log_sse") {
      lifecycle::deprecate_warn("1.5.0", "cvasi::calibrate(err_fun='must not use `log_sse`')", with="cvasi::calibrate(log_scale)")
      err_f <- log_sse
      err_desc <- "Sum of squared errors on log data"
    } else {
      stop("Argument 'err_fun' contains the name of an unsupported error function: ", err_fun)
    }
  } else if(is.function(err_fun)) {
    err_desc <- "Custom error function"
    err_f <- err_fun
  } else {
    stop("Argument 'err_fun' must be a string or function")
  }

  # deprecated and removed parameters
  if(lifecycle::is_present(as_tibble)) {
    lifecycle::deprecate_warn("1.1.0", "cvasi::calibrate(as_tibble)", details="Results can no longer be returned as a tibble")
  }
  if(lifecycle::is_present(catch_errors)) {
    lifecycle::deprecate_warn("1.1.0", "cvasi::calibrate(catch_errors)", details="Catching errors is always enabled")
  }

  # parameter names have to be re-set for Brent method in each iteration
  par_names <- names(par)

  # some stats
  if(verbose) {
    message("Parameter fitting based on ", length(x), " data set", ifelse(length(x) > 1, "s", ""))
    message("Fitted parameter", ifelse(length(par_names) > 1, "s", ""), ": ")
    for(i in seq_along(par_names)) {
      message("  ", par_names[i], " [", lower[i], "; ", upper[i], "]")
    }
    message("Output variable: ", output)
    message("Error function: ", err_desc)
    message("Optimization method: ", method)
  }

  # make sure that the column name of observed data is identical to the
  # output variable's name in all datasets
  for(i in seq_along(x)) {
    x[[i]]@data <- setNames(x[[i]]@data, c("time", output))
  }

  # create an environment object to enable passing data back from inside `optim()`
  # elements:
  # `desolve_diagn`: only set in case of numerical errors
  # `cvasi_status` :   "   "   "   "  "   "         "
  # `tested`       : list of tested parameter sets
  # `err`          : error of best fit
  # `obs`          : observed values
  # `pred`         : predicted values of best fit
  # `times`        : time points of observed values
  env <- new.env()
  env$tested <- list()
  env$success <- logical(0)
  env$err <- NA_real_
  env$obs <- NA_real_
  env$pred <- NA_real_
  env$times <- NA_real_

  # start optimization
  fit <- optim(par=par, fn=optim_set, sets=x, par_names=par_names, output=output,
               err_fun=err_f, log_scale=log_scale, hessian=hessian, verbose=verbose, env=env,
               ode_method=ode_method, lower=lower, upper=upper, method=method, ...)

  # re-set names in case method 'Brent' dropped them
  names(fit$par) <- par_names
  # did all simulations fail?
  # TODO check reason for failures, e.g. if all due to numerics, then report this to user
  if(all(!env$success) & length(env$success) > 0) { # length check necessary to avoid spurious warnings during tests
    warning("All simulations failed. Please check arguments.", call.=FALSE)
  }
  # optimization successful?
  else if(fit$convergence > 0)
  {
    # set a fit message of well-defined return codes if nothing else is present
    # see `?optim()` for reference
    if(is.null(fit$message)) {
      if(fit$convergence == 1)
        fit$message <- "Iteration limit `maxit` reached"
      else if(fit$convergence == 10)
        fit$message <- "Nelder-Mead simplex has degenerated"
    }
    warning("  Possible convergence problem during optimization: `optim()` convergence code = ",
                  fit$convergence, ifelse(is.null(fit$message), "", paste0("\n  '", fit$message, "'")),
            call. = FALSE)
  }

  if(verbose) {
    message("Best fitting parameter", ifelse(length(par_names) > 1, "s", ""), ":")
    for(nm in par_names) {
      message("  ", nm, " = ", fit$par[[nm]])
    }
  }

  ##
  ## Enrich return value with some relevant metadata
  ##
  # assign solver info, if it exists (in case of serious errors)
  if(!is.null(env$desolve_diagn)) {
    attr(fit, "desolve_diagn") <- env$desolve_diagn
    attr(fit, "cvasi_status") <- env$cvasi_status
  }
  # evaluated data points of best-fit
  fit$data <- data.frame(time=env$times, obs=env$obs, pred=env$pred)
  # was data transformed for error calculation?
  if(log_scale) {
    fit$data <- dplyr::mutate(fit$data, log_obs=log(obs), log_pred=log(pred), residuals=log_obs - log_pred)
  } else {
    fit$data <- dplyr::mutate(fit$data, residuals=obs - pred)
  }

  # set row & col names of Hessian, if missing
  if("hessian" %in% names(fit)) {
    if(is.null(row.names(fit$hessian)))
      row.names(fit$hessian) <- par_names
    if(is.null(colnames(fit$hessian)))
      colnames(fit$hessian) <- par_names
  }

  # save arguments which were provided to calibrate()
  args <- list(
    "x" = x,               # list of calisets
    "par" = par,           # starting values
    "err_fun" = err_fun,   # error function
    "method" = method,     # optim method
    #"ode_method" = ode_method,
    "output" = output,
    "log_scale" = log_scale
  )
  attr(fit, "args") <- args

  # return fit
  class(fit) <- c("cvasi_fit", "list")
  fit
}

# Calculate error for a single set of parameters, functions is called by
# `optim()` repeatedly
optim_set <- function(par, sets, par_names, err_fun, log_scale, verbose=verbose, env, ...) {
  if(verbose) {
    message(paste("Testing:", paste(par_names, sprintf("%g", par), sep="=", collapse="; ")), appendLF=FALSE)
  }
  # check if parameter names got lost, happens sometimes
  if(is.null(names(par))) {
    names(par) <- par_names
  }
  env$tested <- append(env$tested, list(par))

  rs <- eval_cs(set_param(sets, par), verbose=FALSE, .diagnosis=TRUE, ...)

  # any errors or issues during simulations?
  any_issue <-   any(rs$is_error | rs$is_issue)
  env$success <- c(env$success, !any_issue)
  if(any_issue) {
    if(verbose)
      message("; Simulation failed")
    # save diagnosis data of first caliset that caused issues/errors
    idx_first_issue <- head(which(rs$is_error | rs$is_issue), n=1)
    env$desolve_diagn <- rs$desolve_diagn[[idx_first_issue]]
    env$cvasi_status <- rs$cvasi_status[[idx_first_issue]]

    # penalize parameters causing issues, requires a finite value for method
    # 'L-BFGS-B' to work, therefore returning `Inf` or `NA` is not an option
    return(1e100)
  }

  # calculate error term/target function value on all data points at once
  if(log_scale) {
    err_total <- err_fun(log(rs$obs), log(rs$pred), rs$wgts, rs$tags)
  } else {
    err_total <- err_fun(rs$obs, rs$pred, rs$wgts, rs$tags)
  }

  # save data in environment
  # this assumes that we are always minimizing the error, be careful
  if(is.na(env$err) | err_total < env$err) {
    env$obs <- rs$obs
    env$err <- err_total
    env$pred <- rs$pred
    env$times <- rs$times
  }

  if(verbose) {
    message("; Error: ", sprintf("%g", err_total))
  }

  err_total
}

# @param sets list of [caliset] objects
# @param output character, output column to collect as *predicted* values
# @param verbose logical, if TRUE will print messages about occurred issues etc
# @param method character, chooses the numerical integration scheme of [deSolve::ode()]
# @param ode_method character, overwrites contents of argument `method`
# @param .diagnosis logical, if TRUE then diagnostic data from [deSolve::ode()]
#   will be appended to the result
# @param .suppress logical, if TRUE then messages raised by [simulate()]
#   will be suppressed and the process will not stop in case of errors
# @return named list, has keys such as `obs`, `pred`, and `times` containing
#    *observed*, *predicted*, and output times for all calisets
eval_cs <- function(sets, output, verbose=TRUE, method=NULL, ode_method, .diagnosis=FALSE,
                    .suppress=TRUE, .ignore_method=FALSE, ...) {
  if(!is.list(sets))
    stop("Argument 'sets' must be a list")
  if(length(output) != 1)
    stop("Argument 'output' must be of length one")
  if(.ignore_method)
    method <- NULL
  if(!missing(ode_method))
    method <- ode_method

  rs <- list(
    obs=numeric(0),
    pred=numeric(0),
    times=numeric(0),
    wgts=numeric(0),
    tags=list(),
    is_error=logical(0),
    is_issue=logical(0),
    desolve_diagn=list(),
    cvasi_status=list(),
    err_msg=character(0)
  )

  # cycle through all calibration sets, i.e. scenario and data combinations
  for(i in seq_along(sets))
  {
    # current calibration set
    set <- sets[[i]]

    # coerce obs data to data.frame to avoid issues with data.frame-like types
    data <- as.data.frame(set@data)

    # run simulation
    # FIXME of sim fails e.g. due to invalid forcings format, then this error will not be apparent in the return value, error goes unnoticed?
    if(is.null(method)) {
      out <- try(set@scenario %>% set_times(data[, 1]) %>% simulate(..., .suppress=TRUE), silent=TRUE)
    } else {
      out <- try(set@scenario %>% set_times(data[, 1]) %>% simulate(method=method, .suppress=TRUE, ...), silent=TRUE)
    }

    # check if simulation result contains errors
    is_error <- FALSE
    is_issue <- FALSE
    msg <- NA_character_
    if(is(out, "try-error")) {
      is_error <- TRUE
      msg <- attr(out, "condition")$message
    }
    else if(num_aborted(out)) {
      is_issue <- TRUE
      msg <- "simulation terminated early"
    }
    else if(num_error(out)) {
      is_issue <- TRUE
      msg <- "simulation failed"
    }
    else if(!missing(output)) {
      if(!(all(output %in% names(out)))) {
        is_error <- TRUE
        msg <- "output column not in simulation results"
      }
      else if(any(is.infinite(out[, output]))) {
        is_issue <- TRUE
        msg <- "output column contains infinite values"
      }
      else if(any(is.nan(out[, output]))) {
        is_issue <- TRUE
        msg <- "output column contains NaN values"
      }
      else if(any(is.na(out[, output]))) {
        is_issue <- TRUE
        msg <- "output column contains NAs"
      }
    }

    if(is_issue | is_error)
    {
      msg2 <- paste0("Issue with caliset #", i, "/", length(sets), ": ", msg)
      if(verbose) {
        message(msg2)
      } else if(is_error & !.suppress) {
        warning(msg2, call.=FALSE)
      }
    }

    rs$obs <- c(rs$obs, data[, 2])
    # make sure predictions are present and always have the required length
    if(!is.data.frame(out)) {
      rs$pred <- c(rs$pred, rep(NA_real_, nrow(data)))
    } else if(!(output %in% names(out))) {
      rs$pred <- c(rs$pred, rep(NA_real_, nrow(data)))
    } else {
      pred <- out[, output]
      if(length(pred) != nrow(data)) {
        if(length(pred) < nrow(data))
          pred <- c(pred, rep(NA_real_, nrow(data) - length(pred)))
        else
          pred <- pred[seq(1, nrow(data))]
      }
      rs$pred <- c(rs$pred, pred)
    }
    rs$times <- c(rs$times, data[, 1])
    # if only one weight defined, then apply the same weight to all data points
    if(length(set@weight) == 1) {
      rs$wgts <- c(rs$wgts, rep(set@weight, times=nrow(data)))
    } else if(length(set@weight) == nrow(data))  {
      rs$wgts <- c(rs$wgts, set@weight)
    } else {
      stop("Length mismatch of weights and predicted data, caliset #", i)
    }
    # tags can also be non-atomic types, so we put them in a list
    rs$tags <- append(rs$tags, rep(list(set@tag), times=nrow(data)))
    rs$is_error <- c(rs$is_error, is_error)
    rs$is_issue <- c(rs$is_issue, is_issue)
    if(.diagnosis) {
      rs$desolve_diagn <- append(rs$desolve_diagn, list(attr(out, "desolve_diagn")))
      rs$cvasi_status <- append(rs$cvasi_status, list(attr(out, "cvasi_status")))
    } else {
      rs$desolve_diagn <- append(rs$desolve_diagn, NA)
      rs$cvasi_status <- append(rs$cvasi_status, NA)
    }
    rs$err_msg <- c(rs$err_msg, msg)
  }

  rs
}

# Sum of squared errors with optional weights
# @param orig observed data
# @param pred predicted data
# @param weights weighting factor for each data point
# @param tags optional study tag, e.g. to identify where data originates from
sse <- function(obs, pred, weights=1, tags=NULL) {
  if(length(obs) != length(pred))
    stop("Length mismatch of observed and predicted data.")
  if(length(weights) != 1 & length(weights) != length(obs))
    stop("Length mismatch of weights and observed data.")
  sum((obs - pred)^2 * weights)
}

# Sum of squared errors on logarithmized data
# @param orig observed data
# @param pred predicted data
# @param ... additional parameters passed through to [sse()]
# Only kept for backwards compatibility
log_sse <- function(obs, pred, ...) {
  sse(log(obs), log(pred), ...)
}

#' A fitted GUTS-RED-SD scenario of the fathead minnow
#'
#' The example scenario consists of a fitted [GUTS-RED-SD][GUTS_RED_SD]
#' model and a constant exposure series. Model parameters were derived from a
#' typical four-day acute fish toxicity study of the *fathead minnow* by
#' Geiger *et al.* (1988).
#' The study evaluated the effect of *chlorpyrifos* concentrations in water on
#' survival of *fathead minnows*.
#'
#' The toxicity dataset used for parameter calibration is also referred to as
#' *GUTS Ring-test dataset C* by EFSA (2018). Fitted parameters were estimated
#' using the *morse* package.
#'
#' The exposure series of the example scenario is a constant concentration of 1.0 µmol/L
#' over a period of four days with a daily time step.
#'
#' @seealso [GUTS-RED-models]
#' @references
#' Geiger D.L., Call D.J., and Brooke L.T., 1988: *Acute toxicities of organic
#' chemicals to fathead minnows (Pimephales promelas): Volume IV*, pp. 195-197.
#' University of Wisconsin-Superior, Center for Lake Superior Environmental Studies.
#' ISBN 9780961496838.
#'
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#' @source https://mosaic.univ-lyon1.fr/guts
#' @examples
#' # Print scenario parameters
#' minnow_sd
#'
#' # Run the example scenario
#' minnow_sd %>% simulate()
"minnow_sd"

#' A fitted GUTS-RED-IT scenario of the fathead minnow
#'
#' The example scenario consists of a fitted [GUTS-RED-IT][GUTS_RED_IT]
#' model and a constant exposure series. Model parameters were derived from a
#' typical four-day acute fish toxicity study of the *fathead minnow* by
#' Geiger *et al.* (1988).
#' The study evaluated the effect of *chlorpyrifos* concentrations in water on
#' survival of *fathead minnows*.
#'
#' The toxicity dataset used for parameter calibration is also referred to as
#' *GUTS Ring-test dataset C* by EFSA (2018). Fitted parameters were estimated
#' using the *morse* package.
#'
#' The exposure series of the example scenario is a constant concentration of 1.0 µmol/L
#' over a period of four days with a daily time step.
#'
#' @seealso [GUTS-RED-models]
#' @references
#' Geiger D.L., Call D.J., and Brooke L.T., 1988: *Acute toxicities of organic
#' chemicals to fathead minnows (Pimephales promelas): Volume IV*, pp. 195-197.
#' University of Wisconsin-Superior, Center for Lake Superior Environmental Studies.
#' ISBN 9780961496838.
#'
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#' @source https://mosaic.univ-lyon1.fr/guts
#' @examples
#' # Print scenario parameters
#' minnow_it
#'
#' # Run the example scenario
#' minnow_it %>% simulate()
#'
"minnow_it"

#' Lemna model fitted to *metsulfuron* effect data
#'
#' Data set for the parametrisation of a mechanistic combined
#' toxicokinetic-toxicodynamic (TK/TD) and growth model for the aquatic
#' macrophytes Lemna spp. as published by Schmitt *et al.* (2013).
#' The growth model was parameterised by Schmitt et al. based on these
#' data while toxicokinetic and toxicodynamic parameters were determined by
#' fitting the model using substance specific effect data of the
#' herbicide *metsulfuron-methyl*.
#'
#' The scenario is based on the [Lemna_Schmitt] model.
#'
#' @seealso [Lemna-models], [schmitt2013]
#' @references
#' Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013: *Mechanistic TK/TD-model
#' simulating the effect of growth inhibitors on Lemna populations*. Ecol Model 255,
#' pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
"metsulfuron"

#' A Lemna_SETAC scenario with variable environment
#'
#' A mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and growth
#' model for the aquatic macrophytes Lemna spp. as published by
#' Klein *et al.* (2021).
#'
#' The scenario will simulate a period of 365 days, a start
#' population of 80 g/m² dry weight, variable environmental conditions, and a
#' complex, time-varying exposure pattern.
#'
#' The scenario setup was published by Hommen *et al*. (2015). Exposure pattern
#' and substance specific parameters are of exemplary character
#' and represent the herbicide *metsulfuron-methyl*. The parameters were
#' derived by Schmitt et al. (2013) based on literature
#' data.
#'
#' @seealso [Lemna-models]
#' @references
#' Hommen U., Schmitt W., Heine S., Brock Theo CM., Duquesne S., Manson P., Meregalli G.,
#'   Ochoa-Acuña H., van Vliet P., Arts G., 2015: How TK-TD and Population Models for
#'   Aquatic Macrophytes Could Support the Risk Assessment for Plant Protection
#'   Products. Integr Environ Assess Manag 12(1), pp. 82-95.
#'   \doi{10.1002/ieam.1715}
#'
#' Klein J., Cedergreen N., Heine S., Reichenberger S., Rendal C.,
#'   Schmitt W., Hommen U., 2021: Refined description of the *Lemna* TKTD growth model
#'   based on *Schmitt et al.* (2013) - equation system and default parameters.
#'   Report of the working group *Lemna* of the SETAC Europe Interest Group Effect
#'   Modeling. Version 1.1, uploaded on 09 May 2022.
#'   <https://www.setac.org/group/effect-modeling.html>
#'
#' Schmitt W., Bruns E., Dollinger M., Sowig P., 2013: Mechanistic TK/TD-model
#'   simulating the effect of growth inhibitors on *Lemna* populations. Ecol Model
#'   255, pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
#'
#' @examples
#' # Simulate the example scenario
#' focusd1 %>% simulate()
#'
"focusd1"

#' A DEB abj scenario of Americamysis bahia
#'
#' Species parameters were collected from the AddMyPet database entry on
#' Americamysis bahia (Opossum shrimp). The exposure series consists of a
#' constant exposure resulting in medium effects on length and reproduction.
#'
#' @seealso [DEB_abj()]
#' @source <https://www.bio.vu.nl/thb/deb/deblab/add_my_pet/entries_web/Americamysis_bahia/Americamysis_bahia_res.html>
#'
"americamysis"

#' A DEBtox scenario of Daphnia magna
#'
#' Species and substance parameters were collected from test runs of the
#' original [DEBtox] Daphnia model.
#'
#' @seealso [DEBtox()]
"dmagna"

#' A Lemna effect study data set with multiple treatment levels
#'
#' Data was collected from Schmitt et al. (2013). The study observed the number of
#' Lemna fronds during the study period of 14 days. The organisms were exposed
#' to multiple concentrations of the sulfonyl urea herbicide *metsulfuron-methyl*.
#'
#' The number of Lemna fronds was recorded at the beginning of the experiment (t=0)
#' and after 3, 5, 7, 10, 12, and 14 days. The exposure to *metsulfuron-methyl*
#' was only present during the first seven days, which was followed by a recovery
#' period of another seven days without exposure to the substance.
#'
#' The tabular dataset consists of four columns:
#'
#' - First column: time (days)
#' - Second column: observed number of fronds (-)
#' - Third column: trial id (-)
#' - Fourth column: metsulfuron-methyl concentration (ug/L)
#'
#' An example [scenario] with parameters fitted to the experimental data set
#' as conducted by Schmitt et al. is available as [`metsulfuron`].
#'
#' @seealso [Lemna-models], [metsulfuron]
#' @references
#' Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013: *Mechanistic TK/TD-model
#' simulating the effect of growth inhibitors on Lemna populations*. Ecol Model 255,
#' pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
"schmitt2013"

#' A Weber scenario of algae exposed to isoproturon
#'
#' The scenario recreates the conditions of algae species *R. subcapitata* exposed
#' to isoproturon in a flow-through reactor experiment, see Weber et al. (2012)
#' publication. Scenario parameters and exposure series were set according to
#' reported values to represent conditions in reactor A, as depicted in
#' Figure 4A in Weber et al.
#'
#' @seealso [Algae_Weber]
#' @references
#' Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
#' and Ratte HT, 2012. Combination of a higher-tier flow-through system and
#' population modeling to assess the effects of time-variable exposure of
#' isoproturon on the green algae Desmodesmus subspictatus and
#' Pseudokirchneriella subcapitata. Environmental Toxicology and
#' Chemistry, 31, 899-908. \doi{10.1002/etc.1765}
"rsubcapitata"

#' EFSA Ringtest Dataset C
#'
#' This dataset is part of EFSA's *GUTS* ringtest to compare results from software
#' implementations of [GUTS-RED models][GUTS-RED-models] (EFSA 2018). The ringtest
#' focused on [GUTS-RED-IT][GUTS_RED_IT] and [GUTS-RED-SD][GUTS_RED_SD] models.
#'
#' *Dataset C* in this ringtest came from the work of Geiger et al. (1988) on effects of
#' *Diazinon* on *fathead minnow*. The dataset contains four columns, identifying the control
#' and treatment levels, the concentrations in these treatments, the timepoint of measurement
#' and the number of survivors.
#'
#' @seealso [GUTS-RED-models]
#' @references
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#' Geiger DL, Call DJ, and Brooke LT, 1988: Acute Toxicities of Organic Chemicals to
#' Fathead Minnow (*Pimephales promelas*), Vol IV. Center for Lake Superior Environmental
#' Studies, University of Wisconsin-Superior, Superior, WI, USA.
#'
#' https://cran.r-project.org/web/packages/GUTS/vignettes/ringTest.html
"ringtest_c"

#' EFSA Ringtest Dataset A SD
#'
#' This dataset is part of EFSA's *GUTS* ringtest to compare results from software
#' implementations of [GUTS-RED models][GUTS-RED-models] (EFSA 2018). The ringtest
#' focused on [GUTS-RED-IT][GUTS_RED_IT] and [GUTS-RED-SD][GUTS_RED_SD] models.
#'
#' *Dataset A SD* for the GUTS ring test is a Hypothetical data set constructed with SD.
#'
#' @seealso [GUTS-RED-models]
#' @references
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#'
#' https://cran.r-project.org/web/packages/GUTS/vignettes/ringTest.html
"ringtest_sd"

#' EFSA Ringtest Dataset A SD
#'
#' This dataset is part of EFSA's *GUTS* ringtest to compare results from software
#' implementations of [GUTS-RED models][GUTS-RED-models] (EFSA 2018). The ringtest
#' focused on [GUTS-RED-IT][GUTS_RED_IT] and [GUTS-RED-SD][GUTS_RED_SD] models.
#'
#' *Dataset A IT* for the GUTS ring test is a Hypothetical data set constructed with IT.
#'
#' @seealso [GUTS-RED-models]
#' @references
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, et al., 2018: *Scientific Opinion on the
#' state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
#' regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
#' 16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
#'
#'
#' https://cran.r-project.org/web/packages/GUTS/vignettes/ringTest.html
"ringtest_it"

# Generated by roxyglobals: do not edit by hand

utils::globalVariables(c(
  "obs", # <calibrate_list>
  "pred", # <calibrate_list>
  "log_obs", # <calibrate_list>
  "log_pred", # <calibrate_list>
  "endpoint", # <dose_response>
  "window.start", # <effect_scenario>
  "window.end", # <effect_scenario>
  "scenario", # <effect_scenario>
  "endpoint", # <epx_min_win>
  "level", # <epx_min_win>
  "EPx", # <epx_min_win>
  "window.start", # <epx_min_win>
  "size", # <epx_min_win>
  "ep", # <epx_binary_search>
  "level", # <epx_binary_search>
  "mf", # <epx_binary_search>
  "tags", # <explore_space>
  "LLR_accept", # <explore_space>
  "time", # <fit_growth_lemna>
  "time", # <fit_growth_magma>
  "time", # <fit_tktd_lemna_setac>
  "time", # <fit_tktd_magma>
  "V1", # <import_toxswa>
  "V3", # <import_toxswa>
  "value", # <import_toxswa>
  "name", # <import_toxswa>
  "time", # <import_toxswa>
  "timestamp", # <import_toxswa>
  "tags", # <profile_par>
  ".", # <format_table>
  "par_value", # <plot.lik_profile>
  "log_lik_rat", # <plot.lik_profile>
  "start", # <plot.lik_profile>
  "LL", # <plot.param_space>
  "LLR_quality", # <plot.param_space>
  "trial", # <plot_sd>
  "time", # <plot_ppc>
  "trial", # <plot_ppc>
  "obs", # <plot_ppc_combi>
  "PPC", # <plot_ppc_combi>
  "pred", # <plot_ppc_combi>
  "study", # <plot_ppc_combi>
  "window.start", # <plot_epx>
  "window.end", # <plot_epx>
  "endpoint", # <plot_epx>
  "len", # <plot_epx>
  "EPx", # <plot_epx>
  ".data", # <plot_epx>
  "scenario", # <pull_metadata>
  "context", # <pull_metadata>
  "tag", # <set_param_multi>
  "model", # <set_param_multi>
  "name", # <set_param_multi>
  "desc", # <set_param_multi>
  "S", # <survival>
  "time", # <surv_data>
  "Nsurv", # <surv_data>
  "conc", # <surv_data>
  NULL
))

#' Calculate EPx values for a series of moving time window
#'
#' Calls `epx()` to calculate the EPx value (i.e. the multiplication factors of
#' an exposure profile that cause x% of effect) for moving windows with length
#' `window_length` that move timesteps defined by `window_interval`.
#'
#' @param x a [scenario]
#' @param level The target effect level of the effect, ie. the x of EPx.
#' @param factor_cutoff above which cutoff is the EPx is not relevant
#' @param window_length the length of the moving time window
#' @param window_interval the interval that the moving time window moves
#' @param ... arguments passed to `epx`
#'
#' @export
#' @return a tibble with five columns
#' \itemize{
#'  \item window.start
#'  \item window.end
#'  \item endpoint
#'  \item level
#'  \item EPx
#'  }
#'
#' @examples
#' metsulfuron %>%
#'   set_window(length=7, interval=1) %>%
#'   epx_mtw()
epx_mtw <- function(x,
                    level = c(10, 50),
                    factor_cutoff = 1000,
                    window_length = 7,
                    window_interval = 1,
                    ...) {
  x <- x %>%
    set_window(length = window_length, interval = window_interval)

  windows <- x %>% window_candidates()

  res <- lapply(windows, function(window_i) {
    scenario <- x %>%
      clip_scenario(window_i)
    epx_res <- epx(scenario,
      level = level,
      factor_cutoff = factor_cutoff,
      long_format = TRUE,
      ep_only = TRUE,
      ...
    )

    dplyr::bind_cols(t(window_i), epx_res)
  })

  return(do.call(rbind, res))
}


# Detect worst case EPx window
#
# Filters the minimal EPx per endpoint and level from the `epx_mtw` object and
# returns it as the worst case windows. If there is more than one worst case
# window, only the first window will be returned (if `only_first == TRUE`), or
# all windows will be returned (if `only_first == FALSE`).
#
# @param epx_mtw result of `epx_mtw()`
# @param only_first TRUE = first window only
# @noRd
#' @autoglobal
#
# @return a tibble with
# \itemize{
#  \item window.start
#  \item window.end
#  \item endpoint
#  \item level
#  \item EPx
#  }
epx_min_win <- function(epx_mtw, only_first = TRUE){
  wc_window <- epx_mtw %>%
    dplyr::group_by(endpoint, level) %>%
    dplyr::filter(EPx == min(EPx)) %>%
    dplyr::arrange(window.start, endpoint, level) %>%
    dplyr::ungroup()

  # check if there is more than one worst case window per endpoint
  incr_wc_window <- wc_window %>%
    dplyr::group_by(endpoint) %>%
    dplyr::summarise(size = dplyr::n()) %>%
    dplyr::filter(size > 1) %>%
    dplyr::select(endpoint)

  # remove all but the first moving time window if more than one exist
  if (only_first){
    wc_window <- wc_window %>%
      dplyr::group_by(endpoint, level) %>%
      dplyr::filter(window.start == min(window.start)) %>%
      dplyr::arrange(window.start, endpoint, level) %>%
      dplyr::ungroup()
  }

  if (incr_wc_window %>% nrow() > 0){
    warning(
      paste0("More than one worst case window found for ",
             incr_wc_window,
             ifelse(only_first, ". All later instances removed","")
      )
    )
  }

  return(wc_window)
}

#' Effect profiles (EPx values)
#'
#' Derives one or more EPx/LPx values for the supplied effect scenarios, i.e. it
#' calculates the multiplication factors of an exposure profile that cause
#' x% of effect. Scenarios are processed in parallel, if possible.
#'
#' To estimate EPx values, a *binary search* on multiplication factors is conducted.
#' The algorithm can achieve arbitrary precision in terms of effects. The
#' same approach is implemented in the `morse` package in the `MFx()` function.
#' Convergence is often achieved in less than 10 iterations per effect level and
#' endpoint.
#'
#' Internally, a knowledge base of all tried factors and resulting effect levels is
#' kept to speed up convergence if more than one endpoint or effect level was
#' requested. The algorithm will automatically sweep the range of multiplication
#' factors as needed but hard cutoff values are implemented to avoid infinite loops;
#' the algorithm will halt with an error message if tried factors are
#' smaller than `1e-30` or greater than `1e30`.
#'
#' ### Numerical precision
#' The precision of reported *EPx* values is controlled by the argument
#' `effect_tolerance` and is given as the upper absolute error threshold of
#' effects that is deemed acceptable. The default value of `0.001` ensures that
#' a derived *EPx* will result in an effect of x% ± 0.1. Decreasing the
#' `effect_tolerance` will result in additional model iterations and longer
#' runtime. Setting an extremely small tolerance value may lead to a breakdown
#' of the algorithm due to the occurrence of extremely small, quasi-random
#' numerical errors in simulation results.
#'
#' @param scenarios table or vector of `EffectScenario` objects
#' @param level effect levels in percent (%), defaults to `c(10,50)`
#' @param effect_tolerance `numeric`, minimum absolute accuracy of effect levels
#' @param factor_cutoff optional `numeric`, the search for a multiplication factor
#' will be cut short if  tried factors exceed this value; the result will report
#' the cutoff value as the final EPx value.
#' @param verbose `logic`, if `TRUE` then infos about model evaluations are displayed
#' @param ... additional arguments passed on to [effect()]
#' @param min_factor `numeric`, if tried factors fall below this threshold, the algorithm
#' will halt with an error
#' @param max_factor  `numeric`, if tried factors exceed this threshold, the algorithm
#' will halt with an error
#' @param ep_only `logical`, if `TRUE` then only EPx values are part of the output,
#'   any contextual information such as `EffectScenario` objects are left out
#' @param long_format `logical`, if `TRUE` then EPx values are returned as
#'   a table in long format, any contextual information will be duplicated
#'
#' @return
#' The original `tibble` with additional columns named after the request effect levels, e.g. `L.EP10.`
#' If no tibble was used as argument, then a new one is created. The first column `scenario` will contain
#' the supplied `EffectScenario` objects.
#' @export
#' @importFrom stats setNames
#'
#' @examples
#' minnow_sd %>% epx()
#' minnow_sd %>% epx(level=c(10,23,42))
#'
#' # displays infos about tested multiplication factors
#' minnow_sd %>% epx(verbose=TRUE)
#'
#' # return results as a table in wide format
#' minnow_sd %>% epx(long_format=TRUE)
epx <- function(scenarios, level=c(10,50), effect_tolerance=0.001, factor_cutoff=NA,
                min_factor=1e-30, max_factor=1e30, verbose=FALSE, ep_only=FALSE,
                long_format=FALSE, ...)
{
  if(any(level>100) | any(level<1))
    stop("epx: requested effect levels out of range")

  # gracious argument handling if single object or vector was supplied
  if(!is.vector(scenarios))
    scenarios <- tibble::tibble(scenario=c(scenarios))
  else
    scenarios <- tibble::tibble(scenario=scenarios)

  # errors should not break code flow, instead catch them
  searchfun <- purrr::safely(epx_binary_search, otherwise=NA_real_)

  # sequential processing requested?
  if(getOption("cvasi.pll.off", FALSE)) {
    ret <- lapply(scenarios$scenario, searchfun, level=level,
             effect_tolerance=effect_tolerance, factor_cutoff=factor_cutoff,
             min_factor=min_factor, max_factor=max_factor,
             verbose=verbose, ...)
  } else {
    # disable collection of globals by furrr to increase performance
    ret <- furrr::future_map(scenarios$scenario, searchfun, level=level,
                      effect_tolerance=effect_tolerance, factor_cutoff=factor_cutoff,
                      min_factor=min_factor, max_factor=max_factor,
                      verbose=verbose, .options=furrr::furrr_options(globals=FALSE), ...)
  }

  failed <- is.na(unlist(ret))
  if(any(failed)) {
    # collect error messages, 1st from binary search
    # 2nd, any unexpected occurrences encountered by purrr/furrr
    rs <- unlist(ret)
    errors <- rs[names(rs) == "result.error" | names(rs) == "error"]
    msg <- paste0("Some scenarios have failed (n=", sum(failed), "), result will include NAs")
    if(length(errors) > 0)
      msg <- paste0(msg, "\n", paste(paste0("  ** ", errors), collapse="\n"))
    warning(msg)
    rm(rs, errors, msg)
  }
  rs <- dplyr::bind_rows(lapply(ret, `[[`, "result"))

  # convert to long format table
  if(long_format) {
    if(!ep_only)
      rs <- dplyr::bind_cols(scenarios, rs)
    return(tidyr::pivot_longer(rs, tidyr::contains("."), names_to=c("endpoint", "level"),
                        names_pattern="(.+)\\.EP(.+)", values_to="EPx"))
  }

  # extend original table by EPx columns
  if(ep_only)
    rs
  else
    dplyr::bind_cols(scenarios, rs)
}

# EPx values are estimated using a binary search on the range of multiplication
# factors. The same approach is used in the morse package. Internally, a
# knowledge base (kbs) is employed in order to avoid re-evaluation of MFs and to
# achieve a faster start-up and convergence in case of multiple endpoints or
# effect levels. The approach has several advantages:
# - it usually requires a lot less effect() calls than alternative methods
#   which is especially useful for slow running models
# - it can achieve arbitrary precision w.r.t. to effects
# - and it automatically covers the whole range of multiplication factors if necessary
#' @global ep level mf
epx_binary_search <- function(scenario, level, effect_tolerance, factor_cutoff, verbose,
                              min_factor, max_factor, .cache, ...)
{
  # init result vector before anything else can fail
  result <- expand.grid(ep=scenario@endpoints, level=level) %>%
    dplyr::mutate(name=paste0(ep, ".EP", level)) %>%
    dplyr::pull(3)
  result <- as.list(setNames(rep(NA_real_, length(result)), result))

  tryCatch({
    # control scenarios?
    if(missing(.cache)) {
      .cache <- cache_windows(scenario, ...)
    }

    # rounded effects in kbs need to be more precise than the requested tolerance
    e.decimals <- max(4, ceiling(abs(log10(effect_tolerance / 10))))
    # start value of the range sweep
    f.test <- 10
    if(verbose) {
      message("epx: screening multiplication factors")
      message(paste0("  start: ", f.test))
    }
    e.test <- effect(scenario, factor=f.test, ep_only=TRUE, .cache=.cache, ...)
  }, error=function(e) {
    result$error <<- e$message
  })

  if(!exists("e.test"))
    return(result)
  if(all(is.na(e.test)))
    return(result)

  # create the knowledge base
  kbs <- as.data.frame(as.list(c(mf=f.test, e.test)))

  for(ep in scenario@endpoints)
  {
    for(lv in level)
    {
      tryCatch({
        ep_name <- paste0(ep, ".EP", lv)
        # Out Of Range error conditions
        oor_error <- FALSE

        e.tgt <- lv / 100
        # if MF smaller than requested level not found yet, test smaller powers of ten
        while(min(kbs[[ep]]) > e.tgt) {
          f.test <- min(kbs[["mf"]]) / 10
          e.test <- effect(scenario, f.test, ep_only=TRUE, .cache=.cache, ...)
          kbs <- dplyr::bind_rows(kbs, c(mf=f.test, e.test))
          if(verbose)
            message(paste0("  ", ep_name, ": ", f.test, " <<<"))
          if(f.test < min_factor) {
            result$error <- paste(result$error, "multiplication factor out of range:", ep_name, "<", min_factor)
            oor_error <- TRUE
            break
          }

        }
        # if MF larger than requested level not found yet, test larger powers of ten
        while(max(kbs[[ep]]) < e.tgt) {
          f.test <- max(kbs[["mf"]]) * 10
          e.test <- effect(scenario, f.test, ep_only=TRUE, .cache=.cache, ...)
          kbs <- dplyr::bind_rows(kbs, c(mf=f.test, e.test))
          if(verbose)
            message(paste0("  ", ep_name, ": ", f.test, " >>>"))
          if(any(f.test >= factor_cutoff, na.rm=TRUE)) break
          else if(f.test > max_factor) {
            result$error <- paste(result$error, "multiplication factor out of range:", ep_name, ">", max_factor)
            oor_error <- TRUE
            break
          }
        }
        # skip this endpoint
        if(oor_error) {
          next
        }

        # sometimes effects in kbs will not increase continuously with multiplication
        # factors due to small numerical errors. these inconsistencies are irrelevant
        # if they appear e.g. in the 10th+ decimal place. therefore, effect values
        # are rounded at a suitable decimal
        kbs[[ep]] <- round(kbs[[ep]], e.decimals)
        # sort knowledge base
        dplyr::arrange(kbs, mf) -> kbs
        # check if effects are also ordered in an ascending fashion, otherwise
        # something bad has happened
        if(!all(kbs[[ep]] == sort(kbs[[ep]]))) {
          show(kbs)
          stop("effects do not fully correlate with multiplication factors, please check numerical stability", call.=FALSE)
        }
        # find lower factor boundary
        f.min <- kbs$mf[sum(kbs[[ep]] <= e.tgt)]
        # find upper factor boundary, if possible
        f.max <- rev(kbs$mf)[sum(kbs[[ep]] > e.tgt)]
        # upper boundary was not included in kbs, can only happen if we reached
        # the cutoff criterion
        if(length(f.max) == 0) f.max <- f.min

        # starting point of binary search
        f.cur <- f.min
        e.cur <- kbs[[ep]][sum(kbs[[ep]] <= e.tgt)]

        # keep halving the search interval until tolerance level is met
        while(abs(e.cur - e.tgt) > effect_tolerance) {
          if(e.cur < e.tgt) {
            f.min <- f.cur
            f.cur <- f.min + (f.max-f.min)/2
          } else if(e.cur > e.tgt) {
            f.max <- f.cur
            f.cur <- f.max - (f.max-f.min)/2
          }
          if(any(f.cur>=factor_cutoff, na.rm=TRUE)) {
            f.cur <- factor_cutoff
            break
          }
          # If max and min factor are almost identical, then search ran into a dead
          # end. This is caused by instable numerics which results in sudden leaps
          # in apparent effects and we end up with an inconsistent KBS.
          if(f.max - f.min < 10*.Machine$double.eps*f.min) # if TRUE, then values are almost equal
            stop("binary search failed, please check numerical stability", call.=FALSE)

          if(f.cur<min_factor) {
            result$error <- paste(result$error, "multiplication factor out of range:", ep_name, "<", min_factor)
            f.cur <- NA_real_
            break
          }
          if(f.cur>max_factor) {
            result$error <- paste(result$error, "multiplication factor out of range:", ep_name, ">", max_factor)
            f.cur <- NA_real_
            break
          }

          e.cur <- effect(scenario,f.cur,ep_only=TRUE, .cache=.cache,...)
          if(any(e.cur<0 | e.cur>1 | is.nan(e.cur)))
            stop("invalid effect level occurred, check numerical stability", call.=FALSE)
          kbs <- dplyr::bind_rows(kbs, c(mf=f.cur,e.cur))
          e.cur <- e.cur[[ep]]
          if(verbose) message(paste0("  ",ep_name,": ",f.cur))
        }
        result[[ep_name]] <- f.cur
      }, error=function(e) {
        result$error <- e$message
        return(result)
      })
    }
  }
  result
}

#' Generic to calculate effects for a particular scenario
#'
#' @param scenario [scenario] object
#' @param ... additional parameters
#' @return numeric named vector
#' @export
setGeneric("fx", function(scenario, ...) standardGeneric("fx"), signature = "scenario")

# Default effects on state variables at end of the simulated period
#' @describeIn fx Use state variables at end of simulation
setMethod("fx", "ANY", function(scenario, ...) fx_default(scenario, ...))

#' @describeIn fx Wrapper for [sequences][sequence]
setMethod("fx", "ScenarioSequence", function(scenario, ...) fx_sequence(scenario, ...))

# Use value of state variable at end of simulation to derive effect
fx_default <- function(scenario, ...) {
  res <- simulate(scenario, ...)
  if(!num_success(res)) {
    return(NA)
  }

  row <- tail_nm(res)
  # setNames() is required if some endpoints are not present in the output
  # of simulate(). in this case the return value would contain columns named
  # `<NA>`, which we want to avoid.
  setNames(row[scenario@endpoints], scenario@endpoints)
}

#' @importFrom methods selectMethod
fx_sequence <- function(scenario, ...) {
  # use type of first scenario in sequence to select an appropriate `fx` function
  cls <- class(scenario[[1]])[1]
  fun <- selectMethod("fx", signature=c(scenario=cls))
  fun(scenario, ...)
}

# return the last row of a data.frame or matrix as a vector and assures that
# column names are retained
#' @importFrom utils head tail
tail_nm <- function(data) {
  if(is.data.frame(data))
    row <- unlist(tail(data,1))
  else if(is.matrix(data)) {
    row <- as.vector(tail(data,1))
    names(row) <- colnames(data)
  } else {
    stop("unknown type")
  }
  row
}

test_that("default fx", {
  sc <- minnow_it %>% set_endpoints(c("D", "H"))
  sim <- unlist(tail(simulate(sc), n=1))

  # the return value of `fx_default()` is simply the value at the end of the
  # simulation

  # single endpoint
  sc <- sc %>% set_endpoints("D")
  expect_equal(fx_default(sc), sim["D"])
  # multiple endpoints
  sc <- sc %>% set_endpoints(c("D", "H"))
  expect_equal(fx_default(sc), sim[c("D", "H")])
  # endpoint not contained in simulation output
  sc <- sc %>% set_endpoints(c("D", "foobar"))
  expect_equal(fx_default(sc), c(sim[c("D")], "foobar"=NA_real_))

  # if simulate() failed for some reason, e.g. the numerical method became instable,
  # fx() should return `NA`
  source(test_path("dummy.R"), local = TRUE)
  fails <- new("DummyScenario", simulate=function(...) NA_real_)
  expect_true(is.na(fx_default(fails)))
})

test_that("tail_nm", {
  # data.frame
  expect_equal(tail_nm(data.frame(a=0, b=1)), c(a=0, b=1))
  expect_equal(tail_nm(data.frame(a=0:3, b=1:4)), c(a=3, b=4))

  # matrix
  expect_equal(tail_nm(as.matrix(data.frame(a=0, b=1))), c(a=0, b=1))
  expect_equal(tail_nm(as.matrix(data.frame(a=0:3, b=1:4))), c(a=3, b=4))

  # other types
  expect_error(tail_nm(c(a=1, b=2)))
  expect_error(tail_nm(list(a=1, b=2)))
})

show_scenario <- function(object, inline=FALSE, ...) {
  if(!inline)
    cli::cli_text("'{object@name}' scenario")

  if(object@tag != "" & !is.na(object@tag))
    cli::cli_text("tag: {object@tag}")

  if(length(object@param) > 0)
    cli::cli_text(cli::col_blue(paste("param:", paste(names(object@param), unlist(object@param), sep="=", collapse=", "))))
  else
    cli::cli_text(cli::col_blue("param: none"))

  if(length(object@init) > 0)
    cli::cli_text(cli::col_blue(paste("init:", paste(names(object@init), unlist(object@init), sep="=", collapse=", "))))
  else
    cli::cli_text(cli::col_blue("init: none"))

  if(length(object@endpoints) > 0)
    cli::cli_text(cli::col_blue(paste("endpt:", paste(object@endpoints, sep=",", collapse=", "))))
  else
    cli::cli_text(cli::col_blue("endpt: none"))

  if(length(object@times) > 0)
    cli::cli_text(cli::col_blue(paste0("times: [", min(object@times), ",", max(object@times), "] n=",
                         length(object@times), ", ", ifelse(is_regular_series(object@times), "regular", "irregular"))))
  else
    cli::cli_text(cli::col_blue("times: none"))

  # biomass transfers
  if(has_transfer(object)) {
    # regular transfers
    if(has_regular_transfer(object)) {
      desc <- paste0("regular, interval=", object@transfer.interval )
    } else {
      tps <- paste(sprintf("%g", head(object@transfer.times, n=6)), collapse=",")
      desc <- paste0("custom, times=c(", tps, ifelse(length(object@transfer.times) > 6, ",...", ""), ")")
    }
    cli::cli_text(cli::col_br_magenta("transf: ", desc, ", ", object@transfer.comp.biomass, "=", sprintf("%g", object@transfer.biomass)))
  }

  # moving windows
  if(has_windows(object)) {
    cli::cli_text(cli::col_br_magenta("windws: length=", sprintf("%g", object@window.length), ", interval=", sprintf("%g", object@window.interval)))
  }

  cli::cli_text(cli::col_green(paste("forcs: ",ifelse(length(object@forcings) == 0, "none", paste0(names(object@forcings), collapse=", ")))))
  show_exposure(object@exposure, inline=TRUE, ...)
}

show_exposure <- function(x, inline=FALSE, show_exposure_series=TRUE) {
  color <- "green"
  nofile <- identical(x@file, character(0))
  if(!nofile) # the expression yields logical(0) for @file=character(0)
    nofile <- x@file %in% c("unknown","")

  if(!inline) {
    cli::cli_text("ExposureSeries object")
    cli::cli_text(cli::col_green("file: {ifelse(nofile, 'none', x@file)}"))
  } else {
    cli::cli_text(cli::col_green("expsr: {ifelse(nofile, '', x@file)}"))
  }

  if(nrow(x@series) > 0) {
    if(show_exposure_series) {
      print(x@series, max=20)
    }
  } else {
    cli::cli_text(cli::col_green(">> exposure series is empty"))
  }
}

#' @importFrom methods show
setMethod("show", "EffectScenario", function(object) show_scenario(object))
setMethod("show", "ExposureSeries", function(object) show_exposure(object))

test_that("show_scenario", {
  capture.output(suppressMessages(show_scenario(minnow_it)))
  succeed()
})

test_that("show_exposure", {
  capture.output(suppressMessages(show_exposure(minnow_it@exposure)))
  succeed()
})

test_that("show biomass transfers", {
  out <- cli::cli_fmt(capture.output(show_scenario(metsulfuron %>% set_transfer(interval=23, biomass=42))))
  expect_true(any(stringr::str_detect(out, "regular.*interval=23.*BM=42")))

  out <- cli::cli_fmt(capture.output(show_scenario(metsulfuron %>% set_transfer(times=c(0,1,2), biomass=42))))
  expect_true(any(stringr::str_detect(out, "times=c\\(0,1,2\\).*BM=42")))
})

test_that("show moving windows", {
  out <- cli::cli_fmt(capture.output(show_scenario(metsulfuron %>% set_window(length=1, interval=2))))
  expect_true(any(stringr::str_detect(out, "length=1.*interval=2")))
})

#' S3 plotting functions
#'
#' These functions overload [base::plot()] to provide simple plotting
#' routines to display various time-series and scenario objects.
#'
#' @name plot
#' @param x object to plot
#' @param y unused parameter
#' @param ... unused parameters
#' @returns (*ggplot2*) plot object
NULL

# S3 overload to plot dose response curves
#
#' @param scale_x character, controls how the x-axis is scaled. `log10` for
#'   a log10-scaled axis, `none` for no scaling, and `auto` for automatic selection
#' @describeIn plot Plot dose response curves
#' @export
plot.cvasi_drc <- function(x, y, scale_x=c("auto", "log10", "none"), ...) {
  if(!missing(y))
    warning("Parameter `y` is unused, ignoring argument")
  if(!missing(...))
    warning("Parameter `...` is unused, ignoring additional arguments")

  scale_x <- match.arg(scale_x)
  if(scale_x == "auto") {
    exp_min <- floor(log10(min(x$mf)))
    exp_max <- ceiling(log10(max(x$mf)))
    if(is.na(exp_min) | is.na(exp_max))
      scale_x <- "none"
    else if(abs(exp_max - exp_min) > 2)
      scale_x <- "log10"
    else
      scale_x <- "none"
  }

  plot <- ggplot2::ggplot(x) +
    ggplot2::geom_line(ggplot2::aes(mf, effect, color=endpoint))
  if(scale_x == "log10") {
    plot <- plot + ggplot2::scale_x_log10()
  }

  # create a nice plot title
  subtitle <- ggplot2::element_blank()
  sc <- attr(x, "scenario")
  if(!is.null(sc)) {
    nm <- get_model_name(sc)
    tg <- get_tag(sc)
    if(is.na(tg))
      subtitle <- nm
    else
      subtitle <- paste0(nm, " #", tg)
  }

  plot +
    ggplot2::scale_color_discrete(name="Endpoint") +
    ggplot2::labs(title="Dose response curve",
                  subtitle=subtitle,
                  x=ifelse(scale_x == "log10", "Multiplication factor [log-scale] (-)", "Multiplication factor (-)"),
                  y="Effect (-)") +
    ggplot2::theme_bw()
}

#' @describeIn plot Plot return value of [simulate()]
#' @export
plot.cvasi_simulate <- function(x, y, ...) {
  if(!missing(y))
    warning("Parameter `y` is unsused, ignoring argument")
  if(!missing(...))
    warning("Parameter `...` is unused, ignoring additional arguments")

  # pivot table in wide-format to long-format. we have to assume that the
  # first column represents time. the remaining numerical columns will be
  # plotted.
  tidyr::pivot_longer(x, cols=!c(1) & dplyr::where(is.numeric)) %>%
    dplyr::rename(time=1) %>%                                  # make sure we have a well defined name
    dplyr::mutate(name=factor(name, levels=unique(name))) %>%  # enforce ordering of state variables in plot
    ggplot2::ggplot() +
    ggplot2::geom_line(ggplot2::aes(time, value, color=name)) +
    ggplot2::facet_wrap(~ name, scales="free") +
    ggplot2::labs(title="Simulation result", x="Time", y="Value (?)") +
    ggplot2::guides(color="none") +
    ggplot2::theme_bw()
}

#' @describeIn plot Plot likelihood profiles.
#' @autoglobal
#' @export
plot.lik_profile <- function(x, y, ...) {
  if(!missing(y))
    warning("Parameter `y` is unsused, ignoring argument")
  if(!missing(...))
    warning("Parameter `...` is unused, ignoring additional arguments")

  npars <- length(x)
  n1 <- ceiling(sqrt(npars))

  args <- attr(x, "args", exact=TRUE)

  # create one plot per profiled parameter
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  plots <- list()
  for(nm in names(x)) {
    df <- x[[nm]][["likelihood_profile"]]
    plot <- df %>%
      ggplot2::ggplot() +
      # # zone of acceptabiity
      # ggplot2::geom_rect(
      #   ggplot2::aes(xmin=-Inf, xmax=Inf, ymin=min(log_lik_rat), ymax=args$chi_crit_s),
      #   fill = "darkolivegreen3", alpha = 0.2
      # ) +
      ggplot2::annotate("rect", xmin=-Inf, xmax=Inf, ymin=min(df$log_lik_rat), ymax=args$chi_crit_s,
                        alpha=0.3, fill="darkolivegreen3") +
      ggplot2::geom_hline(yintercept=args$chi_crit_s, color="tomato", linewidth=1.2) +
      # points for values
      ggplot2::geom_point(ggplot2::aes(x=par_value, y=log_lik_rat, color=start), size=2) +
      ggplot2::scale_color_manual(values = c("black", "orange")) +
      # connect points
      ggplot2::geom_line(ggplot2::aes(x=par_value, y=log_lik_rat)) +
      # could add 95% CI
      ggplot2::geom_vline(xintercept=x[[nm]][["confidence_interval"]], linetype="dashed", color="grey30", linewidth=.5) +
      ggplot2::labs(x=nm, y="LL ratio (-)") +
      # cosmetics
      ggplot2::scale_y_continuous(limits=c(min(df$log_lik_rat), NA), expand=c(0, 0)) +
      ggplot2::theme_classic()
    plots <- append(plots, list(plot))
  }
  do.call(gridExtra::grid.arrange, c(plots, list(nrow = n1)))
}

#' @describeIn plot Alias of `plot.lik_profile()` for backwards-compatibility.
#' @export
plot_lik_profile <- function(x) {
  lifecycle::deprecate_soft("1.5.0", "cvasi::plot_lik_profile()", "plot()")
  plot(x=x)
}


#' Plot profiled parameter space
#'
#' The function provides bivariate parameter space plots indicating
#' parameter draws (from the 95% confidence intervals per parameter obtained
#' through likelihood profiling) that fall within the inner rim (in green, i.e.
#' parameter sets which are not significantly different from the original, based
#' on a chi-square test). The original parameter set is also indicated (in orange),
#' and, if different from the original set, the best fit parameter set is indicated
#' (in red).
#'
#' @param x Return value of [explore_space()]
#' @param y Unused argument
#' @param ... Unused arguments
#' @returns plot object
#' @autoglobal
#' @export
plot.param_space <- function(x, y, ...){
  if(!missing(y))
    warning("Parameter `y` is unsused, ignoring argument")
  if(!missing(...))
    warning("Parameter `...` is unused, ignoring additional arguments")

  # find best LLR value, and compare with original LLR (which is per definition = 0)
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if (min(x$LLR) < 0) {
    plot_colors <- c("tomato", "orange", "#4DAC26",  "#0571B0")
    plot_size <- c(3, 3, 1.5, 1.5)
    index <- which(x$LLR_quality %in% c("Original fit", "Best fit"))
    x$LLR_quality <- as.factor(x$LLR_quality)
    x$LLR_quality <- factor(x$LLR_quality,
                            levels = c("Best fit", "Original fit", "Inner", "Outer"))
  } else {
    plot_colors <- c( "orange","#4DAC26", "#0571B0")
    plot_size <- c(3, 1.5, 1.5)
    index <- which(x$LLR_quality %in% "Original fit")
    x$LLR_quality <- as.factor(x$LLR_quality)
    x$LLR_quality <- factor(x$LLR_quality,
                            levels = c("Original fit", "Inner", "Outer"))
  }


  # reorder so that the original and best fit are the last datapoints (and
  # thus are plotted on top)
  x_best <- x[index,]
  x_withoutbest <- x[-index,]
  x <- rbind(x_withoutbest, x_best)

  # plot of all correlations
  # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # dummy plot to get a legend
  p <- x %>%
    ggplot2::ggplot(
      ggplot2::aes(
        x = colnames(x)[1],
        y = LL,
        color = LLR_quality)) +
    ggplot2::geom_point() +
    ggplot2::scale_color_manual(values = plot_colors)
  leg <- GGally::grab_legend(p)

  # actual plot, complete
  full_plot <- x %>%
    GGally::ggpairs(
      columns = colnames(x)[-which(colnames(x) %in% c("LL", "LLR", "LLR_quality"))],
      ggplot2::aes(
        color = LLR_quality,
        size = as.factor(LLR_quality)
      ),
      upper = list(continuous = "blank"),
      diag = list(continuous = "blankDiag"),
      title = "Parameter space",
      legend = leg
    ) +
    ggplot2::scale_color_manual(values = plot_colors) +
    ggplot2::scale_size_manual(values = plot_size) +
    ggplot2::theme_classic()

  # function to modify ggpairs output (trim empty space)
  gpairs_lower <- function(g) {
    g$plots <- g$plots[-(1:g$nrow)]
    g$yAxisLabels <- g$yAxisLabels[-1]
    g$nrow <- g$nrow - 1

    g$plots <- g$plots[-(seq(g$ncol, length(g$plots), by = g$ncol))]
    g$xAxisLabels <- g$xAxisLabels[-g$ncol]
    g$ncol <- g$ncol - 1

    g
  }

  # modification of the plot
  full_plot <- gpairs_lower(full_plot)

  return(full_plot)
}

#' @describeIn plot.param_space Alias of `plot.param_space()` for backwards-compatibility.
#' @export
plot_param_space <- function(x) {
  lifecycle::deprecate_soft("1.5.0", "cvasi::plot_param_space()", "plot()")
  plot(x=x)
}

#' Get scenario parameters
#'
#' For scenario [sequences][sequence], only the parameters of the first scenario
#' in the sequence is returned. To access parameters of a specific scenario in
#' the sequence, use `get_param()` on the individual scenario object.
#'
#' @param x object to fetch parameters from
#' @return (list of) list(s) with key-value pairs
#' @aliases get_param,ANY-method
#' @seealso [set_param()]
#' @export
#' @examples
#' minnow_it %>% get_param()
setGeneric("get_param", function(x) standardGeneric("get_param"))
#' @export
setMethod("get_param", "ANY", function(x) stop("Object type is not supported"))
#' @export
#' @describeIn get_param Returns a list of parameter lists (if applicable)
setMethod("get_param", "list", function(x) lapply(x, get_param))
#' @export
#' @describeIn get_param Returns a list parameters for a single [scenario]
setMethod("get_param", "EffectScenario", function(x) return(x@param))
#' @export
#' @describeIn get_param Returns a list of parameter lists, one for each scenario in the sequence
setMethod("get_param", "ScenarioSequence", function(x) return(get_param(x[[1]])))
#' @export
#' @describeIn get_param Returns a list of parameters for a single [parameter_set]
setMethod("get_param", "ParameterSet", function(x) return(x@param))

# Returns the required parameters of a scenario or scenario-like object
setGeneric("get_rparam", function(x) standardGeneric("get_rparam"))
setMethod("get_rparam", "ANY", function(x) stop("Object type is not supported"))
setMethod("get_rparam", "list", function(x) lapply(x, get_rparam))
setMethod("get_rparam", "EffectScenario", function(x) return(x@param.req))
setMethod("get_rparam", "ScenarioSequence", function(x) return(get_rparam(x[[1]])))

#
# load results from the original Daphnia magna model
#
tfs_result <- data.frame()
for(part in c("damage","length","repro","surv")) {
  state <- toupper(substring(part, 1, 1))
  fn <- test_path(paste0("../data/DEBtox_Daphnia/test_",part,".txt"))
  df <- read.table(fn, stringsAsFactors = FALSE)
  names(df) <- c("time","set1_control","set1_control2",paste0("set1_VAR1_",1:3),
                 "set2_control","set2_control2",paste0("set2_VAR2_",1:3),
                 "set3_control","set3_control2",paste0("set3_CST_",1:5))
  df$var <- state

  df %>%
    tidyr::pivot_longer(!c(time,var), names_to="scenario") %>%
    dplyr::mutate(set=substring(scenario,1,4), run=substring(scenario,6)) %>%
    dplyr::select(time,var,set,run,value) %>%
    dplyr::filter(run != "control2") %>%
    dplyr::bind_rows(tfs_result) -> tfs_result

}
rm(part,fn,state,df)

tfs_result %>%
  tidyr::pivot_wider(names_from=var, values_from=value) %>%
  dplyr::arrange(set,run,time) %>%
  dplyr::select(time,set,run,D,L,R,S) -> tfs_result

#
# load original exposure series
#
tfs_exposure <- list()

exp <- read.table(test_path("../data/DEBtox_Daphnia/Data_exposure_profile_Dmagna_TFS_Cst.txt"), stringsAsFactors=FALSE, header=FALSE)
names(exp) <- c("time","control", paste0("exp.",seq(length(exp)-2)))
# fix constant exposure series
if(nrow(exp)==2) {
  exp[2,-1] <- exp[1,-1]
}
tfs_exposure$cst <- exp


# variable exposure series No. 1
exp <- read.table(test_path("../data/DEBtox_Daphnia/Data_exposure_profile_Dmagna_TFS_Var1.txt"), stringsAsFactors=FALSE, header=FALSE)
names(exp) <- c("time","control", paste0("exp.",seq(length(exp)-2)))
tfs_exposure$var1 <- exp

# variable exposure series No. 2
exp <- read.table(test_path("../data/DEBtox_Daphnia/Data_exposure_profile_Dmagna_TFS_Var2.txt"), stringsAsFactors=FALSE, header=FALSE)
names(exp) <- c("time","control", paste0("exp.",seq(length(exp)-2)))
tfs_exposure$var2 <- exp
rm(exp)

#
# model parameters
#
tfs_param <- list(
  L0 = 0.9676,
  Lp = 2.42,
  Lm = 4.614,
  rB = 0.1106,
  Rm = 16.69,
  f = 1,
  hb = 0.001608,
  a = 1, # disable new parameter
  Lf = 0,
  Lj = 0,
  Tlag = 0,
  kd = 2.060,
  zb = 5.201,
  bb = 0.04422,
  zs = 11.75,
  bs = 0.1759,
  FBV    = 0.02,
  KRV    = 1,
  kap    = 0.8,
  yP     = 0.8*0.8,
  Lm_ref = 4,
  len    = 2,
  Tbp    = 3,
  MoA = 1,
  FB = 0
)

#
# initial state
#
tfs_init <- c(D=0, L=tfs_param$L0, R=0, S=1)

#
# Originally, a DEBtox model as implemented in BYOM, DEBtox2019 module, version 4.5,
# last modified 27 Nov 2021. Source: https://www.debtox.info/byom.html
# However, the model equations are almost identical to DEBtox 2019 version 4.7
# which is available here in the package. Therefore, the older version is no longer
# implemented explicitly.
#
# See `DEBtox()` for the newer model version.
#

########################
## Class definition
########################

#' @include model-debtox.R solver.R fx.R
#' @export
setClass("DebDaphnia", contains="DebTox")

########################
## Constructor
########################

#' @describeIn DEBtox Deprecated model variant of `DEBtox()`
#' @export
DEB_Daphnia <- function() {
  lifecycle::deprecate_soft("1.2.0", "DEB_Daphnia()", "DEBtox()",
                            details="The model constructor has been renamed to `DEBtox()`")

    new("DebDaphnia",
        name="DEBtox_Daphnia",
        param.req=c("L0", "Lp", "Lm", "rB", "Rm", "f", "hb", "Lf", "Lj", "Tlag",
                    "kd", "zb", "bb", "zs", "bs", "FBV", "KRV", "kap", "yP",
                    "Lm_ref", "len", "Tbp", "MoA", "FB"),
        endpoints=c("L","R","S"),
        control.req=TRUE,
        init=c("D"=0,"L"=0,"R"=0,"S"=1)
    ) %>% set_noexposure()
}

########################
## Simulation
########################

# For backwards-compatibility
solver_debtox_daphnia <- function(scenario, ...) {
  # only effective difference between model versions is the introduction of
  # parameter `a` and the limit of hazard rate `h` to 99% per hour in the newer
  # DEBtox version.

  # Setting the default value of a=1 disables the Weibull background hazard
  scenario <- set_param(scenario, c("a"=1))
  # Continue as a DEBtox scenario
  solver_debtox(scenario, ...)
}

#' @describeIn solver (deprecated) Numerically integrates *DEBtox_Daphnia* scenarios
setMethod("solver", "DebDaphnia", solver_debtox_daphnia)

########################
## Effects
########################

# uses the default fx implementation, nothing more required

#' Set boundaries of model parameters
#'
#' Modifies the boundaries of model parameters for one or more [scenario] or
#' [caliset] objects.
#'
#'
#' @param x vector of [scenario] or [caliset] objects
#' @param bounds named list of numerical vectors, where the first level lists the parameters
#' by name, and the second level lists the lower and upper boundary
#' @return [scenario] or [caliset] with modified parameter boundaries
#' @export
#' @include class-EffectScenario.R
#' @examples
#' metsulfuron %>%
#'    set_bounds(list(k_phot_max = c(0, 30),
#'                    k_resp = c(0, 10)))
#'
setGeneric("set_bounds", function(x, bounds) standardGeneric("set_bounds"))

#' @rdname set_bounds
#' @export
setMethod("set_bounds", c("EffectScenario","list"), function(x, bounds) {
  check_bounds(bounds)

  # if required parameters are set, check if arguments conform
  if(length(x@param.req) > 0) {
    unused <- setdiff(names(bounds), x@param.req)
    if(length(unused)>0)
      cli::cli_warn("unknown parameter{?s}, ignoring boundaries for '{unused}'")
    # limit bounds to known/required parameters
    ints <- intersect(x@param.req, names(bounds))
    bounds <- bounds[ints]
  }

  # convert lists to vectors
  for(nm in names(bounds)) {
    if(is.list(bounds[[nm]]))
      bounds[[nm]] <- unlist(bounds[[nm]])
  }

  x@param.bounds[names(bounds)] <- bounds
  x
})

# CalibrationSet, single parameter_set
#' @rdname set_bounds
#' @export
setMethod("set_bounds", c("CalibrationSet","list"), function(x, bounds) {
  check_bounds(bounds)
  x@scenario <- set_bounds(x@scenario, bounds)
  x
})

# multiple scenarios, single parameter_set
#' @rdname set_bounds
#' @export
setMethod("set_bounds", c("list","list"), function(x, bounds) {
  check_bounds(bounds)
  sapply(x, set_bounds, bounds)
})

#' @rdname set_bounds
#' @export
setMethod("set_bounds", c("ScenarioSequence","list"), function(x, bounds) {
  check_bounds(bounds)
  for(i in seq_along(x)) {
    x[[i]] <- set_bounds(x[[i]], bounds)
  }
  x
})

get_bounds <- function(x) {
  if(is.vector(x) & length(x) != 1)
    stop("Argument `x` must be of length one")
  if(is_sequence(x))
    return(get_bounds(x[[1]]))
  if(is_caliset(x))
    return(get_bounds(x@scenario))
  if(!is_scenario(x))
    stop("Argument `x` must be a scenario")

  x@param.bounds
}

check_bounds <- function(bounds) {
  ## check names
  if(!is.list(bounds))
    stop("Argument `bounds` must be a list")
  if(length(bounds) == 0)
    stop("Argument `bounds` is empty")
  nms <- names(bounds)
  if(is.null(nms))
    stop("Argument `bounds` must be a named list")
  if("" %in% nms)
    stop("Argument `bounds` contains unnamed elements")

  ## check values
  errs <- FALSE
  # check that each name has a numeric vector of length two
  for(nm in nms) {
    if(length(bounds[[nm]]) != 2) {
      cli::cli_warn(paste0("boundaries for '", nm, "' are not of length two"))
      errs <- TRUE
    }
    if(is.list(bounds[[nm]]))
      bounds[[nm]] <- unlist(bounds[[nm]])
    if(!is.numeric(bounds[[nm]])) {
      cli::cli_warn(paste0("boundaries for '", nm, "' must be numeric"))
      errs <- TRUE
    }
    if(any(is.na(bounds[[nm]]))) {
      cli::cli_warn(paste0("boundaries for '", nm, "' contain NA values"))
      errs <- TRUE
    }
    if(min(bounds[[nm]]) != bounds[[nm]][[1]]) {
      cli::cli_warn(paste0("boundaries for '", nm, "' have invalid order"))
      errs <- TRUE
    }
  }
  if(errs)
    stop("Boundaries contain invalid values")
}

test_that("scenario", {
  # single boundary
  bds <- list(kd=c(2, 3))
  sc <- new("EffectScenario") %>% set_bounds(bds)
  expect_equal(get_bounds(sc), bds)

  # update boundary
  bds <- list(kd=c(23, 24))
  sc <- sc %>% set_bounds(bds)
  expect_equal(get_bounds(sc), bds)

  # multiple boundaries
  bds <- list(kd=c(0, 1), hb=c(2, 3))
  sc <- new("EffectScenario") %>% set_bounds(bds)
  expect_equal(get_bounds(sc), bds)
})

test_that("sequence", {
  bds <- list(kd=c(23, 24))
  sc <- new("EffectScenario") %>% set_times(0:5)
  suppressWarnings(sq <- sequence(list(sc)))
  sq <- set_bounds(sq, bds)
  expect_equal(get_bounds(sq), bds)
})

test_that("caliset", {
  bds <- list(kd=c(23, 24))
  cs <- caliset(new("EffectScenario"), data.frame(t=0, a=1))
  cs <- set_bounds(cs, bds)
  expect_equal(get_bounds(cs), bds)
})

test_that("list of scenarios", {
  bds <- list(kd=c(2, 3))
  scs <- list(GUTS_RED_IT(), GUTS_RED_IT()) %>%
    set_bounds(bds)

  expect_equal(get_bounds(scs[[1]]), bds)
  expect_equal(get_bounds(scs[[2]]), bds)
})

test_that("invalid arguments", {
  sc <- GUTS_RED_IT() %>% set_bounds(list(kd=c(2, 3)))
  # unknown parameter
  expect_warning(set_bounds(sc, list(foobar=c(1,2))))
  # not a list
  expect_error(check_bounds("a")) # check function explicitly
  # no elements
  expect_error(set_bounds(sc, list()))
  # no names
  expect_error(set_bounds(sc, list(1)))
  # unnamed elements
  expect_error(set_bounds(sc, list(a=c(1,2), "b")))
  # values not of length 2
  expect_warning(expect_error(set_bounds(sc, list(a=c(1)))))
  # values not numeric
  expect_warning(expect_error(set_bounds(sc, list(a=c("a","b")))))
  # values contain NA
  expect_warning(expect_error(set_bounds(sc, list(a=c(1, NA_real_)))))
  # wrong order
  expect_warning(expect_error(set_bounds(sc, list(kd=c(3, 2)))))
})

#' Set transfer events
#'
#' A *transfer* refers to an event where a certain amount of biomass
#' is moved to a new medium after a period of time. Effectively, this resets
#' the scenario's state variable representing biomass and re-scales all state
#' variables which are correlated with biomass, such as adsorbed chemical mass.
#' This feature replicates a procedure occurring e.g. in *Lemna* effect studies
#' and may be necessary to recreate study results.
#'
#' If a transfer occurs, simulation results of that time point will report the model state
#' **before** the transfer. Be aware that if transfers are defined using the
#' `interval` argument, the transfers will always occur relative to time point
#' zero (`t = 0`). As an example, setting a regular transfer of seven days,
#' `interval =  7`, will result at transfers occurring at time points which are
#' integer multiplicates of seven, such as `t=0`, `t=7`, `t=14` and so forth.
#' The starting and end times of a scenario do not influece **when** a regular
#' transfer occurs, only **if** it occurs.
#'
#' ### Transferred biomass
#'
#' At each transfer, a defined amount of biomass is transferred to a new medium.
#' This is modeled by interrupting the simulation at a transfer time point, modifying
#' the biomass level `BM`, and scaling affected compartments according to new
#' biomass levels. Scaling of compartments depending on biomass, such as
#' *internal toxicant mass*, is necessary to correctly reflect mass balances and
#' concentrations over time.
#'
#' Transferred biomass is set using the `biomass` parameter. Is is either a
#' single numerical value in which case the same biomass level is set at each
#' transfer. Or it is a vector of numerical values with the same length as the
#' `times` parameter in which case a custom biomass level can be set for each
#' transfer. Multiple biomass levels can only be set in conjunction with
#' custom transfer time points.
#'
#' Some scenario types define default values for transferred biomass based
#' on common study set ups.
#'
#' ### Regular and custom transfer time points
#'
#' Transfers can occur either in regular intervals of time or at selected, custom
#' time points. For regular intervals, the parameter `interval` is set to a single
#' numeric value which has the same unit as the scenario's time dimension. As an
#' example: if a scenario uses the unit of *days* for time, the transfer interval
#' is also specified in *days*:
#'
#' Transfers occurring at custom time points are set by passing a numerical vector
#' to the parameter `times`. The time points' units must match with the unit of
#' time in the scenario. A custom transfer time point **must not occur at the
#' starting time point of a simulation**.
#'
#' ### Affected compartments
#'
#' Some compartments depend on biomass to correctly reflect mass balances and
#' concentrations over time, such as *internal toxicant mass*. These compartments
#' need to be scaled linearly to reflect the change in biomass levels.
#' The parameter `scaled_comp` accepts a character vector of compartment names
#' which are scaled at each transfer. This parameter should only be used with
#' custom, user-defined models. If no compartment needs to be scaled, set or
#' use the default value of `character(0)`.
#'
#'
#' @param x vector of `EffectScenario` objects
#' @param times optional `numeric` vector of time points where transfers occur
#' @param interval optional `numeric`, interval in time units of the scenario,
#'  set to `-1` to disable transfers.
#' @param biomass optional `numeric` vector, amount of biomass that is being
#'    transferred at each transfer
#' @param scaled_comp optional `character` vector of affected compartments
#'    that are scaled according to new biomass levels
#' @seealso [Lemna-models]
#' @return Modified [scenario] objects
#' @export
#' @include class-EffectScenario.R
#' @rdname set_transfer
#' @examples
#' # Simulate biomass transfer of 50 *g/m²* at a regular interval of 7 *days*
#' metsulfuron %>%
#'   set_transfer(interval=7, biomass=50) %>%
#'   simulate()
#'
#' # Simulate irregular biomass transfers occuring at days 5, 10, and 12
#' metsulfuron %>%
#'   set_transfer(times=c(5, 10, 12), biomass=50) %>%
#'   simulate()
#'
#' # Simulate irregular transfers with changing amounts of transferred biomass
#' metsulfuron %>%
#'   set_transfer(times=c(5, 10, 12), biomass=c(50, 20, 10)) %>%
#'   simulate()
#'
#' # Disable all biomass transfers
#' metsulfuron %>%
#'   set_notransfer() %>%
#'   simulate()
setGeneric("set_transfer",
           function(x, interval, times, biomass, scaled_comp) standardGeneric("set_transfer"),
           signature = "x"
)

#' @rdname set_transfer
setMethod("set_transfer", "ANY", function(x, interval, times, biomass, scaled_comp) {
  stop("type/scenario does not support transfers")
})

#' @rdname set_transfer
setMethod("set_transfer","Transferable", function(x, interval, times, biomass, scaled_comp) set_transfer2(x, interval, times, biomass, scaled_comp) )

# Function body was moved here from generic in order to be able to debug in
# RStudio. Otherwise, breakpoints would not be hit.
set_transfer2 <- function(x, interval, times, biomass, scaled_comp) {
  if(!missing(interval)) {
    if(!is.numeric(interval))
      stop("interval must be numeric")
    if(is.nan(interval) | interval <= 0)
      interval <- -1
  }
  if(!missing(times)) {
    if(length(times) == 0) # no date == empty vector
      times <- numeric(0)
    else {
      times <- unlist(times)
      if(any(is.na(times) | is.infinite(times) | !is.numeric(times)))
        stop("argument `times` contains invalid values")
      times <- sort(unique(times))
    }
  }
  if(!missing(interval) | !missing(times)) {
    if(missing(interval))
      interval <- -1
    if(missing(times))
      times <- numeric(0)
    # check  if settings are consistent
    if(interval > 0 & length(times) > 0)
      stop("arguments `interval` and `times` cannot be used at the same time")

    x@transfer.interval <- interval
    x@transfer.times <- times
  }

  if(!missing(biomass)) {
    biomass <- unlist(biomass)
    if(!is.numeric(biomass))
      stop("biomass has invalid value")
    if(any(is.na(biomass) | is.nan(biomass)) | any(biomass <= 0))
      stop("biomass has invalid value")
    if(length(biomass) > 1 & length(x@transfer.times) != length(biomass))
      stop("length of biomass and transfer times vectors do not match")
    x@transfer.biomass <- biomass
  }
  if(!missing(scaled_comp)) {
    if(length(scaled_comp) == 0)
      stop("list of scaled compartments must not be empty")
    if(length(setdiff(unname(scaled_comp), names(x@init))) > 0)
      stop("scaled compartments must be state variables")
    x@transfer.comp.scaled <- scaled_comp
  }
  if(any(is.na(x@transfer.comp.biomass))) {
    stop("biomass state variable `transfer.comp.scaled` not set yet")
  }
  if(length(x@transfer.comp.biomass) != 1) {
    stop("biomass state variable `transfer.comp.scaled` must have length 1")
  }
  x
}

#' @export
#' @describeIn set_transfer Disable biomass transfers
set_notransfer <- function(x) {
  if(length(x) > 1)
    return(sapply(x, set_notransfer))

  set_transfer(x, interval=-1)
}

#' Get scenario tag
#'
#' Returns the user-defined, custom tag of a scenario, if available.
#' Tags can be helpful to quickly distinguish scenarios by e.g. a user-specified
#' string. The function supports vectorized inputs. If more
#' than one scenario is supplied in argument `x`, then a list of tags is
#' returned.
#'
#' @param x (vector of) [scenarios] or [parameter_set] objects
#'
#' @return (list of) tag(s), returns `NA` if no tag was set
#' @aliases get_tag,ANY-method
#' @seealso [set_tag()]
#' @export
#'
#' @examples
#' # returns `fathead minnow`
#' get_tag(minnow_it)
#'
#' # update or set a tag
#' myscenario <- GUTS_RED_IT() %>% set_tag("My Custom Tag")
#' # returns `My Custom Tag`
#' get_tag(myscenario)
setGeneric("get_tag", function(x) standardGeneric("get_tag"))
#' @export
setMethod("get_tag", "ANY", function(x) stop("Object type is not supported"))
#' @export
#' @describeIn get_tag Returns a list of tags (if applicable)
setMethod("get_tag", "list", function(x) lapply(x, get_tag))
#' @export
#' @describeIn get_tag Returns the tag of a single [scenario]
setMethod("get_tag", "EffectScenario", function(x) return(x@tag))
#' @export
#' @describeIn get_tag Returns a list of tags, one for each scenario in the sequence
setMethod("get_tag", "ScenarioSequence", function(x) return(get_tag(x[[1]])))
#' @export
#' @describeIn get_tag Returns the tag of a single [parameter_set]
setMethod("get_tag", "ParameterSet", function(x) return(x@tag))

#' Calibration set
#'
#' A *calibration set* combines a [scenario], observed data, and an
#' optional weighting factor into one object. The *calibration set* is used to fit
#' model parameters to observed data using [calibrate()].
#'
#' A *calibration set* usually represents a single experiment or trial.
#' Multiple experimental replicates can be combined into a single *set*, if model
#' parameters are identical between trials.
#' If model parameters were modified during a trial, e.g. a pump failure occurred
#' or flow rates changed, this can be represented by using a *scenario sequence*
#' instead of a basic [scenario]. Please refer to [sequence()] for details.
#'
#' ### Weighting
#' An optional weighting factor can be used to scale the error term of a
#' whole *set* or of individual data points when fitting parameters using e.g.
#' [calibrate()].
#'
#' The vector of weights must either be of length one or have the same length
#' as the dataset. In the former case, the same weight will be applied to all
#' values in the dataset. In the latter, individual weights are applied
#' for each data point.
#'
#' @usage caliset(scenario, data, weight = 1.0, tag = NULL)
#'
#' @param scenario a [scenario] describing conditions during the experiment
#' @param data a `data.frame` with observed data in long format containing two
#'   columns: the 1st column with `numeric` time points and 2nd column with
#'   `numeric` data to fit to. Rows with observed `NA` values will be removed.
#' @param weight optional `numeric` weight to be applied when calculating the
#'  error term for each data point. Default value is `1.0`, i.e. no weighting.
#' @param tag optional value to identify the data, e.g. a study number
#'
#' @returns `caliset()` returns a *calibration set* object
#' @export
#'
#' @examples
#' library(dplyr)
#'
#' # Get observed biomass during control experiment by Schmitt et al. (2013)
#' observed <- schmitt2013 %>%
#'   filter(trial == "T0") %>%
#'   select(time, obs)
#'
#' # Create a scenario that represents conditions during experiment
#' scenario <- metsulfuron %>%
#'   set_param(c(k_phot_fix=TRUE, k_resp=0, Emax=1)) %>%
#'   set_init(c(BM=12)) %>%
#'   set_noexposure() %>%
#'   set_bounds(list(k_phot_max=c(0, 0.5)))
#'
#' # Create a calibration set
#' cs <- caliset(scenario, observed)
#'
#' # Fit parameter 'k_phot_max' to observed biomass growth from experiment
#' calibrate(
#'   cs,
#'   par=c(k_phot_max=1),
#'   output="BM",
#'   method="Brent" # Brent is recommended for one-dimensional optimization
#' ) -> fit
#' fit$par
#'
#' @name CalibrationSet
#' @aliases CalibrationSet-class caliset CalibrationSet
NULL

#' @export
#' @aliases CalibrationSet
setClass("CalibrationSet",
  slots=list(
   scenario="ANY",
   data="data.frame",
   weight="numeric",
   tag="ANY"
  )
)

# Creates a *calibration set*
#' @export
caliset <- function(scenario, data, weight=1, tag=NULL) {
  #
  # check scenario argument
  #
  if(missing(scenario))
    stop("Argument `scenario` is missing")
  if(is.vector(scenario) | any(!is_scenario(scenario) & !is_sequence(scenario)))
    stop("Argument `scenario`  must be a scenario or sequence object")

  #
  # check data
  #
  if(missing(data))
    stop("Argument `data` is missing")
  if(!is.data.frame(data))
    stop("Argument `data` must be a data.frame")
  if(length(data) < 2)
    stop("Argument `data` must have at least two columns")
  # get rid of tibbles and other tabular data structures which differ from
  # data.frame in subtle ways when selecting subsets with base R
  data <- as.data.frame(data)
  if(nrow(data) == 0)
    stop("Argument `data` must not be empty")

  # check that first column, time, is numerical and in ascending order
  colnames <- names(data)
  if(any(is.na(data[, 1]) | is.nan(data[, 1]) | is.infinite(data[, 1])))
    stop("First data column (", colnames[1], ") contains invalid values such as Inf, NaN, or NA")
  if(any(!is.numeric(data[, 1])))
    stop("First data column (", colnames[1], ") must be numerical")
  if(any(diff(data[, 1]) < 0))
    stop("First data column (", colnames[1], ") must be in ascending order")

  # check that second column is numerical, but NAs are allowed for now
  if(any(is.nan(data[, 2]) | is.infinite(data[, 2])))
    stop("Second data column (", colnames[2], ") contains invalid values such as Inf or NaN")
  if(any(!is.numeric(data[, 2])))
    stop("Second data column (", colnames[2], " must be numerical")

  # remove NAs in observed column, if present
  if(any(is.na(data[, 2]))) {
    message("Excluding NAs from observed values ...")
    data <- data[!is.na(data[, 2]), ]
  }

  #
  # check weight
  #
  if(any(is.infinite(weight) | is.na(weight) | is.nan(weight) | !is.numeric(weight)))
    stop("Argument `weight` must only contain finite numeric values")
  # apply the same weight to all data points in dataset
  if(length(weight) == 1) {
    weight <- rep(weight, times=nrow(data))
  }
  # else: length of weight vector must match the dataset
  if(length(weight) != nrow(data)) {
    stop("Argument `weight` must have the same length as `data`")
  }

  ### create object
  new("CalibrationSet",
      scenario=scenario,
      data=data,
      weight=weight,
      tag=tag
      )
}

# Outdated alias for `caliset()`
#' @export
CalibrationSet <- function(...) {
  lifecycle::deprecate_soft("1.2.0", "CalibrationSet()", "caliset()")
  caliset(...)
}

########################
## Organizing man pages
########################

#' Algae models
#'
#' Overview of supported *Algae* models
#'
#' - [Algae_Weber()] by Weber *et al.* (2012)
#' - [Algae_TKTD()] based on Weber *et al.* (2012), but with scaled damage
#' - [Algae_Simple()] simplified growth model without additional forcing variables
#'
#' @inheritSection Transferable Biomass transfer
#'
#' @references
#' Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
#' and Ratte HT, 2012. Combination of a higher-tier flow-through system and
#' population modeling to assess the effects of time-variable exposure of
#' isoproturon on the green algae Desmodesmus subspictatus and
#' Pseudokirchneriella subcapitata. Environmental Toxicology and
#' Chemistry, 31(4), 899-908. \doi{10.1002/etc.1765}
#'
#' EFSA Panel on Plant Protection Products and their Residues, 2018. Scientific
#' opinion on the state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect
#' models for regulatory risk assessment of pesticides for aquatic organisms.
#' EFSA journal 16:5377 \doi{10.2903/j.efsa.2018.5377}
#'
#' @name Algae-models
#' @family algae models
#' @family models
#' @seealso [Lemna-models], [Transferable]
#' @include class-Transferable.R
#' @aliases Algae-class
NULL


########################
## Class definitions
########################

# Generic Algae class
#' @export
setClass("Algae", contains = c("Transferable", "EffectScenario"))

# Algae model with exponential growth and forcings (P, I)
#' @export
setClass("AlgaeWeber", contains = "Algae")
# for backwards compatibility
#' @export
setClass("AlgaeWeberScenario", contains = "AlgaeWeber")

# Algae model with exponential growth and scaled internal damage
#' @export
setClass("AlgaeTKTD", contains = "Algae")
# for backwards compatibility
#' @export
setClass("AlgaeTKTDScenario", contains = "AlgaeTKTD")

# Algae model with exponential growth without forcings
#' @export
setClass("AlgaeSimple", contains = "Algae")
# for backwards compatibility
#' @export
setClass("AlgaeSimpleScenario", contains = "AlgaeSimple")


########################
## Constructor
########################

#' Algae model, *SAM-X* (Weber et al. 2012)
#'
#' The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
#' growth model for algae. The model simulates the development of algal biomass
#' under laboratory and environmental conditions and was developed by
#' Weber et al. (2012) as cited in EFSA TKTD opinion (2018). The growth of the
#' algae population is simulated on the basis of growth rates, which are
#' dependent on environmental conditions (radiation, temperature, and phosphorus).
#' The toxicodynamic sub-model describes the effects of growth-inhibiting
#' substances through a corresponding reduction in the photosynthesis rate on
#' the basis of internal concentrations.
#'
#' Deviating from the equations described by Weber et al., this model implementation
#' uses a user-defined time-series to represent (environmental) concentrations.
#' Therefore, state-variable `C` and its differential equation was removed and
#' model output `C` is identical to the exposure time-series. The implementation
#' of Weber et al. (2012) was followed where units differ from EFSA (2018).
#'
#' @section State variables:
#' The model has three state variables:
#' - `A`, Biomass (µg fresh wt/mL, cells/mL *10^4)
#' - `Q`, Mass of phosphorous internal (mg P/L, or µg P/mL)
#' - `P`, Mass of phosphorous external (mg P/L, or µg P/mL)
#'
#' The original model by Weber et al. contains an additional state variable
#' `C` which models the external stressor concentration. However, the model
#' implementation in this packages uses a user-defined time-series to represent
#' environmental concentrations.
#' Therefore, state variable `C` and accompanying parameters are not present here.
#'
#' @section Model parameters:
#' - Growth model
#'   - `mu_max`, Maximum growth rate (d-1)
#'   - `Q_min`, Minimum intracellular P (µg P/µg fresh wt)
#'   - `Q_max`, Maximum intracellular P (µg P/µg fresh wt)
#'   - `v_max`, Maximum P-uptake rate at non-limited growth (µg P/µg fresh wt/d)
#'   - `k_s`,   Half-saturation constant for extracellular P (mg P/L)
#'   - `m_max`, Natural mortality rate (1/d)
#'   - `I_opt`, Optimum light intensity for growth (uE/m²/s)
#'   - `T_opt`, Optimum temperature for growth (°C)
#'   - `T_max`, Maximum temperature for growth (°C)
#'   - `T_min`, Minimum temperature for growth (°C)
#'   - `D`, Dilution rate (1/d)
#'   - `R_0`, Influx concentration of P (mg P/L)
#'
#' - Concentration response (Toxicodynamics)
#'   - `EC_50`, Effect concentration of 50% inhibition of growth rate (µg/L)
#'   - `b`, slope of concentration effect curve at EC_50 (-)
#'
#' @section Forcings:
#' The Weber model requires two environmental properties as time-series input:
#'  - `T_act`, temperature (°C), and
#'  - `I`, irradiance (uE/m²/s).
#'
#' The following constant default values are used for these properties:
#' - `T_act` = 23 °C
#' - `I` = 100 uE/m²/s
#'
#' Forcings time-series are represented by `data.frame` objects consisting of two
#' columns. The first for time and the second for the environmental factor in question.
#'
#' Entries of the `data.frame` need to be ordered chronologically. A time-series
#' can consist of only a single row; in this case it will represent constant
#' environmental conditions. See [scenarios] for more details.
#'
#' @section Simulation output:
#' Simulation results will contain the state variables Biomass (`A`), mass of
#' internal phosphorous (`Q`), mass of external phosphorous (`P`) and the external
#' concentration (`C`).
#'
#' It is possible to amend the output of [simulate()] with additional model
#' quantities that are not state variables, for e.g. debugging purposes or to
#' analyze model behavior. To enable or disable additional outputs, use the
#' optional argument `nout` of [simulate()]. As an example, set `nout=2` to
#' enable reporting of external concentration and model derivative `dA`.
#' Set `nout=0` to disable additional outputs. The default is `nout=1`.
#'
#' The available output levels are as follows:
#'
#'  - `nout >= 1`: `C`, external concentration (µg/L)
#'  - `nout >= 2`: `f(T)`, temperature dependence (-)
#'  - `nout >= 3`: `f(I)`, light dependence (-)
#'  - `nout >= 4`: `f(Q)`, nutrient dependence (-)
#'  - `nout >= 5`: `f(Q, P)`, uptake flow reduction (-)
#'  - `nout >= 6`: `f(C)`, effect of chemical stressor (-)
#'  - `nout >= 7`: `dA`, biomass derivative (µg)
#'  - `nout >= 8`: `dQ`, internal phosphorous derivative (mg P/µg fresh wt)
#'  - `nout >= 9`: `dP`, external phosphorous derivative (mg P L-1)
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `hmax = 0.1`<br>
#'    Maximum step length in time suitable for most simulations.
#'
#' @section Parameter boundaries:
#' Default values for parameter boundaries are set for all parameters by expert
#' judgement, for calibration purposes. Values can be access from the object, and
#' defaults overwritten.
#'
#' @section Model history and changes:
#' - cvasi v1.5.0
#'   - Unused state variable `C` and parameter `k` removed from documentation
#'     and code. External concentration `C` added to simulation output by means
#'     of optional output level `nout=1`.
#'
#' @references
#' Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
#' and Ratte HT, 2012. Combination of a higher-tier flow-through system and
#' population modeling to assess the effects of time-variable exposure of
#' isoproturon on the green algae Desmodesmus subspictatus and
#' Pseudokirchneriella subcapitata. Environmental Toxicology and
#' Chemistry, 31(4), 899-908. \doi{10.1002/etc.1765}
#'
#' EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
#' Ockleford C, Adriaanse P, Berny P, Brock T, Duquesne S, Grilli S,
#' Hernandez-Jerez AF, Bennekou SH,Klein M, Kuhl T, Laskowski R, Machera K,
#' Pelkonen O, Pieper S, Smith RH, Stemmer M, Sundh I, Tiktak A,Topping CJ,
#' Wolterink G, Cedergreen N, Charles S, Focks A, Reed M, Arena M, Ippolito A,
#' Byers H andTeodorovic I, 2018. Scientific Opinion on the state of the art of
#' Toxicokinetic/Toxicodynamic (TKTD)effect models for regulatory risk assessment
#' of pesticides for aquatic organisms. EFSA Journal, 16(8), 5377.
#' \doi{10.2903/j.efsa.2018.5377}
#'
#' @return an S4 object of type [AlgaeWeber-class]
#' @seealso [Scenarios], [Transferable]
#' @family algae models
#' @export
#' @aliases SAM-X AlgaeWeber-class AlgaeWeberScenario-class
Algae_Weber <- function() {
  new("AlgaeWeber",
      name = "Algae_Weber",
      param.req = c("mu_max", "m_max", "v_max", "k_s", "Q_min", "Q_max", "R_0", "D",
                    "T_opt", "T_min", "T_max", "I_opt", "EC_50", "b"
      ),
      # default values as defined by Weber et al. (2012)
      param = list(mu_max = 1.7380, m_max = 0.0500, v_max = 0.0520, k_s = 0.0680,
                   Q_min = 0.0011, Q_max = 0.0144, R_0 = 0.36, D = 0.5,
                   T_opt = 27, T_min = 0, T_max = 35, I_opt = 120
      ),
      # boundary presets defined by expert judgement
      param.bounds = list(mu_max=c(0, 4),  EC_50=c(0, 1e6), b=c(0.1, 20)),
      endpoints = c("A", "r"),
      forcings.req=c("T_act", "I"),
      forcings=list("T_act"=data.frame(time=0, temp=23),
                    "I"=data.frame(time=0, irr=100)),
      control.req = TRUE,
      init = c(A = 1, Q = 0.01, P = 0.18),
      transfer.interval = -1,
      transfer.biomass = 1,
      transfer.comp.biomass = "A",
      transfer.comp.scaled = "Q"
  ) %>% set_noexposure()
}

#' @export
#' @describeIn Algae_Weber Alias using original model name.
SamX <- Algae_Weber

#' Algae model variant (Weber et al. 2012) with scaled damage
#'
#' The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
#' growth model for algae. The model simulates the development of algal biomass
#' under laboratory and environmental conditions. The growth of the algae
#' population is simulated on the basis of growth rates, which are dependent on
#' environmental conditions (radiation, temperature and phosphorus).
#' The model is a variant of the [Algae_Weber()] model (Weber 2012) as cited
#' in EFSA TKTD opinion (2018). This Algae model, [Algae_TKTD()], provides an
#' additional possibility (probit) to simulate the dose-response curve and
#' considers a scaled internal damage instead of the external concentration.
#'
#' @section State variables:
#' The model has four state variables:
#' - `A`, Biomass (µg fresh wt/mL, cells/mL *10^4)
#' - `Q`, Mass of phosphorous internal (µg P/µg fresh wt)
#' - `P`, Mass of phosphorous external (µg P/L)
#' - `Dw`, Damage concentration (µg/L)
#'
#' @section Model parameters:
#' - Growth model
#'   - `mu_max`, Maximum growth rate (d-1)
#'   - `Q_min`, Minimum intracellular P (µg P/µg fresh wt)
#'   - `Q_max`, Maximum intracellular P (µg P/µg fresh wt)
#'   - `v_max`, Maximum P-uptake rate at non-limited growth (µg P/µg fresh wt/d)
#'   - `k_s`,   Half-saturation constant for extracellular P (mg P/L)
#'   - `m_max`, Natural mortality rate (1/d)
#'   - `I_opt`, Optimum light intensity for growth (µE/m²/s)
#'   - `T_opt`, Optimum temperature for growth (°C)
#'   - `T_max`, Maximum temperature for growth (°C)
#'   - `T_min`, Minimum temperature for growth (°C)
#'   - `D`, Dilution rate (1/d)
#'   - `R_0`, Influx concentration of P (mg P/L)
#'
#' - Concentration response (Toxicodynamics)
#'   - `EC_50`, Effect concentration of 50% inhibition of growth rate (µg L-1)
#'   - `b`, slope of concentration effect curve at EC_50 (-)
#'   - `dose_resp`, shape of the dose response curve (0 = logit, 1 = probit)
#'
#' - External concentration (Toxicokinetics)
#'   - `kD`, dominant rate constant (d-1)
#'
#' @section Forcings:
#' The Weber model variant requires two environmental properties as time-series input:
#'  - `T_act`, temperature (°C), and
#'  - `I`, irradiance (uE/m²/s).
#'
#' The following constant default values are used for these properties:
#' - `T_act` = 23 °C
#' - `I` = 100 uE/m²/s
#'
#' Forcings time-series are represented by `data.frame` objects consisting of two
#' columns. The first for time and the second for the environmental factor in question.
#'
#' Entries of the `data.frame` need to be ordered chronologically. A time-series
#' can consist of only a single row; in this case it will represent constant
#' environmental conditions. See [scenarios] for more details.
#'
#' @section Simulation output:
#' Simulation results will contain the state variables Biomass (`A`), mass of
#' internal phosphorous (`Q`), mass of external phosphorous (`P`) and the damage
#' concentration (`Dw`).
#'
#' It is possible to amend the output of [simulate()] with additional model
#' quantities that are not state variables, for e.g. debugging purposes or to
#' analyze model behavior. To enable or disable additional outputs, use the
#' optional argument `nout` of [simulate()]. As an example, set `nout=2` to
#' enable reporting of model derivatives `dA` and `dQ`. Set `nout=0` to disable
#' additional outputs (default).
#'
#' The available output levels are as follows:
#'
#'  - `nout >= 1`: `C`, external concentration (µg/L)
#'  - `nout >= 2`: `f(T)`, temperature dependence (-)
#'  - `nout >= 3`: `f(I)`, light dependence (-)
#'  - `nout >= 4`: `f(Q)`, nutrient dependence (-)
#'  - `nout >= 5`: `f(Q, P)`, uptake flow reduction (-)
#'  - `nout >= 6`: `f(C)`, effect of chemical stressor (-)
#'  - `nout >= 7`: `dA`, biomass derivative (µg)
#'  - `nout >= 8`: `dQ`, internal phosphorous derivative (mg P/µg fresh wt)
#'  - `nout >= 9`: `dP`, external phosphorous derivative (mg P L-1)
#'  - `nout >= 10`: `dDw`, damage concentration derivative (µg L-1)
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `hmax = 0.1`<br>
#'    Maximum step length in time suitable for most simulations.
#'
#' @section Model history and changes:
#' - cvasi v1.5.0
#'   - Support for simulating flow-through conditions by introducing new parameters
#'     `D` and `R_0` and adapting the ODEs according to the [Algae_Weber] model.
#'   - ODE of external phosphorous concentration `P` corrected, which contained
#'     an erroneous growth term before.
#'   - Response functions added to optional simulation outputs, order of output
#'     levels modified.
#'
#' @references
#' Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
#' and Ratte HT, 2012. Combination of a higher-tier flow-through system and
#' population modeling to assess the effects of time-variable exposure of
#' isoproturon on the green algae Desmodesmus subspictatus and
#' Pseudokirchneriella subcapitata. Environmental Toxicology and
#' Chemistry, 31(4), 899-908. \doi{10.1002/etc.1765}
#'
#' @return an S4 object of type [AlgaeTKTD-class]
#' @seealso [Scenarios], [Transferable]
#' @family algae models
#' @export
#' @aliases AlgaeTKTD-class AlgaeTKTDScenario-class
Algae_TKTD <- function() {
  new("AlgaeTKTD",
      name = "Algae_TKTD",
      param.req = c("mu_max", "m_max", "v_max", "k_s", "Q_min", "Q_max", "R_0",
                    "D", "T_opt", "T_min", "T_max", "I_opt", "EC_50", "b", "kD",
                    "dose_resp"),
      # default values as defined by Weber et al. (2012)
      param = list(mu_max = 1.7380, m_max = 0.0500, v_max = 0.0520, k_s = 0.0680,
                   Q_min = 0.0011, Q_max = 0.0144, R_0 = 0.36, D = 0.5,
                   T_opt = 27, T_min = 0, T_max = 35, I_opt = 120,
                   dose_resp = 0
      ),
      # boundary presets defined by expert judgement
      param.bounds = list(mu_max=c(0, 4),  EC_50=c(0, 1e6), b=c(0.1, 20), kD=c(0, 10)),
      endpoints = c("A", "r"),
      forcings.req=c("T_act", "I"),
      forcings=list("T_act"=data.frame(time=0, temp=23),
                    "I"=data.frame(time=0, irr=100)),
      control.req = TRUE,
      init = c(A = 1, Q = 0.01, P = 0.18, Dw = 0),
      transfer.interval = -1,
      transfer.biomass = 1,
      transfer.comp.biomass = "A",
      transfer.comp.scaled = "Q"
  ) %>% set_noexposure()
}

#' Simple algae model without environment (Rendal et al. 2023)
#'
#' The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
#' growth model for algae. It follows the concept of a simplified algae model
#' described in Rendal et al. (2023). The model simulates the development of
#' algal biomass. The growth of the algae population is simulated on the basis
#' of growth rates, which are, in contrast to the Weber model, independent on
#' environmental conditions which are usually optimal in laboratory effect studies.
#' The toxicodynamic sub-model describes the effects of growth-inhibiting
#' substances through a corresponding reduction in the photosynthesis rate on the
#' basis of either external or internal concentrations (depending on user choice
#' of *scaled* parameter setting).
#'
#' @section State variables:
#' The model has two state variables:
#' - `A`, Biomass (µg fresh wt/mL, cells/mL *10^4)
#' - `Dw`, scaled internal damage, only takes effect if parameter `scaled = 1`
#'
#' @section Model parameters:
#' - Growth model
#'   - `mu_max`, Maximum growth rate (d-1)
#'
#' - Concentration response (Toxicodynamics)
#'   - `EC_50`, Effect concentration of 50% inhibition of growth rate (µg L-1)
#'   - `b`, slope of concentration effect curve at EC_50 (-)
#'   - `dose_response`, shape of the dose response curve (0 = logit, 1 = probit)
#'
#' - External concentration (Toxicokinetics)
#'   - `scaled`, 1 = scaled internal damage, 0 = no internal damage / 1 = yes (-)
#'   - `kD`, dominant rate constant of toxicant in aquatic environments (d-1), only
#'      takes effect if parameter `scaled = 1`
#'
#' @section Forcings:
#' Simplified model without additional forcings for e.g. irradiation or temperature
#' as implemented in `Algae_Weber`. A constant growth over time is assumed.
#' In case that growth is time dependent, a forcing variable (f_growth) can be set.
#' Forcing time-series are represented by `data.frame` objects consisting of two
#' columns. The first for time and the second for a scaling factor of mu_max.
#' The input format for all forcings is a list of the data frames. If f_growth is
#' not set, a default scaling factor of 1 is used.
#'
#' @section Parameter boundaries:
#' Upper and lower parameter boundaries are set by default for each parameter. This,
#' to avoid extreme values during calibration (particularly likelihood profiling)
#'
#' @section Simulation output:
#' Simulation results will contain the state variables biomass (`A`) and
#' scaled damage concentration (`Dw`).
#'
#' It is possible to amend the output of [simulate()] with additional model
#' quantities that are not state variables, for e.g. debugging purposes or to
#' analyze model behavior. To enable or disable additional outputs, use the
#' optional argument `nout` of [simulate()]. As an example, set `nout=2` to
#' enable reporting of external concentration (`Cw`) and growth scaling factor
#' (`f_growth`). Set `nout=0` to disable additional outputs (default).
#'
#' The available output levels are as follows:
#' - `nout >= 1`: `Cw` external concentration (µg L-1)
#' - `nout >= 2`: `f_growth` growth scaling factor (-)
#' - `nout >= 3`: `dA`, biomass derivative (µg)
#' - `nout >= 4`: `dDw`, damage concentration derivative (µg L-1)
#'
#' @section Solver settings:
#' The arguments to ODE solver [deSolve::ode()] control how model equations
#' are numerically integrated. The settings influence stability of the numerical
#' integration scheme as well as numerical precision of model outputs. Generally, the
#' default settings as defined by *deSolve* are used, but all *deSolve* settings
#' can be modified in *cvasi* workflows by the user, if needed. Please refer
#' to e.g. [simulate()] on how to pass arguments to *deSolve* in *cvasi*
#' workflows.
#'
#' Some default settings of *deSolve* were adapted for this model by expert
#' judgement to enable precise, but also computationally efficient, simulations
#' for most model parameters. These settings can be modified by the user,
#' if needed:
#'
#' - `hmax = 0.01`<br>
#'    Maximum step length in time suitable for most simulations.
#'
#' @references
#' Rendal C, Witt J, Preuss TG, Ashauer R, 2023: A Framework for Algae Modeling
#' in Regulatory Risk Assessment. Environ. Toxicol. Chem. 42(8), pp. 1823-1838.
#' \doi{10.1002/etc.5649}
#'
#' Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
#' and Ratte HT, 2012: Combination of a higher-tier flow-through system and
#' population modeling to assess the effects of time-variable exposure of
#' isoproturon on the green algae Desmodesmus subspictatus and
#' Pseudokirchneriella subcapitata. Environ. Toxicol. Chem. 31(4), pp. 899-908.
#' \doi{10.1002/etc.1765}
#'
#' @return an S4 object of type [AlgaeSimple-class]
#' @seealso [Scenarios], [Transferable]
#' @family algae models
#' @export
#' @aliases AlgaeSimple-class AlgaeSimpleScenario-class
Algae_Simple <- function() {
  new("AlgaeSimple",
      name = "Algae_Simple",
      param.req = c("mu_max", "EC_50", "b", "kD", "scaled", "dose_response"),
      # default values as defined by Weber et al. (2012)
      param = list(mu_max = 1.7380, dose_response = 0, scaled = 0),
      # boundary presets defined by expert judgement
      param.bounds = list(mu_max=c(0, 4), EC_50=c(0, 1e6), b=c(0.1, 20), kD=c(0, 10)),
      endpoints = c("A", "r"),
      forcings.req = c("f_growth"),
      forcings = list(f_growth = data.frame(time = 0, f_growth = 1)),
      control.req = TRUE,
      init = c(A = 1, Dw = 0),
      transfer.interval = -1,
      transfer.biomass = 1,
      transfer.comp.biomass = "A"
  ) %>% set_noexposure()
}

########################
## Simulation
########################

# Solver function for Algae_Weber models
# @param scenario Scenario object
# @param method string, numerical solver used by [deSolve::ode()]
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_algae_weber <- function(scenario, method = "lsoda", hmax = 0.1, nout = 1, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # check for missing parameters
  params.missing <- setdiff(scenario@param.req, names(params))
  if(length(params.missing) > 0)
    stop(paste("parameter missing:", paste(params.missing, collapse=", ")))

  # reorder parameters for deSolve
  params <- params[scenario@param.req]
  # check if any parameter has no value
  if(any(is.na(params) | is.nan(params)))
    stop(paste("parameter value missing:", paste(names(params)[which(is.na(params) | is.nan(params))], collapse=", ")))

  # create forcings list
  forcings <- list(
    scenario@exposure@series,
    scenario@forcings$I,
    scenario@forcings$T_act
  )

  # set names of additional output variables
  outnames <- c("C", "f_T", "f_I", "f_Q", "f_QP", "f_C", "dA", "dQ", "dP")

  # run solver
  ode(y = scenario@init, times=scenario@times, initfunc = "algae_init",
        func = "algae_func", initforc = "algae_forc", parms = params,
        forcings = forcings, dllname = "cvasi", method = method, hmax = hmax,
        outnames = outnames, nout = nout, ...)
}

#' @include solver.R
#' @describeIn solver numerically integrates Algae_Weber scenarios
setMethod("solver", "AlgaeWeber", solver_algae_weber)

# Solver function for Algae_TKTD models
# @param scenario Scenario object
# @param method string, numerical solver used by [deSolve::ode()]
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_algae_tktd <- function(scenario, method = "lsoda", hmax = 0.1, ...) {
  # keep for backwards compatibility, older version used wrong list in constructor
  param_order = c("mu_max", "m_max", "v_max", "k_s", "Q_min", "Q_max", "T_opt",
                    "T_min", "T_max", "I_opt", "EC_50", "b", "kD", "dose_resp",
                    "D", "R_0")

  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # check for missing parameters
  params.missing <- setdiff(param_order, names(params))
  if(length(params.missing) > 0)
    stop(paste("parameter missing:", paste(params.missing, collapse=", ")))

  # reorder parameters for deSolve
  params <- params[param_order]
  # check if any parameter has no value
  if(any(is.na(params) | is.nan(params)))
    stop(paste("parameter value missing:", paste(names(params)[which(is.na(params) | is.nan(params))], collapse=", ")))

  # create forcings list
  forcings <- list(
    scenario@exposure@series,
    scenario@forcings$I,
    scenario@forcings$T_act
  )

  # set names of additional output variables
  outnames <- c("C", "f_T", "f_I", "f_Q", "f_QP", "f_C", "dA", "dQ", "dP", "dDw")

  # run solver
  ode(y = scenario@init, times=scenario@times, initfunc = "algae_tktd_init",
        func = "algae_tktd_func", initforc = "algae_tktd_forc",
        parms = params, forcings = forcings,
        dllname = "cvasi", method = method, hmax = hmax, outnames = outnames, ...)
}

#' @describeIn solver numerically integrates Algae_TKTD scenarios
setMethod("solver", "AlgaeTKTD", solver_algae_tktd)


# Solver function for Algae_Weber models
# @param scenario Scenario object
# @param method string, numerical solver used by [deSolve::ode()]
# @param hmax numeric, maximum step length in time, see [deSolve::ode()]
# @param ... additional arguments passed to [deSolve::ode()]
#' @importFrom deSolve ode
solver_algae_simple <- function(scenario, method = "lsoda", hmax = 0.1, ...) {
  params <- scenario@param
  if(is.list(params))
    params <- unlist(params)

  # create forcings list
  forcings <- list(scenario@exposure@series, scenario@forcings$f_growth)

  # keep for backwards compatibility, older version used wrong list in constructor
  param.req = c("mu_max", "EC_50", "b", "kD", "scaled", "dose_response")

  # check for missing parameters
  params.missing <- setdiff(param.req, names(params))
  if(length(params.missing) > 0)
    stop(paste("parameter missing:", paste(params.missing, collapse=", ")))

  # reorder parameters for deSolve
  params <- params[param.req]
  # check if any parameter has no value
  if(any(is.na(params) | is.nan(params)))
    stop(paste("parameter value missing:", paste(names(params)[which(is.na(params) | is.nan(params))], collapse=", ")))

  # set names of additional output variables
  outnames <- c("Cw", "f_growth", "dA", "dDw")

  # run solver
  ode(y = scenario@init, times=scenario@times, initfunc = "algae_simple_init",
        func = "algae_simple_func", initforc = "algae_simple_forc",
        parms = params, forcings = forcings,
        dllname = "cvasi", method = method, hmax = hmax, outnames = outnames,
        ...)
}

#' @describeIn solver numerically integrates Algae_Simple scenarios
setMethod("solver", "AlgaeSimple", solver_algae_simple)
#setMethod("solver", "AlgaeSimple", function(scenario, ...) solver_algae_simple(scenario, ...))


########################
## Effects
########################

# Calculate effect of Algae scenario
fx_algae <- function(scenario, ...) {
  efx_r <- "r" %in% scenario@endpoints
  # TODO move to a validate_scenario function, this takes precious time on every effect() call
  if(efx_r & has_transfer(scenario))
    stop("endpoint r is incompatible with biomass transfers")

  out <- simulate(scenario, ...)

  efx <- c("A"=tail(out$A, 1))
  if(efx_r) # we skip the log() operation if we can
    efx["r"] <- log(tail(out$A,1) / out$A[1]) / (tail(out[,1],1) - out[1,1])

  efx
}

#' @include fx.R
#' @describeIn fx Effect at end of simulation of [Algae-models]
setMethod("fx", "Algae", fx_algae)
#setMethod("fx", "Algae", function(scenario, ...) fx_algae(scenario, ...))

# test routine ---------------------------------------------------------------------
# Simulation with the Algae_Weber model, compare against reference values
# derived from a simulation with this model for R. subcapitata and isoproturon
# (i.e., recreating the fig in EFSA Scientific Opinion on TKTD, Fig 32
# doi.org/10.2903/j.efsa.2018.5377, showing the flow through exp. of Weber et
# al. 2012) This EFSA example uses the Algae_Weber model, however, to use it for
# testing the Algae_TKTD model (where there is no dilution, but there is an
# internal scaled damage), the background mortality parameter was set so it
# included both the value of background mortality and dilution from Weber's flow
# through, additionally, a high KD parameter value was taken to represent high
# uptake hence the scaled damage tracks almost instantly the water concentration
# test_that("Algae_Weber simulation", {
#   tol <- 1e-4
#
#   # Simulate Algae_TKTD for Rsubcapitata exposed to isoproturon
#   # sim setup
#   sim_end <- 72
#   y_0 <- c(A = 1, Q = 0.01, P = 0.36 * 0.5)
#   times <- seq(from = 0, to = sim_end, by = sim_end / 1000)
#   # parms
#   params <- c(mu_max = 1.7380, m_max = 0.5500, v_max = 0.0520, k_s = 0.0680,
#               Q_min = 0.0011, Q_max = 0.0144,
#               T_opt = 27, T_min = 0, T_max = 35, I_opt = 120,
#               EC_50 = 115, b = 1.268, k = 0.2
#   )
#   # forcings
#   forc_I <- data.frame(times = sim_end, I = rep(100, sim_end))
#   forc_T <- data.frame(times = sim_end, T_act = rep(24, sim_end))
#   forcings <- list(forc_I, forc_T)
#   # exposure
#   weber_exposure <- rsubcapitata@exposure@series
#   # Create Eff.Scen.
#   Rsubcap_Isopr <- Algae_Weber() %>%
#     set_param(params) %>%
#     set_exposure(weber_exposure) %>%
#     set_forcings(I = forc_I,
#                  T_act = forc_T) %>%
#     set_times(times)
#   # simulate
#   Rsubcap_Isopr %>% simulate() -> out
#   # calc % biomass
#   out <- out %>%
#     dplyr::mutate(perc = A/max(A)*100)
#
#   # tests for starting values and sim setup
#   expect_equal(out[, "time"], seq(from = 0, to = 72, 72 / 1000)) # simulation duration
#   expect_equal(out[[1, "A"]], 1) # starting biomass value
#   expect_equal(out[[1, "Q"]], 0.01) # starting Q value
#   expect_equal(out[[1, "P"]], 0.18) # starting P value
#   expect_equal(out[[1, "perc"]], 0.758399, tolerance = tol) # starting %A value
#
#   # discard burnin to steady state
#   out <- out %>%
#     dplyr::filter(time > 12)
#   # identify largest drop in biomass time and % (timing is derived from "out")
#   drop_time <- out[which(out$A == min(out$A)), "time"]
#   drop_perc <- out[which(out$A == min(out$A)), "perc"]
#   expect_equal(drop_time, 32.256, tolerance = tol)
#   expect_equal(drop_perc, 7.096122, tolerance = tol)
#
#   # check timing and magnitude of other peaks (timing is based on known values)
#   # first drop
#   expect_equal(out[[225, "time"]], 28.152)
#   expect_equal(out[[225, "perc"]], 18.6998, tolerance = tol)
#   # second drop
#   expect_equal(out[[286, "time"]], 32.544)
#   expect_equal(out[[286, "perc"]], 7.194282, tolerance = tol)
#   # third drop
#   expect_equal(out[[407, "time"]], 41.256)
#   expect_equal(out[[407, "perc"]], 97.89552, tolerance = tol)
#
# })


# test routine ---------------------------------------------------------------------
# Simulation with the Algae_TKTD model, compare against reference values
# derived from a simulation with this model for R. subcapitata and isoproturon
# (i.e., recreating the fig in EFSA Scientific Opinion on TKTD, Fig 32
# doi.org/10.2903/j.efsa.2018.5377, showing the flow through exp. of Weber et
# al. 2012) This EFSA example uses the Algae_Weber model, however, to use it for
# testing the Algae_TKTD model (where there is no dilution, but there is an
# internal scaled damage), the background mortality parameter was set so it
# included both the value of background mortality and dilution from Weber's flow
# through, additionally, a high KD parameter value was taken to represent high
# uptake hence the scaled damage tracks almost instantly the water concentration
# test_that("Algae_TKTD simulation", {
#   tol <- 1e-5
#
#   # Simulate Algae_TKTD for R. subcapitata exposed to isoproturon
#
#   # sim setup
#   sim_end <- 72
#   y_0 <- c(A = 1, Q = 0.01, P = 0.36 * 0.5, Dw = 0)
#   times <- seq(from = 0, to = sim_end, by = sim_end / 1000)
#
#   # parms
#   params <- c(mu_max = 1.7380, m_max = 0.5500, v_max = 0.0520, k_s = 0.0680,
#               Q_min = 0.0011, Q_max = 0.0144, D = 0, R_0 = 0,
#               T_opt = 27, T_min = 0, T_max = 35, I_opt = 120,
#               EC_50 = 115, b = 1.268, kD = 100, dose_resp = 0
#   )
#   # exposure
#   weber_exposure <- rsubcapitata@exposure@series
#   # Create Eff.Scen.
#   Rsubcap_Isopr <- Algae_TKTD() %>%
#     set_param(params) %>%
#     set_exposure(weber_exposure) %>%
#     set_forcings(I = 100, T_act = 24) %>%
#     set_times(times)
#   # simulate
#   out <- Rsubcap_Isopr %>%
#     simulate() #%>%
#     #dplyr::filter(time >= 10)
#   # calc % biomass
#   out <- out %>%
#     dplyr::mutate(perc = A/max(A)*100)
#
#   # tests for starting values and sim setup
#   expect_equal(out[, "time"], seq(from = 0, to = 72, 72 / 1000)) # simulation duration
#   expect_equal(out[[1, "A"]], 1) # starting biomass value
#   expect_equal(out[[1, "Q"]], 0.01) # starting Q value
#   expect_equal(out[[1, "P"]], 0.18) # starting P value
#   expect_equal(out[[1, "Dw"]], 0) # starting exposure value
#   expect_equal(out[[1, "perc"]], 5.179975e-04, tolerance = tol) # starting %A value
#
#   # discard burnin to steady state
#   out <- out %>%
#     dplyr::filter(time > 12)
#   # identify largest drop in biomass time and % (timing is derived from "out")
#   drop_time <- out[which(out$A == min(out$A)), "time"]
#   drop_perc <- out[which(out$A == min(out$A)), "perc"]
#   expect_equal(drop_time, 31.104, tolerance = tol)
#   expect_equal(drop_perc, 57.79968, tolerance = tol)
#
#   # check timing and magnitude of other peaks (timing is based on known values)
#   # first drop
#   expect_equal(out[[225, "time"]], 28.152)
#   expect_equal(out[[225, "perc"]], 59.80685, tolerance = tol)
#   # second drop
#   expect_equal(out[[286, "time"]], 32.544)
#   expect_equal(out[[286, "perc"]], 65.50502, tolerance = tol)
#   # third drop
#   expect_equal(out[[407, "time"]], 41.256)
#   expect_equal(out[[407, "perc"]], 99.91928, tolerance = tol)
#
# })


# test routine ---------------------------------------------------------------------
# Simulation with the Algae_Simple model, compare against reference values
test_that("Algae_Simple simulation", {
  tol <- 1e-5

  # Simulate Algae_Simple

  model_base <- Algae_Simple()

  # sim setup
  sim_end <- 7
  times <- seq(from = 0, to = sim_end, by = 1)
  y_init <- c(A = 1, Dw = 0)

  # parms
  parms <- c(mu_max = 1,
             EC_50 = 1,
             b = 2,
             scaled = 0,
             kD = 200,
             dose_response = 0)

  # forcings
  forc_fgrowth <- data.frame(times = times, f_growth = rep(1, length(times)))
  forc_C_in <- data.frame(times = times, C = rep(0, length(times)))

  # control run
  effect_scenario <- model_base %>%
    set_param(parms) %>%
    set_tag("control run") %>%
    set_exposure(forc_C_in) %>%
    set_times(times) %>%
    set_forcings(f_growth = forc_fgrowth) %>%
    set_init(y_init)
  results <- effect_scenario %>% simulate(nout = 5)

  # check biomass growth against values from analytical solution
  # generated in R with initial_value * exp(growth_max * t)
  expect_equal(results[[2, "time"]], 1)
  expect_equal(results[[2, "A"]], 2.718282, tolerance = tol)

  expect_equal(results[[3, "time"]], 2)
  expect_equal(results[[3, "A"]], 7.389056, tolerance = tol)

  expect_equal(results[[4, "time"]], 3)
  expect_equal(results[[4, "A"]], 20.085537, tolerance = tol)

  expect_equal(results[[5, "time"]], 4)
  expect_equal(results[[5, "A"]], 54.598150, tolerance = tol)

  expect_equal(results[[8, "time"]], 7)
  expect_equal(results[[8, "A"]], 1096.633158, tolerance = tol)

  #------------------------------------------------------------------------------#

  # control run, constant growth
  # parms
  parms <- c(mu_max = 1,
             EC_50 = 1,
             b = 2,
             scaled = 0,
             kD = 200,
             dose_response = 0)

  # forcings
  forc_fgrowth <- data.frame(times = times, f_growth = rep(0, length(times)))
  forc_C_in <- data.frame(times = times, C = rep(0, length(times)))

  # control run
  effect_scenario <- model_base %>%
    set_param(parms) %>%
    set_tag("control run") %>%
    set_exposure(forc_C_in) %>%
    set_times(times) %>%
    set_init(y_init)
  results <- effect_scenario %>% simulate(nout = 5)

  # check biomass growth against values from analytical solution
  # generated in R with initial_value * exp(growth_max * t)
  expect_equal(results[[2, "time"]], 1)
  expect_equal(results[[2, "A"]], 2.718282, tolerance = tol)

  expect_equal(results[[3, "time"]], 2)
  expect_equal(results[[3, "A"]], 7.389056, tolerance = tol)

  expect_equal(results[[4, "time"]], 3)
  expect_equal(results[[4, "A"]], 20.085537, tolerance = tol)

  expect_equal(results[[5, "time"]], 4)
  expect_equal(results[[5, "A"]], 54.598150, tolerance = tol)

  expect_equal(results[[8, "time"]], 7)
  expect_equal(results[[8, "A"]], 1096.633158, tolerance = tol)

  #------------------------------------------------------------------------------#

  #EC50 run logit
  # forcings
  forc_C_in <- data.frame(times = times, C = rep(1, length(times)))

  effect_scenario <- model_base %>%
    set_param(parms) %>%
    set_tag("control run") %>%
    set_exposure(forc_C_in) %>%
    set_times(times) %>%
    set_init(y_init)
  result_epx <- epx(effect_scenario, level = 50)

  expect_equal(result_epx$r.EP50, 1)

  #------------------------------------------------------------------------------#

  #EC50 run probit
  # parms
  parms <- c(mu_max = 1,
             EC_50 = 1,
             b = 2,
             scaled = 0,
             kD = 200,
             dose_response = 1)

  effect_scenario <- effect_scenario %>%
    set_param(parms)

  result_epx <- epx(effect_scenario, level = 50)

  expect_equal(result_epx$r.EP50, 1)

  #------------------------------------------------------------------------------#

  #EC50 run probit scaled
  # parms
  parms <- c(mu_max = 1,
             EC_50 = 1,
             b = 2,
             scaled = 0,
             kD = 200,
             dose_response = 1)

  effect_scenario <- effect_scenario %>%
    set_param(parms)
  result_epx <- epx(effect_scenario, level = 50)

  expect_equal(result_epx$r.EP50, 1)
})


test_that("output variables", {
  sc <- rsubcapitata
  rs <- sc %>% simulate(nout=0)
  rs2 <- sc %>% simulate(nout=9)

  expect_equal(length(rs2), length(rs) + 9)
  expect_equal(tail(names(rs2), n=3), c("dA", "dQ", "dP"))
  expect_true(any(rs2[, "dA"] != 0))
  expect_true(any(rs2[, "dQ"] != 0))
  expect_true(any(rs2[, "dP"] != 0))
})


test_that("algae effects", {
  sc <- rsubcapitata
  ctrl <- sc %>% set_noexposure() %>% simulate()
  t1 <- sc %>% simulate()

  myeffect <- 1 - tail(t1$A, n=1)/tail(ctrl$A, n=1)
  expect_equal(effect(sc)$A[1], myeffect, tolerance=1e-5)

  # growth rate cannot be determined if transfers are enabled
  sc2 <- sc %>% set_transfer(interval=7)
  expect_error(effect(sc2), regexp="biomass transfer")
})

#' Set output times
#'
#' Minimum and maximum output times define the simulated period for a scenario.
#' Simulation results will be returned for each output time, see [simulate()].
#'
#' Be aware that output times may be modified by [set_exposure()]. Precision of
#' simulation results may be influenced by chosen output times, see [simulate()]
#' for more information.
#'
#' @seealso [simulate()],  [get_times()]
#' @param x vector of [scenarios]
#' @param times `numerical` vector
#' @return Vector of modified [scenarios]
#' @aliases set_times,ANY-method set_times,list-method set_times,EffectScenario-method
#'   set_times,ScenarioSequence-method
#' @include class-EffectScenario.R sequence.R
#' @export
#' @examples
#' # Set simulated period to [2,4] with output intervals of length 1
#' minnow_it %>% set_times(c(2,3,4))
#'
#' # Decrease output interval length to 0.1
#' minnow_it %>% set_times(seq(2, 4, 0.1))
setGeneric("set_times", function(x, times) standardGeneric("set_times"), signature="x")

#' @export
setMethod("set_times", "ANY", function(x, times) stop("Object type is not supported"))

#' @export
setMethod("set_times", "list", function(x, times) lapply(x, set_times, times=times))

#' @export
setMethod("set_times", "EffectScenario", function(x, times) set_times_scenario(x, times))

#' @export
setMethod("set_times", "ScenarioSequence", function(x, times) set_times_sequence(x, times))

set_times_scenario <- function(x, times) {
  if(is(times, "units")) {
    times <- units::drop_units(times)
  }
  if(!is.vector(times) | !is.numeric(times)) {
    stop("Argument 'times' has invalid type")
  }
  if(length(times) < 2) {
    stop("Argument 'times' is too short, must have at least two elements")
  }

  x@times <- times
  x
}

set_times_sequence <- function(x, times) {
  # for backwards compatibility
  if(!methods::.hasSlot(x, "breaks")) {
    attr(x, "breaks") <- breaks_from_sequence(x)
  }
  # assign times to each element in sequence
  for(i in seq_along(x@scenarios)) {
    x@scenarios[[i]] <- set_times(x@scenarios[[i]], times)
  }
  # modify sequence to consider breaks between elements
  x <- split_sequence(x, .messages=FALSE)
  x
}

has_forcings <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_forcings))

  length(x@forcings) > 0
}

has_exposure <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_exposure))

  nrow(x@exposure@series) > 0
}

# Checks if a scenario has no or constant forcing functions & exposure
#
# This is an important property of a scenario, because in case of constant
# forcings we only need to simulate a single moving window instead of all of
# them. This is due to the fact that results for all windows will be identical.
has_constant_forcings <- function(x) {
  # the heuristic would be too complicated/fragile to perform for sequences
  if(any(is_sequence(x)))
    return(FALSE)
  if(length(x) > 1)
    return(sapply(x, has_constant_forcings))

  nrow(x@exposure@series)<2 &
    ( length(x@forcings)==0 |
        all(sapply(names(x@forcings), function(nm) nrow(x@forcings[[nm]])<2)) )
}

has_controls <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_controls))

  length(x@control)>0
}

has_windows <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_windows))

  x@window.length>0
}

has_transfer <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_transfer))

  if(!is(x, "Transferable"))
    return(FALSE)
  if(has_regular_transfer(x))
    return(TRUE)
  has_irregular_transfer(x)
}

has_regular_transfer <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_regular_transfer))

  if(!is(x, "Transferable"))
    return(FALSE)
  x@transfer.interval > 0
}

has_irregular_transfer <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_irregular_transfer))

  if(!is(x, "Transferable"))
    return(FALSE)
  length(x@transfer.times) > 0
}

has_units <- function(x) {
  if(length(x) > 1)
    return(sapply(x, has_units))
  is(x, 'units')
}

test_that("has_forcings", {
  scenario <- Lemna_Schmitt() %>% set_forcings(temp=10)
  expect_true(has_forcings(scenario))
  expect_equal(has_forcings(c(scenario,scenario)), c(TRUE,TRUE))
  # no forcings
  expect_false(has_forcings(GUTS_RED_IT()))
  # invalid arguments
  expect_error(has_forcings(NA))
  expect_error(has_forcings(NULL))
  expect_error(has_forcings(1))
  expect_error(any(has_forcings(1:5)))
})

test_that("has_exposure", {
  # exposure is set
  scenario <- new("EffectScenario", exposure=ExposureSeries(series=data.frame(t=1,c=1)))
  expect_true(has_exposure(scenario))
  expect_true(has_exposure(new("EffectScenario", exposure=ExposureSeries(series=data.frame(t=1:5,c=1:5)))))
  expect_equal(c(TRUE,TRUE), has_exposure(c(scenario, scenario)))

  # no exposure
  scenario <- new("EffectScenario")
  expect_false(has_exposure(scenario))
  expect_equal(c(FALSE,FALSE), has_exposure(c(scenario,scenario)))

  # invalid arguments
  expect_error(has_exposure(NA))
  expect_error(has_exposure(NULL))
  expect_error(has_exposure(1))
  expect_error(any(has_exposure(1:5)))
})

test_that("has_constant_forcings", {
  expect_true(has_constant_forcings(Lemna_Schmitt() %>%
                                      set_exposure(data.frame(t=0,c=0), FALSE)))
  expect_true(has_constant_forcings(Lemna_Schmitt() %>%
                                      set_exposure(data.frame(t=0,c=0), FALSE) %>%
                                      set_forcings(temp=1,rad=1)))
  # non-constant forcings
  expect_false(has_constant_forcings(Lemna_Schmitt() %>%
                                       set_exposure(data.frame(t=0:1,c=0:1))))
  expect_false(has_constant_forcings(Lemna_Schmitt() %>%
                                       set_exposure(data.frame(t=0:1,c=0:1)) %>%
                                       set_forcings(temp=1,rad=1)))
  expect_false(has_constant_forcings(Lemna_Schmitt() %>%
                                       set_exposure(data.frame(t=0,c=0), FALSE) %>%
                                       set_forcings(temp=1,rad=data.frame(t=0:1,r=0:1))))
  # invalid arguments
  expect_error(has_constant_forcings(NA))
  expect_error(has_constant_forcings(NULL))
  expect_error(has_constant_forcings(1))
  expect_error(any(has_constant_forcings(1:5)))
})

test_that("has_controls", {
  expect_true(has_controls(new("EffectScenario", control=c(1))))
  expect_true(has_controls(new("EffectScenario", control=list(1))))
  # on controls
  expect_false(has_controls(new("EffectScenario")))
  # invalid arguments
  expect_error(has_controls(NA))
  expect_error(has_controls(NULL))
  expect_error(has_controls(1))
  expect_error(any(has_controls(1:5)))
})

test_that("has_windows", {
  expect_true(has_windows(new("EffectScenario", window.length=1)))
  # on controls
  expect_false(has_windows(new("EffectScenario")))
  # invalid arguments
  expect_error(has_windows(NA))
  expect_error(has_windows(NULL))
  expect_error(has_windows(1))
  expect_error(any(has_windows(1:5)))
})

test_that("has_transfer", {
  has <- Lemna_Schmitt() %>% set_transfer(interval=1)
  has2 <- Lemna_Schmitt() %>% set_transfer(times=c(0,1,3))
  hasnot <- Lemna_Schmitt() %>%  set_transfer(interval=-1)

  # single
  expect_true(has_transfer(has))
  expect_true(has_transfer(has2))
  expect_false(has_transfer(hasnot))

  # multiple
  expect_equal(has_transfer(c(has, has)), c(TRUE, TRUE))
  expect_equal(has_transfer(c(has, has2)), c(TRUE, TRUE))
  expect_equal(has_transfer(c(has, hasnot)), c(TRUE, FALSE))
  expect_equal(has_transfer(c(hasnot, hasnot)), c(FALSE, FALSE))
})

test_that("has_regular_transfer", {
  expect_true(has_regular_transfer(Lemna_Schmitt() %>% set_transfer(interval=1)))
  expect_true(has_transfer(Lemna_Schmitt() %>% set_transfer(interval=1)))
  # on controls
  expect_false(has_regular_transfer(Lemna_Schmitt() %>% set_transfer(interval=-1)))
  expect_false(has_regular_transfer(Lemna_Schmitt() %>% set_transfer(times=1)))
  # invalid arguments
  expect_false(has_regular_transfer(NA))
  expect_false(has_regular_transfer(NULL))
  expect_false(has_regular_transfer(1))
  expect_false(any(has_regular_transfer(1:5)))
})

test_that("has_irregular_transfer", {
  expect_true(has_irregular_transfer(Lemna_Schmitt() %>% set_transfer(times=1)))
  expect_true(has_transfer(Lemna_Schmitt() %>% set_transfer(times=1)))
  # on controls
  expect_false(has_irregular_transfer(Lemna_Schmitt() %>% set_transfer(interval=-1)))
  expect_false(has_irregular_transfer(Lemna_Schmitt() %>% set_transfer(interval=1)))
  # invalid arguments
  expect_false(has_irregular_transfer(NA))
  expect_false(has_irregular_transfer(NULL))
  expect_false(has_irregular_transfer(1))
  expect_false(any(has_irregular_transfer(1:5)))
})

test_that("has_units", {
  bar <- 1
  foo <- units::set_units(1, sec)

  expect_true(has_units(foo))
  expect_true(all(has_units(list(foo, foo))))
  expect_false(has_units(bar))
  expect_false(any(has_units(list(bar, bar))))
  expect_true(any(has_units(list(foo, bar))))
})

library(morse)

# simple fit from morse tutorial
data("propiconazole")
survData = survData(propiconazole)
fit = survFit(survData, model_type="SD")

# clip MCMC samples to a reasonable length
fit$mcmc[[1]] <- fit$mcmc[[1]][1:50,]
fit$mcmc[[2]] <- fit$mcmc[[2]][1:50,]
fit$mcmc[[3]] <- fit$mcmc[[3]][1:50,]
saveRDS(fit, file=testthat::test_path("../data/morse/fit_sd.rds"))
save(fit, file=testthat::test_path("../data/morse/fit_sd.rdata"))

# fit IT model
fit = survFit(survData, model_type="IT")
fit$mcmc[[1]] <- fit$mcmc[[1]][1:50,]
fit$mcmc[[2]] <- fit$mcmc[[2]][1:50,]
fit$mcmc[[3]] <- fit$mcmc[[3]][1:50,]
saveRDS(fit, file=testthat::test_path("../data/morse/fit_it.rds"))

rm(propiconazole, survData, fit)

#' Set initial state
#'
#' The *initial state* represents the starting values of a scenario's state
#' variables when starting a simulation. A scenario's default initial state
#' may be insufficient to get sensible results.
#'
#' In theory, a scenarios's state variables can be renamed by modifying the names
#' of the initial state vector. However, this is strongly
#' discouraged as this will affect other routines such as [effect()] and [epx()]
#' and may render results useless.
#'
#' @param x vector of `EffectScenario` objects
#' @param init named numeric vector
#'
#' @return modified `EffectScenario` objects
#' @export
#' @include class-EffectScenario.R
#' @rdname set_init
#' @examples
#' # Set initial biomass to 1.0
#' metsulfuron %>% set_init(c(BM=1.0)) %>% simulate()
setGeneric("set_init", function(x, init) standardGeneric("set_init"), signature = "x")

## TODO support non-standard evaluation?
#' @rdname set_init
setMethod("set_init","EffectScenario", function(x, init) {
  if(missing(init))
    stop("Argument `init` is missing")

  if(length(init) == 0) {
    return(x)
  }
  if(is.list(init)) {
    init <- unlist(init)
  }

  # if state vars were defined before, check if these conform to the provided values
  if(length(get_vars(x)) > 0)
  {
    nms <- names(init)
    unused <- setdiff(nms, get_vars(x))
    if("" %in% nms | is.null(nms))
      stop("Argument `init` contains unnamed elements.")
    if(length(unused) > 0)
      warning(paste("Argument `init` contains names which are not state variables:", paste(unused, collapse=", ")))
    ints <- intersect(nms, get_vars(x))
    init <- init[ints]

    if(any(is.nan(init) | is.infinite(init)))
      warning("Argument `init` contains invalid values such as NaN or Inf.")
    if(!is.numeric(init))
      stop("Argument `init`: initial states must be numeric.")
    x@init[ints] <- init[ints]
  }  # otherwise, just use values as they are
  else {
    x@init[names(init)] <- init
  }
  x
})

# set initial condition of scenario sequences
#' @rdname set_init
setMethod("set_init", "ScenarioSequence", function(x, init) {
  for(i in seq_along(x@scenarios)) {
    x@scenarios[[i]] <- set_init(x@scenarios[[i]], init)
  }
  x
})

# convenience function for lists of scenarios
#' @rdname set_init
setMethod("set_init", "vector", function(x, init) {
  sapply(x, set_init, init)
})

test_that("no solver defined", {
  expect_error(solver(new("EffectScenario")), "solver missing")
})

test_that("custom solver", {
  setClass("test_model", contains="EffectScenario")
  setMethod("solver", "test_model", function(scenario, ...) return("foobar"))
  on.exit(removeMethod("solver", "test_model"))
  on.exit(removeClass("test_model"))

  rs <- new("test_model") %>% set_times(0:7) %>% simulate()
  expect_equal(rs, "foobar", ignore_attr=TRUE)
})

---
title: "Modeling Howto"
author: "N. Kehrein and contributors"
date: "24 October, 2024"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
  github_document:
    toc: true
  word_document:
    toc: true
vignette: >
  %\VignetteIndexEntry{Modeling Howto}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  fig.path = "../doc/figures/howto-",
  comment = "#>"
)
```

```{r setup, include=FALSE}
library(cvasi)
library(dplyr)
#options(conflicts.policy=FALSE,warn.conflicts=FALSE)
```

This Howto provides instructions on how to address certain modeling challenges and
offers additional details and context for certain features of the package. A final 
section provides a complete worked out example.
For a more general overview, please refer to the [manual](cvasi-1-manual.html).


## How to access scenario properties

The package provides a number of `set*()` functions to modify scenario
properties such as `set_param()` and `set_times()`. Analogous `set*` functions
are not provided because, generally, it should not be necessary to retrieve data
from scenarios. However, it is possible (but not recommended) to access all
scenario settings via objects *slots*. A *slot* is the name for an object
attribute in *R*. The slots of an object can be accessed by using the `@`
operator. It behaves similar to the `$` operator on named lists:


```{r}
# Create a new scenario object
myscenario <- Lemna_Schmitt()

# Get model name
get_model(myscenario)

# Set a custom tag to identify the scenario
myscenario %>%
  set_tag("Lab experiment #1") -> myscenario

# Get custom tag
get_tag(myscenario)

# The tag is also displayed when printing scenario properties
myscenario

# Accessing scenario slots and their default values
myscenario@forcings.req # forcings required for effect calculations
myscenario@endpoints    # available effect endpoints
myscenario@control.req  # are control runs required for effect calculation?
```

The previous example displays some of the default values of a
*Lemna_Schmitt* scenario. The set of available slots depends on the model
type and is documented in the package help. For instance, scenario
properties shared by all models are documented in the effect scenario
class:

  
```{r, eval=FALSE}
# Call the help page of effect scenarios class
?scenarios
```

A scenario class inherits all slots from its ancestors. A notable class
which modifies simulation behavior and provides additional scenario properties
is `Transferable`: it provides capabilities
to consider biomass transfers at defined time points during simulation.
Details about its class slots and functionality are described in the help pages.

```{r, eval=FALSE}
# Call the help page of the biomass transfer class
?Transferable
```

## Using *tidyr* syntax

The *tidyr* syntax, popularized by the *tidyverse* packages in R, provides a 
coherent and efficient approach to data manipulation and analysis. The *tidyverse*,
which includes the widely used *dplyr* and *ggplot2* packages, follows a 
standardized grammar that makes code more readable and intuitive. *tidyr* 
syntax emphasizes the use of functions with clear and descriptive names. This 
makes it easier for users to understand and reproduce analyses. The `%>%` (pipe) 
operator is a key component of *tidyr* syntax and enables fluent and expressive
concatenation of operations. Overall, the introduction of 
*tidyr* syntax improves code readability, reproducibility, and collaboration, 
resulting in maintainable data analysis pipelines.

In brief, the advantages of *tidyr* syntax are:

- a series of statements can be combined to an intuitive workflow using the
  pipeline (`%>%`) operator
- a short cut for the pipe (`%>%`) operator is Ctrl+Shift+M
- pipelines reduce the need for intermediary variables (but thoughtful intermediates are recommended)
- *tidy* verbs generally take `list` data types as input and return these also as output
- some verbs enrich outputs with additional information, effectively extending
  the output data to a table

```{r}
# The example scenario `metsulfuron` based on the Lemna model by Schmitt et al. (2013)
# is modified by setting a new exposure series and initial state. Then, it is
# simulated.
  metsulfuron %>%
    set_noexposure() %>%  # set no exposure (i.e., a control run)
    set_init(c(BM = 50)) %>%  # set initial biomass
    simulate()
```

## Predictions

TKTD models have both species and substance specificity. If risks are identified
at Tier 1 (standard test species approach) and exposure duration is expected to 
be shorter than in standard tests, the simplest solution is to develop TKTD 
models for standard test species. However, if Tier 2A (geometric mean/evidence 
weighting approach) or Tier 2B (species sensitivity distribution approach) 
information is available, it may be more appropriate to develop TKTD models for 
a wider range of species to increase the accuracy of the risk assessment. 
Validated TKTD models for these different species can be used as an alternative 
to assess specific risks using available field exposure profiles. This includes 
the calculation of exposure profile specific LPx/EPx values (multiplication 
factor for a specific overall exposure profile causing x % lethality or effect) 
based on an appropriate aquatic exposure assessment.

```{r}
# Initialize the random numbers generator
set.seed(123)
# Generating a random exposure series spanning 14 days
random_conc <- runif(15, min=0, max=0.1)
exposure_profile <- data.frame(time=0:14, conc=random_conc)
# Run EPx calculations
minnow_it %>%
  set_exposure(exposure_profile) %>%  # set a specific exposure scenario
  epx()  # run EPx calculations
```

## Moving exposure windows

A moving time window is a computing technique used in data analysis. Data is 
systematically analysed within a window of fixed length that moves or slides 
through the data set, in our case an exposure time-series. The window captures a 
subset of successive data points, and as it moves through the data; simulations
and effect calculations are performed for each window.


`effect()` can report effect levels for all evaluated exposure windows
on demand:
```{r}
# Derive effect levels of all exposure windows
metsulfuron %>% 
  set_window(length=7, interval=1) %>%
  effect(max_only=FALSE)
```

The resulting table describes how effect levels change when the exposure
window moves along the exposure series. It is also possible to specify the marginal
effect threshold of reported results (this prevents overinterpretation of spurious 
effect levels originating from marginal numerical errors introduced during 
simulation), as in the following example:

```{r}
# Only report effect levels larger than 1e-5 = 0.001%
metsulfuron %>% 
  set_window(length=7, interval=1) %>%
  effect(max_only=FALSE, marginal_effect=1e-5)
```

The effect over all moving windows can be visualized using `ggplot`:

```{r}
set.seed(123)
# Generate a random exposure series spanning 14 days
ts <- data.frame(time = 0:14,
                 conc = runif(15, min=0, max=0.1))

# Run EPx calculations for a window length of 7 days and a step size of 1 day
metsulfuron %>%
  set_exposure(ts) %>%
  set_window(length=7, interval=1) %>%
  effect(max_only=FALSE) -> results
results

# Create a plot of effects over time
library(ggplot2)
ggplot(results) +
  geom_point(aes(dat.start, BM*100)) +
  labs(x="Start of window (day)", y="Effect on biomass (%)")
```
The effect plot shows the effect for the time point where each window starts.
Effects are not available, and therefore not plotted, for time points where the
window exceeds the simulated timeframe.

## Simulating biomass transfers

A transfer refers to an event where a certain amount of biomass (BM) is moved to
a new medium after a period of time. This feature replicates a procedure 
occurring e.g. in *Lemna* effect studies and may be necessary to recreate study 
results. At each transfer, a defined amount of biomass is transferred to a new 
medium. This is modeled by interrupting the simulation at a transfer time point, 
modifying the biomass level BM, and scaling affected compartments according to 
new biomass levels. Scaling of compartments depending on biomass, such as 
internal toxicant mass, is necessary to correctly reflect mass balances and 
concentrations over time.

Option 1: Regular intervals
```{r}
metsulfuron %>%
  set_init(c(BM=1)) %>%
  set_noexposure() %>%
  set_transfer(interval=3, biomass=1) %>%
  simulate() -> result
result

library(ggplot2)
ggplot(result) +
  geom_line(aes(time, BM)) +
  labs(x="Time (days)", y="Biomass (g_dw/m2)", title="Biomass transfer every three days")

```

Option 2: Custom time points and custom biomass
```{r}
metsulfuron %>%
  set_init(c(BM=1)) %>%
  set_noexposure() %>%
  set_transfer(times=c(3, 6), biomass=c(1, 0.5)) %>%
  simulate() -> result2

library(ggplot2)
ggplot(result2) +
  geom_line(aes(time, BM)) +
  labs(x="Time (days)", y="Biomass (g_dw/m2)", title="Biomass transfer at custom time points")
```

```{r, eval=FALSE}
# Call the help page of set_transfer
?set_transfer
```

## Fitting model parameters

Parameters of a model can be fitted (calibrated) against observed effect data.
To fit a scenario to observed effect data:

1. Combine the necessary inputs (model, exposure time-series, effect data).
2. Fit the scenario for the given parameters and observed effect data.

For the first step, two options are available:

* When only one exposure level (e.g., one experimental treatment) and 
corresponding effect dataset (effect data only contain observations corresponding 
to one exposure scenario) are of interest, then a scenario consisting of
model parameters and exposure time-series can directly be fitted to
effect data using the `calibrate()` function. 

* As a more complex and versatile option, scenarios and effect data are combined into
one or more *calibration sets*. Each *calibration set* has exactly one scenario that
describes the parameters of the experiment. If more than one *calibration set*
is defined, scenario parameters can differ between set, but don't have to,
A numerical weight can be assigned to each *calibration set* to control its
impact on the calculated error term and derived fitting procedure. 
All *calibration sets* are then passed to the `calibrate()` function.

The following describes an example which fits selected parameters of the
*Lemna* model by Schmitt *et al.* (2013) to observed frond numbers from
experiments with the herbicide *metsulfuron*: 

Option 1: direct calibration with one scenario and one effect dataset

```{r warning=FALSE}
library(dplyr)
# set k_phot_fix, k_resp and Emax to the defaults recommended in Klein et al. (2022)
# for Tier 2C studies. Set initial biomass to 12.0 (original data is in fronds
# rather than biomass, but for sake of simplicity, no conversion was performed).
control <- metsulfuron %>% 
  set_param(c(k_phot_fix=TRUE, k_resp=0, Emax=1)) %>% 
  set_init(c(BM=12)) %>%
  set_noexposure() %>%
  set_bounds(list(k_phot_max=c(0, 0.5)))
# `metsulfuron` is an example scenario based on Schmitt et al. (2013) and therefore
# uses the historical, superseded nomenclature by Schmitt et al. We recommend using
# the newer SETAC Lemna model by Klein et al. (2022) for current applications,
# see `Lemna_SETAC()` for details.

# Get control data from Schmitt et al. (2013)
obs_control <- schmitt2013 %>%
  filter(trial == "T0")

# Fit parameter `k_phot_max` to observed data: `k_phot_max` is a physiological
# parameter which is typically fitted to the control data
fit1 <- calibrate(
  x = control,
  par = c(k_phot_max = 1),
  data = obs_control,
  output = "BM",
  method="Brent" # Brent is recommended for one-dimensional optimization
)
fit1$par

# Update the scenario with fitted parameter and simulate it
fitted_growth <- control %>% 
  set_param(fit1$par) %>%
  set_bounds(list(
    EC50=c(0, 1000),
    b=c(0.1, 10), 
    P_up=c(0, 0.1)
  ))
sim_mean <- fitted_growth %>%
  simulate() %>%
  mutate(trial="control")

# Exposure and observations in long format for plotting
treatments <- obs_control %>%
  select(time, conc) %>%
  mutate(trial="control")
obs_mean <- obs_control %>%
  select(time, BM=obs) %>%
  mutate(trial="control")

# Plot results
plot_sd(
  model_base = fitted_growth,
  treatments = treatments,
  rs_mean = sim_mean,
  obs_mean = obs_mean
)
```

Option 2: Create a list of *calibration sets*  and then fit TK/TD model parameters
  on all datasets and exposure levels at the same time:

```{r warning=FALSE}
# Use all trials and exposure series from Schmitt et al. (2013) as input for fitting
fit2 <- calibrate(
  fitted_growth,
  data=schmitt2013,
  par=c(EC50=0.3, b=4.16, P_up=0.005),
  output="BM",
  method="L-BFGS-B",
  verbose=FALSE
)
fit2$par

# Update the scenario with fitted parameter and simulate all trials
fitted_tktd <- fitted_growth %>%
  set_param(fit2$par)

treatments <- schmitt2013 %>% select(time, conc, trial)
rs_mean <- fitted_tktd %>%
  batch(treatments) %>%
  simulate()

# Observations in long format for plotting
obs_mean <- schmitt2013 %>% select(time, obs, trial)

# Plot results
plot_sd(
  model_base = fitted_tktd,
  treatments = treatments,
  rs_mean = rs_mean,
  obs_mean = obs_mean
)
```

The resulting scenario with fitted parameters shows a very good fit with
the observed effects from experiments.

## Obtaining confidence intervals for parameters

After model parameters have been calibrated during model fitting against 
observational data, the parameter estimates obtained can be 
further evaluated by assessing the 95% confidence intervals for each parameter 
through likelihood profiling. 

First, the profiling is conducted using the `lik_profile()` function. Then, the 
profiles can be visualized using `plot()`.

```{r}
# using the calibration set and calibrated parameters from the previous Lemna 
# Schmitt example, a likelihood profiling is done
# We set parameter boundaries to constrain the likelihood profiling within these 
# boundaries (this is optional)

# Calibrate a model to a list of calibration sets
fit <- calibrate(
  fitted_growth,
  data=schmitt2013,
  par=c(EC50=0.3, b=4.16, P_up=0.005),
  output="BM",
  method="L-BFGS-B",
  verbose=FALSE
)

# Conduct profiling
res <- lik_profile(x = fitted_growth,
                   data = schmitt2013,
                   par = fit$par, # the parameter values after calibration
                   output = "BM", # the observational output against which to 
                                  # compare the model fit
                   bounds = list(EC50_int = list(0.1, 4), 
                                 b = list(1, 5),
                                 P = list(0.0001, 0.2)))
# Visualise profiling results
plot(res)

# Access 95% confidence intervals of profiled parameters
res$EC50$confidence_interval
res$b$confidence_interval

```

Optionally, we can adapt caliset weights and tags, for example, a study 
was done by 2 labs (A and B), where the data of lab 2 should 
be considered to have more weight. In this case, the `individual` argument
needs to be set to TRUE.

```{r}
# Create calisets
cs <- td2cs(fitted_growth, tox_data(schmitt2013))
# Define tags and weights
lab_tags <- c("A", "B", "B", "A", "B", "B", "B")
lab_weights <- c(1, 1.5, 1.5, 1, 1.5, 1.5, 1.5)

for(i in 1:length(cs)){
  cs[[i]]@tag <- lab_tags[i]
  cs[[i]]@weight <- lab_weights[i]
}

res_2 <- lik_profile(x = cs, # the calibration set
                   par = fit$par, # the parameter values after calibration
                   output = "BM", # the observational output against which to 
                   # compare the model fit
                   bounds = list(EC50_int = list(0.1, 4), 
                                 b = list(1, 5),
                                 P = list(0.0001, 0.2)),
                   individual = TRUE)
                   
plot(res_2)


```

## Assessing parameter correlations and identifiability

Relations between estimates of the calibrated parameters can be 
visualized using a parameter space explorer. This supports identifying potential
parameter correlations and identifiability issues.

The parameter space explorer draws random parameter sets from the 95% confidence
intervals obtained for each parameter during likelihood profiling, and evaluates 
the fit (likelihood) of each of these parameter sets by comparing the model with
the original parameter set against the model with the new, randomly draw set.
 
First, the parameter space exploration is conducted using the `explore_space()` 
function. Then, the space can be visualized using `plot()`.


```{r, eval=FALSE}
# Call the help page for more information about the parameter space explorer
?explore_space

```

```{r}

# Calibrate a model to a list of calibration sets
cs <- td2cs(fitted_growth, tox_data(schmitt2013))
fit <- calibrate(
  cs,
  par=c(EC50=0.3, b=4.16, P_up=0.005),
  output="BM",
  method="L-BFGS-B",
  verbose=FALSE
)
# conduct profiling
res <- lik_profile(x = cs, # the calibration set
                   par = fit$par, # the parameter values after calibration
                   output = "BM", # the observational output against which to 
                                  # compare the model fit
                   bounds = list(EC50_int = list(0.1, 4), 
                                 b = list(1, 5),
                                 P = list(0.0001, 0.2)))
# Conduct space exploration
res_space <- explore_space(
  x = cs,
  par = fit2$par,
  res = res, # output of the likelihood profiling function
  output = "BM")

# Visualize the parameter space
plot(res_space)

```

## Changes in parameter values over time

Sequences can be used to represent changing conditions over time, such as a
change in model parameters which would otherwise be constant. This can be
used to represent events such as a pump failure or change in temperature.

The function `sequence()` creates an object which represents a sequence of several
scenarios. The sequence is treated as a single scenario and each scenario is
simulated one after the other.


```{r}
# A base scenario is created for the whole period, but parameters are only
# valid until day 7
sc1 <- metsulfuron %>%
  set_times(0:14)

# A parameter change occurs at day 7: k_loss increases from 0 to 0.1 d-1
sc2 <- sc1 %>%
  set_param(c(k_loss=0.1))

# Combine both scenarios to a sequence with a break at *t = 7*:
# scenario `sc1` will be simulated for *t = [0, 7]*, and
# scenario `sc2` will be simulated for *t = [7, 14]*.
sq <- sequence(list(sc1, sc2), breaks=7)
simulate(sq)
```
For more information:

```{r, eval=FALSE}
# Call the help page of `sequence`
?sequence
```

## Decrease assessment runtime

There are multiple ways to decrease the runtime of a simulation. One option 
is to increase the maximum step length of the solver, `hmax`. It is an
optional argument which can be passed to either `simulate()`, `effect()`, or
`epx()`. The larger `hmax`, the faster the simulation generally completes. 
However, be careful: The larger `hmax`, the greater the risk that simulation 
results will be inaccurate and simulations may even fail due to numerical issues.

```{r}
# Simulations with a maximum solver step length of hmax=0.01
metsulfuron %>%
  set_times(0:7) %>%
  simulate(hmax=0.1)
```

The calculation of *EPx* values may take a lot of time if one or more of the following
conditions apply:

1) a large number of scenarios is assessed
2) the simulated exposure time-series are very long
3) moving exposure windows are considered
   which are very small in comparison to the length of the exposure series

The `epx()` function is implemented in a way that it can parallelize
calculations by using more than one CPU. To enable parallel processing,
a command like the following needs to preceed the call to `epx()`:

```{r eval=FALSE}
future::plan(future::multisession)
```

If the computer running the calculation has `n` physical CPU cores, then the
time necessary to calculate *EPx* values will (in the best case) decrease by
the same factor.
For more information on how to make use of parallelization in R, please refer to the
`future` package:

```{r eval=FALSE}
vignette("future-1-overview", package="future")
```


## Implementing custom models

The set of supported models can be extended by users as needed.
Experimental or one-time use models can be implemented in a user's script and
inserted into the package's workflow.

The starting point to add a new model is an implementation of the model's
rules and dynamics in *R* code. In most cases, this will be a code snippet
which describes the model's Ordinary Differential Equations (ODE):

```{r, warning=FALSE}
## An exemplary implementation of the GUTS-RED-SD TKTD model ##
# Model ODEs following the deSolve specification
sd_ode <- function(t, state, params) {
  with(as.list(c(state, params)), {
    dDw <- kd * (Cw(t) - Dw)            # dDw/dt, scaled damage
    dH <- kk * max(0, Dw - z)           # dH/dt, cumulative hazard
    
    list(c(dDw, dH))                    # return derivatives
  })
}
```

The previous example implements the model equations in a way that
can be processed by the `deSolve` package for numerical integration. For a
detailed description on how to use `deSolve`, please refer to its vignette:

```{r, eval=FALSE}
vignette("deSolve", package="deSolve")
```

With some additional scenario data such as initial state, output
time points, and model parameters, we are able to simulate our custom
TKTD model:

```{r, warning=FALSE}
## Properties of a sample scenario ##
init <- c(Dw=0, H=0)                         # initial state
times <- 0:5                                 # output time points [0,5]
param <- c(kd=22, hb=0.01, z=0.5, kk=0.08)   # model parameters
exp <- data.frame(time=0:5,                  # exposure time-series
                  conc=c(0,1,1,0.5,0.6,0.2))

# Create a linear interpolation of the exposure time-series
expf <- approxfun(x=exp$time, y=exp$conc, method="linear", rule=2)
# Extend parameter set by interpolated exposure series
paramx <- as.list(c(param, Cw=expf))

# Numerically solve the ODE
deSolve::ode(y=init, times=times, parms=paramx, func=sd_ode)
```

When we have made sure that our custom model can be simulated and works as expected,
we can continue with integrating it into the package's workflow. First, we will
define a new scenario class that identifies our custom model by a unique name.
Next, we have to tell the package how the models of this class can be simulated.
In a final step, we may have to describe how effects are calculated for these
models.

We start by defining a new scenario class that derives from a suitable ancestor.
In general, we can inherit from the base class `EffectScenario` which
provides the general scenario capabilities. However, if the
custom model is just a variant of an existing model category, then it will be
easier to derive from specialized scenario class that already provides certain
features such as effect calculation. The following
class tree gives an overview of the scenario classes in use:

    EffectScenario
    |    
    |   Transferable
    |___|__ Lemna
    |   |   |__ LemnaSchmitt
    |   |   |__ LemnaSetac
    |   |__ Magma
    |   |__ Algae
    |       |__ AlgaeWeber
    |       |__ AlgaeTKTD
    |       |__ AlgaeSimple 
    |
    |__ GutsRedSd
    |__ GutsRedIt
    |
    |__ Deb
        |__ DebAbj
        |__ DebTox

To give an example, to implement a variant of a *Lemna* model, it would be advisable
to derive from the class `Lemna` to benefit from already implemented features
such as effect endpoint calculation and simulation of biomass transfers.
For the custom GUTS-RED-SD model, the ideal choice would be to derive from
`GutsRedSd` to minimize the implementation overhead. However, we will
derive from the general `EffectScenario` class for the sake of our example and
create a scenario object:

```{r}
## Integrate a new model class into the package workflow ##
# Create a unique class that derives from 'EffectScenario'
setClass("MyGuts", contains="EffectScenario")

# Create an object of the new class and assign scenario properties
new("MyGuts", name="My custom model") %>%
  set_init(init) %>%
  set_times(times) %>%
  set_param(param) %>%
  set_exposure(exp, reset_times=FALSE) %>%
  set_endpoints("L") -> myscenario

myscenario
```

The object `myscenario` now carries all the settings and data which we also
used for our test simulation. In the next step, the `solver()` function will
be overloaded to handle objects
of the newly defined scenario class `MyGuts`. The adapted `solver()` function will
collect the properties of a given scenario object, call the ODE solver, and return
simulation results:

```{r}
# the actual function calling deSolve can have a different signature
solver_myguts <- function(scenario, ...) {
  # get relevant data from scenario
  init <- scenario@init
  param <- scenario@param
  times <- scenario@times
  exp <- scenario@exposure@series
  if(nrow(exp) == 1) { # extend exposure series to have at least two rows
    row2 <- exp[1,]
    row2[[1]] <- row2[[1]]+1
    exp <- rbind(exp, row2)
  }
  
  # Create a linear interpolation of the exposure time-series
  expf <- approxfun(x=exp[,1], y=exp[,2], method="linear", rule=2)
  # Extend parameter set by interpolated exposure series
  paramx <- as.list(c(param, Cw=expf))
  
  # pass everything to ODE solver for numerical integration
  as.data.frame(deSolve::ode(y=init, times=times, parms=paramx, func=sd_ode, ...))
}

## Overload the solver() function ##
# The functions signature, i.e. the number and names of its arguments, must stay as is
setMethod("solver", "MyGuts", function(scenario, ...) solver_myguts(scenario, ...))
```

Overloading an *S4* function is done using `setMethod()`. The first argument
to `setMethod()` identifies which function we want to overload, in this case it is
`solver`. The second argument, `MyGuts`, defines which object type our
overloaded function will accept. In this case, we want to provide an implementation
to simulate `MyGuts` scenarios. *R* will decide during runtime which of the
function candidates to execute when `solver()` is called based on the type
of the first argument to `solver()`. The third argument, `function(object, ...) solver_myguts(scenario=object, ...)`
forwards calls to an appropriate function which can handle
`MyGuts` objects. The function signature `function(object, ...)` must stay exactly 
as it is and must not be modified.

The code within the body of `solver_myguts()` is almost identical to the original
code used to simulate the prototyped model. We have to deal with one corner
case, though: some exposure time-series will contain only a single row, representing
constant exposure over time, but this would raise an error in `approxfun()`.
In order to avoid this error, we will duplicate the first row and append it to
the series. Technically, this issue should be handled in the prototyped code as
well but was left out for reasons of brevity.

The parameterized scenario object `myscenario` of class `MyGuts` can be
passed to the overloaded `simulate()` function:

```{r}
myscenario %>% simulate()
```


The custom model can now be simulated using the framework but it is still
missing effect endpoints. If, on the other hand, we had decided to inherit our
scenario class from a suitable ancestor, such as `GutsRedSd`, we could
have stopped at this point as the ancestor would provide routines to calculate
effects.

By default, any state variable is available
as an effect endpoint and the calculated effect would reflect the change in the
state variables' value at the end of the simulation. However, we need to
calculate the *survival* endpoint in a different manner for GUTS-RED type models.
To implement this specialized endpoint, we need to overload the
function `fx()` which calculates effect endpoints:

```{r}
## Overload effect endpoint calculation ##
# fx() is called by effect()
setMethod("fx", "MyGuts", function(scenario, ...) fx_myguts(scenario, ...))

# @param scenario Scenario object to assess
# @param ... any additional parameters
fx_myguts <- function(scenario, ...) {
  # simulate the scenario (it is already clipped to the moving exposure window)
  out <- simulate(scenario, ...)
  # only use model state at the end of the simulation
  end <- tail(out, 1)
  # length of the simulated period, first column of result contains time
  t <- tail(out[, 1], 1)- out[[1, 1]]
  # calculate survival according to EFSA Scientific Opinion on TKTD models
  # p. 33, doi:10.2903/j.efsa.2018.5377
  survival <- exp(-end$H) * exp(-scenario@param$hb * t)
  return(c("L"=survival))
}

# Derive effect levels for our sample scenario
myscenario %>% effect()
```

`fx()` is used by `effect()` to derive effect endpoints for each model.
By convention, any overload of the `fx()` function must return a named
numerical vector containing the effect endpoints for the provided scenario
and moving exposure window. The scenario will already be parameterized to
only simulate the current exposure window. The return value of `fx()` will then
be used to calculate effect levels. In case of models requiring a control
scenario, the effect level will generally be calculated as `1 - effect/control`. For models not
requiring a control, the overloaded `fx()` must return the final effect
value.


## Complete working example

In this section, a complete example is given to make predictions and 
evaluate toxic effects using a calibrated model, as might be conducted in a risk
assessment context. Specifically, the 
*Lemna* model will be set up to match the study of
Schmitt et al. (2003, doi: 10.1016/j.ecolmodel.2013.01.017) who exposed *Lemna*
to metsulfuron-methyl.

The model will be parameterized with the values as described in the study. Then,
the model will be inspected and used to make predictions for 
exposure scenarios, plot results, and get EPx calculations.


### Setting up a scenario
Setting up the model (i.e. creating a scenario) involves defining the parameters,
the environmental variables (external forcings and chemical exposure), and the initial
conditions. Further, a tag can be assigned to easily identify the 
scenario by name. Also, for primary producer models, a transfer of biomass
can be defined to match the experimental design of the study.

```{r}
## Get all model parameters
# Lemna model parameters taken from file 'mm2.r' included
# in supporting material of Schmitt et al. (2013)
param_study <- list(
  #     - Effect -
  Emax     = 0.784,   # [same as conc. data]  maximum effect concentration
  EC50     = 0.3,     # [same as conc. data]  Midpoint of effect curve
  b        = 4.16,    # [-]  Slope of effect curve
  #     - Toxicokinetics -
  P_up     = 0.0054,  # [cm/d]  Permeability for uptake
  AperBM   = 1000,    # [cm^2/g_dw]  Frond area/dry weight
  Kbm      = 0.75,    # []  Biomass(fw)-water partition coefficient
  P_Temp   = F,       # Switch for temperature dependence of cuticle permeability
  MolWeight = 381,    # [g/mol] Molmass of molecule (determines Q10_permeability)
  #     - Fate of biomass -
  k_phot_fix  = F,    # If True k_G_max is not changed by environmental factors
  k_phot_max  = 0.47, # [1/d]  Maximum photosynthesis rate
  k_resp   = 0.05,    # [1/d]  Respiration rate at ref. temperature
  k_loss   = 0.0,     # [1/d]  Some rate of loss (e.g. Flow rate)
  #     - Temperature dependence -
  Tmin     = 8.0  ,   # [°C]  Minimum growth temperature 
  Tmax     = 40.5 ,   # [°C]  Maximum growth temperature
  Topt     = 26.7 ,   # [°C]  Optimum growth temperature 
  t_ref    = 25,      # [°C]  Reference temperature for respiration rate
  Q10      = 2,       # [-]   Temperature dependent factor for respiration rate
  #     - Light dependence (Linear) -
  k_0      = 3 ,      # [1/d]  Intercept of linear part
  a_k      = 5E-5 ,   # [(1/d)/(kJ/m^2/d)]   Slope of linear part
  #     - Phosphorus dependence (Hill like dependence) -
  C_P      = 0.3,     # [mg/L]  Phosporus concentration in water
  CP50     = 0.0043,  # [mg/L]  P-conc. where growth rate is half
  a_P      = 1,       # []  Hill coefficient
  KiP      = 101,     # [mg/L]  P-inhibition constant for very high P-conc.
  #     - Nitrogen dependence (Hill like dependence) -
  C_N      = 0.6,     # [mg/L]  Nitrogen concentration in water
  CN50     = 0.034,   # [mg/L]  N-conc. where growth rate is half
  a_N      = 1,       # []  Hill coefficient
  KiN      = 604,     # [mg/L]  n-inhibition constant for very high P-conc.
  #     - Density dependence -
  BM50     = 176,     # [g_dw/m^2] Cut off BM
  #     - Others -
  mass_per_frond = 0.0001,  # [g_dw/frond]  Dry weight per frond
  BMw2BMd  = 16.7     # [g_fresh/g_dry]  Fresh- / dryweight 
)


## Define the forcing variables
forc_temp <- data.frame(t = 0, tmp = 12) # [°C]  Current temperature (may also be a table)
forc_rad  <- data.frame(t = 0, rad = 15000) # [kJ/m²/d]  Radiation  (may also be given as table)


## Define a simple exposure pattern
# t 0..6  concentration 1 ug/L
# t 7..14 concentration 0 ug/L
exposure <- data.frame(time = 0:14, 
                       conc = c(rep(1, 7), rep(0, 8))
                       )


## Set initial values 
# given in file 'mmc2.r' of Schmitt et al. (2013)
init <- c(
  BM       = 50,     # [g_dw/m^2]  Dry Biomass dry weight per m2
  E        = 1,      # (0-1)  (Toxic) Effect = Factor on growth rate  (Range: 0 - 1, 1=no effect)
  M_int    = 0       # [ug]   Amount of toxicant in biomass
)

## create a scenario object, containing the model (with parameters) and the exposure time-series
Lemna_Schmitt() %>%               # the Lemna model by Schmitt et al. (2013)
  set_tag("metsulfuron") %>%      # set a tage for the specific implementation of the model
  set_init(init) %>%              # set the starting values (as prepared above)
  set_param(param_study) %>%      # set the parameters (as prepared above)
  set_exposure(exposure) %>%      # set the exposure scenario (exposure time series)
  set_forcings(temp=forc_temp, rad=forc_rad) -> metsulfuron # set the external forcing 
# variables, and save everything as an object


```


### Simulating a scenario and plotting

```{r out.width = "100%"}
## simulate with model, under a range of different exposure scenarios
# create several exposure scenarios
exp_scen <- schmitt2013 %>% select(time, conc, trial)

# simulate for all these scenarios
results <- metsulfuron %>%
  batch(exp_scen) %>%
  simulate()

# plot results
plot_sd(
  model_base = metsulfuron,
  treatments = exp_scen,
  rs_mean = results,
)


## simulate with model, under a range of different exposure scenarios, and including 
## a biomass transfer
# simulate for all scenarios
results <- metsulfuron %>%
  set_transfer(interval = c(5), biomass = 10) %>% # implement a biomass transfer every 5 days
  batch(exp_scen) %>%
  simulate()

# plot results
plot_sd(
  model_base = metsulfuron,
  treatments = exp_scen,
  rs_mean = results,
)

```

---
title: "User Manual"
author: "N. Kehrein"
date: "24 October, 2024"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
  github_document:
    toc: true
  word_document:
    toc: true
vignette: >
  %\VignetteIndexEntry{User Manual}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  fig.path = "../doc/figures/manual-",
  comment = "#>"
)
```

```{r setup, include=FALSE}
library(cvasi)
library(purrr)
library(dplyr)
```

The *cvasi* package is a software library that extends the features of the
programming language [*R*](https://www.r-project.org/) by routines for ecotox
effect modeling. The routines are intended to be used for scientific purposes as
well as in the context of regulatory risk assessments. *cvasi*  implements model
equations and default parameters to simulate a set of toxicokinetic-toxicodynamic
(TKTD) models. The models are intended to assess the effects of substance exposure
on effect endpoints, such as growth and reproduction, for a range of standard test
species.

The package offers a standardized framework which can be extended by (almost) any
TKTD model that can be implemented in *R*. Its modeling workflows are designed to be
self-explanatory and as intuitive as possible to make modeling approaches more
accessible. In addition, the framework's routines are optimized to reduce the
computing time required for extensive risk assessments.

# Required skills

The *cvasi* package and its routines can be used by anyone with basic to intermediate
knowledge of the *R* programming language. The routines can be used in a
procedural fashion, commonly referred to as *base R*, in which function statements
are executed, assigned to a variable, and then passed on to the next function:

```{r}
library(cvasi)

## Sample base R workflow ##
# Create a scenario object of model GUTS-RED-IT
my_it <- GUTS_RED_IT()
# Set model parameters
my_it <- set_param(my_it, c(kd=1.2, hb=0, alpha=9.2, beta=4.3))
# Print scenario details
my_it
```

A more advanced approach is to use the *tidyr* syntax (cf. Wickham & Grolemund 2017)
which uses verb-like statements to modify data. *tidyr* syntax can be used to
chain functions calls which pass the result of one function on to the next:

```{r}
## Example 'tidyr' workflow ##
# the pipeline (%>%) symbol passes results to the next statement
GUTS_RED_IT() %>%
  set_param(c(kd=1.2, hb=0, alpha=9.2, beta=4.3))
```

Both workflow styles achieve the same result, but *tidyr* is terser,
especially if several operations are performed on the same object.
*tidyr* style is the recommended way to interact with the *cvasi* package.

# How to install

The simplest way to install the *cvasi* package is by downloading it from *CRAN*:

```{r eval=FALSE,include=FALSE}
install.packages("cvasi", dependencies=TRUE)
```


The latest development version can be installed from GitHub:

```{r eval=FALSE}
install.packages("remotes", dependencies=TRUE)
remotes::install_github("cvasi-tktd/cvasi", dependencies=TRUE)
```

The package and its source code is also available on [GitHub](https://github.com/cvasi-tktd/cvasi/).


# Modeling and assessment workflow

## Overview

The central concept within the *cvasi* package is the *scenario*. It encapsulates
all key properties to run a simulation such as initial state and model parameters.
The scenario can be thought of as the collection of
data and parameters necessary to reproduce a certain study condition or situation.
The set of available scenario properties depends on the model type but may include:

 - Initial state,
 - Model parameters,
 - Assessment parameters,
 - Exposure time-series, and
 - Environmental factor time-series

Scenario properties can be modified using intuitively named *tidyr* verbs: in
general, each verb accepts a scenario as an argument and returns a modified
scenario as the result. For instance, the `set_param()` function will update 
a scenario's parameters to new values:

```{r}
library(cvasi)

# Create a new GUTS-RED-IT scenario and set its model parameters
GUTS_RED_IT() %>%
  set_param(c(kd=1.2, hb=0, alpha=9.2, beta=4.3))
```

If a scenario is properly defined, it can be passed to dedicated functions that,
for instance, run a simulation or derive assessment endpoints. These higher-level functions
do not return a scenario object but a table of e.g. simulation results for that
particular scenario. As an example, simulating
the sample scenario `minnow_it` will run a *GUTS-RED-IT* model using
substance-specific parameters:

```{r}
# Example GUTS-RED-IT scenario derived from an acute fish toxicity study
# of the fathead minnow and Chlorpyrifos (Geiger et al. 1988)
minnow_it %>%
  simulate()
```

The table returned by `simulate()` contains the value of each state variable at
the requested output time points. In case of the *GUTS-RED-IT* model, this
comprises the state variables `D` (scaled damage) and `H` (cumulative hazard).
Effect levels, effect profiles, and dose-response graphs
can be derived in a similar manner. Please refer to EFSA PPR Panel (2018) or the
package help for more information on *GUTS-RED* type models:

```{r, eval=FALSE}
# Access the package help on GUTS-RED type models
?"GUTS-RED-models"
```

## Setting up a scenario

A scenario is created by first calling a constructor function that is dedicated to
a particular model. The *cvasi* package supports multiple TKTD models, for instance:

- Reduced General Unified Threshold models of Survival (*GUTS-RED*)
   - `GUTS_RED_SD()`
   - `GUTS_RED_IT()`
- Macrophyte models
   - `Lemna_Schmitt()`, published by Schmitt *et al.* (2013)
   - `Lemna_SETAC()`, published by Klein *et al.* (2021)
   - `Magma()`, a generic macrophyte model by Witt et al., sbumitted
- Algae models
   - `Algae_Weber()`, published by Weber *et al.* (2012)
   - `Algae_TKTD()`, based on Weber *et al.* (2012), but with scaled damage
   - `Algae_Simple()`, simplified growth model without external forcing variables
- Dynamic Energy Budget (*DEB*) models
   - `DEB_abj()`, the *abj* model with type M acceleration
   - `DEBtox()`, as implemented in the *DEBtox2019* module of [*BYOM*](https://www.debtox.info/byom.html)

The listed functions create objects that will carry all
data and settings to fully describe a scenario. In case of a *GUTS-RED-IT*
scenario, one needs to provide at least the four model parameters and an exposure
time-series to fully define a scenario:

```{r}
# Define an exposure time-series
myexposure <- data.frame(time=c(0, 1, 1.01, 5),
                         conc=c(10, 10, 0, 0))

# Create and parameterize a scenario
GUTS_RED_IT() %>%
  set_param(c(kd=1.2, hb=0, alpha=9.2, beta=4.3)) %>%
  set_exposure(myexposure) -> myscenario
```

One can examine the main scenario properties by passing the scenario to the
console like in the next example. The console will print the model name,
model parameters (*param*), initial state (*init*), enabled
effect endpoints (*endpt*), output times (*times*), environmental forcings
(*forcs*), and exposure series (*expsr*). Depending on model type, the
scenario may carry additional properties that are not necessarily displayed in the
console output.

```{r}
# Print information about the scenario
myscenario

```

Several `set_*()` verbs are available to modify scenario properties, such as:

 - `set_tag()`, to assign a custom tag to identify the scenario
 - `set_init()`, to set the initial state
 - `set_param()`, to modify model parameters
 - `set_times()`, to define output time points
 - `set_forcings()`, to set environmental factors (i.e. *forcings*)
 - `set_exposure()`, to set an exposure time-series
 - `set_endpoints()`, to enable/disable assessment endpoints
 - `set_transfer()`, to define transfer intervals of *Lemna* fronds
 - `set_moa()`, to select a model's mode of action
 - `set_window()`, to control moving exposure windows

Please note that `set_exposure()` will, by default, use the exposure series' time points
also as output time points. If this behavior is undesired, it can be disabled
by setting the argument `reset_times=FALSE`:

```{r}
# Update the exposure time-series but keep former output time points
myscenario %>%
  set_noexposure()
```

Note that `times` still has the original four elements although the assigned exposure
series has only a single entry. Output time points can be
explicitly set with `set_times()`. For deriving effect endpoints, the simulated
time frame can be split up into moving exposure windows of a certain length using
`set_window()`. A more complex scenario setup may look like the following workflow:

```{r}
# Selected model parameters
myparam <- c(p_M=3211, v=0.023, k_J=0.63)
# Initial body length L
myinit <- c(L=0.02)
# Constant non-zero exposure
myexposure <- data.frame(time=0, conc=1.72)

DEB_abj() %>%
  set_param(myparam) %>%
  set_init(myinit) %>%
  set_exposure(myexposure, reset_times=FALSE) %>%
  set_times(0:10) %>%              # Output times 0,1,2,...,10
  set_mode_of_action(4) %>%        # Method of Action #4 to be activated
  set_window(length=3, interval=1) # Using moving exposure windows of length 3 days
```

## Running a simulation

After a scenario is fully set up, it can be passed to `simulate()` to return
model results:

```{r}
# Example scenario of the Lemna TKTD model
metsulfuron %>%
  set_times(0:7) %>%
  simulate()
```

`simulate()` returns the model state for each output time point. In this example,
it returns results for the period from zero to seven with an equi-distant time
step length of one.
The temporal resolution of results can be increased by specifying additional
output times:

```{r}
# Same simulation period, but with a smaller step length of 0.1
metsulfuron %>%
  set_times(seq(0, 7, 0.1)) %>%
  simulate() %>%
  tail()
```

The resulting table now contains ten times as many rows because we decreased
the step length by a factor of ten but simulated the same period. The temporal
resolution of simulations can have an influence on results, i.e. the numerical
values of the returned variables.
These differences usually originate from small numerical errors introduced by the solver
of the model’s Ordinary Differential Equations. Generally, parameters such as the
step-length in time can have influence on the precision of simulation results.

There are several ways to increase the precision of simulation results without
modifying a scenario's output times. One option is to decrease
the solver’s maximum step length in time by setting the optional argument `hmax`.
The smaller `hmax`, the more precise the results:

```{r}
# Original simulation period, but with a maximum solver step length of hmax=0.01
metsulfuron %>%
  set_times(0:7) %>%
  simulate(hmax=0.01)
```

Although very effective, reducing `hmax` to small values may lead to inefficiently
long runtimes of simulations. If numerical precision is an issue,
it is advisable to also test other solver settings such as absolute and relative
error tolerances (`atol` and `rtol`) or to switch to another solver method.
Numerical precision and stability is a complex and advanced topic, please
refer to the manual of `simulate()` for additional information:

```{r, eval=FALSE}
?cvasi::simulate
```

## Deriving assessment endpoints

Simulation results are the basis for deriving assessment endpoints such
as effect levels and effect profiles. The *cvasi* package implements routines
to derive these endpoints efficiently independent of the actual model that is being
assessed. Hence, if a model is available within the package framework, it can be
used to derive assessment endpoints. As an example, the following statement
derives effect levels for a sample *GUTS-RED-IT* scenario:

```{r}
# GUTS-RED-IT scenario of the fathead minnow and chlorpyrifos
minnow_it %>% effect()
```
```{r, include=FALSE}
# make sure that value in text are still up to date
testthat::expect_equal(minnow_it %>% effect() %>% dplyr::pull(L), 6.297e-5, tolerance=0.001)
```
The lethality endpoint `L` has a value of `6.3e-5` which means that lethality
has only marginally increased by 0.0063% compared to a control scenario without
exposure to the contaminant.
The columns `L.dat.start` and `L.dat.end` denote the period for which the
effect level was observed. In case of *GUTS-RED* models, this usually refers to the
whole simulation period. For certain assessments it may be necessary to evaluate
all exposure periods of a certain length within the original series separately.
This approach is referred to as a *moving exposure window*:

```{r}
# Setting up a custom scenario with a total simulation period of 14 days and
# an exposure window length of 7 to assess a trapezoidal exposure pattern
americamysis %>%
  set_window(7) %>%
  set_exposure(data.frame(t=c(0,3,4,7,8), c=c(0,0,3,3,0))) %>%
  set_times(0:14) -> mydeb

# Derive maximum effect level of all exposure windows
mydeb %>% effect()
```
```{r, include=FALSE}
# make sure that value in text are still up to date
testthat::expect_equal(mydeb %>% effect() %>% dplyr::pull(L), 0.0521, tolerance=0.001)
```

The `americamysis` scenario provides two effect endpoints: structural length (`L`)
and reproduction (`R`). Exposure to the simulated toxicant results at maximum in a 5.21%
decrease in body length compared to control scenarios. The
maximum effect was observed in the exposure window starting at time `2.0` and
ending at `9.0`, cf. columns `L.dat.start` and `L.dat.end`. The toxicant had no effect
on reproduction due to the simulation being too short for reproduction to occur.

Scenarios can also be assessed regarding to so-called effect profiles, or *EPx* values for
short. An *EPx* value quantifies the multiplication factor that is necessary to
produce an effect level of x% for that particular scenario. As an example,
an *EP50* of `17.0` would mean that the scenario's exposure series needs to be
multiplied by a factor of 17 to observe a 50% effect. The function
`epx()` can derive effect profiles efficiently with arbitrary precision:

```{r}
# Restrict assessed endpoints to structural length (L)
mydeb %>% 
  set_endpoints("L") %>%
  epx()
```

```{r, include=FALSE}
# make sure that value in text are still up to date
testthat::expect_equal(mydeb %>%  set_endpoints("L") %>% epx(ep_only=TRUE) %>% unlist(use.names=FALSE), c(1.162598, 1.711914), ignore_attr=TRUE, tolerance=0.01)
```

In this example, a factor of `1.16` resulted in 10% effect on structural
length and a factor of `1.71` in 50% effect. `epx()` provides means to control the precision
of derived *EPx* values, the range of tested multiplication factors, factor
cutoff thresholds, and debugging capabilities. To examine how a particular
*EPx* value was derived, simply set the argument `verbose=TRUE`:

```{r}
# Examine how the EP20 value is derived
minnow_it %>% epx(level=20, verbose=TRUE)
```

In this example, the algorithm of `epx()` first tests a multiplication factor of
`10` and then continues testing additional factors using a binary search algorithm
until a factor is found that causes 20% effect.

__Please note__: Depending on model
and chosen parameters, it may be impossible to calculate *EPx* values (for selected
effect levels), because the model or organism cannot reach the required state.
As an example: If the simulated period of a *DEB_abj* scenario is too short,
reproduction cannot occur. Therefore, effects on reproduction will always be
zero independent of the exposure level. Consequently, *EPx* values for the
reproduction endpoint will not be available.





# Data format of model inputs

Input data for the *cvasi* package relies on common *R* data structures and
imposes only minimal requirements to set up scenarios:

1. **Data representing numerical values should be represented by
   numerical types**. Practically, this means that numerical scenario properties
   such as output time points must not be passed as string literals even if
   the string may be representing a number. Due to *R* being a dynamically typed
   language, parameter values and their types may have to be checked on import by
   the user.
1. **Time-series data must be represented by a `data.frame` with exactly
   two numerical columns**: The first column for time, the second
   for the series' value. The column names are irrelevant but sensible names
   may help documenting the data. Constant values can be represented by
   time-series with only a single row or entry.

Therefore, it should be possible to apply the package's routines on existing
datasets with minimal effort. Please refer to Wickham and Grolemund (2017)
on how to implement plain and consistent processes to import and transform data
using *tidyr* syntax.

*cvasi* does support loading data from common sources, such as the exposure model
*TOXSWA*, to ease data processing. Notable file formats and data structures
that are supported are exposure time-series from the model FOCUS *TOXSWA*, as
well as fitted *GUTS-RED* model parameter sets from
[*morse*](https://cran.r-project.org/package=morse).



# How to derive model inputs

In general, all models require parameter values specific to the simulated
species and substance. Some scenario types contain default values for selected model
parameters as recommended by model authors. Scenarios providing default values
include:

- `Lemna_Schmitt()` and derived models are parameterized according to Schmitt
   *et al.* (2013)
- `Lemna_SETAC()`, `Magma()`, and derived models are parameterized as recommended
   by Klein *et al.* (2021)

In addition, some models require values for environmental factors such as
temperature and nutrient concentrations. These values or time-series are specific
to a certain scenario and must be chosen accordingly.

Parameters of TKTD sub-models are substance-specific and are usually obtained
by fitting model parameters to observed data. The *cvasi* package supports this process with
the function `calibrate()` which uses a frequentist approach for parameter
fitting. However, parameters can also be derived by other means. Generally,
the user is responsible for choosing model parameters and the package does
not impose any restrictions in this regard.

The package function `calibrate()` supports the following operation modes:

- Fitting of a single parameter or multiple parameters simultaneously
- Fit to one or more independent datasets
- Weighting of datasets
- Choosing the error function which is minimized
- Selecting one of multiple optimization algorithms

Parameter fitting builds upon the *R* function `stats::optim()` and it exposes full
control over the fitting process. In this context, a dataset is represented by a
*calibration set* object which consists of a scenario, observed data, and an
optional weighting term.
Please refer to the [Modeling Howto](cvasi-2-howto.html) for a description of the
fitting process using `calibrate()`.

As an alternative for *GUTS-RED* type models, model parameters can also be
obtained  using the [*morse* package](https://cran.r-project.org/package=morse).
It uses a Bayesian approach for parameter fitting.

# Description of model outputs

The *cvasi* package does not use or define any custom file formats for model outputs
but strictly relies on common *R* data structures. Simulation results and
assessment endpoints are returned in tabular formats which can easily be
serialized to common file formats such as `.csv` or *Microsoft Excel* files
for further processing and data storage.

The following subsections will describe in detail the return values of common
package routines for simulating and assessing scenarios.



## Simulation results

The return value of `simulate()` is a `data.frame` containing the value of each
state variable at each requested output time. The first column represents *time*,
followed by columns for each state variable:

```{r}
## Display selected scenario properties
metsulfuron@times  # Output times
metsulfuron@init   # Initial state

## Simulate the sample scenario
metsulfuron %>% simulate()
```

The model state in the first row is, by definition, equal to the initial state
of the scenario. The simulation result may contain additional columns that
are derived variables or intermediate model variables. In case of the
`metsulfuron` scenario, the internal toxicant concentration (`C_int`) was derived
from `BM`, `M_int`, and model parameters. The number of fronds (`FrondNo`)
was derived from `BM` and model parameters. Additional outputs such as `C_int`
and `FrondNo` are included in results for user convenience and to support model
understanding. Additional outputs can be disabled by setting the optional 
argument `nout=0`:

```{r}
metsulfuron %>%
  simulate(nout=0) %>%
  head(5)
```

Intermediate model variables can be enabled in the same manner:

```{r}
metsulfuron %>%
  simulate(nout=8) %>%
  head(5)
```

## Effect levels

The return value of `effect()` is a `data.frame` containing the effect level of
each selected endpoint. The first column, `scenario`, contains the original 
scenario which was the basis of the assessment. Followed by the effect level and
the exposure window in which the maximum effect was observed:

```{r}
metsulfuron %>%
  set_times(0:7) %>%  # restrict scenario to the period [0,7]
  effect()
```

Effect levels are reported as a fraction relative to a control scenario. Effects
can be converted to percentages by multiplying the value with `100` (%). For
scenarios with moving exposure windows disabled, the reported window will cover
the entire simulation period. If multiple windows result in identical maximum
effect levels, the first matching window will be reported.
The table's content can be reduced to endpoints only by setting the
argument `ep_only=TRUE`.

`effect()` can also be instructed to return the effect levels of all assessed
exposure windows by setting the argument `max_only=FALSE`:

```{r}
metsulfuron %>%
  set_window(length=7, interval=1) %>%   # enable moving exposure windows
  effect(max_only=FALSE)                 # return effects of all windows
```

In few cases, `effect()` may report spurious non-zero effect levels that are
very close to zero. These can either be avoided by increasing the numerical
precision of simulations or by setting a marginal effect threshold which instructs
`effect()` to zero any effect below that threshold. A warning will be shown
if a marginal effect threshold greater than 1% is chosen. Applying a threshold
of 1% to the previous example results in zeroing the effect in the last and
second to last exposure window but leaving the remaining results unchanged:

```{r}
metsulfuron %>%
  set_window(length=7, interval=1) %>%         # enable moving exposure windows
  effect(max_only=FALSE, marginal_effect=0.01) # return effects of all windows
```




## Effect profiles

The return value of `epx()` is a `data.frame` containing the effect profile (*EPx*) of
each selected endpoint and effect level. The first column, `scenario`, contains 
the original  scenario which was the basis of the assessment. Followed by the
derived effect profiles, i.e. the multiplication factors to achieve x% effect:

```{r}
metsulfuron %>% epx()
```

For the sample `metsulfuron` scenario, a factor of `0.395` would need to be applied
to the exposure series to achieve an effect level of 10% on biomass (*BM*). An
effect level of 50% on biomass is unattainable with the tested factors ranging
from `1e-30` to `1e30` which is demonstrated by the warning message and the
`NA` value for `BM.EP50`.

The table's content can be reduced to endpoints only by setting the
argument `ep_only=TRUE`. The range of tested factors can be controlled by
setting the arguments `min_factor` and `max_factor`. Multiplication factors
exceeding this range will raise a warning and `NA` will be returned for
that *EPx*. Finding multiplication factors can be cut short if a
certain margin of safety is reached by using the argument `factor_cutoff`:
if the algorithm identifies that an *EPx* greater or equal to `factor_cutoff` is required,
then `factor_cutoff` is reported as the result. This approach may
be used to save computational resources for long running models if the
exact *EPx* value above the cutoff is not of interest.

The precision of *EPx* values is controlled by the argument `effect_tolerance`
and is given as the upper absolute error threshold of effects that
is deemed acceptable. The default value of `0.001` ensures that a derived
*EPx* will result in an effect of x% ± 0.1. Decreasing the `effect_tolerance`
will result in additional model iterations and longer runtime.

The algorithm's inner workings can be examined by setting the argument `verbose=TRUE`:

```{r}
metsulfuron %>%
  set_endpoints("BM") %>%
  epx(verbose=TRUE)
```

In this example, the binary search finds a `BM.EP10` with acceptable precision in
less than ten iterations. For the `BM.EP50`, the algorithm re-uses the
knowledge gained from previously performed calculations but is ultimately unable
to find a suitable multiplication factor within the allowed range of multiplication
factors.



# References
- EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
  Ockleford C., et al., 2018: *Scientific Opinion on the
  state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
  regulatory risk assessment of pesticides for aquatic organisms*. EFSA Journal 2018;
  16(8):5377, 188 pp. DOI: [10.2903/j.efsa.2018.5377](https://doi.org/10.2903/j.efsa.2018.5377)
- Geiger D.L., et al., 1988: *Acute toxicities of organic
  chemicals to fathead minnows (Pimephales promelas): Volume IV*, pp. 195-197.
  University of Wisconsin-Superior, Center for Lake Superior Environmental Studies.
  ISBN 9780961496838.
- Klein J., et al., 2021: *Refined description of the Lemna TKTD growth model
  based on Schmitt et al. (2013) - equation system and default parameters*.
  Report of the working group *Lemna* of the SETAC Europe Interest Group Effect
  Modeling. Version 1.1, uploaded on 09 May 2022.
  https://www.setac.org/group/effect-modeling.html
- Schmitt W., et al., 2013: *Mechanistic TK/TD-model
  simulating the effect of growth inhibitors on Lemna populations*. Ecol Model 255,
  pp. 1-10. DOI: [10.1016/j.ecolmodel.2013.01.017](https://doi.org/10.1016/j.ecolmodel.2013.01.017)
- Weber D., et al., 2012: *Combination of a higher-tier flow-through system and
  population modeling to assess the effects of time-variable exposure of
  isoproturon on the green algae Desmodesmus subspictatus and Pseudokirchneriella
  subcapitata*. Environ Toxicol Chem 31(4), pp. 899-908. DOI: [10.1002/etc.1765](https://doi.org/10.1002/etc.1765)
- Wickham H. and Grolemund G., 2017: *R for Data Science: Import, Tidy, Transform,
  Visualize, and Model Data* (1st ed.). O'Reilly Media. https://r4ds.had.co.nz

# scaled Ci, M=Ci

    WAoAAAACAAQDAAACAwAAAAMTAAAABQAAAA4AAAApAAAAAAAAAAA/uZmZmZmZmj/JmZmZmZma
    P9MzMzMzMzQ/2ZmZmZmZmj/gAAAAAAAAP+MzMzMzMzQ/5mZmZmZmZz/pmZmZmZmaP+zMzMzM
    zM0/8AAAAAAAAD/xmZmZmZmaP/MzMzMzMzQ/9MzMzMzMzT/2ZmZmZmZnP/gAAAAAAAA/+ZmZ
    mZmZmj/7MzMzMzM0P/zMzMzMzM0//mZmZmZmZ0AAAAAAAAAAQADMzMzMzM1AAZmZmZmZmkAC
    ZmZmZmZnQAMzMzMzMzRABAAAAAAAAEAEzMzMzMzNQAWZmZmZmZpABmZmZmZmZ0AHMzMzMzM0
    QAgAAAAAAABACMzMzMzMzUAJmZmZmZmaQApmZmZmZmdACzMzMzMzNEAMAAAAAAAAQAzMzMzM
    zM1ADZmZmZmZmkAOZmZmZmZnQA8zMzMzMzRAEAAAAAAAAAAAAA4AAAApAAAAAAAAAABANAyu
    0zjw50A/lRzdcWYwQEMbtP83YWdARQQoIjRnbUBGHR+kHo/gQEa+veTN9odARxu1QXaZ+UBH
    US8pZOwAQEdv8eQEwVRAR4GjmsQ1UUBHi9ExkBEiQEeRq/wN795AR5UKHxh3/0BHlvoKivxr
    QEeYF04NfgNAR5i7ZQzoUEBHmRnIYYYoQEeZUBOmRjhAR5lvTs7xw0BHmYFFy8yRQEeZi5s7
    ostAR5mRjPH8aEBHmZT4RFiBQEeZlu/FWV1AR5mYEWWvC0BHmZi3/xpsQEeZmRfT9/RAR5mZ
    TvPNMUBHmZluqTtlQEeZmYDmjZNAR5mZi2RydEBHmZmRbW5lQEeZmZTmI7FAR5mZluVX8kBH
    mZmYC2YvQEeZmZi0i9pAR5mZmRXX50BHmZmZTc+OQEeZmZluASBAR5mZmYCF2gAAAA4AAAAp
    f/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAA
    B6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAA
    AAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4
    AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAei
    f/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAA
    B6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAA
    AAAHogAAAA4AAAApAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD8k
    nXtAp5AiP4N49AmChas/nXENXbNzKT+rfK7vUcovP7T28LyVXfc/vKrvh8s9Uj/CUvSmW12q
    P8Zk27PrQpE/yoKA9eHynz/OpudzrtrqP9FnmFFiVwM/03za/oEwFz/VksI7dWM6P9epCCFA
    Eiw/2b+EekrrpT/b1iAlfhJqP93szdToxWM/4AHC8K+z7D/hDSHx9n1EP+IYgqoov7o/4yPk
    XtTmeT/kL0aku9v+P+U6qT4s3rU/5kYMB6uZMT/nUW7szopDP+hc0eHX30Y/6Wg04AaUnT/q
    c5fjeBlZP+t++unwVOg/7Ipd8iYyWD/tlcD7XGYgP+6hJAUmDU0/76yHD0SFmD/wW/UMyePf
    P/DhppH/jSc/8WdYFz1IwwAAAA4AAAApP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAA
    AAA/8AAAAAAAAD/v/rYu789BP++yeq53ERU/7xfSJImoUD/uU83aZNYMP+173gyhdSQ/7Jwi
    EYXi5z/ru1Zi6+6XP+rdJ84HcNM/6gOKPAYwpT/pL3s2xQwgP+hhb19jd9o/55mPuyWGnj/m
    19vtYrWJP+YcPUl61S0/5WaRbfbi1j/ktrAnSEUKP+QMbrFAkXE/42ehgaXFlD/iyB1CKZrf
    P+Itt1ZiByE/4ZhGIVYVdT/hB6EnVd2KP+B7oRwZOBo/3+g/zBQVOD/e4fE3WTE7P93kDvvL
    fhE/3O5T4UunNj/cAHzloSInP9saSSuy6JM/2jt56qvnRT/ZY9JdUetRP9iTF7G0ek4/18kQ
    +UaAYD/XBYcZa4AyP9ZIRLx7THA/1ZEWQzw8nQAABAIAAAABAAQACQAAAAVuYW1lcwAAABAA
    AAAFAAQACQAAAAR0aW1lAAQACQAAAAJDaQAEAAkAAAABRAAEAAkAAAABSAAEAAkAAAABUwAA
    BAIAAAABAAQACQAAAAlyb3cubmFtZXMAAAANAAAAAoAAAAD////XAAAEAgAAAAEABAAJAAAA
    DWRlc29sdmVfZGlhZ24AAAITAAAABQAAAA0AAAACAAAAKQAAAAQAAAANAAAAFQAAAAAAAAGR
    AAAJaIAAAACAAAAAgAAAAIAAAACAAAAAgAAAAIAAAACAAAAAgAAAAAAAAACAAAAAgAAAAIAA
    AACAAAAAAAAABIAAAACAAAAAgAAAAAAAABMAAAACAAAA/gAAABAAAAAEAAQACQAAAAR0aW1l
    AAQACQAAAAJDaQAEAAkAAAABRAAEAAkAAAABSAAAABAAAAACAAQACQAAAAdkZVNvbHZlAAQA
    CQAAAAZtYXRyaXgAAAAQAAAAAQAEAAkAAAACcmsAAAQCAAAB/wAAABAAAAAFAAQACQAAAANk
    aW0ABAAJAAAABmlzdGF0ZQAEAAkAAAAIZGltbmFtZXMABAAJAAAABWNsYXNzAAQACQAAAAR0
    eXBlAAAA/gAABAIAAAABAAQACQAAAAVjbGFzcwAAABAAAAACAAQACQAAAA5jdmFzaV9zaW11
    bGF0ZQAEAAkAAAAKZGF0YS5mcmFtZQAABAIAAAABAAQACQAAAAxjdmFzaV9zdGF0dXMAAAAQ
    AAAAAQAEAAkAAAAHc3VjY2VzcwAABAIAAAABAAQACQAAAA5kZXNvbHZlX291dHB1dAAAABAA
    AAAAAAAEAgAAAAEABAAJAAAADWRlc29sdmVfY29uZHMAAAATAAAAAAAAAP4=

---

    WAoAAAACAAQDAAACAwAAAAMTAAAABQAAAA4AAAApAAAAAAAAAAA/uZmZmZmZmj/JmZmZmZma
    P9MzMzMzMzQ/2ZmZmZmZmj/gAAAAAAAAP+MzMzMzMzQ/5mZmZmZmZz/pmZmZmZmaP+zMzMzM
    zM0/8AAAAAAAAD/xmZmZmZmaP/MzMzMzMzQ/9MzMzMzMzT/2ZmZmZmZnP/gAAAAAAAA/+ZmZ
    mZmZmj/7MzMzMzM0P/zMzMzMzM0//mZmZmZmZ0AAAAAAAAAAQADMzMzMzM1AAZmZmZmZmkAC
    ZmZmZmZnQAMzMzMzMzRABAAAAAAAAEAEzMzMzMzNQAWZmZmZmZpABmZmZmZmZ0AHMzMzMzM0
    QAgAAAAAAABACMzMzMzMzUAJmZmZmZmaQApmZmZmZmdACzMzMzMzNEAMAAAAAAAAQAzMzMzM
    zM1ADZmZmZmZmkAOZmZmZmZnQA8zMzMzMzRAEAAAAAAAAAAAAA4AAAApAAAAAAAAAABAQRN6
    3R4OQkBK5gym6Cy7QFBGPfNmfUNAUeZAjfIOU0BS1YzYy+2EQFNfMxZlp6lAU65g4g9RPEBT
    2+yBoM1VQFP2H199ipxAVAUxTqlHcUBUDdx5ZhdEQFQS2PyjvARAVBW3Q7RMKUBUF12jA2P/
    QFQYUJgsCJxAVBjcWVGAbkBUGSy9GqdnQFQZWvsKXFhAVBl1lHoEI0BUGYThaTxTQFQZja6E
    JcNAVBmSvo0DcUBUGZWoDtHhQFQZl1TjvgFAVBmYS5AYVEBUGZjZdGODQFQZmSsS58pAVBmZ
    WgXhvkBUGZl1B3TCQFQZmYSQSwVAVBmZjX/a+EBUGZmSo7XkQFQZmZWYnmdAVBmZl0wCOkBU
    GZmYRnRRQFQZmZjWhCBAVBmZmSliLkBUGZmZWQz4QFQZmZl0eEVAVBmZmYQ97QAAAA4AAAAp
    f/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAA
    B6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAA
    AAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4
    AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAei
    f/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAA
    B6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAAAAAHon/4AAAAAAeif/gAAAAAB6J/+AAA
    AAAHogAAAA4AAAApAAAAAAAAAAAAAAAAAAAAAD+egGkp1KTmP8ioQ1MVByo/3GrVJ7792j/n
    4flHDOuHP/E/Qj6RKwM/9tKNpx9U6j/8jY0c2ARPQAEvsZO1z5JABB8uIzKSP0AHEnH0uRE8
    QAoH4il7KSJADP6Sag4p6EAP9frDk3+XQBF25oFXaqdAEvLuFedo60AUbwcveV8CQBXrKlzq
    2D9AF2dTVlv2jkAY43+lbdjDQBpfrd+GL8BAG9vdNBFJ2kAdWA0rFMxzQB7UPX+NKJJAICg3
    BOPNwkAg5k9ZdydJQCGkZ7bvTVZAImKAGYUdUkAjIJh/DEYcQCPesOZExkZAJJzJTnaK30Al
    WuG3N7HAQCYY+iBLUuFAJtcSiY5lX0AnlSry7MIjQChTQ1xa0Y9AKRFbxdHom0Apz3QvTjFY
    QCqNjJjNdvJAK0ulAk50hAAAAA4AAAApP/AAAAAAAAA/8AAAAAAAAD/vD5Xkf551P+pkn0f8
    g8U/5IbG2ZlTHT/eV6wlzQkeP9XHYlUhn7A/zr3rVt02CT/FfM19SPVGP73e/J/6Aqw/tLI9
    tm+erz+soEbQOmoMP6PGr/ppH3o/m07GIKVf9D+S2OJmUu14P4oCuKcKeUs/gfI/J27Vsj94
    w0NG/4rqP3EVX3aZ9Yc/Z5IhQSQuCT9gQriE5rTVP1ZvXnq7ZaU/TvQuHdtR0D9FWoUDnRvM
    Pz12Lff9Plk/NFL74kQlbj8sCpG7Y3i8PyNYJKm0YsA/GrB7G+Pb1T8SaWSGKUy8PwlnE1BD
    +p8/AYYm1RvUVT74LYznwqhlPvCt3bfxJQg+5wMkB9D/Sz7fwAQNfMyvPtXnHi2Dp9Q+zjgm
    JxojOj7E2MnL1G/6PrzDLMLIJ6g+s9d94pJXAwAABAIAAAABAAQACQAAAAVuYW1lcwAAABAA
    AAAFAAQACQAAAAR0aW1lAAQACQAAAAJDaQAEAAkAAAABRAAEAAkAAAABSAAEAAkAAAABUwAA
    BAIAAAABAAQACQAAAAlyb3cubmFtZXMAAAANAAAAAoAAAAD////XAAAEAgAAAAEABAAJAAAA
    DWRlc29sdmVfZGlhZ24AAAITAAAABQAAAA0AAAACAAAAKQAAAAQAAAANAAAAFQAAAAAAAAGR
    AAAJaIAAAACAAAAAgAAAAIAAAACAAAAAgAAAAIAAAACAAAAAgAAAAAAAAACAAAAAgAAAAIAA
    AACAAAAAAAAABIAAAACAAAAAgAAAAAAAABMAAAACAAAA/gAAABAAAAAEAAQACQAAAAR0aW1l
    AAQACQAAAAJDaQAEAAkAAAABRAAEAAkAAAABSAAAABAAAAACAAQACQAAAAdkZVNvbHZlAAQA
    CQAAAAZtYXRyaXgAAAAQAAAAAQAEAAkAAAACcmsAAAQCAAAB/wAAABAAAAAFAAQACQAAAANk
    aW0ABAAJAAAABmlzdGF0ZQAEAAkAAAAIZGltbmFtZXMABAAJAAAABWNsYXNzAAQACQAAAAR0
    eXBlAAAA/gAABAIAAAABAAQACQAAAAVjbGFzcwAAABAAAAACAAQACQAAAA5jdmFzaV9zaW11
    bGF0ZQAEAAkAAAAKZGF0YS5mcmFtZQAABAIAAAABAAQACQAAAAxjdmFzaV9zdGF0dXMAAAAQ
    AAAAAQAEAAkAAAAHc3VjY2VzcwAABAIAAAABAAQACQAAAA5kZXNvbHZlX291dHB1dAAAABAA
    AAAAAAAEAgAAAAEABAAJAAAADWRlc29sdmVfY29uZHMAAAATAAAAAAAAAP4=

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.r
\name{plot_sd}
\alias{plot_sd}
\title{Creates plot of model results (uncertainties optional)}
\usage{
plot_sd(
  model_base,
  treatments,
  rs_mean,
  rs_range = NULL,
  obs_mean = NULL,
  obs_full = NULL,
  x_breaks = NULL,
  y_lim = NULL,
  grid_labels = NULL,
  grid_ncol = 2,
  plot_col = 2,
  y_title = NULL,
  ...
)
}
\arguments{
\item{model_base}{effect scenario object with mean parameters}

\item{treatments}{treatments exposure levels as data frame}

\item{rs_mean}{\code{data.frame}, model results best fit params}

\item{rs_range}{\code{data.frame}, uncertainties as data frame}

\item{obs_mean}{\code{data.frame}, observation data with means per treatment level}

\item{obs_full}{\code{data.frame}, full set including results for replicates}

\item{x_breaks}{optional vector of breaks of x-axis}

\item{y_lim}{optional vector containing limits of y-axis}

\item{grid_labels}{optional labels of grid headers}

\item{grid_ncol}{optional number of grid columns}

\item{plot_col}{output column which should be plotted}

\item{y_title}{optional title of y-axis}

\item{...}{any additional parameters}
}
\value{
a ggplot2 plot object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\details{
All parameter combinations and exposure patterns are simulated and the mean
of predictions is derived for a single study. The uncertainty is
passed to the function due to computation time. Results are displayed by
plotting the time series including the uncertainty interval. Observation data
can be optionally displayed. Data should be provided in long format. Function
plots the time (column 1) and the predictions (column 2, can be changed by
the user plot_col)
}
\examples{
set.seed(124)
exposure <- data.frame(
  time = 0:21,
  conc = rnorm(n = 22, mean = 0.1, sd = 0.06),
  trial = "T1"
)
forcings <- list(temp = 12, rad = 15000)
param <- list(EC50 = 0.3, b = 4.16, P_up = 0.0054)
inits <- list(BM = 0.0012, E = 1, M_int = 0)

scenario <- Lemna_Schmitt() \%>\%
  set_forcings(forcings) \%>\%
  set_param(param) \%>\%
  set_init(inits)

sim_result <- simulate_batch(
  model_base = scenario,
  treatments = exposure,
  param_sample = NULL
)

plot_sd(
  model_base = scenario,
  treatments = exposure,
  rs_mean = sim_result
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_bounds.R
\name{set_bounds}
\alias{set_bounds}
\alias{set_bounds,EffectScenario,list-method}
\alias{set_bounds,CalibrationSet,list-method}
\alias{set_bounds,list,list-method}
\alias{set_bounds,ScenarioSequence,list-method}
\title{Set boundaries of model parameters}
\usage{
set_bounds(x, bounds)

\S4method{set_bounds}{EffectScenario,list}(x, bounds)

\S4method{set_bounds}{CalibrationSet,list}(x, bounds)

\S4method{set_bounds}{list,list}(x, bounds)

\S4method{set_bounds}{ScenarioSequence,list}(x, bounds)
}
\arguments{
\item{x}{vector of \link{scenario} or \link{caliset} objects}

\item{bounds}{named list of numerical vectors, where the first level lists the parameters
by name, and the second level lists the lower and upper boundary}
}
\value{
\link{scenario} or \link{caliset} with modified parameter boundaries
}
\description{
Modifies the boundaries of model parameters for one or more \link{scenario} or
\link{caliset} objects.
}
\examples{
metsulfuron \%>\%
   set_bounds(list(k_phot_max = c(0, 30),
                   k_resp = c(0, 10)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-ParameterSet.R
\name{parameter_set}
\alias{parameter_set}
\alias{ParameterSet-class}
\alias{parameter_set-class}
\title{Set of model parameters}
\usage{
parameter_set(model, param = list(), tag = NA_character_)
}
\arguments{
\item{model}{\code{character}, a string containing a model name, e.g. \code{"GUTS-RED-IT"}}

\item{param}{named \code{list} of model parameters}

\item{tag}{\code{character}, an optional identifier}
}
\value{
an S4 object of type \linkS4class{ParameterSet}
}
\description{
Set of model parameters
}
\section{Slots}{

\describe{
\item{\code{model}}{\code{character}, a string containing a model name, e.g. \code{"GUTS-RED-IT"}}

\item{\code{tag}}{\code{character}, an optional identifier}

\item{\code{param}}{named \code{list} of model parameters}
}}

\examples{
# create a parameter set and assign it
ps <- parameter_set("GUTS-RED-IT", list(kd=0.12, hb=0.3))
GUTS_RED_IT() \%>\% set_param(ps)

# multiple scenarios can be modified at once
c(GUTS_RED_IT(), GUTS_RED_IT()) \%>\%
  set_param(ps)

# model names must match, otherwise an error will be raised
try(GUTS_RED_SD() \%>\% set_param(ps))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{focusd1}
\alias{focusd1}
\title{A Lemna_SETAC scenario with variable environment}
\format{
An object of class \code{LemnaSetac} of length 1.
}
\usage{
focusd1
}
\description{
A mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and growth
model for the aquatic macrophytes Lemna spp. as published by
Klein \emph{et al.} (2021).
}
\details{
The scenario will simulate a period of 365 days, a start
population of 80 g/m² dry weight, variable environmental conditions, and a
complex, time-varying exposure pattern.

The scenario setup was published by Hommen \emph{et al}. (2015). Exposure pattern
and substance specific parameters are of exemplary character
and represent the herbicide \emph{metsulfuron-methyl}. The parameters were
derived by Schmitt et al. (2013) based on literature
data.
}
\examples{
# Simulate the example scenario
focusd1 \%>\% simulate()

}
\references{
Hommen U., Schmitt W., Heine S., Brock Theo CM., Duquesne S., Manson P., Meregalli G.,
Ochoa-Acuña H., van Vliet P., Arts G., 2015: How TK-TD and Population Models for
Aquatic Macrophytes Could Support the Risk Assessment for Plant Protection
Products. Integr Environ Assess Manag 12(1), pp. 82-95.
\doi{10.1002/ieam.1715}

Klein J., Cedergreen N., Heine S., Reichenberger S., Rendal C.,
Schmitt W., Hommen U., 2021: Refined description of the \emph{Lemna} TKTD growth model
based on \emph{Schmitt et al.} (2013) - equation system and default parameters.
Report of the working group \emph{Lemna} of the SETAC Europe Interest Group Effect
Modeling. Version 1.1, uploaded on 09 May 2022.
\url{https://www.setac.org/group/effect-modeling.html}

Schmitt W., Bruns E., Dollinger M., Sowig P., 2013: Mechanistic TK/TD-model
simulating the effect of growth inhibitors on \emph{Lemna} populations. Ecol Model
255, pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
}
\seealso{
\link{Lemna-models}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence.R
\name{sequence.extract}
\alias{sequence.extract}
\alias{[[<-,ScenarioSequence,numeric,ANY,ANY-method}
\alias{[,ScenarioSequence,numeric,missing,missing-method}
\alias{[[,ScenarioSequence,numeric-method}
\alias{[[<-,ScenarioSequence,numeric,missing,EffectScenario-method}
\alias{length,ScenarioSequence-method}
\title{Extract and replace elements of a sequence}
\usage{
\S4method{[}{ScenarioSequence,numeric,missing,missing}(x, i)

\S4method{[[}{ScenarioSequence,numeric}(x, i)

\S4method{[[}{ScenarioSequence,numeric,missing,EffectScenario}(x, i, j) <- value

\S4method{length}{ScenarioSequence}(x)
}
\arguments{
\item{x}{\link{sequence}}

\item{i}{index of elements to extract or replace}

\item{j}{\emph{not used}}

\item{value}{new scenario}
}
\value{
various
}
\description{
The array accessor generics allow extracting and replacing scenarios within
am existing sequence. \code{[} and \code{[[} work identical to
}
\section{Functions}{
\itemize{
\item \code{x[i}: Returns a list of scenarios from the sequence.

\item \code{x[[i}: Returns a single scenario from the sequence.

\item \code{`[[`(x = ScenarioSequence, i = numeric, j = missing) <- value}: Replaces a single scenario in the sequence.

\item \code{length(ScenarioSequence)}: Returns the number of scenarios in the sequence.

}}
\examples{
# create a sequence
seq <- sequence(list(minnow_it, minnow_it), breaks=3)

seq[1]       # first element, as a list of scenarios
seq[c(1)]    # the same
seq[c(1, 2)] # both elements as a list of scenarios
seq[[1]]     # first element as a scenario

# replacing single elements
seq[[1]] <- minnow_sd \%>\% set_times(1:3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_LemnaThreshold}
\alias{is_LemnaThreshold}
\title{Test if argument is a LemnaThreshold scenario}
\usage{
is_LemnaThreshold(x)
}
\arguments{
\item{x}{vector of \link{scenarios} objects}
}
\value{
vector of \code{logical} values
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
}
\seealso{
\code{\link[=is_lemna]{is_lemna()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-EffectScenario.R
\name{Scenarios}
\alias{Scenarios}
\alias{scenario}
\alias{scenarios}
\alias{EffectScenario-class}
\title{Effect scenario classes}
\description{
The \code{EffectScenario} class is the base for all of the basic scenario
types and models. It contains slots for data and settings that are
required by most models such as a vector of model parameters and a vector
of initial states. For each particular model, the class's slots are
filled with certain default or fixed values. Some models derive from this
class and add slots to store additional data.
}
\details{
Certain behaviors that are required to model complex processes cannot be
represented by a single EffectScenario. As an example, the parameters of a scenario
are generally fixed during the simulated time period. In order to represent a
change in parameter values, the original scenario would
need to split into two scenarios \emph{A} and \emph{B} which differ by parameter values
and simulated time period. By combining these scenarios to a \emph{scenario sequence},
the sequence would be treated as a single, complex scenario. See
\code{\link[=sequence]{sequence()}} for more information.
\subsection{Parameters}{

Most parameters are represented by numerical types but other types
are possible depending on model. Please refer to the model description
which parameters are required and in which unit. Some or all parameters may
be required to start a simulation. If required parameters are missing,
simulation will fail with an error message.
}

\subsection{Initial state}{

The \emph{initial state} represents the starting values of state variables when
starting a simulation. A scenario's default initial state may be insufficient
to get sensible results. It is advisable to set an initial state explicitly
when creating a new scenario, see \code{\link[=set_init]{set_init()}}.

In theory, a scenario's state variables can be renamed by modifying the names
of the initial state vector. However, this is strongly
discouraged as this will affect other routines such as \code{\link[=effect]{effect()}} and \code{\link[=epx]{epx()}}
and may render results useless.
}

\subsection{Exposure}{

\emph{Exposure} refers to the concentration of toxicant an organism is exposed to.
In case of aquatic organisms, this would commonly be the concentration of a
toxicant in water. Other interpretations are possible depending on model
assumptions.

Exposure time-series are generally represented by a \code{data.frame} containing two
columns. The first column representing time, the second representing the
exposure level. The ordering of columns is mandatory. The column names
are essentially irrelevant but sensible names may help documenting the
scenario and its data. The rows must be ordered chronologically. A time-series
can consist of only a single row; in this case it will represent constant
exposure. Exposure time-series are set to a scenario using \code{\link[=set_exposure]{set_exposure()}}.

Handling time-series is a costly task for the ODE solver due to consistency
checks and interpolation between time steps. How the solver interpolates
the time-series can be controlled by certain arguments to functions
such as \code{\link[=simulate]{simulate()}} and \code{\link[=effect]{effect()}}. Please refer to \code{\link[=simulate]{simulate()}} for a brief
overview and \link[deSolve:forcings]{deSolve::forcings} for a detailed description.

Exposure time-series should be kept as short as possible and as complex as
needed for optimal computational efficiency.
}

\subsection{Environmental forcings}{

\emph{Forcings} generally refer to model parameters that change over time as part
of an external function such as environmental temperature and exposure levels.
Due to the importance of exposure in regulatory assessments, this R package
explicitly distinguishes between environmental forcings and exposure. However,
the same restrictions and features apply to both of them.

Forcing time-series are handled the same way as exposure time-series, i.e.
they are represented by a \code{data.frame} containing two columns. The first column
representing time, the second representing the parameter that is a function of
time. The ordering of columns is mandatory. The rows must be ordered
chronologically. Forcings time-series are set using \code{\link[=set_forcings]{set_forcings()}}.
Please refer to the \emph{Exposure} section for more information on how time-series
are handled.
}

\subsection{Output times}{

A scenario's simulated time period is defined by its minimum and maximum output
time. Simulation results will only be returned for the defined output times
even though the ODE solver may use smaller time steps between output times.
Output times can be explicitly set using \code{\link[=set_times]{set_times()}}. The number and
distance of output times may have influence on the precision of simulation
results and numerical stability, cf. \code{\link[=simulate]{simulate()}}.

Be aware that \code{\link[=set_exposure]{set_exposure()}} will overwrite previously defined output times
if not requested otherwise.
}

\subsection{Effects}{

Generally, all state variables can be used as effect endpoints but models may
provide additional endpoints. Use \code{\link[=set_endpoints]{set_endpoints()}} to enable or disable
endpoints for a scenario.

Some scenarios or models require control runs to calculate effects under
exposure. Generally, control simulations will run automatically
where needed.
}

\subsection{Moving exposure windows}{

The time frame relevant for effects may be much shorter than the assessed
exposure time-series for certain organisms. This fact can be represented
by moving exposure windows which divide a long time period in a number of
consecutive windows of the same length. Each window is simulated individually and effects are
calculated. By default, methods such as \code{\link[=effect]{effect()}} will only return the
maximum effect of all considered windows but detailed results can be presented
on demand.

To use moving exposure windows, the exposure time-series must be regular,
i.e. must have an equidistant step length in time. The length of the window
is defined as the number of time steps of the exposure time-series. As an
example, assume the time-series has daily granularity and a moving window
of seven days length is required. In this case, the moving window must have
a length of seven (7) time steps. If the exposure time-series had hourly granularity,
the same window would need to have a length of 168 (=7*24) time steps. Please
refer to \code{\link[=set_window]{set_window()}} for details.
}
}
\section{Slots}{

\describe{
\item{\code{name}}{\code{character}, unique model name}

\item{\code{tag}}{\code{character}, an optional identifier}

\item{\code{param}}{\code{list} of parameter key-value pairs}

\item{\code{param.bounds}}{named \code{list} of parameter boundaries}

\item{\code{param.req}}{\code{character} vector of required parameters}

\item{\code{forcings}}{\code{list} of \code{data.frame}s representing forcing time-series}

\item{\code{forcings.req}}{\code{character} vector or required model forcings data, e.g. temperature}

\item{\code{init}}{\code{list} of initial model states}

\item{\code{times}}{\code{numeric} vector of output times, beginning and end also define
the simulated period}

\item{\code{endpoints}}{\code{character} vector of endpoints to calculate results for}

\item{\code{exposure}}{\code{data.frame} with two columns representing an exposure time-series}

\item{\code{control}}{\code{list} of named numerical vectors, contains the control values
for all relevant moving windows}

\item{\code{control.req}}{\code{logical}, if \code{TRUE} then control values are required to
calculate effects}

\item{\code{window.length}}{\code{numeric}, maximum length of the simulated period, if
\code{window.length} is shorter than the exposure pattern, then all possible
exposure sub-patterns are evaluated for effect calculation. This is also
referred to as a moving window approach.}

\item{\code{window.interval}}{\code{numeric}, interval determining distance between moving
windows during effect calculation. First window starts at first time point
in exposure pattern.}
}}

\seealso{
Other scenario-related: 
\code{\link{Transferable}}
}
\concept{scenario-related}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{simulate_batch}
\alias{simulate_batch}
\title{Batch simulation using multiple exposure series}
\usage{
simulate_batch(model_base, treatments, param_sample = deprecated(), ...)
}
\arguments{
\item{model_base}{effect scenario object with mean parameters}

\item{treatments}{treatments exposure levels as data frame (time, conc, trial)}

\item{param_sample}{\emph{deprecated} parameter, no longer in use}

\item{...}{additional parameters passed through to \code{\link[=simulate]{simulate()}}}
}
\value{
a \code{data.frame} with simulation results
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
}
\details{
A convenience function to simulate a single base scenario with one or more
exposure series. This aims at reproducing the setup and results of common
effect studies.

A scenario contains only one exposure series. However, laboratory experiments
commonly examine the effects of multiple exposure levels on a biological system.
A batch simulation approach would involve running multiple simulations with
varying exposure or treatment conditions. To illustrate, if the objective is
to examine the impact of a substance on cell growth, the simulation model
could be designed to replicate the cell growth dynamics under varying
concentrations of the substance. Each simulation run would represent a
specific exposure level, ranging from low to high concentrations of the
chemical. To simulate such a laboratory experiment, the simulate_batch
function can be used. All exposure series are saved in the treatment argument.
The first column contains the time, the second column the concentration, and
the third column the trial name (exposure level, e.g. 'T1', 'T2', 'T3').
}
\examples{
t1 <- data.frame(time=0:10, conc=0, trial="control")  # 1st treatment level
t2 <- data.frame(time=0:10, conc=1, trial="T1")       # 2nd treatment level
treatments <- rbind(t1, t2)

metsulfuron \%>\%
  simulate_batch(treatments)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explore_space.R
\name{explore_space}
\alias{explore_space}
\title{Explore parameter space}
\usage{
explore_space(
  x,
  par,
  res,
  output,
  data,
  sample_size = 1000,
  max_iter = 30,
  nr_accept = 100,
  sample_factor = 1.2,
  individual = FALSE,
  log_scale = FALSE,
  data_type = c("continuous", "count"),
  max_runs = deprecated(),
  ...
)
}
\arguments{
\item{x}{a list of \link{caliset} objects}

\item{par}{best fit parameters from joined calibration}

\item{res}{output of \code{\link[=lik_profile]{lik_profile()}} function}

\item{output}{character vector, name of output column of \code{\link[=simulate]{simulate()}} that
is used in calibration}

\item{data}{only needed if \code{x} is a \link{scenario}}

\item{sample_size}{number of samples to draw from each parameter interval}

\item{max_iter}{max number of iterations to redraw samples (within a smaller space), and repeat the process}

\item{nr_accept}{threshold for number of points sampled within the inner circle}

\item{sample_factor}{multiplication factor for sampling (95\% interval * sample factor)}

\item{individual}{if \code{FALSE} (default), the log likelihood is calculated across
the whole dataset. Alternatively, if \code{TRUE}, log likelihoods are calculated for
each (group of) \emph{set}(s) individually.}

\item{log_scale}{\code{FALSE} (default), option to calculate the log likelihood on a
log scale (i.e., observations and predictions are log transformed during calculation)}

\item{data_type}{Character argument, \code{"continuous"} (default) or \code{"count"}, to specify the data type
for the log likelihood calculations.}

\item{max_runs}{\emph{deprecated} alias for \code{max_iter} parameter}

\item{...}{additional parameters passed through to \code{\link[=simulate]{simulate()}}}
}
\value{
a list containing a plot to explore the parameter space, and the \code{data.frame}
supporting it
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

The function is aimed at getting an idea of how the parameter space
of a model behaves, so that parameter identifiability problems and correlations
between parameters can be explored. Therefore, the function samples a large
number of parameter sets by randomly drawing from each parameter's 95\%
confidence interval (generated by \code{\link[=lik_profile]{lik_profile()}}). It then
checks how many of the parameter sets are within acceptable limits by comparing
the likelihood ratio of a parameter set vs. the original parameter set against
a chi-square distribution as degrees of freedom (df) the total number of profile
parameters (outer rim) or one df (inner rim). If needed, the function resamples
until at least \code{nr_accept} parameters sets are within the inner rim
}
\examples{
\donttest{
library(dplyr)
# Example with Lemna model - physiological params
# Before applying the function, a model needs to be calibrated and its parameters profiled
# Inputs for likelihood profiling

# observations - control run
obs <- schmitt2013 \%>\%
  filter(trial == "T0")

# parameters after calibration
params <- c(
  k_phot_max = 5.663571,
  k_resp = 1.938689,
  Topt = 26.7
)

# set parameter boundaries (if different from defaults)
bounds <- list(
  k_resp = list(0, 10),
  k_phot_max = list(0, 30),
  Topt = list(20, 30)
)

# update metsulfuron
myscenario <- metsulfuron \%>\%
  set_init(c(BM = 1.2, E = 1, M_int = 0)) \%>\%
  set_param(list(
    k_0 = 5E-5,
    a_k = 0.25,
    BM50 = 17600,
    mass_per_frond = 0.1
  )) \%>\%
  set_noexposure() \%>\%
  set_param(params) \%>\%
  set_bounds(bounds)

# Likelihood profiling
res <- lik_profile(
  x = myscenario,
  data = obs,
  output = "FrondNo",
  par = params,
  refit = FALSE,
  type = "fine",
  method = "Brent"
)
# plot
plot(res)

# parameter space explorer
set.seed(1) # for reproducibility
res_space <- explore_space(
  x = myscenario,
  data = obs,
  par = params,
  res = res,
  output = "FrondNo",
  sample_size = 1000,
  max_iter = 20,
  nr_accept = 100)

plot(res_space)

}
}
\seealso{
\link{plot.param_space}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{metsulfuron}
\alias{metsulfuron}
\title{Lemna model fitted to \emph{metsulfuron} effect data}
\format{
An object of class \code{LemnaSchmitt} of length 1.
}
\usage{
metsulfuron
}
\description{
Data set for the parametrisation of a mechanistic combined
toxicokinetic-toxicodynamic (TK/TD) and growth model for the aquatic
macrophytes Lemna spp. as published by Schmitt \emph{et al.} (2013).
The growth model was parameterised by Schmitt et al. based on these
data while toxicokinetic and toxicodynamic parameters were determined by
fitting the model using substance specific effect data of the
herbicide \emph{metsulfuron-methyl}.
}
\details{
The scenario is based on the \link{Lemna_Schmitt} model.
}
\references{
Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013: \emph{Mechanistic TK/TD-model
simulating the effect of growth inhibitors on Lemna populations}. Ecol Model 255,
pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
}
\seealso{
\link{Lemna-models}, \link{schmitt2013}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fx.R, R/model-lemna_setac.R, R/model-magma.R,
%   R/model-algae.R, R/model-guts.R, R/model-guts_red.R
\name{fx}
\alias{fx}
\alias{fx,ANY-method}
\alias{fx,ScenarioSequence-method}
\alias{fx,Lemna-method}
\alias{fx,Magma-method}
\alias{fx,Algae-method}
\alias{fx,GutsSd-method}
\alias{fx,GutsIt-method}
\alias{fx,GutsRedSd-method}
\alias{fx,GutsRedIt-method}
\title{Generic to calculate effects for a particular scenario}
\usage{
fx(scenario, ...)

\S4method{fx}{ANY}(scenario, ...)

\S4method{fx}{ScenarioSequence}(scenario, ...)

\S4method{fx}{Lemna}(scenario, ...)

\S4method{fx}{Magma}(scenario, ...)

\S4method{fx}{Algae}(scenario, ...)

\S4method{fx}{GutsSd}(scenario, ...)

\S4method{fx}{GutsIt}(scenario, ...)

\S4method{fx}{GutsRedSd}(scenario, ...)

\S4method{fx}{GutsRedIt}(scenario, ...)
}
\arguments{
\item{scenario}{\link{scenario} object}

\item{...}{additional parameters}
}
\value{
numeric named vector
}
\description{
Generic to calculate effects for a particular scenario
}
\section{Methods (by class)}{
\itemize{
\item \code{fx(ANY)}: Use state variables at end of simulation

\item \code{fx(ScenarioSequence)}: Wrapper for \link[=sequence]{sequences}

\item \code{fx(Lemna)}: Effect at end of simulation of \link{Lemna-models}

\item \code{fx(Magma)}: Effect at end of simulation of \link{Magma} scenarios

\item \code{fx(Algae)}: Effect at end of simulation of \link{Algae-models}

\item \code{fx(GutsSd)}: Calculates lethality of \link{GUTS-SD} scenarios

\item \code{fx(GutsIt)}: Calculates lethality of \link{GUTS-IT} scenarios

\item \code{fx(GutsRedSd)}: Calculates lethality of \link{GUTS-RED-SD} scenarios

\item \code{fx(GutsRedIt)}: Calculates lethality of \link{GUTS-RED-IT} scenarios

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_toxswa.R
\name{import_toxswa}
\alias{import_toxswa}
\title{Import \emph{TOXSWA} exposure series}
\usage{
import_toxswa(
  files,
  alias = NA,
  output_var = "ConLiqWatLay",
  output_unit = "ug/L",
  time_unit = "days",
  substance = NULL,
  split = TRUE
)
}
\arguments{
\item{files}{vector of strings with absolute or relative paths to files}

\item{alias}{optional vector with strings, will be used as an alias to identify
a TOXSWA series instead of its filename}

\item{output_var}{character, single output variable from \emph{TOXSWA} that is
imported, defaults to \emph{ConLiqWatLay}}

\item{output_unit}{character, target unit of the imported output variable,
defaults to \emph{ug/L}, syntax must be compatible with \code{\link[units:units]{units::units()}}}

\item{time_unit}{character, target unit of the imported time scale,
defaults to \emph{days}, syntax must be compatible with \code{\link[units:units]{units::units()}}}

\item{substance}{optional vector of characters, if set, only the substance
codes defined in this vector are imported}

\item{split}{logical, if \code{TRUE} then one series will be returned for each
substance found in the \emph{TOXSWA} files, else all substances per file will
be in one \emph{data.frame}. Defaults to \code{TRUE}}
}
\value{
list of \emph{data.frame} objects with exposure series. Each \emph{data.frame} has at
least three columns:
\itemize{
\item \code{time}: numerical time scale, always starts at zero
\item \code{timestamp}: time as datetime objects such as \code{POSIXct}
\item one or more additional columns for each imported substance
}
}
\description{
Read one or more \emph{TOXSWA} exposure series from \emph{TOXSWA}'s \code{.out} files. By default, the
concentration dissolved in water (\emph{ConLiqWatLay}) at the end of the simulated waterbody
(i.e. at the maximum of the \emph{x} dimension) is returned. The unit of the time
scale as well as of the imported model output variable can be scaled as needed.
}
\details{
The numerical time scale is shifted to always start at time zero (0.0).
Numerical columns of the returned \emph{data.frame} objects will be of type
\link[units:units]{units::units}. Please be aware that the use of \code{units} objects may not be
supported by all functions in this package. However, \code{\link[=set_times]{set_times()}} and
\code{\link[=set_exposure]{set_exposure()}} can handle \code{units} objects safely.

Incomplete list of alternative \emph{TOXSWA} v5.5.3 output variables:
\itemize{
\item \emph{ConLiqWatLay}: Concentration dissolved in water (g/m3)
\item \emph{ConLiqSed}: Concentration in pore water sediment (g/m3)
\item \emph{ConSysWatLay}: Total concentration in water (g/m3)
\item \emph{CntSorSusSol}: Content sorbed to suspended solids (g/kg)
\item \emph{CntSorSed}: Content sorbed to sediment (g/kg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival.R
\name{survival}
\alias{survival}
\title{Survival rate}
\usage{
survival(scenario, ...)
}
\arguments{
\item{scenario}{an \code{EffectScenario} to simulate}

\item{...}{additional parameters passed on to \code{\link[=simulate]{simulate()}}}
}
\value{
a \code{data.frame} containing simulation results
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Derives the survival rate of individuals for
\emph{Reduced GUTS} models. Function was replaced by output of \code{\link[=simulate]{simulate()}} and
will be removed in a later version.
}
\details{
The survival rate describes the survival probability at each
time point. The function simulates the \emph{GUTS} scenario and appends a column
\code{survival} to the simulation result. A value of one (\code{1.0}) denotes that
all individuals survive. A value of zero (\code{0.0}) denotes that no individuals
survived.

Only available for \emph{Reduced GUTS} models, see \link{GUTS-RED-models}.
The equations were described by EFSA (2018).
}
\examples{
# calculate survival rate
minnow_it \%>\% survival()

# plot survival over time based on a random exposure profile
minnow_sd \%>\%
  set_exposure(data.frame(t=1:100, c=runif(100)*10)) \%>\%
  survival() -> df
plot(df$time, df$survival, "l")
}
\references{
EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, et al., 2018: \emph{Scientific Opinion on the
state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
regulatory risk assessment of pesticides for aquatic organisms}. EFSA Journal 2018;
16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
}
\seealso{
\link{GUTS-RED-models}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{minnow_sd}
\alias{minnow_sd}
\title{A fitted GUTS-RED-SD scenario of the fathead minnow}
\format{
An object of class \code{GutsRedSd} of length 1.
}
\source{
https://mosaic.univ-lyon1.fr/guts
}
\usage{
minnow_sd
}
\description{
The example scenario consists of a fitted \link[=GUTS_RED_SD]{GUTS-RED-SD}
model and a constant exposure series. Model parameters were derived from a
typical four-day acute fish toxicity study of the \emph{fathead minnow} by
Geiger \emph{et al.} (1988).
The study evaluated the effect of \emph{chlorpyrifos} concentrations in water on
survival of \emph{fathead minnows}.
}
\details{
The toxicity dataset used for parameter calibration is also referred to as
\emph{GUTS Ring-test dataset C} by EFSA (2018). Fitted parameters were estimated
using the \emph{morse} package.

The exposure series of the example scenario is a constant concentration of 1.0 µmol/L
over a period of four days with a daily time step.
}
\examples{
# Print scenario parameters
minnow_sd

# Run the example scenario
minnow_sd \%>\% simulate()
}
\references{
Geiger D.L., Call D.J., and Brooke L.T., 1988: \emph{Acute toxicities of organic
chemicals to fathead minnows (Pimephales promelas): Volume IV}, pp. 195-197.
University of Wisconsin-Superior, Center for Lake Superior Environmental Studies.
ISBN 9780961496838.

EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, et al., 2018: \emph{Scientific Opinion on the
state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
regulatory risk assessment of pesticides for aquatic organisms}. EFSA Journal 2018;
16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
}
\seealso{
\link{GUTS-RED-models}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{log_scenarios}
\alias{log_scenarios}
\title{Log scenario properties}
\usage{
log_scenarios(x, header = TRUE)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}

\item{header}{\code{logical}, if \code{TRUE} a header line will be printed}
}
\value{
unmodified argument \code{x}
}
\description{
Log scenario properties
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pull.R
\name{pull_metadata}
\alias{pull_metadata}
\title{Pull metadata from scenarios}
\usage{
pull_metadata(x, model = TRUE, exposure = TRUE)
}
\arguments{
\item{x}{vector of \link{scenarios} or a \code{data.frame} containing a column
\code{scenario} with \code{EffectScenario} objects}

\item{model}{\code{logical}, if \code{TRUE} then model metadata is pulled}

\item{exposure}{\code{logical}, if \code{TRUE} then exposure series metadata is pulled}
}
\value{
a \code{data.frame}
}
\description{
The method pulls available metadata from scenario objects and returns a
table with additional columns. If the argument already was a \code{data.frame}
object, the columns are appended. May overwrite existing columns of the
same name.
}
\examples{
metsulfuron \%>\%
  pull_metadata()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.r
\name{plot_ppc}
\alias{plot_ppc}
\title{Creates a PPC plot for a single dataset}
\usage{
plot_ppc(
  rs_mean,
  rs_range,
  col_number = 2,
  obs_mean = NULL,
  obs_full = NULL,
  xy_lim = NULL,
  study = NULL
)
}
\arguments{
\item{rs_mean}{\code{data.frame}, model results best fit params}

\item{rs_range}{\code{data.frame}, predictions (min, max from param.sample run)}

\item{col_number, }{column to plot, default = 2}

\item{obs_mean}{\code{data.frame}, observations with means per treatment level}

\item{obs_full}{\code{data.frame}, full data set including results for replicates}

\item{xy_lim}{optional \code{numeric}, limits of x and y axis for plotting}

\item{study}{optional \code{string}, name of study which can be used as key}
}
\value{
a ggplot2 plot object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\details{
A sample of parameters representing the uncertainty within the dataset
is passed to the function. All parameter combinations and exposure patterns
are simulated and the range of predicted frond numbers is derived for a
single study. The uncertainty is displayed by a Posterior Predictive Plot
(PPC). The data (rs_mean, obs_mean and obs_full) must have the
following format (col1 = time, col2 = data of interest, col3 = trial name).
Data for uncertainties (rs_range) must have the format: col1 = time,
col2 = lower boundaries, col3 = upper boundaries, col4 = trial. The user
should take care of the input data and consider whether control data and
data at time zero should be included in the model check.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-CalibrationSet.R
\name{CalibrationSet}
\alias{CalibrationSet}
\alias{CalibrationSet-class}
\alias{caliset}
\title{Calibration set}
\usage{
caliset(scenario, data, weight = 1.0, tag = NULL)
}
\arguments{
\item{scenario}{a \link{scenario} describing conditions during the experiment}

\item{data}{a \code{data.frame} with observed data in long format containing two
columns: the 1st column with \code{numeric} time points and 2nd column with
\code{numeric} data to fit to. Rows with observed \code{NA} values will be removed.}

\item{weight}{optional \code{numeric} weight to be applied when calculating the
error term for each data point. Default value is \code{1.0}, i.e. no weighting.}

\item{tag}{optional value to identify the data, e.g. a study number}
}
\value{
\code{caliset()} returns a \emph{calibration set} object
}
\description{
A \emph{calibration set} combines a \link{scenario}, observed data, and an
optional weighting factor into one object. The \emph{calibration set} is used to fit
model parameters to observed data using \code{\link[=calibrate]{calibrate()}}.
}
\details{
A \emph{calibration set} usually represents a single experiment or trial.
Multiple experimental replicates can be combined into a single \emph{set}, if model
parameters are identical between trials.
If model parameters were modified during a trial, e.g. a pump failure occurred
or flow rates changed, this can be represented by using a \emph{scenario sequence}
instead of a basic \link{scenario}. Please refer to \code{\link[=sequence]{sequence()}} for details.
\subsection{Weighting}{

An optional weighting factor can be used to scale the error term of a
whole \emph{set} or of individual data points when fitting parameters using e.g.
\code{\link[=calibrate]{calibrate()}}.

The vector of weights must either be of length one or have the same length
as the dataset. In the former case, the same weight will be applied to all
values in the dataset. In the latter, individual weights are applied
for each data point.
}
}
\examples{
library(dplyr)

# Get observed biomass during control experiment by Schmitt et al. (2013)
observed <- schmitt2013 \%>\%
  filter(trial == "T0") \%>\%
  select(time, obs)

# Create a scenario that represents conditions during experiment
scenario <- metsulfuron \%>\%
  set_param(c(k_phot_fix=TRUE, k_resp=0, Emax=1)) \%>\%
  set_init(c(BM=12)) \%>\%
  set_noexposure() \%>\%
  set_bounds(list(k_phot_max=c(0, 0.5)))

# Create a calibration set
cs <- caliset(scenario, observed)

# Fit parameter 'k_phot_max' to observed biomass growth from experiment
calibrate(
  cs,
  par=c(k_phot_max=1),
  output="BM",
  method="Brent" # Brent is recommended for one-dimensional optimization
) -> fit
fit$par

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_tktd.R
\name{fit_tktd}
\alias{fit_tktd}
\alias{fit_tktd,ANY,ANY-method}
\alias{fit_tktd,list,missing-method}
\alias{fit_tktd,cvasi_fit,ANY-method}
\alias{fit_tktd,CalibrationSet,ANY-method}
\alias{fit_tktd,CalibrationSet,missing-method}
\alias{fit_tktd,EffectScenario,ANY-method}
\alias{fit_tktd,ScenarioSequence,ANY-method}
\alias{fit_tktd,LemnaSetac,missing-method}
\alias{fit_tktd,LemnaSchmitt,missing-method}
\alias{fit_tktd,Magma,missing-method}
\title{Fit TK/TD parameters}
\usage{
fit_tktd(x, data, ...)

\S4method{fit_tktd}{LemnaSetac,missing}(x, data, par, log_scale = TRUE, verbose = FALSE, ...)

\S4method{fit_tktd}{LemnaSchmitt,missing}(x, data, par, log_scale = TRUE, verbose = FALSE, ...)

\S4method{fit_tktd}{Magma,missing}(x, data, par, log_scale = TRUE, verbose = FALSE, ...)
}
\arguments{
\item{x}{a \link{scenario} or a list of \link{caliset} objects}

\item{data}{toxicological trial data to fit TK/TD parameters to: required if
\code{x} is either a \link{scenario} or the result of a fit. See Section \emph{Data}
for details.}

\item{...}{additional arguments passed through to \code{\link[=calibrate]{calibrate()}}}

\item{par}{named vector, of parameters to fit and their starting values}

\item{log_scale}{logical, if \code{TRUE} then fitting will be performed on log-transformed
observations and predictions, else the data will be used as-is}

\item{verbose}{logical, if \code{TRUE} then info messages are printed to the console}
}
\value{
a list
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} \if{html}{\out{<br>}}
High-level function to fit TK/TD parameters to observed data. It eases the
use of \code{\link[=calibrate]{calibrate()}} by providing sensible defaults for various (model dependent)
settings.
}
\details{
\subsection{Data}{

The function can be used in three basic ways: Fit parameters of
\enumerate{
\item a single \link{scenario} to a data set
\item a list of \link{caliset}s
\item a list of \link{caliset}s in a chain of \code{fit} functions
}

For option 1), scenario and data are supplied separately. In this case,
supported types of argument \code{data} are \code{data.frame}s and \link{tox_data} objects.
Any \code{data.frame} must have a format that is compatible with the \code{\link[=tox_data]{tox_data()}}
function.

In option 2), all conditions are fully described by a list of calibration sets.
The user is responsible to set up all \link{caliset} objects to their needs.

Option 3) is for convenience purposes and allows the chaining of fit functions.
The latter alternative accepts the return value of e.g. \code{\link[=fit_growth]{fit_growth()}}
as argument \code{x} and the fitted parameter values are applied to all
calibration sets in argument \code{data}.
}
}
\section{Methods (by class)}{
\itemize{
\item \code{fit_tktd(x = LemnaSetac, data = missing)}: Fit TK/TD parameters of \link{Lemna_SETAC} scenarios

\item \code{fit_tktd(x = LemnaSchmitt, data = missing)}: Fit TK/TD parameters of \link{Lemna_Schmitt} scenarios

\item \code{fit_tktd(x = Magma, data = missing)}: Fit TK/TD parameters of \link{Magma} scenarios

}}
\examples{
# Use experimental data from control trial ('T0') to fit growth rate
ctrl <- schmitt2013[schmitt2013$trial == "T0" ,]

# Set up a scenario, provide dummy parameter values where necessary
sc <- Lemna_Schmitt() \%>\%
  set_init(c(BM=0.0012)) \%>\%
  set_param(c(EC50=1, b=1, P_up=1))

# Run fitting routine
fit_growth(sc, data=ctrl, verbose=TRUE)

# Use fitted growth parameter to adapt scenario
sc2 <- sc \%>\% set_param(c(k_phot_max=0.43925))

# Use experimental ecotox data for various concentrations of 'metsulfuron-methyl'
trials <- schmitt2013[schmitt2013$trial != "T0" ,]

# Fit remaining TK/TD parameters
fit_tktd(sc2, data=trials, verbose=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{log_envir}
\alias{log_envir}
\title{Log R environment properties}
\usage{
log_envir()
}
\value{
no return value
}
\description{
Log R environment properties
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot}
\alias{plot}
\alias{plot.cvasi_drc}
\alias{plot.cvasi_simulate}
\alias{plot.lik_profile}
\alias{plot_lik_profile}
\title{S3 plotting functions}
\usage{
\method{plot}{cvasi_drc}(x, y, scale_x = c("auto", "log10", "none"), ...)

\method{plot}{cvasi_simulate}(x, y, ...)

\method{plot}{lik_profile}(x, y, ...)

plot_lik_profile(x)
}
\arguments{
\item{x}{object to plot}

\item{y}{unused parameter}

\item{scale_x}{character, controls how the x-axis is scaled. \code{log10} for
a log10-scaled axis, \code{none} for no scaling, and \code{auto} for automatic selection}

\item{...}{unused parameters}
}
\value{
(\emph{ggplot2}) plot object
}
\description{
These functions overload \code{\link[base:plot]{base::plot()}} to provide simple plotting
routines to display various time-series and scenario objects.
}
\section{Methods (by class)}{
\itemize{
\item \code{plot(cvasi_drc)}: Plot dose response curves

\item \code{plot(cvasi_simulate)}: Plot return value of \code{\link[=simulate]{simulate()}}

\item \code{plot(lik_profile)}: Plot likelihood profiles.

}}
\section{Functions}{
\itemize{
\item \code{plot_lik_profile()}: Alias of \code{plot.lik_profile()} for backwards-compatibility.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-algae.R
\name{Algae-models}
\alias{Algae-models}
\alias{Algae-class}
\title{Algae models}
\description{
Overview of supported \emph{Algae} models
}
\details{
\itemize{
\item \code{\link[=Algae_Weber]{Algae_Weber()}} by Weber \emph{et al.} (2012)
\item \code{\link[=Algae_TKTD]{Algae_TKTD()}} based on Weber \emph{et al.} (2012), but with scaled damage
\item \code{\link[=Algae_Simple]{Algae_Simple()}} simplified growth model without additional forcing variables
}
}
\section{Biomass transfer}{

Models supporting biomass transfer can be instructed to move a fixed amount
of biomass to a new medium after a period of time. This feature replicates
a procedure occurring in e.g. \emph{Lemna} effect studies and may be necessary to
recreate study results.

The biomass transfer feature assumes that always a fixed amount of
biomass is transferred. Transfers can occur at any fixed point in time or
in regular intervals. During a transfer, the biomass is reset to the
transferred amount and additional compartments can be scaled 1:1 accordingly,
to e.g. reflect the change in internal toxicant mass when biomass is modified.
Transfer settings can be modified using \code{\link[=set_transfer]{set_transfer()}}.

If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
}

\references{
Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
and Ratte HT, 2012. Combination of a higher-tier flow-through system and
population modeling to assess the effects of time-variable exposure of
isoproturon on the green algae Desmodesmus subspictatus and
Pseudokirchneriella subcapitata. Environmental Toxicology and
Chemistry, 31(4), 899-908. \doi{10.1002/etc.1765}

EFSA Panel on Plant Protection Products and their Residues, 2018. Scientific
opinion on the state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect
models for regulatory risk assessment of pesticides for aquatic organisms.
EFSA journal 16:5377 \doi{10.2903/j.efsa.2018.5377}
}
\seealso{
\link{Lemna-models}, \link{Transferable}

Other algae models: 
\code{\link{Algae_Simple}()},
\code{\link{Algae_TKTD}()},
\code{\link{Algae_Weber}()}

Other models: 
\code{\link{DEB-models}},
\code{\link{GUTS-RED-models}},
\code{\link{Lemna-models}},
\code{\link{Macrophyte-models}}
}
\concept{algae models}
\concept{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-magma.R
\docType{class}
\name{Magma-class}
\alias{Magma-class}
\title{Magma scenario class}
\description{
This entry documents the class definition used for \link{Magma}-type scenarios.
For details regarding the model, please refer to the \link[=Magma]{Magma model} manual.
}
\section{Slots}{

\describe{
\item{\code{growth_model}}{\emph{character}, selects the growth model, such as \emph{exponential}
or \emph{logistic} growth.}
}}

\seealso{
\link{Macrophyte-models}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{get_model}
\alias{get_model}
\alias{get_model,ANY-method}
\alias{get_model,list-method}
\alias{get_model,EffectScenario-method}
\alias{get_model,ParameterSet-method}
\alias{get_model_name}
\title{Get model name}
\usage{
get_model(x)
}
\arguments{
\item{x}{(vector of) \link{scenarios} or \code{\link{parameter_set}} objects}
}
\value{
vector of \code{character}
}
\description{
Returns the unique model name that is associated with a scenario, e.g.
\code{GUTS-RED-IT}. The function supports vectorized arguments.
}
\examples{
# returns `GUTS-RED-IT`
get_model(minnow_it)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numerics.R
\name{num_info}
\alias{num_info}
\alias{num_info.default}
\alias{num_info.cvasi_simulate}
\alias{num_info.cvasi_fit}
\title{Print information about numerical solver result}
\usage{
num_info(obj)

\method{num_info}{default}(obj)

\method{num_info}{cvasi_simulate}(obj)

\method{num_info}{cvasi_fit}(obj)
}
\arguments{
\item{obj}{Return value of \code{\link[=simulate]{simulate()}}}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\details{
Prints information on the status of a return value from \code{\link[=simulate]{simulate()}}, e.g.
if it was successful and what, if any, issues occurred. Also provides tips
on solving frequently occurring issues.

The function requires certain metadata which is created by \code{\link[deSolve:ode]{deSolve::ode()}}
and is passed through to the result of \code{simulate()}. The metadata may be lost
if the \code{data.frame} returned by \code{simulate()} is converted or cast to other types.
}
\examples{
# A simulation without any issues
minnow_it \%>\% simulate() \%>\% num_info()

# A simulation which terminated early due to the solver
# taking too many numerical steps
rs <- suppressWarnings(minnow_it \%>\% simulate(hmax=1e-80))
num_info(rs)

# Print deSolve diagnostics for additional information
diagnostics(rs)
}
\seealso{
\code{\link[=diagnostics]{diagnostics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_controls}
\alias{cache_controls}
\title{Cache control simulations}
\usage{
cache_controls(x, ...)
}
\arguments{
\item{x}{parameter not used}

\item{...}{parameters not used}
}
\value{
nothing
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
}
\details{
Handling of cached control simulations has been modified and is solely
managed by package functions. \code{cache_controls()} is no longer needed and
will raise an error if called.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-algae.R
\name{Algae_TKTD}
\alias{Algae_TKTD}
\alias{AlgaeTKTD-class}
\alias{AlgaeTKTDScenario-class}
\title{Algae model variant (Weber et al. 2012) with scaled damage}
\usage{
Algae_TKTD()
}
\value{
an S4 object of type \linkS4class{AlgaeTKTD}
}
\description{
The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
growth model for algae. The model simulates the development of algal biomass
under laboratory and environmental conditions. The growth of the algae
population is simulated on the basis of growth rates, which are dependent on
environmental conditions (radiation, temperature and phosphorus).
The model is a variant of the \code{\link[=Algae_Weber]{Algae_Weber()}} model (Weber 2012) as cited
in EFSA TKTD opinion (2018). This Algae model, \code{\link[=Algae_TKTD]{Algae_TKTD()}}, provides an
additional possibility (probit) to simulate the dose-response curve and
considers a scaled internal damage instead of the external concentration.
}
\section{State variables}{

The model has four state variables:
\itemize{
\item \code{A}, Biomass (µg fresh wt/mL, cells/mL *10^4)
\item \code{Q}, Mass of phosphorous internal (µg P/µg fresh wt)
\item \code{P}, Mass of phosphorous external (µg P/L)
\item \code{Dw}, Damage concentration (µg/L)
}
}

\section{Model parameters}{

\itemize{
\item Growth model
\itemize{
\item \code{mu_max}, Maximum growth rate (d-1)
\item \code{Q_min}, Minimum intracellular P (µg P/µg fresh wt)
\item \code{Q_max}, Maximum intracellular P (µg P/µg fresh wt)
\item \code{v_max}, Maximum P-uptake rate at non-limited growth (µg P/µg fresh wt/d)
\item \code{k_s},   Half-saturation constant for extracellular P (mg P/L)
\item \code{m_max}, Natural mortality rate (1/d)
\item \code{I_opt}, Optimum light intensity for growth (µE/m²/s)
\item \code{T_opt}, Optimum temperature for growth (°C)
\item \code{T_max}, Maximum temperature for growth (°C)
\item \code{T_min}, Minimum temperature for growth (°C)
\item \code{D}, Dilution rate (1/d)
\item \code{R_0}, Influx concentration of P (mg P/L)
}
\item Concentration response (Toxicodynamics)
\itemize{
\item \code{EC_50}, Effect concentration of 50\% inhibition of growth rate (µg L-1)
\item \code{b}, slope of concentration effect curve at EC_50 (-)
\item \code{dose_resp}, shape of the dose response curve (0 = logit, 1 = probit)
}
\item External concentration (Toxicokinetics)
\itemize{
\item \code{kD}, dominant rate constant (d-1)
}
}
}

\section{Forcings}{

The Weber model variant requires two environmental properties as time-series input:
\itemize{
\item \code{T_act}, temperature (°C), and
\item \code{I}, irradiance (uE/m²/s).
}

The following constant default values are used for these properties:
\itemize{
\item \code{T_act} = 23 °C
\item \code{I} = 100 uE/m²/s
}

Forcings time-series are represented by \code{data.frame} objects consisting of two
columns. The first for time and the second for the environmental factor in question.

Entries of the \code{data.frame} need to be ordered chronologically. A time-series
can consist of only a single row; in this case it will represent constant
environmental conditions. See \link{scenarios} for more details.
}

\section{Simulation output}{

Simulation results will contain the state variables Biomass (\code{A}), mass of
internal phosphorous (\code{Q}), mass of external phosphorous (\code{P}) and the damage
concentration (\code{Dw}).

It is possible to amend the output of \code{\link[=simulate]{simulate()}} with additional model
quantities that are not state variables, for e.g. debugging purposes or to
analyze model behavior. To enable or disable additional outputs, use the
optional argument \code{nout} of \code{\link[=simulate]{simulate()}}. As an example, set \code{nout=2} to
enable reporting of model derivatives \code{dA} and \code{dQ}. Set \code{nout=0} to disable
additional outputs (default).

The available output levels are as follows:
\itemize{
\item \code{nout >= 1}: \code{C}, external concentration (µg/L)
\item \code{nout >= 2}: \code{f(T)}, temperature dependence (-)
\item \code{nout >= 3}: \code{f(I)}, light dependence (-)
\item \code{nout >= 4}: \code{f(Q)}, nutrient dependence (-)
\item \code{nout >= 5}: \code{f(Q, P)}, uptake flow reduction (-)
\item \code{nout >= 6}: \code{f(C)}, effect of chemical stressor (-)
\item \code{nout >= 7}: \code{dA}, biomass derivative (µg)
\item \code{nout >= 8}: \code{dQ}, internal phosphorous derivative (mg P/µg fresh wt)
\item \code{nout >= 9}: \code{dP}, external phosphorous derivative (mg P L-1)
\item \code{nout >= 10}: \code{dDw}, damage concentration derivative (µg L-1)
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{hmax = 0.1}\if{html}{\out{<br>}}
Maximum step length in time suitable for most simulations.
}
}

\section{Model history and changes}{

\itemize{
\item cvasi v1.5.0
\itemize{
\item Support for simulating flow-through conditions by introducing new parameters
\code{D} and \code{R_0} and adapting the ODEs according to the \link{Algae_Weber} model.
\item ODE of external phosphorous concentration \code{P} corrected, which contained
an erroneous growth term before.
\item Response functions added to optional simulation outputs, order of output
levels modified.
}
}
}

\references{
Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
and Ratte HT, 2012. Combination of a higher-tier flow-through system and
population modeling to assess the effects of time-variable exposure of
isoproturon on the green algae Desmodesmus subspictatus and
Pseudokirchneriella subcapitata. Environmental Toxicology and
Chemistry, 31(4), 899-908. \doi{10.1002/etc.1765}
}
\seealso{
\link{Scenarios}, \link{Transferable}

Other algae models: 
\code{\link{Algae-models}},
\code{\link{Algae_Simple}()},
\code{\link{Algae_Weber}()}
}
\concept{algae models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver.R, R/model-lemna_setac.R,
%   R/model-magma.R, R/model-lemna_schmitt.R, R/model-algae.R,
%   R/model-deb_abj.R, R/model-debtox.R, R/model-deb_daphnia.R, R/model-guts.R,
%   R/model-guts_red.R
\name{solver}
\alias{solver}
\alias{solver,ANY-method}
\alias{solver,LemnaSetac-method}
\alias{solver,Magma-method}
\alias{solver,LemnaSchmitt-method}
\alias{solver,AlgaeWeber-method}
\alias{solver,AlgaeTKTD-method}
\alias{solver,AlgaeSimple-method}
\alias{solver,DebAbj-method}
\alias{solver,DebTox-method}
\alias{solver,DebDaphnia-method}
\alias{solver,GutsSd-method}
\alias{solver,GutsIt-method}
\alias{solver,GutsRedSd-method}
\alias{solver,GutsRedIt-method}
\title{Calls ODE solver for a particular model}
\usage{
solver(scenario, ...)

\S4method{solver}{ANY}(scenario, ...)

\S4method{solver}{LemnaSetac}(scenario, ...)

\S4method{solver}{Magma}(scenario, ...)

\S4method{solver}{LemnaSchmitt}(scenario, ...)

\S4method{solver}{AlgaeWeber}(scenario, method = "lsoda", hmax = 0.1, nout = 1, ...)

\S4method{solver}{AlgaeTKTD}(scenario, method = "lsoda", hmax = 0.1, ...)

\S4method{solver}{AlgaeSimple}(scenario, method = "lsoda", hmax = 0.1, ...)

\S4method{solver}{DebAbj}(scenario, ...)

\S4method{solver}{DebTox}(scenario, method = "ode45", ...)

\S4method{solver}{DebDaphnia}(scenario, ...)

\S4method{solver}{GutsSd}(scenario, ...)

\S4method{solver}{GutsIt}(scenario, ...)

\S4method{solver}{GutsRedSd}(scenario, ...)

\S4method{solver}{GutsRedIt}(scenario, ...)
}
\arguments{
\item{scenario}{\link{scenario} object}

\item{...}{additional parameters passed on to \code{\link[deSolve:ode]{deSolve::ode()}}}

\item{method}{string, selects the numerical solver used by \code{\link[deSolve:ode]{deSolve::ode()}}}

\item{hmax}{numeric, sets the maximum step length in time, see \code{\link[deSolve:ode]{deSolve::ode()}}}

\item{nout}{numeric, the  number of additional output variables returned by
the model, see \code{vignette("compiledCode", "deSolve")} for details.
If and which output variables are available depends on the scenario
type, please refer to the documentation of the model in question.}
}
\value{
\code{data.frame} with simulation results
}
\description{
Please refer to the \emph{Modeling Howto} vignette on how to implement custom
models by overloading the \code{solver} function.
}
\details{
Some solvers may set reasonable default values for e.g. maximum step
length in time (\code{hmax}), but not all do. Please check the model documentation for
details.
}
\section{Methods (by class)}{
\itemize{
\item \code{solver(ANY)}: Default solver, raises an error

\item \code{solver(LemnaSetac)}: Numerically integrates Lemna_SETAC scenarios

\item \code{solver(Magma)}: Numerically integrates \code{Magma} scenarios

\item \code{solver(LemnaSchmitt)}: Numerically integrates Lemna_Schmitt scenarios

\item \code{solver(AlgaeWeber)}: numerically integrates Algae_Weber scenarios

\item \code{solver(AlgaeTKTD)}: numerically integrates Algae_TKTD scenarios

\item \code{solver(AlgaeSimple)}: numerically integrates Algae_Simple scenarios

\item \code{solver(DebAbj)}: Numerically integrates DEB_abj scenarios

\item \code{solver(DebTox)}: Numerically integrates \emph{DEBtox} scenarios

\item \code{solver(DebDaphnia)}: (deprecated) Numerically integrates \emph{DEBtox_Daphnia} scenarios

\item \code{solver(GutsSd)}: Numerically integrates GUTS-SD scenarios

\item \code{solver(GutsIt)}: Numerically integrates GUTS-IT scenarios

\item \code{solver(GutsRedSd)}: Numerically integrates GUTS-RED-SD scenarios

\item \code{solver(GutsRedIt)}: Numerically integrates GUTS-RED-IT scenarios

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_init.R
\name{set_init}
\alias{set_init}
\alias{set_init,EffectScenario-method}
\alias{set_init,ScenarioSequence-method}
\alias{set_init,vector-method}
\title{Set initial state}
\usage{
set_init(x, init)

\S4method{set_init}{EffectScenario}(x, init)

\S4method{set_init}{ScenarioSequence}(x, init)

\S4method{set_init}{vector}(x, init)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}

\item{init}{named numeric vector}
}
\value{
modified \code{EffectScenario} objects
}
\description{
The \emph{initial state} represents the starting values of a scenario's state
variables when starting a simulation. A scenario's default initial state
may be insufficient to get sensible results.
}
\details{
In theory, a scenarios's state variables can be renamed by modifying the names
of the initial state vector. However, this is strongly
discouraged as this will affect other routines such as \code{\link[=effect]{effect()}} and \code{\link[=epx]{epx()}}
and may render results useless.
}
\examples{
# Set initial biomass to 1.0
metsulfuron \%>\% set_init(c(BM=1.0)) \%>\% simulate()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{log_enable}
\alias{log_enable}
\alias{log_disable}
\title{Start and stop logging}
\usage{
log_enable(file = NULL, append = TRUE, envir = parent.frame())

log_disable()
}
\arguments{
\item{file}{\code{character}, file name or path to a log file}

\item{append}{\code{logical}, if \code{TRUE} output will be appended to an existing log file,
otherwise the log file will be replaced}

\item{envir}{log will be automatically disabled if \code{environment} is exited,
set to \code{NULL} to disable}
}
\value{
no return value
}
\description{
Start and stop logging
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{log_msg}
\alias{log_msg}
\title{Add a log message}
\usage{
log_msg(...)
}
\arguments{
\item{...}{elements will be concatenated using \code{paste0()}}
}
\value{
no return value
}
\description{
Message will only appear in the console or in log file if logging was
enabled using \code{log_enable()}.
}
\examples{
log_msg("this message will not appear")

log_enable()
log_msg("this message will appear")
log_msg("a number of ","elements to ",42," concatenate")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.r
\name{plot_ppc_combi}
\alias{plot_ppc_combi}
\title{Create PPC plot for one or more datasets}
\usage{
plot_ppc_combi(table, xy_lim = NULL)
}
\arguments{
\item{table}{\code{data.frame} containing return values of calls to \code{plot_ppc()}}

\item{xy_lim}{optional \code{numeric}, limits of x and y axis for plotting}
}
\value{
a ggplot2 plot object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\details{
The function expects a data.frame with five mandatory and one optional
column. The mandatory columns are as follows:
\itemize{
\item \code{pred}: mean of predictions e.g. frond number for lemna
\item \code{max}: maximum of predictions
\item \code{min}: minimum of predictions
\item \code{obs}: observations
\item \code{PPC}: color code
The optional column is to be named \code{study} and contains a study identifier.
If more than one study identifier is present in the table, individual
studies will be plotted in different colors and a legend will be displayed.
The function is called by plot_ppc where the column names are defined
(see rs_ppc object).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_forcings.R
\name{set_forcings}
\alias{set_forcings}
\alias{set_forcings,EffectScenario-method}
\alias{set_forcings,list-method}
\title{Set time-dependent parameters}
\usage{
set_forcings(x, ...)

\S4method{set_forcings}{EffectScenario}(x, ...)

\S4method{set_forcings}{list}(x, ...)
}
\arguments{
\item{x}{(vector of) \link{scenario} objects}

\item{...}{named argument list to set as forcings}
}
\value{
Modified \link{scenarios}
}
\description{
Parameters which change their value over time are referred to as \emph{forcings}.
If and what parameters can vary over time depends on the model in question.
In many cases, \emph{forcings} represent time-series of environmental properties.
}
\details{
Forcing time-series are always represented by a
\code{data.frame} containing two columns. The first column representing time,
the second representing the parameter that is a function of time. The
ordering of columns is mandatory. The column names are essentially irrelevant
but may help documenting the scenario and its data. The rows must be
ordered chronologically. A time-series can consist of only a single row; in
this case it will represent constant conditions.

Handling forcing time-series is a costly task for the ODE solver due to consistency
checks and interpolation between timesteps. How the solver interpolates
the forcing time-series can be controlled by certain arguments to functions
such as \code{\link[=simulate]{simulate()}} and \code{\link[=effect]{effect()}}. Please refer to \code{\link[=simulate]{simulate()}} for a brief
overview and \link[deSolve:forcings]{deSolve::forcings} for a detailed description.

Forcing time-series should be kept as short as possible and as complex as
needed for optimal computational efficiency.
}
\examples{
# constant values will be automatically converted to a data.frame
Lemna_Schmitt() \%>\% set_forcings(temp=20) -> lemna
lemna@forcings

# setting multiple forcings at once
df <- data.frame(t=0:14, temp=rnorm(15, mean=20)) # random temperature series
Lemna_Schmitt() \%>\% set_forcings(temp=df, rad=15000) -> lemna
lemna@forcings

# forcings can also be supplied as a named list
Lemna_Schmitt() \%>\% set_forcings(list(temp=20, rad=15000)) -> lemna
lemna@forcings
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_exposure.R
\name{set_noexposure}
\alias{set_noexposure}
\title{Set zero exposure}
\usage{
set_noexposure(x)
}
\arguments{
\item{x}{vector of \link{scenarios}}
}
\value{
vector of \link{scenarios}
}
\description{
The scenarios current exposure is replaced by a constant exposure time-series
of value zero(\code{0.0}). Output times are unaffected.
}
\examples{
# Derive effect size in sample scenario without toxicant exposure
minnow_it \%>\%
  set_noexposure() \%>\%
  effect()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_param.R
\name{get_param}
\alias{get_param}
\alias{get_param,ANY-method}
\alias{get_param,list-method}
\alias{get_param,EffectScenario-method}
\alias{get_param,ScenarioSequence-method}
\alias{get_param,ParameterSet-method}
\title{Get scenario parameters}
\usage{
get_param(x)

\S4method{get_param}{list}(x)

\S4method{get_param}{EffectScenario}(x)

\S4method{get_param}{ScenarioSequence}(x)

\S4method{get_param}{ParameterSet}(x)
}
\arguments{
\item{x}{object to fetch parameters from}
}
\value{
(list of) list(s) with key-value pairs
}
\description{
For scenario \link[=sequence]{sequences}, only the parameters of the first scenario
in the sequence is returned. To access parameters of a specific scenario in
the sequence, use \code{get_param()} on the individual scenario object.
}
\section{Methods (by class)}{
\itemize{
\item \code{get_param(list)}: Returns a list of parameter lists (if applicable)

\item \code{get_param(EffectScenario)}: Returns a list parameters for a single \link{scenario}

\item \code{get_param(ScenarioSequence)}: Returns a list of parameter lists, one for each scenario in the sequence

\item \code{get_param(ParameterSet)}: Returns a list of parameters for a single \link{parameter_set}

}}
\examples{
minnow_it \%>\% get_param()
}
\seealso{
\code{\link[=set_param]{set_param()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{set_tag}
\alias{set_tag}
\title{Set a tag}
\usage{
set_tag(x, tag)
}
\arguments{
\item{x}{(vector of) \link{scenario} objects}

\item{tag}{(vector of) \code{character} or any other object}
}
\value{
(vector of) modified \link{scenarios}
}
\description{
Sets the user-defined, custom tag of a scenario. Tags
can be helpful to quickly distinguish scenarios by e.g. a user-specified
string.
}
\details{
The function supports vectorized inputs.
If arguments \code{x} and \code{tag} are vectors, then tags are assigned to scenarios
on a 1:1 basis. If the length of both vectors do not match, an error is raised.
}
\examples{
# set a custom tag
myscenario <- GUTS_RED_SD() \%>\% set_tag("My Custom Tag")

# returns `My Custom Tag`
get_tag(myscenario)

# the tag also appears in the scenario overview
myscenario
}
\seealso{
\code{\link[=get_tag]{get_tag()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lik_profile.R
\name{lik_profile}
\alias{lik_profile}
\title{Likelihood profiling}
\usage{
lik_profile(
  x,
  par,
  output,
  data = NULL,
  bounds = NULL,
  refit = TRUE,
  type = c("coarse", "fine"),
  individual = FALSE,
  break_prof = FALSE,
  log_scale = FALSE,
  data_type = c("continuous", "count"),
  ...
)
}
\arguments{
\item{x}{either a single \link{scenario} or a list of \link{caliset} objects}

\item{par}{named vector - parameters (names and values) to be profiled}

\item{output}{character vector, name of output column of \code{\link[=simulate]{simulate()}} that
is used in calibration}

\item{data}{only needed if \code{x} is a \link{scenario}}

\item{bounds}{optional list of lists (including lower and upper bound): uses defaults in \code{x} object, but
can be overwritten here (e.g. bounds <- list(k_resp = list(0,10), k_phot_max = list(0,30)) )}

\item{refit}{if \code{TRUE} (default), refit if a better minimum is found}

\item{type}{\code{"fine"} or \code{"coarse"} (default) likelihood profiling}

\item{individual}{if \code{FALSE} (default), the log likelihood is calculated across
the whole dataset. Alternatively, if \code{TRUE}, log likelihoods are calculated for
each (group of) \emph{set}(s) individually.}

\item{break_prof}{If \code{TRUE}, then stop the profiling if a better optimum is located.
Default is \code{FALSE}.}

\item{log_scale}{\code{FALSE} (default), option to calculate the log likelihood on a
log scale (i.e., observations and predictions are log transformed during calculation)}

\item{data_type}{Character argument, \code{"continuous"} (default) or \code{"count"}, to specify the data type
for the log likelihood calculations.}

\item{...}{additional parameters passed on to \code{\link[=calibrate]{calibrate()}} and \code{\link[=simulate]{simulate()}}. To avoid
parameter confusion, use argument \code{method} to select optimization algorithms
of \code{calibrate()} and argument \code{ode_method} to select numerical integration
schemes of package \code{deSolve}.}
}
\value{
A list containing, for each parameter profiled, the likelihood
profiling results as a dataframe;
the 95\% confidence interval; the original parameter value; the likelihood plot object; and
the recalibrated parameter values (in case a lower optimum was found)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

The aim of the function is two-fold: 1) estimate a 95\% confidence
around each parameter of a calibrated model, and 2) see if perhaps a local minimum was found rather
than a global minimum. To achieve this, the likelihood profiling goes through
every parameter one by one. For each parameter,
the model is sequentially refit with the parameter value set to
increasingly lower and higher values, and the likelihood of the model given the
data calculated (using \code{\link[=log_lik]{log_lik()}}). The likelihood is then compared
to the likelihood of the original model (using a likelihood ratio). This leads
to the development of a likelihood profile, from which a plot a 95\%
confidence interval for the parameter is derived.

The idea of the function is a variable stepwise algorithm:
When the likelihood ratio changes very little (less than \code{l_crit_min}), the stepsize is
increased (up to a maximum, specified by \code{f_step_max}). When the lik.
ratio changes too much (more than \code{l_crit_max}), the algorithm tries again
with a smaller stepsize (also bound to a minimum: \code{f_step_min}). Note that
the stepsize is used as a fraction of the parameter value that is tried.
To prevent very small stepsizes when the value goes towards zero (as can
be the case for effect thresholds), an absolute minimum
stepsize (\code{f_step_abs}), which is specified as a fraction of the best
parameter value (\code{Xhat}) (unless it is zero, then the algorithm takes
something small).

Note that the likelihood of the model given the data can be calculated across
all datasets provided in the calibration set \code{x}, or calculated separately for
each individual dataset before being combined into one likelihood (by adjusting
the optional parameter \code{individual}). The latter
has the advantage that different datasets can be given different weights in
the likelihood calculation (using the "weight" slot of the \link{caliset} objects,\code{x}).
Further, for continuous data (e.g. biomass), the likelihood considers the variance (standard
deviation) in the log likelihood calculation, which can vary between datasets
when the likelihood is calculated for each dataset separately before combining
into an overall likelihood. The latter could be relevant when factors might lead
to variability between datasets (e.g. different labs, different animal culture,...)

To conduct the likelihood calculations on separate datasets, the parameter \code{individual}
which by default is 'FALSE' can be set to 'TRUE'. Then, then log likelihoods
are calculated for each dataset individually (or in subgroups, using the "tag" names of the
\link{caliset} object, if provided, to group datasets with the same "tag"
before calculating the log likelihood). Subsequently, the log
likelihoods for the subsets are combined into an overall likelihood (considering
the \emph{set} weights provided in the "weight" slot of the \link{caliset} object).
Note that for each \emph{set} only 1 weight can be provided (i.e. not individual
weights for each datapoint within the \emph{set}), and that \emph{set} with the same tag
should have identical weight.

The function was inspired by a MatLab BYOM v.6.8 procedure, created by
Tjalling Jager. For details, please refer to BYOM (http://debtox.info/byom.html)
as well as Jager (2021).
}
\examples{
# Example with Lemna model - physiological params
library(dplyr)

# observations - control run
obs <- schmitt2013 \%>\%
  filter(trial == "T0")

# update metsulfuron
myscenario <- metsulfuron \%>\%
  set_param(c(k_phot_fix = TRUE, Emax = 1)) \%>\%
  set_init(c(BM = 0.0012)) \%>\%
  set_noexposure() \%>\%
  set_bounds(list(k_phot_max=c(0, 1)))

fit <- calibrate(
  x = myscenario,
  par = c(k_phot_max = 1),
  data = obs,
  output = "FrondNo",
  method = "Brent"
)

# Likelihood profiling
\donttest{
res <- lik_profile(
  x = myscenario,
  data = obs,
  output = "FrondNo",
  par = fit$par,
  bounds = list(
    k_phot_max = list(0, 30)
  ),
  refit = FALSE,
  type = "fine",
  method = "Brent"
)
# plot
plot(res)
}

}
\references{
Jager T, 2021: Robust Likelihood-Based Optimization and Uncertainty Analysis
of Toxicokinetic-Toxicodynamic Models. Integrated Environmental Assessment and
Management 17:388-397. \doi{10.1002/ieam.4333}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{rsubcapitata}
\alias{rsubcapitata}
\title{A Weber scenario of algae exposed to isoproturon}
\format{
An object of class \code{AlgaeWeber} of length 1.
}
\usage{
rsubcapitata
}
\description{
The scenario recreates the conditions of algae species \emph{R. subcapitata} exposed
to isoproturon in a flow-through reactor experiment, see Weber et al. (2012)
publication. Scenario parameters and exposure series were set according to
reported values to represent conditions in reactor A, as depicted in
Figure 4A in Weber et al.
}
\references{
Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
and Ratte HT, 2012. Combination of a higher-tier flow-through system and
population modeling to assess the effects of time-variable exposure of
isoproturon on the green algae Desmodesmus subspictatus and
Pseudokirchneriella subcapitata. Environmental Toxicology and
Chemistry, 31, 899-908. \doi{10.1002/etc.1765}
}
\seealso{
\link{Algae_Weber}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tox_data.R
\name{tox_data}
\alias{tox_data}
\alias{ToxData-class}
\title{Prepare ecotox study data for fitting}
\usage{
tox_data(data, exposure = NULL)
}
\arguments{
\item{data}{a \code{data.frame} with at least two and at most four columns; the first
column must represent time, the second an observed quantity, the optional third
a trial or treatment ID, and the optional fourth the concentration during
the experiment}

\item{exposure}{an optional named list; names must correspond to trial IDs used for
the \code{data} argument; values can be numeric constants, data.frames, or an
\link{ExposureSeries} object}
}
\value{
a \code{ToxData} object
}
\description{
Takes ecotox study data in \emph{long-form} tabular format and prepares it for
parameter fitting. It supports extracting (optional) exposure concentration from
tabular data (useful for e.g. studies of acute toxicity) or exposure series
can also provided as individual time-series.
}
\section{Tabular format}{
The long-form tabular data must have at least two and at most four columns.
The position of the columns define what they represent, the column names
are ignored:
\itemize{
\item First column: time
\item Second column: observed quantity, e.g. number of individuals
\item (optional) Third column: Trial or treatment ID
\item (optional) Fourth column: Concentration
}

The first two columns, time and observed quantity, must always be present.
The third column, trial ID, is used to split the table by treatment so
that trials can later be handled individually. The fourth column, concentration,
can be used to also define the exposure level during the experiments.
}

\section{Explicit exposure series}{
As an alternative to defining concentrations along observed data, exposure
can also be passed as a list of exposure levels and series with
argument \code{exposure}. It can be used to provide exposure series for each
trial. The following object types are supported to define exposure:
\itemize{
\item Numerical constants
\item Tabular data, e.g. \code{data.frame}s
\item \link[=ExposureSeries]{exposure series} objects
}

If exposure is constant over time, exposure can be defined using a single
constant value. More complex exposure time-series can be defined using e.g.
\code{data.frame}s. Tabular data must have two columns with the first column
representing time and the second column representing exposure/concentrations.
}

\examples{
library(dplyr)

mydata <- schmitt2013 \%>\% tox_data()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-algae.R
\name{Algae_Simple}
\alias{Algae_Simple}
\alias{AlgaeSimple-class}
\alias{AlgaeSimpleScenario-class}
\title{Simple algae model without environment (Rendal et al. 2023)}
\usage{
Algae_Simple()
}
\value{
an S4 object of type \linkS4class{AlgaeSimple}
}
\description{
The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
growth model for algae. It follows the concept of a simplified algae model
described in Rendal et al. (2023). The model simulates the development of
algal biomass. The growth of the algae population is simulated on the basis
of growth rates, which are, in contrast to the Weber model, independent on
environmental conditions which are usually optimal in laboratory effect studies.
The toxicodynamic sub-model describes the effects of growth-inhibiting
substances through a corresponding reduction in the photosynthesis rate on the
basis of either external or internal concentrations (depending on user choice
of \emph{scaled} parameter setting).
}
\section{State variables}{

The model has two state variables:
\itemize{
\item \code{A}, Biomass (µg fresh wt/mL, cells/mL *10^4)
\item \code{Dw}, scaled internal damage, only takes effect if parameter \code{scaled = 1}
}
}

\section{Model parameters}{

\itemize{
\item Growth model
\itemize{
\item \code{mu_max}, Maximum growth rate (d-1)
}
\item Concentration response (Toxicodynamics)
\itemize{
\item \code{EC_50}, Effect concentration of 50\% inhibition of growth rate (µg L-1)
\item \code{b}, slope of concentration effect curve at EC_50 (-)
\item \code{dose_response}, shape of the dose response curve (0 = logit, 1 = probit)
}
\item External concentration (Toxicokinetics)
\itemize{
\item \code{scaled}, 1 = scaled internal damage, 0 = no internal damage / 1 = yes (-)
\item \code{kD}, dominant rate constant of toxicant in aquatic environments (d-1), only
takes effect if parameter \code{scaled = 1}
}
}
}

\section{Forcings}{

Simplified model without additional forcings for e.g. irradiation or temperature
as implemented in \code{Algae_Weber}. A constant growth over time is assumed.
In case that growth is time dependent, a forcing variable (f_growth) can be set.
Forcing time-series are represented by \code{data.frame} objects consisting of two
columns. The first for time and the second for a scaling factor of mu_max.
The input format for all forcings is a list of the data frames. If f_growth is
not set, a default scaling factor of 1 is used.
}

\section{Parameter boundaries}{

Upper and lower parameter boundaries are set by default for each parameter. This,
to avoid extreme values during calibration (particularly likelihood profiling)
}

\section{Simulation output}{

Simulation results will contain the state variables biomass (\code{A}) and
scaled damage concentration (\code{Dw}).

It is possible to amend the output of \code{\link[=simulate]{simulate()}} with additional model
quantities that are not state variables, for e.g. debugging purposes or to
analyze model behavior. To enable or disable additional outputs, use the
optional argument \code{nout} of \code{\link[=simulate]{simulate()}}. As an example, set \code{nout=2} to
enable reporting of external concentration (\code{Cw}) and growth scaling factor
(\code{f_growth}). Set \code{nout=0} to disable additional outputs (default).

The available output levels are as follows:
\itemize{
\item \code{nout >= 1}: \code{Cw} external concentration (µg L-1)
\item \code{nout >= 2}: \code{f_growth} growth scaling factor (-)
\item \code{nout >= 3}: \code{dA}, biomass derivative (µg)
\item \code{nout >= 4}: \code{dDw}, damage concentration derivative (µg L-1)
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{hmax = 0.01}\if{html}{\out{<br>}}
Maximum step length in time suitable for most simulations.
}
}

\references{
Rendal C, Witt J, Preuss TG, Ashauer R, 2023: A Framework for Algae Modeling
in Regulatory Risk Assessment. Environ. Toxicol. Chem. 42(8), pp. 1823-1838.
\doi{10.1002/etc.5649}

Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
and Ratte HT, 2012: Combination of a higher-tier flow-through system and
population modeling to assess the effects of time-variable exposure of
isoproturon on the green algae Desmodesmus subspictatus and
Pseudokirchneriella subcapitata. Environ. Toxicol. Chem. 31(4), pp. 899-908.
\doi{10.1002/etc.1765}
}
\seealso{
\link{Scenarios}, \link{Transferable}

Other algae models: 
\code{\link{Algae-models}},
\code{\link{Algae_TKTD}()},
\code{\link{Algae_Weber}()}
}
\concept{algae models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-guts_red.R
\name{GUTS-RED-models}
\alias{GUTS-RED-models}
\title{GUTS-RED models}
\description{
Reduced \emph{General Unified Threshold models of Survival} (GUTS) with stochastic
death (\emph{SD}) and individual tolerance (\emph{IT})
}
\details{
The TKTD models \emph{\link[=GUTS_RED_SD]{GUTS-RED-SD}} and \emph{\link[=GUTS_RED_IT]{GUTS-RED-IT}}
were described by EFSA (2018).
GUTS-RED models assume a one-compartment model which directly links
external concentration to the scaled damage. The scaled damage is given in
units of concentration, equal to the units of measurement in the external
medium, e.g. ug/L. The damage dynamics is connected to an individual hazard
state variable, resulting in simulated mortality when an internal damage
threshold is exceeded. The death mechanisms stochastic death (\emph{SD}) and
individual threshold (\emph{IT}) are extreme cases of the \emph{GUTS} theory.

For \emph{SD} models, the threshold parameter for lethal effects is fixed and
identical for all individuals of a group, meaning that the variance of the
threshold values is zero. Hence, the killing rate relates the probability of
a mortality event in proportion to the scaled damage. For \emph{IT} models,
the thresholds for effects are distributed among individuals of a group.
Mortality of an individual follows immediately once the individual's tolerance
is exceeded. Meaning in model terms that the killing rate is set to
infinity (EFSA 2018).
}
\section{State variables}{

The following list describes the default names and standard units of \emph{GUTS-RED}
state variables:
\itemize{
\item \code{D}, scaled damage (conc)
\item \code{H}, cumulative hazard (-)
}

The state variables are initialized with zero by default.
}

\section{SD model parameters}{

\itemize{
\item \code{kd}, dominant rate constant (time^-1)
\item \code{hb}, background hazard rate (time^-1)
\item \code{z}, threshold for effects (conc)
\item \code{kk}, killing rate constant (time^-1)
}
}

\section{IT model parameters}{

\itemize{
\item \code{kd}, dominant rate constant (time^-1)
\item \code{hb}, background hazard rate (time^-1)
\item \code{alpha}, median of thresholds (conc)
\item \code{beta}, shape parameter (-)
}
}

\section{Effects}{

The effect endpoint \code{L} (lethality) is available for \emph{GUTS-RED} models.
A value of zero (\code{0.0}) denotes \emph{no effect} on organism survival. A value of
one (\code{1.0}) denotes a lethality rate of 100\%, i.e. no survivors.

The survival probability \code{S} is available in the return value of \code{\link[=simulate]{simulate()}}.
}

\references{
EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, et al., 2018: \emph{Scientific Opinion on the
state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
regulatory risk assessment of pesticides for aquatic organisms}. EFSA Journal 2018;
16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
}
\seealso{
Other GUTS-RED models: 
\code{\link{GUTS_RED_IT}()},
\code{\link{GUTS_RED_SD}()}

Other models: 
\code{\link{Algae-models}},
\code{\link{DEB-models}},
\code{\link{Lemna-models}},
\code{\link{Macrophyte-models}}
}
\concept{GUTS-RED models}
\concept{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.r
\name{plot_epx}
\alias{plot_epx}
\title{Plot EPx values}
\usage{
plot_epx(
  EPx_ts,
  exposure_ts,
  draw = TRUE,
  time_col = "time",
  conc_col = "conc",
  epx_x_title = "Start time",
  conc_y_title = "Exposure conc."
)
}
\arguments{
\item{EPx_ts}{the result of \code{epx_mtw}, ie. a tibble with window.start,
window.end, endpoint, level and EPx}

\item{exposure_ts}{an exposure time series with columns for time 'time' and
concentration 'conc'}

\item{draw}{Should the whole plot be drawn? If FALSE the exposure plot and
the EPx plot are returned as a list for later modification}

\item{time_col}{the name of the time column in the exposure dataset}

\item{conc_col}{the name of the concentration column in the exposure dataset}

\item{epx_x_title}{title of the x-axis of the epx panel}

\item{conc_y_title}{title of the y-axis of the concentration panel}
}
\value{
a grid of ggplots
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
ti <- 0:21
expo <- abs(0.01*ti + rnorm(length(ti), 0, 0.05))
exposure <- data.frame(time = ti, conc = expo)
metsulfuron_epx_mtw <- metsulfuron \%>\%
set_exposure(exposure) \%>\%
epx_mtw(level = 10, factor_cutoff = 1000)
metsulfuron_epx_mtw
plot_epx(EPx_ts = metsulfuron_epx_mtw,
exposure_ts = exposure, conc_y_title = "env. concentration [µg/L]")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dose_response.R
\name{dose_response}
\alias{dose_response}
\title{Calculate a dose response curve}
\usage{
dose_response(
  scenario,
  range = c(1, 99),
  n = 20,
  strategy = c("exponential", "decadic", "vanilla"),
  verbose = FALSE,
  ...
)
}
\arguments{
\item{scenario}{used for calculation}

\item{range}{numeric vector specifying the required range of effect levels in
percent (\%), defaults to \code{c(1,99)}}

\item{n}{minimum number of points on the dose response curve}

\item{strategy}{controls how multiplication factors are chosen, \code{vanilla} uses a fixed
set of multiplication factors, \code{decadic} and \code{exponential} have varying step lengths.}

\item{verbose}{logical, set to \code{TRUE} for additional status messages}

\item{...}{additional arguments passed on to \code{\link[=effect]{effect()}}}
}
\value{
\code{data.frame} with two columns, i.e. \code{mf} and \code{effect}
}
\description{
Returns a \code{data.frame} with points on the dose response curve for the given
effect scenario.
}
\details{
Derives a dose response curve from a \link{scenario}. The result will
cover the requested range of effect levels. The tested multiplication factors
can be chosen by different strategies, i.e. a \code{vanilla} approach using a
fixed set of factors, or \code{decadic} and \code{exponential} approaches
employing logarithmic and exponential factor scaling, respectively.
}
\examples{
# basic dose response curve
minnow_sd \%>\% dose_response()

# modify the minimum number of points on the curve
minnow_sd \%>\% dose_response(n=10)

# select a subset of the effect range
minnow_sd \%>\% dose_response(range=c(10,20))

# use an alternative strategy for the selection of multiplication factors
minnow_sd \%>\% dose_response(strategy="decadic")

# provide additional output how multiplication factors were selected
minnow_sd \%>\% dose_response(verbose=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_transfer.R
\name{set_transfer}
\alias{set_transfer}
\alias{set_transfer,ANY-method}
\alias{set_transfer,Transferable-method}
\alias{set_notransfer}
\title{Set transfer events}
\usage{
set_transfer(x, interval, times, biomass, scaled_comp)

\S4method{set_transfer}{ANY}(x, interval, times, biomass, scaled_comp)

\S4method{set_transfer}{Transferable}(x, interval, times, biomass, scaled_comp)

set_notransfer(x)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}

\item{interval}{optional \code{numeric}, interval in time units of the scenario,
set to \code{-1} to disable transfers.}

\item{times}{optional \code{numeric} vector of time points where transfers occur}

\item{biomass}{optional \code{numeric} vector, amount of biomass that is being
transferred at each transfer}

\item{scaled_comp}{optional \code{character} vector of affected compartments
that are scaled according to new biomass levels}
}
\value{
Modified \link{scenario} objects
}
\description{
A \emph{transfer} refers to an event where a certain amount of biomass
is moved to a new medium after a period of time. Effectively, this resets
the scenario's state variable representing biomass and re-scales all state
variables which are correlated with biomass, such as adsorbed chemical mass.
This feature replicates a procedure occurring e.g. in \emph{Lemna} effect studies
and may be necessary to recreate study results.
}
\details{
If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
\subsection{Transferred biomass}{

At each transfer, a defined amount of biomass is transferred to a new medium.
This is modeled by interrupting the simulation at a transfer time point, modifying
the biomass level \code{BM}, and scaling affected compartments according to new
biomass levels. Scaling of compartments depending on biomass, such as
\emph{internal toxicant mass}, is necessary to correctly reflect mass balances and
concentrations over time.

Transferred biomass is set using the \code{biomass} parameter. Is is either a
single numerical value in which case the same biomass level is set at each
transfer. Or it is a vector of numerical values with the same length as the
\code{times} parameter in which case a custom biomass level can be set for each
transfer. Multiple biomass levels can only be set in conjunction with
custom transfer time points.

Some scenario types define default values for transferred biomass based
on common study set ups.
}

\subsection{Regular and custom transfer time points}{

Transfers can occur either in regular intervals of time or at selected, custom
time points. For regular intervals, the parameter \code{interval} is set to a single
numeric value which has the same unit as the scenario's time dimension. As an
example: if a scenario uses the unit of \emph{days} for time, the transfer interval
is also specified in \emph{days}:

Transfers occurring at custom time points are set by passing a numerical vector
to the parameter \code{times}. The time points' units must match with the unit of
time in the scenario. A custom transfer time point \strong{must not occur at the
starting time point of a simulation}.
}

\subsection{Affected compartments}{

Some compartments depend on biomass to correctly reflect mass balances and
concentrations over time, such as \emph{internal toxicant mass}. These compartments
need to be scaled linearly to reflect the change in biomass levels.
The parameter \code{scaled_comp} accepts a character vector of compartment names
which are scaled at each transfer. This parameter should only be used with
custom, user-defined models. If no compartment needs to be scaled, set or
use the default value of \code{character(0)}.
}
}
\section{Functions}{
\itemize{
\item \code{set_notransfer()}: Disable biomass transfers

}}
\examples{
# Simulate biomass transfer of 50 *g/m²* at a regular interval of 7 *days*
metsulfuron \%>\%
  set_transfer(interval=7, biomass=50) \%>\%
  simulate()

# Simulate irregular biomass transfers occuring at days 5, 10, and 12
metsulfuron \%>\%
  set_transfer(times=c(5, 10, 12), biomass=50) \%>\%
  simulate()

# Simulate irregular transfers with changing amounts of transferred biomass
metsulfuron \%>\%
  set_transfer(times=c(5, 10, 12), biomass=c(50, 20, 10)) \%>\%
  simulate()

# Disable all biomass transfers
metsulfuron \%>\%
  set_notransfer() \%>\%
  simulate()
}
\seealso{
\link{Lemna-models}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R
\name{effect}
\alias{effect}
\alias{effect,EffectScenario-method}
\alias{effect,ScenarioSequence-method}
\title{Effect level}
\usage{
effect(x, ...)

\S4method{effect}{EffectScenario}(x, factor = 1, max_only = TRUE, ep_only = FALSE, marginal_effect, ...)

\S4method{effect}{ScenarioSequence}(x, ...)
}
\arguments{
\item{x}{a \link{scenario} objects}

\item{...}{additional parameters passed on to \code{\link[=simulate]{simulate()}}}

\item{factor}{optional numeric value which scales the exposure time-series}

\item{max_only}{\code{logical}, if \code{TRUE} only the maximum effect is returned, else
results for all effect windows are reported}

\item{ep_only}{logical, if TRUE only effect endpoints are returned as a vector}

\item{marginal_effect}{\code{numeric}, if set, any effect smaller than this threshold will
be reported as zero to exclude pseudo-effects originating from small numerical
errors}
}
\value{
a \code{tibble}, by default containing scenarios, effect levels, and the
exposure window where the maximum effect level occurred. The number of columns
depends on the enabled effect endpoints and function arguments.

By default, the first column, named \code{scenarios}, contains the original scenario
objects that were the basis of the calculation. For each effect endpoint, it
will be followed by one column with the maximum effect level and two columns
containing start and end time of the associated exposure window. If exposure
windows are disabled, the columns will just contain the start and end time of
the simulation. The effect level column will have the name of the effect
endpoint, start and end time will additionally have the suffixes \code{.dat.start}
and \code{.dat.end}, respectively.
}
\description{
Derives the effect level due to toxicant exposure in the supplied scenarios.
Either relative to a control scenario or derived directly from model
endpoints, depending on model type. For scenarios with moving exposure windows,
the maximum effect is returned.
}
\details{
By default, only the maximum effect in all moving exposure windows will
be returned. If argument \code{max_only=FALSE} is set, the returned table will
be converted to long-format and will contain effect levels for each
assessed exposure window.
\subsection{Calculation}{

Effects are calculated similarly to \emph{relative errors}, i.e. the difference
between control and treatment scenarios is divided by the absolute value
of the control. Effects are usually in the interval \verb{[0,1]}, but values
larger than one or smaller than zero can occur. As a special case, if the
endpoint from the control scenario is zero, then the effect is either
\itemize{
\item zero, if also the treatment is zero
\item positive infinity, if the treatment is smaller than zero
\item negative infinity, if the treatment is greater than zero
}

As an example, a control scenario achieves a biomass of
1.0 and the treatment scenario achieves a biomass of 0.9, the effect will
be equal to 0.1 or 10\%. However, effects can take on any real value. If,
for example, the biomass of the previously mentioned treatment scenario drops
below zero, then an effect larger than 1.0 will be calculated If, instead,
the biomass in the treatment scenario is greater than in the control, then the
effect will be negative.
\subsection{Output formatting}{

Start and end time of exposure windows can be disabled by setting \code{ep_only=TRUE}.
Effect levels smaller than a certain threshold can be automatically set to
zero (\code{0.0}) to avoid spurious effect levels introduced by numerical errors.
Set \code{marginal_effect} to an adequate value less than 1\%.
}

\subsection{Computational efficiency}{

Calculations can be sped up by providing a \code{data.frame} of pre-calculated
control scenarios for each assessed time window. As control scenarios are
by definition independent of any exposure multiplication factor, they can
be reused for repeated calculations, e.g. to derive effect profiles or
dose-response relationships.
}

}
}
\section{Methods (by class)}{
\itemize{
\item \code{effect(EffectScenario)}: Default for all generic \link{scenarios}

\item \code{effect(ScenarioSequence)}: For scenario \link[=sequence]{sequences}

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_deb}
\alias{is_deb}
\alias{is_DEB}
\title{Test if argument is a DEB scenario}
\usage{
is_deb(x)

is_DEB(x)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}
}
\value{
vector of \code{logical} values
}
\description{
Test if argument is a DEB scenario
}
\section{Functions}{
\itemize{
\item \code{is_DEB()}: Deprecated alias.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_scenario}
\alias{is_scenario}
\title{Test if argument is an effect scenario}
\usage{
is_scenario(x)
}
\arguments{
\item{x}{Some value or object}
}
\value{
vector of \code{logical} values
}
\description{
Supports vectorized arguments.
}
\examples{
# returns `TRUE`
is_scenario(minnow_it)

# returns `FALSE`
is_scenario(list())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{set_mode_of_action}
\alias{set_mode_of_action}
\alias{set_moa}
\title{Set mode of action}
\usage{
set_mode_of_action(x, code)

set_moa(x, code)
}
\arguments{
\item{x}{vector of \link{scenarios}}

\item{code}{a code for a mode of action, refer to model description for
details}
}
\value{
modified \link{scenarios}
}
\description{
Updates the model parameter \code{MoA} to a certain value
}
\section{Functions}{
\itemize{
\item \code{set_moa()}: Shorthand version

}}
\examples{
# Set MoA=8, i.e. hazard during oogenesis
americamysis \%>\%
  set_mode_of_action(8) \%>\%
  effect(method="ode45")

# alternative approach using the parameter directly
americamysis \%>\%
  set_param(c(MoA=8)) \%>\%
  effect(method="ode45")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{simulate}
\alias{simulate}
\alias{simulate,EffectScenario-method}
\alias{simulate,Transferable-method}
\alias{simulate,ScenarioSequence-method}
\alias{simulate,SimulationBatch-method}
\title{Simulate an effect scenario}
\usage{
simulate(x, ...)

\S4method{simulate}{EffectScenario}(x, ...)

\S4method{simulate}{Transferable}(x, ...)

\S4method{simulate}{ScenarioSequence}(x, ...)

\S4method{simulate}{SimulationBatch}(x, ...)
}
\arguments{
\item{x}{\link{scenario} to simulate}

\item{...}{additional parameters passed on to ODE solver}
}
\value{
A \code{data.frame} with the time-series of simulation results
}
\description{
The supplied \code{EffectScenario} is passed on to the ODE solver for numerical
integration. Internally, \code{simulate()} is split up into several functions
dedicated to particular models, e.g. one for GUTS and one for Lemna type
models. The package will take care of using the correct function
for each model when \code{simulate()} is called.
}
\details{
Simulation results are returned as a time-series for each state
variable. Some models provide additional properties describing the model state,
e.g. the internal concentration of a toxicant within the organism. Refer
to the respective \link{scenario} for more information.

Additional arguments to \code{simulate()} will be passed on to \code{\link[deSolve:ode]{deSolve::ode()}}
which enables control of the numerical integration parameters.
\subsection{Output times and windows}{

The minimum and maximum of given time points define the simulated
period. However, the simulation can also be limited to
a subset of time points by enabling a moving exposure window, see \code{\link[=set_window]{set_window()}}.

Results will be returned for each output time point. Precision of the numeric solver
may be affected by chosen output times in certain cases. Hence, small deviations
in results should be expected if different output times are set. This effect
can be mitigated by either defining are sufficiently small time step for the solver
using argument \code{hmax} or by decreasing the error tolerances \code{atol} and \code{rtol}.
These arguments are passed to the solver, see e.g. \code{\link[deSolve:lsoda]{deSolve::lsoda()}} for details.
}

\subsection{Optional output variables}{

Some models support adding intermediary model variables to the return value
of \code{simulate()}. Analyzing the additional outputs may be helpful to understand
model behavior and support finding potential issues in model parameterization.

Optional outputs are enabled by setting the parameter \code{nout} to a value greater
than zero. If \code{nout} is set to \code{n}, then the first \code{n} optional output columns
will be returned along the normal simulation result.

Which optional outputs are available depends on the model/scenario at hand.
Please refer to the model documentation for details.
As an example, the \link[=GUTS_RED_IT]{GUTS-RED-IT} model supports adding the
external toxicant concentration to the output by setting \code{nout=1}:

\code{minnow_it \%>\% simulate(nout=1)}
}

\subsection{Numerical precision and stability}{

Each model was assigned a default ODE solver which handles most of the
occurring inputs well. In most cases, this will be an explicit numerical
scheme of the Runge-Kutta family with variable step width. For certain extreme
parameters settings, such as very high uptake/permeability of the contaminant
or exposure series which represent step functions, the numerical approximation
might deteriorate and yield invalid results. In this case try to decrease the
allowed max step width by setting the argument \code{hmax} with various values.
Start with \code{hmax=1} and decrease the value by orders of 10. It is not
possible or desirable to reduce \code{hmax} to extremely small values, as the
ODE solver will require more CPU time and simulation will become inefficient.

Oftentimes, it will be computational more efficient to adapt the solver's
error tolerances \code{atol} and \code{rtol} than reducing the step width \code{hmax} to achieve
stable numerics. Start by decreasing deSolve's default values by
orders of ten until the simulation yields acceptable results, see e.g.
\code{\link[deSolve:lsoda]{deSolve::lsoda()}} for more information on error tolerances.

As an alternative to adapting solver parameters, it might be worthwhile to
try other numerical schemes which can handle stiff ODEs, such as Radau, LSODA,
or LSODES. To change solvers, set the \code{method} argument.
To select e.g. the Radau scheme, set \code{method="radau"}. For LSODA, set \code{method="lsoda"}.
Radau performs better than LSODA in some cases, as the latter method can return
biologically nonsensical results without raising an error. See \code{\link[deSolve:ode]{deSolve::ode()}}
for details on available ODE solvers.
}
}
\examples{
# base R syntax
simulate(minnow_sd)
# tidy syntax with the same result
minnow_sd \%>\% simulate()

# Extend the simulated time frame to the interval [0, 10]
minnow_sd \%>\%
  set_times(seq(0, 10)) \%>\%
  simulate()

# Use an alternative exposure profile, but keep the original output times
minnow_sd \%>\%
  set_exposure(data.frame(t=0, c=10), reset_times=FALSE) \%>\%
  simulate()

##
## Precision of results

# A large number of output times forces smaller solver time steps
minnow_it \%>\%
  set_times(seq(0, 10, 0.001)) \%>\%
  simulate() \%>\%
  tail()

# Defining only two output times allows the ODE solver to make larger steps
# in time during numerical integration. However, results can become
# imprecise.
minnow_long <- minnow_it \%>\% set_times(c(0, 10))
minnow_long \%>\% simulate()

# Numerical precision of results can be increased by limiting the solver's
# maximum step length in time using argument `hmax`.
minnow_long \%>\% simulate(hmax=0.005)

# A similar numerical precision can be achieved by switching to an alternative
# numerical integration scheme, such as the Radau scheme, without limiting
# the step length.
minnow_long \%>\% simulate(method="radau")

# Reducing the step length even further may increase numerical precision, but
# may exceed the solver's allowed number of integration steps per output interval.
# The following simulation will be aborted with a solver error:
try(
  minnow_long \%>\% simulate(hmax=0.001)
)

# However, the solver's maximum number of allowed steps can be increased,
# if needed, using the argument `maxsteps`:
minnow_long \%>\% simulate(hmax=0.001, maxsteps=10^5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_growth.R
\name{ci_from_hessian}
\alias{ci_from_hessian}
\title{Parameter confidence intervals from fit}
\usage{
ci_from_hessian(fit, dof, level = 0.95)
}
\arguments{
\item{fit}{return value from \code{\link[=calibrate]{calibrate()}}}

\item{dof}{integer, Degrees of Freedom, commonly the number of independent observations
minus the number of fitted parameters}

\item{level}{numeric, desired confidence level, i.e a value between zero and one.
Defaults to 95\% (\code{0.95}).}
}
\value{
name list of numeric vectors of length two, the elements
representing the lower and upper confidence limit, respectively.
}
\description{
Calculates parameter confidence intervals for a fit returned by \code{\link[=calibrate]{calibrate()}}.
The fit must provide a valid \emph{Hessian} matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch.R
\name{batch}
\alias{batch}
\title{Batch simulation of multiple exposure levels}
\usage{
batch(
  scenario,
  exposure,
  id_col = "trial",
  format = c("long", "wide"),
  times_from = c("scenario", "exposure"),
  select = NULL
)
}
\arguments{
\item{scenario}{a \link{scenario} object}

\item{exposure}{a named \code{list()} or a \code{data.frame} with three columns}

\item{id_col}{\code{character}, name of column in resulting ´data.frame` which
contains a trial's name or ID}

\item{format}{\code{character}, set to \code{'long'} for long tabular format, or
\code{'wide'} for wide format}

\item{times_from}{\code{character}, set to \code{'scenario'} to use output times from
scenario, or \code{'exposure'} to take output times from each exposure series}

\item{select}{optional \code{character} vector to select columns from the
simulation output}
}
\value{
a simulation batch object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\details{
A convenience function to simulate a single base scenario with one or more
exposure levels. The functions aims at reproducing the setup and result format
of common effect studies.

Simulating a \link{scenario} is generally limited to assessing a single exposure
series. However, laboratory experiments commonly examine the effects of
multiple exposure levels on a biological system.
A \emph{batch simulation} approach involves running multiple simulations with
varying exposure or \emph{treatment} conditions. To illustrate: if the objective is
to examine the impact of a chemical on cell growth, multiple scenarios need
to be simulated to reproduce the cell growth dynamics under varying
concentrations of the assessed chemical. Each simulation run will represent a
specific exposure level, ranging from low to high concentrations of the
chemical.

To simulate the conditions of such a laboratory experiment, the scenarios and
exposure levels can either be created and simulated individually, or the
\code{batch()} function can be used for ease of use.
\subsection{Exposure series}{

The set of exposure levels can be represented by one of the following types:
\itemize{
\item A (named) list: Each element represents an exposure level or exposure
series. An exposure level can be represented by a constant numeric, a
\code{data.frame} with two columns, or an ExposureSeries object. The names of
the list elements specify the study ID.
\item Or alternatively, a \code{data.frame} with three columns: One column for time,
one for the exposure level, and one character column to specify the
study IDs.
}

Each exposure level will be simulated using the base scenario. If the exposure
levels are provided as a named list, the names will also appear in the return
value of \code{simulate()}. This behavior can be used, for example, to define unique
study IDs for particular exposure levels.
}

\subsection{Exposure IDs}{

The list of exposure levels can be supplied as a named list. The names
will be used as unique (study) IDs, so that the simulation results belonging
to any exposure level can be identified in the output.
If no IDs are defined by the user, generic IDs of the form \code{'trial{n}'} will
be assigned, with \code{{n}} being replaced by consecutive integers starting at one.

If the batch is passed on to \code{\link[=simulate]{simulate()}}`, the IDs will be contained in its
return value, e.g. as a dedicated column (long format) or as part of the
column names (wide format).
}

\subsection{Output format}{

The return value of \code{\link[=simulate]{simulate()}} is by default in long format, i.e. it will
contain one row for each output time and exposure level. It is possible to
pivot the tabular data to wide format, by setting the argument
\code{format = 'wide'}.

In wide format, the output columns of each exposure level are pasted
next to each other. If more than one column is pivoted per exposure level,
then the exposure or study ID is added as a suffix to column names.
If the output per exposure level contains only a single column (besides time
and the exposure ID itself), then original column name is dropped and only
exposure IDs are used. See the examples section for reference.
}

\subsection{Select output columns}{

Often, only a single output column is of interest in batch simulations,
such as the number of surviving individuals. To ease the interpretation and
handling of the output of batch simulations, the columns contained in the
output of each simulated exposure level can be filtered. One or more columns
can be selected. By default, no filtering of output columns is conducted.

As an example, to create an overview of survival probabilities (\emph{S}) in the
GUTS-RED-IT example scenario \code{minnow_it}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{minnow_it \%>\%
  batch(exposure=list(0, 5, 10), select="S", format="wide") \%>\%
  simulate()
}\if{html}{\out{</div>}}
}
}
\examples{
# Simulate a batch experiment with three constant exposure levels of
# 0.0, 2.0, and 5.0 µmol/L
simulate(batch(minnow_it, list(0, 2, 5)))

# Alternatively, in tidyr style syntax
trials_list1 <- list(0, 2, 5)
minnow_it \%>\%
  batch(trials_list1) \%>\%
  simulate()

# Assign unique IDs to each exposure level
trials_list2 <- list(Control=0, TrialA=2, TrialB=5)
minnow_it \%>\%
  batch(trials_list2) \%>\%
  simulate()

# Alternatively, define multiple exposure levels in a single data.frame
trials_table <- data.frame(time=c(0, 0, 0),
                           conc=c(0, 2, 5),
                           trial=c("Control", "TrialA", "TrialB"))
minnow_it \%>\%
  batch(trials_table) \%>\%
  simulate()

# Limit simulation output to column 'S' (survival probability)
minnow_it \%>\%
  batch(trials_list2, select="S") \%>\%
  simulate()

# Return data in wide-format, unique IDs will be used as column names
minnow_it \%>\%
  batch(trials_list2, select="S", format="wide") \%>\%
  simulate()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_toxswa.R
\name{import_swash}
\alias{import_swash}
\title{SWASH project exposure profile import}
\usage{
import_swash(swash_dir, ...)
}
\arguments{
\item{swash_dir}{path to the SWASH project directory}

\item{...}{arguments passed on to \code{\link[=import_toxswa]{import_toxswa()}}}
}
\value{
a list of imported exposure series, see \code{\link[=import_toxswa]{import_toxswa()}} for details
}
\description{
Read all TOXSWA files within a SWASH project directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-guts_red.R
\name{GUTS_RED_IT}
\alias{GUTS_RED_IT}
\alias{GutsRedIt-class}
\alias{GUTS-RED-IT}
\title{GUTS-RED-IT scenario}
\usage{
GUTS_RED_IT(param, init)
}
\arguments{
\item{param}{optional named \code{list} or \code{vector} with model parameters}

\item{init}{optional named numeric \code{vector} to use as initial state}
}
\value{
an S4 object of type \linkS4class{GutsRedIt}
}
\description{
Reduced \emph{General Unified Threshold models of Survival} (GUTS) with
individual tolerance (\emph{IT}).
}
\section{Simulation output}{

The return value of \code{\link[=simulate]{simulate()}} will contain values for the state variables,
as well as an additional column \code{S} which represents the survival probability
for each time point. \code{S} is calculated as described in EFSA (2018) as
\emph{S = (1- F(t))}. The background hazard rate \code{hb} is already considered in state
variable \code{H} and therefore does not occur as an additional term to derive \code{S}.
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.
}

\section{Model history and changes}{

\itemize{
\item cvasi v1.0.0
\itemize{
\item Model and parameters as described by EFSA PPR (2018).
}
\item cvasi v1.5.0
\itemize{
\item Cumulative damage over time will be calculated on ODE level, instead on
results of \code{\link[=simulate]{simulate()}}, to avoid influence of selected output times
on predicted survival.
}
}
}

\section{State variables}{

The following list describes the default names and standard units of \emph{GUTS-RED}
state variables:
\itemize{
\item \code{D}, scaled damage (conc)
\item \code{H}, cumulative hazard (-)
}

The state variables are initialized with zero by default.
}

\section{IT model parameters}{

\itemize{
\item \code{kd}, dominant rate constant (time^-1)
\item \code{hb}, background hazard rate (time^-1)
\item \code{alpha}, median of thresholds (conc)
\item \code{beta}, shape parameter (-)
}
}

\section{Effects}{

The effect endpoint \code{L} (lethality) is available for \emph{GUTS-RED} models.
A value of zero (\code{0.0}) denotes \emph{no effect} on organism survival. A value of
one (\code{1.0}) denotes a lethality rate of 100\%, i.e. no survivors.

The survival probability \code{S} is available in the return value of \code{\link[=simulate]{simulate()}}.
}

\references{
EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, et al., 2018: \emph{Scientific Opinion on the
state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
regulatory risk assessment of pesticides for aquatic organisms}. EFSA Journal 2018;
16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}
}
\seealso{
Other GUTS-RED models: 
\code{\link{GUTS-RED-models}},
\code{\link{GUTS_RED_SD}()}
}
\concept{GUTS-RED models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{schmitt2013}
\alias{schmitt2013}
\title{A Lemna effect study data set with multiple treatment levels}
\format{
An object of class \code{data.frame} with 56 rows and 4 columns.
}
\usage{
schmitt2013
}
\description{
Data was collected from Schmitt et al. (2013). The study observed the number of
Lemna fronds during the study period of 14 days. The organisms were exposed
to multiple concentrations of the sulfonyl urea herbicide \emph{metsulfuron-methyl}.
}
\details{
The number of Lemna fronds was recorded at the beginning of the experiment (t=0)
and after 3, 5, 7, 10, 12, and 14 days. The exposure to \emph{metsulfuron-methyl}
was only present during the first seven days, which was followed by a recovery
period of another seven days without exposure to the substance.

The tabular dataset consists of four columns:
\itemize{
\item First column: time (days)
\item Second column: observed number of fronds (-)
\item Third column: trial id (-)
\item Fourth column: metsulfuron-methyl concentration (ug/L)
}

An example \link{scenario} with parameters fitted to the experimental data set
as conducted by Schmitt et al. is available as \code{\link{metsulfuron}}.
}
\references{
Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013: \emph{Mechanistic TK/TD-model
simulating the effect of growth inhibitors on Lemna populations}. Ecol Model 255,
pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
}
\seealso{
\link{Lemna-models}, \link{metsulfuron}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pll.R
\name{pll_debug}
\alias{pll_debug}
\title{Disable parallelization for debugging}
\usage{
pll_debug(state = TRUE)
}
\arguments{
\item{state}{\code{logical}, if \code{TRUE} then parallelization is disabled}
}
\value{
no return value
}
\description{
In certain cases it might be beneficial to disable parallel execution
of e.g. effect profile calculations. By disabling, all processes run
sequentially and instantly pass messages to the console which would be
delayed during parallel processing. This makes it easier to pinpoint
problems within the data or algorithm.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lik_profile.R
\name{log_lik}
\alias{log_lik}
\title{Calculate log likelihood}
\usage{
log_lik(
  obs,
  pred,
  data_type = c("continuous", "count"),
  log_scale = FALSE,
  npars = NULL
)
}
\arguments{
\item{obs}{numeric vector of observed values}

\item{pred}{numeric vector of predicted values}

\item{data_type}{determines the if likelihood profiling is conducted for
\code{"continuous"} (default) or \code{"count"} data}

\item{log_scale}{\code{FALSE} (default), option to calculate the log likelihood on a
log scale (i.e., observations and predictions are log transformed during calculation)}

\item{npars}{named numeric vector of parameters that the model was calibrated on,
required for \code{"continuous"} data type, optional for \code{"count"}.}
}
\value{
the log likelihood value
}
\description{
Calculates the sum of log likelihoods of each observation given
the model parameterization.

Current implementations enable log likelihood calculations for:
\enumerate{
\item continuous data, considering a normal distribution around the prediction
for each datapoint,
\item count data, considering a multinomial distribution for data reporting the
number of survivors over time.
}

The log likelihood calculation for count data was inspired by a MatLab BYOM v.6.8
procedure, created by Tjalling Jager. For details, please refer to BYOM
(http://debtox.info/byom.html) as well as Jager (2021).
}
\examples{
# simple example for continuous data #####
# observations
obs <- c(12, 38, 92, 176, 176, 627, 1283, 2640)
# intercept, a, and slope, b, of a Poisson regression fitted through obs
pars <- c(a = 2, b = 0.73)
# predictions with the Poisson regression
pred <- c(15.43, 32.15, 66.99, 139.57, 290.82, 605.94, 1262.52, 2630.58)
# example plot
plot(seq(1:length(obs)), obs)
lines(seq(1:length(obs)), pred)
log_lik(
  obs = obs,
  pred = pred,
  npars = length(pars),
)

# example with count data and GUTS model #####
library(dplyr)
# observational data
dt <- ringtest_c \%>\% filter(replicate == "E")
myexposure <- dt \%>\% select(time, conc)
obs <- dt \%>\%
  mutate(S=Nsurv / max(Nsurv)) \%>\%
  select(time, S)
# GUTS model
GUTS_RED_IT() \%>\%
  set_param(c(hb = 0)) \%>\%
  set_exposure(myexposure) -> myscenario
# fit
fit <- calibrate(
  x = myscenario,
  par = c(kd=1.2, alpha=9.2, beta=4.3),
  data = obs,
  output = "S")
# update
myscenario <- myscenario \%>\% set_param(fit$par)
# simulate
pred <- myscenario \%>\% simulate()
pred <- pred$S #* max(obs$S)
obs <- obs$S
# calc likelihood
log_lik(obs,
    pred,
    data_type = "count")

}
\references{
Jager T, 2021. Robust Likelihood-Based Optimization and Uncertainty Analysis
of Toxicokinetic-Toxicodynamic Models. Integrated Environmental Assessment and
Management 17:388-397. \doi{10.1002/ieam.4333}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_guts}
\alias{is_guts}
\alias{is_GUTS}
\alias{is_guts_it}
\alias{is_GUTS_IT}
\alias{is_guts_sd}
\alias{is_GUTS_SD}
\title{Test if argument is a GUTS scenario}
\usage{
is_guts(x)

is_GUTS(x)

is_guts_it(x)

is_GUTS_IT(x)

is_guts_sd(x)

is_GUTS_SD(x)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}
}
\value{
vector of \code{logical} values
}
\description{
Test if argument is a GUTS scenario
}
\section{Functions}{
\itemize{
\item \code{is_GUTS()}: Deprecated alias.

\item \code{is_guts_it()}: Test if argument is a GUTS-IT scenario

\item \code{is_GUTS_IT()}: Deprecated alias.

\item \code{is_guts_sd()}: Test if argument is a GUTS-SD scenario

\item \code{is_GUTS_SD()}: Deprecated alias.

}}
\examples{
# returns `TRUE`
is_guts(minnow_it)
is_guts(GUTS_RED_IT())

# returns `c(TRUE,TRUE,TRUE)`
is_guts(c(minnow_it, minnow_it, minnow_it))

# returns `FALSE`
is_guts_sd(minnow_it)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ringtest_sd}
\alias{ringtest_sd}
\title{EFSA Ringtest Dataset A SD}
\format{
An object of class \code{data.frame} with 42 rows and 4 columns.
}
\usage{
ringtest_sd
}
\description{
This dataset is part of EFSA's \emph{GUTS} ringtest to compare results from software
implementations of \link[=GUTS-RED-models]{GUTS-RED models} (EFSA 2018). The ringtest
focused on \link[=GUTS_RED_IT]{GUTS-RED-IT} and \link[=GUTS_RED_SD]{GUTS-RED-SD} models.
}
\details{
\emph{Dataset A SD} for the GUTS ring test is a Hypothetical data set constructed with SD.
}
\references{
EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, et al., 2018: \emph{Scientific Opinion on the
state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
regulatory risk assessment of pesticides for aquatic organisms}. EFSA Journal 2018;
16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}

https://cran.r-project.org/web/packages/GUTS/vignettes/ringTest.html
}
\seealso{
\link{GUTS-RED-models}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numerics.R
\name{diagnostics}
\alias{diagnostics}
\alias{diagnostics.default}
\alias{diagnostics.cvasi_simulate}
\title{Diagnostics of solvers}
\usage{
diagnostics(obj, ...)

\method{diagnostics}{default}(obj, ...)

\method{diagnostics}{cvasi_simulate}(obj, ...)
}
\arguments{
\item{obj}{return value of a simulation}

\item{...}{unused parameters}
}
\description{
Prints several diagnostics of the simulation to the console, e.g. number of
steps taken, the last step size, etc. The information is provided by
\code{\link[deSolve:diagnostics]{deSolve::diagnostics()}}.
}
\seealso{
\code{\link[deSolve:diagnostics]{deSolve::diagnostics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-deb_abj.R
\name{DEB_abj}
\alias{DEB_abj}
\alias{DebAbj-class}
\title{DEB_abj}
\usage{
DEB_abj()
}
\value{
an S4 object of type \linkS4class{DebAbj}
}
\description{
Creates a \emph{DEB abj} scenario. The \emph{abj} model with type M acceleration is
like model \emph{std}, but acceleration occurs between birth and metamorphosis (V1-morph).
Isomorphy is assumed before and after acceleration. Metamorphosis is before
puberty and occurs at maturity \code{E_Hj}, which might or might not correspond with
changes in morphology. The \emph{abj} model is a one-parameter extension of model \emph{std}
\href{https://debportal.debtheory.org/docs/Typified_models.html}{(DEB Wiki)}.
}
\details{
\subsection{State variables}{

The following list describes the default names and standard units of the model's
state variables:
\itemize{
\item \code{L}, structural length (cm)
\item \code{E}, energy reserve (J)
\item \code{H}, energy invested in maturity (J)
\item \code{R}, reproduction buffer (J)
\item \code{cV}, internal concentration (\link{C})
\item \code{Lmax}, maximum structural length (cm)
}

All state variables are initialized with zero. See \code{\link[=set_init]{set_init()}} on how to set
the initial state.
}

\subsection{Parameters}{

The following model parameters are required:
\itemize{
\item \code{p_M}, vol-spec somatic maintenance (J/d.cm^3)
\item \code{v}, energy conductance (cm/d)
\item \code{k_J}, maturity maint rate coefficient (1/d)
\item \code{p_Am}, surface-area specific maximum assimilation rate (J/d.cm^2)
\item \code{kap}, allocation fraction to soma (-)
\item \code{E_G}, spec cost for structure (J/cm^3)
\item \code{f}, scaled functional response (-)
\item \code{E_Hj}, maturity at metamorphosis (J)
\item \code{E_Hp}, maturity at puberty (J)
\item \code{kap_R}, reproduction efficiency (-)
\item \code{L_b}, structural length at birth (cm)
\item \code{L_j}, structural length at metamorphosis (cm)
\item \code{ke}, elimination rate constant (d-1)
\item \code{c0}, no-effect concentration sub-lethal (\link{C})
\item \code{cT}, tolerance concentration (\link{C})
\item \code{MoA}, mode of action switch (-)
}
}

\subsection{Mode of Actions}{

Any combination of the following mode of actions (MoA) can be considered by
the model:
\itemize{
\item \code{MoA = 1}: effect on feeding
\item \code{MoA = 2}: effect on maintenance costs
\item \code{MoA = 4}: effect on overhead costs for making an egg
\item \code{MoA = 8}: hazard during oogenesis
\item \code{MoA = 16}: energy conductance
}

To activate more than one MoA, simply add up the corresponding
codes. To disable all MoAs, set the parameter to zero.
See also \code{\link[=set_mode_of_action]{set_mode_of_action()}}.
}

\subsection{Effects}{

The state variables \emph{L} (structural length) and \emph{R} (reproduction buffer) are
set as effect endpoints by default. All state variables are available as
potential endpoints. The list of considered endpoints can be modified
by using \code{\link[=set_endpoints]{set_endpoints()}}.

To calculate effects, each \emph{DEB} scenario is simulated twice: One simulation
which considers exposure to a toxicant and one simulation without exposure, i.e.
a control. See also \code{\link[=effect]{effect()}}.
}
}
\section{Simulation output}{

Simulation results will contain the state variables.
It is possible to amend the output of \code{\link[=simulate]{simulate()}} with additional model
quantities that are not state variables, for e.g. debugging purposes or to
analyze model behavior. To enable or disable additional outputs, use the
optional argument \code{nout} of \code{\link[=simulate]{simulate()}}. As an example, set \code{nout=2} to
enable reporting of the acceleration factor (\code{MV}) and the mobilization flux
(\code{pC}). Set \code{nout=0} to disable additional outputs (default).

The available output levels are as follows:
\itemize{
\item \code{nout} >= 1: \code{MV} acceleration factor (-)
\item \code{nout} >= 2: \code{pC} mobilization flux (J/d)
\item \code{nout} >= 3: \code{pA} assimilation flux (J/d)
\item \code{nout} >= 4: \code{pJ} energy invested in maturity flux (J/d)
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.
}

\examples{
# Create an abj scenario from scratch and simulate it
DEB_abj() \%>\%
  set_init(c(L=0.02,E=0.1,H=0.01)) \%>\%
  set_param(c(p_M=3000,v=0.02,k_J=0.6,p_Am=300,kap=0.9,E_G=4000,f=1,
              E_Hj=0.05,E_Hp=0.3,kap_R=0.9,ke=1,c0=0,cT=1,L_b=0.02,
              L_j=0.04,MoA=0)) \%>\%
  set_noexposure() \%>\%
  set_times(0:10) \%>\%
  simulate()

# Print information about sample scenario 'americamysis'
americamysis

# Simulate 'americamysis' scenario
americamysis \%>\% simulate()
}
\seealso{
Other DEB models: 
\code{\link{DEB-models}},
\code{\link{DEBtox}()}
}
\concept{DEB models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_growth.R
\name{fit_growth}
\alias{fit_growth}
\alias{fit_growth,list,missing-method}
\alias{fit_growth,EffectScenario,ANY-method}
\alias{fit_growth,ScenarioSequence,ANY-method}
\alias{fit_growth,ANY,ANY-method}
\alias{fit_growth,LemnaSetac,missing-method}
\alias{fit_growth,LemnaSchmitt,missing-method}
\alias{fit_growth,Magma,missing-method}
\title{Fit growth parameters}
\usage{
fit_growth(x, data, ...)

\S4method{fit_growth}{ANY,ANY}(x, data, ...)

\S4method{fit_growth}{LemnaSetac,missing}(x, data, par, log_scale = TRUE, verbose = FALSE, ...)

\S4method{fit_growth}{LemnaSchmitt,missing}(x, data, par, log_scale = TRUE, verbose = FALSE, ...)

\S4method{fit_growth}{Magma,missing}(x, data, par, log_scale = TRUE, verbose = FALSE, ...)
}
\arguments{
\item{x}{a \link{scenario} or a list of \link{caliset} objects}

\item{data}{toxicological trial data to fit growth parameters to: required if
\code{x} is either a \link{scenario} or the result of a fit. See Section \emph{Data}
for details.}

\item{...}{additional arguments passed through to \code{\link[=calibrate]{calibrate()}}}

\item{par}{named vector, of parameters to fit and their starting values}

\item{log_scale}{logical, if \code{TRUE} then fitting will be performed on log-transformed
observations and predictions, else the data will be used as-is}

\item{verbose}{logical, if \code{TRUE} then info messages are printed to the console}
}
\value{
a list
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} \if{html}{\out{<br>}}
High-level function to fit growth and/or loss parameters to observed data. It eases the
use of \code{\link[=calibrate]{calibrate()}} by providing sensible defaults for various (model dependent)
settings.
}
\section{Methods (by class)}{
\itemize{
\item \code{fit_growth(x = ANY, data = ANY)}: Default handler

\item \code{fit_growth(x = LemnaSetac, data = missing)}: Fit growth parameters of \link{Lemna_SETAC} scenarios

\item \code{fit_growth(x = LemnaSchmitt, data = missing)}: Fit growth parameters of \link{Lemna_Schmitt} scenarios

\item \code{fit_growth(x = Magma, data = missing)}: Fit growth parameters of \link{Magma} scenarios

}}
\examples{
# Use experimental data from control trial ('T0') to fit growth rate
ctrl <- schmitt2013[schmitt2013$trial == "T0" ,]

# Set up a scenario, provide dummy parameter values where necessary
sc <- Lemna_Schmitt() \%>\%
  set_init(c(BM=0.0012)) \%>\%
  set_param(c(EC50=1, b=1, P_up=1))

# Run fitting routine
fit_growth(sc, data=ctrl, verbose=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-algae.R
\name{Algae_Weber}
\alias{Algae_Weber}
\alias{SAM-X}
\alias{AlgaeWeber-class}
\alias{AlgaeWeberScenario-class}
\alias{SamX}
\title{Algae model, \emph{SAM-X} (Weber et al. 2012)}
\usage{
Algae_Weber()

SamX()
}
\value{
an S4 object of type \linkS4class{AlgaeWeber}
}
\description{
The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
growth model for algae. The model simulates the development of algal biomass
under laboratory and environmental conditions and was developed by
Weber et al. (2012) as cited in EFSA TKTD opinion (2018). The growth of the
algae population is simulated on the basis of growth rates, which are
dependent on environmental conditions (radiation, temperature, and phosphorus).
The toxicodynamic sub-model describes the effects of growth-inhibiting
substances through a corresponding reduction in the photosynthesis rate on
the basis of internal concentrations.
}
\details{
Deviating from the equations described by Weber et al., this model implementation
uses a user-defined time-series to represent (environmental) concentrations.
Therefore, state-variable \code{C} and its differential equation was removed and
model output \code{C} is identical to the exposure time-series. The implementation
of Weber et al. (2012) was followed where units differ from EFSA (2018).
}
\section{Functions}{
\itemize{
\item \code{SamX()}: Alias using original model name.

}}
\section{State variables}{

The model has three state variables:
\itemize{
\item \code{A}, Biomass (µg fresh wt/mL, cells/mL *10^4)
\item \code{Q}, Mass of phosphorous internal (mg P/L, or µg P/mL)
\item \code{P}, Mass of phosphorous external (mg P/L, or µg P/mL)
}

The original model by Weber et al. contains an additional state variable
\code{C} which models the external stressor concentration. However, the model
implementation in this packages uses a user-defined time-series to represent
environmental concentrations.
Therefore, state variable \code{C} and accompanying parameters are not present here.
}

\section{Model parameters}{

\itemize{
\item Growth model
\itemize{
\item \code{mu_max}, Maximum growth rate (d-1)
\item \code{Q_min}, Minimum intracellular P (µg P/µg fresh wt)
\item \code{Q_max}, Maximum intracellular P (µg P/µg fresh wt)
\item \code{v_max}, Maximum P-uptake rate at non-limited growth (µg P/µg fresh wt/d)
\item \code{k_s},   Half-saturation constant for extracellular P (mg P/L)
\item \code{m_max}, Natural mortality rate (1/d)
\item \code{I_opt}, Optimum light intensity for growth (uE/m²/s)
\item \code{T_opt}, Optimum temperature for growth (°C)
\item \code{T_max}, Maximum temperature for growth (°C)
\item \code{T_min}, Minimum temperature for growth (°C)
\item \code{D}, Dilution rate (1/d)
\item \code{R_0}, Influx concentration of P (mg P/L)
}
\item Concentration response (Toxicodynamics)
\itemize{
\item \code{EC_50}, Effect concentration of 50\% inhibition of growth rate (µg/L)
\item \code{b}, slope of concentration effect curve at EC_50 (-)
}
}
}

\section{Forcings}{

The Weber model requires two environmental properties as time-series input:
\itemize{
\item \code{T_act}, temperature (°C), and
\item \code{I}, irradiance (uE/m²/s).
}

The following constant default values are used for these properties:
\itemize{
\item \code{T_act} = 23 °C
\item \code{I} = 100 uE/m²/s
}

Forcings time-series are represented by \code{data.frame} objects consisting of two
columns. The first for time and the second for the environmental factor in question.

Entries of the \code{data.frame} need to be ordered chronologically. A time-series
can consist of only a single row; in this case it will represent constant
environmental conditions. See \link{scenarios} for more details.
}

\section{Simulation output}{

Simulation results will contain the state variables Biomass (\code{A}), mass of
internal phosphorous (\code{Q}), mass of external phosphorous (\code{P}) and the external
concentration (\code{C}).

It is possible to amend the output of \code{\link[=simulate]{simulate()}} with additional model
quantities that are not state variables, for e.g. debugging purposes or to
analyze model behavior. To enable or disable additional outputs, use the
optional argument \code{nout} of \code{\link[=simulate]{simulate()}}. As an example, set \code{nout=2} to
enable reporting of external concentration and model derivative \code{dA}.
Set \code{nout=0} to disable additional outputs. The default is \code{nout=1}.

The available output levels are as follows:
\itemize{
\item \code{nout >= 1}: \code{C}, external concentration (µg/L)
\item \code{nout >= 2}: \code{f(T)}, temperature dependence (-)
\item \code{nout >= 3}: \code{f(I)}, light dependence (-)
\item \code{nout >= 4}: \code{f(Q)}, nutrient dependence (-)
\item \code{nout >= 5}: \code{f(Q, P)}, uptake flow reduction (-)
\item \code{nout >= 6}: \code{f(C)}, effect of chemical stressor (-)
\item \code{nout >= 7}: \code{dA}, biomass derivative (µg)
\item \code{nout >= 8}: \code{dQ}, internal phosphorous derivative (mg P/µg fresh wt)
\item \code{nout >= 9}: \code{dP}, external phosphorous derivative (mg P L-1)
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{hmax = 0.1}\if{html}{\out{<br>}}
Maximum step length in time suitable for most simulations.
}
}

\section{Parameter boundaries}{

Default values for parameter boundaries are set for all parameters by expert
judgement, for calibration purposes. Values can be access from the object, and
defaults overwritten.
}

\section{Model history and changes}{

\itemize{
\item cvasi v1.5.0
\itemize{
\item Unused state variable \code{C} and parameter \code{k} removed from documentation
and code. External concentration \code{C} added to simulation output by means
of optional output level \code{nout=1}.
}
}
}

\references{
Weber D, Schaefer D, Dorgerloh M, Bruns E, Goerlitz G, Hammel K, Preuss TG
and Ratte HT, 2012. Combination of a higher-tier flow-through system and
population modeling to assess the effects of time-variable exposure of
isoproturon on the green algae Desmodesmus subspictatus and
Pseudokirchneriella subcapitata. Environmental Toxicology and
Chemistry, 31(4), 899-908. \doi{10.1002/etc.1765}

EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, Brock T, Duquesne S, Grilli S,
Hernandez-Jerez AF, Bennekou SH,Klein M, Kuhl T, Laskowski R, Machera K,
Pelkonen O, Pieper S, Smith RH, Stemmer M, Sundh I, Tiktak A,Topping CJ,
Wolterink G, Cedergreen N, Charles S, Focks A, Reed M, Arena M, Ippolito A,
Byers H andTeodorovic I, 2018. Scientific Opinion on the state of the art of
Toxicokinetic/Toxicodynamic (TKTD)effect models for regulatory risk assessment
of pesticides for aquatic organisms. EFSA Journal, 16(8), 5377.
\doi{10.2903/j.efsa.2018.5377}
}
\seealso{
\link{Scenarios}, \link{Transferable}

Other algae models: 
\code{\link{Algae-models}},
\code{\link{Algae_Simple}()},
\code{\link{Algae_TKTD}()}
}
\concept{algae models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_lemna}
\alias{is_lemna}
\alias{is_Lemna}
\title{Test if argument is a Lemna scenario}
\usage{
is_lemna(x)

is_Lemna(x)
}
\arguments{
\item{x}{vector of \link{scenario} objects}
}
\value{
vector of logical values
}
\description{
Test if argument is a Lemna scenario
}
\section{Functions}{
\itemize{
\item \code{is_Lemna()}: Deprecated alias.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/man-deb.R
\name{DEB-models}
\alias{DEB-models}
\alias{Deb-class}
\title{Dynamic Energy Budget (DEB) models}
\description{
Supported models:
\itemize{
\item \link{DEB_abj}
\item \link{DEBtox}
}
}
\seealso{
Other DEB models: 
\code{\link{DEB_abj}()},
\code{\link{DEBtox}()}

Other models: 
\code{\link{Algae-models}},
\code{\link{GUTS-RED-models}},
\code{\link{Lemna-models}},
\code{\link{Macrophyte-models}}
}
\concept{DEB models}
\concept{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibrate.R
\name{calibrate}
\alias{calibrate}
\alias{calibrate,EffectScenario-method}
\alias{calibrate,CalibrationSet-method}
\alias{calibrate,list-method}
\alias{calibrate,ScenarioSequence-method}
\title{Fit model parameters to experimental data}
\usage{
\S4method{calibrate}{EffectScenario}(x, output, data, ...)

\S4method{calibrate}{list}(
  x,
  par,
  output,
  err_fun = "sse",
  log_scale = FALSE,
  verbose = TRUE,
  ode_method = NULL,
  ...
)
}
\arguments{
\item{x}{either a single \link{scenario}. \link{sequence}, or a list of \link{caliset} objects
to be fitted. If only a scenario or sequence is supplied, the additional argument
\code{data} is required.}

\item{output}{\code{character}, name of a single output column of \code{\link[=simulate]{simulate()}} to
optimize on}

\item{data}{a \code{data.frame} or return value of \code{\link[=tox_data]{tox_data()}}; the scenarios's
output is fitted to the (observed) data. See \code{\link[=tox_data]{tox_data()}} for valid tabular
data formats.}

\item{...}{additional parameters passed on to \code{\link[stats:optim]{stats::optim()}} and \code{\link[=simulate]{simulate()}}}

\item{par}{named numeric vector with parameters to fit and their start values}

\item{err_fun}{a \code{character} choosing one of the pre-defined error functions,
or alternatively a function implementing a custom error function. Defaults
to \emph{Sum of squared errors} (\code{"sse"}).}

\item{log_scale}{logical, if \code{TRUE} then observed and predicted values are
log-transformed before the error function is evaluated. Defaults to \code{FALSE}.}

\item{verbose}{\code{logical}, if \code{TRUE} then debug outputs are displayed during
optimization}

\item{ode_method}{optional \code{character} to select an ODE solver for \code{\link[=simulate]{simulate()}}}
}
\value{
A list of fitted parameters (as produced by \code{\link[stats:optim]{stats::optim()}})
is returned.
}
\description{
The function \code{calibrate()} performs the calibration (fitting) of model
parameters to observed data. The data can originate from one or more experiments or
trials. Experimental conditions, such as model parameters and exposure
level, can differ between trials; fitting can be performed on all datasets
at the same time.
}
\details{
Fitting of model parameters can be performed in two ways:
\enumerate{
\item A single \link{scenario} is fitted to a single dataset.
The dataset must represent a time-series of an output variable of the
model, e.g. observed biomass over time (effect data). The dataset can represent
results of one or more experimental replicates under identical conditions.
\item One or more datasets of observed data are fitted each to a scenario which
describes the experimental conditions during observation, such as exposure
level and environmental properties. Each combination of dataset and scenario
is represented by a \link[=caliset]{calibration set}. During fitting,
all \emph{calibration sets} are evaluated and a total error term is calculated
over all observed and predicted values.
}
\subsection{Observed data}{

Experimental, or effect, data must be supplied as a \code{data.frame} in long format
with at least two columns: the first column contains \code{numeric} timestamps and
the remaining columns must contain the observed quantity. The observed quantitiy
must match in unit and meaning with the model output defined by argument \code{output}.

As an example, the simulation result of \link{Lemna_Schmitt} model contains the
output column \emph{biomass} (\code{BM}), amongst others. To fit model parameters of said
\emph{Lemna_Schmitt} scenario based on observed biomass, the observed data must
represent biomass as well. A minimal observed dataset could look like this:

\if{html}{\out{<div class="sourceCode">}}\preformatted{observed <- data.frame(time=c(0,  7, 14, 21),
                       BM=c( 12, 23, 37, 56))
}\if{html}{\out{</div>}}
}

\subsection{Error function}{

The error function quantifies the deviations between simulated and observed data.
The decision for an error function can have influence on the result of the
fitting procedure. Two error functions are pre-defined by the package and
can be selected by the user, but custom error functions can be used as well.

Available pre-defined error functions:
\itemize{
\item \code{"sse"}: Sum of squared errors
}

By default, the sum of squared errors is used as the error function which
gets minimized during fitting. A custom error function must accept four vectorized
arguments and return a numeric of length one, i.e. the total error value.
The arguments to the error function will all have the same length and are
defined as follows:
\itemize{
\item First argument: all observed data points
\item Second argument: all simulated data points
\item Third argument: optional weights for each data point
\item Fourth argument: a list of optional caliset tags
}

You can choose to ignore certain arguments, such as weights and tags, in your
custom error function. An example of a custom error function which returns
the sum of absolute errors looks as follow:

\if{html}{\out{<div class="sourceCode">}}\preformatted{my_absolute_error <- function(observed, predicted, weights, tags) \{
  sum(abs(observed - predicted))
\}
}\if{html}{\out{</div>}}

As tags are optional, the fourth argument may be a list containing \code{NULL} values.
The fourth argument can be used to pass additional information to the error
function: For example, the tag may identify the study from where the data
originates from and the error function could group and evaluate the data
accordingly.
}
}
\section{Methods (by class)}{
\itemize{
\item \code{calibrate(EffectScenario)}: Fit single \link{scenario} to a (\link{tox_data}) dataset

}}
\examples{
# Create an artificial data set of observed frond numbers.
# It assumes exponential growth with an effective growth rate of 0.38
trial <- data.frame(time=0:14,
                   fronds=12 * exp(0:14 * 0.38))
plot(trial)

# Create a Lemna scenario that represents unrestricted, exponential growth.
scenario <- Lemna_Schmitt() \%>\%
  set_param(c(k_phot_max=1, k_resp=0, EC50=1, b=1, P_up=1)) \%>\%
  set_init(c(BM=12)) \%>\%
  set_noexposure()

# Fit scenario parameter 'k_phot_max' to observed frond numbers:
fit <- calibrate(
  scenario,
  par="k_phot_max",
  data=trial,
  output="BM"
)

# The fitted value of 'k_phot_max' matches the effective growth rate which
# was used to create the artificial data set:
fit$par
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_exposure.R
\name{set_exposure}
\alias{set_exposure}
\alias{set_exposure,ANY,ANY-method}
\alias{set_exposure,EffectScenario,data.frame-method}
\alias{set_exposure,EffectScenario,ExposureSeries-method}
\alias{set_exposure,EffectScenario,NoExposureSeries-method}
\alias{set_exposure,EffectScenario,list-method}
\alias{set_exposure,list,list-method}
\alias{set_exposure,list,ANY-method}
\alias{set_exposure,ScenarioSequence,ANY-method}
\title{Set exposure time-series}
\usage{
set_exposure(scenarios, series, ...)

\S4method{set_exposure}{ANY,ANY}(scenarios, series, ...)

\S4method{set_exposure}{EffectScenario,data.frame}(scenarios, series, ...)

\S4method{set_exposure}{EffectScenario,ExposureSeries}(scenarios, series, reset_times = TRUE)

\S4method{set_exposure}{EffectScenario,NoExposureSeries}(scenarios, series, ...)

\S4method{set_exposure}{EffectScenario,list}(scenarios, series, ...)

\S4method{set_exposure}{list,list}(scenarios, series, ...)

\S4method{set_exposure}{list,ANY}(scenarios, series, ...)

\S4method{set_exposure}{ScenarioSequence,ANY}(scenarios, series, ...)
}
\arguments{
\item{scenarios}{\code{vector} of \link{scenarios}}

\item{series}{\code{vector} of  \link{ExposureSeries} objects or a single \code{data.frame}}

\item{...}{additional arguments}

\item{reset_times}{\code{logical}, if \code{TRUE}, the exposure time-series' time points
will be set as output times. Defaults to \code{TRUE}}
}
\value{
\code{list} of \code{EffectScenario} objects
}
\description{
\emph{Exposure} refers to the toxicant concentration an organism is exposed to.
In case of aquatic organisms, this would commonly be the concentration of a
toxicant in water. Other interpretations are possible depending on model
assumptions.
}
\details{
Exposure time-series are generally represented by a \code{data.frame} containing two
columns. The first column for time, the second representing the
exposure level. The ordering of columns is mandatory. The column names
are non-relevant but sensible names may help documenting the
scenario and its data. The \code{data.frame}'s rows must be ordered chronologically.
A time-series can consist of only a single row; in this case it will represent constant
exposure.

For convenience, a time-series with zero exposure can be set using
\code{\link[=set_noexposure]{set_noexposure()}}.
\subsection{Computational efficiency}{

Handling time-series is a costly task for the ODE solver due to consistency
checks and interpolation between time steps. How the solver interpolates
the time-series can be controlled by optional arguments to functions
such as \code{\link[=simulate]{simulate()}} and \code{\link[=effect]{effect()}}. Please refer to \code{\link[=simulate]{simulate()}} for a brief
overview and \link[deSolve:forcings]{deSolve::forcings} for a detailed description.

Exposure time-series should be kept as short as possible and as complex as
needed for optimal computational efficiency.
}

\subsection{Output times}{

By default, the exposure time-series' time points will also be used as output
times of the scenario. Any output times previously set by \code{\link[=set_times]{set_times()}} will
be lost. If this behavior is undesired, set the function argument
\code{reset_times=FALSE}.
}

\subsection{Multiple exposure series and scenarios}{

The functions supports modifying multiple scenarios at once: by
calling it with lists of \link{scenario} and \link{ExposureSeries}
objects. The cartesian product of all scenarios and exposure series will
be returned, iff the parameter \code{expand = TRUE} is set.

As an example for the \emph{expand} mode, two scenarios \code{A} and \code{B} and one
exposure series \code{g} will result in two scenarios \code{Ag} and \code{Bg}, both
using exposure series \code{g}. Two scenarios \code{A} and \code{B} as well as two
exposure seres \code{g} and \code{h} will result in four scenarios \code{Ag},\code{Ah},\code{Bg},
and \code{Bh}.
}
}
\examples{
# Set a data.frame as exposure series
df <- data.frame(time=c(0, 1, 2, 3), conc=c(1, 1, 0, 0))
Lemna_Schmitt() \%>\% set_exposure(df)

# Create and set an ExposureSeries object
es1 <- ExposureSeries(df)
Lemna_Schmitt() \%>\% set_exposure(es1)

# By default, the time points of the exposure series will also be used as
# as output times. To avoid overriding existing output times, set reset_times=FALSE
Lemna_Schmitt() \%>\%
  set_times(0:10) \%>\%
  set_exposure(es1, reset_times=FALSE)

# Setting two series with one function call, creates two scenarios
es2 <- ExposureSeries(data.frame(time=5:10, conc=1))
Lemna_Schmitt() \%>\% set_exposure(c(es1, es2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-lemna_schmitt.R
\name{Lemna_Schmitt}
\alias{Lemna_Schmitt}
\alias{LemnaSchmitt-class}
\alias{LemnaSchmittScenario-class}
\alias{Lemna_SchmittThold}
\title{Lemna model (Schmitt et al. 2013)}
\usage{
Lemna_Schmitt(param, init)

Lemna_SchmittThold(param, init)
}
\arguments{
\item{param}{optional named \code{list} or \code{vector} of model parameters}

\item{init}{optional named numeric \code{vector} of initial state values}
}
\value{
an S4 object of type \linkS4class{LemnaSchmitt}
}
\description{
The model is a mechanistic combined toxicokinetic-toxicodynamic (TK/TD) and
growth model for the aquatic macrophytes \emph{Lemna spp.}
The model simulates the development of \emph{Lemna} biomass under laboratory and
environmental conditions and was developed by Schmitt \emph{et al.} (2013). Growth
of the \emph{Lemna} population is simulated on basis of photosynthesis and respiration
rates which are functions of environmental conditions.
The toxicodynamic sub-model describes the effects of growth-inhibiting
substances by a respective reduction in the photosynthesis rate based on
internal concentrations. This is the historical version of the Lemna model.
For current uses, we recommend the Lemna (SETAC) model, which is a more recent
version of the Schmitt model.
}
\details{
Constructors to ease creation of scenarios based on the \emph{Lemna} model by
Schmitt \emph{et al.} (2013).
A variant of this \emph{Lemna} model, \code{Lemna_SchmittThold()}, provides an additional
cumulative exposure threshold parameter. The Lemna biomass stops growing
if the integral of exposure over time exceeds the threshold. The integral
of exposure is internally accounted for by an additional state variable
\code{AUC} (Area Under Curve).
}
\section{Functions}{
\itemize{
\item \code{Lemna_SchmittThold()}: model variant with cumulative exposure threshold

}}
\section{State variables}{

The following list describes the default names and standard units of the model's
state variables:
\itemize{
\item BM, g_dw/m2, dry weight biomass per square meter
\item E, -, effect [0,1]
\item M_int, ug, internal toxicant mass
\item AUC, ug/L, cumulative exposure (\strong{only} for \code{LemnaThreshold} model)
}

Biomass (BM) and internal toxicant mass (M_int) are initialized to zero by
default. See \code{\link[=set_init]{set_init()}} on how to set the initial states.
}

\section{Model parameters}{

The following model parameters are required:
\itemize{
\item Fate and biomass
\itemize{
\item \code{k_phot_fix}, logical, TRUE then k_phot_max is not changed by environmental factors, else FALSE
\item \code{k_phot_max}, 1/d, maximum photosynthesis rate
\item \code{k_resp}, 1/d, respiration rate
\item \code{k_loss}, 1/d, rate of loss (e.g. flow rate)
\item \code{mass_per_frond}, g_dw/frond, dry weight per frond
\item \code{BMw2BMd}, g_fw/g_dw, Fresh weight/dry weight
}
\item Effect
\itemize{
\item \code{Emax}, -, maximum effect [0,1]
\item \code{EC50}, ug/L, midpoint of effect curve
\item \code{b}, -, slope of effect curve
}
\item Toxicokinetics
\itemize{
\item \code{P_up}, cm/d, Permeability for uptake
\item \code{AperBM}, cm2/g_dw, A_leaf / d_leaf = 1/d_leaf (for circular disc, d=0.05 cm)
\item \code{Kbm}, -, Biomass(fw) : water partition coefficient
\item \code{P_Temp}, logical, TRUE to enable temperature dependence of cuticle permeability, else FALSE
\item \code{MolWeight}, g/mol, Molmass of molecule (determines Q10_permeability)
}
\item Temperature dependence
\itemize{
\item \code{Tmin}, deg C, minimum temperature for growth
\item \code{Tmax}, deg C, maximum temperature for growth
\item \code{Topt}, deg C, optimal temperature for growth
\item \code{t_ref}, deg C, reference temperature for respiration rate
\item \code{Q10}, -, temperature dependence factor for respiration rate
}
\item Light dependence
\itemize{
\item \code{k_0}, 1/d, light dependence: intercept of linear part
\item \code{a_k}, (1/d)/(kJ/m2.d), light dependence: slope of linear part
}
\item Phosphorus dependence (Hill like dep.)
\itemize{
\item \code{C_P}, mg/L, phosphorus concentration in water
\item \code{CP50}, mg/L, phosphorus conc. where growth rate is halfed
\item \code{a_p}, -, Hill coefficient
\item \code{KiP}, mg/L, p-inhibition constant for very high p-conc.
}
\item Nitrogen dependence (Hill like dep.)
\itemize{
\item \code{C_N}, mg/L, nitrogen concentration in water
\item \code{CN50}, mg/L, n-conc. where growth rate is halfed
\item \code{a_N}, -, Hill coefficient
\item \code{KiN}, mg/L, n-inhibition constant for very high p-conc.
}
\item Density dependence
\itemize{
\item \code{BM50}, g_dw/m2, cut off BM
}
}

The \code{Lemna_SchmittThold} model requires the following additional parameter:
\itemize{
\item \code{threshold}, ug/L, cumulative exposure threshold
}
}

\section{Forcings}{

Besides exposure, the model requires four environmental properties as
time-series input:
\itemize{
\item \code{temp}, temperature (°C)
\item \code{rad}, global irradiation (kJ m-2 d-1)
}

The following constant default values are used for these properties:
\itemize{
\item \code{temp} = 12 °C
\item \code{rad} = 15,000 kJ m-2 d-1
}

Forcings time-series are represented by \code{data.frame} objects consisting of two
columns. The first for time and the second for the environmental factor in question.

Entries of the \code{data.frame} need to be ordered chronologically. A time-series
can consist of only a single row; in this case it will represent constant
environmental conditions. See \link{scenarios} for more details.
}

\section{Effects}{

Supported effect endpoints include \emph{BM} (biomass) and \emph{r} (average
growth rate during simulation). The effect on biomass is calculated from
the last state of a simulation. Be aware that endpoint \emph{r} is incompatible
with frond transfers.
}

\section{Parameter boundaries}{

Default values for parameter boundaries are set for all parameters by expert
judgement, for calibration purposes. Values can be access from the object, and
defaults overwritten.
}

\section{Simulation output}{

Simulation results will contain two additional columns besides state variables:
\itemize{
\item \code{C_int}, ug/L, internal concentration of toxicant
\item \code{FrondNo}, -, number of fronds
}

It is possible to amend the output of \code{\link[=simulate]{simulate()}} with additional model
quantities that are not state variables, for e.g. debugging purposes or to
analyze model behavior. To enable or disable additional outputs, use the
optional argument \code{nout} of \code{\link[=simulate]{simulate()}}, see examples below. \code{nout=1}
enables reporting of internal concentration (C_int), \code{nout=14} enables all
additional outputs, and \code{nout=0} will disable additional outputs.

The available output levels are as follows:
\itemize{
\item \code{nout >= 1}: \code{C_int}, internal concentration (ug/L)
\item \code{nout >= 2}: \code{FrondNo}, number of fronds (-)
\item \code{nout >= 3}: \code{C_int_u}, unbound internal concentration (ug/l)
\item Growth and TK/TD
\itemize{
\item \code{nout >= 4}: \code{BM_fresh}, fresh weight biomass (g_fw/m2)
\item \code{nout >= 5}: \code{k_photo_eff}, current photosynthesis rate (1/d)
\item \code{nout >= 6}: \code{k_resp_eff}, current respiration rate (1/d)
\item \code{nout >= 7}: \code{f_Eff}, toxic effect factor (-)
\item \code{nout >= 8}: \code{P_up_eff}, current permeability for uptake (cm/d)
}
\item Environmental variables
\itemize{
\item \code{nout >= 9}: \code{actConc}, current toxicant concentration in surrounding medium (ug/L)
\item \code{nout >= 10}: \code{actTemp}, current environmental temperature (deg C)
\item \code{nout >= 11}: \code{actRad}, current environmental radiation (kJ/m2.d)
}
\item Derivatives
\itemize{
\item \code{nout >= 12}: \verb{d BM/dt}, current change in state variable BM
\item \code{nout >= 13}: \verb{d E/dt}, current change in effect
\item \code{nout >= 14}: \verb{d M_int/dt}, current change in internal toxicant mass
}
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{hmax = 0.1}\if{html}{\out{<br>}}
Maximum step length in time suitable for most simulations.
}
}

\section{Biomass transfer}{

Models supporting biomass transfer can be instructed to move a fixed amount
of biomass to a new medium after a period of time. This feature replicates
a procedure occurring in e.g. \emph{Lemna} effect studies and may be necessary to
recreate study results.

The biomass transfer feature assumes that always a fixed amount of
biomass is transferred. Transfers can occur at any fixed point in time or
in regular intervals. During a transfer, the biomass is reset to the
transferred amount and additional compartments can be scaled 1:1 accordingly,
to e.g. reflect the change in internal toxicant mass when biomass is modified.
Transfer settings can be modified using \code{\link[=set_transfer]{set_transfer()}}.

If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
}

\references{
Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013:
\emph{Mechanistic TK/TD-model simulating the effect of growth inhibitors on
Lemna populations}. Ecol Model 255, pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
}
\seealso{
\link{Lemna-models}, \link{Macrophyte-models}, \link{Transferable}, \link{Scenarios}

Other Lemna models: 
\code{\link{Lemna-models}},
\code{\link{Lemna_SETAC}()}

Other macrophyte models: 
\code{\link{Lemna_SETAC}()},
\code{\link{Macrophyte-models}},
\code{\link{Magma}}
}
\concept{Lemna models}
\concept{macrophyte models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/epx.R
\name{epx}
\alias{epx}
\title{Effect profiles (EPx values)}
\usage{
epx(
  scenarios,
  level = c(10, 50),
  effect_tolerance = 0.001,
  factor_cutoff = NA,
  min_factor = 1e-30,
  max_factor = 1e+30,
  verbose = FALSE,
  ep_only = FALSE,
  long_format = FALSE,
  ...
)
}
\arguments{
\item{scenarios}{table or vector of \code{EffectScenario} objects}

\item{level}{effect levels in percent (\%), defaults to \code{c(10,50)}}

\item{effect_tolerance}{\code{numeric}, minimum absolute accuracy of effect levels}

\item{factor_cutoff}{optional \code{numeric}, the search for a multiplication factor
will be cut short if  tried factors exceed this value; the result will report
the cutoff value as the final EPx value.}

\item{min_factor}{\code{numeric}, if tried factors fall below this threshold, the algorithm
will halt with an error}

\item{max_factor}{\code{numeric}, if tried factors exceed this threshold, the algorithm
will halt with an error}

\item{verbose}{\code{logic}, if \code{TRUE} then infos about model evaluations are displayed}

\item{ep_only}{\code{logical}, if \code{TRUE} then only EPx values are part of the output,
any contextual information such as \code{EffectScenario} objects are left out}

\item{long_format}{\code{logical}, if \code{TRUE} then EPx values are returned as
a table in long format, any contextual information will be duplicated}

\item{...}{additional arguments passed on to \code{\link[=effect]{effect()}}}
}
\value{
The original \code{tibble} with additional columns named after the request effect levels, e.g. \code{L.EP10.}
If no tibble was used as argument, then a new one is created. The first column \code{scenario} will contain
the supplied \code{EffectScenario} objects.
}
\description{
Derives one or more EPx/LPx values for the supplied effect scenarios, i.e. it
calculates the multiplication factors of an exposure profile that cause
x\% of effect. Scenarios are processed in parallel, if possible.
}
\details{
To estimate EPx values, a \emph{binary search} on multiplication factors is conducted.
The algorithm can achieve arbitrary precision in terms of effects. The
same approach is implemented in the \code{morse} package in the \code{MFx()} function.
Convergence is often achieved in less than 10 iterations per effect level and
endpoint.

Internally, a knowledge base of all tried factors and resulting effect levels is
kept to speed up convergence if more than one endpoint or effect level was
requested. The algorithm will automatically sweep the range of multiplication
factors as needed but hard cutoff values are implemented to avoid infinite loops;
the algorithm will halt with an error message if tried factors are
smaller than \code{1e-30} or greater than \code{1e30}.
\subsection{Numerical precision}{

The precision of reported \emph{EPx} values is controlled by the argument
\code{effect_tolerance} and is given as the upper absolute error threshold of
effects that is deemed acceptable. The default value of \code{0.001} ensures that
a derived \emph{EPx} will result in an effect of x\% ± 0.1. Decreasing the
\code{effect_tolerance} will result in additional model iterations and longer
runtime. Setting an extremely small tolerance value may lead to a breakdown
of the algorithm due to the occurrence of extremely small, quasi-random
numerical errors in simulation results.
}
}
\examples{
minnow_sd \%>\% epx()
minnow_sd \%>\% epx(level=c(10,23,42))

# displays infos about tested multiplication factors
minnow_sd \%>\% epx(verbose=TRUE)

# return results as a table in wide format
minnow_sd \%>\% epx(long_format=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-debtox.R, R/model-deb_daphnia.R
\name{DEBtox}
\alias{DEBtox}
\alias{DEBtox2019}
\alias{DebTox-class}
\alias{DEB_Daphnia}
\alias{DebDaphnia-class}
\title{DEBtox model}
\usage{
DEBtox()

DEB_Daphnia()
}
\value{
an S4 object of type \linkS4class{DebTox}
}
\description{
Creates a  \emph{DEBtox} scenario as described by Jager (2020). It
represents a simplified \emph{DEBtox} model based on \emph{DEBkiss}. In the \emph{BYOM}
application [\href{https://www.debtox.info/byom.html}{link}], this model is
referred to as \emph{DEBtox 2019}, version 4.7.
It supports an optional feature of the \emph{ERA special} model variant, which
can consider a reference \emph{Lm} parameter to compare results of multiple
datasets.
}
\details{
\subsection{State variables}{

The following list describes the names and units of the
model's state variables:
\itemize{
\item \code{D}, scaled damage ([C])
\item \code{L}, body length (mm)
\item \code{R}, cumulative reproduction (-)
\item \code{S}, survival probability (-)
}

State variables \code{D}, \code{L}, and \code{R} are initialized with zero. Variable \code{S}
is initialized with one (\code{1.0}). See \code{\link[=set_init]{set_init()}} on how to set
the initial state manually.
}

\subsection{Parameters}{

The following parameters are required:
\itemize{
\item General
\itemize{
\item \code{L0}, body length at start (mm)
\item \code{Lp}, body length at puberty (mm)
\item \code{Lm}, maximum body length (mm)
\item \code{rB}, von Bertalanffy growth rate constant (1/d)
\item \code{Rm}, maximum reproduction rate (#/d)
\item \code{f}, scaled functional response (-)
\item \code{hb}, background hazard rate (d-1)
\item \code{a}, Weibull background hazard coefficient (-). Set to \code{1} to disable.
}
\item Extra parameters
\itemize{
\item \code{Lf}, body length at half-saturation feeding (mm)
\item \code{Lj}, body length at which acceleration stops (mm)
\item \code{Tlag}, lag time for start development (d)
}
\item TK/TD parameters
\itemize{
\item \code{kd}, dominant rate constant (d-1)
\item \code{zb}, effect threshold energy budget ([C])
\item \code{bb}, effect strength energy-budget effects (1/[C])
\item \code{zs}, effect threshold survival ([C])
\item \code{bs}, effect strength survival (1/([C] d))
}
\item Other parameters (formerly globals in \emph{BYOM})
\itemize{
\item \code{FBV}, dry weight egg as fraction of structural body weight (-)
\item \code{KRV}, part. coeff. repro buffer and structure (kg/kg) (for losses with reproduction)
\item \code{kap}, approximation for kappa (for starvation response)
\item \code{yP}, product of yVA and yAV (for starvation response)
\item \code{Lm_ref}, optional reference max length for scaling rate constants (mm).
Set to zero to disable the reference length. Disabled by default.
\item \code{len}, a switch to control body length dynamics: \code{1} organism can shrink,
\code{2} organism cannot shrink. Default value is \code{1}.
\item \code{Tbp}, optional brood-pouch delay (d). Set to \code{NA} or zero to disable.
Default value is \code{0}.
\item \code{MoA}, mode of action switches (-). Default value is \code{0}.
\item \code{FB}, feedback on damage dynamics switches (-). Default value is \code{0}.
}
}

A reference \code{Lm_ref} is needed to properly compare different data sets,
or when calibrating on more than one data set. If \code{Lm} differs, one would not
want to have different rate constants at the same length.
\subsection{Mode of Action}{

Any combination of the following mode of actions (\emph{MoA}) can be considered by
the model:
\itemize{
\item \code{MoA = 1}: assimilation/feeding
\item \code{MoA = 2}: costs for maintenance
\item \code{MoA = 4}: costs for growth and reproduction
\item \code{MoA = 8}: costs for reproduction
\item \code{MoA = 16}: hazard for reproduction
}

To activate more than one mode of action, simply add up the corresponding
codes and set parameter \code{MoA} to the desired value. To disable all mode of
actions, set parameter \code{MoA} to zero. See also \code{\link[=set_moa]{set_moa()}}.

As an example, to consider effects on feeding and maintenance, set the
mode of action to three (\code{3}):

\code{DEBtox() \%>\% set_param(c(MoA=3))}
}

\subsection{Feedbacks}{

Any combination of the following damage feedbacks can be considered by
the model:
\itemize{
\item \code{1}: surf:vol scaling uptake rate
\item \code{2}: surf:vol scaling elimination rate
\item \code{4}: growth dilution
\item \code{8}: losses with reproduction
}

To activate more than one feedback, simply add up the corresponding
codes. To disable all feedbacks, set the parameter to zero.
}

}

\subsection{Effects}{

The state variables \emph{L} (body length), \emph{R} (cumulative reproduction), and
\emph{S} (survival probability) are set as effect endpoints by default. All state
variables are available as potential endpoints. The list of considered
endpoints can be modified by using \code{\link[=set_endpoints]{set_endpoints()}}.

To calculate effects, each \emph{DEBtox} scenario is simulated twice: One simulation
which considers exposure to a toxicant and one simulation without exposure, i.e.
a control. See also \code{\link[=effect]{effect()}}.
}

\subsection{Simulation output}{

The following intermediary model variables can be added to the model
output on demand. Simply set the optional parameter \code{nout} to the
required output level and pass it to \code{\link[=simulate]{simulate()}}.
\itemize{
\item \code{nout} >= 1: \code{f}, actual scaled response
\item \code{nout} >= 2: \code{fR}, actual f considering starvation
\item \code{nout} >= 3: \code{kd}, actual kd
\item \code{nout} >= 4: \code{s}, stress level
\item \code{nout} >= 5: \code{h}, hazard rate
\item \code{nout} >= 6: \code{sA}, stress factor on assimilation/feeding
\item \code{nout} >= 7: \code{sM}, stress factor on maintenance
\item \code{nout} >= 8: \code{sG}, stress factor on growth costs
\item \code{nout} >= 9: \code{sR}, stress factor on reproduction costs
\item \code{nout} >= 10: \code{sH}, stress factor on hazard to reproduction
\item \code{nout} >= 11: \code{xu}, damage feedback factor for surf:vol scaling uptake rate
\item \code{nout} >= 12: \code{xe}, damage feedback factor for surf:vol scaling elimination rate
\item \code{nout} >= 13: \code{xG}, damage feedback factor for growth dilution
\item \code{nout} >= 14: \code{xR}, damage feedback factor for losses with repro
}
}

\subsection{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{method = 'ode45'}\if{html}{\out{<br>}}
Selects the Dormand-Prince 4(5) method of the Runge-Kutta family, see
\code{\link[deSolve:rkMethod]{deSolve::rkMethod()}} for details.
}
}

\subsection{Model history and changes}{
\itemize{
\item cvasi v1.0.0
\itemize{
\item The \code{DEB_Daphnia()} model implemented BYOM's \emph{DEBtox 2019}
model version 4.5
}
\item cvasi v1.2.0
\itemize{
\item The model equations were updated to conform with BYOM's \emph{DEBtox 2019}
version 4.7. This introduced a new model parameter \code{a}, the Weibull
background hazard coefficient, and limited the maximum hazard rate to
99\% per hour.
\item The scenario constructor was renamed to \code{DEBtox()}.
\item Additional intermediary model variables available as optional simulation
output
}
}
}
}
\section{Functions}{
\itemize{
\item \code{DEB_Daphnia()}: Deprecated model variant of \code{DEBtox()}

}}
\references{
Jager T, 2020: Revisiting simplified DEBtox models for analysing
ecotoxicity data. Ecol Model 416. \doi{10.1016/j.ecolmodel.2019.108904}

Romoli et al., 2024: Environmental risk assessment with energy budget
models: a comparison between two models of different complexity.
Environ Toxicol Chem 43(2):440-449. \doi{10.1002/etc.5795}
}
\seealso{
Other DEB models: 
\code{\link{DEB-models}},
\code{\link{DEB_abj}()}
}
\concept{DEB models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{americamysis}
\alias{americamysis}
\title{A DEB abj scenario of Americamysis bahia}
\format{
An object of class \code{DebAbj} of length 1.
}
\source{
\url{https://www.bio.vu.nl/thb/deb/deblab/add_my_pet/entries_web/Americamysis_bahia/Americamysis_bahia_res.html}
}
\usage{
americamysis
}
\description{
Species parameters were collected from the AddMyPet database entry on
Americamysis bahia (Opossum shrimp). The exposure series consists of a
constant exposure resulting in medium effects on length and reproduction.
}
\seealso{
\code{\link[=DEB_abj]{DEB_abj()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.r
\name{plot_scenario}
\alias{plot_scenario}
\title{Creates a prediction plot for one effect scenario}
\usage{
plot_scenario(model_base, plot_col = 2, trial_number = NULL)
}
\arguments{
\item{model_base}{effect scenario object with mean parameters}

\item{plot_col}{output column which should be plotted, default = 2}

\item{trial_number}{name for model run (if available tag is used)}
}
\value{
plot of the results for one \verb{effect scenario}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
}
\details{
This function has been deprecated and replaced by the generic \code{\link[=plot]{plot()}}.

Sometimes it is helpful if the user can plot results of one effect
scenario. This is for instance the case for test simulations or predictions
for one profile. This function runs the simulation for one effect scenario
and plots the results. Function plots the time (column 1) and the predictions
(column 2, can be changed by the user plot_col)
}
\examples{
# Please use `plot()` instead
metsulfuron \%>\%
  simulate() \%>\%
  plot()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/man-macrophytes.R
\name{Macrophyte-models}
\alias{Macrophyte-models}
\title{Macrophyte models}
\description{
Population models of standard test macrophytes, such as \emph{Lemna spp.}
}
\details{
Available macrophyte models:
\itemize{
\item \link[=Lemna-models]{Lemna}
\item \link[=Magma]{Magma}
}
}
\section{Biomass transfer}{

Models supporting biomass transfer can be instructed to move a fixed amount
of biomass to a new medium after a period of time. This feature replicates
a procedure occurring in e.g. \emph{Lemna} effect studies and may be necessary to
recreate study results.

The biomass transfer feature assumes that always a fixed amount of
biomass is transferred. Transfers can occur at any fixed point in time or
in regular intervals. During a transfer, the biomass is reset to the
transferred amount and additional compartments can be scaled 1:1 accordingly,
to e.g. reflect the change in internal toxicant mass when biomass is modified.
Transfer settings can be modified using \code{\link[=set_transfer]{set_transfer()}}.

If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
}

\seealso{
\link{Scenarios}

Other macrophyte models: 
\code{\link{Lemna_SETAC}()},
\code{\link{Lemna_Schmitt}()},
\code{\link{Magma}}

Other models: 
\code{\link{Algae-models}},
\code{\link{DEB-models}},
\code{\link{GUTS-RED-models}},
\code{\link{Lemna-models}}
}
\concept{macrophyte models}
\concept{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.param_space}
\alias{plot.param_space}
\alias{plot_param_space}
\title{Plot profiled parameter space}
\usage{
\method{plot}{param_space}(x, y, ...)

plot_param_space(x)
}
\arguments{
\item{x}{Return value of \code{\link[=explore_space]{explore_space()}}}

\item{y}{Unused argument}

\item{...}{Unused arguments}
}
\value{
plot object
}
\description{
The function provides bivariate parameter space plots indicating
parameter draws (from the 95\% confidence intervals per parameter obtained
through likelihood profiling) that fall within the inner rim (in green, i.e.
parameter sets which are not significantly different from the original, based
on a chi-square test). The original parameter set is also indicated (in orange),
and, if different from the original set, the best fit parameter set is indicated
(in red).
}
\section{Functions}{
\itemize{
\item \code{plot_param_space()}: Alias of \code{plot.param_space()} for backwards-compatibility.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_morse.R
\name{import_morse}
\alias{import_morse}
\alias{morse}
\title{Import \code{morse} model parameters}
\usage{
import_morse(
  fit,
  find_sd = TRUE,
  find_it = TRUE,
  reset_hb = FALSE,
  params = c("estim", "all"),
  mcmc_size,
  find.SD = deprecated(),
  find.IT = deprecated(),
  reset.hb = deprecated(),
  mcmc.size = deprecated(),
  file = deprecated()
)

morse(...)
}
\arguments{
\item{fit}{Either a string with a file path to an \emph{.Rdata} or \emph{.RDS} file
containing a \emph{morse} fit, or a \emph{morse} fit object itself}

\item{find_sd}{a logical value. If \code{TRUE}, it will try to find fitted parameters of a \emph{GUTS-RED-SD} model}

\item{find_it}{a logical value. If \code{TRUE}, it will try to find fitted parameters of a \emph{GUTS-RED-IT} model}

\item{reset_hb}{a logical value. If \code{TRUE}, the background hazard rate \code{hb} is set to zero}

\item{params}{\code{character}, if set to \code{"estim"} then only the best-fit parameters are
imported, else all parameter sets in the MCM chains are returned}

\item{mcmc_size}{optional \code{integer}, sets the maximum number of imported parameter sets per MCMC.
By default, all MSMS parameter samples are imported.}

\item{find.SD}{\emph{deprecated}\verb{,} alias for parameter \code{find_sd}}

\item{find.IT}{\emph{deprecated}\verb{,} alias for parameter \code{find_it}}

\item{reset.hb}{\emph{deprecated}\verb{,} alias for parameter \code{rest_hb}}

\item{mcmc.size}{\emph{deprecated}\verb{,} alias for parameter \code{mcmc_size}}

\item{file}{\emph{deprecated}\verb{,} alias for parameter \code{fit}}

\item{...}{Arguments passed on to \code{import_morse()}}
}
\value{
list of \link{parameter_set} objects
}
\description{
Loads GUTS model parameters which were fitted by the morse package.
}
\section{Functions}{
\itemize{
\item \code{morse()}: deprecated alias

}}
\examples{
# import all parameter fits
try(import_morse("path/to/morse_fit.RData"))

# import parameters for a specific model
try(import_morse("path/to/morse_fit.RData", find_it=TRUE, find_sd=FALSE))

# modify model objects
try(models \%>\% set_param(import_morse("path/to/morse_fit.RData")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set.R
\name{set_endpoints}
\alias{set_endpoints}
\title{Set effect endpoints}
\usage{
set_endpoints(x, endpoints)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}

\item{endpoints}{\code{character} vector of endpoint names}
}
\value{
Modified \code{EffectScenario} objects
}
\description{
Effect endpoints calculated by functions such as \code{\link[=effect]{effect()}} and \code{\link[=epx]{epx()}}
can be enabled and disabled. If an endpoint is not required for an assessment,
it should be disabled for reasons of computational efficiency. Please refer
to the model description for a list of available endpoints.
}
\examples{
# Only enable reproduction (R) endpoint for americamysis scenario
americamysis \%>\%
  set_endpoints("R") \%>\%
  effect()

# Enable endpoints length (L) and reproduction (R)
americamysis \%>\%
  set_endpoints(c("L","R")) \%>\%
  effect()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-ExposureSeries.R
\name{no_exposure}
\alias{no_exposure}
\title{Zero exposure}
\usage{
no_exposure()
}
\value{
an S4 object of type \linkS4class{ExposureSeries}
}
\description{
Creates an \link{ExposureSeries} with zero concentration. When setting
the zero exposure, pay attention not to accidentally reset the output times
of your scenario as the zero exposure series contains only a single time point.
See the examples.
}
\examples{
# Set exposure to zero, but keep the original output times
minnow_it \%>\%
  set_noexposure() \%>\%
  simulate()
}
\seealso{
\code{\link[=set_noexposure]{set_noexposure()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-ExposureSeries.R
\name{ExposureSeries}
\alias{ExposureSeries}
\alias{ExposureSeries-class}
\alias{NoExposureSeries-class}
\title{Exposure time-series}
\usage{
ExposureSeries(series, dates, file, meta, context)
}
\arguments{
\item{series}{\code{data.frame} with two columns containing a time-series}

\item{dates}{\code{vector}, optional original list of time stamps}

\item{file}{\code{character}, optional file name where data originates from}

\item{meta}{\code{list}, optional metadata}

\item{context}{\code{list} optional contextual metadata such as project ids}
}
\value{
an S4 object of type \linkS4class{ExposureSeries}
}
\description{
Creates an object that encapsulates an exposure time-series with its
metadata, such as formatted datetime strings and file name where the
series was loaded from. \code{\link[=no_exposure]{no_exposure()}} is shorthand to create a time-series
of constant zero exposure.
}
\section{Slots}{

\describe{
\item{\code{dates}}{original time points of time-series, e.g. time stamps of the form \verb{2000-01-01 12:00}}

\item{\code{file}}{\code{character}, file name where data originates from, may be empty}

\item{\code{meta}}{\code{list}, contains metadata}

\item{\code{context}}{\code{list}, contains contextual metadata, such as project ids}

\item{\code{series}}{\code{data.frame} containing the actual time-series}
}}

\seealso{
\code{\link[=no_exposure]{no_exposure()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_tag.R
\name{get_tag}
\alias{get_tag}
\alias{get_tag,ANY-method}
\alias{get_tag,list-method}
\alias{get_tag,EffectScenario-method}
\alias{get_tag,ScenarioSequence-method}
\alias{get_tag,ParameterSet-method}
\title{Get scenario tag}
\usage{
get_tag(x)

\S4method{get_tag}{list}(x)

\S4method{get_tag}{EffectScenario}(x)

\S4method{get_tag}{ScenarioSequence}(x)

\S4method{get_tag}{ParameterSet}(x)
}
\arguments{
\item{x}{(vector of) \link{scenarios} or \link{parameter_set} objects}
}
\value{
(list of) tag(s), returns \code{NA} if no tag was set
}
\description{
Returns the user-defined, custom tag of a scenario, if available.
Tags can be helpful to quickly distinguish scenarios by e.g. a user-specified
string. The function supports vectorized inputs. If more
than one scenario is supplied in argument \code{x}, then a list of tags is
returned.
}
\section{Methods (by class)}{
\itemize{
\item \code{get_tag(list)}: Returns a list of tags (if applicable)

\item \code{get_tag(EffectScenario)}: Returns the tag of a single \link{scenario}

\item \code{get_tag(ScenarioSequence)}: Returns a list of tags, one for each scenario in the sequence

\item \code{get_tag(ParameterSet)}: Returns the tag of a single \link{parameter_set}

}}
\examples{
# returns `fathead minnow`
get_tag(minnow_it)

# update or set a tag
myscenario <- GUTS_RED_IT() \%>\% set_tag("My Custom Tag")
# returns `My Custom Tag`
get_tag(myscenario)
}
\seealso{
\code{\link[=set_tag]{set_tag()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_times.R
\name{get_times}
\alias{get_times}
\alias{get_times,ANY-method}
\alias{get_times,list-method}
\alias{get_times,EffectScenario-method}
\alias{get_times,ScenarioSequence-method}
\title{Get output times}
\usage{
get_times(x)
}
\arguments{
\item{x}{(vector of) \link{scenario} objects}
}
\value{
(list of) times vector
}
\description{
Get output times
}
\examples{
# Create a scenario
myscenario <- GUTS_RED_IT() \%>\% set_times(0:5)
# Returns the defined output times
get_times(myscenario)
}
\seealso{
\code{\link[=set_times]{set_times()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-lemna_setac.R
\name{Lemna_SETAC}
\alias{Lemna_SETAC}
\alias{LemnaSetacScenario-class}
\alias{LemnaSetac-class}
\title{Lemna model v1.3 (Klein et al. 2021)}
\usage{
Lemna_SETAC()
}
\value{
an S4 object of type \linkS4class{LemnaSetac}
}
\description{
The model was described and published by the SETAC Europe Interest Group
Effect Modeling (Klein et al. 2021). It is based on the \emph{Lemna} model
by Schmitt (2013). The model is a mechanistic combined
toxicokinetic-toxicodynamic (TK/TD) and growth model for the aquatic
macrophytes \emph{Lemna spp.}. The model simulates the development of Lemna biomass
under laboratory and environmental conditions. Growth of the Lemna population
is simulated on basis of photosynthesis and respiration rates which are
functions of environmental conditions. The toxicodynamic sub-model describes
the effects of growth-inhibiting substances by a respective reduction in the
photosynthesis rate based on internal concentrations.
}
\section{State variables}{

The model has two state variables:
\itemize{
\item \code{BM}, Biomass (g dw)
\item \code{M_int}, Mass of toxicant in plant population (ng)
}

The  units of the state variables \code{BM} and \code{M_int} are to be interpreted
as \emph{per vessel} in laboratory tests, and as \emph{per square meter} for field populations.
The same interpretation applies to all parameters relating biomass and toxicant mass to
other entities.
}

\section{Model parameters}{

\itemize{
\item Growth model
\itemize{
\item \code{k_photo_fixed}, Model switch for unlimited growth conditions (TRUE/FALSE)
\item \code{k_photo_max}, Maximum photosynthesis rate (d-1)
\item \code{k_loss}, Reference loss rate (d-1)
\item \code{BM_threshold}, Lower biomass abundance threshold,  (g dw)
\item \code{BM_min}, Reservoir for biomass recovery,  (g dw)
}
\item Temperature response of photosynthesis
\itemize{
\item \code{T_opt}, Optimum growth temperature (°C)
\item \code{T_min}, Minimum growth temperature (°C)
\item \code{T_max}, Maximum growth temperature (°C)
}
\item Temperature response of biomass loss rate
\itemize{
\item \code{Q10}, Temperature coefficient (-)
\item \code{T_ref}, Reference temperature for response=1 (°C)
}
\item Irradiance reponse of photosynthesis
\itemize{
\item \code{alpha}, Slope of irradiance response (m2 d kJ-1)
\item \code{beta}, Intercept of irradiance response (-)
}
\item Nutrient response of photosynthesis
\itemize{
\item \code{N_50}, Half-saturation constant of Nitrogen (mg N L-1)
\item \code{P_50}, Half-saturation constant of Phosphorus (mg P L-1)
}
\item Density dependence of photosynthesis
\itemize{
\item \code{BM_L}, Carrying capacity (g dw)
}
\item Concentration response (Toxicodynamics)
\itemize{
\item \code{EC50_int}, Internal concentration resulting in 50\% effect (ug L-1)
\item \code{E_max}, Maximum inhibition (-)
\item \code{b}, Slope parameter (-)
}
\item Internal concentration (Toxicokinetics)
\itemize{
\item \code{P}, Permeability (cm d-1)
\item \code{r_A_DW}, Area per dry-weight ratio (cm2 g-1)
\item \code{r_FW_DW}, Fresh weight per dry weight ratio (-)
\item \code{r_FW_V}, Fresh weight density (g cm-3)
\item \code{r_DW_FN}, Dry weight per frond ratio  (g dw)
\item \code{K_pw}, Partitioning coefficient plant:water (-)
\item \code{k_met}, Metabolisation rate (d-1)
}
}
}

\section{Forcings}{

External concentrations, alias \emph{exposure},  have to be provided as \emph{ug L-1}.

Besides exposure, the model requires four environmental properties as
input:
\itemize{
\item \code{tmp}, ambient temperature (°C)
\item \code{irr}, irradiance (kJ m-2 d-1)
\item \code{P}, Phosphorus concentration (mg P L-1)
\item \code{N}, Nitrogen concentration (mg N L-1)
}

The following constant default values are used for these properties:
\itemize{
\item \code{tmp} = 12 °C
\item \code{irr} = 15,000 kJ m-2 d-1
\item \code{P} = 0.3 mg L-1
\item \code{N} = 0.6 mg L-1
}

Forcings time-series are represented by \code{data.frame} objects consisting of two
columns. The first for time and the second for the environmental factor in question.

Entries of the \code{data.frame} need to be ordered chronologically. A time-series
can consist of only a single row; in this case it will represent constant
environmental conditions. See \link{scenarios} for more details.
}

\section{Effects}{

Supported effect endpoints include \emph{BM} (biomass) and \emph{r} (average
growth rate during simulation). The effect on biomass is calculated from
the last state of a simulation. Be aware that endpoint \emph{r} is incompatible
with biomass transfers.
}

\section{Simulation output}{

For reasons of convenience, the return value of \code{\link[=simulate]{simulate()}} contains by
default two additional variables derived from simulation results:
the internal concentration \code{C_int} as well as the number of fronds \code{FrondNo}.
These can be disabled by setting the argument \code{nout = 0}.

The available output levels are as follows:
\itemize{
\item \code{nout >= 1}: \code{C_int}, internal concentration (mass per volume)
\item \code{nout >= 2}: \code{FrondNo}, frond number (-)
\item Response functions
\itemize{
\item \code{nout >= 3}: \code{f_loss}, respiration dependency function (-)
\item \code{nout >= 4}: \code{f_photo}, photosynthesis dependency function (-)
\item \code{nout >= 5}: \code{fT_photo}, temperature response of photosynthesis (-)
\item \code{nout >= 6}: \code{fI_photo}, irradiance response of photosynthesis (-)
\item \code{nout >= 7}: \code{fP_photo}, phosphorus response of photosynthesis (-)
\item \code{nout >= 8}: \code{fN_photo}, nitrogen response of photosynthesis (-)
\item \code{nout >= 9}: \code{fBM_photo}, density response of photosynthesis (-)
\item \code{nout >= 10}: \code{fCint_photo}, concentration response of photosynthesis (-)
}
\item Environmental variables
\itemize{
\item \code{nout >= 11}: \code{C_int_unb}, unbound internal concentration (mass per volume)
\item \code{nout >= 12}: \code{C_ext}, external concentration (mass per volume)
\item \code{nout >= 13}: \code{Tmp}, temperature (deg C)
\item \code{nout >= 14}: \code{Irr}, irradiance (kJ m-2 d-1)
\item \code{nout >= 15}: \code{Phs}, Phosphorus concentration (mg P L-1)
\item \code{nout >= 16}: \code{Ntr}, Nitrogen concentration (mg N L-1)
}
\item Derivatives
\itemize{
\item \code{nout >= 17}: \code{dBM}, biomass derivative (g dw m-2 d-1)
\item \code{nout >= 18}: \code{dM_int}, mass of toxicant in plants derivative (mass per m2 d-1)
}
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{hmax = 0.1}\if{html}{\out{<br>}}
Maximum step length in time suitable for most simulations.
}
}

\section{Model history and changes}{

\itemize{
\item cvasi v1.0.0
\itemize{
\item Model and parameters as described by Klein et al. (2022) report version 1.1
}
\item cvasi v1.5.0
\itemize{
\item Default value of parameter \code{beta} modified due to typo in previous report
versions, now conforms with Klein et al. (2025), report version 1.2
\item New value: \code{beta=0.25}, old value: \code{beta=0.025}
}
}
}

\section{Biomass transfer}{

Models supporting biomass transfer can be instructed to move a fixed amount
of biomass to a new medium after a period of time. This feature replicates
a procedure occurring in e.g. \emph{Lemna} effect studies and may be necessary to
recreate study results.

The biomass transfer feature assumes that always a fixed amount of
biomass is transferred. Transfers can occur at any fixed point in time or
in regular intervals. During a transfer, the biomass is reset to the
transferred amount and additional compartments can be scaled 1:1 accordingly,
to e.g. reflect the change in internal toxicant mass when biomass is modified.
Transfer settings can be modified using \code{\link[=set_transfer]{set_transfer()}}.

If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
}

\references{
Klein J., Cedergreen N., Heine S., Reichenberger S., Rendal C.,
Schmitt W., Hommen U., 2021: \emph{Refined description of the Lemna TKTD growth model
based on Schmitt et al. (2013) - equation system and default parameters}.
Report of the working group \emph{Lemna} of the SETAC Europe Interest Group Effect
Modeling. Version 1.3, uploaded in July 2025.
https://www.setac.org/group/effect-modeling.html

Schmitt W., Bruns E., Dollinger M., and Sowig P., 2013:
\emph{Mechanistic TK/TD-model simulating the effect of growth inhibitors on
Lemna populations}. Ecol Model 255, pp. 1-10. \doi{10.1016/j.ecolmodel.2013.01.017}
}
\seealso{
\link{Lemna-models}, \link{Macrophyte-models}, \link{Transferable}, \link{Scenarios}

Other Lemna models: 
\code{\link{Lemna-models}},
\code{\link{Lemna_Schmitt}()}

Other macrophyte models: 
\code{\link{Lemna_Schmitt}()},
\code{\link{Macrophyte-models}},
\code{\link{Magma}}
}
\concept{Lemna models}
\concept{macrophyte models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_param.R
\name{set_param}
\alias{set_param}
\alias{set_param,EffectScenario,vector-method}
\alias{set_param,EffectScenario,ParameterSet-method}
\alias{set_param,list,ParameterSet-method}
\alias{set_param,list,vector-method}
\alias{set_param,ScenarioSequence,ANY-method}
\alias{set_param,CalibrationSet,ANY-method}
\title{Set scenario parameters}
\usage{
set_param(x, param)

\S4method{set_param}{EffectScenario,vector}(x, param)

\S4method{set_param}{EffectScenario,ParameterSet}(x, param)

\S4method{set_param}{list,ParameterSet}(x, param)

\S4method{set_param}{list,vector}(x, param)

\S4method{set_param}{ScenarioSequence,ANY}(x, param)

\S4method{set_param}{CalibrationSet,ANY}(x, param)
}
\arguments{
\item{x}{object(s) to modify}

\item{param}{named numeric \code{vector} with parameter names and value OR a
list of \linkS4class{parameter_set} objects}
}
\value{
Vector of modified objects
}
\description{
Modifies the parameters of one or more \link{scenario} objects.
}
\details{
Most parameters are represented by numerical types but other types
are possible depending on model. Please refer to the model description
which parameters are required and in which unit. Some or all parameters may
be required to start a simulation. If required parameters are missing,
simulation will fail with an error message.
}
\examples{
Lemna_Schmitt() \%>\% set_param(c(Emax=1,EC50=0.12))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dmagna}
\alias{dmagna}
\title{A DEBtox scenario of Daphnia magna}
\format{
An object of class \code{DebTox} of length 1.
}
\usage{
dmagna
}
\description{
Species and substance parameters were collected from test runs of the
original \link{DEBtox} Daphnia model.
}
\seealso{
\code{\link[=DEBtox]{DEBtox()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_times.R
\name{set_times}
\alias{set_times}
\alias{set_times,ANY-method}
\alias{set_times,list-method}
\alias{set_times,EffectScenario-method}
\alias{set_times,ScenarioSequence-method}
\title{Set output times}
\usage{
set_times(x, times)
}
\arguments{
\item{x}{vector of \link{scenarios}}

\item{times}{\code{numerical} vector}
}
\value{
Vector of modified \link{scenarios}
}
\description{
Minimum and maximum output times define the simulated period for a scenario.
Simulation results will be returned for each output time, see \code{\link[=simulate]{simulate()}}.
}
\details{
Be aware that output times may be modified by \code{\link[=set_exposure]{set_exposure()}}. Precision of
simulation results may be influenced by chosen output times, see \code{\link[=simulate]{simulate()}}
for more information.
}
\examples{
# Set simulated period to [2,4] with output intervals of length 1
minnow_it \%>\% set_times(c(2,3,4))

# Decrease output interval length to 0.1
minnow_it \%>\% set_times(seq(2, 4, 0.1))
}
\seealso{
\code{\link[=simulate]{simulate()}},  \code{\link[=get_times]{get_times()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-Transferable.R
\name{Transferable}
\alias{Transferable}
\alias{Biomass-transfer}
\alias{Transferable-class}
\title{Biomass transfer class}
\description{
By inheriting from class \code{Transferable}, a scenario's behavior can be
extended to support transfer and reset of biomass at dedicated points during
simulation.
}
\section{Slots}{

\describe{
\item{\code{transfer.times}}{\code{numeric}, vector of custom time points at which transfers occur,
e.g. \code{c(2,5,14)}}

\item{\code{transfer.interval}}{\code{numeric}, length of regular interval until biomass transfer to new
medium, regular transfers always occur relative to time point zero}

\item{\code{transfer.biomass}}{\code{numeric}, amount of biomass transferred to new medium}

\item{\code{transfer.comp.biomass}}{\code{character} state variable which describes
biomass}

\item{\code{transfer.comp.scaled}}{\code{character} vector of state variable which will
be scaled 1:1 when biomass is modified, e.g. internal toxicant mass}
}}

\section{Biomass transfer}{

Models supporting biomass transfer can be instructed to move a fixed amount
of biomass to a new medium after a period of time. This feature replicates
a procedure occurring in e.g. \emph{Lemna} effect studies and may be necessary to
recreate study results.

The biomass transfer feature assumes that always a fixed amount of
biomass is transferred. Transfers can occur at any fixed point in time or
in regular intervals. During a transfer, the biomass is reset to the
transferred amount and additional compartments can be scaled 1:1 accordingly,
to e.g. reflect the change in internal toxicant mass when biomass is modified.
Transfer settings can be modified using \code{\link[=set_transfer]{set_transfer()}}.

If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
}

\examples{
# Simulation without biomass transfers
metsulfuron \%>\%
  set_noexposure() \%>\%
  set_notransfer() \%>\%
  simulate()

# With biomass transfer every 7 days, biomass is reset to 50 *g/m²* on transfer
metsulfuron \%>\%
  set_noexposure() \%>\%
  set_transfer(interval=7, biomass=50) \%>\%
  simulate()
}
\seealso{
\code{\link[=set_transfer]{set_transfer()}}

Other scenario-related: 
\code{\link{Scenarios}}
}
\concept{scenario-related}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-guts.R
\name{GUTS_SD}
\alias{GUTS_SD}
\alias{Guts-class}
\alias{GutsSd-class}
\alias{GUTS-SD}
\title{GUTS-SD scenario}
\usage{
GUTS_SD(scaled_ci = FALSE, dose_metric = c("D", "Ci", "Cw"))
}
\arguments{
\item{scaled_ci}{\code{logical}, switch to enable scaling. If \code{TRUE}, the model
will use the scaled internal concentration \verb{Ci*}. Else no scaling. Default
value is \code{FALSE}.}

\item{dose_metric}{\code{character}, selects the dose metric. \code{'D'} to use damage, \code{'Ci'} to
for internal concentration, or \code{'Cw'} for the external concentration. Default
value is \code{'D'}.}
}
\value{
an S4 object of type \code{GutsSd-class}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}\if{html}{\out{<br>}}
Full \emph{General Unified Threshold models of Survival} (GUTS) with stochastic
death (\emph{SD}). The model was defined by Jager et al. (2011). It is
compatible with the \emph{Full GUTS} model as described by EFSA (2018), but
some parameter names may differ.
}
\section{State variables}{

The following list describes the default names and standard units of \emph{GUTS}
state variables:
\itemize{
\item \code{Ci}, (scaled) internal concentration (conc)
\item \code{D}, (scaled) damage (*)
\item \code{H}, cumulative hazard (-)
}

The state variables are initialized with zero by default.
}

\section{SD model parameters}{

The set of parameters and their names follows the definition by Jager et al
(2011). The actual number of required parameters depends on the selected model
variant, i.e. if the internal concentration is scaled or not, as well as the
selected dose metric. The full set of parameters is as follows
\itemize{
\item \code{ki}, accumulation rate into body (time^-1)
\item \code{ke}, elimination rate (time^-1)
\item \code{Kiw}, scaling constant for external concentration (*)
\item \code{kr}, damage recovery rate (time^-1)
\item \code{kk}, killing rate constant (time^-1)
\item \code{hb}, background hazard rate (time^-1)
\item \code{z}, threshold for effects (*)
}
}

\section{Effects}{

The effect endpoint \code{L} (lethality) is available for \emph{GUTS} models.
A value of zero (\code{0.0}) denotes \emph{no effect} on organism survival. A value of
one (\code{1.0}) denotes a lethality rate of 100\%, i.e. no survivors.

The survival probability \code{S} is available in the return value of \code{\link[=simulate]{simulate()}}.
}

\references{
EFSA PPR Panel (EFSA Panel on Plant Protection Products and their Residues),
Ockleford C, Adriaanse P, Berny P, et al., 2018: \emph{Scientific Opinion on the
state of the art of Toxicokinetic/Toxicodynamic (TKTD) effect models for
regulatory risk assessment of pesticides for aquatic organisms}. EFSA Journal 2018;
16(8):5377, 188 pp. \doi{10.2903/j.efsa.2018.5377}

Jager T., Albert C., Preuss T.G., and Ashauer R., 2011: General Unified Threshold
Model of Survival - a Toxicokinetic-Toxicodynamic Framework for Ecotoxicology.
Environ. Sci. Technol. 45(7), pp. 2529-2540. \doi{10.1021/es103092a}
}
\seealso{
\link{GUTS-RED-SD}

Other GUTS models: 
\code{\link{GUTS_IT}()}
}
\concept{GUTS models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tox_data.R
\name{td2cs}
\alias{td2cs}
\title{Create \emph{calisets} from \emph{tox_data}}
\usage{
td2cs(scenario, data, output_var = NULL)
}
\arguments{
\item{scenario}{a base \link{scenario}}

\item{data}{return value of \code{\link[=tox_data]{tox_data()}}}

\item{output_var}{optional \emph{character}, rename observed data column to this value}
}
\value{
list of \link{caliset}s
}
\description{
This is a convenience function which eases the creation of \link{caliset}s
from a base scenario and a \link{tox_data} object. The scenario
will be used as is. If exposure series are defined by the \code{tox_data}
object, then these will be assigned to the scenario(s) accordingly.
}
\examples{
# Import trial data from Schmitt et al. (2013), including exposure
mydata <- tox_data(schmitt2013)

# Example trial contained in data set: trial 'T0.32'
mydata@data[["T0.32"]]        # observed quantities
mydata@exposure[["T0.32"]]    # associated exposure series

# Create list of calisets from full data set
lst <- td2cs(Lemna_SETAC(), mydata)

# Example caliset representing conditions of trial 'T0.32'
lst[[2]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_window.R
\name{set_window}
\alias{set_window}
\alias{set_nowindow}
\title{Set window length}
\usage{
set_window(x, length, interval)

set_nowindow(x)
}
\arguments{
\item{x}{vector of \code{EffectScenario} objects}

\item{length}{\code{numeric}, length of exposure window to consider for effect
calculation, set \code{length=-1} to disable moving windows}

\item{interval}{\code{numeric}, interval between considered exposure windows}
}
\value{
modified \code{EffectScenario} objects
}
\description{
Exposure windows are defined as a period of time at the scale of the exposure series.
As an example: if an exposure series has an hourly time step, a window length
of \code{24} will consider the exposure within 24 hours intervals for effect
calculation. The same applies for the window interval, i.e. the period between
considered exposure windows. Set \code{length=-1} to disable moving windows.
}
\section{Functions}{
\itemize{
\item \code{set_nowindow()}: Disable moving windows

}}
\examples{
# Calculate the maximum effect for all windows of 10 days length
metsulfuron \%>\%
  set_window(length=10, interval=1) \%>\%
  effect()

# Disable moving exposure windows
metsulfuron \%>\%
  set_nowindow() \%>\%
  effect()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-magma.R
\name{Magma}
\alias{Magma}
\alias{Myrio}
\alias{Myrio_log}
\title{Magma model (Witt et al., submitted)}
\usage{
Magma(growth = c("exp", "log"))

Myrio()

Myrio_log()
}
\arguments{
\item{growth}{\code{character}, growth model to simulate: \code{"exp"} for exponential
growth or \code{"log"} for logistic growth. Default is \code{"exp"}.}
}
\value{
an S4 object of type \code{Magma}
}
\description{
The \emph{Magma} model interprets the Tier 2C version of the \emph{Lemna} model by
\link[=Lemna_SETAC]{Klein et al. (2021)}, as a generic macrophyte model.
It is mathematically equivalent to the Tier 2C version of the model
by Klein et al. (2021) with the recommended Tier 2C settings \code{k_photo_fixed=TRUE}
and \code{k_resp=0}.
}
\details{
In particular, the growth model is a simple exponential growth model,
which is considered to be the typical situation for a laboratory macrophyte
study. Instead of frond numbers as for Lemna, the biomass is also returned as
total shoot length (\emph{TSL}) in simulation results. Consequently, the model has
the additional parameter \code{r_DW_TSL} (dry weight per total shoot length ratio)
instead of \code{r_DW_FN} (dry weight per frond number ratio). A model variant
with an option for logistic growth is provided as well.
}
\section{State variables}{

The model has two state variables:
\itemize{
\item \code{BM}, Biomass (g dw)
\item \code{M_int}, Mass of toxicant in plant population (ng)
}
}

\section{Model parameters}{

The growth model can either simulate exponential growth (the default) or
logistic growth. For logistic growth, an additional parameter \code{D_L} describing
the limit density or carrying capacity needs to be provided.
\itemize{
\item Growth model
\itemize{
\item \code{mu_control}, Maximum photosynthesis rate (d-1), default: \code{0.47}
\item (optional) \code{D_L}, Limit density (g dw)
}
\item Concentration response (Toxicodynamics)
\itemize{
\item \code{EC50_int}, Internal concentration resulting in 50\% effect (ug L-1)
\item \code{E_max}, Maximum inhibition (-), default: \code{1}
\item \code{b}, Slope parameter (-)
}
\item Internal concentration (Toxicokinetics)
\itemize{
\item \code{P}, Permeability (cm d-1)
\item \code{r_A_DW}, Area per dry-weight ratio (cm2 g-1), default: \code{1000}
\item \code{r_FW_DW}, Fresh weight per dry weight ratio (-), default: \code{16.7}
\item \code{r_FW_V}, Fresh weight density (g cm-3), default: \code{1}
\item \code{r_DW_TSL}, Dry weight per total shoot length ratio  (g dw cm-1)
\item \code{K_pw}, Partitioning coefficient plant:water (-), default: \code{1}
\item \code{k_met}, Metabolisation rate (d-1), default: \code{0}
}
}
}

\section{Environmental factors}{


None.
}

\section{Parameter boundaries}{

Default values for parameter boundaries are set for all parameters by expert
judgement, for calibration purposes. Values can be modified using \code{\link[=set_bounds]{set_bounds()}}.
}

\section{Simulation output}{

Simulation results will contain the state variables biomass (\code{BM}) and
mass of internal toxicant (\code{M_int}).

It is possible to amend the output of \code{\link[=simulate]{simulate()}} with additional model
quantities that are not state variables, for e.g. debugging purposes or to
analyze model behavior. To enable or disable additional outputs, use the
optional argument \code{nout} of \code{\link[=simulate]{simulate()}}. As an example, set \code{nout=2} to
enable reporting of total shoot length (\code{TSL}) and internal concentration
(\code{C_int}). Set \code{nout=0} to disable additional outputs. The default is \code{nout=1}.

The available output levels are as follows:
\itemize{
\item \code{nout} >= 1: \code{TSL}, total shoot length (cm)
\item \code{nout} >= 2: \code{C_int}, internal concentration (ug L-1)
\item \code{nout} >= 3: \code{f_photo}, photosynthesis dependency function (-)
\item \code{nout} >= 4: \code{C_int_unb}, unbound internal concentration (ug L-1)
\item \code{nout} >= 5: \code{C_ext}, external concentration (ug L-1)
\item \code{nout} >= 6: \code{dBM}, biomass derivative (g dw d-1)
\item \code{nout} >= 7: \code{dM_int}, mass of toxicant in plants derivative (ng d-1)
}
}

\section{Solver settings}{

The arguments to ODE solver \code{\link[deSolve:ode]{deSolve::ode()}} control how model equations
are numerically integrated. The settings influence stability of the numerical
integration scheme as well as numerical precision of model outputs. Generally, the
default settings as defined by \emph{deSolve} are used, but all \emph{deSolve} settings
can be modified in \emph{cvasi} workflows by the user, if needed. Please refer
to e.g. \code{\link[=simulate]{simulate()}} on how to pass arguments to \emph{deSolve} in \emph{cvasi}
workflows.

Some default settings of \emph{deSolve} were adapted for this model by expert
judgement to enable precise, but also computationally efficient, simulations
for most model parameters. These settings can be modified by the user,
if needed:
\itemize{
\item \code{hmax = 0.1}\if{html}{\out{<br>}}
Maximum step length in time suitable for most simulations.
}
}

\section{Effects}{

Supported effect endpoints include \emph{BM} (biomass) and \emph{r} (average
growth rate during simulation). The effect on biomass is calculated from
the last state of a simulation. Be aware that endpoint \emph{r} is incompatible
with biomass transfers.
}

\section{Biomass transfer}{

Models supporting biomass transfer can be instructed to move a fixed amount
of biomass to a new medium after a period of time. This feature replicates
a procedure occurring in e.g. \emph{Lemna} effect studies and may be necessary to
recreate study results.

The biomass transfer feature assumes that always a fixed amount of
biomass is transferred. Transfers can occur at any fixed point in time or
in regular intervals. During a transfer, the biomass is reset to the
transferred amount and additional compartments can be scaled 1:1 accordingly,
to e.g. reflect the change in internal toxicant mass when biomass is modified.
Transfer settings can be modified using \code{\link[=set_transfer]{set_transfer()}}.

If a transfer occurs, simulation results of that time point will report the model state
\strong{before} the transfer. Be aware that if transfers are defined using the
\code{interval} argument, the transfers will always occur relative to time point
zero (\code{t = 0}). As an example, setting a regular transfer of seven days,
\code{interval =  7}, will result at transfers occurring at time points which are
integer multiplicates of seven, such as \code{t=0}, \code{t=7}, \code{t=14} and so forth.
The starting and end times of a scenario do not influece \strong{when} a regular
transfer occurs, only \strong{if} it occurs.
}

\references{
Witt et al., submitted

Klein J., Cedergreen N., Heine S., Reichenberger S., Rendal C.,
Schmitt W., Hommen U., 2021: \emph{Refined description of the Lemna TKTD growth model
based on Schmitt et al. (2013) - equation system and default parameters}.
Report of the working group \emph{Lemna} of the SETAC Europe Interest Group Effect
Modeling. Version 1.1, uploaded on 09 May 2022.
https://www.setac.org/group/effect-modeling.html
}
\seealso{
\link{Macrophyte-models}, \link{Lemna-models}, \link{Transferable}, \link{Scenarios}

Other macrophyte models: 
\code{\link{Lemna_SETAC}()},
\code{\link{Lemna_Schmitt}()},
\code{\link{Macrophyte-models}}
}
\concept{macrophyte models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence.R
\name{sequence}
\alias{sequence}
\alias{ScenarioSequence-class}
\title{Sequence of scenarios}
\usage{
sequence(seq, breaks = NULL)
}
\arguments{
\item{seq}{list of \link{scenario} objects}

\item{breaks}{optional vector of \emph{numerics}, scenarios' output times will
be modified so that one scenario ends at the break and the next one begins}
}
\value{
an S4 object of type \linkS4class{ScenarioSequence}
}
\description{
Scenario sequences can be used to implement changes in model parameters
over time, which otherwise would remain constant for the duration of a simulation.
A sequence of scenarios is treated as a single scenario and each scenario
is simulated one after the other. If scenario \code{n} in a sequence was simulated,
scenario \code{n+1} will start off in the model state where \code{n} had ended.
}
\details{
Sequences are generally treated the same as scenarios: Sequences can be
simulated, as well as effects and EPx can be derived.
\subsection{Requirements}{

All scenarios in a sequence must fulfill the following requirements:
\itemize{
\item All scenarios must have identical state variables
\item The \emph{output times} of all scenarios must represent a continuous time series
without gaps or overlaps
}

Using the \code{breaks} argument, the function can split up the scenarios'
output times at the given break points. The break points must be within
the interval defined by the superset of all output times in the sequence.
}
}
\examples{
# Create a scenario with background mortality only
scen1 <- minnow_it \%>\%
  set_noexposure() \%>\%
  set_times(0:10)
# Modify a scenario parameter, e.g. set background mortality to zero
scen2 <- scen1 \%>\% set_param(c(hb=0))

# Create a sequence of scenarios, scenario #1 will be simulated for the
# time period [0, 4], and #2 for [4, 10]
sq <- sequence(list(scen1, scen2), breaks=c(4))

# Simulate the sequence: the mortality stops after t=4.0, due to scenario #2
# being simulated after t=4.0, which disabled the background mortality
simulate(sq)
# Effect endpoints can also be calculated
effect(sq)

}
\seealso{
\link{sequence.extract}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/epx.R
\name{epx_mtw}
\alias{epx_mtw}
\title{Calculate EPx values for a series of moving time window}
\usage{
epx_mtw(
  x,
  level = c(10, 50),
  factor_cutoff = 1000,
  window_length = 7,
  window_interval = 1,
  ...
)
}
\arguments{
\item{x}{a \link{scenario}}

\item{level}{The target effect level of the effect, ie. the x of EPx.}

\item{factor_cutoff}{above which cutoff is the EPx is not relevant}

\item{window_length}{the length of the moving time window}

\item{window_interval}{the interval that the moving time window moves}

\item{...}{arguments passed to \code{epx}}
}
\value{
a tibble with five columns
\itemize{
\item window.start
\item window.end
\item endpoint
\item level
\item EPx
}
}
\description{
Calls \code{epx()} to calculate the EPx value (i.e. the multiplication factors of
an exposure profile that cause x\% of effect) for moving windows with length
\code{window_length} that move timesteps defined by \code{window_interval}.
}
\examples{
metsulfuron \%>\%
  set_window(length=7, interval=1) \%>\%
  epx_mtw()
}
